PROGRAM MAIN
! 
! This is the main program in an implementation of a sigma-coordinate
! ocean model called "BOM", version 5.
! 
! Build a fresh BOM with the commands  
!    make clean
!    make mod
!    make
!
! This must typically be done when you make changes in the modules
! or move the code to another platform. When only minor changes are
! done to the code, a single "make" may do.  
!
  USE STATE
  USE DIVINTERFACES
  USE MOD_BOUND
  USE CONTROL_PARAMETERS
  USE SETUP2
  USE TIME_UTIL
  USE MOD_CONCENTRATION
  USE MOD_FORCING, ONLY : GET_AND_APPLY_FORCING
  USE MOD_RAMPUP
  USE CHECKPOINTING
  USE PGPLOT
  USE MOD_PARTICLE
  USE GRID_UTIL, ONLY : NUMGRIDS, GRID, MYGID
  USE ETAIMPLICIT
  
  IMPLICIT NONE
  INTEGER(8) :: T1=0, T2=0, TOTTIME=0 
  INTEGER i, tid(8)
  REAL(8) STEPTIME
  
  
! first, initialize the MPI sub system, if present. link in
! qmpi_dummy.f90 if the code is compiled without mpi support. If the
! file "multibom.dat" is present, it will be parsed and specifies init
! file names for each BOM grid, which may be processor dependent.
  
  CALL INITIALIZE_PARALLELIZATION() 

  CALL DATE_AND_TIME(VALUES=TID)
  IF(MASTER) WRITE(*,'(a,i4.4,"-",i2.2,"-",i2.2,"T",i2.2,":",i2.2,":",i2.2,/)') "BOM started ", tid(1:3),tid(5:7)
  
! On each processor, read the case description file. If a command line
! argument is present, use it as the config file name instead of the
! default "setupfile2.dat".

  IF(NUMGRIDS == 1 .AND. COMMAND_ARGUMENT_COUNT() == 1)THEN
      CALL GET_COMMAND_ARGUMENT(1, GRID(1)%INIT_FILE)
  END IF

  CALL READ_INIT_FILE(GRID(MYGID)%INIT_FILE)
  
! open all datafiles that are kept open during a run in this routine

  CALL FILES()
  
! Compute HU, HV, FSM, DUM, DVM, COR. (a stripped down dypgrin)

  CALL INITIALIZE_GRID()
! 
! initialize all prognostic fields. (a stripped down fieldinit)
!
  CALL INITIALIZE_FIELDS()
!
!  Call NCALFA to compute the relaxation parameters ALPHA and ALPHAE
!  to be used in the FRS-zones at the open boundaries for the 3-D
!  variables and 2-D variables respectively.
!
  CALL NCALFA()

! initialize the lagrangian particle model if the file particle_setup.dat is present
  CALL SETUP_PARTICLE()
  
! Override the new setup system or add initialization not present in
! setup2 here. If desired this is a good point to call the old dypgrin
! or fieldinit routines.
  
  CALL USER_INIT()

! If desired, plot the initial fields
  CALL OUTPUT()
  
! initialization mostly done. time to find bugs in the setup...
  CALL SANITY_CHECK()

  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
! START TIMESTEPPING
  
  DO ISTEP = FIRST_STEP, NUMSTEP     
    T1 = RDTSC()

! increment model time by DT seconds, NDATE is an integer clock, TIME double precision.
    CALL TICTAC(REAL(DT,8), NDATE, TIME)
    
    IF(MASTER) WRITE (*,'(/,"STEP ",i8.8," ",i4.4,"-",i2.2,"-",i2.2,"T",i2.2,":",i2.2,":",i2.2," t: ",f0.1)') &
         ISTEP, NDATE, TIME

! some routines, e.g. timeseries need to know if we will do a checkpoint this timestep.
    DO_CHECKPOINT = ARE_WE_GOING_TO_CHECKPOINT()

! run time control of a limited, safe, subset of options
    IF(MOD(ISTEP, RUNTIME_INTERVAL) == 0) CALL RUNTIME_CONTROL()
!
! activate tracer concentration
!
    CALL ACTIVATE_CONCENTRATION()
    
! set the forcing arrays (xxx_bound) as specified in the setup file.
    CALL GET_AND_APPLY_FORCING()
    CALL APPLY_RAMPUPS()
!
    IF(TEST_FORCING_MODE) GOTO 2
!    
! Atmospheric part
!
    IF (WIND_FORCING) THEN
        CALL WINDFORCING
    ENDIF
    
    IF (ATMPRESS) THEN
        CALL ATMOSP
        ETA_TIDE = ETA_TIDE + ETABFRS
    ENDIF
!
!  Update the density field.
!
    IF(.NOT.ADVECT_DENSITY)THEN
        SELECT CASE(EOS)
        CASE("wang")
            CALL DENS
        CASE("unesco")
            CALL DENSUNESCO
        case("user")
            CALL USER_EOS
        END SELECT
    END IF
    
     CALL TAG_GHOST(D)
     CALL TAG_GHOST(DU)
     CALL TAG_GHOST(DV)      
!     CALL TAG_GHOST(IDU)
!     CALL TAG_GHOST(IDV)
     CALL TAG_GHOST(WUSURF,EAST,1)
     CALL TAG_GHOST(WVSURF,NORTH,1)
     CALL TAG_GHOST(WUBOT,EAST,1)
     CALL TAG_GHOST(WVBOT,NORTH,1)
     CALL TAG_GHOST(U)
     CALL TAG_GHOST(V)
     CALL TAG_GHOST(RHO)
     CALL UPDATE_GHOST
 
!
!  Compute vertical diffusivities and viscositities
!
     SELECT CASE(VERTICAL_CLOSURE)
     CASE(0)
     CASE(1)
         CALL MY2HALV
         
         IF (ADVTURB) THEN
             CALL UPSTREAMQ(Q2)
             CALL UPSTREAMQ(Q2L)
         ENDIF

         CALL BOUND2(6)
!    CASE(X)
!       CALL ...
     CASE DEFAULT
         CALL ABORT("Unknown vertical closure")
     END SELECT
    
!
!  Store old values.
!
     ETAP = ETA

! alternative timestepping with implicit solver for ETA
     IF(IMPLICIT_ETA) THEN
        CALL IMAIN()
        GOTO 1
     END IF

!
!  ver 4.0 initialization
!
     UNM1 = UN
     UN = U
     DELUNM1 = DELUN
     DELUN = 0.
     DELUSTAR = 0.
!
     VNM1 = VN
     VN = V
     DELVNM1 = DELVN
     DELVN = 0.
     DELVSTAR = 0.
!
!  In 5.0
!
     WNM1 = WN
     WN = W 
     DELWNM1 = DELWN
     DELWN = 0.
     DELWSTAR = 0.
!
     IF(ADVECT_DENSITY)THEN
         RHONM1 = RHON
         RHON = RHO
         DELRHONM1 = DELRHON
         DELRHON = 0.
         DELRHOSTAR = 0.
     END IF
     IF(ADVECT_S)THEN
         SNM1 = SN
         SN = S
         DELSNM1 = DELSN
         DELSN = 0.
         DELSSTAR = 0.
     END IF
     IF(ADVECT_T)THEN
         TNM1 = TN
         TN = T
         DELTNM1 = DELTN
         DELTN = 0.
         DELTSTAR = 0.
     END IF
!
     DO i=1,MAXNUMCONC
         IF(CO(I)%ACTIVE)THEN
             CO(I)%CONCNM1 = CO(I)%CONCN
             CO(I)%CONCN = CO(I)%CONC
             CO(I)%DELCONCNM1 = CO(I)%DELCONCN
             CO(I)%DELCONCN = 0.
             CO(I)%DELCONCSTAR = 0.
         END IF
     END DO
!
!  Take an explicit predictor step in ETA, U, V, S and T
!  and store coorrections DELETA, DELUE, DELVE, DELSE and DELTE.
!
!
!  Compute contributions from the internal pressure from old values of RHO
!
     SELECT CASE(INTERNAL_PRESSURE)
     CASE("fourth")
         CALL INTERNAL4
     CASE("pom")
         CALL INTERNAL
     CASE("balanced")
         CALL INTERNALOPT()
     END SELECT
!
!  Compute contributions from the advective terms in the momentum equations
!
     CALL SUPERBEEUV
!
!  In MODESPLIT the U and V fields are split in 2-D external parts and 3-D internal parts
!  and propagated in time
!
     CALL MODESPLIT(1)
     
     CALL DIAGNOSE(1)
     
     CALL BOUND2(3)
       
     CALL TAG_GHOST(ETA, WEST+SOUTH, 1)
     CALL TAG_GHOST(U)
     CALL TAG_GHOST(V)
     CALL TAG_GHOST(WADV, WIDTH=1)
     CALL TAG_GHOST(W, WIDTH=1)
     CALL TAG_GHOST(UADV)
     CALL TAG_GHOST(VADV)
     CALL TAG_GHOST(KM, WEST+SOUTH, 1)
     CALL UPDATE_GHOST
     
     CALL UPDATEDD
     
     CALL TAG_GHOST(D)
     CALL TAG_GHOST(DU)
     CALL TAG_GHOST(DV)
     CALL TAG_GHOST(DUDZ,EAST+WEST,1)
     CALL TAG_GHOST(DVDZ,NORTH+SOUTH, 1)
     IF(ADVECT_DENSITY)THEN
         CALL TAG_GHOST(RHO)
     ELSE
         CALL TAG_GHOST(S)
         CALL TAG_GHOST(T)
     END IF
     CALL UPDATE_GHOST

     IF(NUMCONC > 0) THEN
         DO I=1,MAXNUMCONC
             IF(CO(I)%ACTIVE) CALL TAG_GHOST(CO(I)%CONC)
         END DO
         CALL UPDATE_GHOST
     END IF

!
! Include advection and horizontal diffusivity of S and T 
!
     IF (FOURTH) THEN
       IF(ADVECT_DENSITY) CALL ADVECT(RHON,RHO,DELRHON,RHOAVER,1)
       IF(ADVECT_S)       CALL ADVECT(SN,S,DELSN,SAVER,1)
       IF(ADVECT_T)       CALL ADVECT(TN,T,DELTN,TAVER,2)
       DO i=1,MAXNUMCONC
           IF(CO(I)%ACTIVE) CALL ADVECT(CO(I)%CONCN, CO(I)%CONC, CO(I)%DELCONCN, CO(I)%AVER,i+2)
       END DO
    ELSE
       IF(ADVECT_DENSITY) CALL SUPERBEEF(RHON,RHO,DELRHON,RHOAVER,1)
       IF(ADVECT_S)       CALL SUPERBEEF(SN,S,DELSN,SAVER,1)
       IF(ADVECT_T)       CALL SUPERBEEF(TN,T,DELTN,TAVER,2)

       DO i=1,MAXNUMCONC
           IF(CO(I)%ACTIVE) CALL SUPERBEEF(CO(I)%CONCN, CO(I)%CONC, CO(I)%DELCONCN, CO(I)%AVER,i+2)
       END DO
     ENDIF
!
!  In 4.0 : Leapfrog predictor step.
!
     IF(ADVECT_DENSITY) RHO = RHONM1 + 2.0*DELRHON
     IF(ADVECT_S)       S   = SNM1   + 2.0*DELSN
     IF(ADVECT_T)       T   = TNM1   + 2.0*DELTN
     CALL BOUND2(4)

     IF(NUMCONC > 0) THEN
         DO i=1,MAXNUMCONC
             IF(CO(I)%ACTIVE) CO(I)%CONC = CO(I)%CONCNM1 + 2.0*CO(I)%DELCONCN
         END DO
         CALL BOUND2(10)
     END IF
     
!
!  Update the density field.
!
     IF(.NOT.ADVECT_DENSITY)THEN
       SELECT CASE(EOS)
       CASE("wang")
           CALL DENS
       CASE("unesco")
           CALL DENSUNESCO
       CASE("user")
           CALL USER_EOS
       END SELECT
     END IF
!
!  Take an implicit corrector step in U, V, S and T
!  and store coorrections DELUI, DELVI, DELSI and DELTI.
!
!
!  Compute contributions from the internal pressure from new values of RHO
!
       CALL TAG_GHOST(RHO)
       CALL UPDATE_GHOST

       SELECT CASE(INTERNAL_PRESSURE)
       CASE("fourth")
           CALL INTERNAL4
       CASE("pom")
           CALL INTERNAL
       CASE("balanced")
           CALL INTERNALOPT()
       END SELECT       
!
!  Compute contributions from the advective terms in the momentum equations
!
       CALL SUPERBEEUV
!
!  In MODESPLIT the U and V fields are split in 2-D external parts and 3-D internal parts
!  and propagated in time
!
!  Start from the previous water-level again 
!
       ETA = ETAP

       CALL MODESPLIT(2)       
       CALL DIAGNOSE(1)
       CALL BOUND2(3)
       
       CALL TAG_GHOST(ETA, WIDTH=1)
       CALL TAG_GHOST(U)
       CALL TAG_GHOST(V)
       CALL TAG_GHOST(WADV, WIDTH=1)
       CALL TAG_GHOST(W, WIDTH=1)
       CALL TAG_GHOST(UADV)
       CALL TAG_GHOST(VADV) 
       CALL UPDATE_GHOST
       
       CALL UPDATEDD
       
       CALL TAG_GHOST(D)
       CALL TAG_GHOST(DU)
       CALL TAG_GHOST(DV)
       CALL TAG_GHOST(DUDZ, EAST+WEST, 1)
       CALL TAG_GHOST(DVDZ, NORTH+SOUTH, 1)
       IF(ADVECT_DENSITY) CALL TAG_GHOST(RHO)
       IF(ADVECT_S)       CALL TAG_GHOST(S)
       IF(ADVECT_T)       CALL TAG_GHOST(T)
       CALL UPDATE_GHOST

       DO I=1,MAXNUMCONC
           IF(CO(I)%ACTIVE) CALL TAG_GHOST(CO(I)%CONC)
       END DO
       IF(NUMCONC > 0) CALL UPDATE_GHOST       
       
       IF (FOURTH) THEN
         IF(ADVECT_DENSITY) CALL ADVECT(RHON,RHO,DELRHOSTAR,RHOAVER,1)
         IF(ADVECT_S)       CALL ADVECT(SN,S,DELSSTAR,SAVER,1)
         IF(ADVECT_T)       CALL ADVECT(TN,T,DELTSTAR,TAVER,2)

         DO i=1,MAXNUMCONC
             IF(CO(I)%ACTIVE) CALL ADVECT(CO(I)%CONCN, CO(I)%CONC, CO(I)%DELCONCSTAR, CO(I)%AVER,i+2)
         END DO
       ELSE
         IF(ADVECT_DENSITY) CALL SUPERBEEF(RHON,RHO,DELRHOSTAR,RHOAVER,1)
         IF(ADVECT_S)       CALL SUPERBEEF(SN,S,DELSSTAR,SAVER,1)
         IF(ADVECT_T)       CALL SUPERBEEF(TN,T,DELTSTAR,TAVER,2)

         DO i=1,MAXNUMCONC
             IF(CO(I)%ACTIVE) CALL SUPERBEEF(CO(I)%CONCN, CO(I)%CONC, CO(I)%DELCONCSTAR, CO(I)%AVER,i+2)
         END DO
       ENDIF
!
!  In 4.0 AM-2 Corrector
!
       IF (AM2) THEN
         IF(ADVECT_DENSITY) RHO = RHON + (5.0/12.0)*DELRHOSTAR + (2.0/3.0)*DELRHON - (1.0/12.0)*DELRHONM1
         IF(ADVECT_S)       S = SN + (5.0/12.0)*DELSSTAR + (2.0/3.0)*DELSN - (1.0/12.0)*DELSNM1
         IF(ADVECT_T)       T = TN + (5.0/12.0)*DELTSTAR + (2.0/3.0)*DELTN - (1.0/12.0)*DELTNM1

         DO i=1,MAXNUMCONC
           IF(CO(I)%ACTIVE) CO(I)%CONC = CO(I)%CONCN+(5.0/12.0)*CO(I)%DELCONCSTAR+(2.0/3.0)*CO(I)%DELCONCN-(1.0/12.0)*CO(I)%DELCONCNM1
         END DO
       ELSE
         IF(ADVECT_DENSITY) RHO = RHON + DELRHOSTAR
         IF(ADVECT_S)       S   = SN + DELSSTAR 
         IF(ADVECT_T)       T   = TN + DELTSTAR
         DO I=1,MAXNUMCONC
             IF(CO(I)%ACTIVE) CO(I)%CONC = CO(I)%CONCN + CO(I)%DELCONCSTAR
         END DO
       ENDIF
!
!  Call WREAL to get the z-coordinate vertical velocities
!
1    CALL WREAL
!
!  Vertical diffusion of S and T
!
     IF(ADVECT_DENSITY) CALL VERTDIFF(RHO,RHOSURF)
     IF(ADVECT_S)       CALL VERTDIFF(S,SSURF)
     IF(ADVECT_T)       CALL VERTDIFF(T,TSURF)
     CALL BOUND2(4)

     IF(NUMCONC > 0) THEN
         DO i=1,MAXNUMCONC
             IF(CO(I)%ACTIVE) CALL VERTDIFF(CO(I)%CONC, CO(I)%CONCSURF)
         END DO
         CALL BOUND2(10)
     END IF

     CALL MOVE_PARTICLE()
     
     CALL USER_STEP()
          
2    CALL OUTPUT()

     IF(DO_CHECKPOINT) CALL CHECKPOINT()
     
     T2 = RDTSC()
     
     TOTTIME=(TOTTIME+T2)-T1
     STEPTIME = REAL(T2-T1,8)/REAL(CLOCK_FREQUENCY,8)
     IF(DEBUG > 1 .AND. MASTER) WRITE(*,'(a,f0.6,a,f0.6)')'Timestep took ',STEPTIME,'s'
     
  END DO  ! Main time stepping loop

  
  IF(DEBUG > 0 .AND. MASTER)THEN
      WRITE(*,'(/,f0.3,"s in total was spent timestepping.")') REAL(TOTTIME,8)/REAL(CLOCK_FREQUENCY,8)

      IF(NUMPROC>1) WRITE(*,'(f0.3,"s in MPI routines exchanging halos,",&
           &/,f0.3,"s copying data for the sends, and",&
           &/,f0.3,"s for various horizontal sums")') &
           REAL(TOTCOMMTIME,8)/REAL(CLOCK_FREQUENCY,8), &
           REAL(TOTCOPYTIME,8)/REAL(CLOCK_FREQUENCY,8), &
           REAL(TOTSUMTIME,8)/REAL(CLOCK_FREQUENCY,8) 
  END IF
  
  IF(MASTER.AND.PGPLOT_NUMPLOTS >0) CALL PGCLOS()
  CALL PART_FINALIZE()
  CALL STOP_QMPI()
  CALL DEALLOCATE_BOM()

END PROGRAM MAIN
