module adcirc_tide_fac
!
! Computes nodal factors and equilibrium arguments
! Based on tide_fac.f from ADCIRC: http://www.adcirc.org/
!
! Adapted for BOM5 (22/01/2010)
!
! usage: 
!    call get_astronomical_arguments(...)
!

  private
  public get_astronomical_arguments

  real ds,dp,dh,dp1,dn,di,dnu,dxi,dnup,dnup2,dpc
  real fndcst(37),eqcst(37)

! Constituent names
  
  character(8) :: cst(37)=&
                  (/'M2      ','S2      ','N2      ','K1      ',&
     &              'M4      ','O1      ','M6      ','MK3     ',&
     &              'S4      ','MN4     ','NU2     ','S6      ',&
     &              'MU2     ','2N2     ','OO1     ','LAMBDA2 ',&
     &              'S1      ','M1      ','J1      ','MM      ',&
     &              'SSA     ','SA      ','MSF     ','MF      ',&
     &              'RHO1    ','Q1      ','T2      ','R2      ',&
     &              '2Q1     ','P1      ','2SM2    ','M3      ',&
     &              'L2      ','2MK3    ','K2      ','M8      ',&
     &              'MS4     '/)

! Orbital speeds (degrees/hour)
  
   real :: acst(37) = (/28.9841042,30.0,28.4397295,15.0410686,57.9682084, &
      13.9430356,86.9523127,44.0251729,60.0,57.4238337,28.5125831,90.0, &
      27.9682084,27.8953548,16.1391017,29.4556253,15.0,14.4966939, &
      15.5854433,0.5443747,0.0821373,0.0410686,1.0158958,1.0980331, &
      13.4715145,13.3986609,29.9589333,30.0410667,12.8542862,14.9589314, &
      31.0158958,43.4761563,29.5284789,42.9271398,30.0821373, &
      115.9364169,58.9841042/)

! Number of tide cycles per day per constituent
   
   real :: pcst(37) = (/2.,2.,2.,1.,4.,1.,6.,3.,4.,4.,2.,6.,2.,2.,1.,2.,1.,1.,&
        1.,0.,0.,0.,0.,0.,1.,1.,2.,2.,1.,1.,2.,3.,2.,3.,2.,8.,4./)

  
contains  


  subroutine get_astronomical_arguments(date, numstep, timezone, dt, &
       &  freqlist,constlist, eqlist, nodfaclist)
!
! Input:
!   date(6)    : Start time for simulations
!   numstep,dt : Used to calculate runtime for simulation
!   freqlist   : frequency of each tidal constituent used in the simulation
! Output:
!   eqlist     : array with equilibrium arguments for each frequency provided in freqlist
!   nodfaclist : nodal factors for frequencies in freqlist
    
    implicit none
    
    integer, intent(in) :: date(6), numstep
    real, intent(in) :: timezone, dt, freqlist(:)
    character(len=8), intent(out) :: constlist(size(freqlist))
    real, intent(out) :: eqlist(size(freqlist)), nodfaclist(size(freqlist))
    integer i,n
    real :: rhrs,hrm,dayj,bhr

! Find length of run time in hours, and time half way through run
    rhrs = numstep*dt/3600.
    bhr = real(date(4)) + real(date(5))/60. + real(date(6))/3600. - timezone
    hrm = bhr + rhrs/2.

! Determine the Julian time at the beginning of record
    dayj = dayjul(date(1), date(2), date(3))
  
! Determine node factors at middle of record
    call nfacs(date(1), dayj, hrm)
    
! all 37 node factors are now in FNDCST

! Determine Greenwich equil. terms at beginning of record
    call gterms(date(1),dayj,bhr,dayj,hrm)
    
! all equilibrium arguments are now in EQCST

! search for freqlist(i) in ACST, assign pos. of closest frequency to n
!  (const name is  CST(n) etc)
    
    do i=1, size(freqlist)
        n = findfreq(freqlist(i))
        eqlist(i)     = eqcst(n)
        nodfaclist(i) = fndcst(n)
        constlist(i)  = cst(n)
    end do

  end subroutine get_astronomical_arguments



    SUBROUTINE NFACS(YR,DAYJ,HR)
      IMPLICIT NONE
!-- CALCULATES NODE FACTORS FOR CONSTITUENT TIDAL SIGNAL

!-- THE EQUATIONS USED IN THIS ROUTINE COME FROM:
!         "MANUAL OF HARMONIC ANALYSIS AND PREDICTION OF TIDES"
!         BY PAUL SCHUREMAN, SPECIAL PUBLICATION #98, US COAST
!         AND GEODETIC SURVEY, DEPARTMENT OF COMMERCE (1958).

!-- IF DAYM AND HRM CORRESPOND TO MIDYEAR, THEN THIS ROUTINE
!-- RETURNS THE SAME VALUES AS FOUND IN TABLE 14 OF SCHUREMAN.
!---------------------------------------------------------------------

      integer :: yr
      real :: dayj,hr
      real :: pi180,p,pc,xi,qainv,rainv
      real :: sini,sini2,sin2i,cosi2,tani2
      real :: EQ73,EQ74,EQ75,EQ76,EQ77,EQ78,EQ149,EQ207,EQ215,EQ227,EQ235
      REAL :: I,N,NU
      PI180=3.14159265/180.
      CALL ORBIT(YR,DAYJ,HR)
      N=DN*PI180
      I=DI*PI180
      NU=DNU*PI180
      XI=DXI*PI180
      P=DP*PI180
      PC=DPC*PI180
      SINI=SIN(I)
      SINI2=SIN(I/2.)
      SIN2I=SIN(2.*I)
      COSI2=COS(I/2.)
      TANI2=TAN(I/2.)
!-- EQUATION 197, SCHUREMAN
      QAINV=SQRT(2.310+1.435*COS(2.*PC))
!-- EQUATION 213, SCHUREMAN
      RAINV=SQRT(1.-12.*TANI2**2*COS(2.*PC)+36.*TANI2**4)
!-- VARIABLE NAMES REFER TO EQUATION NUMBERS IN SCHUREMAN
      EQ73=(2./3.-SINI**2)/.5021
      EQ74=SINI**2/.1578
      EQ75=SINI*COSI2**2/.37988
      EQ76=SIN(2*I)/.7214
      EQ77=SINI*SINI2**2/.0164
      EQ78=(COSI2**4)/.91544
      EQ149=COSI2**6/.8758
      EQ207=EQ75*QAINV
      EQ215=EQ78*RAINV
      EQ227=SQRT(.8965*SIN2I**2+.6001*SIN2I*COS(NU)+.1006)
      EQ235=.001+SQRT(19.0444*SINI**4+2.7702*SINI**2*COS(2.*NU)+.0981)
!-- NODE FACTORS FOR 37 CONSTITUENTS:
      FNDCST(1)=EQ78
      FNDCST(2)=1.0
      FNDCST(3)=EQ78
      FNDCST(4)=EQ227
      FNDCST(5)=FNDCST(1)**2
      FNDCST(6)=EQ75
      FNDCST(7)=FNDCST(1)**3
      FNDCST(8)=FNDCST(1)*FNDCST(4)
      FNDCST(9)=1.0
      FNDCST(10)=FNDCST(1)**2
      FNDCST(11)=EQ78
      FNDCST(12)=1.0
      FNDCST(13)=EQ78
      FNDCST(14)=EQ78
      FNDCST(15)=EQ77
      FNDCST(16)=EQ78
      FNDCST(17)=1.0
!** EQUATION 207 NOT PRODUCING CORRECT ANSWER FOR M1
!**SET NODE FACTOR FOR M1 = 0 UNTIL CAN FURTHER RESEARCH
      FNDCST(18)=0.
!     FNDCST(18)=EQ207
      FNDCST(19)=EQ76
      FNDCST(20)=EQ73
      FNDCST(21)=1.0
      FNDCST(22)=1.0
      FNDCST(23)=EQ78
      FNDCST(24)=EQ74
      FNDCST(25)=EQ75
      FNDCST(26)=EQ75
      FNDCST(27)=1.0
      FNDCST(28)=1.0
      FNDCST(29)=EQ75
      FNDCST(30)=1.0
      FNDCST(31)=EQ78
      FNDCST(32)=EQ149
!** EQUATION 215 NOT PRODUCING CORRECT ANSWER FOR L2
!** SET NODE FACTOR FOR L2 = 0 UNTIL CAN FURTHER RESEARCH
      FNDCST(33)=0.
!     FNDCST(33)=EQ215
      FNDCST(34)=FNDCST(1)**2*FNDCST(4)
      FNDCST(35)=EQ235
      FNDCST(36)=FNDCST(1)**4
      FNDCST(37)=EQ78
    END SUBROUTINE NFACS

    SUBROUTINE GTERMS(YR,DAYJ,HR,DAYM,HRM)
!-- CALCULATES EQUILIBRIUM ARGUMENTS V0+U FOR CONSTITUENT TIDE

!-- THE EQUATIONS USED IN THIS ROUTINE COME FROM:
!         "MANUAL OF HARMONIC ANALYSIS AND PREDICTION OF TIDES"
!         BY PAUL SCHUREMAN, SPECIAL PUBLICATION #98, US COAST
!         AND GEODETIC SURVEY, DEPARTMENT OF COMMERCE (1958).

!-- IF DAYM AND HRM CORRESPOND TO MIDYEAR, THEN THIS ROUTINE
!-- RETURNS THE SAME VALUES AS FOUND IN TABLE 15 OF SCHUREMAN.
!---------------------------------------------------------------------
      IMPLICIT NONE
      integer :: yr
      real :: DAYJ,HR,DAYM,HRM
      integer :: ih
      real :: pi180,s,p,h,p1,t,r,pc,xi
      real :: top,bottom,q
      REAL NU,NUP,NUP2,I
      PI180=3.14159265/180.
!* OBTAINING ORBITAL VALUES AT BEGINNING OF SERIES FOR V0
      CALL ORBIT(YR,DAYJ,HR)
      S=DS
      P=DP
      H=DH
      P1=DP1
      T=ANGLE(180.+HR*(360./24.))
!** OBTAINING ORBITAL VALUES AT MIDDLE OF SERIES FOR U
      CALL ORBIT(YR,DAYM,HRM)
      NU=DNU
      XI=DXI
      NUP=DNUP
      NUP2=DNUP2
!* SUMMING TERMS TO OBTAIN EQUILIBRIUM ARGUMENTS
      EQCST(1)=2.*(T-S+H)+2.*(XI-NU)
      EQCST(2)=2.*T
      EQCST(3)=2.*(T+H)-3.*S+P+2.*(XI-NU)
      EQCST(4)=T+H-90.-NUP
      EQCST(5)=4.*(T-S+H)+4.*(XI-NU)
      EQCST(6)=T-2.*S+H+90.+2.*XI-NU
      EQCST(7)=6.*(T-S+H)+6.*(XI-NU)
      EQCST(8)=3.*(T+H)-2.*S-90.+2.*(XI-NU)-NUP
      EQCST(9)=4.*T
      EQCST(10)=4.*(T+H)-5.*S+P+4.*(XI-NU)
      EQCST(11)=2.*T-3.*S+4.*H-P+2.*(XI-NU)
      EQCST(12)=6.*T
      EQCST(13)=2.*(T+2.*(H-S))+2.*(XI-NU)
      EQCST(14)=2.*(T-2.*S+H+P)+2.*(XI-NU)
      EQCST(15)=T+2.*S+H-90.-2.*XI-NU
      EQCST(16)=2.*T-S+P+180.+2.*(XI-NU)
      EQCST(17)=T
      I=DI*PI180
      PC=DPC*PI180
      TOP=(5.*COS(I)-1.)*SIN(PC)
      BOTTOM=(7.*COS(I)+1.)*COS(PC)
      Q=ARCTAN(TOP,BOTTOM,1)
      EQCST(18)=T-S+H-90.+XI-NU+Q
      EQCST(19)=T+S+H-P-90.-NU
      EQCST(20)=S-P
      EQCST(21)=2.*H
      EQCST(22)=H
      EQCST(23)=2.*(S-H)
      EQCST(24)=2.*S-2.*XI
      EQCST(25)=T+3.*(H-S)-P+90.+2.*XI-NU
      EQCST(26)=T-3.*S+H+P+90.+2.*XI-NU
      EQCST(27)=2.*T-H+P1
      EQCST(28)=2.*T+H-P1+180.
      EQCST(29)=T-4.*S+H+2.*P+90.+2.*XI-NU
      EQCST(30)=T-H+90.
      EQCST(31)=2.*(T+S-H)+2.*(NU-XI)
      EQCST(32)=3.*(T-S+H)+3.*(XI-NU)
      R=SIN(2.*PC)/((1./6.)*(1./TAN(.5*I))**2-COS(2.*PC))
      R=ATAN(R)/PI180
      EQCST(33)=2.*(T+H)-S-P+180.+2.*(XI-NU)-R
      EQCST(34)=3.*(T+H)-4.*S+90.+4.*(XI-NU)+NUP
      EQCST(35)=2.*(T+H)-2.*NUP2
      EQCST(36)=8.*(T-S+H)+8.*(XI-NU)
      EQCST(37)=2.*(2.*T-S+H)+2.*(XI-NU)
      DO IH=1,37
         EQCST(IH)=ANGLE(EQCST(IH))
      END DO
    END SUBROUTINE GTERMS


    SUBROUTINE ORBIT(YR,DAYJ,HR)

!-- DETERMINATION OF PRIMARY AND SECONDARY ORBITAL FUNCTIONS

!-- THE EQUATIONS PROGRAMMED HERE ARE NOT REPRESENTED BY EQUATIONS IN
!   SCHUREMAN.  THE CODING IN THIS ROUTINE DERIVES FROM A PROGRAM BY
!   THE NATIONAL OCEANIC AND ATMOSPHERIC ADMINISTRATION (NOAA).
!   HOWEVER, TABULAR VALUES OF THE ORBITAL FUNCTIONS CAN BE FOUND IN
!   TABLE 1 OF SCHUREMAN.


!---------------------------------------------------------------------
      IMPLICIT NONE
      integer :: yr
      real :: dayj,hr
      real :: pi180,x,dyr,dday,p,xi
      REAL I,N,NU,NUP,NUP2

      PI180=3.14159265/180.
      X=AINT((REAL(YR)-1901.)/4.)
      DYR=REAL(YR)-1900.
      DDAY=DAYJ+X-1.
!-- DN IS THE MOON'S NODE (CAPITAL N, TABLE 1, SCHUREMAN)
      DN=259.1560564-19.328185764*DYR-.0529539336*DDAY-.0022064139*HR
      DN=ANGLE(DN)
      N=DN*PI180
!-- DP IS THE LUNAR PERIGEE (SMALL P, TABLE 1)
      DP=334.3837214+40.66246584*DYR+.111404016*DDAY+.004641834*HR
      DP=ANGLE(DP)
      P=DP*PI180
      I=ACOS(.9136949-.0356926*COS(N))
      DI=ANGLE(I/PI180)
      NU=ASIN(.0897056*SIN(N)/SIN(I))
      DNU=NU/PI180
      XI=N-2.*ATAN(.64412*TAN(N/2.))-NU
      DXI=XI/PI180
      DPC=ANGLE(DP-DXI)
!-- DH IS THE MEAN LONGITUDE OF THE SUN (SMALL H, TABLE 1)
      DH=280.1895014-.238724988*DYR+.9856473288*DDAY+.0410686387*HR
      DH=ANGLE(DH)
!-- DP1 IS THE SOLAR PERIGEE (SMALL P1, TABLE 1)
      DP1=281.2208569+.01717836*DYR+.000047064*DDAY+.000001961*HR
      DP1=ANGLE(DP1)
!-- DS IS THE MEAN LONGITUDE OF THE MOON (SMALL S, TABLE 1)
      DS=277.0256206+129.38482032*DYR+13.176396768*DDAY+.549016532*HR
      DS=ANGLE(DS)
      NUP=ATAN(SIN(NU)/(COS(NU)+.334766/SIN(2.*I)))
      DNUP=NUP/PI180
      NUP2=ATAN(SIN(2.*NU)/(COS(2.*NU)+.0726184/SIN(I)**2))/2.
      DNUP2=NUP2/PI180
    END SUBROUTINE

    FUNCTION ANGLE(ARG)
!
!*** THIS ROUTINE PLACES AN ANGLE IN 0-360 (+) FORMAT
!
      IMPLICIT NONE
      real :: angle,arg
      integer :: m
      M=-IFIX(ARG/360.)
      ANGLE=ARG+FLOAT(M)*360.
      IF(ANGLE .LT. 0.) ANGLE=ANGLE+360.
    END FUNCTION ANGLE

    FUNCTION ARCTAN(TOP,BOTTOM,KEY)
!** DETERMINE ARCTANGENT AND PLACE IN CORRECT QUADRANT
!   IF KEY EQ 0  NO QUADRANT SELECTION MADE
!   IF KEY .NE. 0 PROPER QUADRANT IS SELECTED

      IF(BOTTOM .NE. 0.0) GO TO 4
      IF(TOP) 2,9,3
    2 ARCTAN=270.
      RETURN
    3 ARCTAN=90.
      RETURN
    4 ARCTAN=ATAN(TOP/BOTTOM)*57.2957795
      IF(KEY.EQ.0) RETURN
      IF(TOP) 5,5,7
    5 IF(BOTTOM) 6,9,8
    6 ARCTAN=ARCTAN+180.
      RETURN
    7 IF(BOTTOM) 6,3,10
    8 ARCTAN=ARCTAN+360.
      RETURN
    9 ARCTAN=0.
   10 RETURN
    END FUNCTION ARCTAN


    FUNCTION DAYJUL(YR,XMONTH,DAY)
!
!*** THIS ROUTINE COMPUTES THE JULIAN DAY (AS A REAL VARIABLE)
!
      IMPLICIT NONE
      real :: dayjul
      integer :: yr,xmonth,day
      integer :: yrlp,i
      real :: dinc
      REAL DAYT(12),DAYS(12)
      DATA DAYT/0.,31.,59.,90.,120.,151.,181.,212.,243.,273.,304.,334./
      DATA DAYS(1),DAYS(2) /0.,31./
      DINC=0.
      YRLP=MOD((YR-1900),4)
      IF(YRLP .EQ. 0) DINC=1.
      DO I=3,12
         DAYS(I)= DAYT(I)+DINC
      END DO
      DAYJUL=DAYS(XMONTH) + REAL(DAY)
    END FUNCTION DAYJUL

  function findfreq(f)
    integer findfreq
    real f, mindiff

    mindiff = 360.
    do i=1,37
        if(abs(f-acst(i)) < mindiff) then
            findfreq=i
            mindiff=abs(f-acst(i))
        end if
    end do
  end function findfreq


end module
