SUBROUTINE ADVECT(FIELDOLD,FIELD,DELFIELD,FAVER,ICODE)
!
!           ADVECT integrates the equation
!
!           d(FIELD*D)/d(T) + d(FIELD*U*D)/d(X) + d(FIELD*V*D)/d(Y) +
!       
!           d(FIELD*W)/d(SIGMA) = d(AH*d(FIELD*D)/d(X))/d(X) +
!                                 d(AH*d(FIELD*D)/d(Y))/d(Y)
!
!           from timestep n to timestep n+1
!           with a superbee limiter TVD scheme.
!
!           d(*) is the differential operator,
!           FIELD is the scalar field to be integrated,
!           D is the dynamic depth,
!           T is time,
!           X and Y are the horisontal coordinates,
!           U is the horisontal velocity in X-direction,
!           V is the horisontal velocity in Y-direction,
!           SIGMA is the vertical coordinate,
!           W is the vertical velocity. 
!           
!           FAVER is the average value of the scalar field which is 
!                 subtracted from FIELD before differences are 
!                 computed to reduce the cancellation errors.
!           ICODE is an integer flag to signal which scalar field
!                 that is advected. ICODE is needed when river-runoff
!                 is included.
!
!***ROUTINES CALLED-NONE
!***END PROLOGUE
!
!   Global variables.
!
  USE STATE
  USE CONTROL_PARAMETERS
  USE MOD_BOUND
  IMPLICIT NONE
  REAL, POINTER :: FIELDOLD(:,:,:),FIELD(:,:,:),DELFIELD(:,:,:)
  REAL FAVER
  INTEGER ICODE
!
!   Local variables.
!
  REAL, DIMENSION(I1:I2,J1:J2,KB) :: ZFLUX,XFLUXD,YFLUXD
  REAL, DIMENSION(I1:I2,J1:J2,KB) :: DIFF1,FNEW
  REAL, DIMENSION(:,:,:), POINTER :: XFLUX,YFLUX,DIFF2
!
  REAL AREA,NETFLUX,VOLUME,VOLUMEP
  REAL AREAT,AREATU,AREATV,AREATW
  REAL EPSILON,ONESIX
  INTEGER I,J,K,RIV,WELL
  REAL FRAC,FLUX,FIELDRIVER,FIELDWELL
  REAL SUM,PROD
  LOGICAL INBLOCK
!
  EPSILON = 1.E-5
  ONESIX = 1.D0/6.D0

  allocate(xflux(i1:i2,j1:j2,kb), yflux(i1:i2,j1:j2,kb), diff2(i1:i2,j1:j2,kb))
  
  call tag_ghost( field, width= 1)
  call update_ghost

!
!   Define FIELD(KB) = FIELD(KB-1)
!
  FIELD(:,:,KB) = FIELD(:,:,KB-1)

  IF(.NOT.ADVECT_DENSITY)THEN
      FIELD = FIELD - FAVER
      FIELDOLD = FIELDOLD - FAVER
  END IF
  
  DELFIELD = 0.
!
  XFLUX = 0.
  XFLUXD = 0.
  YFLUX = 0.
  YFLUXD = 0.
  ZFLUX = 0.
!
!   Compute F*U*D in U-points of the
!   Arakawa C-grid.
!
!   Compute the sum of the advective and diffusive
!   fluxes  in X-direction.
!
  IF (RIVLOG) THEN
      DO RIV = 1,NUMC
          I=IRIVER(RIV)
          J=JRIVER(RIV)
          CALL GLOBALIND2LOCAL(I,J,INBLOCK)
    
          IF(INBLOCK)THEN
             SELECT CASE(ICODE)
             CASE(1)
                 FIELDRIVER = SRIVER(riv)
             CASE(2)
                 FIELDRIVER = TRIVER(riv)
             CASE(3:MAXNUMCONC+2)
                 FIELDRIVER = CRIVER(ICODE-2,RIV)
             CASE DEFAULT
                 WRITE(*,*)'ERROR IN ADVECT'
                 STOP
             END SELECT
             
             FIELDRIVER = FIELDRIVER - FAVER
             DO K = KRIVER1(RIV),KRIVER2(RIV)
                 FRAC = DZ(K)/DZRIVER(RIV)
                 FLUX = FRAC*FRIVER(RIV)*FIELDRIVER 
                 IF (CODERIVER(RIV).EQ.1) THEN
                     XFLUX(I,J,K) = FLUX
                 ELSEIF (CODERIVER(RIV).EQ.2) THEN
                     YFLUX(I,J,K) = FLUX
                 ELSEIF (CODERIVER(RIV).EQ.3) THEN
                     XFLUX(I+1,J,K) = -FLUX
                 ELSEIF (CODERIVER(RIV).EQ.4) THEN
                     YFLUX(I,J+1,K) = -FLUX
                 ELSE
                     WRITE(*,*)'ERROR IN ADVECT-2'
                     STOP
                 ENDIF
             END DO
         end if
     END DO
  ENDIF
  
! 
!  DIFF1
!
  DIFF1 = 0.

  DO J = 1,J_EQ_JM
      DO I = I_EQ_2,i2  !I_EQ_IM
          DO K = 1,KB-1
              DIFF1(I,J,K) = (FIELD(I,J,K) -FIELD(I-1,J,K))*DUM(I,J)
          END DO
      END DO
  END DO
!
!  DIFF2
!
  DIFF2 = 0.

  DO J = 1,J_EQ_JM
      DO I = 1,IM_MI_1
          DO K = 1,KB-1
              SUM = DIFF1(I,J,K)+DIFF1(I+1,J,K)
              PROD = DIFF1(I,J,K)*DIFF1(I+1,J,K)
              IF (ABS(SUM).GT.EPSILON .AND. PROD.GT.EPSILON) THEN
                  DIFF2(I,J,K) = 2.0*PROD/SUM
                  DIFF2(I,J,K) = DIFF2(I,J,K)*FSM(I,J)
              ENDIF
          END DO
      END DO
  END DO
!
!  FNEW
!
  FNEW = 0.

call tag_ghost( diff2, west, 1)
call update_ghost


  DO J = 1,J_EQ_JM
      DO I = I_EQ_2,I_EQ_IM
          DO K = 1,KB-1
              FNEW(I,J,K) = 0.5*(FIELD(I,J,K)+FIELD(I-1,J,K)) - ONESIX*(DIFF2(I,J,K)-DIFF2(I-1,J,K))
              FNEW(I,J,K) = FNEW(I,J,K)*DUM(I,J)
          END DO
      END DO
  END DO



  DO J = J_EQ_2,JM_MI_1
      DO I = I_EQ_2,I_EQ_IM
          IF (DUM(I,J).NE.0.) THEN
              DO K = 1,KB-1
                  AREAT = DY*DUDZ(I,J,K)
                  AREATU = AREAT*UADV(I,J,K)
                  XFLUXD(I,J,K) = AH(I,J,K)*AREAT*(FIELD(I,J,K)-FIELD(I-1,J,K))/DX
                  XFLUX(I,J,K) = AREATU*FNEW(I,J,K)
                  XFLUX(I,J,K) = XFLUX(I,J,K) - XFLUXD(I,J,K)
              END DO
          ENDIF
      END DO
  END DO

  
  IF(JM > 1)THEN
! JM==1 => all differences in j direction is 0
      
!
!  DIFF1
!
  DIFF1 = 0.

  DO J = J_EQ_2,j2  !J_EQ_JM
      DO I = 1,I_EQ_IM
          DO K = 1,KB-1
              DIFF1(I,J,K) = (FIELD(I,J,K) -FIELD(I,J-1,K))*DVM(I,J)
          END DO
      END DO
  END DO
!
!  DIFF2
!
  DIFF2 = 0.

  DO J = 1,JM_MI_1
      DO I = 1,IM_MI_1
          DO K = 1,KB-1
              SUM = DIFF1(I,J,K)+DIFF1(I,J+1,K)
              PROD = DIFF1(I,J,K)*DIFF1(I,J+1,K)
              IF (ABS(SUM).GT.EPSILON .AND. PROD.GT.EPSILON) THEN
                  DIFF2(I,J,K) = 2.0*PROD/SUM
                  DIFF2(I,J,K) = DIFF2(I,J,K)*FSM(I,J)
              ENDIF
          END DO
      END DO
  END DO
!
!  FNEW
!
  FNEW = 0.

call tag_ghost( diff2, south, 1)
call update_ghost


  DO J = J_EQ_2,J_EQ_JM
      DO I = 1,I_EQ_IM
          DO K = 1,KB-1
              FNEW(I,J,K) = 0.5*(FIELD(I,J,K)+FIELD(I,J-1,K)) - ONESIX*(DIFF2(I,J,K)-DIFF2(I,J-1,K))
              FNEW(I,J,K) = FNEW(I,J,K)*DVM(I,J)
          END DO
      END DO
  END DO
!
!   Compute the sum of the advective and diffusive
!   fluxes  in Y-direction.
!

  DO J = J_EQ_2,J_EQ_JM
      DO I = I_EQ_2,IM_MI_1
          IF (DVM(I,J).NE.0.) THEN
              DO K = 1,KB-1
                  AREAT = DX*DVDZ(I,J,K)
                  AREATV = AREAT*VADV(I,J,K)
                  YFLUXD(I,J,K) = AREAT*AH(I,J,K)*((FIELD(I,J,K)-FIELD(I,J-1,K))/DY)
                  YFLUX(I,J,K) = AREATV*FNEW(I,J,K)
                  YFLUX(I,J,K) = YFLUX(I,J,K) - YFLUXD(I,J,K)
              END DO
          ENDIF
      END DO
  END DO

  END IF   ! JM>1

!
!   Compute the advective fluxes in sigma-direction.
!
!
!  DIFF1
!
  DIFF1 = 0.

  DO J = 1,J_EQ_JM
      DO I = 1,I_EQ_IM
          DO K = 2,KB
              DIFF1(I,J,K) = (FIELD(I,J,K-1) -FIELD(I,J,K))*FSM(I,J)
          END DO
      END DO
  END DO
!
!  DIFF2
!
  DIFF2 = 0.

  DO J = 1,J_EQ_JM
      DO I = 1,IM_MI_1
          DO K = 1,KB-1
              SUM = DIFF1(I,J,K)+DIFF1(I,J,K+1)
              PROD = DIFF1(I,J,K)*DIFF1(I,J,K+1)
              IF (ABS(SUM).GT.EPSILON .AND. PROD.GT.EPSILON) THEN
                  DIFF2(I,J,K) = 2.0*PROD/SUM
                  DIFF2(I,J,K) = DIFF2(I,J,K)*FSM(I,J)
              ENDIF
          END DO
      END DO
  END DO
!
!  FNEW
!
  FNEW = 0.

  DO J = J_EQ_2,J_EQ_JM
      DO I = 1,I_EQ_IM
          DO K = 2,KB-1
              FNEW(I,J,K) = 0.5*(FIELD(I,J,K)+FIELD(I,J,K-1)) - ONESIX*(DIFF2(I,J,K-1)-DIFF2(I,J,K))
              FNEW(I,J,K) = FNEW(I,J,K)*FSM(I,J)
          END DO
      END DO
  END DO
  AREA = DX*DY

  DO J = J_EQ_2,JM_MI_1
      DO I = I_EQ_2,IM_MI_1
          IF(FSM(I,J).NE.0.) THEN
              DO K = 2,KB
                  AREATW = AREA*W(I,J,K)
                  ZFLUX(I,J,K) = AREATW*FNEW(I,J,K)
              END DO
          ENDIF
      END DO
  END DO
!
!  If UPWELL, then also specify a flux through the sea bed.
!
  IF (UPWELL) THEN
      DO WELL = 1,NUMWELL
          I=IWELL(WELL)
          J=JWELL(WELL)
          CALL GLOBALIND2LOCAL(I,J,INBLOCK)
    
          IF(INBLOCK)THEN
              SELECT CASE(ICODE)
              CASE(1)
                  FIELDWELL = SWELL(WELL)
              CASE(2)
                  FIELDWELL = TWELL(WELL)
              CASE(3:MAXNUMCONC+2)
                  FIELDWELL = CWELL(ICODE-2,WELL)
              CASE DEFAULT
                  WRITE(*,*)'ERROR IN ADVECT'
                  STOP
              END SELECT
              FIELDWELL = FIELDWELL - FAVER
              ZFLUX(I,J,KB) = FWELL(WELL)*FIELDWELL
          END IF
      END DO
  ENDIF
!
!   Take a step to timestep n+1  by adding up the fluxes in and out of each cell.
!
call tag_ghost( xflux, east,  1)
call tag_ghost( yflux, north, 1)
call update_ghost


  DO J = J_EQ_2,JM_MI_1
      DO I =I_EQ_2,IM_MI_1
          IF (FSM(I,J).GT.0.0) THEN
              DO K = 1,KB-1
                  NETFLUX = (ZFLUX(I,J,K)-ZFLUX(I,J,K+1)) + &
                       &             XFLUX(I+1,J,K)-XFLUX(I,J,K)
                  if(jm > 1)then
                     netflux = netflux + YFLUX(I,J+1,K) - YFLUX(I,J,K)
                  end if
                  VOLUME = DX*DY*DZ(K)*(H(I,J)+ETA(I,J))
                  VOLUMEP = DX*DY*DZ(K)*(H(I,J)+ETAP(I,J))
                  FIELD(I,J,K) = (FIELDOLD(I,J,K)*VOLUMEP - DT*NETFLUX)/VOLUME
              END DO
          ENDIF
      END DO
  END DO
!
  DELFIELD = FIELD - FIELDOLD

  IF(.NOT.ADVECT_DENSITY)THEN
      FIELD = FIELD + FAVER
      FIELDOLD = FIELDOLD + FAVER
  END IF

  CALL DUMP_STATE("advect")
  deallocate(xflux, yflux, diff2)
END SUBROUTINE ADVECT
