module autodecomp
  private
  public get_decomp
  
contains
  
  subroutine get_decomp(np, h, m, n)
!    
! Find a good MxN decomposition of 2D array H, given an estimate for
! the number of blocks, NP, where H is > 0.
!    
    implicit none
    integer, intent(in) :: np         ! # processors
    real, intent(in) :: h(:,:)        ! bathymetry
    integer, intent(out) :: m,n       ! blocks in i and j direction
    integer im,jm,i,j,k   
    integer dry,om,on,inactive
    real drypercent
    type searchobj
       integer m,n,dry,inactive,di,dj
       real drypercent, ratio
    end type searchobj
    integer, parameter :: maxresults=5000
    type(searchobj) result(maxresults)
    integer numres, opt, res, di, dj
    real :: rectangularity
    
    im=size(h,1)
    jm=size(h,2)
    
! estimate a optimal block configuration OM,ON:
! we seek as quadratic blocks as possible to maximize the number of
! grid cells wrt circumference, so m/n should be close to im/jm.
!   numproc = wet + inactive    
!   numproc = numbli*numblj - dry(numbli,numblj) + inactive  
!   inactive = numproc + dry(numbli,numblj) - numbli*numblj    
! we want to minimize this. 
!
! the percentage of dry cells L is known. then the number of wet
! blocks must be larger than (1-L)*numbli*numblj, while optimally
! numblj=numbli*jm/im    
!   numproc = (1-L)*numbli*jm/im*numbli
!   numbli = sqrt(numproc*im/jm/(1-L))    
    
    call drycells(h, dry, drypercent)
    
    om = int(sqrt(im*np/(1 - 0.01*drypercent)/jm))
    on = (jm*om)/im

!    print *,'om,on=',om,on,'i',max(1,om-20),om+2,'j',max(1,on-20),on+2
    
    opt=np
    numres=0
    res=0

! first decide how rectangular we should allow blocks to be using
! max(di,dj)/min(di,dj)-1.0 as a criterium.
! large block counts should get us closer to the optimum, so we can be more strict:
!     
    rectangularity = 0.2
    if(np < 64) rectangularity = 1
    if(np < 24) rectangularity = 2

! search over a span of block configurations around the optimal number
! from above, and find all block configs where the rectangularity
! requirement is met:
    do j=max(1,on-20),on+2
        do i=max(1,om-20),om+2
            dry = i*j - numwetblocks(h, i, j, di, dj)
            inactive = np + dry - i*j                   

            if(inactive >= 0 .and. real(max(di,dj))/min(di,dj)-1.0 < rectangularity) then
                numres = numres + 1

                if(numres > maxresults) then
                    write(*,*) "get_decomp warning: more than ",maxresults,"results. present max is ",&
                         (om+2-max(1,om-20)+1) *  (on+2 - max(1,on-20))
                    goto 101
                end if
                
                result(numres)%m = i
                result(numres)%n = j
                result(numres)%di = di
                result(numres)%dj = dj                
                result(numres)%dry = dry
                result(numres)%inactive = inactive
                result(numres)%drypercent = 100*dry/(i*j)
                result(numres)%ratio = real(max(di,dj))/min(di,dj)-1.0
                if(result(numres)%inactive < opt) then
                    opt = result(numres)%inactive
                    res = numres
                end if
!                print '(i7,i7,i7,i9,1x,f10.2,1x,f10.2)',result(numres)
            end if
        end do
    end do
    
101 continue
    
    if(res==0) then
        print *,'no block config found'
    else
        m=result(res)%m
        n=result(res)%n
    end if

! for small processor counts, 1d decomposition is often better (unless
! the grid is really small?)
    
    if(opt > 0 .and. np < 11) then
        if(im > jm) then
            m = np ; n=1
        else
            m = 1 ; n = np
        end if
    end if

    write(*,*) "Autodecomp:"
    write(*,*)'  decomposition M,N=',m,n
    
    if(opt < np)then
        write(*,*)'  inactive processors:',       result(res)%inactive
        write(*,*)'  block size:',                result(res)%di,result(res)%dj
        write(*,*)'  grid side ratio IM/JM:',     real(im)/real(jm)
        write(*,*)'  block rectangularity:',      result(res)%ratio
        write(*,*)'  optimal decomposition M,N:', om,on
        write(*,*)'  drycellpercent:',            drypercent
        write(*,*)'  dryblockpercent:',           result(res)%drypercent
    end if
  end subroutine get_decomp
  
  subroutine drycells(h, dry, drypercent)
! given a 2D array H, return the number of elements DRY where H <= 0.,
! and the percentage of such cells in DRYPERCENT
    real h(:,:), drypercent
    integer dry,i,j
    dry=0
    do j=1,size(h,2)
        do i=1,size(h,1)
            if(h(i,j) <= 0.0) then
                dry=dry+1
            end if
        end do
    end do
    drypercent=100*real(dry)/real(size(h,1)*size(h,2))
  end subroutine drycells

  integer function numwetblocks(f, m, n, di, dj)
! decompose array F into MxN blocks. return the number of blocks where
! all F on block is > 0.
    implicit none
    real f(:,:)
    integer i,j,m,n,gi1,gi2,gj1,gj2,im,jm,imt,jmt,di, dj

    im = size(f,1)
    jm = size(f,2)

    imt=im
    do
        if(mod(imt,m)== 0) exit
        imt=imt+1
    end do
    jmt=jm
    do
        if(mod(jmt,n)==0) exit
        jmt=jmt+1
    end do

    di=imt/m
    dj=jmt/n

    numwetblocks=0

    do j=1,n
        do i=1,m
!! global range for a block
            gi1 = (i-1)*di + 1
            gi2 = min(im, i*di)
            gj1 = (j-1)*dj + 1
            gj2 = min(jm, j*dj)

!            if(maxval(f(gi1:gi2,gj1:gj2)) > 0.0) then
!            if(sum(f(gi1:gi2,gj1:gj2)) > 0.0) then
            if(anywet(gi1,gi2,gj1,gj2)) then
                numwetblocks = numwetblocks + 1
            end if
        end do
    end do
  contains
    logical function anywet(i1,i2,j1,j2)
      integer i1,i2,j1,j2,i,j
      anywet=.true.
      do j=j1,j2
          do i=i1,i2
              if(f(i,j) > 0.0) return
          end do
      end do
      anywet=.false.      
    end function anywet
  end function numwetblocks
end module autodecomp
