module bomplot
! 
! Data dumping routines for BOM, usually called from output.f90.
! please report bugs/feature requests to
!    Helge Avlesen <avle@ii.uib.no>, para//ab
! 
! **Usage via the setup file
!  
! See documentation for keywords starting with "bomplot_" in
! setup2.f90 or via running bomhelp.py
!  
! Basically each plotting keyword specified, adds a plot object to the
! list "plot" below. the routine batch_bomplot (called once per
! timestep) loops over the list and does whatever is specified in the
! object.
!  
! **Calling the library routines directly
!  
! Some of the arguments are common for many of the subroutines:
! 
! the optional argument REGION specifies a horizontal region from
! which to dump data.  REGION is an integer array of length 4,
! containing AI,AJ,BI,BJ where AI,AJ is the global indice coordinate
! of the 'south west' corner of the reduced size region to
! dump. (BI,BJ upper right corner)
!  
! The optional argument FULLPREC on specify whether the data should be
! dumped to the ascii files with full precision unformatted output
! (.TRUE.), or a reduced size formatted format (.FALSE.).
! 
!  SUBROUTINE SLICE2D(LABEL, FIELD, RANGE)  
!  SUBROUTINE SLICE3D(LABEL, FIELD, RANGE)  
!   let R=RANGE, SLICExD dumps FIELD(r(1):r(2),r(3):r(4) [,r(5):r(6)] ) to 
!   the file "<label>-<idim>x<jdim>x<kdim>.data"  
!  
!  SUBROUTINE PARBINDUMP(LABEL, FIELD [, RANGE])
!   similar to SLICExD above, but dumps to unformatted binary files
!   FIELD, can be a 2D or 3D distributed array.
!   If RANGE is not present, dumps the whole field.
!
! SUBROUTINE VSECT(LABEL,X,Y,FULLPREC, FIELD1, FIELD2)
!   optional: FULLPREC, FIELD1, FIELD2
!  
!   save UN(normal component),UT(tangential component),WR(vertical
!   component), S,T for the section along the polygon with N points,
!   defined in vectors X,Y, to a filename constructed from LABEL,
!   ISTEP and the resulting grid size. if the velocity, S and T is not the
!   desired data from the section, the user can provide the optional,
!   real, 3d arrays FIELD1 and/or FIELD2.
!
! SUBROUTINE DAVER(LABEL, FIELD, REGION)
!   given a 3D FIELD, compute and dump the depth average, optionally from REGION
!  
! SUBROUTINE HSECT(LABEL, FIELD, DEPTH, LEVEL, REGION,FULLPREC,POS)
!   optional: FIELD and REGION, and one of DEPTH, LEVEL.
!   if field is not given, save U,V,S,T for the horisontal section at DEPTH
!   meters below static 0 level, or at sigma level LEVEL if DEPTH is not given.
!   if DEPTH<0, sample the section at |depth| meters above bottom.
!   the filename constructed from LABEL, ISTEP and the grid size.
!   POS is "u","v" or "c" (center=default) the grid type of FIELD.
!  
! SUBROUTINE HSECT2D(LABEL, FIELD, REGION,FULLPREC)
!   optional: FIELD, REGION
!   quick dump horisontal fields for ETA, UA, VA to a file which name is
!   constructed from LABEL  
!   if a 2d array FIELD is given, only dump that one array.
!   example: to dump the bathymetry to a file:
!       ...  
!       call hsect2d('bathymetry', field=h)
!       ...
!   a reduced size region can be specified as for HSECT above.
!   to dump the bathymetry from the region i=10,20 and j=20,40 ;
!       call hsect2d('bath-corner1', field=h, region=(/10,20,20,40/) )
!  
! SUBROUTINE VPROF(LABEL,AX,AY,FIELD,FULLPREC)
!   save the vertical profile of U,V,S and T under the point AX,AY
!   filename constructed from LABEL, ISTEP and the vertical grid size
!   if the optional 3d field FIELD is given, only dump that ones profile.
!  
! SUBROUTINE TSVPROF(LABEL, IP,JP,FIELD,FULLPREC)
!   append a vertical profile for horisontal indice IP,JP of U,V,S,T,...
!   each time called, to a file with name constructed from LABEL.
!   the vertical profiles will be stored in batches of bomplot_vprof_interval 
!   samples(profiles) in each file (defaults to 100).
!   if the optional 3d field FIELD is given, only store the profile of FIELD.
!  
! SUBROUTINE ISOSURF(LABEL, FIELD, ISOVAL, REGION,FULLPREC)
!   optional: REGION
!  
!   Given a 3d FIELD, ISOSURF finds the location of the isosurface
!   with value ISOVAL, as well as the velocity components U,V and WR
!   on this surface. The filename is constructed from LABEL and
!   ISTEP. The isolevel is found by searching each water column from
!   the top. Optionally only store the isosurface for the Fortran
!   region defined in RANGE. FIELD is assumed to be cell centered, like S,T.
!
  use mod_mpi
  
  use pgplot_core, only : pgundef, pgplot_current_totsec, pg_ssec, &
       pg_usec, pg_utsec, pg_vx, pg_vz, pg_f, pg_u, pg_v

! if false, no plotting will be done. 
  logical :: enable_bomplot=.true.

! the directory in which to store data files from bomplot. should end
! with a /.

  character(len=80) :: bomplot_workdir='./'

  character(len=80) :: &
       hsect_dir = "HSECT/",&
       vsect_dir = "VSECT/",&
       vprof_dir = "VPROF/",&
       slice_dir = "SLICE/"
    
! how many vertical profiles in each dumped vprof file? filenames are
! ordered, so they can be cat'ed together after run to create a full
! profile.

  integer :: bomplot_vprof_interval=100

! for polygon vsect
  integer, parameter :: pmaxsec=20  
  
! how to save data: 0=use gather to store global fields, 1=each
! processor stores a block, 2=each processor stores unformatted stream
  
  integer :: bomplot_dumping_method=0

! if true, bomplot_gnuplot makes dump2d add a blank line between rows,
! which enables direct plots of files with the splot command.
! only works if bomplot_compress is set to false.  
  logical :: bomplot_gnuplot=.false.

! the default action if the fullprec argument to hsect etc. is not
! given.  if compress is true, write ascii data with minimal precision
! and width in order to save space.
  logical :: bomplot_compress=.false.

  integer, parameter :: filenamesize=200    ! max length of full path to a bomplot outfile

  

! general purpose object describing a plot to be carried out by batch_bomplot routine
! typically only some of the values here are actually used.  
  type plot_object
     character(20) :: label=" "                       ! identifier for what to dump data for
     integer :: type=0
     real(8) :: last_plot_time=-3600.                 ! time specification for when and how often to dump
     real :: plotdt=0.0, start=0.0, end=0.0
     integer :: fileunit=0                            ! to where, in case of open units.
     integer :: i,j,k,n                               ! various parameters defining from where to pick the data
     integer range(6)
     real :: depth=10000.0
     real :: isoval=0.0   
     real :: true_range(4) = 0.0
     integer :: level=-1
     integer :: nsecp1=0
     real, dimension(pmaxsec) :: px, py
     real x1,y1,x2,y2
     real, pointer, dimension(:,:) :: f2d => null()   ! pointer to 2d field from which to interpolate
     real, pointer, dimension(:,:,:) :: f3d => null() ! pointer to 3d field from which to interpolate
     character(20) :: what=" "                        ! what to slice?
     real, allocatable :: data4d(:,:,:,:)               ! storage for accumslice
  end type plot_object

  integer, parameter :: maxpts=10000

  integer, parameter :: bomplot_maxplots=100
  integer :: bomplot_numplots=0
! a list of plots to be made:
  type(plot_object), allocatable :: bomplots(:)

  character(30) :: plotdesc(0:6)=(/&
       "hsect                      ",&
       "hsect                      ",&
       "hsect                      ",&
       "vsect                      ",&
       "hsect2d                    ",&
       "hsect2d                    ",&
       "vsect with polygon         "/)


  interface parbindump
! subroutine parbindump(label, f[, range])
!   dump distributed array f to an unformatted file, which name is constructed from "label"
!   and the array dimensions of f.
     module procedure bindump2d, bindump3d
  end interface


contains  


  subroutine batch_bomplot
! 
! to be called each time step. in setup2 a list of bomplot task
! objects, PLOT, has been made. each object contains control
! parameters for what to plot, when etc.  each plot task e.g. has a
! start time (plot(i)%start), an end time, and a "plotdt", the minimum
! time between dumps. this routine loops over all plot objects and
! executes the appropriate dumping routines.
! 
    use time_util, only : time
    use state, only : debug, rho, jm
    implicit none
    integer i

    do i=1, bomplot_numplots
        if(  time >= bomplots(i)%last_plot_time + bomplots(i)%plotdt  .and. &
             time >= bomplots(i)%start .and. time <= bomplots(i)%end ) then

            bomplots(i)%last_plot_time = time

            select case(bomplots(i)%type)
            case(0)
                call hsect(trim(bomplots(i)%label), position=bomplots(i)%depth)
            case(1)
                call hsect(trim(bomplots(i)%label), level=bomplots(i)%level)
            case(2)
                call hsect(trim(bomplots(i)%label), position=bomplots(i)%depth, range=bomplots(i)%range)
            case(4)
                call hsect2d(trim(bomplots(i)%label))
            case(5)
                call hsect2d(trim(bomplots(i)%label), range=bomplots(i)%range)
            case(6)
                call vsect(trim(bomplots(i)%label), bomplots(i)%px(1:bomplots(i)%nsecp1), bomplots(i)%py(1:bomplots(i)%nsecp1))
            case(7)
                call isosurf(trim(bomplots(i)%label), rho, bomplots(i)%isoval, range=bomplots(i)%range)                
            case(8)
                call tsvprof(trim(bomplots(i)%label), bomplots(i)%i, bomplots(i)%j)
            case(9)
                call do_slice(i)
            case(10)
                call hsect(trim(bomplots(i)%label), field=bomplots(i)%f3d, level=bomplots(i)%level, range=bomplots(i)%range)
            case(11)
                call hsect(trim(bomplots(i)%label), field=bomplots(i)%f3d, position=bomplots(i)%depth, range=bomplots(i)%range)
            case(12)
                call do_binslice(i)
            case(13)
                call do_accumslice(i)
            case(14)
                call daver(trim(bomplots(i)%label), field=bomplots(i)%f3d, range=bomplots(i)%range)
            case default
                write(*,*)'batch_bomplot error; unknown plot type=', bomplots(i)%type
            end select
        end if
    end do
  end subroutine batch_bomplot

  
  subroutine allocate_bomplot()
    implicit none
    
    allocate(bomplots(bomplot_maxplots))
  end subroutine allocate_bomplot
  
  subroutine deallocate_bomplot()
    implicit none
    
    deallocate(bomplots)
  end subroutine deallocate_bomplot


  subroutine do_binslice(i)   
! redirect calls to 2d and 3d variants    
    integer i

    if(associated(bomplots(i)%f3d)) call parbindump(trim(slice_dir)//trim(bomplots(i)%label), bomplots(i)%f3d, bomplots(i)%range)
    if(associated(bomplots(i)%f2d)) call parbindump(trim(slice_dir)//trim(bomplots(i)%label), bomplots(i)%f2d, bomplots(i)%range(1:4))
  end subroutine do_binslice 

  
  subroutine do_slice(i)
    
! redirect calls to 2d and 3d variants
    
    use state, only : kb
    integer i

    if(associated(bomplots(i)%f3d)) then
        call slice3d(trim(bomplots(i)%label), bomplots(i)%f3d, bomplots(i)%range)
        return
    end if
    if(associated(bomplots(i)%f2d)) then
        call slice2d(trim(bomplots(i)%label), bomplots(i)%f2d, bomplots(i)%range)
        return
    end if    
  end subroutine do_slice 


  subroutine do_accumslice(n)
!
! for bomplot object n, insert array slice from %f3d into %data4d every time step until %nsecp1 steps, then dump to file and repeat.
! only do this if we are in the given bomplot object time interval given by %start and %end
!
    use state, only : istep, numstep, debug
    implicit none
    
    character(filenamesize) filename
    integer i,j,k,r(6),kdim,n
    real, allocatable :: gf(:,:,:)
    character(8) step,dim1,dim2,dim3,dim4,rkind
    
    r = bomplots(n)%range
    kdim = r(6)-r(5)+1
    
    allocate(gf(gim,gjm,kdim))

!TODO optimize this call to use a gather only on the covered blocks (use a local communicator via qmpi)
    call gatherfield(bomplots(n)%f3d(:,:,r(5):r(6)), gf, gim, gjm, kdim)

    if(bomplots(n)%n < bomplots(n)%nsecp1) then
        bomplots(n)%n = bomplots(n)%n + 1
        if(master) bomplots(n)%data4d(:,:,:,bomplots(n)%n) = gf(r(1):r(2),r(3):r(4),:)
    end if

    if(bomplots(n)%n == bomplots(n)%nsecp1 .or. istep == numstep) then
        if(master)then
            write(rkind,'("r",i1)') kind(bomplots(n)%data4d)
            write(dim1,'(i0)') size(bomplots(n)%data4d,1)
            write(dim2,'(i0)') size(bomplots(n)%data4d,2)
            write(dim3,'(i0)') size(bomplots(n)%data4d,3)
            write(dim4,'(i0)') bomplots(n)%n
            write(step,'(i8.8)') istep
            write(filename,'(13a)') trim(adjustl(bomplots(n)%label))//"-accstep"//step//"-"//trim(dim1)//"x"//trim(dim2)//"x"//&
                 trim(dim3)//"x"//trim(dim4)//".s"//rkind
            
            if(debug > 0) write(*,*) "Accumslice binary dump to ",trim(filename)
            
            open(99,file=trim(slice_dir)//trim(filename), access="stream", form="unformatted")
            write(99) bomplots(n)%data4d(:,:,:,1:bomplots(n)%n)
            close(99)
        end if
        bomplots(n)%n = 0
    end if
    deallocate(gf)
  end subroutine do_accumslice

  
  subroutine slice3d(label, field, range)
!    
!   if R=RANGE, dump FIELD(r(1):r(2),r(3):r(4),r(5):r(6)) to 
!     file=<label>-<idim>x<jdim>x<kdim>.data
!
! the slice routines are meant to be used for debugging mainly, so we
! just gather to master and dump there.
!     
    use state, only : istep
    implicit none
    character(*) label
    integer range(:), r(size(range))
    real field(1-ghost_wi:,1-ghost_wj:,:)
    character(filenamesize) datafile
    integer i,j,k
    real, allocatable :: gf(:,:,:)
    
    allocate(gf(gim,gjm,size(field,3)))
    call gatherfield(field, gf, gim, gjm, size(field,3))   
    if(master)then        
        r=range
        datafile=" "
        write(datafile,'(a,"-",i8.8,"-",i4.4,"x",i4.4,"x",i3.3,".data")') &
             label, istep, r(2)-r(1)+1,r(4)-r(3)+1,r(6)-r(5)+1
        
        open(unit=1000009, file=(trim(slice_dir)//trim(datafile)) )
        write(*,'(4a)') 'slice: dumping "',label,'" to ',trim(datafile)
        write(1000009,'(f0.7)') (((gf(i,j,k),i=r(1),r(2)),j=r(3),r(4)), k=r(5),r(6))    
        close(1000009)
    end if
    deallocate(gf)
  end subroutine slice3d


  subroutine slice2d(label, field, range)
!    
!   if R=RANGE, dump FIELD(r(1):r(2),r(3):r(4)) to 
!     file=<label>-<idim>x<jdim>.data
!    
    use state, only : istep
    implicit none
    character(*) label
    integer range(:), r(size(range))
    real field(1-ghost_wi:,1-ghost_wj:)
    character(filenamesize) datafile
    integer i,j
    real, allocatable :: gf(:,:)

    if(master)then        
        allocate(gf(gim,gjm))
        call gatherfield(field, gf, gim, gjm, 1)

        r=range
        datafile=" "
        write(datafile,'(a,"-",i8.8,"-",i4.4,"x",i4.4,".data")') &
             label, istep, r(2)-r(1)+1,r(4)-r(3)+1

        open(unit=1000009, file=(trim(slice_dir)//trim(datafile)) )
        write(*,'(4a)') 'slice: dumping "',label,'" to ',trim(datafile)
        write(1000009,'(f0.7)') ((gf(i,j),i=r(1),r(2)),j=r(3),r(4))
        close(1000009)
    else
        allocate(gf(1,1))
        call gatherfield(field, gf, 1, 1, 1)
    end if
    deallocate(gf)
  end subroutine slice2d  


  
  subroutine vsect(label,sx,sy,fullprec,field1,field2,pgplot)
! 
! Dump U V S and T for a series of sections given by the polygon with
! vertices in x,y.  also dump the "static" coordinates for these
! points to a file - where static means to neglect the elevation wr
! depth in the computation of the depth of each point.
! 
!  example: a simple section between two points xa,ya and xb,yb:
!     call vsect('sec1', (/xa,xb/), (/ya,yb/))
!  example2: assemble the subsections along the 4 pt polygon into one
!  section:
!     trasex=(/ 10., 40., 40.0, 100. /)
!     trasey=(/ 10., 10., 40.0, 100. /)
!     call vsect('sectiona',trasex,trasey,4)
                                !
! if the optional, real, 3d arrays field1 and/or field2 is present,
! only dump the cross section of those variables.
                                !
! if pgplot is true, save into preallocated pgplot vsect array.
!    
    use state
    use control_parameters
    implicit none

    character(len=*), intent(in) :: label
    real, intent(in), dimension(:) :: sx,sy
    logical, optional, intent(in) :: fullprec, pgplot
    real, optional :: field1(1-ghost_wi:,1-ghost_wj:,:), field2(1-ghost_wi:,1-ghost_wj:,:)    

    character(filenamesize) :: datafile, gridfile
    integer :: maxsec, totsec
    real, dimension(:),allocatable :: secdep, secpos, seclen
    real, dimension(:,:),allocatable :: usec,utsec,wsec,ssec,tsec,rhosec

    logical file_is_there, compress
    integer, parameter :: dump_uvst=0, dump_f1=1, dump_f2=2, dump_f1f2=3
    integer :: dump_what

    real, dimension(im+jm) :: x,y,enx,eny,etx,ety   
    integer k,i,n, npsec,ps1(numproc),ps2(numproc),psproc(numproc),psec
    logical no_pg,ib(im+jm)
    
!...............

    no_pg=.true.
    if(present(pgplot))then
        no_pg = .not. pgplot
    end if

    n=size(sx)
    if(size(sy).ne.n) then
        write(*,*)'bomplot.f90: vsect error: x and y are not of equal length'
        return
    end if

    if(present(fullprec)) then
        compress = .not. fullprec
    else
        compress = bomplot_compress
    end if

    dump_what=dump_uvst
    if( present(field1) .and. .not.present(field2) ) dump_what=dump_f1
    if( present(field2) .and. .not.present(field1) ) dump_what=dump_f2
    if( present(field1) .and. present(Field2) ) dump_what=dump_f1f2

    maxsec=im+jm
    
    allocate( &
         secpos(maxsec), &
         seclen(maxsec), &
         secdep(maxsec), &
         usec(maxsec,kb), &
         utsec(maxsec,kb), &
         wsec(maxsec,kb), &
         ssec(maxsec,kb), &
         tsec(maxsec,kb), &
         rhosec(maxsec,kb))

! convert the global section sx,sy into local point lists x,y, with
! corresponding unit normal and tangential vectors en and et, and
! section length seclen. n is the number of sections/points.
    
    call walkngetpts(sx,sy, x,y,enx,eny,etx,ety,secpos,seclen,n, npsec,ps1,ps2,psproc,ib)
    pgplot_current_totsec = n
    totsec=n ! only temporary for 1proc
    
    do psec=1,npsec
        if(psproc(psec) == myid) then
            select case(dump_what)
            case(dump_f1)
                call sect(ps1(psec),ps2(psec),x, y, enx, eny, etx, ety, usec, utsec, wsec, ssec, tsec, secdep, s=field1)
            case(dump_f2)
                call sect(ps1(psec),ps2(psec), x, y, enx, eny, etx, ety, usec, utsec, wsec, ssec, tsec, secdep, t=field2)
            case(dump_f1f2)
                call sect(ps1(psec),ps2(psec), x, y, enx, eny, etx, ety, usec, utsec, wsec, ssec, tsec, secdep, s=field1, t=field2)
            case default
                if(advect_density)then
                    call sect(ps1(psec),ps2(psec), x, y, enx, eny, etx, ety, usec, utsec, wsec, ssec, tsec, secdep, u,v,wr,s=rho)
                else
                    call sect(ps1(psec),ps2(psec), x, y, enx, eny, etx, ety, usec, utsec, wsec, ssec, tsec, secdep, u,v,wr,s,t)
                end if
            end select
            
            if(no_pg) then
                write(gridfile,'(a,"-",i3.3,"-",i5.5,"x",i4.4,"-grid.data")') label, psec, ps2(psec)-ps1(psec)+1, kb-1
                inquire( file=trim(vsect_dir)//trim(gridfile), exist=file_is_there )                
                if(.not.file_is_there) call dump_grid()
                
                write(datafile,'(a,"-",i3.3,"-",i8.8,"-",i5.5,"x",i3.3,".data")') &
                     label, psec, istep, ps2(psec)-ps1(psec)+1, kb-1
                if(master) write(*,'(4a)') 'vsect: dumping field "',label,'" to ',trim(datafile)
                
                open(unit=1000009, file=(trim(vsect_dir)//trim(datafile)) ,recl=200)
                if(compress)then
                    call dump_data_compressed()
                else
                    call dump_data_fullprec()
                end if
                close(1000009)
            else
                select case(dump_what)
                case(dump_f1)
                    pg_ssec(ps1(psec):ps2(psec),:)  = ssec(ps1(psec):ps2(psec),:)
                case(dump_uvst)
                    pg_usec(ps1(psec):ps2(psec),:)  = usec(ps1(psec):ps2(psec),:)
                    pg_utsec(ps1(psec):ps2(psec),:)  = utsec(ps1(psec):ps2(psec),:)
                end select
                
                if(myid > 0) then
                    call isend(secdep(ps1(psec):ps2(psec)), 0, tag=4)
              
                    select case(dump_what)
                    case(dump_f1)
                        call isend(pg_ssec(ps1(psec):ps2(psec),:), 0, tag=1)
                    case(dump_uvst)
                        call isend(pg_usec(ps1(psec):ps2(psec),:),  0, tag=2)
                        call isend(pg_utsec(ps1(psec):ps2(psec),:), 0, tag=3)
                    end select
                end if
            end if
        end if
        
        if(.not.no_pg .and. master .and. numproc>1 .and. psproc(psec)>0)then
            call receive(secdep(ps1(psec):ps2(psec)), psproc(psec), tag=4)

            select case(dump_what)
            case(dump_f1)
                call receive(pg_ssec(ps1(psec):ps2(psec),:), psproc(psec), tag=1)                
            case(dump_uvst)
                call receive(pg_usec(ps1(psec):ps2(psec),:), psproc(psec), tag=2)
                call receive(pg_utsec(ps1(psec):ps2(psec),:), psproc(psec), tag=3)
            end select
        end if
    end do
    
    if(.not.no_pg)then
        if(numproc > 1) call waitall()

        if(master)then
            do psec=1,npsec             
                do k=1,kb-1
                    do i=ps1(psec),ps2(psec)
                        pg_vx(i,k) = secpos(i)*dx
                        pg_vz(i,k) = zz(k)*secdep(i)
                    end do
                end do
            end do
        end if
    end if
    
    deallocate(secpos, seclen, secdep)
    deallocate(usec,utsec,wsec,ssec,tsec,rhosec)

  contains
    
    subroutine dump_grid()
      !! dump coordinates for current section psec. note that k is
      !! written before i, so that files from the different
      !! sections can be catted toghether into a complete file.
      !! if in pgplot mode, and numproc>1 set up isends and irecvs for secdep.
      implicit none
      
      open(unit=1000009, file=(trim(vsect_dir)//trim(gridfile)), recl=200)
      if(compress) then
          write(1000009,'(2(e12.5,1x))') &
               ((secpos(i), zz(k)*secdep(i),k=1,kb-1),i=ps1(psec),ps2(psec))
      else
          do i=ps1(psec),ps2(psec)
              do k=1,kb-1
                  write(1000009,*) secpos(i), zz(k)*secdep(i)
              end do
          end do
      end if
      close(1000009)
    end subroutine dump_grid

    
    subroutine dump_data_compressed()      
!! note that k is innermost in order to enable catting of files together into a complete 2d field.
      select case(dump_what)
      case(dump_uvst)
          if(.not.advect_density) then
              write(1000009,'((2(f0.4,1x),e10.3,1x,2(f0.4,1x)))') &
                   ((usec(i,k),utsec(i,k),wsec(i,k),ssec(i,k),tsec(i,k), i=ps1(psec),ps2(psec)),k=1,kb-1)
          else ! s is rho
              write(1000009,'((2(f0.4,1x),e10.3,1x,2(f0.4,1x)))') &
                   ((usec(i,k),utsec(i,k),wsec(i,k),ssec(i,k)+rhoaver, i=ps1(psec),ps2(psec)),k=1,kb-1)
          end if
      case(dump_f2)
          write(1000009,'(f0.4)') ((tsec(i,k), k=1,kb-1), i=ps1(psec),ps2(psec))
      case(dump_f1)
          write(1000009,'(f0.4)') ((ssec(i,k), k=1,kb-1), i=ps1(psec),ps2(psec))
      case(dump_f1f2)
          write(1000009,'((2(f0.4,1x)))') ((ssec(i,k),tsec(i,k), k=1,kb-1), i=ps1(psec),ps2(psec))
      end select
    end subroutine dump_data_compressed

    
    subroutine dump_data_fullprec()
        character(130) listiostring   ! large enough to keep 5 double prec. list io numbers from pgf90.
        select case(dump_what)
        case(dump_uvst)
            do i=ps1(psec),ps2(psec)
                do k=1,kb-1
                    if(.not.advect_density)then
                        write(listiostring,*) usec(i,k),utsec(i,k),wsec(i,k),ssec(i,k),tsec(i,k)
                    else
                        write(listiostring,*) usec(i,k),utsec(i,k),wsec(i,k),ssec(i,k)+rhoaver
                    end if                        
                    write(1000009,'(a)') trim(listiostring)
                end do
            end do
        case(dump_f1)
            do i=ps1(psec),ps2(psec)
                do k=1,kb-1
                    write(1000009,*) ssec(i,k)
                end do
            end do
        case(dump_f2)
            do i=ps1(psec),ps2(psec)
                do k=1,kb-1
                    write(1000009,*) tsec(i,k)
                end do
            end do
        case(dump_f1f2)
            do i=ps1(psec),ps2(psec)
                do k=1,kb-1
                    write(1000009,*) ssec(i,k), tsec(i,k)
                end do
            end do
        end select
    end subroutine dump_data_fullprec
    
  end subroutine vsect



  subroutine walkngetpts(sx,sy,x,y,enx,eny,etx,ety,s,seglen,n,sections,section_start,section_end,section_owner,inblock)
!
! discretize the linear sections between (global) points in
! SX(:),SY(:) into short line segments. N is the number of segments
! found. X(N),Y(N) is the mid-segment coordinate of each segment, with
! corresponding unit normal and tangential vectors ENX(N),ENY(N) and
! ETX(N),ETY(N), and segment length SEGLEN(N). SECTIONS is the number
! of processors the range of segments cross, SECTION_START(SECTIONS),
! SECTION_END(SECTIONS) are the start and end segment of each
! processor/block crossed. SECTION_OWNER(SECTIONS) is the processor id
! of each of these sub ranges.
!
    use state, only : debug,jm,dx
    use mod_mpi, only : bltab,i_blocksize,j_blocksize
    implicit none
    real, dimension(:) :: sx,sy, x,y,enx,eny,etx,ety,seglen,s
    integer n,sections,section_start(:),section_end(:),section_owner(:)

    integer nsegpts, spt, pt, lpt
    real deltax, deltay, cslen, ds
    integer, dimension(size(sx)) :: npts
    real, dimension(size(sx)) ::  cetx,cety,cenx,ceny,segpos
    logical inblock(:)
    integer prevbli,prevblj,bli,blj
    real tmpx,tmpy

    if(jm > 1) then    
        nsegpts = size(sx)
        segpos(1) = 0.0

! find the lengths and directions of the long linear segments given by
! the user.
    
        do spt=1,nsegpts-1
            deltax = sx(spt+1)-sx(spt)
            deltay = sy(spt+1)-sy(spt)
            cslen  = sqrt(deltax**2 + deltay**2)
            npts(spt) = nint(cslen)
            if(npts(spt) == 0) npts(spt)=1
            
            segpos(spt+1) = segpos(spt) + cslen
            
            cetx(spt) = deltax/cslen
            cety(spt) = deltay/cslen
            cenx(spt) =-cety(spt)
            ceny(spt) = cetx(spt)
        end do

! now discretizise along the long segments: npts above is chosen so
! that the distance between the points should be close to 1, the first
! and last point along a long segment match perfectly.
    
        s(1) = 0.0
        x(1) = sx(1)
        y(1) = sy(1)
        pt = 0          ! grid point index
    
        do spt=1, nsegpts-1
            ds = (segpos(spt+1)-segpos(spt))/(npts(spt)-1)
            
            do lpt=1, npts(spt)-1
                pt = pt + 1
                
                s(pt+1) = s(pt) + ds
                x(pt+1) = x(pt) + ds*cetx(spt)
                y(pt+1) = y(pt) + ds*cety(spt)
                
                seglen(pt) = ds
                
                etx(pt) = cetx(spt)
                ety(pt) = cety(spt)
                enx(pt) = cenx(spt)
                eny(pt) = ceny(spt)
            end do
        end do
    
        n = pt
    
! "finally", find the mid point coordinates that will be used for the
! interpolation

        do pt=1, n
            x(pt) = 0.5*(x(pt) + x(pt+1))
            y(pt) = 0.5*(y(pt) + y(pt+1))
            s(pt) = 0.5*(s(pt) + s(pt+1))
        end do
        
    else
        !! jm == 1:
        !! only need to return s,seglen,n, npsec,ps1,ps2,psproc

        ! end cell - start cell + 1
        n = nint(sx(2)+0.5) - nint(sx(1)+0.5) + 1
        seglen(:) = 1.0
        do pt=1, n
            s(pt) = (pt-1)                              ! s is 0 at the beginning of the section.
            x(pt) = (nint(sx(1)+0.5) + pt - 1.0 - 0.5)  ! used to get i, 
            y(pt) = 0.0
        end do
    end if

        
! now walk along the global list of points, and each time we cross a
! processor boundary, record the start and end indices for each
! processor point range.
    
    if(numproc > 1) then
        sections=1
        section_start(1)=1
        
        prevbli = (nint(x(1)+0.5) - 1)/i_blocksize + 1
        prevblj = (nint(y(1)+0.5) - 1)/j_blocksize + 1   
        ! the above is not bullet proof. if y is 0, and we add 0.5, is
        ! it 0.499999999 or 0.5000000001? override to be sure:
        if(jm == 1) prevblj = 1
        section_owner(sections) = bltab(prevbli,prevblj)%prn
        
        do pt=1,n
            bli = (nint(x(pt)+0.5) - 1)/i_blocksize + 1
            blj = (nint(y(pt)+0.5) - 1)/j_blocksize + 1
            if(jm == 1) blj = 1
            
            if(any((/bli,blj/) /= (/prevbli,prevblj/))) then
                !! we have crossed a processor boundary
                sections = sections + 1
                section_owner(sections) = bltab(bli,blj)%prn
                section_start(sections) = pt
                section_end(sections-1) = pt-1
            end if
            prevbli = bli
            prevblj = blj
        end do

        section_end(sections) = n
    else
        section_start(1) = 1
        section_end(1) = n
        section_owner(1) = myid
        sections = 1
    end if

! convert global point coordinates to local block coordinates

    if(numproc > 1) then
        do pt=1, n
            tmpx = x(pt)
            tmpy = y(pt)
            call globalcoo2local(tmpx,tmpy,inblock(pt))
            if(inblock(pt))then
                x(pt) = tmpx
                y(pt) = tmpy
            end if
        end do
    end if       
  end subroutine walkngetpts
  

  subroutine hsect(label, field, position, level, range, fullprec, output, spv, ru, rv, rw, rs, rt, pos)
!  
! Quick dump horisontal fields for S,T,U and V to file constructed from label
! LABEL at POSITION. if POSITION>0 the field position is measured POSITION
! meters below z=0 level, if POSITION<0 the field is measured
! ABS(POSITION) meters above bottom
! 
! usage: 50 meters below zero:
!  call HSECT('50mbz', 50.0 )
! will generate file: 50mbz-ISTEP-IMxJM.data
!
! if RANGE =(is,ie,js,je) is present, data will only be dumped for the horizontal
! index range f(is:ie,js:je).    
! if FULLPREC is present, and true, ascii dumps will be done with full real precision.   
! OUTPUT is an optional string defining how to do output. if not
! present hsect dumps to file. if OUTPUT="pgplot" hsect dumps to
! pgplot arrays. if OUTPUT="none" no output is done.
! if SPV is present, the value will be assigned to land points.    

    use state
    use util, only : binsearch
    use control_parameters
    implicit none
    character(len=*), intent(in) :: label
    integer, intent(in), optional :: range(4), level
    real, intent(in), optional :: field(1-ghost_wi:,1-ghost_wj:,:), position
    logical, intent(in), optional :: fullprec
    character(*), optional :: output
    real, optional :: spv
    real, optional, dimension(i1:i2,j1:j2) :: ru, rv, rw, rs, rt ! optionally copy results in here
    character(1), optional :: pos
    real landspv
    logical compress, nodump
    integer i,j,k,is,ie,js,je,gis,gie,gjs,gje
    real, dimension(i1:i2,j1:j2) :: uz,vz,wz,sz,tz
    real, dimension(kb-1) :: zpos

    integer iincr,jincr,r(4),gr(4), ipos
    real depth, lambda
    logical single_field, from_pgplot, inblock

    if(present(field))then
        single_field=.true.
    else
        single_field=.false.
    end if

    ipos=1 ! center
    if(present(pos))then
        select case(pos)
        case("u")
            ipos=2
        case("v")
            ipos=3
        end select
    end if
    
    from_pgplot=.false.
    nodump=.false.
    if(present(output))then
        if(output=="pgplot") from_pgplot=.true.
        if(output=="none") nodump=.true.
    end if
    
    if(present(fullprec)) then
        compress = .not. fullprec
    else
        compress = bomplot_compress
    end if
    
    landspv = -pgundef
    if(present(spv)) landspv = spv
    
    if( present(position).and.present(level))then
        if(master) write(*,*) 'hsect error (bomplot.f90): &
             &cannot dump both at level and position for label ',label
        return
    end if

    if(.not.present(position).and..not.present(level))then
        if(master)write(*,*) 'hsect error (bomplot.f90): &
             &you did not specify position or level at which to dump ',label
        return
    end if

    if(.not.present(range))then
        gis=1 ; gie=im ; gjs=1 ; gje=jm
    else
        gis=range(1) ; gie=range(2) ; gjs=range(3) ; gje=range(4)
    end if
    call globalrect2local(gis,gie,gjs,gje,is,ie,js,je,inblock)
    gr = (/gis,gie,gjs,gje/)
    r = (/is,ie,js,je/)

! get the data

    if (present(position) .and. inblock) then        
        do j=js,je
            do i=is,ie
                ! default value is land
                if (.not. single_field)then
                    uz(i,j) = landspv
                    vz(i,j) = landspv
                    wz(i,j) = landspv
                    sz(i,j) = landspv
                end if
                tz(i,j) = landspv
                
                if(h(i,j) <= 0.0) cycle ! nothing to do for land cells
                                
                if (position < 0) then
                    depth = (d(i,j) - (-position))
                else
                    depth = position + eta(i,j)
                end if

                if(depth > d(i,j) .or. depth < 0.0) cycle ! point not in water column.
                
                zpos(:) = -zz(1:kb-1)*d(i,j)
                
                k = binsearch(depth, zpos)
                
                if (k > 1 .and. k < kb) then
                    lambda = (depth - zpos(k-1))/(zpos(k) - zpos(k-1))
                else if (k == kb) then
                    k = kb-1 ; lambda = 1.0  ! return bottom layer value
                else if (k == 0) then        ! return surface layer
                    k = 2 ; lambda = 0.0
                else
                    call abort("bomplot hsect: should not happen")
                end if

                iincr=1 ; if(i == im) iincr = 0
                jincr=1 ; if(j == jm .or. jm == 1) jincr = 0
                
                if (.not. single_field)then
                    uz(i,j) = lambda*0.5*(u(i,j,k)  + u(i+iincr,j,k)) &
                         + (1-lambda)*0.5*(u(i,j,k-1)+ u(i+iincr,j,k-1))
                    vz(i,j) = lambda*0.5*(v(i,j,k)  + v(i,j+jincr,k)) &
                         + (1-lambda)*0.5*(v(i,j,k-1)+ v(i,j+jincr,k-1))
                    wz(i,j) = lambda*wr(i,j,k) + (1-lambda)*wr(i,j,k-1)
                    if(advect_density) then
                        sz(i,j) = lambda*rho(i,j,k)  + (1-lambda)*rho(i,j,k-1) + rhoaver + 1000.
                    else
                        sz(i,j) = lambda*s(i,j,k)  + (1-lambda)*s(i,j,k-1)
                        tz(i,j) = lambda*t(i,j,k)  + (1-lambda)*t(i,j,k-1)
                    end if
                else
                    select case(ipos)
                    case(1)
                        tz(i,j) = lambda*field(i,j,k)  + (1-lambda)*field(i,j,k-1)
                    case(2)
                        tz(i,j) = lambda*0.5*(field(i,j,k)  + field(i+iincr,j,k)) &
                             + (1-lambda)*0.5*(field(i,j,k-1)+ field(i+iincr,j,k-1))
                    case(3)
                        tz(i,j) = lambda*0.5*(field(i,j,k)  + field(i,j+jincr,k)) &
                             + (1-lambda)*0.5*(field(i,j,k-1)+ field(i,j+jincr,k-1))
                    end select
                end if
            end do
        end do
    end if

    if( present(level))then
        if(single_field)then
            tz = field(:,:,level)
        else
            uz = u(:,:,level)
            vz = v(:,:,level)
            wz = wr(:,:,level)
            if(advect_density)then
                sz = rho(:,:,level)+rhoaver+1000.
            else
                sz = s(:,:,level)
                tz = t(:,:,level)
            end if
        end if
    end if

    if(present(ru)) ru = uz
    if(present(rv)) rv = vz
    if(present(rw)) rw = wz
    if(present(rs)) rs = sz
    if(present(rt)) rt = tz

    if(nodump) return
    
    select case(bomplot_dumping_method)
    case(0)
        if(single_field) then
            if(.not.from_pgplot)then
                call dump2d(label, gr,tz,comp=compress)
            else
                call dumppg(tz, spv=landspv)
            end if
        else
            if(.not.from_pgplot)then
                if(advect_density)then
                    call dump2d(label, gr,uz,vz,wz,sz,comp=compress)
                else
                    call dump2d(label, gr,uz,vz,wz,sz,tz,comp=compress)
                end if
            else
                call dumppg(f2=uz, f3=vz, spv=landspv)
            end if
        end if
    case(1)
        if(single_field) then
            call dump2d(label, r,tz, comp=compress, parallel=.true., inblock=inblock)
        else
            if(advect_density)then
                call dump2d(label, r,uz,vz,wz,sz,comp=compress, parallel=.true., inblock=inblock)
            else
                call dump2d(label, r,uz,vz,wz,sz,tz,comp=compress, parallel=.true., inblock=inblock)
            end if
        end if

    end select
  end subroutine hsect

  
  subroutine dumppg(f1, f2, f3, spv)
!
! gather and store up to 3 2d distributed arrays in the pgplot global arrays.
! the three fields are always:
! if f1 is present, it is a scalar field. f2 and f3 can be omitted.
! if f2,f3 are present, they are u and v.
!    
    implicit none
    real, dimension(:,:), optional :: f1,f2,f3
    real, optional :: spv
      
! count the number of fields to dump, and allocate global space if we
! are going to gather.
    
    if(present(f1)) then
        call gatherfield(f1, pg_f, gim, gjm, 1, spv)
    end if
    if(present(f2)) then
        call gatherfield(f2, pg_u, gim, gjm, 1, spv)
    end if
    if(present(f3)) then
        call gatherfield(f3, pg_v, gim, gjm, 1, spv)
    end if    
  end subroutine dumppg
  
  
  subroutine dump2d(label, r, f1, f2, f3, f4, f5, comp, prec, parallel,inblock)
!    
! dump up to 5 2d fields to ascii file(s) with given compression level/precision
!   if f3 is present, f1 and f2 must also be etc.
! if prec is present, it is the format of the floating point numbers, e.g. prec="f0.5"
!
      use state, only : im,jm,istep
      implicit none
      character(*) label
      integer r(4)
      real, dimension(:,:), optional :: f1,f2,f3,f4,f5
      logical, optional :: comp, parallel, inblock
      character(*), optional :: prec
      character(20) format
      character(filenamesize) datafile
      logical compress, par,file_open, dump, inbl
      integer narr,is,ie,js,je,i,j
      real, dimension(:,:,:), allocatable :: gf
      character(4) :: nan(6)=(/"nan ","nan ","nan ","nan ","nan ","nan "/)
      character(5*26) listiostring
      
      is=r(1)
      ie=r(2)
      js=r(3)
      je=r(4)
      
      compress=.false.
      if(present(comp))then
          if(comp)compress=comp
      end if
      par=.false.
      if(present(parallel)) then
          if(parallel) par=parallel
      end if
      inbl=.false.
      if(present(inblock))then
          inbl=inblock
      end if
      
! count the number of fields to dump, and allocate global space if we
! are going to gather.
      if(.not.par) then
          allocate(gf(im,jm,5))
      else
          allocate(gf(i1:i2,j1:j2,5))
      end if
      
      narr=0
      if(present(f1)) then
          narr=narr+1
          if(.not.par) then
              call gatherfield(f1, gf(:,:,narr), im, jm, 1)
          else
              gf(:,:,1) = f1
          end if
      end if
      if(present(f2)) then
          narr=narr+1 
          if(.not.par) then
              call gatherfield(f2, gf(:,:,narr), im, jm, 1)
          else
              gf(:,:,2) = f2
          end if
      end if
      if(present(f3)) then
          narr=narr+1
          if(.not.par) then
              call gatherfield(f3, gf(:,:,narr), im, jm, 1)
          else
              gf(:,:,3) = f3
          end if
      end if
      if(present(f4)) then
          narr=narr+1
          if(.not.par) then
              call gatherfield(f4, gf(:,:,narr), im, jm, 1)
          else
              gf(:,:,4) = f4
          end if
      end if
      if(present(f5)) then
          narr=narr+1
          if(.not.par) then
              call gatherfield(f5, gf(:,:,narr), im, jm, 1)
          else
              gf(:,:,5) = f5
          end if
      end if


! find local file name, and open file(s)

      select case(bomplot_dumping_method)
      case(0)
          if(master) then
              write(datafile,'(a,"-",i8.8,"-",i4.4,"x",i4.4,".data")') &
                   label, istep, ie-is+1,je-js+1
              open(unit=1000009, file=(trim(hsect_dir)//trim(datafile)), recl=200)
              write(*,'(4a)') 'hsect: dumping field "',label,'" to ',trim(datafile)
          end if
      case(1)
          if(inblock)then        
              write(datafile,'(a,"-",i2.2,"x",i2.2,"-",i8.8,"-",i4.4,"x",i4.4,".data")') &
                   label, block%ibn, block%jbn, istep, ie-is+1,je-js+1
              open(unit=1000009, file=(trim(hsect_dir)//trim(datafile)), recl=200)
              write(*,'(4a)') 'hsect: dumping field "',label,'" to ',trim(datafile)
          end if
      case default
          if(master) write(*,*)'bomplot dumping method=',bomplot_dumping_method,'not implemented'
          stop
      end select    

! now decide whether to dump or not:
!   if par=false dump only on master
!   if par=true and inblock, always dump.
      
      dump = (.not.par .and. master) .or. (par.and.inbl)

      if(dump)then
          if(compress)then                 
              if(present(prec))then
                  write(format,"('(',i1,'(',a,',1x))')") narr, prec
              else
                  select case(narr)
                  case(5) 
                      format = '(5(f0.6,1x))'
                  case(1) 
                      format = '(f0.6))'
                  case default
                      call message("narr unknown", 2, ival=(/narr/))
                  end select
              end if
              write(1000009,trim(format)) ((gf(i,j,1:narr), i=r(1),r(2)),j=r(3),r(4))
          else
              do j=r(3),r(4)
                  do i=r(1),r(2)
                      if(bomplot_gnuplot)then
                          if(gridbox(i,j)==1)then
                              write(listiostring,*) gf(i,j,1:narr)
                          else
                              write(listiostring,*) nan(1:narr)
                          end if                              
                      else
                          write(listiostring,*) gf(i,j,1:narr)                          
                      end if
                      write(1000009,'(a)') trim(listiostring)
                  end do
                  if(bomplot_gnuplot) write(1000009,*)
              end do
          end if
      end if

      inquire(unit=1000009, opened=file_open)
      if(file_open) close(1000009)

      if(allocated(gf)) deallocate(gf)
    end subroutine dump2d


    subroutine bindump2d(label, f, range)
      use state, only : im,jm,istep,debug
      implicit none
      character(*) label
      real f(i1:,j1:)
      real, allocatable :: gf(:,:)
      integer, optional :: range(4)
      integer r(4)
      character(4) dim1,dim2
      character(2) rkind
      character(8) step
      character(200) filename

      if(present(range)) then
         r = range
      else
         r = (/1,im,1,jm/)
      end if

      allocate(gf(im,jm))
      call gatherfield(f, gf, im, jm, 1)

      if(master)then
         write(rkind,'("r",i1)') kind(f)
         write(dim1,'(i0)') r(2)-r(1)+1
         write(dim2,'(i0)') r(4)-r(3)+1
         write(step,'(i8.8)') istep
         write(filename,'(9a)') label//"-"//step//"-"//trim(dim1)//"x"//trim(dim2)//".s"//rkind
         if(debug > 0) write(*,*) "Binary dump to ",trim(filename)
         open(99,file=filename, access="stream", form="unformatted")
         write(99) gf(r(1):r(2),r(3):r(4))
         close(99)
      end if
      deallocate(gf)
    end subroutine bindump2d

    subroutine bindump3d(label,f,range)
      use state, only : im,jm,kb,istep,debug
      implicit none
      character(*) label
      real f(i1:,j1:,:)
      real, allocatable :: gf(:,:,:)
      integer, optional :: range(6)
      integer r(6)
      character(4) dim1,dim2,dim3
      character(2) rkind
      character(8) step
      character(200) filename
      integer kdim

      kdim = size(f,3)
      if(present(range)) then
         r = range
      else
         r = (/1,im,1,jm,1,kdim/)
      end if

      allocate(gf(im,jm,kdim))
      call gatherfield(f, gf, im, jm, kdim)

      if(master)then
         write(rkind,'("r",i1)') kind(f)
         write(dim1,'(i0)') r(2)-r(1)+1
         write(dim2,'(i0)') r(4)-r(3)+1
         write(dim3,'(i0)') r(6)-r(5)+1
         write(step,'(i8.8)') istep
         write(filename,'(11a)') label//"-"//step//"-"//trim(dim1)//"x"//trim(dim2)//"x"//trim(dim3)//".s"//rkind
         if(debug > 0) write(*,*) "Binary dump to ",trim(filename)
         open(99,file=filename, access="stream", form="unformatted")
         write(99) gf(r(1):r(2),r(3):r(4),r(5):r(6))
         close(99)
      end if
      deallocate(gf)
    end subroutine bindump3d


    
    subroutine hsect2d(label,field,range,fullprec)
!  
! quick dump horisontal fields for eta, ua, va to file constructed from name
! label
!    
      use state
      implicit none
      character(len=*), intent(in):: label
      integer, optional, intent(in) :: range(4)
      real, intent(in), optional :: field(i1:,j1:)
      logical, intent(in), optional :: fullprec
      logical compress, inblock
      integer is,ie,js,je,gjs,gje,gis,gie,gr(4),r(4)

      gis=1 ; gie=im ; gjs=1 ; gje=jm
      if(present(range))then
          gis=range(1) ; gie=range(2)
          gjs=range(3) ; gje=range(4)
      end if
      call globalrect2local(gis,gie,gjs,gje,is,ie,js,je,inblock)
      gr=(/gis,gie,gjs,gje/)
      r=(/is,ie,js,je/)

      if(present(fullprec)) then
          compress=.not.fullprec
      else
          compress=bomplot_compress
      end if

      select case(bomplot_dumping_method)
      case(0)
          if(present(field))then
              call dump2d(label, gr, field, comp=compress, prec="f0.4")
          else
              call dump2d(label, gr, eta,ua,va, comp=compress, prec="f0.4")
          end if
      case(1)
          if(present(field))then
              call dump2d(label, r, field, comp=compress, prec="f0.4", parallel=.true., inblock=inblock)
          else
              call dump2d(label, r, eta,ua,va, comp=compress, prec="f0.4", parallel=.true., inblock=inblock)
          end if
      end select
    end subroutine hsect2d



    subroutine isosurf(label, field, isoval, range, fullprec, spv)
!
! isosurf finds the isosurface of ISOVAL(ue) given FIELD, as well as
! the velocity components U,V and WR on this surface. the datafile is
! constructed from LABEL and ISTEP. optionally a rectangular
! horisontal range may be specified in RANGE containing
! (/i1,i2,j1,j2/) FIELD is assumed to be given in S/T points.
!
! the isolevel is found by searching the water column from the top
!    
    use state
    implicit none

    character(len=*), intent(in) :: label
    real, intent(in) :: isoval, field(i1:,j1:,:)
    integer, intent(in), optional, dimension(4) :: range
    logical, intent(in), optional :: fullprec
    logical compress, inblock
    real, optional :: spv
    real landspv
    real, dimension(im,jm) :: isolev,ulev,vlev,wlev
    integer di,dj,i,j,k,is,ie,js,je,gis,gie,gjs,gje,gr(4),r(4)
    real dep, fracz, fdist
    real, parameter :: eps=0.001    
    
    gis=1 ; gie=im  ; gjs=1 ; gje=jm
    if( present(range) )then
        gis=range(1) ; gie=range(2)
        gjs=range(3) ; gje=range(4)
    end if
    call globalrect2local(gis,gie,gjs,gje,is,ie,js,je,inblock)

    gr=(/gis,gie,gjs,gje/)
    r=(/is,ie,js,je/)
    
    if(present(fullprec)) then
        compress=.not.fullprec
    else
        compress=bomplot_compress
    end if

    landspv = -pgundef
    if(present(spv)) landspv=spv
    

    do j=js,je
        do i=is,ie
            isolev(i,j) = landspv
            ulev(i,j) = landspv
            vlev(i,j) = landspv
            wlev(i,j) = landspv

            if (fsm(i,j) > 0.0) then

                do k = 1,kb-2
                    if( (isoval >= field(i,j,k) .and. &
                         isoval < field(i,j,k+1)) .OR. &
                         (isoval <= field(i,j,k) .and. &
                         isoval > field(i,j,k+1)) ) then

                        di=1 ;  dj=1
                        if(i==im) di=0
                        if(j==jm) dj=0
                        if(jm==1) dj=0

                        fdist = field(i,j,k+1) - field(i,j,k)

                        if( abs(fdist) > eps ) then

! the vertical fraction of the cell from isoval
! to the k level above

                            fracz= (isoval - field(i,j,k) )/fdist

! the sigma coordinate depth of isoval

                            dep = zz(k) - dzz(k)*fracz
                            ulev(i,j)=0.5*((1-fracz)*(u(i,j,k)+u(i+di,j,k)) &
                                 + fracz*(u(i,j,k+1) + u(i+di,j,k+1) ) )
                            vlev(i,j)=0.5*((1-fracz)*(v(i,j,k)+u(i,j+dj,k)) &
                                 + fracz*(v(i,j,k+1) + v(i,j+dj,k+1) ) )
                            wlev(i,j)=(1-fracz)*wr(i,j,k)+fracz*wr(i,j,k+1)
                        else

! levels are pretty close, so just get the average to
! avoid interpolation error (?)

                            dep = 0.5*(zz(k)+zz(k+1))
                            ulev(i,j) = 0.25*(u(i,j,k)+u(i+di,j,k)+&
                                 u(i,j,k+1)+u(i+di,j,k+1) )
                            vlev(i,j) = 0.25*(v(i,j,k)+v(i,j+dj,k)+&
                                 v(i,j,k+1)+v(i,j+dj,k+1) )
                            wlev(i,j) = 0.5*(wr(i,j,k)+wr(i,j,k+1))
                        end if
! store a positive isolevel
                        isolev(i,j) = -(dep*d(i,j)+eta(i,j))      
                        exit
                    end if
                end do
            end if

        end do
    end do
    
    select case(bomplot_dumping_method)
    case(0)
        call dump2d(label, gr, isolev, ulev, vlev, wlev, comp=compress, prec="f0.4")
    case(1)
        call dump2d(label, r, isolev, ulev, vlev, wlev, comp=compress, prec="f0.4", parallel=.true., inblock=inblock)
    end select
  end subroutine isosurf


  subroutine vprofz(fz, i,j, zdepths)
    use state
    use control_parameters, only : advect_density
    implicit none
    real fz(:,:), zdepths(:), sigdepths(kb)
    integer i,j,k,kz,di,dj
    real depth,lam,uz,vz,sz,tz,rhoz,wrz

    if(size(zdepths) .ne. size(fz,2)) call message("vprofz fz dim 2 must be equal to zdepths",2)
    if(size(fz,1) < 6) call message("vprofz: fz dim 1 must be at least 6",2)
    
    sigdepths = -zz(:)*d(i,j)

    di=1
    dj=1
    if(block%fi+i-1 == im) di=0
    if(block%fj+j-1 == im) dj=0
    
    k=1
    do kz=1,size(zdepths)
    
       ! distance from surf to z level in meters
        depth = zdepths(kz) + eta(i,j)

        if(depth > d(i,j) .or. depth < 0.0) then
            cycle
        end if
        
        if(depth < sigdepths(k))then
        ! this kz is before current sigma point. flag that or extrapolate upwards? then next kz
            fz(:,kz) = (/u(i,j,k),v(i,j,k),s(i,j,k),t(i,j,k),rho(i,j,k)+rhoaver/)
            cycle
        end if
        
        if(depth > sigdepths(kb-1))then
         ! kz after last sigma depth, extrapolate or just exit kz loop?
            fz(:,kz) = (/u(i,j,kb-1),v(i,j,kb-1),s(i,j,kb-1),t(i,j,kb-1),rho(i,j,kb-1)+rhoaver/)
            cycle
        end if
      
        if(depth > sigdepths(k+1)) then
        ! increment k until sigdepths surrounds depth
            do
                k=k+1
                if(k >= kb-1) exit
                if(depth > sigdepths(k) .and. depth <= sigdepths(k+1)) exit
            end do
        end if
        
        lam = 1 - (depth - sigdepths(k))/(sigdepths(k+1)-sigdepths(k))

        uz = lam*0.5*(u(i,j,k)+u(i+di,j,k)) + (1-lam)*0.5*(u(i,j,k+1)+u(i+di,j,k+1))
        vz = lam*0.5*(v(i,j,k)+v(i,j+dj,k)) + (1-lam)*0.5*(v(i,j,k+1)+v(i,j+dj,k+1))
        wrz = lam*wr(i,j,k) + (1-lam)*wr(i,j,k+1)
        if(.not.advect_density)then
            sz = lam*s(i,j,k) + (1-lam)*s(i,j,k+1)
            tz = lam*t(i,j,k) + (1-lam)*t(i,j,k+1)
        end if
        rhoz = lam*rho(i,j,k) + (1-lam)*rho(i,j,k+1)
        
        fz(:,kz) = (/uz,vz,wrz,sz,tz,rhoz+rhoaver/)
    end do   
  end subroutine vprofz
  

  subroutine vprof(label,x,y,field,fullprec)
!   
! dump u v wr s and t for a given horisontal position with arbitary label 
! "label" to a file, which name is constructed from this info: 
!  
!  example
!     call vprof('tp12', x,y )   =>   tp12-vpro-51-00100.data 
! 
    use state
    use control_parameters
    
    implicit none
    character(len=*), intent(in) :: label
    real, intent(in) :: x,y
    real, intent(in), optional :: field(1-ghost_wi:,1-ghost_wj:,:)
    logical, intent(in), optional :: fullprec
    logical compress,inblock
    character(filenamesize) :: datafile  
    integer :: i,j,k,jof
    character(6*26) listiostring
    
    jof=0
    if(jm>1) jof=1
    
    if(present(fullprec)) then
        compress=.not.fullprec
    else
        compress=bomplot_compress
    end if

    i=nint(x+0.5) ; j=nint(y+0.5)

    call globalind2local(i,j,inblock)
    
    write(datafile,'(a,"-vpro-",i3.3,"-",i8.8,".data")') label, kb-1, istep
    if(master)write(*,'(4a)') 'vprof: dumping "',label,'" to file: ',trim(datafile)

    if(inblock)then
        open(unit=1000009, file=(trim(vprof_dir)//trim(datafile)), recl=200)
        if(compress)then
            if(.not.present(field))then
                if(.not.advect_density)then
                    write(1000009,'(6(e13.6,1x))') (  -zz(k)*d(i,j)-eta(i,j),    &
                         0.5*(u(i,j,k)+u(i+1,j,k)), &
                         0.5*(v(i,j,k)+v(i,j+jof,k)), &
                         wr(i,j,k),                 &
                         s(i,j,k),                  &
                         t(i,j,k),                  &
                         k=1,kb-1)
                else
                    write(1000009,'(5(e13.6,1x))') (  -zz(k)*d(i,j)-eta(i,j),    &
                         0.5*(u(i,j,k)+u(i+1,j,k)), &
                         0.5*(v(i,j,k)+v(i,j+jof,k)), &
                         wr(i,j,k),                 &
                         rho(i,j,k)+rhoaver,       &
                         k=1,kb-1)
                end if
            else
                write(1000009,'(2(e13.6,1x))') (  -zz(k)*d(i,j)-eta(i,j),    &
                     field(i,j,k), &
                     k=1,kb-1)
            end if
        else
            if(.not.present(field))then
                if(.not.advect_density)then
                    do k=1,kb-1
                        write(listiostring,*) -zz(k)*d(i,j)-eta(i,j),    &
                             0.5*(u(i,j,k)+u(i+1,j,k)), &
                             0.5*(v(i,j,k)+v(i,j+jof,k)), &
                             wr(i,j,k),                 &
                             s(i,j,k),                  &
                             t(i,j,k)
                        write(1000009,'(a)') trim(listiostring)
                    end do
                else
                    do k=1,kb-1
                        write(listiostring,*) -zz(k)*d(i,j)-eta(i,j),    &
                             0.5*(u(i,j,k)+u(i+1,j,k)), &
                             0.5*(v(i,j,k)+v(i,j+jof,k)), &
                             wr(i,j,k),                 &
                             rho(i,j,k)+rhoaver
                        write(1000009,'(a)') trim(listiostring)
                    end do
                end if
            else
                do k=1,kb-1
                    write(1000009,*) -zz(k)*d(i,j)-eta(i,j),    &
                         field(i,j,k)
                end do
            end if
            
        end if
        close(1000009)
    end if
  end subroutine vprof



  subroutine tsvprof(label, ip,jp, field, fullprec)
! 
! append a vertical profile for horisontal indice ip,jp of u,v , s, t,...
! each time called to a file with name constructed from LABEL.
! the vertical profiles will be stored in batches of bomplot_vprof_interval 
! samples(profiles) in each file
!  
    use state
    use control_parameters
    implicit none
    character(len=*), intent(in):: label
    integer, intent(in):: ip,jp
    real, intent(in), optional :: field(1-ghost_wi:,1-ghost_wj:,:)
    logical, intent(in), optional :: fullprec
    logical compress,inblock,exist
    character(filenamesize) datafile
    real :: umid,vmid,speed, theta, pi,lon
    integer i,j,k,jof
    character(8*26) listiostring

    jof=0
    if(jm>1) jof=1
    
    if(present(fullprec)) then
        compress=.not.fullprec
    else
        compress=bomplot_compress
    end if

    pi = 4.*atan(1.0)
    i=ip ; j=jp
    
    call globalind2local(i,j,inblock,ghost=.false.)

    lon = 0.0 
    if(associated(longitude)) lon = longitude(i,j)
    
    if(.not.inblock) return
    
    datafile=' '
    write(datafile,'(2a,i3.3,a,i4.4,a)') &
         label,'-',kb,'-',istep/bomplot_vprof_interval,'.data'

    inquire(file=trim(vprof_dir)//datafile, exist=exist)

    if(exist)then
! searching for the end of the file for large files may be very time consuming. 
        open(1000009,file=(trim(vprof_dir)//datafile),status='old',position='append', recl=200)
    else
        open(1000009,file=trim(vprof_dir)//datafile, recl=200)
    end if
    if(debug > 2) then
        write(*,*) 'tsvprof: dumping timeseries of vprof ',label,'to file=',trim(vprof_dir)//trim(datafile)
    end if

    do k=1,kb-1
        if(.not.present(field))then
            umid = 0.5*(u(i,j,k)+u(i+1,j,k))
            if(jm > 1)then
                vmid = 0.5*(v(i,j,k)+v(i,j+jof,k))
            else
                vmid = v(i,j,k)
            end if
            speed = sqrt(umid**2 + vmid**2)    
            if (umid.ne.0. .or. vmid.ne.0.) then
                ! for a point on a pol.ster. grid this holds, theta is the current direction wrt north; 90=from west, 0=from south
                !    ylong-longitude + 90 + atan2(v,u) + theta = 180
                theta = 90 - atan2(vmid,umid)*180/pi + lon - ylong
                ! (close to ok for utm grids also...)
            else
                theta = 0.
            endif
            do
                if( theta<0 ) then
                    theta=theta + 360
                else
                    exit
                end if
            end do
            if(compress)then
                if(.not.advect_density)then
                    write(1000009,'(8(f0.4,1x))') zz(k)*h(i,j),umid,vmid,speed,theta,t(i,j,k),&
                         s(i,j,k),rho(i,j,k)+rhoaver
                else
                    write(1000009,'(6(f0.4,1x))') zz(k)*h(i,j),umid,vmid,speed,theta,rho(i,j,k)+rhoaver
                end if
            else
                if(.not.advect_density)then
                    write(listiostring,*) zz(k)*h(i,j),umid,vmid,speed,theta,t(i,j,k),&
                         s(i,j,k),rho(i,j,k)+rhoaver
                else
                    write(listiostring,*) zz(k)*h(i,j),umid,vmid,speed,theta,rho(i,j,k)+rhoaver
                end if
                write(1000009,'(a)') trim(listiostring)
            end if

        else
            if(compress)then
                write(1000009,'(2(f0.4,1x))') zz(k)*h(i,j),field(i,j,k)
            else
                write(1000009,*) zz(k)*h(i,j),field(i,j,k)
            end if

        end if

    end do
    close(1000009)
  end subroutine tsvprof
  
  subroutine find_ucoeffs(gx,gy,lx,ly,i,j,ui,uj,cu)
!    
! given a grid coordinate gx,gy, return interpolation coefficients in
!   u points, cu, so that if f us a u point array
!     F(gx,gy) = cu(1)*f(ui,uj)+cu(2)*f(ui+1,uj)+cu(3)*f(ui+1,uj+1)+cu(4)*f(ui,uj+1)
!    
    implicit none
    real gx,gy           ! global grid coordinate
    real lx,ly           ! distance from global point of interpolation gx,gy to closest sw cell center
    integer i,j          ! closest global cell center to the south west of gx,gy
    integer ui,uj        ! same as i,j above, but for u and v points respectively
    real cu(4)           ! interpolation weights
    real ulx,uly
    
    ui=int(gx+1.0)
    uj=j
    ulx=gx-(ui-1.0)
    uly=ly

    cu(1)=(1.-ulx)*(1.-uly)
    cu(2)=ulx*(1.-uly)
    cu(3)=ulx*uly
    cu(4)=(1.-ulx)*uly
  end subroutine find_ucoeffs


  subroutine find_vcoeffs(gx,gy, lx,ly, i,j, vi,vj,cv)
!    
! given a grid coordinate gx,gy, return interpolation coefficients in
!   v points, cv, so that if f us a v point array
!     F(gx,gy) = cv(1)*f(vi,vj)+cv(2)*f(vi+1,vj)+cv(3)*f(vi+1,vj+1)+cv(4)*f(vi,vj+1)
!    
    implicit none
    real gx,gy           ! global grid coordinate
    real lx,ly           ! distance from global point of interpolation gx,gy to closest sw cell center
    integer i,j          ! closest global cell center to the south west of gx,gy
    integer vi,vj        ! same as i,j above, but for u and v points respectively
    real cv(4)           ! interpolation weights
    real vlx,vly
    
    vi=i
    vj=int(gy+1.0)
    vlx=lx
    vly=gy-(vj-1.0)

    cv(1)=(1.-vlx)*(1.-vly)
    cv(2)=vlx*(1.-vly)
    cv(3)=vlx*vly
    cv(4)=(1.-vlx)*vly      
  end subroutine find_vcoeffs
  
  subroutine find_coeffs(lx,ly,i,j,ci,cj,c)
!      
! we are going to try to find the value of an array in the global
! point gx,gy via interpolation.  give the coordinate lx,ly which is
! the distance from gx,gy to the closest sw cell center. return interpolation
! weights c(4) so that if f is an array in RHO points, we can write
!   F(lx,ly) = c(1)*f(i,j) +c(2)*f(i+1,j) +c(3)*f(i+1,j+1) +c(4)*f(i,j+1)      
! if one or more of the 4 cells surrounding gx,gy is land, we need to
! fall back to simpler interpolation techniques: with 3 cells
! interpolation in a plane, 2 cells gives linear interpolation, 1cell
! gives extrapolation.
!      
      use state, only : h
      implicit none
      real lx,ly          ! distance from global point of interpolation gx,gy to closest sw cell center
      integer i,j         ! block local indices of closest sw cell
      integer ci,cj       ! block local indices of cell containing global point gx,gy
      real c(4)           ! interpolation weights
      integer ifsm(4),fsmsum

      !! default coeffs. u and v coeffs are valid also with mixed
      !! ocean/land pts
      
      c(:) = 0.0
      if(h(ci,cj) <= 0.0) return

      where((/h(i,j),h(i+1,j),h(i+1,j+1),h(i,j+1)/) <= 0.)
          ifsm = 0
      elsewhere
          ifsm = 1
      end where
      
      fsmsum = ifsm(1) + ifsm(2) + ifsm(3) + ifsm(4)
      
      select case(fsmsum)
      case(4)
          !! bilinear interpolation
          c(1) = (1.-lx)*(1.-ly)
          c(2) = lx*(1.-ly)
          c(3) = lx*ly
          c(4) = (1.-lx)*ly
      case(3)
          !! interpolation using plane defined by the 3 wet points

          if(ifsm(1) == 0) then
              c(2) = 1.-ly ; c(3)=lx+ly ; c(4)=-lx
          else if(ifsm(2) == 0) then
              c(1)=1.-ly ; c(3)=lx ; c(4)=ly-lx
          else if(ifsm(3) == 0) then
              c(1)=1-lx-ly ; c(2)=lx ; c(4)=ly
          else if(ifsm(4) == 0) then
              c(1)=1.-lx ; c(2)=lx-ly ; c(3)=ly
          end if
      case(2)
          !! linear interpolation using the 2 wet points

          if(ifsm(1) + ifsm(2) == 2) then
              c(1)=1.-lx ; c(2)=lx
          else if(ifsm(2) + ifsm(3) == 2) then
              c(2)=1.-ly ; c(3)=ly
          else if(ifsm(3) + ifsm(4) == 2) then
              c(3)=lx ; c(4)=1.0-lx
          else if(ifsm(1) + ifsm(4) == 2) then
              c(1)=1.-ly ; c(4)=ly
          else            
              !! f1 and f3 or f2 and f4 active: treat as single point case
              i=ci ; j=cj
              c(1)=1.0
          end if

      case(1)
          !! one active point? extrapolate
          i=ci ; j=cj
          c(1)=1.0

      case default
          write(*,*)'??? sum(fsm)=',fsmsum,' i,j=',i,j          
      end select
    end subroutine find_coeffs

    
  
  subroutine sect(n1, n2, x, y, enx, eny, etx, ety, unsec, utsec, &
       wsec, ssec, tsec, secdep, u,v,wr,s,t)
! 
! find the values of velocity, s and t under the horisontal range of
! positions [n1,n2] in vectors x,y by using bilinear interpolation
! along sigma layers.
!
    use state, only : h, i1,i2,j1,j2,jm,kb
    implicit none
    integer, intent(in) :: n1,n2
    real, intent(in), dimension(:) :: x,y,enx,eny,etx,ety
    real, dimension(:,:) :: unsec, utsec, wsec, ssec, tsec
    real, dimension(:) :: secdep
    real, dimension(i1:,j1:,:), intent(in), optional :: u,v,wr,s,t
    
    real gx,gy, um,vm
    integer i,j,k,ui,uj, vi,vj, pt, ci, cj
    real, dimension(4) :: c,cu,cv
    real lx,ly
    logical u_present, v_present, wr_present, s_present, t_present

    
    u_present=.false.
    if(present(u)) u_present=.true.
    v_present=.false.
    if(present(v)) v_present=.true.
    wr_present=.false.
    if(present(wr)) wr_present=.true.
    s_present=.false.
    if(present(s)) s_present=.true.
    t_present=.false.
    if(present(t)) t_present=.true.   

    if(jm == 1) then
        call getslice()
        return
    end if

    do pt=n1,n2
        gx= x(pt)
        gy= y(pt)

        !! closest cell south west
        i=nint(gx)
        j=nint(gy)

        !! cell of coordinate
        ci=nint(gx+0.5)
        cj=nint(gy+0.5)

        lx=gx-(i-0.5)
        ly=gy-(j-0.5)

        if(present(u)) call find_ucoeffs(gx,gy, lx,ly, i,j, ui,uj, cu)
        if(present(v)) call find_vcoeffs(gx,gy, lx,ly, i,j, vi,vj, cv)
        call find_coeffs(lx,ly, i,j, ci,cj, c)

        secdep(pt) = c(1)*h(i,j) + c(2)*h(i+1,j) + c(3)*h(i+1,j+1) + c(4)*h(i,j+1)

        if(present(u) .and. present(v)) then
            do k=1,kb-1
                um=cu(1)*u(ui,uj,k)+cu(2)*u(ui+1,uj,k)+cu(3)*u(ui+1,uj+1,k)+cu(4)*u(ui,uj+1,k)
                vm=cv(1)*v(vi,vj,k)+cv(2)*v(vi+1,vj,k)+cv(3)*v(vi+1,vj+1,k)+cv(4)*v(vi,vj+1,k)
                unsec(pt,k) = um*enx(pt) + vm*eny(pt)
                utsec(pt,k) = um*etx(pt) + vm*ety(pt)
            end do
        end if
        if(present(s)) then
            do k=1,kb-1
                ssec(pt,k)=  c(1)*s(i,j,k)  +c(2)*s(i+1,j,k)  +c(3)*s(i+1,j+1,k)  +c(4)*s(i,j+1,k)
            end do
        end if
        if(present(t)) then
            do k=1,kb-1
                tsec(pt,k)=  c(1)*t(i,j,k)  +c(2)*t(i+1,j,k)  +c(3)*t(i+1,j+1,k)  +c(4)*t(i,j+1,k)
            end do
        end if
        if(present(wr)) then
            do k=1,kb-1
                wsec(pt,k)=  c(1)*wr(i,j,k) +c(2)*wr(i+1,j,k) +c(3)*wr(i+1,j+1,k) +c(4)*wr(i,j+1,k)
            end do
        end if
    end do

  contains
    
    
    subroutine getslice()
!      
! if jm=1, we only need to return the points between nint(xa+0.5) and
! nint(xb+0.5) along the slice, store with processor number to get
! ordering right.
!      
     implicit none
     integer i,j,pt

     j=1
     do pt = n1,n2
         i = nint(x(pt)+0.5)
         
         secdep(pt) = h(i,j)
         
         if(u_present .and. v_present) then
             do k=1,kb-1
                 unsec(pt,k) = v(i,j,k)
                 utsec(pt,k) = 0.5*(u(i,j,k)+u(i+1,j,k))
             end do
         end if
         if(s_present) then
             do k=1,kb-1
                 ssec(pt,k)= s(i,j,k)
             end do
         end if
         if(t_present) then
             do k=1,kb-1
                 tsec(pt,k)= t(i,j,k)
             end do
         end if
         if(wr_present) then
             do k=1,kb-1
                 wsec(pt,k)= wr(i,j,k)
             end do
         end if
     end do
   end subroutine getslice

  end subroutine sect

  
  subroutine check_time_range(pdt,pt1,pt2)
!    
! pdt is a interval between plots or dumps(seconds), pt1, pt2 is the
! beginning and the end time of the plotting (model time unit s).
! test if the given time range makes sense w.r. what we already have.
! if pdt < 0, set it to the model time step length.
!
    use state, only : debug,dt,ndivis,numstep
    implicit none
    real pdt,pt1,pt2

    if(pt2 < pt1) then
        if(debug > 0 .and. master) write(*,*)'dt,t1,t2=', pdt,pt1,pt2
        call message("invalid time range: t2 < t1",2)
    end if
    if(pdt > pt2-pt1) then
        if(debug > 0 .and. master) write(*,*)'dt,t1,t2=', pdt,pt1,pt2
        call message("invalid time range: interval (dt) longer than time window (t2-t1)",2)
    end if
    if(ndivis > 0) then
        if(pt2 > ndivis*dt) then
            if(debug > 0 .and. master) write(*,*)'  NUMSTEP=',numstep,' DT=', dt
            if(debug > 0 .and. master) write(*,*)'  dt,t1,t2=', pdt,pt1,pt2
            call message("end of time range beyond end of simulation=NUMSTEP*DT",1)
        end if
    end if
    if(pdt <= 0.0) then
        pdt = 3600./real(ndivis)
    end if
  end subroutine check_time_range

  
  subroutine add_bomplot(plotid, splitline)
    use state
    use mod_concentration
    use control_parameters, only : nonhyd, vertical_closure
    use grid_util
    implicit none
    character(*) plotid, splitline(:)
    integer n
    integer :: point,k,r(6)
    real reals(4)
    logical tmplogical
    
    call check_allocation()
    
    bomplot_numplots = bomplot_numplots + 1
    n = bomplot_numplots

    if(n > bomplot_maxplots) &
         call message('bomplot: too many plots. increase bomplot_maxplots and recompile',2)

    bomplots(n)%label = splitline(2)
    if(trim(splitline(2))=="") call message("no label found in column 2",2)

    select case(plotid)
    case("bomplot_hsect_depth")
        !! hsect at given depth
        !!   bomplot_hsect_depth <label>  <depth>                     <plotdt> <start> <end>

        !! slightly important; the plot type must correspond to the one in batch_bomplot
        bomplots(n)%type = 0

        read(splitline(3),*, err=102, end=102) bomplots(n)%depth
        if(debug>0.and.master) write(*,*)' ', trim(plotdesc(bomplots(n)%type)),', depth=',bomplots(n)%depth

        call gettimerange(4,6)
    case("bomplot_hsect_level")
        !! hsect at given level
        !!   bomplot_hsect_level <label>  <level>                     <plotdt> <start> <end>
        bomplots(n)%type = 1
        read(splitline(3),*, err=104, end=104) bomplots(n)%level
        if(debug>0.and.master) write(*,*)' ', trim(plotdesc(bomplots(n)%type)),', level=',bomplots(n)%level

        call gettimerange(4,6)
    case("bomplot_hsect_depth_range")
        !! hsect at given depth and range
        !!   bomplot_hsect_depth_range <label>  <depth> <x1> <x2> <y1> <y2> <plotdt> <start> <end>
        bomplots(n)%type = 2
        read(splitline(3),*, err=102, end=102) bomplots(n)%depth
        call get2drange(4,7)
        if(debug>0.and.master) write(*,*)' ', trim(plotdesc(bomplots(n)%type)),', depth=',bomplots(n)%depth, 'range=',bomplots(n)%range

        call gettimerange(8,10)
    case("bomplot_hsect2d")
        !! hsect2d
        !!   bomplot_hsect2d <label>                              <plotdt> <start> <end>
        bomplots(n)%type = 4
        if(debug>0.and.master) write(*,*)' ', trim(plotdesc(bomplots(n)%type))

        call gettimerange(3,5)
    case("bomplot_hsect2d_range")
        !! hsect2d for given range
        !!   bomplot_hsect2d_range <label>  <x1> <x2> <y1> <y2>         <plotdt> <start> <end>
        bomplots(n)%type = 5
        call get2drange(3,6)
        if(debug>0.and.master) write(*,*)' ', trim(plotdesc(bomplots(n)%type)),' range=', bomplots(n)%range

        call gettimerange(7,9)
    case("bomplot_vsect")
        !!vsect under connected line segments
        !!  bomplot_vsect <label>  <N> <x1> <y1> ... <xN> <yN>   <plotdt> <start> <end>
        bomplots(n)%type = 6
        read(splitline(3),*,err=111,end=111) bomplots(n)%nsecp1
        if(debug>0.and.master) write(*,*)'"',trim(plotdesc(bomplots(n)%type)),'" number of points=', bomplots(n)%nsecp1

        read(splitline(4:4+2*bomplots(n)%nsecp1-1),*,err=112,end=112) (bomplots(n)%px(point),bomplots(n)%py(point),point=1,bomplots(n)%nsecp1)

        bomplots(n)%px(1:bomplots(n)%nsecp1) = bomplots(n)%px(1:bomplots(n)%nsecp1)/dx + 1.0
        bomplots(n)%py(1:bomplots(n)%nsecp1) = bomplots(n)%py(1:bomplots(n)%nsecp1)/dy + 1.0
        
        if(jm>1.and.any(bomplots(n)%py(1:bomplots(n)%nsecp1) >= real(jm))) &
             call message("1 or more invalid Y's for vsect, all Y must be < REAL(JM)",2)
        if(any(bomplots(n)%px(1:bomplots(n)%nsecp1) >= real(im))) &
             call message("1 or more invalid X's for vsect, all X must be < REAL(IM)",2)
        if(jm>1.and.any(bomplots(n)%py(1:bomplots(n)%nsecp1) <= 0.0)) &
             call message("1 or more invalid Y's for vsect, all Y must be > 0",2)
        if(any(bomplots(n)%px(1:bomplots(n)%nsecp1) <= 0.0)) &
             call message("1 or more invalid X's for vsect, all X must be > 0",2)       
        if(debug>0.and.master) then
            write(*,*)'polygon points:'
            do point=1,bomplots(n)%nsecp1
                write(*,*)point, bomplots(n)%px(point), bomplots(n)%py(point)
            end do
        end if

        call gettimerange(4+2*bomplots(n)%nsecp1, 4+2*bomplots(n)%nsecp1 + 2)
    case("bomplot_isosurf")
        !! isosurface for rho
        !!   bomplot_isosurf <label>  <rhoisolevel> <2drange> <plotdt> <start> <end>
        bomplots(n)%type = 7
        read(splitline(3),*,err=114,end=114) bomplots(n)%isoval

        call get2drange(4,7)
        if(debug>0.and.master) write(*,*)' ', trim(plotdesc(bomplots(n)%type)),', isolevel=',bomplots(n)%isoval, 'range=',bomplots(n)%range

        call gettimerange(8,10)
    case("bomplot_vprof_ts")
        !! time series of vertical profile 
        !!   bomplot_vprof_ts "1010.2profil" 230.0 450.0   0 0 0
        bomplots(n)%type = 8
        read(splitline(3:4),*, err=117, end=117) reals(1:2)
        call xyz2gind(reals(1),reals(2),1.0,bomplots(n)%i, bomplots(n)%j, k)
        call gettimerange(5,7)

    case('bomplot_depth_average')
! bomplot_depth_avlerage <label> <3D field>  <x1> <x2> <y1> <y2>  <plotdt> <start> <end>
        bomplots(n)%type = 14
        bomplots(n)%label = splitline(2)
        bomplots(n)%what = splitline(3)
        call get2drange(4,7)
        call gettimerange(8,10)
        select case(bomplots(n)%what)
        case("conc1")
            bomplots(n)%f3d => co(1)%conc
        case("conc2")
            bomplots(n)%f3d => co(2)%conc
        case("conc3")
            bomplots(n)%f3d => co(3)%conc
        case("conc4")
            bomplots(n)%f3d => co(4)%conc
        case("conc5")
            bomplots(n)%f3d => co(5)%conc
        case("conc6")
            bomplots(n)%f3d => co(6)%conc
        case("conc7")
            bomplots(n)%f3d => co(7)%conc
        case("conc8")
            bomplots(n)%f3d => co(8)%conc
        case("conc9")
            bomplots(n)%f3d => co(9)%conc
        case("rho")
            bomplots(n)%f3d => rho
        case default
            call message("unknown <what> for genhsect",2)
        end select
        
    case('bomplot_genhsect')
        !! general horisontal section        
        !!    bomplot_genhsect <label> <what> <how> <depth> <2drange> <plotdt> <start> <end>

        bomplots(n)%what = splitline(3)
        bomplots(n)%label = splitline(2)
        
        select case(splitline(4))
        case("above_bottom")
            bomplots(n)%type=11
            read(splitline(5),*,err=118,end=118) bomplots(n)%depth
            bomplots(n)%depth = -bomplots(n)%depth
        case("below_surf")
            bomplots(n)%type=11
            read(splitline(5),*,err=118,end=118) bomplots(n)%depth
        case("level")
            bomplots(n)%type=10
            read(splitline(5),*,err=119,end=119) bomplots(n)%level
        case default
            call message("unknown <how> for bomplot_genhsect",2)
        end select
        
        call gettimerange(10,12)
        call get2drange(6,9)

        select case(bomplots(n)%what)
        case("q2")
            if(vertical_closure.ne.1) call message('vertical closure <> 1, not possible to dump',2)
            bomplots(n)%f3d => q2
        case("q2l")
            if(vertical_closure.ne.1) call message('vertical closure <> 1, not possible to dump',2)
            bomplots(n)%f3d => q2l
        case("conc1")
            bomplots(n)%f3d => co(1)%conc
        case("conc2")
            bomplots(n)%f3d => co(2)%conc
        case("conc3")
            bomplots(n)%f3d => co(3)%conc
        case("conc4")
            bomplots(n)%f3d => co(4)%conc
        case("conc5")
            bomplots(n)%f3d => co(5)%conc
        case("conc6")
            bomplots(n)%f3d => co(6)%conc
        case("conc7")
            bomplots(n)%f3d => co(7)%conc
        case("conc8")
            bomplots(n)%f3d => co(8)%conc
        case("conc9")
            bomplots(n)%f3d => co(9)%conc
        case("am")
            bomplots(n)%f3d => am
        case("ah")
            bomplots(n)%f3d => ah
        case("km")
            bomplots(n)%f3d => km
        case("kh")
            bomplots(n)%f3d => kh
        case("rho")
            bomplots(n)%f3d => rho
        case("wr")
            bomplots(n)%f3d => wr
        case("p")
            if(.not.nonhyd) call message("nonhyd is false, cannot dump P",2)
            bomplots(n)%f3d => p
        case default
            call message("unknown <what> for genhsect",2)
        end select
    case("bomplot_slice","bomplot_binslice","bomplot_accumslice")
        !! dump array slice to file
        !!   bomplot_slice <label> <plotdt> <start> <stop> <what> <i1> <i2> <j1> <j2> [<k1> <k2>]
        !!   bomplot_binslice <label> <plotdt> <start> <stop> <what> <i1> <i2> <j1> <j2> [<k1> <k2>]
        !! the latter is identical except the dump is done to a binary file instead of ascii.
        !! or
        !!   bomplot_accumslice <label> <numaccum> <start> <stop> <what> <i1> <i2> <j1> <j2> [<k1> <k2>]
        !! numaccum is the number of timesteps stored in each file

        bomplots(n)%what = splitline(6)
        
        select case(bomplots(n)%what)
        case("u")
            bomplots(n)%f3d => u
            call get3drange(7,12)
        case("v")
            bomplots(n)%f3d => v
            call get3drange(7,12)
        case("s")
            if(.not.associated(s)) call message('S is not allocated, not possible to dump',2)
            bomplots(n)%f3d => s
            call get3drange(7,12)
        case("t")
            if(.not.associated(t)) call message('T is not allocated, not possible to dump',2)
            bomplots(n)%f3d => t
            call get3drange(7,12)
        case("rho")
            bomplots(n)%f3d => rho
            call get3drange(7,12)
        case("wr")
            bomplots(n)%f3d => wr
            call get3drange(7,12)
        case("w")
            bomplots(n)%f3d => w
            call get3drange(7,12)
        case("km")
            bomplots(n)%f3d => km
            call get3drange(7,12)
        case("kh")
            bomplots(n)%f3d => kh
            call get3drange(7,12)
        case("am")
            bomplots(n)%f3d => am
            call get3drange(7,12)
        case("ah")
            bomplots(n)%f3d => ah
            call get3drange(7,12)
        case("conc1")
            if(.not. co(1)%ALLOCATED) call message("conc1 not allocated, cannot slice",2)
            bomplots(n)%f3d => co(1)%conc
            call get3drange(7,12)
        case("conc2")
            if(.not. co(2)%ALLOCATED) call message("conc2 not allocated, cannot slice",2)
            bomplots(n)%f3d => co(2)%conc
            call get3drange(7,12)
        case("conc3")
            if(.not. co(3)%ALLOCATED) call message("conc3 not allocated, cannot slice",2)
            bomplots(n)%f3d => co(3)%conc
            call get3drange(7,12)
        case("conc4")
            if(.not. co(4)%ALLOCATED) call message("conc4 not allocated, cannot slice",2)
            bomplots(n)%f3d => co(4)%conc
            call get3drange(7,12)
        case("conc5")
            if(.not. co(5)%ALLOCATED) call message("conc5 not allocated, cannot slice",2)
            bomplots(n)%f3d => co(5)%conc
            call get3drange(7,12)
        case("conc6")
            if(.not. co(6)%ALLOCATED) call message("conc6 not allocated, cannot slice",2)
            bomplots(n)%f3d => co(6)%conc
            call get3drange(7,12)
        case("conc7")
            if(.not. co(7)%ALLOCATED) call message("conc7 not allocated, cannot slice",2)
            bomplots(n)%f3d => co(7)%conc
            call get3drange(7,12)
        case("conc8")
            if(.not. co(8)%ALLOCATED) call message("conc8 not allocated, cannot slice",2)
            bomplots(n)%f3d => co(8)%conc
            call get3drange(7,12)
        case("conc9")
            if(.not. co(9)%ALLOCATED) call message("conc9 not allocated, cannot slice",2)
            bomplots(n)%f3d => co(9)%conc
            call get3drange(7,12)
        case("q2")
            if(.not.associated(q2)) call message("q2 not allocated, cannot slice",2)
            bomplots(n)%f3d => q2
            call get3drange(7,12)
        case("q2l")
            if(.not.associated(q2l)) call message("q2l not allocated, cannot slice",2)
            bomplots(n)%f3d => q2l
            call get3drange(7,12)
        case("p")
            if(.not.associated(p)) call message("P not allocated, check value of NONHYD (should be true)",2)
            bomplots(n)%f3d => p
            call get3drange(7,12)
        case("eta")
            bomplots(n)%f2d => eta
            call get2dirange(7,10)
        case("ua")
            bomplots(n)%f2d => ua
            call get2dirange(7,10)
        case("va")
            bomplots(n)%f2d => va
            call get2dirange(7,10)
        case default
            call message("unknown <what> for this keyword",2)
        end select
        
        call gettimerange(3,5)
    end select

        
! the type is used by the batch_bomplot routine
    if(plotid=="bomplot_slice") bomplots(n)%type = 9
    if(plotid=="bomplot_binslice") bomplots(n)%type = 12
    
    if(plotid=="bomplot_accumslice") then
        bomplots(n)%type = 13
        bomplots(n)%nsecp1 = nint(bomplots(n)%plotdt)  ! interval val read into the float for unused plotdt, convert back to int here
        bomplots(n)%plotdt = 0.99*dt                   ! should be dt, hack to make sure we get a field every time step
        bomplots(n)%n = 0
        r = bomplots(n)%range
        allocate(bomplots(n)%data4d(r(2)-r(1)+1, r(4)-r(3)+1, r(6)-r(5)+1, bomplots(n)%nsecp1))
    end if

    if(debug>0.and.master) then
        write(*,'(3x,a,6(i0,1x))') "what="//trim(bomplots(n)%what)//" range=",bomplots(n)%range
        if(all( (/bomplots(n)%plotdt, bomplots(n)%start, bomplots(n)%end/) > 0.0) ) then
            write(*,'(2x,a,f0.1,a,f0.1,a,f0.1)') 'PlotDT=',bomplots(n)%plotdt, ' start=',bomplots(n)%start, ' end=',bomplots(n)%end
        else
            write(*,*)'  dumped every time step'
        end if
    end if

! set end time to two years if 0 (=ignore). never seen a bom simulation that long.
    if( abs(bomplots(n)%end) < 0.00001) bomplots(n)%end = 2*365*24*3600

    return

101 call message('could not read integer for code from column 2',2)
102 call message('could not read float for depth from column 3',2)
104 call message('could not read integer for level from column 3',2)
107 call message('could not read 4 floats for x1,y1,x2,y2 from col 3-6',2)
111 call message('could not get an integer from col 3 for the number of points in the vsect specification',2)
112 call message('failed reading N polygon points from column 4 and onward',2)
114 call message('could not get float for isoval from column 3',2)
117 call message('could not get 2 integers for position of vsect series from column 3-4',2)
118 call message("could not read float for depth from column 5",2)        
119 call message("could not read integer for sigma level column 5",2)

  contains
    
    subroutine get2drange(c1,c2)
      integer c1,c2
      character(80) mess

      read(splitline(c1:c2),*,err=101,end=101) reals
      call apply0000convention(reals)
      bomplots(n)%range(1:4) = true2globrange(reals)
      call check_range(bomplots(n)%range(1:4))
      return
101   write(mess,'("could not get 4 numbers for 2d slice range from column ",i0," to ",i0)') c1,c2
      call message(trim(mess),2)
    end subroutine get2drange

    subroutine get2dirange(c1,c2)
      integer c1,c2
      character(80) mess

      read(splitline(c1:c2),*,err=101,end=101) bomplots(n)%range(1:4)
      if(all(bomplots(n)%range(1:4) == 0)) bomplots(n)%range(1:4) = (/1, im, 1, jm/)          
      call check_range(bomplots(n)%range(1:4))
      return
101   write(mess,'("could not get 4 numbers for 2d slice range from column ",i0," to ",i0)') c1,c2
      call message(trim(mess),2)
    end subroutine get2dirange
    
    subroutine get3drange(c1,c2)
      integer c1,c2
      character(80) mess
! 3d ranges (in bomplot) always integer (only used by slice for now)
      read(splitline(c1:c2),*,err=101,end=101) bomplots(n)%range(1:6)
      if(maxval(bomplots(n)%range(1:6)) == 0) bomplots(n)%range(1:6)=(/1,im,1,jm,1,kb/)
      call check_range(bomplots(n)%range(1:6))
      return
101   write(mess,'("could not get 6 integers for 3d slice from column ",i0," to ",i0)') c1,c2
      call message(trim(mess),2)
    end subroutine get3drange

    subroutine gettimerange(c1,c2)
      integer c1,c2
      character(80) mess
      write(mess,'("could not get 3 floats for plotdt,start,end from column ",i0," to ",i0)') c1,c2
      read(splitline(c1:c2),*, err=101, end=101) bomplots(n)%plotdt, bomplots(n)%start, bomplots(n)%end
      bomplots(n)%last_plot_time = - bomplots(n)%plotdt
      call check_time_range(bomplots(n)%plotdt, bomplots(n)%start, bomplots(n)%end)
      return
101   call message(trim(mess),2)
    end subroutine gettimerange

  end subroutine add_bomplot

  
  subroutine daver(label, field, range, output, spv)
!   given a 3D FIELD, compute and dump the depth average, optionally from REGION
    use state, only : im,jm,kb,h,dz
    implicit none
    character(len=*), intent(in) :: label
    real, intent(in) :: field(1-ghost_wi:,1-ghost_wj:,:)
    integer, intent(in), optional :: range(4)
    character(*), optional :: output
    real, optional :: spv
!    character(1), optional :: pos
    integer i,j,k,is,ie,js,je,gis,gie,gjs,gje,gr(4),r(4)
    real, dimension(i1:i2,j1:j2) :: fav
    logical inblock,from_pgplot
    real landspv

    if(.not.present(range))then
        gis=1 ; gie=im ; gjs=1 ; gje=jm
    else
        gis=range(1) ; gie=range(2) ; gjs=range(3) ; gje=range(4)
    end if
    call globalrect2local(gis,gie,gjs,gje,is,ie,js,je,inblock)
    gr = (/gis,gie,gjs,gje/)
    r = (/is,ie,js,je/)

    from_pgplot=.false.
    if(present(output))then
        if(output=="pgplot") from_pgplot=.true.
    end if
    
    landspv = -pgundef
    if(present(spv)) landspv=spv
    
    fav = 0.0
    if (inblock) then
        do k = 1,kb-1
            do j=js,je
                do i=is,ie
                    if(h(i,j) <= 0.0) then
                        fav(i,j) = landspv          ! default value is land
                        cycle     ! nothing to do for land cells
                    end if
                    fav(i,j) = fav(i,j) + field(i,j,k)*dz(k)
                end do
            end do
        end do
    end if
    
    select case(bomplot_dumping_method)
    case(0)
        if(.not.from_pgplot)then
            call dump2d(label, gr, fav, comp=bomplot_compress)
        else
            call dumppg(fav, spv=landspv)
        end if
    case(1)
        call dump2d(label, r, fav, comp=bomplot_compress, parallel=.true., inblock=inblock)
    end select
  end subroutine daver
  
END MODULE BOMPLOT
