module checkpointing
!
! A restart of BOM is done via the setup2 system
!  
! The restart file is automatically generated by BOM when a
! "checkpoint" is requested, and is accompanied by binary files for
! all necessary fields. Scalars are stored in the restart file
! "restart.dat". All restart data for a given checkpoint is stored in
! a separate directory, as we do for all BOM case's.
!  
! If we later decide we want to do netcdf, the dumpxxx routines below
! is where that would happen, in addition to additional initialization
! methods for arrays in setup2.f90.
!
! Helge Avlesen <avle@bccs.uib.no>
!  
  use state  
  use mod_bound
  use control_parameters
  use time_util
  use mod_concentration
  use timeseries
  use util
  use pgplot_core
  
  private
  public checkpoint, are_we_going_to_checkpoint, checkpoint_interval, checkpoint_time, checkpoint_label, numcptime, &
       max_checkpoint_times, checkpoint_timestep, checkpoint_dir, previous_checkpoint_time, reread_pgplot_timeseries, &
       checkpoint_delay

  interface store     
     module procedure dump2d, dump3d, dumpreal4, dumpreal8, &
          dumpinteger4, dumpinteger8, dumplogical, dumpcharacter, dump1d, &
          dump1dint
  end interface

  real, allocatable :: f3d(:,:,:), f2d(:,:)

  real(8) :: checkpoint_interval=-1                            ! interval between checkpoints in seconds
  real(8) :: previous_checkpoint_time=0.0_8                    ! the previous time we did an interval checkpoint
  integer :: checkpoint_timestep=-1                            ! for those insisting on using time step number to give next checkpoint.
  integer :: checkpoint_delay=2                                ! delay in seconds before writing checkpoint files
  
  integer, parameter :: max_checkpoint_times=50
  real(8) :: checkpoint_time(max_checkpoint_times)             ! a list of checkpoint times in seconds since beginning
  integer :: numcptime=0                                       ! the number of separate checkpoint times
  integer :: next_cptime=1                                     ! a pointer to the "next" checkpoint_time
  character(50) :: checkpoint_label(max_checkpoint_times)=" "  ! these checkpoints also have a label for the dir name

  integer, parameter :: sf=1000008, unf=1000009                ! file units used for checkpointing.

! the type of checkpoint to do  
  logical interval_checkpoint, separate_checkpoint, specific_step_checkpoint

  
contains

  
  LOGICAL FUNCTION ARE_WE_GOING_TO_CHECKPOINT()
!    
! return true if we will do a checkpoint during the current time step.
!    
    interval_checkpoint = .false.
    separate_checkpoint = .false.
    specific_step_checkpoint = .false.
    
    if(checkpoint_interval > 0.0) interval_checkpoint = time - previous_checkpoint_time >= checkpoint_interval
    
    if(numcptime > 0)then
        if(next_cptime <= numcptime) separate_checkpoint = nint(time) >= nint(checkpoint_time(next_cptime))
        if(master.and.debug>2)write(*,*) "NUMCPTIME", numcptime, next_cptime,"<", nint(time), ">",nint(checkpoint_time(next_cptime))
    end if

    if(istep == checkpoint_timestep) specific_step_checkpoint=.true.
    
    are_we_going_to_checkpoint = interval_checkpoint .or. separate_checkpoint .or. specific_step_checkpoint

    if(master.and.debug>3 .and. (interval_checkpoint .or. separate_checkpoint .or. specific_step_checkpoint))then
        write(*,*) 'CHECKPT?', interval_checkpoint, separate_checkpoint, specific_step_checkpoint        
        write(*,*) "CP intval",checkpoint_interval
        write(*,*) "CP time" , time
        write(*,*) "CP prev" , previous_checkpoint_time
    end if
    

! If a checkopoint is to be done, prepare a directory for the restart
! data. We must require NDATE to be set in the original setup file,
! otherwise it will later be difficult to determine what (model) time
! the dump was made...
!    
    if(are_we_going_to_checkpoint)then
        checkpoint_dir=" "
        write(checkpoint_dir,'("restart_",i4.4,"-",i2.2,"-",i2.2,"h",i2.2,"m",i2.2,"s",i2.2)') ndate

! If a checkpoint is requested along with a label, use it to name the restart directory
        if(separate_checkpoint)then
            if(trim(checkpoint_label(next_cptime)) .ne. "") then
                checkpoint_dir = "restart_" // trim(checkpoint_label(next_cptime))
            end if
            next_cptime = next_cptime + 1
        end if

        if(specific_step_checkpoint) write(checkpoint_dir,'("restart_step",i8.8)') istep    

! try to open a file in the checkpoint dir. if it fails, we assume the
! directory does not exist. if ok we overwrite its content.        
        if(master)then
            open(unit=sf, file=trim(checkpoint_dir)//"/thistestfilecanberemoved", iostat=ierr)
            if(ierr > 0) then
                if(debug > 1) write(*,'(3a)') "trying to open restart dir >>",trim(checkpoint_dir),"<<"
                call system("mkdir "//trim(checkpoint_dir))
            else
                close(sf)
            end if
        end if
        call ensureendslash(checkpoint_dir)
        
! wait a couple of seconds. we had a couple of crashes for cases where
! mkdir did not create a restart directory fast enough, e.g. when the
! timestep took 0.2s to complete.
        call fsleep(checkpoint_delay)
        
        call barrier()
        
    end if
  END FUNCTION ARE_WE_GOING_TO_CHECKPOINT
  
  
  SUBROUTINE CHECKPOINT()
!    
! We reuse the old setup file, which is saved into the
! control_parameter module, some keywords must be prepended to that
! file, some replaced and others appended.
!         
    implicit none
    integer i, timearr(8)
    integer ierr
    
    previous_checkpoint_time = time
    
! Storage for the gather operation in "store"
    
    if(master)then
        allocate(f3d(im,jm,kb), f2d(im,jm))
    else
        allocate(f3d(1,1,1), f2d(1,1))
    end if    

! Try to open a restart file, if not successful, we for some reason
! cannot dump, but we choose to continue simulation.
    call open_restart_file(ierr)
    if(ierr.ne.0) return    
    
! keep the original setup file and use it as a template for each
! restart file.
    newlines = lines
    newkeywords = keywords
    
!    
! This is where we start dumping (a minimum of) data to the setup and binary files.
!    
! The "store" call saves fields to unformatted files, and creates the
! proper line for the setup/restart file.  The second argument to
! "store" must be the same as the keyword used to initialize the
! field.
!
! "store" will check if the keyword is in the old setup file. if so,
! overwrite that line. if not, append a new line to the file, unless
! the optional argument prepend is .true.
!
    call store(advect_density, "advect_density", "prepend")
    call store(istep, "istep", "prepend")
    call store(initial_time, "initial_time", "prepend")
    call store(ndate, "ndate", "prepend")
    call store(time, "time", "prepend")
    call store(rhoaver, "rhoaver", "prepend")

    call store(data_path, "data_path", "comment")
    
    call store(h, "h", "replace")
    
    call store(u, "u")
    call store(un, "un")
    call store(delun, "delun")
    call store(uadv, "uadv")
    call store(wubot, "wubot")
    
    call store(v, "v")
    call store(vn, "vn")
    call store(delvn, "delvn")
    call store(vadv, "vadv")
    call store(wvbot, "wvbot")
    
    call store(w, "w")
    call store(wn, "wn")
    call store(delwn, "delwn")
    call store(wadv, "wadv")
       
    call store(eta, "eta")

    if(nonhyd_allocated)then
        call store(p, "p")
    end if    

    call store(km, "km")
    call store(kh, "kh")
    
    if(vertical_closure == 1)then
        call store(kq, "kq")
    end if
    
    if(advect_density)then
        call store(rho, "rho")
        call store(rhon, "rhon")
        call store(rhonm1, "rhonm1")
        call store(rho_bound, "rho_bound")
    end if
    if(advect_s) then
        call store(s, "s")
        call store(sn, "sn")
        call store(snm1, "snm1")
        call store(s_bound, "s_bound")
    end if    
    if(advect_t) then
        call store(t, "t")
        call store(tn, "tn")
        call store(tnm1, "tnm1")
        call store(t_bound, "t_bound")
    end if

    if(binreg_set)then
        if(.not.advect_density)then
            call store(s_initial, "s_initial")
            call store(t_initial, "t_initial")
        else
            call store(rho_initial, "rho_initial")
        end if
    end if
    
    if(vertical_closure == 1) then
        call store(q2, "q2")
        call store(q2l, "q2l")
        call store(q2_bound, "q2_bound")
        call store(q2l_bound, "q2l_bound")
    end if

! mod_bound: most stuff comes in via setup2, except arrays

    call store(u_bound, "u_bound")
    call store(v_bound, "v_bound")
    call store(w_bound, "w_bound")
    call store(ua_bound, "ua_bound")
    call store(va_bound, "va_bound")
    call store(eta_bound, "eta_bound")
    call store(etabfrs, "etabfrs")
    call store(etatfrs, "etatfrs")

! mod_concentration:   
!  concentration restart is not as accurate as S,T,RHO as we only store CONC
    
    if(numconc > 0) then
        do i=1,maxnumconc
            if(co(i)%allocated) then
                call store(co(i)%conc, trim(co(i)%name))
                call store(co(i)%conc_bound, trim(co(i)%name)//"_bound")
                if(binreg_set) call store(co(i)%conc_initial, trim(co(i)%name)//"_initial")
            end if
        end do
    end if

! timeseries. the actual data for timeseries is dumped regularly
! anyway and not needed for a clean restart, but we checkpoint the
! data used by pgplot
    if(num_ts > 0)then
        call store(timeseries_dump_start, "timeseries_dump_start", "prepend")
        call store(timeseries_file_number, "timeseries_file_number", "prepend")
        call store(.true., "reread_pgplot_timeseries")

        call checkpoint_pgplot_timeseries(istep, trim(checkpoint_dir))
    end if
    
    deallocate(f2d, f3d)

    
! Now write the restart "setup file"

! The first keyword in a restart file should always be "RESTART T"
    
    if(master)then
        call date_and_time(values=timearr)
        write(sf,'("!! -*-f90-*-",/,"!!",/,a)', advance='no') "!! Restart file written by BOM5 on "
        write(sf,'(i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2)') timearr(1:3),timearr(5:7)
        write(sf,'(a)') "!!"       
        write(sf,*,err=102) "restart t"
        write(sf,*,err=102) "sanity_check_fatal f"
        
        do i=1,newkeywords
            if(trim(newlines(i)) .ne. "") then
                write(sf,*) trim(newlines(i))
            end if
        end do    
        close(sf)
    end if
    
    return
    
101 if(master) write(*,'(///,a,///)') '******ERROR:  COULD NOT OPEN RESTART FILE *********'
    return
102 if(master) write(*,'(///,a,///)') '******ERROR:  COULD NOT WRITE TO RESTART FILE *********'
    return
  END SUBROUTINE CHECKPOINT


  SUBROUTINE OPEN_RESTART_FILE(error)
!    
! try to open the restart file. if the creation of the restart
! directory with the name reflecting dump time has failed, try the
! fallback directory "restart" instead. if also that fails, flag an
! error.
!    
    integer error,ierr
  
    call ensureendslash(checkpoint_dir)
    error=0
    if(master)then
        open(unit=sf, file=trim(checkpoint_dir)//"RESTARTFILE.DAT", err=1, recl=200)
        goto 3

! first wait a couple of seconds, try again        
1       call fsleep(checkpoint_delay)
        open(unit=sf, file=trim(checkpoint_dir)//"RESTARTFILE.DAT", err=5, recl=200)
        goto 3

! fail. try the default dir name
5       checkpoint_dir = "restart/"
        if(debug > 0) write(*,*) "FAILED opening",trim(checkpoint_dir)//"RESTARTFILE.DAT, trying default"
        open(unit=sf, file=trim(checkpoint_dir)//"RESTARTFILE.DAT", err=2, recl=200)
        goto 3
        
2       write(*,*) "also failed opening restartfile in fallback dir 'restart', skipping checkpoint"
        error=1
        goto 4
        
3       if(debug > 0) write(*,*) "Restart file successfully opened in checkpoint directory: ",trim(checkpoint_dir)
4       continue
    end if

    call broadcast(error)
  END SUBROUTINE



  SUBROUTINE CHECK_KEYWORD(KEYWORD, NEWLINE, ACTION)
!    
! Called by every store call to check if the keyword is in the old setup file or not.
! not at all efficient, but dumping for restart should be done sparingly anyway...
!    
    use util
    implicit none
    character(*) keyword, newline
    character(*), optional :: action
    character(132) splitline(50)    
    integer i
    logical do_prepend, do_replace, do_append, first, do_comment

! make "append" the default action of the routine
    do_prepend=.false.
    do_replace=.false.
    do_append=.true.
    do_comment=.false.
    
    if(present(action))then
        select case(action)
        case("prepend")
            do_prepend=.true.
        case("replace")
            do_replace=.true.
        case("append")
            do_append=.true.
        case("comment")
            do_comment=.true.
        case default
            call abort("check_keyword error: unknown action argument")
        end select
    end if

    !! first, check keyword against the current setup file. if action
    !! is replace, and it is the first occurence, replace. otherwise
    !! comment out.
    
    first=.true.
    do i=1,newkeywords
        call split(newlines(i), splitline)
        if(keyword == splitline(1)) then
            if(do_replace .and. first) then
                newlines(i) = trim(newline)
                first=.false.
                cycle
            end if            
            newlines(i)="!! "//trim(newlines(i))//" !commented out by checkpoint"
        end if
    end do

    !! if in replace or comment mode, we are done
    if(do_replace .or. do_comment) return
    
    if(do_prepend) then
        call prepend2list(newline, newlines, newkeywords)
        return
    end if
    
    call append2list(newline, newlines, newkeywords)
  END SUBROUTINE CHECK_KEYWORD


  SUBROUTINE APPEND2LIST(ITEM, LIST, N)
    implicit none
    character(*) item, list(:)
    integer n
    
    if(n+1 > size(list)) call abort("checkpoint.f90 routine append: the list is full, increase list dimension")
    n=n+1
    list(n) = item    
  END SUBROUTINE APPEND2LIST

  SUBROUTINE PREPEND2LIST(ITEM, LIST, N)
    implicit none
    character(*) item, list(:)
    integer i,n
    
    if(n+1 > size(list)) call abort("checkpoint.f90 routine prepend: the list is full, increase list dimension")
    n=n+1
    do i=n,2,-1
        list(i) = list(i-1)
    end do
    list(1) = item    
  END SUBROUTINE PREPEND2LIST
  

! now define routines to populate the "setup" interface above. one
! routine for each datatype as usual.
  
  subroutine dump3d(f, name, action)
    implicit none
    real f(i1:,j1:,:)
    character(*) name
    character(*), optional :: action
    character(200) newline
    
    call gatherfield(f, f3d, im, jm ,size(f,3))

    if(master)then
        if(debug > 2) write(*,*) "dump3d:",name
        call bindump(trim(checkpoint_dir)//name, f3d, access="stream")
        write(newline,'(a,2x,i0,2x,a)') name, 12, unfname(f3d, name, access="stream")
        call check_keyword(name, trim(newline), action)
    end if
  end subroutine dump3d
  
  subroutine dump2d(f, name, action)
    implicit none
    real f(i1:,j1:)
    character(*) name
    character(*), optional :: action
    character(200) newline
    
    call gatherfield(f, f2d, im, jm ,1)

    if(master)then
        if(debug > 2) write(*,*) "dump2d:",name
        call bindump(trim(checkpoint_dir)//name, f2d, access="stream")       
        write(newline,'(a,2x,i0,2x,a)') name, 12, unfname(f2d, name, access="stream")
        call check_keyword(name, trim(newline), action)
    end if
  end subroutine dump2d

  subroutine dump1d(f, name, action)
! 1d arrays are assumed to be "vertical" or other fully broadcast arrays.    
    implicit none
    real f(:)
    character(*) name
    character(*), optional :: action
    character(200) newline
    
    if(master)then
        if(debug > 2) write(*,*) "dump1d:",name
        call bindump(trim(checkpoint_dir)//name, f, access="stream")              
        write(newline,'(a,2x,i0,2x,a)') name, 12, unfname(f, name, access="stream")
        call check_keyword(name, trim(newline), action)
    end if
  end subroutine dump1d

  subroutine dump1dint(f, name, action)
! 1d arrays are assumed to be "vertical" or other fully broadcast arrays.    
    implicit none
    integer f(:)
    character(*) name
    character(*), optional :: action
    character(200) newline
    
    if(master)then
        if(debug > 2) write(*,*) "dump1dint:",name
        call bindump(trim(checkpoint_dir)//name, f, access="stream")                     
        write(newline,'(a,2x,i0,2x,a)') name, 12, unfname(f, name, access="stream")
        call check_keyword(name, trim(newline), action)
    end if
  end subroutine dump1dint
  
! it should be sufficient to store scalars in the restart file only
  
  subroutine dumpreal4(f, name, action)
    implicit none
    real(4) f
    character(*) name
    character(*), optional :: action
    character(200) newline
    
    if(master) then
        write(newline,*) name, f
        call check_keyword(name, trim(newline), action)
    end if
  end subroutine dumpreal4
  
  subroutine dumpreal8(f, name, action)
    implicit none
    real(8) f
    character(*) name
    character(*), optional :: action
    character(200) newline
    
    if(master) then
        write(newline,*) name, f
        call check_keyword(name, trim(newline), action)
    end if
  end subroutine dumpreal8

  subroutine dumpinteger4(f, name, action)
    implicit none
    integer(4) f
    character(*) name
    character(*), optional :: action
    character(200) newline
    
    if(master) then
        write(newline,*) name, f
        call check_keyword(name, trim(newline), action)
    end if
  end subroutine dumpinteger4
  
  subroutine dumpinteger8(f, name, action)
    implicit none
    integer(8) f
    character(*) name
    character(*), optional :: action
    character(200) newline
    
    if(master) then
        write(newline,*) name, f
        call check_keyword(name, trim(newline), action)
    end if
  end subroutine dumpinteger8

  subroutine dumplogical(f, name, action)
    implicit none
    logical f
    character(*) name
    character(*), optional :: action
    character(200) newline

    if(master) then
        write(newline,*) name, f
        call check_keyword(name, trim(newline), action)
    end if
  end subroutine dumplogical

  subroutine dumpcharacter(f, name, action)
    implicit none
    character(*) f, name
    character(*), optional :: action
    character(300) newline

    if(master) then
        write(newline,*) name, f
        call check_keyword(name, trim(newline), action)
    end if
  end subroutine dumpcharacter


  subroutine checkpoint_pgplot_timeseries(timestep, restart_dir)
! backup pgplot timeseries to unformatted files
    implicit none
    integer timestep
    character(*) restart_dir
    character(200) filename
    integer i
    real dummy
    
    if(master)then
        do i=1, pgplot_numplots
            if(pgplots(i)%type==4)then
                write(filename,'("pgplottimeseries-",i0,"-steps-",i0,".sr",i0)') i, timestep+1, kind(dummy)
                open(1000007, file=restart_dir//filename, access="stream", form='unformatted')
                write(1000007) pgplots(i)%data1d(0:timestep)
                close(1000007)
            end if
        end do
    end if
  end subroutine checkpoint_pgplot_timeseries
      
      
  subroutine reread_pgplot_timeseries(timestep)
! loop over the current timeseries, if a file that matches our dump
! file name format from "checkpoint_timeseries" is found, read it.
    implicit none
    character(200) filename
    integer timestep, i
    logical got_you
    real dummy
    
    if(master)then
        do i=1, pgplot_numplots
            if(pgplots(i)%type==4)then
                write(filename,'("pgplottimeseries-",i0,"-steps-",i0,".sr",i0)') i, timestep+1, kind(dummy)
                inquire(file=filename, exist=got_you)
                if(got_you) then
                    if(debug > 1.and.master) write(*,'(1x,a,a,a,i0,":",i0)') "Found timeseries restart file: ",trim(filename)
                    open(1000007, file=filename, access="stream", form='unformatted')
                    read(1000007) pgplots(i)%data1d(0:timestep)
                    close(1000007)
                end if
            end if
        end do
    end if
  end subroutine reread_pgplot_timeseries
  
end module checkpointing
