SUBROUTINE DYNP
!
!PURPOSE DYNPRESSURE approximates the solution to the equations
!
! dUD/dt = -1/RHO0 dP/dx
!
! dVD/dt = -1/RHO0 dP/dy
!
! dw/dt = -1/RHO0 dP/dz
!
! where P is the pressure caused by the dynamics, using an implicit 
! method.
!
! Global variables
!
USE STATE, west_side=>west, east_side=>east, north_side=>north, south_side=>south
USE MOD_BOUND  
USE SORITER
IMPLICIT NONE
!
! Local variables
!
INTEGER I,J,K,INFO,RIV,WELL
REAL FACTX,FACTY
REAL, POINTER :: RS(:,:,:), &
     WEST(:,:,:),EAST(:,:,:),SOUTH(:,:,:),NORTH(:,:,:), &
     UP(:,:,:),DOWN(:,:,:),TOTSUM(:,:,:)

LOGICAL INBLOCK

ALLOCATE(RS(I1:I2,J1:J2,KB),WEST(I1:I2,J1:J2,KB),EAST(I1:I2,J1:J2,KB),SOUTH(I1:I2,J1:J2,KB),NORTH(I1:I2,J1:J2,KB),&
     UP(I1:I2,J1:J2,KB),DOWN(I1:I2,J1:J2,KB),TOTSUM(I1:I2,J1:J2,KB))

!
!**********Set up the 'matrix' (7-diagonals) ********************
!
FACTX=(DT/(DX*DX))*(1/RHO0)
FACTY=(DT/(DY*DY))*(1/RHO0)
DO K=1,KB
  do j=1,j_eq_jm
    do i=1,i_eq_im
                 WEST(I,J,K)=FACTX*DU(I,J)
        IF(JM>1) SOUTH(I,J,K)=FACTY*DV(I,J)
    end do    
  end do
END DO

call tag_ghost(west, east_side, 1)
call tag_ghost(south, north_side, 1)
call update_ghost
! 
EAST=0.0E0
do k=1,kb
  do j=1,j_eq_jm
    do i=1,im_mi_1
        east(i,j,k) = west(i+1,j,k)
    end do
  end do
end do

IF(JM > 1)THEN
NORTH=0.0E0
do k=1,kb
  do j=1,jm_mi_1
    do i=1,i_eq_im
        north(i,j,k) = south(i,j+1,k)
    end do
  end do
end do
END IF


UP=0.0E0
DOWN=0.0E0
DO J=j_eq_2,jm_mi_1
  DO I=i_eq_2,IM_mi_1
     IF (FSM(I,J).NE.0.0) THEN
      DO K=2,KB-1
         UP(I,J,K)=(DT/RHO0)*(1./D(I,J))*(1./DZ(K))*(1./DZZ(K-1))
       END DO
       DO K=1,KB-2
          DOWN(I,J,K)=(DT/RHO0)*(1./D(I,J))*(1./DZ(K))*(1./DZZ(K))
       END DO
     ENDIF
  END DO
END DO
UP(:,:,1)=0.0E0
DOWN(:,:,KB-1:KB)=0.0E0
!
! In the case of a source/sink at the sea bed, this must be considered in DOWN
! and RS.
!
IF (UPWELL) THEN
   DO WELL = 1,NUMWELL
      i=iwell(WELL)
      j=jwell(WELL)
      call globalind2local(i,j,inblock)
      if(inblock)then
         DOWN(I,J,KB-1) = (DT/RHO0)*(1./D(I,J))*(1./DZ(KB-1))*(1./DZZ(KB-2))
      end if
   END DO
ENDIF

!
!
TOTSUM=0.0E0
RS=0.0E0

IF(JM > 1)THEN
DO K=1,KB-1
  DO J=J_EQ_2,JM_MI_1
    DO I=I_EQ_2,IM_MI_1      
      TOTSUM(I,J,K)=-(WEST(I,J,K)+EAST(I,J,K)+  &
        &           SOUTH(I,J,K)+NORTH(I,J,K)+  &
        &           UP(I,J,K)+DOWN(I,J,K))
    END DO
  END DO
END DO
ELSE
DO K=1,KB-1
  DO J=J_EQ_2,JM_MI_1
    DO I=I_EQ_2,IM_MI_1      
      TOTSUM(I,J,K)=-(WEST(I,J,K)+EAST(I,J,K)+  &
        &           UP(I,J,K)+DOWN(I,J,K))
    END DO
  END DO
END DO
END IF
!
!************ Calculate right hand side. ************************
!
DO K=1,KB-1
   DO J=J_EQ_2,JM_MI_1
       DO I=I_EQ_2,IM_MI_1
           RS(I,J,K)= RS(I,J,K) + (ETA(I,J)-ETAP(I,J))/DT &
         & + (UADV(I+1,J,K)*DU(I+1,J)-UADV(I,J,K)*DU(I,J))/DX  &
         & + (W(I,J,K)-W(I,J,K+1))/DZ(K)

           IF(JM>1) RS(I,J,K)= RS(I,J,K) + (VADV(I,J+1,K)*DV(I,J+1)-VADV(I,J,K)*DV(I,J))/DY 
          
           RS(I,J,K)=RS(I,J,K)*FSM(I,J)
        END DO
   END DO
END DO

IF(SOR_RESID < 0.0) THEN
   SOR_RESID = 1.0E-2
END IF
IF(SOR_NUMITER < 0) THEN
   SOR_NUMITER = 20
END IF

IF(JM > 1)THEN
   CALL SORITER3D(RS,P,SOR_NUMITER,SOR_RESID,INFO,IM,JM,KB,  &
       &       EAST,WEST,SOUTH,NORTH,UP,DOWN,TOTSUM,FSM)
ELSE
   CALL SORITER2D(RS,P,SOR_NUMITER,SOR_RESID,INFO,IM,   KB , &
               EAST,WEST,            UP,DOWN,TOTSUM,FSM)
END IF

IF(AVPEQ0>0)THEN
    IF(MOD(ISTEP,AVPEQ0)==0) CALL ENFORCE0AVP()
END IF

CALL DUMP_STATE('dynp')

DEALLOCATE(RS,WEST,EAST,SOUTH,NORTH,UP,DOWN,TOTSUM)


CONTAINS

  SUBROUTINE ENFORCE0AVP
!
!  Force the average P equal to 0
!
    USE NORMS
    IMPLICIT NONE
    REAL MEANP
    
    MEANP = PARSUM(P)    
    MEANP = MEANP/(NUMWET*(KB-1))
    
    IF (DEBUG > 0) THEN
        IF(MASTER) PRINT *,'MEANP=',MEANP
    END IF
    
    DO K=1,KB-1
        DO J=1, J_EQ_JM
            DO I=1,I_EQ_IM
                P(I,J,K) = FSM(I,J)*(P(I,J,K) - MEANP)
            END DO
        END DO
    END DO    
  END SUBROUTINE ENFORCE0AVP
END SUBROUTINE DYNP

