module file_util
  use qmpi, only : message
  use time_util, only : fsleep
  
  private
  public readfield, genopen, gettype, getdimension, &
         checkunit, get_fields, createncheckdir

  interface get_fields
!    
! get 1D,2D or 3D fields from spreadsheet type file. data can be fetched from
! up to 8 columns. if specific columns are given in the argument list
! COLS, only those will be fetched. if no COLS argument is given, the
! first column of the file will be assigned to the first argument
! after the file name, and so on.
!    
! example1: get u,v,s,t from "test1.dat"
!      call get_fields("test1.dat", U,V,S,T)
! example2: get vertical position z and s from column 4 and 2, and the number of lines read n
!      call get_fields("test3.dat", Z, S, cols=(/4,2/), lines=n)
     module procedure get_1d_fields4, get_1d_fields8, &
          get_2d_fields4, get_2d_fields8, &
          get_3d_fields4, get_3d_fields8
  end interface get_fields
  
  
  interface readfield
!     
! SUBROUTINE READFIELD(F, FILENAME)     
! try to read F from the given file FILENAME. the file contains a
! single field. the interface guess the type and dimension of the
! field from the substring of the file name between the last "-" and
! the last "."
!      use filereader
!      real, allocatable :: arr(:,:)
!      ...
!      call readfield(arr, "arr-10x10.sr4")
! "arr" should be allocatable, and is allocated in the routine on need
!     
     module procedure getfield_1dr4, getfield_1dr8, getfield_1di4, &
                      getfield_2dr4, getfield_2dr8, getfield_2di4, &
                      getfield_3dr4, getfield_3dr8, getfield_3di4
  end interface
  
  interface check_args
!  subroutine check_args(field, i, f, n, s)
!    field is a 1d, 2d or 3d array(field) to be read from a columnar file.     
!    i is the column of the field     
!    set f to the number of elements in field
!    set n to i if i>n.
!    s is the fortran shape of the field

     module procedure check_args1d4, check_args2d4, check_args3d4,&
          check_args1d8, check_args2d8, check_args3d8
  end interface

  
contains



  SUBROUTINE CREATENCHECKDIR(DIR ,PATIENCE, ENDSLASH, COMMENT)
!   try to create a new directory DIR, ignore error if it exists.
!   stop if it is not possible to create a file. optionally be patient
!   for PATIENCE seconds, retry. optionally ensure DIR ends with a "/" on exit.
!   if present, COMMENT is a string which is written to the temporary file    
    implicit none
    character(*) dir
    integer, optional :: patience
    logical, optional :: endslash
    character(*), optional :: comment
    integer n

! temporary remove trailing slash    
    n = len(trim(dir))    
    if(dir(n:n) == "/") then
        dir(n:) = " "
    end if

! linux, osx, ...
    call system("mkdir -p " // trim(dir))
    
    open(99, file=trim(dir)//"/readordeleteme.now", err=1)
! success
    if(present(comment)) write(99,*) comment
    goto 3

! first wait a couple of seconds, try again
1   if(present(patience)) then
        call fsleep(patience)
        open(99, file=trim(dir)//"/readordeleteme.NOW", err=5)
        if(present(comment)) write(99,*) comment
        goto 3
    end if
    
! fail. return dot slash
5   write(*,*) "FAILED opening directory: ",trim(dir)," reverting DIR to default ./"
    dir = "./"
    return
        
3   close(99)

! if callee asks for a slash...
4   if(present(endslash)) then
        if(endslash) then
            n = len(trim(dir))    
            dir(n+1:n+1) = "/"
        end if
    end if
  end SUBROUTINE CREATENCHECKDIR

  
  subroutine getfield_1dr4(f, name)
    real(4), allocatable :: f(:)
    include "filereader1d.f90"
  end subroutine getfield_1dr4

  subroutine getfield_1dr8(f, name)
    real(8), allocatable :: f(:)
    include "filereader1d.f90"
  end subroutine getfield_1dr8
  
  subroutine getfield_1di4(f, name)
    integer(4), allocatable :: f(:)
    include "filereader1d.f90"
  end subroutine getfield_1di4

  subroutine getfield_2dr4(f, name)
    real(4), allocatable :: f(:,:)
    include "filereader2d.f90"
  end subroutine getfield_2dr4

  subroutine getfield_2dr8(f, name)
    real(8), allocatable :: f(:,:)
    include "filereader2d.f90"
  end subroutine getfield_2dr8

  subroutine getfield_2di4(f, name)
    integer(4), allocatable :: f(:,:)
    include "filereader2d.f90"
  end subroutine getfield_2di4
 
  subroutine getfield_3dr4(f, name)
    real(4), allocatable :: f(:,:,:)
    include "filereader3d.f90"
  end subroutine getfield_3dr4

  subroutine getfield_3dr8(f, name)
    real(8), allocatable :: f(:,:,:)
    include "filereader3d.f90"
  end subroutine getfield_3dr8
  
  subroutine getfield_3di4(f, name)
    integer(4), allocatable :: f(:,:,:)
    include "filereader3d.f90"
  end subroutine getfield_3di4

  
  subroutine genopen(unit, filename, type)
! shortcut routine to open FILENAME connected to UNIT according to predefined codes TYPE;
! 1=stream format, 2=fortran unformatted sequential, anything else "text".
    integer unit
    character(*) filename
    integer type
    
    select case(type)
    case(1)
        open(unit, file=filename, status="old", access="stream", form="unformatted")
    case(2)
        open(unit, file=filename, status="old", form="unformatted")
    case default
        open(unit, file=filename, status="old")
    end select
  end subroutine genopen

  
  integer function gettype(string)
    character(*) string
    gettype = 0
    if(index(string, ".sr") > 0) gettype=1
    if(index(string, ".unfr") > 0) gettype=2
    if(index(string, ".si") > 0) gettype=1
    if(index(string, ".unfi") > 0) gettype=2
  end function gettype


  function getdimension(string,adim)    
! parse a file name backwards from the dot until the minus to detect the
! dimensions of the field in the array, e.g.    
!     test-40x23x10.unfr4
! should return 3 40 23 10, i.e. 3 dimensions, then the size of each.    
    implicit none
    character(*) string
    integer getdimension(10),xpos(5),p1,p2,numit,minuspos,dotpos,numx,tmp,adim

    getdimension = 0
    
! search backwards for the dot and minus
    dotpos = index(string, ".", back=.true.)
    if(dotpos <= 0) return
    minuspos = index(string, "-", back=.true.) 
    if(minuspos <= 0) return
    if(minuspos > dotpos .or. dotpos-minuspos==1) return

! 0 x can mean 1d, 1 x can mean 2d, etc.    
    numx = 0
    p1=minuspos+1
    p2=dotpos-1
    do numit=1,10
        tmp = index(string(p1:p2), "x", back=.true.)
        if(tmp > 0) then
            numx = numx + 1
            xpos(numx) = p1 - 1 + tmp
            p2 = p1 - 1 + tmp - 1
            if(numx > 2) exit
        end if
    end do

    xpos(1:numx) = xpos(numx:1:-1)
!    print *,'numx',numx,'minuspos',minuspos,'dotpos',dotpos,'pos',xpos(1:numx)
    
! try to read integer from the gaps, fail silently if there is
! anything else than digits in the sub strings
    select case(numx)
    case(0)
        read(string(minuspos+1:dotpos-1),*,err=101,end=101) getdimension(2)
        getdimension(1) = 1
    case(1)
        read(string(minuspos+1:xpos(1)-1),*,err=101,end=101) getdimension(2)
        read(string(xpos(1)+1:dotpos-1),*,err=101,end=101) getdimension(3)
        getdimension(1) = 2
    case(2)
        read(string(minuspos+1:xpos(1)-1),*,err=101,end=101) getdimension(2)
        read(string(xpos(1)+1:xpos(2)-1),*,err=101,end=101) getdimension(3)
        read(string(xpos(2)+1:dotpos-1),*,err=101,end=101) getdimension(4)
        getdimension(1) = 3
    case(3)
        read(string(minuspos+1:xpos(1)-1),*,err=101,end=101) getdimension(2)
        read(string(xpos(1)+1:xpos(2)-1),*,err=101,end=101) getdimension(3)
        read(string(xpos(2)+1:xpos(3)-1),*,err=101,end=101) getdimension(4)       
        read(string(xpos(3)+1:dotpos-1),*,err=101,end=101) getdimension(5)
        getdimension(1) = 4
    case default
        write(*,*) "filereader getdimension: not taught how to parse more than 3 x's"
    end select
101 if(adim .ne. getdimension(1)) then
        write(*,*) "filereader getdimension error: expected dimension=",adim," does not match dimension from string=",trim(string)
        stop
    end if
  end function getdimension


  subroutine checkunit(unit,label)
    logical taken
    integer unit
    character(*),optional :: label
    inquire(unit=unit, opened=taken)
    if(taken) then
        if(present(label)) write(*,*)label
        write(*,*)'unit=',unit,'already open'
        stop
    end if
  end subroutine checkunit


  subroutine check_args1d4(field, i, f, n,s,m)
    implicit none
    real(4) field(:)
    integer i, f(:), n, s(:), stmp(size(s)),m

    if(all(s==0)) s=shape(field)
    stmp=shape(field)
    if(.not. all(s==stmp)) then
        write(*,*)'get_1D_fields error: fields provided via col1-col8 were of different size 1:',s,'2:',stmp
        stop
    end if
    s=stmp
    f(i)=s(1)
    if(i>n) n=i
    m=m+1
  end subroutine
  subroutine check_args2d4(field, i, f, n,s,m)
    implicit none
    real(4) field(:,:)
    integer i, f(:), n, s(:), stmp(size(s)), m
    
    if(all(s==0)) s=shape(field)
    stmp=shape(field)
    if(.not. all(s==stmp)) then
        write(*,*)'get_2D_fields error: fields provided via col1-col8 were of different size 1:',s,'2:',stmp
        stop
    end if
    s=stmp
    f(i)=s(1)*s(2)
    if(i>n) n=i
    m=m+1
  end subroutine
  subroutine check_args3d4(field, i, f, n,s,m)
    implicit none
    real(4) field(:,:,:)
    integer i, f(:), n, s(:), stmp(size(s)),m
    
    if(all(s==0)) s=shape(field)
    stmp=shape(field)
    if(.not. all(s==stmp)) then
        write(*,*)'get_3D_fields error: fields provided via col1-col8 were of different size 1:',s,'2:',stmp
        stop
    end if
    s=stmp
    f(i)=s(1)*s(2)*s(3)
    if(i>n) n=i
    m=m+1
  end subroutine
   


  subroutine check_args1d8(field, i, f, n,s,m)
    implicit none
    real(8) field(:)
    integer i, f(:), n, s(:), stmp(size(s)),m

    if(all(s==0)) s=shape(field)
    stmp=shape(field)
    if(.not. all(s==stmp)) then
        write(*,*)'get_1D_fields error: fields provided via col1-col8 were of different size 1:',s,'2:',stmp
        stop
    end if
    s=stmp
    f(i)=s(1)
    if(i>n) n=i
    m=m+1
  end subroutine
  subroutine check_args2d8(field, i, f, n,s,m)
    implicit none
    real(8) field(:,:)
    integer i, f(:), n, s(:), stmp(size(s)), m
    
    if(all(s==0)) s=shape(field)
    stmp=shape(field)
    if(.not. all(s==stmp)) then
        write(*,*)'get_2D_fields error: fields provided via col1-col8 were of different size 1:',s,'2:',stmp
        stop
    end if
    s=stmp
    f(i)=s(1)*s(2)
    if(i>n) n=i
    m=m+1
  end subroutine
  subroutine check_args3d8(field, i, f, n,s,m)
    implicit none
    real(8) field(:,:,:)
    integer i, f(:), n, s(:), stmp(size(s)),m
    
    if(all(s==0)) s=shape(field)
    stmp=shape(field)
    if(.not. all(s==stmp)) then
        write(*,*)'get_3D_fields error: fields provided via col1-col8 were of different size 1:',s,'2:',stmp
        stop
    end if
    s=stmp
    f(i)=s(1)*s(2)*s(3)
    if(i>n) n=i
    m=m+1
  end subroutine

  
  subroutine get_1d_fields4(file, col1, col2, col3, col4, col5, col6, col7, col8, cols, lines, file_region, file_shape,ierr)
!    
! get 1d, 2d or 3d fields from file. data can be fetched from up to 8
! columns and from a given region in these fields. if a region is
! specified, the shape of the fields in the file must be given.
! if specific columns are given in the argument list, only those will be fetched.
! the provided region size must match the size of the given field.
!    
! example1: get u,v,s,t from "test1.dat"
!      call get_data("test1.dat", U,V,S,T)
! example2: get vertical position z and s from column 4 and 2, and the number of lines read n
!      call get_data("test3.dat", Z, S, cols=(/4,2/), lines=n)
    
    implicit none
    integer, parameter :: localkind=4

    include "get1dfields.f90"   
  end subroutine

  subroutine get_1d_fields8(file, col1, col2, col3, col4, col5, col6, col7, col8, cols, lines, file_region, file_shape,ierr)
    implicit none
    integer, parameter :: localkind=8

    include "get1dfields.f90"   
  end subroutine
  
  subroutine get_2d_fields4(file, col1, col2, col3, col4, col5, col6, col7, col8, cols, file_region, file_shape, ierr)
    implicit none
    integer, parameter :: localkind=4

    include "get2dfields.f90"
  end subroutine
  
  subroutine get_2d_fields8(file, col1, col2, col3, col4, col5, col6, col7, col8, cols, file_region, file_shape, ierr)
    implicit none
    integer, parameter :: localkind=8

    include "get2dfields.f90"
  end subroutine
  
  subroutine get_3d_fields4(file, col1, col2, col3, col4, col5, col6, col7, col8, cols, file_region, file_shape, ierr)
    implicit none
    integer, parameter :: localkind=4

    include "get3dfields.f90"
  end subroutine get_3d_fields4
  
  subroutine get_3d_fields8(file, col1, col2, col3, col4, col5, col6, col7, col8, cols, file_region, file_shape, ierr)
    implicit none
    integer, parameter :: localkind=8

    include "get3dfields.f90"
  end subroutine get_3d_fields8
  
end module
