module grid_util
!  
! various utility routines specific to BOM type grids
!  
! SUBROUTINE XYZ2GIND(X,Y,VZ,I,J,K)
!   return i,j,k for box containing global grid position x,y,vz (meters)  
! SUBROUTINE CHECK_RANGE(R)
!   test if global range r=(i1,i2[,j1,j2,k1,k2]) makes sense wrt im,jm etc.
! SUBROUTINE CHECKPOS(X)
!   abort if position in vector x is outside grid  
! SUBROUTINE APPLY0000CONVENTION(R)
!   if coordinate range r=(/x1,x2,y1,y2/) are all 0, set r to max possible extent.
! SUBROUTINE GR2SPH(XGRID, YGRID, XPOLE, YPOLE, DX, DY, YLONG,LONG, LATT, IFAIL)
!   convert from model grid coordinate to geographical coordinate using the model polar stereographic transform.  
! SUBROUTINE SPH2GR(LONG, LATT, XPOLE, YPOLE, DX, DY, YLONG,XGRID, YGRID, IFAIL)
!   convert from geographical coordinates to the model grid coordinate  


  use qmpi, only : communicator_object  

  
! Data for the multiple grid setup.  
! The grid id of the processor local grid, and total number of grids
  
  integer :: mygid=1, numgrids=1
  
  type grid_object
     character(20) :: label=" "
     character(200) :: init_file=" "      ! the setupfile for the grid
     integer :: rank1=0, rank2=0          ! first and last global rank of grid
     real :: workload=0.                  ! a number saying how much work the grid has relative to the others
  end type grid_object
  
  type(grid_object) :: grid(10)
  
! each grid has its own communicator
  type(communicator_object) :: gridcomm(10)

! the projection of the model grid, currently supported are "std"
! which is a polar stereographic transform (see gr2sph),  
!  "utm33","utm32","polst". For the UTM system the global grid
!  coordinate is relative to the UTM coordinate x0,y0 of the south
!  west corner
  
  character(5) :: grid_projection="polst"
  real :: x0, y0
  
contains

  subroutine xyz2gind(x,y,vz,i,j,k)
!
! given a global grid position in meters, return the global index of
! the surrounding box. it is assumed that ETA is 0 at the time of the
! call.
!
    use mod_mpi, only : globval
    use state, only : dx,dy,z,h,jm,kb
    implicit none
    real x,y,vz,depth
    integer i,j,k,k2
    
    i = nint(x/dx + 1.5)
    j = nint(y/dy + 1.5)
    if(jm==1) j=1
    
    depth = globval(h,i,j)
    if(vz > 0.0) then
        k = -2e9
        return
    end if
    if(vz == depth) then
        k = kb-1
        return
    end if
    
    k = 2e9
    
    do k2=2,kb
        if(vz <= z(k2-1)*depth .and. vz > z(k2)*depth) then
            k = k2-1
            exit
        end if
    end do
  end subroutine xyz2gind

  
  subroutine check_range(r)
!
! perform tests to see if global range r=(i1,i2[,j1,j2,k1,k2]) makes sense
! w.r. current grid IM,JM,KB. if not, exit program with error message.
!
    use qmpi, only : message
    use state, only : im,jm,kb
    implicit none
    integer i,r(:),comp_dim(6),ind
    character(200) string

    if(size(r) > 6 .or. mod(size(r),2).ne.0) call message("r does not have correct length",2)

    comp_dim = (/1,im,1,jm,1,kb/)
    
    do i=1,size(r)/2
        ind = i*2-1
        if(r(ind) < comp_dim(ind)) then
            write(string,'(a,i0,a,6(i0,","))') "range r(",ind,") too small, r=", r
            call message(trim(string),2)
        end if

        ind = i*2
        if(r(ind) > comp_dim(ind)) then
            write(string,'(a,i0,a,6(i0,","))') "range r(",ind,") too large, r=",r
            call message(trim(string),2)
        end if

        if(r(ind) < r(ind-1)) then
            write(string,'("range r(",i0,") < r(",i0,")")') ind,ind-1
            call message(trim(string),2)
        end if
    end do
  end subroutine check_range

  
  subroutine checkpos(x)
! abort if position in vector x is outside grid
    use state, only : dx,dy,lx,ly,h,im,jm
    use mod_mpi, only : globval
    use qmpi, only : message
    implicit none
    real x(:)
    integer i,j
    
    if(x(1) < 0.0 .or. x(1) > lx) call message("x position outside grid",2)
    if(jm > 1)then
        if(x(2) < 0.0 .or. x(2) > ly) call message("y position outside grid",2)
    else
        if(x(2) < 0.0 .or. x(2) > 0.0) call message("y position outside grid",2)
    end if
    if(size(x) > 2) then
        i = nint(x(1)/dx + 1.5)
        if(jm > 1)then
            j = nint(x(2)/dy + 1.5)
        else
            j = 1
        end if
        if(x(3) > 0 .or. x(3) < -globval(h,i,j)) call message("z should be <= 0 and >= -abs(h(x,y))",2)
    end if
  end subroutine checkpos

  
  subroutine apply0000convention(r)
! r is a horisontal (true) coordinate range=(/x1,x2,y1,y2/)
! if all r are 0, set them to full extent.
    use state, only : lx,ly,jm
    implicit none
    real r(4)
    if(abs(maxval(r)) < 1e-7) then
        r=(/0.0, lx, 0.0, ly/)
        if(jm==1) r(4)=0.0
    end if
  end subroutine apply0000convention
  
  
  function true2globrange(r)
! let r=(/x1,x2,y1,y2/), a global (true) horisontal coordinate range in meters.
! return global indice range, also always correct for jm=1.
    use state, only : dx,dy,im,jm
    implicit none
    real r(4)
    integer true2globrange(4)

    if(abs(maxval(r)) < 1e-7) then
        true2globrange = (/1, im, 1, jm/)
    else
! its always problematic when coordinates are exactly on the edge. try
! to make the range slightly less than we need:
        true2globrange = nint(r/(/dx,dx,dy,dy/)+(/1.50001,1.49999,1.50001,1.49999/))
        if(jm == 1) true2globrange(3:4) = 1
    end if
  end function true2globrange


  
  SUBROUTINE GR2SPH(XGRID, YGRID, XPOLE, YPOLE, DX, DY, YLONG,LONG, LATT, IFAIL)
!
!***BEGIN PROLOGUE GR2SPH
!***DATE WRITTEN    7. JAN. 1991
!***REVISION DATE  23. JAN. 1991
!***AUTHOR
!            Bjorn Adlandsvik
!            Institute of Marine Research,
!            Postboks 1870 Nordnes
!            N-5024 Bergen, Norway
!            Email..  bjorn@imr.no
!
!  The code is based on an the routine XYBL from 
!  The Norwegian Meteorological Institute.
!
!***PURPOSE
!  Converts from polar stereographic grid coordinates
!  to spherical coordinates (lattitude and longitude).
!
!***PARAMETERS
!***ON ENTRY
!     XGRID  Real,
!        X grid coordinate.
!     YGRID  Real,
!        Y grid coordinate.
!     XPOLE  Real,
!        X grid coordinate to the North Pole
!     YPOLE  Real,
!        Y grid coordinate to the North Pole.
!     DX     Real,
!        Width in kilometers of grid cell in X direction.
!     DY     Real,
!        Width in kilometers of grid cell in Y direction.
!     YLONG  Real,
!        Longitude in (decimal) degrees of meridian (anti)parallell 
!        to Y-axis.
!
!***ON RETURN
!
!     LONG   Real,
!        Longitude in (decimal) degrees,
!        Normalised to -180 <= LONG <= 180.
!     LATT   Real,
!        Lattitude in (decimal) degrees.
!     IFAIL  Integer,
!        Error indicator, returns
!        2 if DX = 0 or DY = 0,
!        1 if XGRID = XPOLE and YGRID = YPOLE,
!        0 otherwise.
!
!***DESCRIPTION
!  
!  For the coordinate systems see description in routine SPH2GR
!  The North and South Poles does not have well defined longitudes 
!    and are not allowed. This means that 
!    (XGRID, YGRID) = (XPOLE, YPOLE) is not allowed.
!
!  The longitude is computed by
!
!     tan(long-ylong) = (dx * (xgrid-xpole)) / -(dy * (ygrid-ypole))
!
!  and the lattitude by
!
!     tan(45 - 0.5*latt) = r / (rearth * (1 + sin(phinul))
!
!  with
!
!     r**2 = (dx * (xgrid-xp))**2 + (dy * (ygrid-yp))**2
!
!  Normalising is done by finding a rotation number k
!  such that.
!     - 180 <= long + k*360 <= 180
!  that is
!    - long/360 - 1/2 <=  k <= -long/360 + 1/2
!    k is the nearest integer to ( - long/360)
!
!***LOCAL VARIABLES
!
!     LAMBDA		Real,
!        Lattitude in radians.
!     PHI		Real,
!        Longitude in radians.
!     ALPHA		Real,
!        YLONG in radians.
!     PI 		Real,
!        3.14.....
!     RAD		Real,
!        Conversion factor from degrees to radians.
!     DEG		Real,
!        Conversion factor from radians to degrees.
!     REARTH		Real,
!        Radius (in kilometers) of Earth.
!     PHINUL	   	Real,
!        Lattitude (in radians) of intersecting plane =
!        lattitude of "true" length.
!     R 		Real,
!        Distance (in kilometers) from North Pole in the plane.
!
!***RELATED ROUTINES
!  This routine is the inverse of SPH2GR
!
!***ROUTINES CALLED-NONE
!***END PROLOGUE GR2SPH
!
!  Parameters
!
    REAL XGRID, YGRID, XPOLE, YPOLE, DX, DY, YLONG, LONG, LATT
    INTEGER IFAIL
!
!  Local variables.
!
    REAL LAMBDA, PHI, ALPHA, PI, RAD, DEG, REARTH, R
!
!  Fixed values.  
!
    PARAMETER (PI = 3.14159265)
    PARAMETER (RAD = PI / 180.) 
    PARAMETER (DEG = 1. / RAD)
    PARAMETER (REARTH = 6370.)
    REAL PHINUL
    PARAMETER (PHINUL = 60. * RAD)
!
!***FIRST EXECUTABLE STATEMENT GR2SPH.
    
    if(index(grid_projection,"utm") > 0) then
        long = xgrid + x0
        latt = ygrid + y0        
        if(grid_projection=="utm33") call utm2geo(long, latt, reverse=.false., zone=33)
        if(grid_projection=="utm32") call utm2geo(long, latt, reverse=.false., zone=32)
        return
    end if   
!    
!  Error treatment.
!
    IFAIL = 0
    IF ((XGRID .EQ. XPOLE) .AND. (YGRID .EQ. YPOLE)) THEN
        IFAIL = 1
        WRITE(*,*) '*** ERROR IN GR2SPH -- North Pole is illegal ***'
        GOTO 999
    ENDIF
    IF ((DX .EQ. 0.) .OR. (DY .EQ. 0.)) THEN
        IFAIL = 2
        WRITE(*,*) '*** ERROR IN GR2SPH -- DX or DY is not pos. ***'
        GOTO 999
    ENDIF
!
!  Convert YLONG to radians.
!
    ALPHA = YLONG * RAD
!
!  Longitude.
! 
    LAMBDA = ALPHA + ATAN2(DX*(XGRID-XPOLE), -DY*(YGRID-YPOLE))
!
!  Lattitude.
!
    R = SQRT(DX*DX*(XGRID-XPOLE)*(XGRID-XPOLE) + DY*DY*(YGRID-YPOLE)*(YGRID-YPOLE))
!
    PHI = .5*PI - 2.*ATAN(R / (REARTH*(1+SIN(PHINUL))))
!
!  Convert to degrees.
!
    LONG = LAMBDA * DEG
    LATT = PHI * DEG
!
!  Normalise longitude.
!
    LONG = LONG - 360. * NINT( LONG/360.)

999 RETURN
!
!***END GR2SPH
!
  END SUBROUTINE GR2SPH


  
  SUBROUTINE SPH2GR(LONG, LATT, XPOLE, YPOLE, DX, DY, YLONG,XGRID, YGRID, IFAIL)
    IMPLICIT NONE
!
!***BEGIN PROLOGUE SPH2GR
!***DATE WRITTEN   910104 (YYMMDD)
!***REVISION DATE  910104 (YYMMDD)
!***AUTHOR
!            Bjorn Adlandsvik
!            Institute of Marine Research,
!            Postboks 1870 Nordnes
!            N-5024 Bergen, Norway
!            Email..  bjorn@imr.no
!
!  The code is based on an the routine BLXY from 
!  The Norwegian Meteorological Institute.
!
!***PURPOSE
!  Converts from spherical coordinates (lattitude and longitude)
!  to polar stereographic grid coordinates.
!
!***PARAMETERS
!***ON ENTRY
!     LONG   Real,	
!  	 Longitude in (decimal) degrees.
!     LATT   Real, 
!        Lattitude in (decimal) degrees,
!        Requirement: -90. < LATT <= 90.    
!     XPOLE  Real,
!        X grid coordinate of the North Pole
!     YPOLE  Real,
!        Y grid coordinate of the North Pole.
!     DX     Real,
!        Width in kilometers of grid cell in X direction.
!     DY     Real,
!        Width in kilometers of grid cell in Y direction.
!     YLONG  Real,
!        Longitude in (decimal) degrees of meridian (anti)parallell 
!        to the Y-axis. 
!
!***ON RETURN
!
!     XGRID  Real,
!        X grid coordinate. 
!     YGRID  Real,
!        Y grid coordinate.
!     IFAIL  Integer,
!        Error indicator, returns
!        2 if DX <= 0 or DY <= 0,
!        1 if illegal value of LATT,
!        0 otherwise.
!
!***DESCRIPTION
!  
!  Spherical coordinates:
!  These are ordinary longitude and lattitude in decimal degrees. 
!  Decimal degrees means that the angle is described as a decimal 
!     number, i.e. 64.5 is used for 64 degrees and 30 minutes.
!  Positive values of longitude means easterly and negative westerly.
!     All values are allowed and are treated modulo 360.
!  Positive values of lattitude means northern and negative southern.
!     The values are restricted to the intervall (-90., 90].
!     The South Pole is not included since the polar stereographic
!     projection is not defined there. The North Pole is treated
!     correctly with any value of longitude.
! 
!  Grid coordinates:
!  The grid coordinate system is based on the polar stereographic
!     projection, "true" at 60 degrees north. This longitude can be 
!     changed by varying the constant PHINUL below.
!  The grids origin in the plane, its orientation and the scaling
!     of the axes are given by XP,YP,YLONG,DX,DY above.
!  The grid coordinates of the center of the (i, j)-th grid cell
!     are (i-0.5, j-0.5)
!
!  The projected radius r is computed by the formula
!     r = rearth * cos(phi) * (1 + sin(phinul)) / (1 + sin(phi))
!
!  Then the grid coordinate are computed as follows
!     xgrid = xpole + r * sin(lambda-alpha) / dx
!     ygrid = ypole - r * cos(lambda-alpha) / dy
!
!***LOCAL VARIABLES
!
!     LAMBDA		Real
!        lattitude in radians.
!     PHI		Real
!        Longitude in radians.
!     ALPHA		Real
!        YLONG in radians.
!     RAD		Real
!        Conversion factor from degrees to radians.
!     REARTH		Real
!        Radius (in kilometers) of Earth.
!     PHINUL
!        Lattitude (in radians) of intersecting plane =
!        lattitude of "true" length.
!     R 		Real
!        Distance (in kilometers) from North Pole in the plane.
!
!***REMARKS
!  Parameters from some often used grids : 
!                           xpole, ypole,  dx,  dy, ylong
!  15570 grid :              181., 128.5, 20., 20.,  58.
!  Barents Sea grid (95,95):  90., 128.5, 20., 20.,  58
!  Hindcast grid :            11.,  65. , 75., 75., -32.
!  North Sea grid            352., 256. , 10., 10.,  58.
!
!***RELATED ROUTINES
!  This routine is the inverse of GR2SPH
!
!***ROUTINES CALLED-NONE
!***END PROLOGUE SPH2GR
!
!  Parameters
!
    REAL LONG, LATT, XPOLE, YPOLE, DX, DY, YLONG, XGRID, YGRID
    INTEGER IFAIL
!
!  Local variables.
!
    REAL LAMBDA, PHI, ALPHA, RAD, REARTH, PHINUL, R 
!
!  Fixed values.  
!
    PARAMETER (RAD = 3.14159265 / 180.)
    PARAMETER (REARTH = 6370.)
    PARAMETER (PHINUL = 60. * RAD)

    if(index(grid_projection,"utm") > 0) then
        xgrid = long
        ygrid = latt        
        if(grid_projection=="utm33") call utm2geo(xgrid, ygrid, reverse=.true., zone=33)
        if(grid_projection=="utm32") call utm2geo(xgrid, ygrid, reverse=.true., zone=32)
        xgrid = xgrid - x0
        ygrid = ygrid - y0
        return
    end if
    
!
!***FIRST EXECUTABLE STATEMENT SPH2GR.
!    
!  Error treatment.
!
    IFAIL = 0
    IF ((LATT .LE. -90.) .OR. (LATT .GT. 90.)) THEN
        IFAIL = 1
        WRITE(*,*) '*** ERROR IN SPH2GR -- Illegal value of LATT ***'
        GOTO 999
    ENDIF
    IF ((DX .LE. 0.) .OR. (DY .LE. 0.)) THEN
        IFAIL = 2
        WRITE(*,*) '*** ERROR IN SPH2GR -- DX or DY not positive ***'
        write(*,*)LONG,LATT
        GOTO 999
    ENDIF
!
!  Convert degrees to radians.
!
    LAMBDA = LONG  * RAD
    PHI    = LATT  * RAD
    ALPHA  = YLONG * RAD
!
!  The computations.
!
    R = REARTH * COS(PHI) * (1 + SIN(PHINUL)) / (1 + SIN(PHI))
!
    XGRID = XPOLE + R * SIN(LAMBDA-ALPHA) / DX
    YGRID = YPOLE - R * COS(LAMBDA-ALPHA) / DY
999 RETURN
  END SUBROUTINE SPH2GR

  
  subroutine utm2geo(x,y,reverse,zone)    
! convert coordinate x,y (in place) from utm wgs84 (x,y in meters)
! with given zone(optional, default=33) to geographical coordinates in
! degrees.  if reverse is present and .true. do the inverse operation
! from geo. to utm.    
    implicit none
    real x,y
    integer, optional :: zone
    logical, optional :: reverse
    
    real(kind=8) crdin(2)  ! input coordinate
    real(8) tparin(15)     ! only used for insys >=2
    integer :: insys=1     ! 0=geographic, 1=utm
    integer :: inzone=33   ! e.g. utm zone
    integer :: inunit=4    ! input unit 2=meters,0=radians,4=degrees
    integer :: insph=12    ! spheroid (12=wgs84)
    integer :: ipr=0       ! print error messages? 0=yes, 1=no 
    integer :: jpr=1       ! print projection parameters? 0=yes, 1=no
    integer :: lemsg=6     ! unit for above printouts, 6=stdout
    integer :: lparm=6 
    real(8) :: crdio(2)    ! output coordinate
    integer :: iosys=0     ! output coordinate system (see insys)
    integer :: iozone=33   ! output coordsyst zone (ala inzone) 
    real(8) :: tpario(15)  ! see tparin
    integer :: iounit=2    ! output unit 2=meters,0=radians,4=degrees
    integer :: ln27        ! file unit used by NAD 1927 
    integer :: ln83        ! file unit used by NAD 1983
    character(128) :: fn27 ! names for above units
    character(128) :: fn83 !
    integer length         ! record length for above files
    integer :: iflg        ! error flag, 0=ok

    crdin = (/x,y/)
    
    insys=1 ; iosys=0 ; inunit=2 ; iounit=4
    
    if(present(reverse)) then
        if(reverse)then
            insys=0 ; iosys=1 ; inunit=4 ; iounit=2
        end if
    end if
    
    inzone=33 ;  iozone=33
    if(present(zone)) then
        inzone=zone  ;  iozone=zone      
    end if
    
!    call gtpz0(crdin,insys,inzone,tparin,inunit,insph,ipr,jpr,lemsg,lparm, &
!               crdio,iosys,iozone,tpario,iounit,ln27,ln83,fn27,fn83,length,iflg)

    x = crdio(1) ; y = crdio(2)
    
    if(iflg .ne. 0)  write(*,*) "gtpz0 error, code=",iflg   
  end subroutine utm2geo
  
end module grid_util
