SUBROUTINE INTERNALOPT
!
!  INTERNAL computes the following terms of the
!  moment equations:
!
!   -(g*D**2/RHO0)*Integral from sigma to 0 of
!                  ((dRHO/dx) - (sigma/D)*(dsigma/dx)*(dRHO/dsigma) )
!
!   -(g*D**2/RHO0)*Integral from sigma to 0 of
!                  ((dRHO/dy) - (sigma/D)*(dsigma/dy)*(dRHO/dsigma)
!
!  where g is gravity, D ia dynamic depth and RHO0 reference density,
!  with an explicit technique, using "optimal" values for sigma in the  
!  integration. The routine depends on having a background
!  stratification (RHOBACK) available, which should be as close to the
!  true stratification (RHO) as possible.
!
!  The results are stored in DELTAU and DELTAV.
!
!  The method otherwise is as in Blumberg and Mellor (1987).
!
  USE STATE
  IMPLICIT NONE
  LOGICAL :: FIRSTIPG=.TRUE.
  REAL REFDENS,XFACTOR,YFACTOR
  INTEGER I,J,K,DJ
  REAL, DIMENSION(I1:I2,J1:J2,KB) :: TERM1,TERM2,FINT
  
  DJ=1
  IF(JM==1) DJ=0
  
  REFDENS = 1.0 / RHO0
!
!  IF FIRSTIPG, compute optimal sigmas
!
!  Here: The Terms are computed centrally between the U/RHO-points
!  except for K = 1 as in standard POM
! 
  IF (FIRSTIPG) THEN
      CALL FIND_OPTIMAL_SIGMA()
      FIRSTIPG=.FALSE.
  ENDIF

!
!  Then compute the 'optimal' IPGs using the weights
!
  TERM1 = 0.0
  TERM2 = 0.0
  FINT = 0.0
  DELTAU = 0.0
!
!  Compute TERM1: dRHO/DX
!
  K = 1
  DO I = I_eq_2,I_eq_IM
      DO J = 1,J_eq_JM
          TERM1(I,J,K) = DUM(I,J)*(RHO(I,J,K)-RHO(I-1,J,K))
      END DO
  END DO
  DO K = 2,KB-1
      DO I = I_eq_2,I_eq_IM
          DO J = 1,J_eq_JM
              TERM1(I,J,K) = 0.5*DUM(I,J)*(RHO(I,J,K-1)- &
                   &      RHO(I-1,J,K-1)+RHO(I,J,K)-RHO(I-1,J,K))
          END DO
      END DO
  END DO
!
!  Compute TERM2: (sigma/H)*(dRHO/dsigma)*(dH/dx)
!

  K = 1
  DO I = I_eq_2,I_eq_IM
      DO J = 1,J_eq_JM
          TERM2(I,J,K)=0.0
      END DO
  END DO
!
  DO K = 2,KB-1
      DO I = I_eq_2,I_eq_IM
          DO J = 1,J_eq_JM
              TERM2(I,J,K)=DUM(I,J)*(0.5*(RHO(I-1,J,K-1)- &
                   &      RHO(I-1,J,K)+RHO(I,J,K-1)- &
                   &      RHO(I,J,K))/DZZ(K-1))* &
                   &      (2.0/(D(I-1,J)+D(I,J)))* &
                   &       (D(I,J)-D(I-1,J))
          END DO
      END DO
  END DO
!
!
!  Compute optimal dRHO/dx
!
!
!  Compute the integrand values at the sigma-levels between U-points
!
  DO K=2,KB-1
      DO J=1,J_eq_JM
          DO I=I_eq_2,I_eq_IM
              FINT(I,J,K)=DUM(I,J)*GRAV*0.5*(D(I-1,J)+D(I,J))* &
                   (TERM1(I,J,K)-SOPTU(I,J,K)*TERM2(I,J,K))*MASKSOPTU(I,J,K)
          END DO
      END DO
  END DO
!      
!  The surface half-cell integrals 
!
  DO  J=J_eq_2,JM_mi_1
      DO  I=1,I_eq_IM
          DELTAU(I,J,1)=FINT(I,J,2)*(-ZZ(1))
      END DO
  END DO

!
!  The standard (METVER=1) is to use trapezoidal rule integration in the
!  interior.
!
  DO K=2,KB-1
      DO J=J_eq_2,JM_mi_1
          DO I=1,I_eq_IM
              DELTAU(I,J,K)=DELTAU(I,J,K-1) + DZZ(K-1)*FINT(I,J,K)
          END DO
      END DO
  END DO
!
!  Compute TERM1: dRHO/DY
!
  TERM1 = 0.0
  TERM2 = 0.0
  FINT = 0.0
  DELTAV = 0.0
!
  K = 1
  DO I = 1,I_eq_IM
      DO J = J_eq_2,J_eq_JM
          TERM1(I,J,K) = DVM(I,J)*(RHO(I,J,K)- RHO(I,J-DJ,K))
      END DO
  END DO
  DO K = 2,KB-1
      DO I = 1,I_eq_IM
          DO J = J_eq_2,J_eq_JM
              TERM1(I,J,K) = 0.5*DVM(I,J)*(RHO(I,J,K-1)- &
                   &      RHO(I,J-DJ,K-1)+RHO(I,J,K)-RHO(I,J-DJ,K))
          END DO
      END DO
  END DO
!
!  Compute TERM2: (1/H)*(dRHO/dsigma)*(dH/dy)
!
  DO K = 2,KB-1
      DO I = 1,I_eq_IM
          DO J = J_eq_2,J_eq_JM
              TERM2(I,J,K)=DVM(I,J)*(0.5*(RHO(I,J-DJ,K-1)- &
                   &      RHO(I,J-DJ,K)+RHO(I,J,K-1)- &
                   &      RHO(I,J,K))/DZZ(K-1))* &
                   &      (2.0/(D(I,J-DJ)+D(I,J)))* &
                   &       (D(I,J)-D(I,J-DJ))
          END DO
      END DO
  END DO
!
!  Compute the integrand values at the sigma-levels
!
  DO K=2,KB-1
      DO J=J_eq_2,J_eq_JM
          DO I=1,I_eq_IM
              FINT(I,J,K)=DVM(I,J)*GRAV*0.5*(D(I,J-DJ)+D(I,J))* &
                   (TERM1(I,J,K)-SOPTV(I,J,K)*TERM2(I,J,K))*MASKSOPTV(I,J,K)
          END DO
      END DO
  END DO
!
!  The surface half-cell integrals 
!
  DO J=J_eq_2,J_eq_JM
      DO I=1,I_eq_IM
          DELTAV(I,J,1)=FINT(I,J,2)*(-ZZ(1))
      END DO
  END DO
!
!  The standard (METVER=1) is to use trapezoidal rule integration in the
!  interior.
!
  DO K=2,KB-1
      DO J=J_eq_2,J_eq_JM
          DO I=1,I_eq_IM
              DELTAV(I,J,K)=DELTAV(I,J,K-1) + DZZ(K-1)*FINT(I,J,K)
          END DO
      END DO
  END DO
!
!  Include effect of internal pressure.
!
  XFACTOR = REFDENS/DX
  YFACTOR = REFDENS/DY
  DO K = 1,KB-1
      DO J = J_eq_2,JM_mi_1
          DO I = I_eq_2,IM_mi_1
              DELTAU(I,J,K) = - XFACTOR*DELTAU(I,J,K)*DU(I,J)
              DELTAV(I,J,K) = - YFACTOR*DELTAV(I,J,K)*DV(I,J)
          END DO
      END DO
  END DO

  
CONTAINS

  
  SUBROUTINE FIND_OPTIMAL_SIGMA()
    IMPLICIT NONE
    INTEGER I,J,K,DJ
    REAL, PARAMETER :: EPSILON = 1.0E-10

    DJ=1
    IF(JM==1) DJ=0

    TERM1 = 0.0
    TERM2 = 0.0
    DO K = 1,KB-1
       DO I = 1,I_EQ_IM
          DO J = 1,J_EQ_JM
             MASKSOPTU(I,J,K) = DUM(I,J)
             MASKSOPTV(I,J,K) = DVM(I,J)
          END DO
       END DO
    END DO
!
!  Compute TERM1: dRHO/DX
!
    call tag_ghost(rhoback)
    call update_ghost()

    K = 1
    DO I = I_eq_2,I_eq_IM
        DO J = 1,J_eq_JM
            TERM1(I,J,K)=DUM(I,J)*(RHOBACK(I,J,K)-RHOBACK(I-1,J,K))
        END DO
    END DO
    DO K = 2,KB-1
        DO I = I_eq_2,I_eq_IM
            DO J = 1,J_eq_JM
                TERM1(I,J,K)=0.5*DUM(I,J)*(RHOBACK(I,J,K-1)- &
                     &   RHOBACK(I-1,J,K-1)+RHOBACK(I,J,K)-RHOBACK(I-1,J,K))
            END DO
        END DO
    END DO
!
!  Compute TERM2: (1/H)*(dRHO/dsigma)*(dH/dx)
!
    DO K = 2,KB-1
        DO I = I_eq_2,I_eq_IM
            DO J = 1,J_eq_JM
              IF(DUM(I,J) .NE. 0.) THEN
                TERM2(I,J,K) = (0.5*(RHOBACK(I-1,J,K-1)- &
                     &      RHOBACK(I-1,J,K)+RHOBACK(I,J,K-1)- &
                     &      RHOBACK(I,J,K))/DZZ(K-1))* &
                     &      (2.0/(D(I-1,J)+D(I,J)))* &
                     &       (D(I,J)-D(I-1,J))
              END IF
            END DO
        END DO
    END DO

    SOPTU = 0.
    SOPTV = 0.
    
    DO K = 2,KB-1
        DO I = I_eq_2,I_eq_IM
            DO J = 1,J_eq_JM
                IF (DUM(I,J).NE.0.0) THEN
                    IF (ABS(TERM2(I,J,K)).GT.EPSILON) THEN
                        SOPTU(I,J,K) = TERM1(I,J,K)/TERM2(I,J,K)
                        IF (SOPTU(I,J,K).GT.0.0) THEN
                           MASKSOPTU(I,J,K) = 0.0
                        ENDIF
                        IF(SOPTU(I,J,K).LT.-1.0) THEN
                           MASKSOPTU(I,J,K) = 0.0
                        ENDIF
                    ELSE
                        SOPTU(I,J,K) = 0.5*(ZZ(K-1)+ZZ(K))
                        MASKSOPTU(I,J,K) = 0.0
                    ENDIF
                END IF
            END DO
        END DO
    END DO
!
!   Repeat comput in y for SOPTV
!
    TERM1 = 0.0
    TERM2 = 0.0
!
!  Compute TERM1: dRHO/DY
!
    K = 1
    DO I = 1,I_eq_IM
        DO J = J_eq_2,J_eq_JM
            TERM1(I,J,K)=DVM(I,J)*(RHOBACK(I,J,K)-RHOBACK(I,J-DJ,K))
        END DO
    END DO
    DO K = 2,KB-1
        DO I = 1,I_eq_IM
            DO J = J_eq_2,J_eq_JM
                TERM1(I,J,K)=0.5*DVM(I,J)*(RHOBACK(I,J,K-1)- &
                     &  RHOBACK(I,J-DJ,K-1)+RHOBACK(I,J,K)-RHOBACK(I,J-DJ,K))
            END DO
        END DO
    END DO
!
!  Compute TERM2: (1/H)*(dRHO/dsigma)*(dH/dy)
!
!  First the dRHO/dsigma terms
!
    DO K = 2,KB-1
        DO I = 1,I_eq_IM
            DO J = J_eq_2,J_eq_JM
                IF(DVM(I,J) .NE. 0) THEN
                   TERM2(I,J,K) = (0.5*(RHOBACK(I,J-DJ,K-1)- &
                     &      RHOBACK(I,J-DJ,K)+RHOBACK(I,J,K-1)- &
                     &      RHOBACK(I,J,K))/DZZ(K-1))* &
                     &      (2.0/(D(I,J-DJ)+D(I,J)))* &
                     &       (D(I,J)-D(I,J-DJ))
              END IF
            END DO
        END DO
    END DO

    DO K = 2,KB-1
        DO I = 1,I_eq_IM
            DO J = J_eq_2,J_eq_JM
                IF (DVM(I,J).NE.0.0) THEN
                    IF (ABS(TERM2(I,J,K)).GT.EPSILON) THEN
                        SOPTV(I,J,K) = TERM1(I,J,K)/TERM2(I,J,K)
                        IF (SOPTV(I,J,K).GT.0.0) THEN
                            MASKSOPTV(I,J,K) = 0.0
                        ENDIF
                        IF (SOPTV(I,J,K).LT.-1.0) THEN
                            MASKSOPTV(I,J,K) = 0.0
                        ENDIF
                    ELSE
                        SOPTV(I,J,K) = 0.5*(ZZ(K-1)+ZZ(K))
                        MASKSOPTV(I,J,K) = 0.0
                    ENDIF
                ENDIF
            END DO
        END DO
    END DO
  END SUBROUTINE FIND_OPTIMAL_SIGMA
  
END SUBROUTINE INTERNALOPT
