module intpol
!
! avle@fimm $ grep -i -h -e "^ *subroutine " intpol.f90
!  
! user interface; optional arguments in brackets:
!
! SUBROUTINE MULTIVPROF_INTPOL(F, FILE_LIST [, REGION, WEIGHTING])
!   interpolate from multiple vertical profiles onto 3D distributed array F  
! SUBROUTINE VPROF_INTPOL(F, FILENAME [, REGION, REPORT])
!   interpolate from a single vertical profile onto 3D distributed array F  
! SUBROUTINE Z2SIGMA(ICODE,FELT,FINN3D,ZINN,LINN,UNDEF)
!   interpolate from a (distributed) 3D Z coordinate grid onto the 3D sigma coordinate model grid
! SUBROUTINE IJSLICE2SIGMA(EDGE, SLICE, SLICEDATA, SIND1, F, FIND1, FIND2, DEPTH, LEVELS [, RANGE], INVALID)
!   interpolation from a 2D vertical slice Z coordinate grid onto the corresponding 2D slice sigma grid of the model.
! SUBROUTINE LL2GRID(LLF, LONS, LATS, F [, POINTTYPE, R1, R2])
!   interpolation of 2D array LLF given on a regular lon lat grid onto the model grid array F
! SUBROUTINE GIP(FS, FT, MAP [, METHOD])
!   Interpolation between regular 2D grids both with arbritary spacing.
!   FS(:,:) is the source grid, FT(:,:) target. MAP(dimensions of FT) is an array of mapping objects.  
!   By default interpolation is triangular, but bilinear interpolation
!   can be chosen by giving the optional arg METHOD="bilinear"
! SUBROUTINE GETTRIMAP3(LON, LAT, MAP)  
!   compute MAP containing weights for triangular interpolation from
!   geographical grid LON(:,:),LAT(:,:) to model grid
! SUBROUTINE GETBILINMAP(LONS, LATS, MAP)
!   Compute map and weights for bilinear interpolation from a regular
!   grid of dimension MxN, where longitudes are defined in LONS(M),
!   latitudes in LATS(N), onto the model grid
! SUBROUTINE GET_GIP_MAP(XS, YS, XT, YT, MAP)
!   find the mapping between the source s and target t grid. the mapping is stored in "m".
! SUBROUTINE VSECTINTPOL(VSECT [, F, FU, FV, F2D])
!   F, FU, FV are distributed arrays in eta, u or v points respectively.
!   given a VSECT object find the horizontally interpolated values of
!   the given arrays under each segment. The result is stored in
!   VSECT%F, VSECT%FUN, VSECT%FUT respectively. Assuming FU,FV to be
!   vector components in the X and Y direction, FUN, FUT are the normal
!   and tangential components of FU,FV wrt each linear section of VSECT.
!   F2D is the 2D equivalent to the 3D F.
! SUBROUTINE INITVSECT(VSECT, LABEL, SX, SY)
!   given a list of horizontal coordinates SX,SY construct a VSECT
!   object, which contains a list of short line segments with coordinates,
!   unit direction vectors and weights needed to interpolate values from
!   the model grid onto the slice below the linear sections defined by SX,SY. 
!   LABEL is just a label.
! SUBROUTINE SIGMA2Z(F, FZ, MAP [, LOC])
!   (vertically) interpolate 3d field F on sigma coordinates onto 3d field FZ defined
!   on a z level grid, using the precomputed coefficients in MAP.
!   the horisontal grid locations are assumed identical.
! SUBROUTINE GETMAP_SIGMA2Z(ZDEPTHS, MAP)
!   find interpolation coefficient lambda for all points in a sigma2z mapping
!   fz(zk) = f(map(zk)%k)*map(zk)%lambda + (1-map(zk)%lambda)*f(map(zk)%k+1)
!
! Helge Avlesen <helge.avlesen@uni.no>
!     
  use state, only : dp
  use file_util
  use control_parameters, only : data_path
  use grid_util
  
  private

  public multivprof_intpol, vprof_intpol, ll2grid, ijslice2sigma, z2sigma,&
       zslice2dtosigma3d, mapping_object, gip, get_gip_map, gettrimap3, getbilinmap, &
       vsect_object, vsectintpol, initvsect, gathervsectdata, pointinpoly, &
       s2zmap_object, sigma2z, getmap_sigma2z
 
  interface z2sigma
     module procedure z2sigma_r4, z2sigma_r8
  end interface
  
  
  integer, parameter :: maxvpts=10000

  type profile_type
     integer size                    ! number of data
     real x,y                        ! horisontal position
     character(len=80) file_name     ! name of data file       
     real :: f(maxvpts), z(maxvpts)  ! the data, vertical pos
  end type profile_type


  type mapping_object
     integer :: i=0, j=0        ! in the source grid: closest grid pt "south west" 
     integer :: weight_type=0   ! 0 no weights, 1=lower triangle, 2 upper, 3 bilinear
     real :: w(4) = 0.0         ! the weights
  end type mapping_object
  

  interface gip
! SUBROUTINE GIP(FS, FT, M, METHOD)
!
! Map the 2d field FS onto the field FT on the model grid. FS should
! cover all of FT for a good result. In parallel FT is distributed, so
! FS should cover FT on each block (including ghost) as a minimum.
!
! M is a 2d array of mapping objects, that for each target grid point
! contains the location of the surrounding source grid points and the
! corresponding interpolation weights W.
!
! by default we use triangular interpolation, but bilinear
! interpolation can be chosen by setting the optional argument
! METHOD="bilinear"
!   
     module procedure gipr4dr, gipr8dr
  end interface gip
  
  interface get_gip_map
!   SUBROUTINE GET_GIP_MAP(XS, YS, XT, YT, MAP)
!   find the mapping between the source s and target t grid. the mapping is stored in MAP.
!   XS .and. YS .and./.or. XT .and. YT can optionally be 1D arrays.
      module procedure get_gip_map2d2d, get_gip_map1d2d, get_gip_map2d1d, get_gip_map1d1d
  end interface get_gip_map


! sections are defined in setupfile2.dat as a list of horizontal positions. the
! straight lines between these points are divided into segments of
! length close to 1 grid units. x,y is the sample point in the
! middle of each of these segments. 

  type segment_object
     real :: x=-1.,y=-1.      ! block grid coordinate for the middle of each segment
     real :: dep=-1.          ! the static depth at gx,gy
     real :: len=-1.          ! the length of the segment
     real :: pos=-1.          ! the position of gx,gy measured along the section
     integer :: i=-2,j=-2     ! closest SW cell center for gx,gy
     integer :: iu=-2,ju=-2   ! closest SW u cell for gx,gy
     integer :: iv=-2,jv=-2   ! closest SW v cell for gx,gy
     real :: enx=0.,eny=0.    ! unit normal vector for segment with gx,gy in the middle
     real :: etx=0.,ety=0.    ! unit tangential vector
     real :: w(4)=0.0         ! weights for bilinear interpolation into gx,gy from surrounding eta points
     real :: wu(4)=0.0        ! as above, but for surrounding u points
     real :: wv(4)=0.0        ! as above, but for surrounding v points
     logical :: inblock       ! true if point is on block
  end type segment_object

  type vsect_object
     character(30) label                                  ! the section name
     logical inblock                                      ! if false, no part of the section is on the block
     integer n                                            ! the number of segments
     type(segment_object), allocatable :: seg(:)          ! the segments
     integer np                                           ! the number of processors crossed by segments
     integer, allocatable :: s1(:), s2(:)                 ! the start and end segment on each processor
     integer, allocatable :: proc(:)                      ! the processor id of each range
     real, allocatable, dimension(:,:) :: f,fun,fut       ! storage for interpolated data along segments
     real, allocatable :: f2d(:)
  end type vsect_object


  ! sigma to z helper structure
  type s2zmap_object
     integer :: k=0
     real :: lambda=1.0
  end type s2zmap_object
  
  
contains


  subroutine sigma2z(f, fz, map, loc)
!    
! (vertically) interpolate 3d field F on sigma coordinates onto 3d field FZ defined
! on a z level grid, using the precomputed coefficients in MAP.
! the horisontal grid locations are assumed identical.
!
    use state, only : i1, j1, i_eq_im, j_eq_jm,block,im,jm,message
    implicit none
    real f(i1:,j1:,:), fz(i1:,j1:,:)
    type(s2zmap_object) map(i1:,j1:,:)
    character, optional :: loc
    integer i,j,k,kz,di,dj
    real lam

    if(size(fz,3) .ne. size(map,3)) call message("fz and map must have equal 3rd dimension",2)

    if(present(loc))then
        select case(loc)
        case("u")
            do k=1,size(fz,3)
                do j=1,j_eq_jm
                    do i=1,i_eq_im
                        kz = map(i,j,k)%k 
                        lam = map(i,j,k)%lambda
                        di=1
                        if(i+block%fi-1==im) di=0
                        fz(i,j,k) = 0.5*(lam*(f(i,j,kz)+f(i+di,j,kz)) + (1-lam)*(f(i,j,kz+1)+f(i+di,j,kz+1)))
                    end do
                end do
            end do
            return
        case("v")
            do k=1,size(fz,3)
                do j=1,j_eq_jm
                    do i=1,i_eq_im
                        kz = map(i,j,k)%k 
                        lam = map(i,j,k)%lambda
                        dj=1
                        if(j+block%fj==jm) dj=0
                        fz(i,j,k) = 0.5*(lam*(f(i,j,kz)+f(i,j+dj,kz)) + (1-lam)*(f(i,j,kz+1)+f(i,j+dj,kz+1)))
                    end do
                end do
            end do            
            return
        case default
            call message("sigma2z unknown loc:", 1, sval=(/loc/))
            return
        end select
    end if
    
    do k=1,size(fz,3)
        do j=1,j_eq_jm
            do i=1,i_eq_im
                kz = map(i,j,k)%k 
                lam = map(i,j,k)%lambda
                fz(i,j,k) = lam*f(i,j,kz) + (1-lam)*f(i,j,kz+1)
            end do
        end do
    end do
  end subroutine sigma2z


  
  subroutine getmap_sigma2z(zdepths, map)
!    
! find interpolation coefficient lambda for all points in a sigma2z mapping
!   fz(zk) = f(map(zk)%k)*map(zk)%lambda + (1-map(zk)%lambda)*f(map(zk)%k+1)
!
    use state, only : kb, zz, d, eta, i1, j1, i_eq_im, j_eq_jm
    
    implicit none
    real zdepths(:)
    type(s2zmap_object) map(i1:,j1:,:)
    integer i,j,k,kz
    real sigdepths(kb), depth
    
    do j=1,j_eq_jm
        do i=1,i_eq_im
            
            ! distance from surface to each sigma level in meters
            
            sigdepths(:) = -zz(:)*d(i,j)

            k=1
            do kz=1,size(zdepths)
                
                ! distance from surf to z level in meters
                depth = zdepths(kz) + eta(i,j)

                if(depth < 0.0 .or. depth > d(i,j)) then
                    map(i,j,kz)%k = 1
                    map(i,j,kz)%lambda = -1e20          
                    cycle
                end if
                
                if(depth < sigdepths(k))then
! this kz is before current sigma point. flag that or extrapolate upwards? then next kz
                    map(i,j,kz)%k = k
                    map(i,j,kz)%lambda = 1
                    cycle
                end if

                if(depth > sigdepths(kb-1))then
! kz after last sigma depth, extrapolate or just exit kz loop?
                    map(i,j,kz)%k = kb-1
                    map(i,j,kz)%lambda = 1
                    cycle
                end if

                if(depth > sigdepths(k+1)) then
! increment k until sigdepths surrounds depth
                    do
                        k=k+1
                        if(k >= kb-1) exit
                        if(depth > sigdepths(k) .and. depth <= sigdepths(k+1)) exit
                    end do
                end if
      
                map(i,j,kz)%k = k
                map(i,j,kz)%lambda =  (sigdepths(k+1) - depth)/(sigdepths(k+1)-sigdepths(k))
            end do
        end do
    end do    
  end subroutine getmap_sigma2z

  
  subroutine zslice2dtosigma3d(fz, x, z, f, r)
    implicit none
    real(4) fz(:,:),x(:),z(:)
    real f(:,:,:)
    integer r(6)
    print *,'TODO zslice2dtosigma3d'
  end subroutine zslice2dtosigma3d

  
  subroutine multivprof_intpol(f, file_list, region, weighting)
!
! Initialize the 3d field F, from a set of vertical profiles, given the
! model geometry H and the vertical sigma grid distribution ZZ.(H,ZZ and more
! are passed via module state.)    
! the string FILE_LIST is a filename for a file containing a list of files with
! vertical profiles. The first line in the FILE_LIST file says how many filenames to read.
! The file format for the vertical profile files is simple:
! two columns, the left column is the depth in meters, right col is the corresponding
! profile value. the depth must be increasing for each line.
!
! F is interpolated in space from the vertical profiles. for each
! point, interpolate each vertical profile linearly in the vertical to
! find the value at the current depth level.  then, in the horisontal
! use inverse distance weighted interpolation to find the final value.
! if the depth is outside the range of the profile data, extrapolate: use the
! vertically closest value from the profiles.
!
! the WEIGHTING argument, by default -1.0 (inverse distance squared),
! is the exponent of the square of the distance between the current
! point and a interpolated profile value, used in the weighting. -0.5
! gives 1/distance as weight, -0.25 gives 1/sqrt(distance) as weight. read code for
! further info.    
!
! if REGION is given, it is a vector with 4 integers denoting a horisontal range
!  (istart,iend,jstart,jend). if present, only initialize F in this region.
!
    use state, only : zz, h, im, jm, kb, debug
    use mod_mpi
    implicit none
    real(4) f(i1:,j1:,:)      
    character(len=*) file_list
    real(4), optional :: weighting
    integer, optional :: region(4)

    type(profile_type), allocatable :: profile(:)
    real, allocatable :: intpolf(:), dist(:)
    integer i,j,k,n, gi1,gi2, gj1,gj2, li1,li2, lj1,lj2,gi,gj
    real depth,x,y, wm
    integer ip, numw, nm, count, botloc, bufsize
    real sumw, weight, topval, botval, topz, botz
    logical inblock
    character, allocatable :: buffer(:)
 
    gi1=1 ; gi2=im ; gj1=1 ; gj2=jm
    if(present(region))then
        gi1=region(1) ; gi2=region(2) ; gj1=region(3) ; gj2=region(4)
    end if

    call globalrect2local(gi1,gi2,gj1,gj2,li1,li2,lj1,lj2,inblock)

! perhaps read this in setup2?    
    wm = -0.5
    if(present(weighting)) then
        wm=weighting
    end if

    if(master)then
        call checkunit(1000003, 'multi_vprof_intpol profile list')    
        open(1000003, file=(trim(data_path)//file_list), status='old', iostat=ierr)
        if(ierr.ne.0)then
            write(*,*)'error: could not open file with list of profiles:',trim(file_list)
            stop
        end if
    
        read(1000003,*) n
    end if

    call broadcast(n)    
    allocate(profile(n), intpolf(n), dist(n))

    if(master)then        
        do i=1,n
            read(1000003,*) profile(i)%x, profile(i)%y, profile(i)%file_name
        end do
        close(1000003)
    
        topval=0.0 ; botval=0.0 ; topz=1e20 ; botz=-1e20
    
        do i=1,n
            call checkunit(1000003, 'multi_vprof_intpol vprof read')

            open(1000003, file=(trim(data_path)//profile(i)%file_name), status='old', iostat=ierr)
            if(ierr.ne.0)then
                write(*,*)'error: could not open vertical profile file:',profile(i)%file_name
                close(1000003)
                stop
            end if
            count=1
            do
                read(1000003,*,end=9000) profile(i)%z(count), profile(i)%f(count)
                count=count+1
                if(count > maxvpts) then
                    write(*,*)'error in bom_input.f90: maxvpts=',maxvpts,' too small. increase and recompile.'
                    return
                end if
            end do
9000        continue
            close(1000003)

! store the number of samples in each profile        
            profile(i)%size = count-1

! find f at the highest and the lowest data point
            botloc = profile(i)%size
        
            if(profile(i)%z(botloc) > botz) then
                botz   = profile(i)%z(botloc)
                botval = profile(i)%f(botloc)
            end if
            if(profile(i)%z(1) < topz) then
                topz   = profile(i)%z(1)
                topval = profile(i)%f(1)
            end if
        
            if(debug>0) write(*,*)'profile ',trim(profile(i)%file_name),' items=',count,'pos=',profile(i)%x, profile(i)%y,&
                 'max:',maxval(profile(i)%f(1:count-1)),'min:',minval(profile(i)%f(1:count-1))
        end do
    end if

    call mbroadcast(botz, botval, topz, topval)
    

!! broadcast of derived type can be done like this:
! transfer the derived type into a character array, first find the number of characters
    bufsize = size(transfer(profile(1:n),(/'x'/)))
    allocate(buffer(bufsize))
! then transfer the array of objects into the buffer and broadcast
    if(master)then
        buffer = transfer(profile(1:n),buffer)
    end if
    call broadcast(buffer)
! then transfer back    
    if(.not.master)then
        profile(1:n) = transfer(buffer, profile(1:n))
    end if
    deallocate(buffer)
    

    if(inblock)then
       do k=1,size(f,3)
          do j=lj1,lj2
             do i=li1,li2
                if(h(i,j)>0.0)then
                    depth = -zz(k)*h(i,j)

                    ! if current depth is outside data range, extrapolate
                    if(depth < topz)then
                        f(i,j,k) = topval
                        cycle
                    end if
                    if(depth > botz)then
                        f(i,j,k) = botval
                        cycle
                    end if

                    gi = block%fi + i - 1
                    gj = block%fj + j - 1
                    
                    x= gi-0.5 ;  y= gj-0.5

                    !! find distance to vertical profiles and use lin interpolation vertically in
                    !! each profile to find values at the current depth 
                    do ip = 1,n
                        dist(ip) = (x-profile(ip)%x)**2 + (y-profile(ip)%y)**2
                        nm = profile(ip)%size
                        intpolf(ip) = vlinintpol(depth, profile(ip)%f(1:nm), profile(ip)%z(1:nm) )
                    end do
                    
                    f(i,j,k) = 0.0
                    numw = 0
                    sumw = 0.
                    do ip = 1,n
                        if (intpolf(ip) > -9000) then
                            numw = numw + 1
                            weight = dist(ip)**wm
                            sumw = sumw + weight
                            f(i,j,k) = f(i,j,k) + weight*intpolf(ip)
                        endif
                    end do
                    
                    if(numw>0) f(i,j,k) = f(i,j,k)/sumw
                end if
            end do
        end do
    end do
   end if
   deallocate(profile, intpolf, dist)
  end subroutine multivprof_intpol



  subroutine vprof_intpol(f, filename, cols, region, report)
!
! given a file FILENAME, with two colums of e.g. in situ data, where
! the left column is the depth in meters(increasing), and the right
! column is the corresponding value of a field, interpolate these
! values onto the model grid field f.  optionally specify a horisontal
! region=(i1,i2,j1,j2) where the field f should be initialized (all
! other points of f will remain untouched) if report=.true. print the
! profile in the deepest point to screen.
! if cols is present, the first item is the column with x the other z.
!
! example usage:
!
!   use intpol
!   ...
!   region1=(/1,10,1,10/)       ! the "south west" corner
!
!   call vprof_intpol(S, 'salinity_reg1.dat', region=region1 )
!   call vprof_intpol(T, 'temperature.dat')
!
    use state
    implicit none
    character(*) :: filename
    real(4) f(1-ghost_wi:,1-ghost_wj:,:)
    integer, optional :: region(4), cols(2)
    real(4), dimension(10000) :: depth, field
    real target_depth,thickness,val
    logical, optional :: report
    integer i,j,k,l,count,is,ie,js,je, loc(3), usecols(2)
    integer previous_start, r(4)
    logical inblock

    inblock=.true.
    is=1 ; ie=i_eq_im ; js=1 ; je=j_eq_jm
    if(present(region))then
        r = region
        call globalrange2local(block%ibn,block%jbn, r, inblock)
        is=r(1) ; ie=r(2) ; js=r(3) ; je=r(4)
    end if

    if(master)then    
        usecols=(/1,2/)
        if(present(cols)) usecols=cols
        call get_fields(filename, depth, field, cols=usecols, lines=count)

        if(maxval(depth(1:count)) <= 0.0) &
             call abort("vprof_intpol: all depths < 0. assumed positive, with increasing value downwards.")
    end if
    
    call broadcast(count)
    call broadcast(depth(1:count))    
    call broadcast(field(1:count))
    
    if(inblock)then
        
    do j=js,je
        do i=is,ie
            if(h(i,j) > 0.0) then
                previous_start=1
                do k=1,kb-1
                    target_depth= -(zz(k)*h(i,j))
                    if(target_depth < depth(1)) then
                        f(i,j,k)=field(1)
                        cycle
                    end if
                    if(target_depth >= depth(count)) then
                        f(i,j,k)=field(count)
                        cycle
                    end if
                    do l=previous_start,count-1
                        if(target_depth >= depth(l) .and. target_depth < depth(l+1)) then
                            thickness=depth(l+1)-depth(l)
                            f(i,j,k)=(field(l+1)*( target_depth-depth(l) )+field(l)*( depth(l+1)-target_depth ))/thickness
                            previous_start = l
                            exit
                        end if
                    end do
                end do
            end if
        end do
    end do

    end if ! inblock

    if(debug > 0) then
        loc(1:2) = depmaxloc(1:2)
        call globalind2local(loc(1), loc(2), inblock, ghost=.false.)
        if(inblock)then
            write(*,*) 'the vertical profile of ',filename,' under point ',loc
            do k=1,kb-1
                write(*,*) zz(k)*h(loc(1),loc(2)), f(loc(1),loc(2),k)
            end do
        end if
        call barrier()
    end if
  end subroutine vprof_intpol
  


  function vlinintpol(depth, f, z)
!
! given the vertical profile f(z), z(k+1)>z(k) for all k,
! find f at depth
!
    use util, only : binsearch
    implicit none
    real vlinintpol, depth, f(:), z(:)
    integer k

    if(size(f).ne.size(z))then
        write(*,*)'vlinintpol error: size(f)=',size(F),'size(z)=',size(z)
        stop
    end if
    k = binsearch(depth, z)
    if(k>0 .and. k<=size(f)) then
        vlinintpol = f(k-1) + (f(k)-f(k-1))/(z(k)-z(k-1)) * (depth-z(k-1))
    else
        vlinintpol = -9999.0
    end if
  end function vlinintpol

  subroutine z2sigma_r8(icode,felt,finn3d,zinn,linn,undef)
    use state
    implicit none
    integer, parameter :: localkind=8

    include "z2sigma.f90"
  end subroutine z2sigma_r8
    
  subroutine z2sigma_r4(icode,felt,finn3d,zinn,linn,undef)
    use state
    implicit none
    integer, parameter :: localkind=4
    
    include "z2sigma.f90"
  end subroutine z2sigma_r4


  subroutine ijslice2sigma(edge, slice, slicedata, sind1, f, find1, find2, depth, levels, range, invalid)
!
! same algoritm as z2sigma above, but instead interpolate from a vertical
! z-coordinate 2d slice onto a 3d sigma coordinate array, optionally
! in the given region.
!    
    use mod_mpi, only : i_eq_im, j_eq_jm, myid, block, east, west, north, south, abort
    use state, only : zz, kb, debug
    implicit none
    integer edge
    character slice    
    integer sind1
    real slicedata(sind1:,:)
    integer find1,find2
    real f(find1:,find2:,:)
    real depth(find1:, find2:), levels(:)
    integer, optional :: range(:)
    real invalid
    integer i,j,k,r(6),numlevels,ll,lbot,sind
    real :: zdyp
    logical fail
! neighbour offsets
    integer no(2), numvalid, n, newi,newj, ugly
    real :: upper, lower, eps=1e-2

    ! offsets to neighbour points. the 4 first are the closest. 
    no = (/-1,1/)
    ugly = 0
    
    r = (/1,i_eq_im, 1,j_eq_jm, 1, kb-1/)
    if(present(range)) r=range

    numlevels = size(levels)
    
    do j=r(3),r(4)
        do i=r(1),r(2)
            if(depth(i,j) <= 0.0) cycle

            ! assume point and edge on same block
            select case(edge)
            case(east)
                if(depth(i_eq_im,j) <= 0.0) cycle
            case(west)
                if(depth(1,j) <= 0.0) cycle
            case(north)
                if(depth(i,j_eq_jm) <= 0.0) cycle
            case(south)
                if(depth(i,1) <= 0.0) cycle
            end select
            
            select case(slice)
            case("i")
                sind = j
            case("j")
                sind = i
            end select

            lbot = -1
            
            ! shallower than the first z data layer. 
            !if(levels(1) > depth(i,j)) lbot=1
            
            ! find lowermost Z data point LBOT:
            do ll = 1, numlevels
!                if (levels(ll) < depth(i,j) .and. abs(slicedata(sind,ll)-invalid) > eps) lbot = ll
                if (levels(ll) < depth(i,j) .and. slicedata(sind,ll) > -10 .and. slicedata(sind,ll) < 50.) lbot = ll                
            end do

            if(lbot == -1) then
                write(*,*) myid, "ijslice2sigma:"
                write(*,*) myid, "ERROR: NO, as in 0, useful Z data points found in water column. Congratulations."
                write(*,*) myid, "gi,gj=", block%fi+i-1,block%fj+j-1, "H(i,j)=", depth(i,j)
                write(*,*) myid, "slicedata(sind,:)=", slicedata(sind,:)
                write(*,*) myid, "levels=",levels
                call abort()
            end if            
                    
            do k = r(5), r(6)
                zdyp = - zz(k) * depth(i,j)
                
                if (zdyp <= levels(1)) then
!                    if(abs(slicedata(sind,1) - invalid) < eps) call abort("invalid top z level")
                    if(slicedata(sind,1) < -10 .and. slicedata(sind,1) > 50.) call abort("invalid top z level")                    
                    
                    f(i,j,k) = slicedata(sind,1)
                    
                elseif (zdyp > levels(lbot)) then
                    
                    if(lbot < numlevels)then
!                        if(abs(slicedata(sind,lbot+1) - invalid) > eps) then
                        if(slicedata(sind,lbot+1) > -10 .and. slicedata(sind,lbot+1) < 50.) then                        
                            f(i,j,k) = slicedata(sind,lbot) + (slicedata(sind,lbot+1)-slicedata(sind,lbot)) &
                                 &      * (zdyp-levels(lbot)) / (levels(lbot+1)-levels(lbot))
                        end if                    
                    end if

                    ! if lbot=numlevels it is actually "safe" to extrapolate
                    f(i,j,k) = slicedata(sind, lbot)
                    
                    !numvalid=0 ;  upper = 0.0  ;  lower = 0.0
                    !do n=1,2
                    !    if(slicedata(sind+no(n),lbot+1) > invalid .and.  &
                    !       slicedata(sind+no(n),lbot)   > invalid) then
                    !        numvalid = numvalid + 1
                    !        upper = upper + slicedata(sind+no(n),lbot)
                    !        lower = lower + slicedata(sind+no(n),lbot+1)
                    !    end if
                    !end do
                    !
                    !if(numvalid > 0)then
                    ! ! find the average of the valid, surrounding z points
                    !    upper = upper/numvalid
                    !    lower = lower/numvalid
                    !    f(i,j,k) = upper + (lower-upper)* (zdyp-levels(lbot)) / (levels(lbot+1)-levels(lbot))
                    !    cycle
                    !else
! call abort("target sigma depth below last valid Z data point")
                    !    f(i,j,j) = slicedata(sind, numlevels)
                    !    ugly = ugly + 1
                    !end if
                    
                else
                    do ll = 1, lbot-1
                        if ((zdyp .gt. levels(ll)) .and. (zdyp .le. levels(ll+1))) then
                            f(i,j,k) = slicedata(sind,ll) + (slicedata(sind,ll+1)-slicedata(sind,ll)) &
                                 &      * (zdyp-levels(ll)) / (levels(ll+1)-levels(ll))
                        endif
                    end do
                endif

                if(f(i,j,k) < -10 .or. f(i,j,k) > 50) then
                    write(*,*) "ijslice2sigma error: f=",f(i,j,k),"seems out of range"
                    write(*,*) "gi,gj,k=", block%fi + i -1, block%fj + j -1, k
                    write(*,*) "edge=",edge
                    write(*,*) "h(i,j)=",depth(i,j)
                    write(*,*) "levels=",levels
                    write(*,*) "lbot=",lbot," sind=",sind," zdyp=",zdyp
                    write(*,*) "slicedata(sind)=", slicedata(sind,:)
                    write(*,*) "slicedata(sind+1)=", slicedata(sind+1,:)
                    call abort()
                end if
            end do
        end do
    end do
    if(debug > 0 .and. ugly > 0) then
        write(*,*) myid, "ijslice2sigma: check z data, ",ugly," extrapolations was done"
    end if
  end subroutine ijslice2sigma
  
  SUBROUTINE LL2GRID(LLF, LONS, LATS, F, POINTTYPE, R1, R2)
!    
! Bilinear interpolation of the lon,lat gridded data in LLF, onto the model grid array F.
! The first dimension of LLF is lon, lat the second.
! LLF is assumed small, in order to have it broadcast to all processors.
! If LLF is not small, it may be faster to interpolate on master and then scatter F.
! By default pointtype is 0 (eta point, i.e. center of box). For u points use 1, v points 2.
!
! If r1 and r2 are present the routine calls abort if F is not in the expected range.
!    
    use util, only : binsearch
    use mod_mpi, only : master, i1,i2,j1,j2,block,abort,fci1,fci2,fcj1,fcj2
    use state, only : xpole,ypole,ylong,dx,dy,longitude,latitude,debug
    implicit none
    integer kdim
    real llf(:,:), lons(:), lats(:), f(i1:i2,j1:j2), gxpole, gypole,lon,lat
    integer, optional :: pointtype
    real, optional :: r1, r2
    logical validate
    integer i,j,k,ifail,type,londim,latdim,lonind,latind
    real x,y,lx,ly

    validate=.false.
    if(present(r1) .and. present(r2)) validate=.true.
    
    londim = size(llf,1)
    latdim = size(llf,2)
    
    if(londim .ne. size(lons) .or. latdim.ne.size(lats)) then
        if(master) write(*,*) "ll2grid error: size of lons,lats does not correspond to size of llf"
    end if
    
    type = 0
    if(present(pointtype)) type=pointtype

    !! find the coordinate of north pole in grid coordinates
    gxpole = (xpole+dx)/dx
    gypole = (ypole+dy)/dy

    f = 0.0
    
    do j=fcj1,fcj2
        do i=fci1,fci2
            !! find the global grid coordinate x,y for u,v or p points
            select case(pointtype)
            case(0)
!  optimization question: what is faster of calling gr2sph for each
!  grid point versus using precomputed values from 2 2d arrays? figure out later.
!                x = block%fi -1 + i-0.5
!                y = block%fj -1 + j-0.5
!                call gr2sph(x, y, gxpole, gypole, dx*0.001, dy*0.001, ylong, lon, lat, ifail)
                lon = longitude(i,j)
                lat = latitude(i,j)
            case(1)
                x = block%fi -1 + i-1
                y = block%fj -1 + j-0.5
                call gr2sph(x, y, gxpole, gypole, dx*0.001, dy*0.001, ylong, lon, lat, ifail)
! the following alternative requires a ghost update. not clear what is
! most expensive w.r. doing gr2sph over and over again.
!                lon = 0.5*(longitude(i,j) + longitude(i-1,j))
!                lat = latitude(i,j)
            case(2)
                x = block%fi -1 + i-0.5
                y = block%fj -1 + j-1
                call gr2sph(x, y, gxpole, gypole, dx*0.001, dy*0.001, ylong, lon, lat, ifail)
!                lon = longitude(i,j)
!                lat = 0.5*(latitude(i,j) + latitude(i,j-1))
            case default
                stop
            end select

            !! search for the indices of surrounding longitude, latitude
            lonind = binsearch(lon, lons) - 1
            latind = binsearch(lat, lats) - 1
            
            if(lonind < 1 .or. lonind >= londim .or. latind < 1 .or. latind >= latdim) then
                write(*,*) 'i,j',i,j, 'lon,lat', longitude(i,j),latitude(i,j), lon, lat, lonind, latind
                call abort("ll2grid error: lon,lat data does not cover grid, cannot interpolate")
            end if
            !! normalized local coordinate
            lx = (lon - lons(lonind))/(lons(lonind+1)-lons(lonind))
            ly = (lat - lats(latind))/(lats(latind+1)-lats(latind))

            if(lx > 1 .or. lx < 0 .or. ly > 1 .or. ly < 0) &
                 call abort("ll2grid error, local normalized coordinate out of range")

            f(i,j) = llf(lonind,latind)*(1-lx)*(1-ly) + llf(lonind+1,latind)*lx*(1-ly) + &
                 llf(lonind,latind+1)*(1-lx)*ly + llf(lonind+1,latind+1)*lx*ly

            if(validate)then
                if(f(i,j) < r1 .or. f(i,j) > r2) then
                    write(*,*) "F=",f(i,j)," i,j=",i,j," r1,r2=", r1,r2
                    write(*,*) "llf   llf_east   llf_north  llf_northeast:"
                    write(*,*) llf(lonind,latind), llf(lonind+1,latind), llf(lonind,latind+1),llf(lonind+1,latind+1)
                    call abort("ll2grid error: interpolated F not in the expected range")
                end if
            end if
            
        end do
    end do
    if(debug>2.and.master) write(*,*) '   ll2grid: max,min f=',maxval(f),minval(f)
  end subroutine ll2grid
  
  
  subroutine gipr4dr(fs, ft, m, method)
    use mod_mpi, only : i1,j1,fci1,fci2,fcj1,fcj2
    implicit none
    integer, parameter :: localkind=4
    include "gip.f90"
  end subroutine gipr4dr
  
  subroutine gipr8dr(fs, ft, m, method)
    use mod_mpi, only : i1,j1,fci1,fci2,fcj1,fcj2
    implicit none
    integer, parameter :: localkind=8
    include "gip.f90"
  end subroutine gipr8dr

  
  subroutine get_gip_map2d2d(xs,ys,xt,yt,m)
! find the mapping m between the source s and target t grid.
! for each point i,j in t, the map gives the source point m,n ;
! the bottom left corner of the source box containing target point XT(I,J),YT(I,J).
! Note that XS .and. YS .and./.or. XT .and. YT can optionally be 1D arrays.
! the box is furter split into triangles t; 1=lower 2=upper.
    implicit none
    real, dimension(:,:) :: xs,ys,xt,yt
    type(mapping_object) m(:,:)
    integer ind(3), i,j,is,js,idim,jdim,t,range(4)
    real(8) x(2),x1(2),x2(2),x3(2),C
    real, dimension(2*(size(xs,1)+size(xs,2))) :: xp,yp

    idim = size(xs,1)
    jdim = size(xs,2)

! for each target grid point find the surrounding source grid
! point i,j and triangle
    
    do j=1,size(xt,2)
        do i=1,size(xt,1)
            range = (/1,idim,1,jdim/)
            
            ind = search((/xt(i,j), yt(i,j)/), r=range)
            
            m(i,j)%i = ind(1)
            m(i,j)%j = ind(2)
            m(i,j)%weight_type = ind(3)
        end do
    end do

! for each target grid point X, compute the interpolation weights W for
! the function values located at each surrounding triangle vertice
    
    do j=1,size(xt,2)
        do i=1,size(xt,1)
            if(m(i,j)%weight_type == 0) cycle

            x = (/xt(i,j), yt(i,j)/)

            is = m(i,j)%i
            js = m(i,j)%j
            x1 = (/xs(is,js),    ys(is,js)/)
            x3 = (/xs(is+1,js+1),ys(is+1,js+1)/)
            
            if(m(i,j)%weight_type == 1) then
                x2 = (/xs(is+1,js),ys(is+1,js)/)   ! lower
            else
                x2 = (/xs(is,js+1),ys(is,js+1)/)   ! upper
            end if
            
            m(i,j)%w(1) = x(1)*(x3(2)-x2(2)) + x(2)*(x2(1) - x3(1)) + x3(1)*x2(2) - x2(1)*x3(2)
            m(i,j)%w(2) = x(1)*(x1(2)-x3(2)) + x(2)*(x3(1) - x1(1)) + x1(1)*x3(2) - x3(1)*x1(2)
            m(i,j)%w(3) = x(1)*(x2(2)-x1(2)) + x(2)*(x1(1) - x2(1)) + x2(1)*x1(2) - x1(1)*x2(2)
            
            C = -1._8/(x1(1)*(x2(2) - x3(2)) + x2(1)*(x3(2) - x1(2)) + x3(1)*(x1(2) - x2(2)))
            m(i,j)%w = m(i,j)%w*C
        end do
    end do
    
  contains   
    
    recursive function search(x,r) result(s)
!
! given a range r=(/i1,i2,j1,j2/) find s=(/i,j,t/) where
! pointinpoly(x)=.true., using recursive 2d bisection. return 0,0,0 on
! failure.
!
      implicit none
      real x(2)
      integer r(4), s(3), n, mp(2), res(4,3),i,j,range(4)

     ! by default return nothing
      s = 0

      if(r(2)-r(1)==0 .or. r(4)-r(3)==0) return      
      if(r(2)-r(1)<0 .or. r(4)-r(3)<0) then ; print *,'WTF! r=',r ; stop ; end if
      
     ! assemble a polygon that goes counterclockwise around the edge of the
     ! box defined by r. unfortunately this is not always convex, so we
     ! must use the pointinpoly check, which is somewhat slower.
      n=0
      do i=r(1),r(2)-1
          n=n+1
          xp(n)=xs(i,r(3))
          yp(n)=ys(i,r(3))
      end do
      do j=r(3),r(4)-1
          n=n+1
          xp(n)=xs(r(2),j)
          yp(n)=ys(r(2),j)
      end do
      do i=r(2),r(1)+1,-1
          n=n+1
          xp(n)=xs(i,r(4))
          yp(n)=ys(i,r(4))
      end do
      do j=r(4),r(3),-1
          n=n+1
          xp(n)=xs(r(1),j)
          yp(n)=ys(r(1),j)
      end do
      
      if(pointinpoly(x, xp, yp, n, .true.) == 0) then
          
! the new search areas should not be smaller than a single box in
! either direction.
          
          if(r(2)-r(1)==1 .and. r(4)-r(3)==1) then
            !! we got the box. now test the lower triangle:
              xp(1)=xs(r(1),r(3))  ; yp(1)=ys(r(1),r(3))
              xp(2)=xs(r(2),r(3))  ; yp(2)=ys(r(2),r(3))
              xp(3)=xs(r(2),r(4))  ; yp(3)=ys(r(2),r(4))
              if(pointinpoly(x, xp, yp, 3, .false.) == 0) then
                  s = (/r(1),r(3),1/)
              else
                  s = (/r(1),r(3),2/)
              end if
              return
          end if
          
      ! otherwise, search deeper, new bisection mid point
          mp = (/(r(1)+r(2))/2, (r(3)+r(4))/2/)
      ! and a recursive search into each resulting quadrant
          range = (/r(1),mp(1),r(3),mp(2)/)
          res(1,:) = search(x, range)
          range = (/mp(1),r(2),r(3),mp(2)/)
          res(2,:) = search(x, range)
          range = (/mp(1),r(2),mp(2),r(4)/)
          res(3,:) = search(x, range)
          range = (/r(1),mp(1),mp(2),r(4)/)
          res(4,:) = search(x, range)
          
          ! return with the first quadrant where we have a hit
          do i=1,4
              if(res(i,1) > 0) then
                  s = res(i,:)
                  return
              end if
          end do
      end if
    end function search
  end subroutine

! sometimes the coordinates are given as 1d arrays.
  
  subroutine get_gip_map2d1d(xs,ys,xt,yt,m)
    implicit none
    real, dimension(:,:) :: xs,ys
    real, dimension(:) :: xt,yt
    type(mapping_object) m(:,:)
    real, dimension(size(xt,1),size(yt,1)) :: xt2d, yt2d
    integer i,j

    do j=1,size(yt,1)
        do i=1,size(xt,1)
            xt2d(i,j) = xt(i)
            yt2d(i,j) = yt(j)
        end do
    end do
    call get_gip_map2d2d(xs,ys,xt2d,yt2d,m)    
  end subroutine get_gip_map2d1d

  subroutine get_gip_map1d2d(xs,ys,xt,yt,m)
    implicit none
    real, dimension(:) :: xs,ys
    real, dimension(:,:) :: xt,yt
    type(mapping_object) m(:,:)
    real, dimension(size(xs,1),size(ys,1)) :: xs2d, ys2d
    integer i,j

    do j=1,size(ys,1)
        do i=1,size(xs,1)
            xs2d(i,j) = xs(i)
            ys2d(i,j) = ys(j)
        end do
    end do
    call get_gip_map2d2d(xs2d,ys2d,xt,yt,m)    
  end subroutine get_gip_map1d2d

  subroutine get_gip_map1d1d(xs,ys,xt,yt,m)
    implicit none
    real, dimension(:) :: xs,ys, xt,yt
    type(mapping_object) m(:,:)
    real, dimension(size(xs,1),size(ys,1)) :: xs2d, ys2d
    real, dimension(size(xt,1),size(yt,1)) :: xt2d, yt2d
    integer i,j

    do j=1,size(ys,1)
        do i=1,size(xs,1)
            xs2d(i,j) = xs(i)
            ys2d(i,j) = ys(j)
        end do
    end do
    do j=1,size(yt,1)
        do i=1,size(xt,1)
            xt2d(i,j) = xt(i)
            yt2d(i,j) = yt(j)
        end do
    end do
    call get_gip_map2d2d(xs2d,ys2d,xt2d,yt2d,m)    
  end subroutine get_gip_map1d1d
  
  integer function pointinpoly(point, x, y, nvert, isclosed)
!    
! Determine whether a point is inside a polygon;
! A vertical line is drawn through the point in question. If it 
! crosses the polygon an odd number of times, then the 
! point is inside of the polygon. 
!
! INPUT PARAMETERS 
!  point     array length 2 with x,y coordinate of point    
!  x,y       2 vectors, length nvert, containing x and y for each point
!            along the polygon The vertices may be listed clockwise or
!            anticlockwise.  The input polygon may be a compound
!            polygon consisting of several separate subpolygons. If
!            so, the first vertex of each subpolygon must be repeated,
!            and when calculating NVERT, these first vertices must be
!            counted twice.
!  nvert     number of polygon vertices
!  isclosed  if true, last record is equal to first record
!            in poly array
!    
! On return, pointinpoly is 0.0 if point is inside x,y. 1.0 otherwise.       
!
! The method is based on InsidePolygon at
!   http://local.wasp.uwa.edu.au/~pbourke/geometry/insidepoly/
! A similar method is available at
!   http://www.ecse.rpi.edu/Homepages/wrf/Research/Short_Notes/pnpoly.html
! Modified version by Tomas Torsvik, Uni Computing, 2/2011.
!
    implicit none
    real, intent(in) :: point(2)
    real, intent(in) :: x(nvert),y(nvert)
    integer, intent(in) :: nvert
    logical, intent(in) :: isclosed
    real :: xintersect
    integer :: i,j,i1,crosscount

    if(isclosed) then
        i1=2 ;  j=1
    else
        i1=1 ;  j=nvert
    end if

    pointinpoly=0
    
    ! Search for crossings along projections parallel to y-axis,
    ! from point coordinate to +infinity
    crosscount = 0
    do i=i1,nvert
        if(point(1)==x(i) .and. point(2)==y(i)) return
        
        ! Crossings only possible if y-coordinate of point between 
        ! y-coordinates of line segment end points
        if((point(2) > min(y(i),y(j))) .and.  &
             (point(2) <= max(y(i),y(j)))) then
            ! Search towards +infinity
            if(point(1) <= max(x(i),x(j))) then
                ! Exclude the case where the point is on the tangent 
                ! of the line segment
                if(y(i) /= y(j)) then
                    xintersect = (point(2) - y(j))*  &
                         (x(i) - x(j))/(y(i) - y(j))  &
                         + x(j)
                    if((x(i) == x(j)) .or. point(1) <= xintersect) then
                        crosscount = crosscount+1
                    end if
                else if ((point(1) > min(x(i),x(j))) .and. (point(1) <= max(x(i),x(j)))) then
                    return   ! point is inside if it is on the line segment
                end if
            end if
        end if
        j=i
    end do
    
! inside true if crosscount is odd
    if(mod(crosscount,2) == 0) pointinpoly = 1
  end function pointinpoly 


  subroutine gettrimap3(lon, lat, map)
!
! given a arbritrarily spaced regular 2D grid with geographical vertice
! coordinates in lon(:,:),lat(:,:), set up the mapping from this grid
! onto the model grid longitude(:,:), latitude(:,:), by dividing the
! source grid into triangles. this routine computes the connections
! between the source grid/triangles and the target grid.
! 
    use state, only : dx,dy,xpole,ypole,ylong,debug,im,jm
    use mod_mpi, only : i1,j1,i2,j2,fci1,fci2,fcj1,fcj2,block,&
         globalrect2local,i_eq_im,j_eq_jm
    use qmpi, only : rdtsc, clock_frequency, myid, master, reduce, dr, sp
    
    implicit none
    real lon(:,:), lat(:,:)
    type(mapping_object) :: map(i1:,j1:)
    integer i,j,t,d1,d2,li1,li2,lj1,lj2,gimin,gimax,gjmin,gjmax,ifail,numip,is,js,li,lj
    real gxpole, gypole,x,y
    real, dimension(size(lon,1),size(lon,2)) :: gx,gy
    real tx(i1:i2), ty(j1:j2)
    real(8) t1,t2

    t1 = rdtsc()
    
! find the global x,y of the source grid points 
    d1 = size(lon,1)
    d2 = size(lat,2)
    gxpole = (xpole+dx)/dx
    gypole = (ypole+dy)/dy
    do j=1,d2
        do i=1,d1
            call sph2gr(lon(i,j), lat(i,j), gxpole, gypole, dx*0.001, dy*0.001, ylong, &
                 x, y, ifail)
            gx(i,j) = x
            gy(i,j) = y
        end do
    end do

    if(master .and. debug > 2) then
        write(*,*) " gettrimap3: max gx,gy=",maxval(gx),maxval(gy),"minval gx,gy=",minval(gx),minval(gy)
    end if

! the target grid
    do i=i1,i2
        tx(i) = block%fi+i-1.5
    end do
    do j=j1,j2
        ty(j) = block%fj+j-1.5
    end do

    call get_gip_map(gx, gy, tx, ty, map)

    t2 = rdtsc()
    
    if(master.and.debug>1) write(*,*) " gettrimap3 done (init of triangular interpolation) t=",real(t2-t1)/clock_frequency
  end subroutine gettrimap3


  subroutine getbilinmap(lons, lats, map)
!    
! Compute weights for bilinear interpolation of data from a regular
! grid with geographical coordinates to the distributed model grid. we
! allow the spacing of the source grid to be arbritary along each
! dimension.
!    
    use util, only : binsearch
    use mod_mpi, only : i1,i2,j1,j2,abort,fci1,fci2,fcj1,fcj2
    use state, only : longitude,latitude,debug
    
    implicit none
    real lons(:), lats(:)
    type(mapping_object) map(i1:i2,j1:j2)
    real lx, ly
    integer i,j,londim,latdim,lonind,latind
   
    londim = size(lons,1)
    latdim = size(lats,1)
    
    do j=fcj1,fcj2
        do i=fci1,fci2
            !! search for the indices of surrounding longitude, latitude
            lonind = binsearch(longitude(i,j), lons) - 1
            latind = binsearch(latitude(i,j),  lats) - 1

            if(lonind < 1 .or. lonind >= londim .or. latind < 1 .or. latind >= latdim) then
                write(*,*) 'i,j',i,j, 'lon,lat', longitude(i,j),latitude(i,j), lonind, latind
                call abort("getbilinmap error: lon,lat data does not cover grid, cannot interpolate")
            end if

            map(i,j)%weight_type = 3
            map(i,j)%i = lonind
            map(i,j)%j = latind
            
            !! normalized local coordinate
            lx = (longitude(i,j) - lons(lonind))/(lons(lonind+1) - lons(lonind))
            ly = (latitude(i,j)  - lats(latind))/(lats(latind+1) - lats(latind))

            if(lx > 1 .or. lx < 0 .or. ly > 1 .or. ly < 0) &
                 call abort("getbilinmap error, local normalized coordinate out of range")

            !the interpolation of llf onto model grid f is then
            !  f(i,j) = llf(lonind,latind)*(1-lx)*(1-ly) + llf(lonind+1,latind)*lx*(1-ly) + &
            !           llf(lonind,latind+1)*(1-lx)*ly + llf(lonind+1,latind+1)*lx*ly            
            map(i,j)%w(1) = (1-lx)*(1-ly)
            map(i,j)%w(2) = lx*(1-ly)
            map(i,j)%w(3) = lx*ly
            map(i,j)%w(4) = (1-lx)*ly
        end do
    end do
  end subroutine getbilinmap

  subroutine vsectintpol(vsect, f, fu, fv, f2d)
! 
! F, FU, FV are distributed arrays in eta, u or v points respectively.
! given a VSECT object find the horizontally interpolated values of
! the given arrays under each segment. The result is stored in
! VSECT%F, VSECT%FUN, VSECT%FUT respectively. Assuming FU,FV to be
! vector components in the X and Y direction, FUN, FUT are the normal
! and tangential components of FU,FV wrt each linear section of VSECT.
! F2D is the 2D equivalent to the 3D F.
!
    use qmpi, only : myid
    use state, only : i1,j1,kb
    implicit none
    type(vsect_object) :: vsect
    real, dimension(i1:,j1:,:), intent(in), optional :: f,fu,fv
    real, dimension(i1:,j1:), intent(in), optional :: f2d
    integer i,j,k,pt,ui,uj,vi,vj,ci,cj
    real um,vm,cu(4),cv(4),c(4)

    if(.not.vsect%inblock) return

    if(present(f)) then
        if(.not.allocated(vsect%f)) allocate(vsect%f(vsect%n,kb))

        do i=1,vsect%np
            if(vsect%proc(i) == myid) then
                do pt=vsect%s1(i),vsect%s2(i)
                    c = vsect%seg(pt)%w
                    ci = vsect%seg(pt)%i
                    cj = vsect%seg(pt)%j
                    do k=1,kb-1
                        vsect%f(pt,k) = c(1)*f(ci,cj,k) + c(2)*f(ci+1,cj,k) + c(3)*f(ci+1,cj+1,k) + c(4)*f(ci,cj+1,k)
                    end do
                end do
            end if
        end do
    end if
    if(present(fu) .and. present(fv)) then
        if(.not.allocated(vsect%fun)) allocate(vsect%fun(vsect%n,kb))
        if(.not.allocated(vsect%fut)) allocate(vsect%fut(vsect%n,kb))

        do i=1,vsect%np
            if(vsect%proc(i) == myid) then
                do pt=vsect%s1(i),vsect%s2(i)                
                    cu = vsect%seg(pt)%wu
                    cv = vsect%seg(pt)%wv
                    ui = vsect%seg(pt)%iu
                    uj = vsect%seg(pt)%ju
                    vi = vsect%seg(pt)%iv
                    vj = vsect%seg(pt)%jv
                    do k=1,kb-1
                        um = cu(1)*fu(ui,uj,k) + cu(2)*fu(ui+1,uj,k) + cu(3)*fu(ui+1,uj+1,k) + cu(4)*fu(ui,uj+1,k)
                        vm = cv(1)*fv(vi,vj,k) + cv(2)*fv(vi+1,vj,k) + cv(3)*fv(vi+1,vj+1,k) + cv(4)*fv(vi,vj+1,k)
                        vsect%fun(pt,k) = um*vsect%seg(pt)%enx + vm*vsect%seg(pt)%eny
                        vsect%fut(pt,k) = um*vsect%seg(pt)%etx + vm*vsect%seg(pt)%ety
                    end do
                end do
            end if
        end do
    end if
    if(present(f2d)) then
        if(.not.allocated(vsect%f2d)) allocate(vsect%f2d(vsect%n))

        do i=1,vsect%np
            if(vsect%proc(i) == myid) then
                do pt=vsect%s1(i),vsect%s2(i)
                    c = vsect%seg(pt)%w
                    ci = vsect%seg(pt)%i
                    cj = vsect%seg(pt)%j
                    vsect%f2d(pt) = c(1)*f2d(ci,cj) + c(2)*f2d(ci+1,cj) + c(3)*f2d(ci+1,cj+1) + c(4)*f2d(ci,cj+1)
                end do
            end if
        end do
    end if
  end subroutine vsectintpol


  subroutine initvsect(vsect, label, sx, sy)
!    
! given a list of horizontal coordinates SX,SY construct a VSECT
! object, which is a list of short line segments with coordinates,
! unit direction vectors and weights needed to interpolate values from
! the model grid onto the slice below the linear sections. LABEL is a
! generic label to identify the VSECT object.
!    
    use state, only : im,jm,h
    use mod_mpi, only : numbli, numblj, myid
    use bomplot, only : walkngetpts, find_coeffs, find_ucoeffs, find_vcoeffs
    implicit none
    type(vsect_object) vsect
    character(*) label
    real sx(:),sy(:)

! guess the max dims of these temporaries
    real, dimension(im+jm) :: x,y,enx,eny,etx,ety,secpos,seclen,secdep
    integer, dimension(numbli+numblj) :: ps1,ps2,psproc
    logical, dimension(im+jm) :: inblock
    integer ns, np, ci,cj,i
    real wght(4),lx,ly

    call walkngetpts(sx,sy, x,y,enx,eny,etx,ety,secpos,seclen,ns,np,ps1,ps2,psproc,inblock)
    
    allocate(vsect%s1(np), vsect%s2(np), vsect%proc(np))

    vsect%label = label
    vsect%n = ns
    vsect%np = np
    vsect%inblock = .false.
    
    do i=1,np
        vsect%proc(i) = psproc(i)
        vsect%s1(i) = ps1(i)
        vsect%s2(i) = ps2(i)
        if(vsect%proc(i) == myid) vsect%inblock = .true.
    end do

    if(vsect%inblock .or. myid==0) then
        allocate(vsect%seg(ns))
    else
        return
    end if

    do i=1, ns
        vsect%seg(i)%x   = x(i) 
        vsect%seg(i)%y   = y(i)
        vsect%seg(i)%len = seclen(i)
        vsect%seg(i)%pos = secpos(i)
        vsect%seg(i)%enx = enx(i)
        vsect%seg(i)%eny = eny(i)
        vsect%seg(i)%etx = etx(i) 
        vsect%seg(i)%ety = ety(i)

! closest cell south west
        vsect%seg(i)%i = nint(x(i))
        vsect%seg(i)%j = nint(y(i))

        lx = x(i) - (vsect%seg(i)%i - 0.5)
        ly = y(i) - (vsect%seg(i)%j - 0.5)

! cell of coordinate
        ci = nint(x(i)+0.5)
        cj = nint(y(i)+0.5)

        vsect%seg(i)%inblock = inblock(i)

        if(inblock(i)) then
            call find_coeffs(lx,ly, vsect%seg(i)%i, vsect%seg(i)%j, ci,cj, vsect%seg(i)%w)
            call find_ucoeffs(x(i),y(i), lx,ly, vsect%seg(i)%i,vsect%seg(i)%j, vsect%seg(i)%iu,vsect%seg(i)%ju, vsect%seg(i)%wu)
            call find_vcoeffs(x(i),y(i), lx,ly, vsect%seg(i)%i,vsect%seg(i)%j, vsect%seg(i)%iv,vsect%seg(i)%jv, vsect%seg(i)%wv)
        else
            cycle
        end if
        
        wght = vsect%seg(i)%w
        ci = vsect%seg(i)%i
        cj = vsect%seg(i)%j
        vsect%seg(i)%dep = wght(1)*h(ci,cj) + wght(2)*h(ci+1,cj) + wght(3)*h(ci+1,cj+1) + wght(4)*h(ci,cj+1)
    end do
  end subroutine initvsect

  subroutine gathervsectdata(vsect, f)
!    
! F is a data array containing the result of interpolation via the
! given VSECT object.  This routine assemble f from all processors
! touched by VSECT on the master processor.
!    
    use qmpi, only : myid, numproc, isend, receive, waitall
    implicit none
    type(vsect_object) :: vsect
    real f(:,:)
    integer i
    
    if(.not. (vsect%inblock .or. myid == 0)) return
    
    do i=1,vsect%np
        if(myid > 0 .and. vsect%proc(i) == myid) then
            call isend(f(vsect%s1(i):vsect%s2(i),:), 0)
        end if
        
        if(myid == 0 .and. numproc > 1 .and. vsect%proc(i) > 0) then
            call receive(f(vsect%s1(i):vsect%s2(i),:), vsect%proc(i))
        end if
    end do
    call waitall()
  end subroutine gathervsectdata
  
end module intpol


