module maxmin
!  
! subroutine dump_maxmin is a diagnostic routine to print maximum and
! minimum of selected variables to screen. normally called from output
! once per time step, but if debug_level > 1 the routine dump_maxmin
! will be called also from dump_state. the logicals that decide if a
! variable gets its max/min computed or not, are set via the new setup
! system (see list below). 
!  
  use state
  use mod_modesplit
  use control_parameters
  use pgplot_core, only : pgplot_object
  implicit none

  private

  public dump_maxmin, maxmins, num_maxmins, max_maxmins, add_maxmin

! which variables to compute max and min for:

  integer :: num_maxmins=0
  integer, parameter :: max_maxmins=20
  type(pgplot_object), save :: maxmins(max_maxmins)

! dump to screen each maxmin_nd'th time step, starting from timestep
! maxmin_start and stopping at step maxmin_end.
  
  integer, public :: maxmin_nd=-1., maxmin_start=-1., maxmin_end=-1.

  logical :: initialize_maxmin=.true.
  
contains

  subroutine dump_maxmin
    implicit none
    integer i
    
    if(initialize_maxmin) then
        if(maxmin_nd < 0) maxmin_nd = 1
        if(maxmin_start < 0) maxmin_start=0
        if(maxmin_end < 0) maxmin_end=numstep
        initialize_maxmin=.false.
    end if
      
    if(mod(istep,maxmin_nd) ==0 .and. &
         istep >= maxmin_start .and. istep <= maxmin_end ) then
        do i=1,num_maxmins
            select case(maxmins(i)%type)
            case(0)
                if(associated(maxmins(i)%f2d)) &
                     call maxminprint2d(trim(maxmins(i)%title), maxmins(i)%f2d, maxmins(i)%range3d)
                if(associated(maxmins(i)%f3d)) &
                     call maxminprint3d(trim(maxmins(i)%title), maxmins(i)%f3d, maxmins(i)%range3d)
            case(1)
                if(associated(maxmins(i)%f2d)) &
                     call monitor2d(trim(maxmins(i)%title), maxmins(i)%f2d, maxmins(i)%range3d)
                if(associated(maxmins(i)%f3d)) &
                     call monitor3d(trim(maxmins(i)%title), maxmins(i)%f3d, maxmins(i)%range3d)                
            end select
        end do       
    end if
  end subroutine dump_maxmin

  subroutine maxminprint3d(string, f, r)
    use util
    implicit none
    character(*) string
    real f(i1:,j1:,:), fmax, fmin
    integer locmax(3),locmin(3), r(6)
    character(80) line
    
    call find_max_reals(kb, f, fmax, locmax, wet=.true., root=0, range=r)
    call find_min_reals(kb, f, fmin, locmin, wet=.true., root=0, range=r)

    ! try to guess if we are looking at density
    if(index(lower(string), "rho") > 0)then
        fmax = fmax + rhoaver
        fmin = fmin + rhoaver
    end if
    
    if(master)then
        write(line,*) trim(string)," max:",fmax
        write(line(37:),'(1x,"i,j,k: ",i0,",",i0,",",i0)') locmax
        write(*,'(a)') trim(line)
        write(line,*) trim(string)," min:",fmin
        write(line(37:),'(1x,"i,j,k: ",i0,",",i0,",",i0)') locmin
        write(*,'(a)') trim(line)
    end if
  end subroutine maxminprint3d
  
  subroutine maxminprint2d(string, f, r)
    implicit none
    character(*) string
    real f(i1:,j1:), fmax, fmin
    integer locmax(3),locmin(3), r(6)
    character(80) line

    r(5:6)=1
    call find_max_reals(1, f, fmax, locmax, wet=.true., root=0, range=r)
    call find_min_reals(1, f, fmin, locmin, wet=.true., root=0, range=r)
    if(master)then    ! 9              5     23
        write(line,*) trim(string)," max:",fmax
        write(line(37:),'(1x,"i,j:",1x,i0,",",i0)') locmax(1:2)
        write(*,*) trim(line)
        write(line,*) trim(string)," min:",fmin
        write(line(37:),'(1x,"i,j:",1x,i0,",",i0)') locmin(1:2)
        write(*,*) trim(line)        
    end if
  end subroutine maxminprint2d

  subroutine monitor3d(string, f, r)
    use util
    implicit none
    character(*) string
    real f(i1:,j1:,:)
    integer r(6),rt(6),grt(6)
    logical inblock

    call globalrect2local(r(1),r(2),r(3),r(4), rt(1),rt(2),rt(3),rt(4), inblock, ghost=.true.)
    if(inblock)then
        grt(1:4) = rt(1:4) + (/block%fi,block%fi,block%fj,block%fj/) - 1
        grt(5:6) = r(5:6)
        write(*,1) myid, string, grt, f(rt(1):rt(2),rt(3):rt(4),r(5):r(6))
    end if
1   format("proc",i4.4,2x,  a,  "(",i0,":",i0,",",i0,":",i0,",",i0,":",i0,")",2x,  40(f0.5,1x))
  end subroutine monitor3d
  
  subroutine monitor2d(string, f, r)
    implicit none
    character(*) string
    real f(i1:,j1:)
    integer r(6),rt(6),grt(6)
    logical inblock
    
    call globalrect2local(r(1),r(2),r(3),r(4), rt(1),rt(2),rt(3),rt(4), inblock, ghost=.true.)
    if(inblock)then
        grt(1:4) = rt(1:4) + (/block%fi,block%fi,block%fj,block%fj/) - 1
        write(*,1) myid, string, grt(1:4), f(rt(1):rt(2),rt(3):rt(4))
    end if
1   format("proc",i4.4,2x,  a,  "(",i0,":",i0,",",i0,":",i0,")",2x,  40(f0.5,1x))
  end subroutine monitor2d
  

  subroutine add_maxmin(splitline)
! keyword
!   maxmin <label> <what> [<range>]
    use pgplot, only : assignslice
    use util, only : lower
    implicit none
    character(*) splitline(:)
    integer r(6), ldim
    
    num_maxmins = num_maxmins + 1
    maxmins(num_maxmins)%ind = num_maxmins

    if(num_maxmins > max_maxmins) &
         call message('too many maxmins. reduce, or increase max_maxmins and recompile',2)
    
    maxmins(num_maxmins)%title = splitline(2)
    maxmins(num_maxmins)%what = lower(splitline(3))   

    select case(splitline(1))
    case("maxmin")
        maxmins(num_maxmins)%type = 0
        if(debug > 0 .and. master) then
            write(*,*)' adding maxmin, title=',trim(maxmins(num_maxmins)%title),' what=',&
                 trim(maxmins(num_maxmins)%what)
        end if
    case("monitor")
        maxmins(num_maxmins)%type = 1
        if(debug > 0 .and. master) then
            write(*,*)' adding monitor, title=',trim(maxmins(num_maxmins)%title),' what=',&
                 trim(maxmins(num_maxmins)%what)
        end if
    end select
    
    call assignslice(maxmins(num_maxmins))

    if(associated(maxmins(num_maxmins)%f3d)) then
        ldim = kb-1
        if(size(maxmins(num_maxmins)%f3d,3) < ldim) ldim = size(maxmins(num_maxmins)%f3d,3)
        r = (/1,im,1,jm,1,ldim/)

        if(trim(splitline(4)) .ne. "") then
            read(splitline(4:9),*,err=116,end=116) r
            if(all(r(1:6) == 0)) r = (/1,im,1,jm,1,ldim/)
        else
            if(maxmins(num_maxmins)%type==1) call message("a range is required for the monitor keyword",2)
        end if        
        if(debug > 0 .and. master) write(*,*) ' range=',r
    end if
    
    if(associated(maxmins(num_maxmins)%f2d)) then
        r = (/1,im,1,jm,1,1/)
        if(trim(splitline(4)) .ne. "") then
            read(splitline(4:7),*,err=117,end=117) r(1:4)
            if(all(r(1:4) == 0)) r = (/1,im,1,jm,1,1/)
            if(debug > 0 .and. master) write(*,*) ' range=',r(1:4)
        else
            if(maxmins(num_maxmins)%type==1) call message("a range is required for the monitor keyword",2)
        end if
    end if
    
    maxmins(num_maxmins)%range3d = r
    if(r(1) < -1 .or. r(2) > im+2) call message("invalid i range",2)
    if(r(3) < 1 .or. r(4) > jm) call message("invalid j range",2)

    if(splitline(1) == "monitor") then
        if(associated(maxmins(num_maxmins)%f3d)) then
            allocate(maxmins(num_maxmins)%data3d(r(1):r(2),r(3):r(4),r(5):r(6)))
        end if
        if(associated(maxmins(num_maxmins)%f2d)) then
            allocate(maxmins(num_maxmins)%data2d(r(1):r(2),r(3):r(4)))
        end if
    end if
    
    return
    116 call message("could not read range",2)
    117 call message("could not read range",2)
  end subroutine add_maxmin
  
end module maxmin
