module mod_averaging
!  
! storage and routines for computing averages of fields. the intention
! is to support averages of horizontal and vertical cross sections, 2d
! fields and 3d fields, perhaps from given regions. 
!
  use intpol, only : vsect_object, vsectintpol, initvsect, sigma2z, s2zmap_object, getmap_sigma2z
  implicit none  
  integer :: numaverages=0
  integer, parameter :: maxnumaverages=30

  character(80) :: average_dir="AVERAGE/"    ! where to dump files by default
  
  type averaging_object
     character(80) :: label="", what=""
     character(1) :: pos="c"          ! src field grid position; "u","v" or "c" (center)
     integer :: type=0
     character(20) :: avtype=""       ! averaging type ("monthly","daily",...)
     integer :: interval=0            ! if manual averaging interval is specified;
     integer :: prevdate(6)           ! the previous date the data was accumulated
     integer :: numav=0               ! current number of accumulated fields
     logical :: distributed=.true.    ! is the data array distributed?
     integer :: r(6)                  ! a list of index pairs, e.g. 3 coordinate ranges.
     real    :: depth=0.0             ! for hsect
     integer :: step=0                ! e.g. s2zmap; which t.s. was current map computed for?
     integer :: i,j,k                 ! a grid position
     
     real, pointer :: src2d(:,:) => null(), src3d(:,:,:)=>null()   ! pointers to source data
     real, allocatable :: data2d(:,:), data3d(:,:,:)               ! storage for actual averaged data
     real, allocatable :: data3d2(:,:,:), data3d3(:,:,:), data3d4(:,:,:), data3d5(:,:,:)
     real, allocatable :: tmp2d(:,:)
     
     type(vsect_object), allocatable :: vsect                      ! interpolation object for vertical sections
     
     type(s2zmap_object), allocatable :: s2zmap(:,:,:)
     real, allocatable :: zlevs(:)
     logical :: inblock=.false.
  end type averaging_object

  type(averaging_object) average(maxnumaverages)

contains
  
  subroutine add_average(splitline)
!    
! usage in setupfile:
!   hsectaverage <label> <what> <how often> <depth> <range: x1 x2 y1 y2>
!   vsectaverage <label> <what> <how often> <section: N x1 y1 ... xN yN>
!   uvstaverage <how often> <depth file>
!   vprofaverage <how often> <x> <y> <depth file> 
!    
    use state, only : debug,dx,dy,im,jm,kb
    use mod_mpi, only : i1,i2,j1,j2,globalind2local
    use qmpi, only : master, message
    use grid_util
    use util, only : lower, check_file
    use time_util, only : ndate
    use file_util
    use control_parameters, only : data_path
    implicit none
    character(*) splitline(:)
    real sx(100),sy(100),x,y
    integer n,pt,na,k,gi,gj
    
    numaverages = numaverages + 1
    na=numaverages
    
    average(na)%prevdate = ndate
    
    select case(lower(splitline(1)))
    case('hsectaverage')
        average(na)%label = splitline(2)
        average(na)%what = lower(splitline(3))
        average(na)%avtype = lower(splitline(4))
        average(na)%distributed = .true.
        average(na)%type = 0
        read(splitline(5),*) average(na)%depth
        call get2drange(6,9)
        call stringassign(average(na))
        
        select case(trim(average(na)%what))
        case("u")
            average(na)%pos = "u"
        case("v")
            average(na)%pos = "v"
        end select
                
        allocate(average(na)%data2d(i1:i2,j1:j2))
        average(na)%data2d = 0
        if(master.and.debug>0) write(*,*) " what=",trim(average(na)%what)," depth=",average(na)%depth, " range=",average(na)%r(1:4)
        
    case('vsectaverage')
        average(na)%label = splitline(2)
        average(na)%what = lower(splitline(3))
        average(na)%avtype = lower(splitline(4))
        average(na)%distributed = .false.        
        allocate(average(na)%vsect)
        
        read(splitline(5),*) n
        read(splitline(6:6+2*n-1),*) (sx(pt),sy(pt),pt=1,n)
        sx(1:n) = sx(1:n)/dx + 1.0
        sy(1:n) = sy(1:n)/dy + 1.0
        if(debug>0.and.master) then
            write(*,*)'polygon points:'
            do pt=1,n
                write(*,*) pt, sx(pt), sy(pt)
            end do
        end if

        call initvsect(average(na)%vsect, trim(average(na)%label), sx(1:n), sy(1:n))

        allocate(average(na)%data2d(average(na)%vsect%n,kb))
        average(na)%data2d = 0
        
        select case(average(na)%what)
        case("un","transport")
            average(na)%type = 2
        case("ut")
            average(na)%type = 8            
        case("speed")
            average(na)%type = 3           
        case default
            average(na)%type = 1
            call stringassign(average(na))            
        end select
        if(master.and.debug>0) write(*,*) " what=",trim(average(na)%what)

    case('uvstaverage')
        average(na)%label = "zav_uvstw"
        average(na)%avtype = lower(splitline(2))
        average(na)%type = 6
        average(na)%distributed = .true.
!        call get2drange(5,8)        ! maybe later, do the full initially
        call check_file(trim(data_path)//trim(splitline(3)))
        call get_fields(trim(data_path)//trim(splitline(3)), sx, cols=(/1/), lines=n)
        average(na)%r = (/1,im,1,jm,1,n/)
        allocate(average(na)%zlevs(n))
        average(na)%zlevs = sx(1:n)
        if(master.and. debug >= 1) write(*,'(2x,a,i0,a,a)') "got ",n," zlevels from file ",trim(data_path)//trim(splitline(3))

        if(master.and.debug>1) then
            write(*,'(2x,a,i0,a,100(f0.1,1x))') "z(1:",n,")=",sx(1:n)
        end if
        
        allocate(average(na)%s2zmap(i1:i2,j1:j2,n))
        
        allocate(average(na)%data3d(i1:i2,j1:j2,n))
        average(na)%data3d = 0
        allocate(average(na)%data3d2(i1:i2,j1:j2,n))
        average(na)%data3d2 = 0
        allocate(average(na)%data3d3(i1:i2,j1:j2,n))
        average(na)%data3d3 = 0
        allocate(average(na)%data3d4(i1:i2,j1:j2,n))
        average(na)%data3d4 = 0
        allocate(average(na)%data3d5(i1:i2,j1:j2,n))
        average(na)%data3d5 = 0
        
    case('vprofaverage')
! vprofaverage <how often> <x> <y> <depth file> 
        average(na)%avtype = lower(splitline(2))
        average(na)%type = 7
        average(na)%distributed = .false.

        read(splitline(3:4),*, err=101, end=101) x,y
        call xyz2gind(x,y, 1.0, gi, gj, k)
        average(na)%i = gi
        average(na)%j = gj
        call globalind2local(average(na)%i,average(na)%j,average(na)%inblock)
        
        call check_file(trim(data_path)//trim(splitline(5)))
        call get_fields(trim(data_path)//trim(splitline(5)), sx, cols=(/1/), lines=n)
        allocate(average(na)%zlevs(n))
        average(na)%zlevs = sx(1:n)

        write(average(na)%label,'("vproav_uvwstr_i",i0,"j",i0,"kdim",i0)') average(na)%i, average(na)%j, n
        
        if(master.and.debug >= 1) write(*,'(2x,a,i0,a,a,a,i0,a,i0)') "got ",n," zlevels from file ",trim(data_path)//trim(splitline(5)), &
             " global i,j=", gi, ",", gj
        
        if(master.and.debug>1) then
            write(*,'(2x,a,i0,a,100(f0.1,1x))') "z(1:",n,")=",sx(1:n)
        end if

        ! u,v,wr,s(i,j,k),t(i,j,k),rho(i,j,k)+rhoaver        
        allocate(average(na)%data2d(6,n), average(na)%tmp2d(6,n))
        average(na)%data2d = 0
        
    case('average')
        average(na)%distributed = .true.
        call get2drange(5,8)       
        call stringassign(average(na))       
        if(associated(average(na)%src3d)) then
            allocate(average(na)%data3d(i1:i2,j1:j2,kb))
            average(na)%data3d = 0
        end if
        if(associated(average(na)%src2d)) then
            allocate(average(na)%data2d(i1:i2,j1:j2))
            average(na)%data2d = 0
        end if        
        if(master.and.debug>0) write(*,*) " what=",trim(average(na)%what)
    end select

    return

101 call message("could not read x,y for vprofaverage from column 3,4",2)
    
  contains

    subroutine stringassign(object)
      use pgplot, only : pgplot_object, assignslice
      implicit none
      type(averaging_object) object
      type(pgplot_object) dummyobject
      
! hack... reuse the pgplot assignslice routine, which connects
! the 2d/3d pointer of the pgplot_object to a bom array according to
! the string "what"
      dummyobject%what = object%what
      call assignslice(dummyobject)
! now make the pointer of the average object point to the same as the dummyobject    
      if(associated(dummyobject%f3d)) object%src3d => dummyobject%f3d
      if(associated(dummyobject%f2d)) object%src2d => dummyobject%f2d      
    end subroutine stringassign
    
    subroutine get2drange(c1,c2)
      implicit none
      integer c1,c2
      character(80) mess
      real reals(4)
      
      read(splitline(c1:c2),*,err=101,end=101) reals
      average(na)%r(1:4) = true2globrange(reals)
      call check_range(average(na)%r(1:4))
      return
101   write(mess,'("could not get 4 numbers for 2d range from column ",i0," to ",i0)') c1,c2
      call message(trim(mess),2)
    end subroutine get2drange
    
  end subroutine add_average
  

  subroutine compute_averages()
    use mod_mpi, only : i1,i2,j1,j2,abort
    use state, only : istep, ndivis,u,v,s,t,wr
    use bomplot, only : hsect, vprofz
    use time_util, only : ndate
    
    implicit none
    integer i,j,interval
    character(20) tmp
    real tmp2d(i1:i2,j1:j2)
    real, allocatable :: tmp3d(:,:,:)
    
    if(istep == 0) return
    
! loop over all averages, compute the actual average and dump to disk if it is time
    
    do i=1,numaverages
        select case(average(i)%avtype)
        case("daily")
            if(average(i)%prevdate(3) .ne. ndate(3)) then
                call compndump(i)
                average(i)%prevdate(3) = ndate(3)
            end if
        case("monthly")
            if(average(i)%prevdate(2) .ne. ndate(2)) then
                call compndump(i)
                average(i)%prevdate(2) = ndate(2)
            end if
        case("weekly")
            if(mod(istep,ndivis*24*7) == 0) call compndump(i)
        case("hourly")
            if(mod(istep,ndivis) == 0) call compndump(i)
        case default
            j = index(average(i)%avtype,"istep") 
            if(j > 0) then
                read(average(i)%avtype(j+5:),*) interval
                if(mod(istep,interval)==0) call compndump(i)
            else
                call abort("the variable AVTYPE in module MOD_AVERAGING is corrupt")
            end if
        end select
    end do


! add yet another time to the sum that will give the average in the end;
! for certain types of averages we need to fetch data first

    do i=1,numaverages
        average(i)%numav = average(i)%numav + 1

        select case(average(i)%type)
        case(0)
          !! hsect of given field
            call hsect("avhsect", field=average(i)%src3d, position=-average(i)%depth, output="none", rt=tmp2d, pos=average(i)%pos)
            average(i)%data2d = average(i)%data2d + tmp2d
        case(1)
          !! vsect of given field
            call vsectintpol(average(i)%vsect, average(i)%src3d)
            if(average(i)%vsect%inblock) then
                average(i)%data2d(:,:) = average(i)%data2d(:,:) + average(i)%vsect%f(:,:)
            end if
        case(2)
          !! vsect of transport normal to sect
            call vsectintpol(average(i)%vsect, fu=u, fv=v)
            if(average(i)%vsect%inblock) then
                average(i)%data2d(:,:) = average(i)%data2d(:,:) + average(i)%vsect%fun(:,:)
            end if            
        case(8)
          !! vsect of transport along sect
            call vsectintpol(average(i)%vsect, fu=u, fv=v)
            if(average(i)%vsect%inblock) then
                average(i)%data2d(:,:) = average(i)%data2d(:,:) + average(i)%vsect%fut(:,:)
            end if            
        case(3)
          !! vsect of speed over sect
            call vsectintpol(average(i)%vsect, fu=u, fv=v)
            if(average(i)%vsect%inblock) then
                average(i)%data2d(:,:) = average(i)%data2d(:,:) + (average(i)%vsect%fun(:,:)**2 + average(i)%vsect%fut(:,:)**2)**0.5
            end if
        case(4)
            !! vsect of square of field
            call vsectintpol(average(i)%vsect, average(i)%src3d)
            if(average(i)%vsect%inblock) then
                average(i)%data2d(:,:) = average(i)%data2d(:,:) + average(i)%vsect%f(:,:)**2
            end if
        case(5)
          !! hsect of square of given field
            call hsect("avhsect", field=average(i)%src3d, position=-average(i)%depth, output="none", rt=tmp2d, pos=average(i)%pos)
            average(i)%data2d = average(i)%data2d + tmp2d**2
        case(6)
          !! uvst 3d average

            if(istep > average(i)%step) then
                ! recompute map once per timestep
                call getmap_sigma2z(average(i)%zlevs, average(i)%s2zmap)
                average(i)%step = istep
            end if

            allocate(tmp3d(i1:i2,j1:j2,size(average(i)%zlevs)))
            
            call sigma2z(u, tmp3d, average(i)%s2zmap, loc="u")
            average(i)%data3d = average(i)%data3d + tmp3d
           
            call sigma2z(v, tmp3d, average(i)%s2zmap, loc="v")
            average(i)%data3d2 = average(i)%data3d2 + tmp3d
            
            call sigma2z(s, tmp3d, average(i)%s2zmap)
            average(i)%data3d3 = average(i)%data3d3 + tmp3d
            
            call sigma2z(t, tmp3d, average(i)%s2zmap)
            average(i)%data3d4 = average(i)%data3d4 + tmp3d

            call sigma2z(wr, tmp3d, average(i)%s2zmap)
            average(i)%data3d5 = average(i)%data3d5 + tmp3d
            
            deallocate(tmp3d)            
        case(7)
          !! vprofaverage
            if(average(i)%inblock) then
                call vprofz(average(i)%tmp2d, average(i)%i, average(i)%j, average(i)%zlevs)
                average(i)%data2d = average(i)%data2d + average(i)%tmp2d
            end if            
        case default
            if(associated(average(i)%src2d)) average(i)%data2d = average(i)%data2d + average(i)%src2d
            if(associated(average(i)%src3d)) average(i)%data3d = average(i)%data3d + average(i)%src3d            
        end select
    end do
  end subroutine compute_averages


  subroutine compndump(i)
!      
! compute the current average from accumulated data,
! dump to file and prepare for next
!
    use bomplot, only : parbindump
    use intpol, only : gathervsectdata
    use util, only : bindump
    use state, only : debug, istep
    use qmpi, only : master
    implicit none
    integer i
    character(8) ss

    write(ss,'(i8.8)') istep
    
    if(allocated(average(i)%data2d)) then
        average(i)%data2d = average(i)%data2d/average(i)%numav
        
        if(average(i)%distributed)then
            call parbindump(trim(average_dir)//trim(average(i)%label), average(i)%data2d, average(i)%r(1:4))
        end if
        
        if(allocated(average(i)%vsect))then
            call gathervsectdata(average(i)%vsect, average(i)%data2d)
            if(master)then
                if(debug > 1) write(*,*) "average bindump2d label=",trim(average(i)%label)
                call bindump(trim(average_dir)//trim(average(i)%label)//ss, average(i)%data2d, access="stream")
            end if
        end if

        if(average(i)%type == 7 .and. average(i)%inblock) then
            if(debug > 1) write(*,*) "average bindump2d label=",trim(average(i)%label)
            call bindump(trim(average_dir)//trim(average(i)%label)//"-"//ss, average(i)%data2d, access="stream")
        end if
        
        average(i)%numav = 0
        average(i)%data2d = 0                    
    end if

    if(allocated(average(i)%data3d)) then
        average(i)%data3d = average(i)%data3d/average(i)%numav
        if(allocated(average(i)%data3d2)) average(i)%data3d2 = average(i)%data3d2/average(i)%numav
        if(allocated(average(i)%data3d3)) average(i)%data3d3 = average(i)%data3d3/average(i)%numav
        if(allocated(average(i)%data3d4)) average(i)%data3d4 = average(i)%data3d4/average(i)%numav
        if(allocated(average(i)%data3d5)) average(i)%data3d5 = average(i)%data3d5/average(i)%numav        
        
        if(average(i)%distributed)then
            call parbindump(trim(average_dir)//trim(average(i)%label), average(i)%data3d, average(i)%r)            
            if(allocated(average(i)%data3d2)) call parbindump(trim(average_dir)//trim(average(i)%label)//"2", average(i)%data3d2, average(i)%r)
            if(allocated(average(i)%data3d3)) call parbindump(trim(average_dir)//trim(average(i)%label)//"3", average(i)%data3d3, average(i)%r)
            if(allocated(average(i)%data3d4)) call parbindump(trim(average_dir)//trim(average(i)%label)//"4", average(i)%data3d4, average(i)%r)
            if(allocated(average(i)%data3d5)) call parbindump(trim(average_dir)//trim(average(i)%label)//"5", average(i)%data3d5, average(i)%r)
        else
            if(master.and.debug>1) write(*,*) "average bindump3d label=",trim(average(i)%label)
            call bindump(trim(average_dir)//trim(average(i)%label)//ss, average(i)%data3d, access="stream")
        end if

        average(i)%numav = 0
        average(i)%data3d = 0
        if(allocated(average(i)%data3d2)) average(i)%data3d2 = 0.
        if(allocated(average(i)%data3d3)) average(i)%data3d3 = 0.
        if(allocated(average(i)%data3d4)) average(i)%data3d4 = 0.
        if(allocated(average(i)%data3d5)) average(i)%data3d5 = 0.        
    end if
  end subroutine compndump

end module mod_averaging
