module mod_cruise
!  
! support routines for sampling modeled S,T (others if needed) along cruise
! trajectories. we read cruise files, e.g. tab delimited
! cruise data from http://doi.pangaea.de/10.1594/PANGAEA.819701,
! and add the position and date to an array of "cruise_objects" below.
!  
  integer :: numcruises=0
  integer, parameter :: maxnumcruises=20
  integer :: cruise_dump_interval=1000
  
  type cruise_object
     character(120) :: label
     character(200) :: filename
     integer :: coordinate_type=0  ! 0=geographical, 1=xy in meters, 2=grid coordinates
     integer :: position=1         ! the next cruise location to be sampled
     integer :: numsamples=0       ! the number of locations to sample
     integer :: dump_start=1       ! when we dump to file, start with this location
     real, allocatable :: time(:), x(:), y(:), z(:), s(:), t(:)
     integer :: timecol=1, loncol=2, latcol=3, depcol=4  ! the default column from where to read time, x/lon, y/lat
  end type cruise_object

  type(cruise_object) :: cruise(maxnumcruises)
  
contains
  
  subroutine add_cruise(splitline)
!     
! sample data along the cruise trajectory given by e.g. TIME,LON,LAT,DEP and
! store in the CRUISE file
!   cruise <label> <cruise file> [<coordinate type> <timecol> <loncol> <latcol>]
!
    use qmpi, only : broadcast, master, message
    use state, only : dx,dy,xpole,ypole,ylong
    use grid_util, only : sph2gr
    use time_util
    use util
    implicit none
    character(*) splitline(:)
    integer cruise_time(6), ierr
    integer, parameter :: maxsamples=1000
    character(200) cruise_lines(maxsamples)
    character(200) cruise_splitline(10)
    integer linenumbers(maxsamples)
    integer i,j,n, numwords, prevnumwords
    logical wordwarning    
    real coo1,coo2
    
    numcruises = numcruises + 1
    if(numcruises > maxnumcruises) call message("too many cruises, increase MAXNUMCRUISES and recompile",2)

    cruise(numcruises)%label = splitline(2)   
    
    if(master) call get_lines(splitline(3), cruise(numcruises)%numsamples, cruise_lines, linenumbers)
    
    call broadcast(cruise(numcruises)%numsamples)
    call broadcast(cruise_lines)
    call broadcast(linenumbers)

    if(trim(splitline(4)) .ne. "") read(splitline(4),*,err=109,end=109) cruise(numcruises)%coordinate_type
    if(trim(splitline(5)) .ne. "") read(splitline(5),*) cruise(numcruises)%timecol
    if(trim(splitline(6)) .ne. "") read(splitline(6),*) cruise(numcruises)%loncol
    if(trim(splitline(7)) .ne. "") read(splitline(7),*) cruise(numcruises)%latcol
    if(trim(splitline(8)) .ne. "") read(splitline(8),*) cruise(numcruises)%depcol
    
    n = cruise(numcruises)%numsamples
    allocate(cruise(numcruises)%time(n), cruise(numcruises)%x(n), cruise(numcruises)%y(n) , cruise(numcruises)%z(n))
    allocate(cruise(numcruises)%s(n), cruise(numcruises)%t(n))

    do i=1,n
        prevnumwords = numwords
        call split(cruise_lines(i), cruise_splitline, numwords)
        if(numwords < 4) call message("error in cruise file, found < 3 reals on line",2)
        
        if(i > 1 .and. numwords .ne. prevnumwords) wordwarning = .true.

! if there is a t in the forcing time string, it must be in iso format
        if(index(cruise_splitline(cruise(numcruises)%timecol),"T") > 0) then
            call get_iso8601(cruise_splitline(cruise(numcruises)%timecol), cruise_time)
            cruise(numcruises)%time(i) = timediff(initial_time, cruise_time)
        else
            read(cruise_splitline(cruise(numcruises)%timecol),*,end=110,err=110) cruise(numcruises)%time(i)
        end if
        
        read(cruise_splitline(cruise(numcruises)%loncol),*,end=111,err=111) coo1
        read(cruise_splitline(cruise(numcruises)%latcol),*,end=111,err=111) coo2
        
        select case(cruise(numcruises)%coordinate_type)
        case(0)
            call sph2gr(coo1, coo2, (xpole+dx)/dx, (ypole+dy)/dy, dx*0.001, dy*0.001, &
                 ylong, cruise(numcruises)%x(i), cruise(numcruises)%y(i), ierr)
        case(1)
            cruise(numcruises)%x(i) = (coo1+dx)/dx
            cruise(numcruises)%y(i) = (coo2+dy)/dy
        case(2)
            cruise(numcruises)%x(i) = coo1
            cruise(numcruises)%y(i) = coo2
        end select        
        read(cruise_splitline(cruise(numcruises)%depcol),*,end=112,err=112) cruise(numcruises)%z(i)
    end do

! set cruise()%position to the first i where the time is > 0    
    do i=1,n
        cruise(numcruises)%position = i
        cruise(numcruises)%dump_start = i
        if(cruise(numcruises)%time(i) > 0.0) exit
    end do
   
    if(master)then
        j = numcruises
        write(*,*) 'cruise',j,' name=',trim(cruise(j)%label)
        write(*,*) 'position=',cruise(j)%position
        write(*,*) 'time x y z:'
        do i=1,n
            write(*,*) cruise(j)%time(i), cruise(j)%x(i), cruise(j)%y(i), cruise(j)%z(i)
        end do
    end if
    return
109 call message("could not read integer for coordinate type from column 4",2)
110 call message("could not read model time in seconds from column 1 in data file",2)
111 call message("could not get horisontal cruise coordinate from column 2 and 3 in data file",2)
112 call message("could not get vertical cruise coordinate from column 4 in data file",2)   
  end subroutine add_cruise


  subroutine sample_cruises(step)
    use time_util, only : time
    use state, only : dt, d, zz, s ,t, kb, numstep
    use util, only : binsearch2    
    implicit none
    integer step
    integer i,j,k,c,n
    real dyndepth, elevation, sigma, s_k, s_km1, t_k, t_km1, alpha
    
    do c=1,numcruises
        if(cruise(c)%position > cruise(c)%numsamples) cycle
        
        n = cruise(c)%position

! get samples at all cruise times until the cruise time is > model TIME
        
        if(time > cruise(c)%time(n)) then
            do
                dyndepth = biip(d, cruise(c)%x(n), cruise(c)%y(n))
                !elevation = biip(eta, cruise(c)%x(n), cruise(c)%y(n))
                
                ! assumpe the depth is given relative to the sea surface
                sigma = cruise(c)%z(n)/dyndepth
                
                k = binsearch2(sigma, zz, last=kb-1)
                if(k < 1) then
                    cruise(c)%s(n) = biip(s(:,:,1), cruise(c)%x(n), cruise(c)%y(n))
                    cruise(c)%t(n) = biip(t(:,:,1), cruise(c)%x(n), cruise(c)%y(n))                    
                else if(k > kb-1) then
                    cruise(c)%s(n) = 0.0 !hmmm
                    cruise(c)%t(n) = 0.0
                else
                    s_k = biip(s(:,:,k), cruise(c)%x(n), cruise(c)%y(n))
                    t_k = biip(t(:,:,k), cruise(c)%x(n), cruise(c)%y(n))                 
                    s_km1 = biip(s(:,:,k-1), cruise(c)%x(n), cruise(c)%y(n))
                    t_km1 = biip(t(:,:,k-1), cruise(c)%x(n), cruise(c)%y(n))

                    ! alpha is 1 if sigma=zz(k-1), 0 for sigma=zz(k):
                    alpha = (sigma - zz(k))/(zz(k-1)-zz(k))
                    
                    cruise(c)%s(n) = s_km1*alpha + s_k*(1-alpha) 
                    cruise(c)%t(n) = t_km1*alpha + t_k*(1-alpha)
                    print *,'sampled',n
                end if                

                n = n + 1
                if(n > cruise(c)%numsamples) exit
                if(cruise(c)%time(n) > time) exit
            end do
            cruise(c)%position = n
        end if
    end do

    if(step > 0 .and. (mod(step, cruise_dump_interval) == 0  .or.  step==numstep)) call store_cruises()
    
  contains

    real function biip(f, gx, gy)
      
! if F is a 2D distributed array of cell centred data, use bilinear
! interpolation to find the value in the point with global coordinate
! GX,GY.      
      use state, only : block, h,i1,j1,i2,j2
      use mod_mpi, only : globalcoo2local
      implicit none
      real f(i1:i2,j1:j2), gx, gy
      integer i,j,ci,cj,sw(2),se(2),nw(2),ne(2)
      logical inblock
      real x,y,a(4),gx2,gy2
      
! if the sample point gx,gy is not on block, return 0
      gx2=gx ; gy2=gy
      call globalcoo2local(gx2,gy2,inblock)     
      biip = 0.0
      if(.not.inblock) return
      
! we'll use 4 cells for the interpolation, this is the index of the southwest cell
      i = INT(gx + 0.5)
      j = INT(gy + 0.5)
            
! local bilinear interpolation coordinate between grid points surrounding gx,gy
      x = gx - (i -0.5)
      y = gy - (j -0.5)
      
! convert i,j to block local coordinates
      i = i - block%fi + 1
      j = j - block%fj + 1

! block local coordinate of cell containing gx,gy
      ci = NINT(gx + 0.5) - block%fi + 1
      cj = NINT(gy + 0.5) - block%fj + 1

! if some of the cells in the interpolation is land, point their
! indices to the cell containing the coordinate. this is then not
! bilinear interpolation any more, but easier than reverting to
! triangular or linear interpolation. 
      sw = (/i,j/)
      se = (/i+1,j/)
      nw = (/i,j+1/)
      ne = (/i+1,j+1/)
      if(h(i,j) <= 0.0    ) sw = (/ci,cj/)
      if(h(i+1,j) <= 0.0  ) se = (/ci,cj/)
      if(h(i,j+1) <= 0.0  ) nw = (/ci,cj/)
      if(h(i+1,j+1) <= 0.0) ne = (/ci,cj/)
      
      a(1) = f(sw(1),sw(2))
      a(2) = f(se(1),se(2)) - a(1)
      a(3) = f(nw(1),nw(2)) - a(1)
      a(4) = f(ne(1),ne(2)) - (a(1)+a(2)+a(3))

      biip = a(1) + a(2)*x + a(3)*y + a(4)*x*y
    end function biip  
  end subroutine sample_cruises

  
  subroutine store_cruises()
!    
! at regular intervals, fetch cruise data from all processors, add
! them (sample=0 if off-block), dump batch on master.
!
    use qmpi, only : reduce, master
    implicit none
    integer i,n1,n2,c
    character(200) filename
    
    do c=1,numcruises
        if(cruise(c)%dump_start > cruise(c)%numsamples) cycle
        
        n1 = cruise(c)%dump_start
! (...%position is the next sample point)
        n2 = min(cruise(c)%position-1, cruise(c)%numsamples)
        call reduce('sum', cruise(c)%s(n1:n2), root=0)
        call reduce('sum', cruise(c)%t(n1:n2), root=0)
        if(master)then
            write(filename,'("cruise-",a,"-loc",i5.5,"-",i5.5,".dat")') trim(cruise(c)%label), n1, n2
            open(99, file=filename)
            write(99,'(3(f0.3,1x))') (cruise(c)%time(i), cruise(c)%s(i), cruise(c)%t(i),i=n1,n2)
            close(99)
        end if
        cruise(c)%dump_start = n2+1
    end do    
  end subroutine store_cruises
  
end module mod_cruise
