module mod_drag_region

  private
  public apply_regional_drag, initialize_drag_region
  
  integer, parameter :: maxregion = 100
  integer :: numregion = 0

  type region_object 
     logical :: set=.false.
     real :: x1,x2,y1,y2,z1,z2 ! location of region
     real :: cd_region  ! drag coeff for region
     character(30) :: name=""

     real :: a_region             ! lateral area of cage (meters**2)
     integer :: i1,i2,j1,j2,k1,k2        ! processor global indices for region
     integer :: li1,li2,lj1,lj2        ! processor local indices for region
     logical :: inblock    ! true if cage is in current processor block
  end type region_object

  type(region_object), save :: drag(maxregion)


contains

  subroutine apply_regional_drag
    use state
    implicit none

    integer i,j,k,id
    integer ni,nj,nk
    integer dni, dnj
    integer sidei, sidej
    real speed, correction_u, correction_v


    if(numregion == 0) return

    do id=1,numregion

        if (drag(id)%inblock) then

! Check on UA and VA what direction the mean flow is
            ni=int((drag(id)%li1+drag(id)%li2)/2)
            nj=int((drag(id)%lj1+drag(id)%lj2)/2)

! Estimate the direction of the flow and find what side to be used for calculating the correction
            if (ua(ni,nj).ge.0.0) then
                dni=1
                sidei=drag(id)%li2
            else
                dni=-1
                sidei=drag(id)%li1
            end if

            if (va(ni,nj).ge.0.0) then
                dnj=1
                sidej=drag(id)%lj2
            else
                dnj=-1
                sidej=drag(id)%lj1
            end if

! update the cell for calculating the correction

            ni=sidei+dni
            nj=sidej+dnj

! Use the velocity from the neighbour cell as the first guess
! Assume small variations in the horizontal
!       ni=drag(id)%li1-1
!       nj=drag(id)%lj2+1

            do k=drag(id)%k1,drag(id)%k2

                speed = sqrt(u(ni,nj,k)**2+v(ni,nj,k)**2) ! Need to find speed in u point and v points seperately
                correction_u = 0.5*drag(id)%cd_region*speed*u(ni,nj,k)*drag(id)%a_region/(dx*dy*dz(k))
                correction_v = 0.5*drag(id)%cd_region*speed*v(ni,nj,k)*drag(id)%a_region/(dx*dy*dz(k))
                do j=drag(id)%lj1,drag(id)%lj2

                    do i=drag(id)%li1,drag(id)%li2
                        if (u(i,j,k).ge.0) then
                            u(i,j,k) = u(i,j,k) - abs(dt*correction_u)
                            if (u(i,j,k).lt.0.0) u(i,j,k)=0.0
                        else
                            u(i,j,k) = u(i,j,k) + abs(dt*correction_u)
                            if (u(i,j,k).gt.0.0) u(i,j,k)=0.0
                        end if
                        if (v(i,j,k).ge.0) then
                            v(i,j,k) = v(i,j,k) - abs(dt*correction_v)
                            if (v(i,j,k).lt.0.0) v(i,j,k)=0.0
                        else
                            v(i,j,k) = v(i,j,k) + abs(dt*correction_v)
                            if (v(i,j,k).gt.0.0) v(i,j,k)=0.0
                        end if
                    end do

                end do

            end do

        end if
    end do
    call dump_state("regdrag")
  end subroutine apply_regional_drag


  function getspeed(i,j,k)

! Simple estimate of the free stream velocity/speed. 
! Assumption is that the boundary layer is thin based on
! delta approx 5.5*x/sqrt(Re) wherer x is the radius of the object.
! For cages around 50m and Reynoldsnumber between 10-1000
! the thickness of the boundary layer is between 10-100m
! which means an that the neighbour grid cell gives a good 
! estimate on the free flow.

    use state
    implicit none

    real getspeed(3)
    integer i,j,k
    integer ni,nj
    real tx,ty,nx,ny,speed

    speed = sqrt(u(i,j,k)**2 + v(i,j,k)**2)

! Normalize the velocity components
    tx = u(i,j,k)/speed
    ty = v(i,j,k)/speed

! Find the normal component to the velocity
    nx = -ty
    ny = tx 

! Find the coordinates to a point one grid cell normal to the initial velocity
    ni = nint((i-0.5 + nx) + 0.5)
    nj = nint((j-0.5 + ny) + 0.5)

! extract the velocity abd assume flow direction is the same in the vertical
    getspeed(3) = sqrt(u(ni,nj,k)**2 + v(ni,nj,k)**2)
    getspeed(1) = u(ni,nj,k) 
    getspeed(2) = v(ni,nj,k)    
  end function getspeed


  subroutine initialize_drag_region(splitline)
    use qmpi, only : master, message
    use mod_mpi, only : globalrect2local
    use state, only : debug, dx, dy, kb, z, h, im, jm
    
    implicit none
    integer id, k    
    character(*) splitline(:)
    real dep 

    numregion=numregion+1

    if(numregion > maxregion) then
        if(master)write(*,*)'numregion=', numregion
        call message('Too many regions! you should increase MAXREGIONS in mod_region.f90 and recompile.',2)
    end if

    if(drag(id)%set) call message("Drag region already set",2)

    read(splitline(2),*,err=100,end=100) id
    write(drag(id)%name,'(a,i0)') "drag_region",id
    if(master .and. debug>0) write(*,'(2x,a,i0)') "setting drag_region ",id

    read(splitline(3),*,err=101,end=101) drag(id)%cd_region
    read(splitline(4),*,err=102,end=102) drag(id)%x1
    read(splitline(5),*,err=102,end=102) drag(id)%x2
    read(splitline(6),*,err=102,end=102) drag(id)%y1
    read(splitline(7),*,err=102,end=102) drag(id)%y2
    read(splitline(8),*,err=103,end=103) drag(id)%z1
    read(splitline(9),*,err=103,end=103) drag(id)%z2

    if(drag(id)%z1 > 0. .or. drag(id)%z2 > 0) call message("z must be negative",2)
    if (drag(id)%z1.gt.drag(id)%z2) call message("z1 must be less than z2",2)
    drag(id)%set = .true.

! compute the drag area. it can be discussed if this is a smart way to do it. Oyvind
    drag(id)%a_region=abs(max((drag(id)%x2-drag(id)%x1),(drag(id)%y2-drag(id)%y1))*(drag(id)%z2-drag(id)%z1))

! compute the global grid coordinates
    drag(id)%i1=int(drag(id)%x1/dx)+1
    drag(id)%i2=int(drag(id)%x2/dx)
    drag(id)%j1=int(drag(id)%y1/dy)+1
    drag(id)%j2=int(drag(id)%y2/dy)
    if(master .and. debug>0) write(*,'(2x,a,4(i0,1x))') "Coordinates of drag region i1,i2,j1,j2 ", drag(id)%i1, drag(id)%i2, drag(id)%j2, drag(id)%j2
    
! check if the grid range seems resonable
    if(drag(id)%i1 < 1 .or. drag(id)%i1 > im) call message("drag region x1 out of range",2)
    if(drag(id)%i2 < 1 .or. drag(id)%i2 > im) call message("drag region x2 out of range",2)
    if(drag(id)%j1 < 1 .or. drag(id)%j1 > jm) call message("drag region y1 out of range",2)
    if(drag(id)%j2 < 1 .or. drag(id)%j2 > jm) call message("drag region y2 out of range",2)
    
! convert the global grid coordinats to local grid coordinates
    call globalrect2local(drag(id)%i1,drag(id)%i2,drag(id)%j1,drag(id)%j2,drag(id)%li1,drag(id)%li2,drag(id)%lj1,drag(id)%lj2,drag(id)%inblock)

! find the approximated vertical grid coordinates (only checks one place) Needs to be improved with the average depth for the region? Oyvind

    if(drag(id)%inblock)then
! find the deepest grid cell in the vertical
        drag(id)%k1 = 0
        drag(id)%k2 = 0
        do k = 1,kb-1
            dep = z(k)*h(drag(id)%li1,drag(id)%lj1)
            if (drag(id)%z1.ge.dep) then
                drag(id)%k2 = k
                exit
            endif
        end do
        if (drag(id)%z1.lt.dep) then
            drag(id)%k2=kb-1
            write(*,*) 'Error drag_region',id,': Setting maximum depth to ', z(k)*h(drag(id)%li1,drag(id)%lj1)
        end if

! find the shallowest grid cell in the vertical
! first check if upper limit is below bottom
        if(drag(id)%z2.lt.-h(drag(id)%li1,drag(id)%lj1)) call message("top of drag region is below sea floor",2)

        do k = 1,kb-1
            dep = z(k)*h(drag(id)%li1,drag(id)%lj1)
            if (drag(id)%z2 .ge. dep) then
                drag(id)%k1 = k-1
                exit
            endif
        end do

! k1 can now be 0:
        if(drag(id)%k1 == 0) then
            drag(id)%k1 = 1
        end if
    end if
    return
100 call message("Could not read integer for drag region",2) 
101 call message("Could not read real value for regional drag coefficient",2) 
102 call message("Could not read real value for horizontal coordinates",2) 
103 call message("Could not read real value for vertical coordinates",2)     
  end subroutine initialize_drag_region

end module mod_drag_region
