module mod_forcing
!
! this module contains the definition of the global array "forcing",
! and routines to fetch and apply all forcings. scalar, time dependent
! data are normally read into the objects during initialization, while
! arrays are read below.
!
! a forcing is an object (type forcing_object) containing a list of
! fields or scalar values (the field_object) to be applied to a
! variable e.g. as a bc. the object also contains "metadata", i.e. 
! data that describes the field_objects, like type, resolution etc.
!
  use mod_bound
  use intpol
  implicit none
  
  type field_object
     !! the field sample time in seconds since the start date of the
     !! simulation
     real(8) :: time=0.0

     !! the data to be applied is stored here, arrays are allocated on
     !! demand to save space.
     real :: data=0.0, data2=0.0, data3=0.0
     real :: conc(maxnumconc) = 0.0
     real, allocatable :: data1d(:)
     real, allocatable :: data2d(:,:), data2d2(:,:), data2d3(:,:)
     real, allocatable :: data3d(:,:,:)
     
     !! files containing fields if applicable
     character(130) :: file=" ", file2=" ", file3=" "

     !! column if applicable
     integer :: col=1     
  end type field_object

  
  type forcing_object
     !! a name for identification, usually splitline(1)
     character(50) :: name=" "

     !! the file from where "meta" data was read
     character(200) :: file_name=" "
     
     !! the type of data stored in the field object
     !! type 0=scalar,1=1d, 2=2d, 3=3d, 4=river, 5=well, 6=uniform wind, ...
     integer :: type=-1

     !! grid coordinate (if applicable)
     real :: xpos=-1., ypos=-1., zpos=-1.

     !! processor local indice
     integer :: i=-10, j=-10, k=-10
     
     !! target grid point type "u", "v" or "p"
     character :: pointtype="p"
     
     !! if true, the alf parameter will be permanently set to zero,
     !! which means always to apply the nearest forcing value
     !! backwards in time
     logical :: lag=.false.
     
     !! for e.g. rivers, we need to store the river number
     integer :: id=-1
     
     !! the number of forcing fields
     integer :: numfields=0
     
     !! the list of forcing fields
     type(field_object), allocatable :: field(:)
     
     !! storage for the current field to be applied, usually a result of
     !! interpolation in time between two field objects.     
     real, allocatable :: data1d(:), data1d2(:), data1d3(:)
     real, allocatable :: data2d(:,:), data2d2(:,:), data2d3(:,:)
     real, allocatable :: data3d(:,:,:), data3d2(:,:,:), data3d3(:,:,:)

     !! storage for scalar values used by some forcing methods
     real :: rdata=0.0
     integer :: idata=0
     
     !! the grid dimension of the forcing fields
     integer :: dim1=0, dim2=0, dim3=0
          
     !! grid spacing of forcing fields
     real :: dx=-1., dy=-1.
     
     !! 3d data are always in z coordinates, this is the z coordinate.
     !! size(z) should be equal to size(data3d,3)
     real, allocatable :: z(:)    
     character(200) :: zdatafile=" "
     
     !! an integer pointer to the next and the previous field in "field"
     integer :: next=2, prev=1

     !! if the forcing field grid has the same resolution as the model
     !! grid, we can specify the location of sw corner in the forcing
     !! grid relative to the model grid with an integer offset.
     integer :: ioffset=0, joffset=0
     
     !! location of sw corner of first data point in physical coordinates (meters)
     real :: xoffset=0.0, yoffset=0.0

     !! these are pointers to source data if applicable.
     real, pointer :: source1d(:) => null(), source2d(:,:) => null(), source3d(:,:,:) => null()
     
     !! this is a pointer to where we want to apply the data, only one
     !! of these should be different from null()
     real, pointer :: target1d(:) => null(), target2d(:,:) => null(), target3d(:,:,:) => null()
     
     !! for each model (eta) grid point, point to source grid triangle and weights
     type(mapping_object), allocatable :: map(:,:)
     
     !! an integer array with the target range. we copy data into this range only.
     !! the range is converted to a block local range in setup2.
     integer :: range(6)=-1
     
     !! edge: (use integer values from mod_mpi)
     integer :: edge=-1

     logical :: timedepconc(maxnumconc) = .false.

     !! true if a region of the forcing field is on the current processor
     logical :: inblock=.false.
  end type forcing_object

  integer, parameter :: max_numforcings = 100

!! the global list of forcings
  type(forcing_object), save :: forcing(max_numforcings)

!! the global number of forcings
  integer :: numforcings = 0

!! storage for various forcing longitudes and latitudes
  real, allocatable :: ncep_lons(:), ncep_lats(:)
  real(4), allocatable :: met_lons(:,:), met_lats(:,:)


contains

  subroutine getalf(i, alf, prev, next)
!    
! for forcing number i, increment prev and next and find the new alf
! coefficient. at any given time, the alf coefficient weights between the
! available next and previous value of the forcing in question,
!    f_bound = (1-alf)*forcing_prev + alf*forcing_next
! if there is only one field, it is in "prev", and alf should be 0.
    use time_util
    use state, only : master, debug
    implicit none
    integer i, prev, next
    real alf

! special case: only one field
    if(forcing(i)%numfields == 1) then
        alf = 0
        if(time < forcing(i)%field(1)%time) alf = time/forcing(i)%field(1)%time
        prev=1
        next=1
        forcing(i)%next =  1
        forcing(i)%prev =  1
        goto 101
    end if
    
    next = forcing(i)%next
    prev = forcing(i)%prev
    
! increment next and prev until the forcing fields surround "time".
! the only time we are allowed to increment prev,next is if time is
! beyond next, and next is <= last.
    do
        if(next < forcing(i)%numfields) then
            if(time > forcing(i)%field(next)%time) then
                prev = next
                next = next + 1
            else
                exit
            end if
        else
            exit
        end if
    end do
        
    forcing(i)%next =  next
    forcing(i)%prev =  prev

! then alfa. 

! first the most common situation, time is inbetween the "prev" and the "next" time:
    if(time > forcing(i)%field(prev)%time .and. time <= forcing(i)%field(next)%time) then
        alf = (time - forcing(i)%field(prev)%time)/(forcing(i)%field(next)%time-forcing(i)%field(prev)%time)
    
! if time is before the first forcing, we will interpolate between the
! initial value of the field, which is in the target anyway, and the
! first forcing field:
    else if(time < forcing(i)%field(1)%time) then
        alf = time/forcing(i)%field(1)%time
    end if

    if(forcing(i)%lag) alf = 0.0
    
! otherwise, if time is beyond the last forcing field, we will use the
! last forcing field, which is in "next", which means alf must be 1.

    if(time > forcing(i)%field(forcing(i)%numfields)%time) alf = 1.0

101 if(master.and.debug >3) then
        write(*,'(i0,3a,i0)') i, ". forcing name: ", trim(forcing(i)%name),"Forcing type: ", forcing(i)%type
        write(*,'(2x,a,i0,1x,i0,1x,f0.2)') "prev,next,alf = ",prev,next,alf        
        select case(forcing(i)%type)
        case(4)
            write(*,'(2x,a,f0.2)') "Flux=",forcing(i)%field(prev)%data
        case default
            write(*,'(2x,a,i0,a,6(i0,1x))') " edge=",forcing(i)%edge, " range=",forcing(i)%range
            if(trim(forcing(i)%field(prev)%file).ne."") write(*,'(2x,2a)') "Forcing prev file: ", trim(forcing(i)%field(prev)%file)
            if(trim(forcing(i)%field(next)%file).ne."") write(*,'(2x,2a)') "Forcing next file: ", trim(forcing(i)%field(next)%file)
        end select
    end if
  end subroutine getalf

  
  subroutine set_initial_val_if_binreg(b,n)
    
! loop over all boundary objects in the list, and set s,t,rho_bound to
! x_initial in the boundary region if xxx_type is 4
    
    use control_parameters, only : advect_density
    implicit none
    type(boundary_object) b(:)
    integer n,i,r(4),j

    do i=1,n
        if(b(i)%inblock) then
            r = b(i)%local_range            
                
            if(.not.advect_density)then
                if(b(i)%s_type == 4) s_bound(r(1):r(2),r(3):r(4),:) = s_initial(r(1):r(2),r(3):r(4),:)                
                if(b(i)%t_type == 4) t_bound(r(1):r(2),r(3):r(4),:) = t_initial(r(1):r(2),r(3):r(4),:)
            else
                if(b(i)%rho_type == 4) rho_bound(r(1):r(2),r(3):r(4),:) = rho_initial(r(1):r(2),r(3):r(4),:)
            end if

            do j=1,maxnumconc
                if(co(j)%allocated .and. b(i)%conc_type(j) == 4) then
                    co(j)%conc_bound(r(1):r(2),r(3):r(4),:) = co(j)%conc_initial(r(1):r(2),r(3):r(4),:)
                end if
            end do
        end if
    end do
  end subroutine set_initial_val_if_binreg

  
  subroutine initialize_density_forcing(mass)
    
! subtract mass from forcing data if the 3d target is RHO_BOUND
    
    use state, only : master, debug
    use control_parameters, only : advect_density
    implicit none
    real mass
    integer i,n
    if(advect_density) then
        if(associated(rho_bound)) then
            do i=1,numforcings
                if(associated(forcing(i)%target3d, rho_bound)) then
                    if(debug > 1 .and. master) then
                        write(*,*) "forcing",i,"is density, subtracting 100+MASS from data. MASS=",mass
                    end if
                    n = forcing(i)%numfields
                    forcing(i)%field(1:n)%data  = forcing(i)%field(1:n)%data - (1000.0 + mass)
                end if
            end do
        end if
    end if
  end subroutine initialize_density_forcing
  
  
  subroutine get_and_apply_forcing()
!    
! iterate through the list of forcings, fetch and assign the forcing
! data to the target range of the given boundary array.
!    
    use state
    use time_util
    use mod_concentration
    use control_parameters
    use intpol
    implicit none
    integer next, prev, fi,i,j,k, r(6), target, li,lj,ind,dim1,dim2,gi,gj,di,dj,loc(3), dist
    real alf,val, fac
    real speed, direction, road0, ws, fact, cdrag, c1

    
! if we use the open boundary bin region stuff, the initial state
! should be assigned to s,t_bound by default, which is overwritten by
! forcings specified below.
    call set_initial_val_if_binreg(boeast, num_zones_east)
    call set_initial_val_if_binreg(bowest, num_zones_west)
    call set_initial_val_if_binreg(bonorth, num_zones_north)
    call set_initial_val_if_binreg(bosouth, num_zones_south)
        
    
    do fi=1, numforcings
        
! not all forcings needs to be interpolated in time
        if(all(forcing(fi)%type .NE. (/9/))) call getalf(fi, alf, prev, next)
        
        select case(forcing(fi)%type)
        case(-1)
            if(master) write(*,*) "forcing no",i,"name=",trim(forcing(fi)%name)," field=",next,&
                 "has not been assigned a type. confused."
            call abort()
        case(0)
            if(.not.forcing(fi)%inblock) cycle

            r = forcing(fi)%range

! todo: test if we can make the target pointer point to the boundary
! region directly.the below technique works ok in mod_mpi, which is why we try it first.
            if(associated(forcing(fi)%target3d)) then
                if(time > forcing(fi)%field(1)%time)then
                    forcing(fi)%target3d(r(1):r(2),r(3):r(4),r(5):r(6)) = &
                         forcing(fi)%field(prev)%data * (1-alf) + alf*forcing(fi)%field(next)%data
                else
                    forcing(fi)%target3d(r(1):r(2),r(3):r(4),r(5):r(6)) = &
                         forcing(fi)%target3d(r(1):r(2),r(3):r(4),r(5):r(6)) * (1-alf) + alf*forcing(fi)%field(1)%data
                end if
            else if(associated(forcing(fi)%target2d)) then
                if(time > forcing(fi)%field(1)%time)then
                    forcing(fi)%target2d(r(1):r(2),r(3):r(4)) = &
                         forcing(fi)%field(prev)%data * (1-alf) + alf*forcing(fi)%field(next)%data
                else
                    forcing(fi)%target2d(r(1):r(2),r(3):r(4)) = &
                         forcing(fi)%target2d(r(1):r(2),r(3):r(4)) * (1-alf) + alf*forcing(fi)%field(1)%data
                end if
            else
                if(master) write(*,*) "forcing no",i,"name=",trim(forcing(fi)%name)," target has not been associated"
                call abort()
            end if          
        case(1)
            if(master) write(*,*) "forcing from 1d fields in file not implemented, forcing name=",trim(forcing(fi)%name)
!                 forcing(fi)%target = forcing(fi)%field(prev)%data1d * (1-alf) + alf*forcing(fi)%field(next)%data1d

        case(2)
!            get bounary values from 2d binary files for the given edge.
            
            if(.not. forcing(fi)%inblock) cycle
            r = forcing(fi)%range
            if(prev > 1) then
                if(allocated(forcing(fi)%field(prev-1)%data2d)) deallocate(forcing(fi)%field(prev-1)%data2d)
            end if

            select case(forcing(fi)%edge)
            case(east,west)
                dim1 = j1 ; dim2 = j2
            case(north,south)
                dim1 = i1 ; dim2 = i2
            end select
            
            if(.not.allocated(forcing(fi)%z))then
                open(unit=1000013, file=(trim(data_path)//forcing(fi)%zdatafile), status="old", access="stream", form="unformatted", err=1)
                allocate(forcing(fi)%z(forcing(fi)%dim2))
                read(1000013) forcing(fi)%z
                close(1000013)
            end if
            if(.not. allocated(forcing(fi)%data2d)) then
                allocate(forcing(fi)%data2d(dim1:dim2,forcing(fi)%dim2))
            end if
            
          !! first interpolate z coordinate data horisontally onto model grid            
            if(.not. allocated(forcing(fi)%field(prev)%data2d)) then
                allocate(forcing(fi)%field(prev)%data2d(dim1:dim2,forcing(fi)%dim2))
                call ip2dforcing(forcing(fi), prev)
            end if            
            if(.not. allocated(forcing(fi)%field(next)%data2d)) then
                allocate(forcing(fi)%field(next)%data2d(dim1:dim2,forcing(fi)%dim2))
                call ip2dforcing(forcing(fi), next)
            end if            
            if(time > forcing(fi)%field(1)%time)then
                forcing(fi)%data2d = forcing(fi)%field(prev)%data2d * (1-alf) + alf*forcing(fi)%field(next)%data2d
            else
                call abort("TODO")
!                forcing(fi)%data2d = initialval * (1-alf) + alf*forcing(fi)%field(next)%data2d
            end if
            
          !! then interpolate vertically onto sigma coordinates            
            select case(forcing(fi)%edge)
            case(east,west)
                call ijslice2sigma(forcing(fi)%edge, "i", forcing(fi)%data2d, dim1, forcing(fi)%target3d, i1, j1, h, forcing(fi)%z, r, -32767.00)
            case(north,south)
                call ijslice2sigma(forcing(fi)%edge, "j", forcing(fi)%data2d, dim1, forcing(fi)%target3d, i1, j1, h, forcing(fi)%z, r, -32767.00)
            end select
            
        case(3)
            
! on the processors sitting on a piece of the boundary block, get
! boundary array values from 3d binary stream files of z coordinate
! data. horisontal resolution same as model.

            if(.not. forcing(fi)%inblock) cycle
            
            if(prev > 1) then
                if(allocated(forcing(fi)%field(prev-1)%data3d)) deallocate(forcing(fi)%field(prev-1)%data3d)
            end if
            
            if(.not.allocated(forcing(fi)%z))then
                open(unit=1000013, file=(trim(data_path)//forcing(fi)%zdatafile), status="old", access="stream", form="unformatted", err=1)
                allocate(forcing(fi)%z(forcing(fi)%dim3))
                read(1000013) forcing(fi)%z
                close(1000013)
            end if
            if(.not. allocated(forcing(fi)%data3d)) then
                allocate(forcing(fi)%data3d(forcing(fi)%dim1,forcing(fi)%dim2,forcing(fi)%dim3))
            end if
            
          !! first interpolate z coordinate data horisontally onto model grid
          ! (just do the allocation, as data is already interpolated horisontally.)
            if(.not. allocated(forcing(fi)%field(prev)%data3d)) then
                allocate(forcing(fi)%field(prev)%data3d(forcing(fi)%dim1,forcing(fi)%dim2,forcing(fi)%dim3))

                open(99, file=(trim(data_path)//forcing(fi)%field(prev)%file), form="unformatted", access="stream")
                read(99) forcing(fi)%field(prev)%data3d
                close(99)
            end if
            if(.not. allocated(forcing(fi)%field(next)%data3d)) then
                allocate(forcing(fi)%field(next)%data3d(forcing(fi)%dim1,forcing(fi)%dim2,forcing(fi)%dim3))
                
                open(99, file=(trim(data_path)//forcing(fi)%field(next)%file), form="unformatted", access="stream")
                read(99) forcing(fi)%field(next)%data3d
                close(99)
            end if            
            if(time > forcing(fi)%field(1)%time)then
                forcing(fi)%data3d = forcing(fi)%field(prev)%data3d * (1-alf) + alf*forcing(fi)%field(next)%data3d
            else
                call abort("TODO time seems less than first forcing time!")
!                forcing(fi)%data2d = initialval * (1-alf) + alf*forcing(fi)%field(next)%data2d
            end if
            
!! then interpolate vertically onto sigma coordinates
       
! TODO: we need this routine
            call vertintpol(forcing(fi))
            
        case(4)
!            rivers
            if(.not.rivlog) cycle
            if(time < forcing(fi)%field(1)%time)then
                if(master)write(*,*) "please add initial values (t=0.0) for river in file=",trim(forcing(fi)%file_name)
                call abort()
            end if
            target = forcing(fi)%id
            friver(target) = forcing(fi)%field(prev)%data  * (1-alf) + alf*forcing(fi)%field(next)%data
            sriver(target) = forcing(fi)%field(prev)%data2 * (1-alf) + alf*forcing(fi)%field(next)%data2
            triver(target) = forcing(fi)%field(prev)%data3 * (1-alf) + alf*forcing(fi)%field(next)%data3
            do i=1,maxnumconc
                if(forcing(fi)%timedepconc(i)) then
                    criver(i, target) = forcing(fi)%field(prev)%conc(i) * (1-alf) + alf*forcing(fi)%field(next)%conc(i)
                end if
            end do
        case(5)
!            well
            if(.not.upwell) cycle
            if(time < forcing(fi)%field(1)%time)then
                if(master)write(*,*) "please add initial values (t=0.0) for well in file=",trim(forcing(fi)%file_name)
                call abort()
            end if
            target = forcing(fi)%id
            fwell(target) = forcing(fi)%field(prev)%data  * (1-alf) + alf*forcing(fi)%field(next)%data
            swell(target) = forcing(fi)%field(prev)%data2 * (1-alf) + alf*forcing(fi)%field(next)%data2
            twell(target) = forcing(fi)%field(prev)%data3 * (1-alf) + alf*forcing(fi)%field(next)%data3
            do i=1,maxnumconc
                if(forcing(fi)%timedepconc(i)) then
                    cwell(i, target) = forcing(fi)%field(prev)%conc(i) * (1-alf) + alf*forcing(fi)%field(next)%conc(i)
                end if
            end do
        case(6)
!            compute wusurf and wvsurf from the wind speeds.
            if(.not.wind_forcing) cycle
            
            if(time < forcing(fi)%field(1)%time)then
                if(master)write(*,*) "please add initial values (t=0.0) for wind in file=",trim(forcing(fi)%file_name)
                call abort()
            end if
            
            speed =      forcing(fi)%field(prev)%data  * (1-alf) + alf*forcing(fi)%field(next)%data
            direction =  forcing(fi)%field(prev)%data2  * (1-alf) + alf*forcing(fi)%field(next)%data2
            wspeedu = speed*cos(direction)
            wspeedv = speed*sin(direction)
            road0=roair/rho0

            do i=1,i_eq_im
                do j=1,j_eq_jm
                    ws=sqrt(wspeedu(i,j)*wspeedu(i,j)+wspeedv(i,j)*wspeedv(i,j))
                    fact=1.2
                    if(ws.ge.11.) fact=0.49 +0.065*ws
                    cdrag=1.0e-3*fact
                    wusurf(i,j)=road0*cdrag*ws*wspeedu(i,j)
                    wvsurf(i,j)=road0*cdrag*ws*wspeedv(i,j)
                end do
            end do
            patm = pmean


        case(7,8,11,15)
            
! read wind speed, atm pressure from various binary files
            
            if(.not.wind_forcing) cycle

            if(time < forcing(fi)%field(1)%time)then
                if(master)then
                    write(*,*) "time is now ",time," forcing number ",fi
                    write(*,*) "first forcing 7 time is ",forcing(fi)%field(1)%time
                    write(*,*) "please add initial values (t=0.0) for wind in file=",trim(forcing(fi)%file_name)
                    call abort()
                end if
            end if

            if(prev > 1) then
                if(allocated(forcing(fi)%field(prev-1)%data2d)) deallocate(forcing(fi)%field(prev-1)%data2d)
                if(allocated(forcing(fi)%field(prev-1)%data2d2)) deallocate(forcing(fi)%field(prev-1)%data2d2)
                if(allocated(forcing(fi)%field(prev-1)%data2d3)) deallocate(forcing(fi)%field(prev-1)%data2d3)
            end if

            if(forcing(fi)%type==7) then
                if(.not.allocated(forcing(fi)%field(prev)%data2d)) call getncep(forcing(fi), prev)
                if(.not.allocated(forcing(fi)%field(next)%data2d)) call getncep(forcing(fi), next)
            end if
            if(forcing(fi)%type==8) then
                if(.not.allocated(forcing(fi)%field(prev)%data2d)) call getstreamwind(forcing(fi), prev)
                if(.not.allocated(forcing(fi)%field(next)%data2d)) call getstreamwind(forcing(fi), next)
            end if
            if(forcing(fi)%type==11) then
                if(.not.allocated(forcing(fi)%field(prev)%data2d)) call getmetwind(forcing(fi), prev)
                if(.not.allocated(forcing(fi)%field(next)%data2d)) call getmetwind(forcing(fi), next)
            end if
            if(forcing(fi)%type==15) then
                if(.not.allocated(forcing(fi)%field(prev)%data2d)) call getstreamwind2(forcing(fi), prev)
                if(.not.allocated(forcing(fi)%field(next)%data2d)) call getstreamwind2(forcing(fi), next)
            end if
            
            ! now, data2d has U wind, data2d2 V, all in P points.
            do j=1,j_eq_jm
                do i=1,i_eq_im
                    gi = block%fi + i -1
                    gj = block%fj + j -1
                    di = 1
                    dj = 1
                    if(gi == 1) di=0
                    if(gj == 1) dj=0                    
                    wspeedu(i,j) = 0.5*(&
                           (1-alf)*(forcing(fi)%field(prev)%data2d(i,j) + forcing(fi)%field(prev)%data2d(i-di,j))   &
                             + alf*(forcing(fi)%field(next)%data2d(i,j) + forcing(fi)%field(next)%data2d(i-di,j)))
                    
                    wspeedv(i,j) = 0.5*(&
                           (1-alf)*(forcing(fi)%field(prev)%data2d2(i,j) + forcing(fi)%field(prev)%data2d2(i,j-dj)) &
                           +   alf*(forcing(fi)%field(next)%data2d2(i,j) + forcing(fi)%field(next)%data2d2(i,j-dj)))
                end do
            end do
            
            if(atmpress .and. forcing(fi)%type .ne. 15)then
                patm = forcing(fi)%field(prev)%data2d3  * (1-alf) + alf*forcing(fi)%field(next)%data2d3
            end if
            
            road0=roair/rho0

            do j=1,j_eq_jm
                do i=1,i_eq_im
                    ws=sqrt(wspeedu(i,j)*wspeedu(i,j)+wspeedv(i,j)*wspeedv(i,j))
                    fact=1.2
                    if(ws.ge.11.) fact=0.49 +0.065*ws
                    cdrag=1.0e-3*fact
                    wusurf(i,j)=road0*cdrag*ws*wspeedu(i,j)
                    wvsurf(i,j)=road0*cdrag*ws*wspeedv(i,j)
                end do
            end do

        case(9)
! depth integrate source3d and put into target2d
            if(.not.forcing(fi)%inblock) cycle
            r = forcing(fi)%range
            forcing(fi)%target2d(r(1):r(2),r(3):r(4)) = 0.0           
            do k = 1,kb-1
                do j = r(3),r(4)
                    do i = r(1),r(2)
                        forcing(fi)%target2d(i,j) = forcing(fi)%target2d(i,j) + forcing(fi)%source3d(i,j,k)*dz(k)
                    end do
                end do
            end do
            if(master.and.debug >3) then
                write(*,'(i0,3a,i0)') fi, ". forcing name: ", trim(forcing(fi)%name)," edge=",forcing(fi)%edge
                write(*,'(2x,a,i0,a,6(i0,1x))') "Copy and depth integrate forcing, type: ", forcing(fi)%type, " range=",forcing(fi)%range
            end if
            
        case(10)
!   TSURF from climatology; compute e.g. as
!            c1 = 1.736E-5
!            tsurf = c1*( Tclim - t(:,:,1))
!   TSURF is the b.c. for T at sigma=0:   const*dT/dSigma = TSURF            
!   we observe that if Tclim is larger than T, the derivative is
!   positive, and we slowly heat up the surface as we want.
            
            if(prev > 1) then
                if(allocated(forcing(fi)%field(prev-1)%data2d)) deallocate(forcing(fi)%field(prev-1)%data2d)
            end if
            if(.not.allocated(forcing(fi)%field(prev)%data2d)) call gettempclim(forcing(fi), prev)
            if(.not.allocated(forcing(fi)%field(next)%data2d)) call gettempclim(forcing(fi), next)

            tsurf = forcing(fi)%rdata * ( &
                                        !  surface temperature from climatology                    - model temp
                 (1-alf)*forcing(fi)%field(prev)%data2d  + alf*forcing(fi)%field(next)%data2d  -  t(:,:,1) )
           
            if(debug > 3)then
                call find_max_reals(1, forcing(fi)%field(next)%data2d, val, loc, wet=.true.)
                if(master) write(*,*) "  max t clim", val, loc
                call find_min_reals(1, forcing(fi)%field(next)%data2d, val, loc, wet=.true.)
                if(master) write(*,*) "  min t clim", val, loc
            end if

        case(12)
! same as 10, but tclim is scalar            
            tsurf = forcing(fi)%rdata * ( &
                 (1-alf)*forcing(fi)%field(prev)%data  + alf*forcing(fi)%field(next)%data  -  t(:,:,1) )
            if(debug > 3)then
                if(master) write(*,*) "  t clim", (1-alf)*forcing(fi)%field(prev)%data  + alf*forcing(fi)%field(next)%data
            end if

        case(13)
! point source is a vertical line    
            if(.not. forcing(fi)%inblock) cycle
            
            if(prev > 1) then
                if(allocated(forcing(fi)%field(prev-1)%data2d)) deallocate(forcing(fi)%field(prev-1)%data2d)
            end if
            
            if(.not.allocated(forcing(fi)%field(prev)%data1d)) call getlinesrc(forcing(fi), prev)
            if(.not.allocated(forcing(fi)%field(next)%data1d)) call getlinesrc(forcing(fi), next)

            i = forcing(fi)%i
            j = forcing(fi)%j
            co(forcing(fi)%id)%conc(i,j,:) = (1-alf)*forcing(fi)%field(prev)%data1d  + alf*forcing(fi)%field(next)%data1d

        case(14)
!  get bounary values from 1d binary files and apply to the given edge of the target 2D field
            
            if(.not. forcing(fi)%inblock) cycle
            r = forcing(fi)%range
            if(prev > 1) then
                if(allocated(forcing(fi)%field(prev-1)%data1d)) deallocate(forcing(fi)%field(prev-1)%data1d)
            end if

            select case(forcing(fi)%edge)
            case(east,west)
                dim1 = j1 ; dim2 = j2
            case(north,south)
                dim1 = i1 ; dim2 = i2
            end select
            
            if(.not. allocated(forcing(fi)%data1d)) then
                allocate(forcing(fi)%data1d(dim1:dim2))
            end if
            
          !! interpolate data horisontally onto model grid            
            if(.not. allocated(forcing(fi)%field(prev)%data1d)) then
                allocate(forcing(fi)%field(prev)%data1d(dim1:dim2))
                call ip1dforcing(forcing(fi), prev)
            end if            
            if(.not. allocated(forcing(fi)%field(next)%data1d)) then
                allocate(forcing(fi)%field(next)%data1d(dim1:dim2))
                call ip1dforcing(forcing(fi), next)
            end if            
            if(time > forcing(fi)%field(1)%time)then
                forcing(fi)%data1d = forcing(fi)%field(prev)%data1d * (1-alf) + alf*forcing(fi)%field(next)%data1d
            else
                call abort("TODO get_and_apply_forcing case(14)")
!                forcing(fi)%data1d = initialval * (1-alf) + alf*forcing(fi)%field(next)%data1d
            end if
           
            
        case default
            if(master) write(*,*) "forcing no",i,"name=",trim(forcing(fi)%name)," field=",next,&
                 "unknown forcing type=",forcing(fi)%type
            call abort()
        end select
    end do
    call dump_state("get_and_apply_forcing")
    return
1   write(*,*) 'GET_AND_APPLY_FORCING ERROR: could not open or read file="',trim(forcing(fi)%zdatafile),&
         '" with z levels for forcing=',trim(forcing(fi)%name) ; call abort()
  end subroutine get_and_apply_forcing


  subroutine vertintpol(forcing)
!    
! interpolate from forcing%data3d to forcing%target3d
!    forcing%data3d is in z coordinates, with levels in forcing%z, dimensions in forcing%dim1 to dim3
!
    use mod_mpi, only : block
    use state, only : zz,kb,h
    implicit none
    type(forcing_object) forcing
    integer r(6), di, dj, li, lj,i,j

! local range
    r = forcing%range
    di = forcing%ioffset
    dj = forcing%joffset

    do lj=r(3),r(4)
        do li=r(1),r(2)
            
! the local 3d data array numbering starts at 1 in each direction. the
! the global coordinate = the local (i&j) + the offset (di&dj)
            i = block%fi + li - 1 - di
            j = block%fj + lj - 1 - dj
            select case(forcing%pointtype)
            case("p")
                call z2sigma1d(forcing%data3d(i,j,:), forcing%z, forcing%target3d(li,lj,:), zz(1:kb-1), h(li,lj))  ! or dynamic depth?
            case("u")
                call z2sigma1d(forcing%data3d(i,j,:), forcing%z, forcing%target3d(li,lj,:), zz(1:kb-1), 0.5*(h(li-1,lj)+h(li,lj)))
            case("v")
                call z2sigma1d(forcing%data3d(i,j,:), forcing%z, forcing%target3d(li,lj,:), zz(1:kb-1), 0.5*(h(li,lj-1)+h(li,lj)))
            end select
        end do
    end do
  end subroutine vertintpol


  subroutine z2sigma1d(fz, zl, fs, sl, dep, elev)
!    
! interpolate from the vertical array FZ(:) on zlevels in ZL(:) onto
! the array FS(:) given on sigma levels SL(:). ELEV the current
! elevation, DEP is the static depth. ZL(i) > 0 and ZL(i+1) > ZL(i) for all i.
!
    use qmpi, only : abort
    implicit none
    real fz(:), fs(:), zl(:), sl(:), dep
    real, optional :: elev
    integer k, numzlevels, lbot, ll
    real :: elevation, undef=-32767.00, eps=1e-6, zdyp
    
    if(dep <= 0.0) return

    numzlevels = size(fz,1)

    elevation = 0.0
    if(present(elev)) elevation = elev
    
! find last Z data point LBOT above the sea floor with valid data
    lbot = -1
    do ll = 1, numzlevels
        if (zl(ll) < dep .and. abs(fz(ll)-undef) > eps)  lbot = ll
    end do
    if(lbot == -1) call abort("1dz2sigma: no Z data points in water column")

    do k = 1, size(sl)
        zdyp = - sl(k) * (dep + elevation) - elevation
        
        if (zdyp <= zl(1)) then
           ! extrapolate upwards   
            fs(k) = fz(1)  
            cycle
        end if
        
        if (zdyp > zl(lbot)) then
            
          ! problem! the sigma layer is below the last z data point.  we
          ! should first check if we have a valid data point below LBOT. if we
          ! do, no problem.
            
            if(lbot < numzlevels)then
                if(fz(lbot+1) > undef) then
                  ! found valid data below sea floor :) interpolate!
                    fs(k) = fz(lbot) + (fz(lbot+1)-fz(lbot)) &
                                     * (zdyp-zl(lbot)) / (zl(lbot+1)-zl(lbot))
                    cycle
                end if
            else
                
              ! lbot=numzlevels? no need to search neighbours for data points below
              ! lbot. should be safe to extrapolate since none of the neighbours can
              ! be interpolated either.

                fs(k) = fz(numzlevels)
                cycle
            end if
            
          ! fallback that can/will create artificial horizontal gradients            
            fs(k) = fz(lbot)
            cycle
        end if
        
      ! most of the points are in this category
        
        do ll = 1, lbot-1
            if ((zdyp > zl(ll)) .and. (zdyp <= zl(ll+1))) then
                fs(k) = fz(ll) + (fz(ll+1)-fz(ll)) * (zdyp-zl(ll))/(zl(ll+1)-zl(ll))
            endif
        end do
    end do ! k    
  end subroutine z2sigma1d

  
  subroutine getlinesrc(forcing, ind)
    use state, only : kb,zz,h,eta,debug
    use control_parameters, only : data_path
    implicit none
    type(forcing_object) forcing
    integer ind,k,nz,i,j
    real(4) tmp1d(500)
    
    if(.not. allocated(forcing%field(ind)%data1d)) allocate(forcing%field(ind)%data1d(kb))

    nz = forcing%dim1   
    i = forcing%i
    j = forcing%j
    
    if(index(forcing%field(ind)%file, ".sr4") > 0) then
        open(1000002, file=(trim(data_path)//forcing%field(ind)%file), access="stream", form="unformatted", status="old")
        read(1000002) tmp1d(1:nz)
    else if(index(forcing%field(ind)%file, ".ur4") > 0) then
        open(1000002, file=(trim(data_path)//forcing%field(ind)%file), form="unformatted", status="old")
        read(1000002) tmp1d(1:nz)
    else
        open(1000002, file=(trim(data_path)//forcing%field(ind)%file), status="old")
        do k=1,nz
            read(1000002,*) tmp1d(k)
        end do
    end if
    close(1000002)

    call z2sigma1d(real(tmp1d(1:nz)), forcing%z, forcing%field(ind)%data1d, zz, h(i,j), eta(i,j))   
    if(debug>2) write(*,*) " got line_source z data: ",trim(forcing%field(ind)%file)," max,min=",maxval(tmp1d(1:nz)),minval(tmp1d(1:nz))
    if(debug>3) then
        write(*,'(2x,a)') "Depth    Value"
        do k=1,kb
            write(*,'(2x,f8.1,1x,f0.1)') eta(i,j)+(eta(i,j)+h(i,j))*zz(k), forcing%field(ind)%data1d(k)
        end do
    end if
  end subroutine getlinesrc

  
  subroutine gettempclim(forcing, ind)
! read "global" climatology field from file into forcing object.
    use mod_mpi, only : i1,i2,j1,j2,master,scatterfield
    use state, only : debug, im,jm
    use control_parameters, only : data_path
    implicit none
    type(forcing_object) forcing
    integer ind
    real tmp2d(im,jm)
    
    if(.not. allocated(forcing%field(ind)%data2d))  allocate(forcing%field(ind)%data2d(i1:i2,j1:j2))

    if(master) then
        open(1000002, file=(trim(data_path)//forcing%field(ind)%file), access="stream", form="unformatted", status="old")
        read(1000002) tmp2d
        close(1000002)
        if(debug>2) write(*,*) " got temperature field: ",trim(forcing%field(ind)%file)," max,min=",maxval(tmp2d),minval(tmp2d)
    end if
    call scatterfield(tmp2d, forcing%field(ind)%data2d)
    
  end subroutine gettempclim

    
  
  subroutine ip2dforcing(forcing, ind)
! 
! fetch the data for FORCING%FIELD(IND)%DATA2D from file. the data is
! stored as 2d arrays "vertical slices" along each edge of the 3d
! gloabl grid, with a resolution and offset that may differ from the
! model grid. this makes it possible to use the data both for normal
! forcing and nesting. we first interpolate the slice horisontally so
! that all grid points match the model grid. then we use a special
! purpose z2sigma to extrude the 2d slice onto the 3d grid in the
! given region.
! 
    use mod_mpi, only : i1,i2,j1,j2,east,west,north,south,i_eq_im,j_eq_jm,block,myid,abort
    use state, only : dx,dy,kb
    use control_parameters, only : data_path
    implicit none
    type(forcing_object) forcing
    integer ind,si,sj,i,j,k
    real sx1,sx2,sy1,sy2,gx,gy
    real, allocatable :: tmp2d(:,:)

! get the global "lateral" array for the given edge    
    if(.not.allocated(tmp2d)) allocate(tmp2d(forcing%dim1, forcing%dim2))
    
    open(99, file=(trim(data_path)//forcing%field(ind)%file), form="unformatted", access="stream")
    read(99) tmp2d
    close(99)

! interpolate the data horisontally along the edge into ...field(ind)%data2d,
! assuming the slice position in the normal direction to the bndr is
! already correct.
!    
! the source data: the first grid point has y position forcing%yoffset
! (measured as meters w.r. our grid origo) and resolution
! forcing%dy. the source data is "global", i.e. each processor has the
! whole bndry array for an edge.
    
    select case(forcing%edge)
    case(east,west)
! increase the source j index "sj" until sj and sj+1 surround the first j
! grid point of the block (block%fj-0.5)
        sj = 0 ; sy1=0. ; sy2=0.
        do j=1,j_eq_jm
            gy = block%fj + j - 1 - 0.5   ! the global y position of the grid point given in grid coordinates

            ! think about this one. should it be <= and < or <= and <=? what if the source grid is on top of the target?
            if(.not.(sy1 <= gy .and. gy <= sy2))then
                do
                    sj = sj + 1
                    if(sj+1 > forcing%dim1) then
                        write(*,*) myid,"gy,sy1,sy2=",gy,sy1,sy2,"j,sj=",j,sj,"forcing%dim1",forcing%dim1,"ibn,jbn",block%ibn,block%jbn
                        call abort("IP2DFORCING: sj beyond last source grid point")
                    end if
                    sy1 = (dy + forcing%yoffset + (sj-1)*forcing%dy)/dy ! global y grid coordinate of source grid pt sj.
                    sy2 = (dy + forcing%yoffset + (sj)*forcing%dy)/dy   ! same for sj + 1
                    if(gy >= sy1 .and. gy <= sy2) exit                   ! if sy1 and sy2 surround first target grid point, done
                end do
            end if
          ! interpolate
            do k=1,forcing%dim2
                if(tmp2d(sj+1,k) > -10 .and. tmp2d(sj,k) > -10) then
                    forcing%field(ind)%data2d(j,k) = (gy - sy1)/(sy2-sy1) * tmp2d(sj+1,k) + (sy2 - gy)/(sy2-sy1) * tmp2d(sj,k)
                else if(tmp2d(sj+1,k) > -10) then
                    forcing%field(ind)%data2d(j,k) = tmp2d(sj+1,k)
                else 
                    forcing%field(ind)%data2d(j,k) = tmp2d(sj,k)
                end if
            end do
        end do
    case(north,south)
        ! same as above, but for x direction
        si = 0 ; sx1=0. ; sx2=0.
        do i=1,i_eq_im
            gx = block%fi + i - 1 - 0.5
            if(.not.(sx1 <= gx .and. gx <= sx2))then
                do
                    si = si + 1
                    if(si+1 > forcing%dim1) call abort("IP2DFORCING: si beyond last source grid point")
                    sx1 = (dx + forcing%xoffset + (si-1)*forcing%dx)/dx
                    sx2 = (dx + forcing%xoffset + (si)*forcing%dx)/dx
                    if(gx >= sx1 .and. gx <= sx2) exit
                end do
            end if
            do k=1,forcing%dim2
                if(tmp2d(si+1,k) > -10 .and. tmp2d(si,k) > -10) then
                    forcing%field(ind)%data2d(i,k) = (gx - sx1)/(sx2-sx1) * tmp2d(si+1,k) + (sx2 - gx)/(sx2-sx1) * tmp2d(si,k)
                else if(tmp2d(si+1,k) > -10) then
                    forcing%field(ind)%data2d(i,k) = tmp2d(si+1,k)
                else 
                    forcing%field(ind)%data2d(i,k) = tmp2d(si,k) ! its valid or invalid anyway.
                end if
            end do
        end do
    end select
    if(allocated(tmp2d)) deallocate(tmp2d)
  end subroutine ip2dforcing
  

  subroutine ip1dforcing(forcing, ind)
! as ip2dforcing, but for 2d target arrays
    use mod_mpi, only : i1,i2,j1,j2,east,west,north,south,i_eq_im,j_eq_jm,block,myid,abort
    use state, only : dx,dy,kb
    use control_parameters, only : data_path
    implicit none
    type(forcing_object) forcing
    integer ind,si,sj,i,j,k
    real sx1,sx2,sy1,sy2,gx,gy
    real, allocatable :: tmp1d(:)

! get the global "lateral" array for the given edge    
    if(.not.allocated(tmp1d)) allocate(tmp1d(forcing%dim1))
    
    open(99, file=(trim(data_path)//forcing%field(ind)%file), form="unformatted", access="stream")
    read(99) tmp1d
    close(99)

! interpolate the data horisontally along the edge into ...field(ind)%data1d,
! assuming the slice position in the normal direction to the bndr is
! already correct.
!    
! the source data: the first grid point has y position forcing%yoffset
! (measured as meters w.r. our grid origo) and resolution
! forcing%dy. the source data is "global", i.e. each processor has the
! whole bndry array for an edge.
    
    select case(forcing%edge)
    case(east,west)
! increase the source j index "sj" until sj and sj+1 surround the first j
! grid point of the block (block%fj-0.5)
        sj = 0 ; sy1=0. ; sy2=0.
        do j=1,j_eq_jm
            gy = block%fj + j - 1 - 0.5   ! the global y position of the grid point given in grid coordinates

            ! think about this one. should it be <= and < or <= and <=? what if the source grid is on top of the target?
            if(.not.(sy1 <= gy .and. gy <= sy2))then
                do
                    sj = sj + 1
                    if(sj+1 > forcing%dim1) then
                        write(*,*) myid,"gy,sy1,sy2=",gy,sy1,sy2,"j,sj=",j,sj,"forcing%dim1",forcing%dim1,"ibn,jbn",block%ibn,block%jbn
                        call abort("IP1DFORCING: sj beyond last source grid point")
                    end if
                    sy1 = (dy + forcing%yoffset + (sj-1)*forcing%dy)/dy ! global y grid coordinate of source grid pt sj.
                    sy2 = (dy + forcing%yoffset + (sj)*forcing%dy)/dy   ! same for sj + 1
                    if(gy >= sy1 .and. gy <= sy2) exit                   ! if sy1 and sy2 surround first target grid point, done
                end do
            end if
          ! interpolate
            if(tmp1d(sj+1) > -10 .and. tmp1d(sj) > -10) then
                forcing%field(ind)%data1d(j) = (gy - sy1)/(sy2-sy1) * tmp1d(sj+1) + (sy2 - gy)/(sy2-sy1) * tmp1d(sj)
            else if(tmp1d(sj+1) > -10) then
                forcing%field(ind)%data1d(j) = tmp1d(sj+1)
            else 
                forcing%field(ind)%data1d(j) = tmp1d(sj)
            end if
        end do
    case(north,south)
        ! same as above, but for x direction
        si = 0 ; sx1=0. ; sx2=0.
        do i=1,i_eq_im
            gx = block%fi + i - 1 - 0.5
            if(.not.(sx1 <= gx .and. gx <= sx2))then
                do
                    si = si + 1
                    if(si+1 > forcing%dim1) call abort("IP1DFORCING: si beyond last source grid point")
                    sx1 = (dx + forcing%xoffset + (si-1)*forcing%dx)/dx
                    sx2 = (dx + forcing%xoffset + (si)*forcing%dx)/dx
                    if(gx >= sx1 .and. gx <= sx2) exit
                end do
            end if
            if(tmp1d(si+1) > -10 .and. tmp1d(si) > -10) then
                forcing%field(ind)%data1d(i) = (gx - sx1)/(sx2-sx1) * tmp1d(si+1) + (sx2 - gx)/(sx2-sx1) * tmp1d(si)
            else if(tmp1d(si+1) > -10) then
                forcing%field(ind)%data1d(i) = tmp1d(si+1)
            else 
                forcing%field(ind)%data1d(i) = tmp1d(si) ! its valid or invalid anyway.
            end if
        end do
    end select
    if(allocated(tmp1d)) deallocate(tmp1d)
  end subroutine ip1dforcing


  
  subroutine getdimensions(file, dim1, dim2)
    use netcdf
    use state, only : master, debug
    use control_parameters, only : data_path
    use mod_mpi, only : mbroadcast
    implicit none
    integer dim1,dim2,i
    character(*) file
    character(20) names(nf90_max_var_dims)
    integer shape(nf90_max_var_dims)
    integer ncfile, varid, dimids(nf90_max_var_dims)

    if(master)then

! get netcdf file handle  
       call nc(nf90_open((trim(data_path)//file), mode=0, ncid=ncfile))

! get the integer handle for a variable that we know we have in the NCEP data:
       call nc(nf90_inq_varid(ncfile, "uwnd", varid))
  
! get id's for the dimensions we know we have in the files
       call nc(nf90_inquire_variable(ncfile, VarId, dimids = dimids))
  
! get the dimensions
       call nc(nf90_inquire_dimension(ncfile, dimids(1), name=names(1), len=shape(1)))
       call nc(nf90_inquire_dimension(ncfile, dimids(2), name=names(2), len=shape(2)))
       call nc(nf90_inquire_dimension(ncfile, dimids(3), name=names(3), len=shape(3)))
       call nc(nf90_inquire_dimension(ncfile, dimids(4), name=names(4), len=shape(4)))
       call nc(nf90_close(ncfile))

       dim1 = shape(1)
       dim2 = shape(2)
    
       if(debug > 1)then
          write(*,*) " Using GETDIMENSIONS to find dim1,dim2 from file ",trim(file)
          write(*,*) "  dimnames=", (trim(names(i))//" ",i=1,4)
          write(*,'(3x,a,4(i0,","))') "shape=", shape(1:4)
       end if
    end if
    call mbroadcast(dim1,dim2)
  end subroutine getdimensions

  
  subroutine getncep(forcing, ind)
! NB certainly does not work with -r8 option!
! Get netcdf variables called "uwnd","vwnd","pres", and
! "lon","lat" from the given netcdf files (FORCING%FIELD(IND)%FILE, FORCING%FIELD(IND)%FILE2, FORCING%FIELD(IND)%FILE3)    
! for forcing number IND. the time index in the netcdf file for the given time is stored in FORCING%FIELD(IND)%COL.
! field dimensions are fetched via GETDIMENSIONS above, and
! we assume all fields to be of the same size and on the same grid :)
!
    use netcdf    
    use mod_mpi, only : i1,i2,j1,j2,master,broadcast,block,mbroadcast,abort
    use grid_util, only : gr2sph
    use state, only : longitude, latitude, xpole, ypole, dx, dy,  &
         ylong, minlon, maxlon, minlat, maxlat, debug
    use control_parameters, only :data_path
    use util, only : binsearch
    implicit none
    
    type(forcing_object) forcing
    integer ind, ncfile, varid, i,j,ifail
    real x,y, gxpole,gypole
    real, allocatable :: tmp2d(:,:), tmp2d2(:,:), angle_bom(:,:), speed(:,:), angle(:,:), tmp1d(:)
    integer lon1,lon2,lat1,lat2
    real pi

    pi = 4*atan(1.0)
    
! we only keep the arrays for prev and next forcing in memory    
    if(.not. allocated(forcing%field(ind)%data2d))  allocate(forcing%field(ind)%data2d(i1:i2,j1:j2))
    if(.not. allocated(forcing%field(ind)%data2d2)) allocate(forcing%field(ind)%data2d2(i1:i2,j1:j2))
    if(.not. allocated(forcing%field(ind)%data2d3)) allocate(forcing%field(ind)%data2d3(i1:i2,j1:j2))

! we got the array dimensions from the netcdf files brute force by
! calling getdimensions from setup2 (should be improved)
    allocate(tmp2d(forcing%dim1, forcing%dim2), tmp2d2(forcing%dim1, forcing%dim2))

! fetch the longitudes and latitudes of the grid only once
    if(.not. allocated(ncep_lons)) then
        allocate(ncep_lons(forcing%dim1), ncep_lats(forcing%dim2))

        if(master) then
            call nc(nf90_open((trim(data_path)//forcing%field(ind)%file), mode=0, ncid=ncfile))
            
            call nc(nf90_inq_varid(ncfile, "lon", varid))        
            call nc(nf90_get_var(ncfile, varid, ncep_lons))

            call nc(nf90_inq_varid(ncfile, "lat", varid))        
            call nc(nf90_get_var(ncfile, varid, ncep_lats))
            
            call nc(nf90_close(ncfile))
        end if

! ncep data is stored with the lon axis ranging from 0 to 360, and the lat axis from 90 to -90.        
! our data has negative lon to the west, and the lat direction is
! positive northward. so we need to swap the lon direction and reverse
! the lat direction. 
        if(mod(forcing%dim1,2).ne.0) call abort("GETNCEP: cannot swap lon axis, dim1 assumed even")
        allocate(tmp1d(forcing%dim1))
        tmp1d(1:forcing%dim1/2) = ncep_lons(forcing%dim1/2+1:) - 360.
        tmp1d(forcing%dim1/2+1:) = ncep_lons(1:forcing%dim1/2)       
        ncep_lons = tmp1d
        deallocate(tmp1d)

        ncep_lats = ncep_lats(forcing%dim2:1:-1)
        call broadcast(ncep_lons)
        call broadcast(ncep_lats)

! find the region we need in the ncep grid, given minlon, maxlat etc.
        lon1 = binsearch(minlon, ncep_lons) - 1
        lon2 = binsearch(maxlon, ncep_lons)
        lat1 = binsearch(minlat, ncep_lats) - 1
        lat2 = binsearch(maxlat, ncep_lats)
        if(master.and.debug > 2) write(*,'(a,i0,1x,i0,a,i0,1x,i0)') "GETNCEP array range lon: ",lon1,lon2," lat: ",lat1,lat2

!       allocate(forcing%map(i1:i2,j1:j2))    
!       call getbilinmap(ncep_lons(lon1:lon2), ncep_lats(lat1:lat2), forcing%map)
    end if

    call mbroadcast(lon1,lon2,lat1,lat2)
    
    if(master) then
        !! get the file handle for the netcdf file containing "u" wind
        call nc(nf90_open((trim(data_path)//forcing%field(ind)%file), mode=0, ncid=ncfile))
        
        !! get the handle for the variable we want
        call nc(nf90_inq_varid(ncfile, "uwnd", varid))
        
        !! read the actual data into tmp2d
        !! the time index is forcing%field(ind)%col, which corresponds to the iso time in the first column of the header file.
        call nc(nf90_get_var(ncfile, varid, tmp2d, start=(/1,1,1,forcing%field(ind)%col/), count=(/forcing%dim1, forcing%dim2,1,1/)))
        call nc(nf90_close(ncfile))
        
! swap the halves as for longitude above
        tmp2d2(1:forcing%dim1/2,:) = tmp2d(forcing%dim1/2+1:,:)        
        tmp2d2(forcing%dim1/2+1:,:) = tmp2d(1:forcing%dim1/2,:)       
! then reverse latitude
        tmp2d = tmp2d2(:,forcing%dim2:1:-1)       
    end if

    !! optimization possibility: instead of broadcasting the (global
    !! field as this) pick out the region that cover our grid and
    !! broadcast that. since the grid is 2d and not to large we just
    !! do this for now
    call broadcast(tmp2d(lon1:lon2,lat1:lat2))
    
    ! work in p (pointtype=0) points until the rotation and all that stuff is ok. 
    call ll2grid(tmp2d(lon1:lon2,lat1:lat2), ncep_lons(lon1:lon2), ncep_lats(lat1:lat2), forcing%field(ind)%data2d, pointtype=0)
 !  call gip(tmp2d(lon1:lon2,lat1:lat2), forcing%field(ind)%data2d, forcing%map, method="bilinear")

    
    ! get the nortward wind field
    if(master)then
        call nc(nf90_open((trim(data_path)//forcing%field(ind)%file2), mode=0, ncid=ncfile))
        call nc(nf90_inq_varid(ncfile, "vwnd", varid))
        call nc(nf90_get_var(ncfile, varid, tmp2d, start=(/1,1,1,forcing%field(ind)%col/), count=(/forcing%dim1, forcing%dim2,1,1/)))
        call nc(nf90_close(ncfile))

! swap'nreverse
        tmp2d2(1:forcing%dim1/2,:) = tmp2d(forcing%dim1/2+1:,:)        
        tmp2d2(forcing%dim1/2+1:,:) = tmp2d(1:forcing%dim1/2,:)       
        tmp2d = tmp2d2(:,forcing%dim2:1:-1)        
    end if
    call broadcast(tmp2d(lon1:lon2,lat1:lat2))
    call ll2grid(tmp2d(lon1:lon2,lat1:lat2), ncep_lons(lon1:lon2), ncep_lats(lat1:lat2), forcing%field(ind)%data2d2, pointtype=0)
!  call gip(tmp2d(lon1:lon2,lat1:lat2), forcing%field(ind)%data2d2, forcing%map, method="bilinear")

    ! and sea level pressure
    if(master)then
        call nc(nf90_open((trim(data_path)//forcing%field(ind)%file3), mode=0, ncid=ncfile))
        call nc(nf90_inq_varid(ncfile, "pres", varid))
        call nc(nf90_get_var(ncfile, varid, tmp2d, start=(/1,1,forcing%field(ind)%col/), count=(/forcing%dim1,forcing%dim2,1/)))
        call nc(nf90_close(ncfile))

! swap'nreverse
        tmp2d2(1:forcing%dim1/2,:) = tmp2d(forcing%dim1/2+1:,:)        
        tmp2d2(forcing%dim1/2+1:,:) = tmp2d(1:forcing%dim1/2,:)       
        tmp2d = tmp2d2(:,forcing%dim2:1:-1)        
    end if
    call broadcast(tmp2d(lon1:lon2,lat1:lat2))
    call ll2grid(tmp2d(lon1:lon2,lat1:lat2), ncep_lons(lon1:lon2), ncep_lats(lat1:lat2), forcing%field(ind)%data2d3, pointtype=0)
!  call gip(tmp2d(lon1:lon2,lat1:lat2), forcing%field(ind)%data2d3, forcing%map, method="bilinear")
    
    allocate(speed(i1:i2,j1:j2), angle(i1:i2,j1:j2), angle_bom(i1:i2,j1:j2))
    
! find speed and angle:
! in lon-lat grids the angle is the angle between the velocity vector
! and the eastward tangent. seen from above, the longitude of each
! point will (alone) define the local rotation w.r. the polar
! stereographic grid. in our case the grid y axis points north along
! the meridian where longitude=ylong. the ATAN2 function gives us this
! angle directly, as it is the same as the principal arg of a complex
! number:
    
    speed = sqrt(forcing%field(ind)%data2d**2 + forcing%field(ind)%data2d2**2)
    angle = atan2(forcing%field(ind)%data2d2, forcing%field(ind)%data2d)

! sanity test:
!  if the wind dir is eastward, angle is 0. if we are in a point where
!  longitude=ylong this means 0 degree angle w.r. the grid.
    
    angle_bom = angle + pi*(longitude - ylong)/180.

! uwind and vwind now in p points but with model direction
    forcing%field(ind)%data2d = speed*cos(angle_bom)
    forcing%field(ind)%data2d2 = speed*sin(angle_bom)
            
    deallocate(tmp2d, tmp2d2, speed, angle, angle_bom)
  end subroutine getncep


  subroutine getstreamwind(forcing, ind)
!
! Get wind u,v and p from files (FORCING%FIELD(IND)%FILE, FORCING%FIELD(IND)%FILE2, FORCING%FIELD(IND)%FILE3)    
! for forcing number IND. 
!
    use mod_mpi, only : i1,i2,j1,j2,master,broadcast,block,mbroadcast
    use intpol, only : ll2grid
    use grid_util, only : gr2sph
    use state, only : longitude, latitude, xpole, ypole, dx, dy,  &
         ylong, minlon, maxlon, minlat, maxlat, debug
    use control_parameters, only : data_path
    implicit none
    
    type(forcing_object) forcing
    integer ind, i,j
    real x,y,pi
    real, allocatable :: tmp2d(:,:), angle_bom(:,:), speed(:,:), angle(:,:)
    
    pi = 4*atan(1.0)
    
! we only keep the arrays for prev and next forcing in memory    
    if(.not. allocated(forcing%field(ind)%data2d))  allocate(forcing%field(ind)%data2d(i1:i2,j1:j2))
    if(.not. allocated(forcing%field(ind)%data2d2)) allocate(forcing%field(ind)%data2d2(i1:i2,j1:j2))
    if(.not. allocated(forcing%field(ind)%data2d3)) allocate(forcing%field(ind)%data2d3(i1:i2,j1:j2))

    allocate(tmp2d(forcing%dim1, forcing%dim2))

    if(master) then
        call gettmp2d(trim(data_path)//forcing%field(ind)%file)        
        if(debug>2) write(*,*) " got eastward wind: ",trim(forcing%field(ind)%file)," max,min=",maxval(tmp2d),minval(tmp2d)
    end if
    call broadcast(tmp2d)
    
    ! work in p (pointtype=0) points until the rotation and all that stuff is ok. 
    call ll2grid(tmp2d, ncep_lons, ncep_lats, forcing%field(ind)%data2d, pointtype=0, r1=-40., r2=40.)    
!    call gip(tmp2d, forcing%field(ind)%data2d, forcing%map, method="bilinear")

    ! get the nortward wind field
    if(master)then
        call gettmp2d(trim(data_path)//forcing%field(ind)%file2)        
        if(debug>2) write(*,*) " got northward wind: ",trim(forcing%field(ind)%file2)," max,min=",maxval(tmp2d),minval(tmp2d)
    end if
    call broadcast(tmp2d)
    call ll2grid(tmp2d, ncep_lons, ncep_lats, forcing%field(ind)%data2d2, pointtype=0, r1=-40., r2=40.)
!    call gip(tmp2d, forcing%field(ind)%data2d2, forcing%map, method="bilinear")

    ! and sea level pressure
    if(master)then
        call gettmp2d(trim(data_path)//forcing%field(ind)%file3)
        if(debug>2) write(*,*) " got pressure: ",trim(forcing%field(ind)%file3)," max,min=",maxval(tmp2d),minval(tmp2d)
    end if
    call broadcast(tmp2d)
    call ll2grid(tmp2d, ncep_lons, ncep_lats, forcing%field(ind)%data2d3, pointtype=0, r1=70000., r2=130000.)
!    call gip(tmp2d, forcing%field(ind)%data2d3, forcing%map, method="bilinear")

    
! find speed and angle:
! in lon-lat grids the angle is the angle between the velocity vector
! and the eastward tangent. seen from above, the longitude of each
! point will (alone) define the local rotation w.r. the polar
! stereographic grid. in our case the grid y axis points north along
! the meridian where longitude=ylong. the ATAN2 function gives us this
! angle directly, as it is the same as the principal arg of a complex
! number:
    
    allocate(speed(i1:i2,j1:j2), angle(i1:i2,j1:j2), angle_bom(i1:i2,j1:j2))
    
    speed = sqrt(forcing%field(ind)%data2d**2 + forcing%field(ind)%data2d2**2)
    angle = atan2(forcing%field(ind)%data2d2, forcing%field(ind)%data2d)
    
! sanity test:
!  if the wind dir is eastward, angle is 0. if we are in a point where
!  longitude=ylong this means 0 degree angle w.r. the grid.
    
    angle_bom = angle + (longitude - ylong)*pi/180.

! uwind and vwind now in p points but with model direction
    forcing%field(ind)%data2d = speed*cos(angle_bom)
    forcing%field(ind)%data2d2 = speed*sin(angle_bom)
            
    deallocate(tmp2d, speed, angle, angle_bom)

  contains

    subroutine gettmp2d(filename)
! kludge to support the -r8 compiler option, assuming all input data is REAL(KIND=4)      
      character(*) filename
      real(4), allocatable :: tmp2dr4(:,:)

      open(1000002, file=filename, access="stream", form="unformatted", status="old")
      if(kind(tmp2d) .ne. 4)then
          allocate(tmp2dr4(forcing%dim1, forcing%dim2))
          read(1000002) tmp2dr4
          tmp2d = tmp2dr4
          deallocate(tmp2dr4)
      else
          read(1000002) tmp2d
      end if
      close(1000002)
    end subroutine gettmp2d   
  end subroutine getstreamwind


  subroutine getstreamwind2(forcing, ind)
!
! Get wind u and v from files (FORCING%FIELD(IND)%FILE, FORCING%FIELD(IND)%FILE2)    
! for forcing number IND. 
!
    use mod_mpi, only : i1,i2,j1,j2,master,scatterfield
    use state, only : im,jm,debug
    use control_parameters, only : data_path
    implicit none
    
    type(forcing_object) forcing
    integer ind, i,j
    real x,y
    real, allocatable :: tmp2d(:,:)  
    
! we only keep the arrays for prev and next forcing in memory
    if(.not. allocated(forcing%field(ind)%data2d))  allocate(forcing%field(ind)%data2d(i1:i2,j1:j2))
    if(.not. allocated(forcing%field(ind)%data2d2)) allocate(forcing%field(ind)%data2d2(i1:i2,j1:j2))

    allocate(tmp2d(im,jm))

    if(master) then
        call gettmp2d(trim(data_path)//forcing%field(ind)%file)        
        if(debug>2) write(*,*) " got eastward wind: ",trim(forcing%field(ind)%file)," max,min=",maxval(tmp2d),minval(tmp2d)
    end if
    call scatterfield(tmp2d, forcing%field(ind)%data2d)
    
! get the nortward wind field
    if(master)then
        call gettmp2d(trim(data_path)//forcing%field(ind)%file2)        
        if(debug>2) write(*,*) " got northward wind: ",trim(forcing%field(ind)%file2)," max,min=",maxval(tmp2d),minval(tmp2d)
    end if
    call scatterfield(tmp2d, forcing%field(ind)%data2d2)
    
    deallocate(tmp2d)

  contains

    subroutine gettmp2d(filename)
! kludge to support the -r8 compiler option, assuming all input data is REAL(KIND=4)      
      character(*) filename
      real(4), allocatable :: tmp2dr4(:,:)

      open(1000002, file=filename, access="stream", form="unformatted", status="old")
      if(kind(tmp2d) .ne. 4)then
          allocate(tmp2dr4(forcing%dim1, forcing%dim2))
          read(1000002) tmp2dr4
          tmp2d = tmp2dr4
          deallocate(tmp2dr4)
      else
          read(1000002) tmp2d
      end if
      close(1000002)
    end subroutine gettmp2d   
  end subroutine getstreamwind2
  

  subroutine getmetwind(forcing, ind)
!
! Very similar to getstreamwind, but gets speed, angle and p from files
! (FORCING%FIELD(IND)%FILE, FORCING%FIELD(IND)%FILE2, FORCING%FIELD(IND)%FILE3) for forcing number IND.
!
    use mod_mpi, only : i1,i2,j1,j2,master,broadcast,block,mbroadcast,dr
    use intpol, only : ll2grid
    use grid_util, only : gr2sph
    use state, only : longitude, latitude, xpole, ypole, dx, dy,  &
         ylong, minlon, maxlon, minlat, maxlat, debug, pmean
    use control_parameters, only : data_path
    use pgplot_core
    implicit none

    type(forcing_object) forcing
    integer ind, i,j,t,r(4)
    real pi,x,y,xp(4),yp(4)
    character(12)lab
    real(kind=4), allocatable :: tmp2d(:,:), tmp2d2(:,:), angle_bom(:,:), speed(:,:), angle(:,:)
        
    pi = 4*atan(1.0)

! we only keep the arrays for prev and next forcing in memory
    if(.not. allocated(forcing%field(ind)%data2d))  allocate(forcing%field(ind)%data2d(i1:i2,j1:j2))
    if(.not. allocated(forcing%field(ind)%data2d2)) allocate(forcing%field(ind)%data2d2(i1:i2,j1:j2))
    if(.not. allocated(forcing%field(ind)%data2d3)) allocate(forcing%field(ind)%data2d3(i1:i2,j1:j2))

    allocate(tmp2d(forcing%dim1, forcing%dim2), tmp2d2(forcing%dim1, forcing%dim2))
    allocate(speed(forcing%dim1, forcing%dim2), angle(forcing%dim1, forcing%dim2), angle_bom(forcing%dim1, forcing%dim2))

    if(master) then
        open(1000002, file=(trim(data_path)//forcing%field(ind)%file), access="stream", form="unformatted", status="old")
        read(1000002) speed
        close(1000002)
        if(debug>2) write(*,*) " got speed: ",trim(forcing%field(ind)%file)," max,min=", maxval(speed), minval(speed)
    end if
    call broadcast(speed)

    if(master)then
        open(1000002, file=(trim(data_path)//forcing%field(ind)%file2), access="stream", form="unformatted", status="old")
        read(1000002) angle
        close(1000002)        
        if(debug>2) write(*,*) " got angle: ",trim(forcing%field(ind)%file2)," max,min=",maxval(angle),minval(angle)
    end if
    call broadcast(angle)


! translate speed and angle on the lon,lat grid into u,v velocity vectors
! in the model coordinate system.
!
! in the expected lon-lat grid ANGLE should be the angle between the
! speed vector and north. the longitude of each point define the local
! rotation w.r. the bom grid. in our polar stereo grid the y axis
! points north along the meridian where longitude=ylong. angle_bom is
! the angle between the speed vector and the x axis (i dim) of the bom
! grid.
!
! sanity tests:
! 1) if a vector with lon=-32 points north (angle=0) and ylong=58, the
!    angle in the bom grid should be 0.
! 2) same, but pointing east, i.e. angle=90, should give bom angle=-90.
! 3) if a vector with lon=ylong points north, angle=0, bom angle should be 90.

! find model direction of speed vectors
    angle_bom = (90 - angle  + met_lons - ylong)*pi/180.


! get uwind and vwind
    tmp2d  = speed*cos(angle_bom)
    tmp2d2 = speed*sin(angle_bom)

! finally, interpolate onto model grid
    call gip(tmp2d, forcing%field(ind)%data2d, forcing%map)
    call gip(tmp2d2, forcing%field(ind)%data2d2, forcing%map)
    
    ! and sea level pressure
    if(master)then
        open(1000002, file=(trim(data_path)//forcing%field(ind)%file3), access="stream", form="unformatted", status="old")
        read(1000002) tmp2d
        close(1000002)
        if(debug>2) write(*,*) " got pressure: ",trim(forcing%field(ind)%file3)," max,min=",maxval(tmp2d),minval(tmp2d)
    end if
    call broadcast(tmp2d)
    
    call gip(tmp2d, forcing%field(ind)%data2d3, forcing%map)
    
    deallocate(tmp2d, tmp2d2, speed, angle, angle_bom)
  end subroutine getmetwind


  subroutine nc(id)
    use netcdf
    use qmpi
! netcdf error handler wrapper, usage:
!    call nc(nf90_routine(....))    
    integer id

    if(myid .ne. 0) call abort("netcdf should only run on master!")

    select case(id)
    case(NF90_NOERR)
        return
    case(NF90_EBADID)
        write(*,*) "getncep: invalid netcdf file id"
    case default
        write(*,*) "netcdf status=",id
        call abort("getncep: not yet taught how to handle this netcdf error status=")
    end select
  end subroutine nc

  subroutine add_line_source(splitline)
!     
!  line_source <name> <concX> <x> <y> <z level file> <forcing file>
!
    use qmpi, only : master, message
    use mod_mpi, only : globalind2local, broadcast
    use state, only : debug, dx, dy
    use time_util, only : initial_time, timediff, get_iso8601
    use control_parameters, only : data_path
    use util, only : split, get_lines
    use file_util, only: readfield
    implicit none
    character(*) splitline(:)
    integer linenumbers(400),i, forcing_time(6),k
    character(200) forcing_lines(400), forcing_splitline(2)
    
    numforcings = numforcings + 1
    if(master) write(*,'(2x,a,i0)') "Forcing id=", numforcings
    
    if(numforcings > max_numforcings) call message("too many forcings with input from file, increase max_numforcings",2)
    
    forcing(numforcings)%name = splitline(2)
    read(splitline(3),*,err=1,end=1) forcing(numforcings)%id   ! the conc id
    read(splitline(4:5),*,err=2,end=2) forcing(numforcings)%xpos, forcing(numforcings)%ypos

    forcing(numforcings)%file_name = trim(data_path)//splitline(7)

    call readfield(forcing(numforcings)%z, trim(data_path)//splitline(6))
    forcing(numforcings)%dim1 = size(forcing(numforcings)%z, 1)
    
    if(master.and.debug>2) then
        write(*,'(2x,a)') "Line source z levels:"
        do k=1,forcing(numforcings)%dim1
            write(*,'(2x,i4,1x,f0.1)') k,forcing(numforcings)%z(k)
        end do        
    end if
   
    
    forcing(numforcings)%xpos = (forcing(numforcings)%xpos + dx)/dx  ! grid coordinate
    forcing(numforcings)%ypos = (forcing(numforcings)%ypos + dy)/dy
    forcing(numforcings)%i = nint(forcing(numforcings)%xpos+0.5)     ! cell containing grid coo
    forcing(numforcings)%j = nint(forcing(numforcings)%ypos+0.5)

    call globalind2local(forcing(numforcings)%i, forcing(numforcings)%j, forcing(numforcings)%inblock)

    if(master) write(*,'(2x,a,f0.1,",",f0.1)') "gx,gy=", forcing(numforcings)%xpos,forcing(numforcings)%ypos
    
    forcing(numforcings)%type = 13    

    if(master) call get_lines(forcing(numforcings)%file_name, forcing(numforcings)%numfields, forcing_lines, linenumbers)
    if(debug>0 .and. master) write(*,*) " got",forcing(numforcings)%numfields,&
         "fields, forcing file=",trim(forcing(numforcings)%file_name)
        
    call broadcast(forcing(numforcings)%numfields)
    call broadcast(forcing_lines)
    call broadcast(linenumbers)
        
    allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))
        
! the file format for the forcing file is lines like this
!   xxxx-xx-xxTxx:xx:xx  file1  
        
    do i=1,forcing(numforcings)%numfields
        call split(forcing_lines(i), forcing_splitline)
            
       !! if there is a t in the forcing time string, it must be in iso format
        if(index(forcing_splitline(1),"T") > 0) then
            call get_iso8601(forcing_splitline(1), forcing_time)
            forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
        else
            read(forcing_splitline(1),*,err=3,end=3) forcing(numforcings)%field(i)%time
        end if
        
       !! the file containing the concentration profile for the given time
        forcing(numforcings)%field(i)%file = forcing_splitline(2)
    end do
    
    return
1   call message("could not read integer id for concentration target for line_source",2)
2   call message("could not read x,y position for line source",2)
3   call message("could not read float for time",2)
4   call message("could not read integer for number of z levels in line source data files",2)
5   call message("could not get line_source z levels from stream binary file",2)
6   call message("could not get line_source z levels from unformatted binary file",2)
7   call message("could not get line_source z levels from ascii file",2)    
  end subroutine add_line_source
  
end module mod_forcing
