MODULE MOD_MPI
!  
! A collection of high level communication primitives to supplement module QMPI
! with supposedly common operations for parallelization of problems on regular grids.
! the library sets up, and works on, 2d and 3d distributed arrays using
! a 2d block topology for the first 2 array dimensions.  in most ocean
! and atmospheric models these 2 dimensions represent the horisontal
! coordinates.
!
! Helge Avlesen <avle@ii.uib.no>, Para//ab
!
!
!The interface routines are:
!
! SUBROUTINE INIT_MPILIB(IM,JM,[KB,FGLOBAL,CRITERIA])
!   initializes mpilib related datastructures and constants.
!   IM,JM is the dimensions of the 2D grid, KB the optional 3rd dimension, if not given assumed to be 1.
!   If FGLOBAL is present, only cells where FGLOBAL > CRITERIA are regarded as active in the topology.
!   a block with only inactive cells may be deleted from the 2d topology if desired. 
! SUBROUTINE SCATTERFIELD(FGLOBAL, FLOCAL)
!   split up fglobal, 2d or 3d global array, and hand out local pieces to the right processors.
!   collective operation that must be called by all processors.
! SUBROUTINE GATHERFIELD(FLOCAL, FGLOBAL, IM,JM,KB)
!   get flocal from each process, and assemble global field fglobal process 0
!   collective operation that must be called by all processors.
! SUBROUTINE TAG_GHOST(FIELD, SIDES, WIDTH)
!   add a 2d or 3d field, along with a list of sides to update, to a list of fields that will
!   get their ghost zones updated. sides is a sum of any of the integer parameters east,west,north,south.
!   width tells how far out the ghost zones should be updated, the default is the full width.
!   normally, only width=1 is needed, so use that to "optimize" the amount of data sent/recived.  
! SUBROUTINE UPDATE_GHOST
!   perform the ghost zone update as specified in the lists built up with tag_ghost.
! SUBROUTINE COPY_REGION(FROM, TO, KB, F,F2,F3,F4)
!   inside distributed array f, copy the numbers in the from range into the to range.
!   perform the same operation on f2,f3,f4 if any of them are given.
! REAL FUNCTION GLOBVAL(F,GI,GJ[,GK])
!   return the value of the (default real) global array f(gi,gj,gk) (omit gk if f is 2d)
!   expensive operation, as it is collective, so use sparingly.  
! SUBROUTINE FIND_MAX_REALS(KB, F, VAL, LOC, ROOT)
!   return the global maximum of distributed array f, with third dim kb. also return
!   global position in the integer array loc(3).
!   expensive operation, as it is collective, so use sparingly.  
! SUBROUTINE FIND_MIN_REALS(KB, F, VAL, LOC, ROOT)
!   return the global minimum and its location.
!   expensive operation, as it is collective, so use sparingly.  
! SUBROUTINE GLOBALIND2LOCAL(GI,GJ,INBLOCK)
!   convert from global to local indices, inblock=.false. if a
!   coordinate is not on the block. _ng routine excludes ghost zone
! SUBROUTINE GLOBALCOO2LOCAL(GX,GY,INBLOCK)
!   similar to previous, but gx,gy are coordinates instead of indices.  
! SUBROUTINE GLOBALRECT2LOCAL(GI1,GI2,GJ1,GJ2,LI1,LI2,LJ1,LJ2,INBLOCK)
!   convert the 2 global coordinate pairs gi1,gj1 gi2,gj2 into local
!   coordinates. set inblock to false if the region gi1:gi2,gj1,gj2 is
!   not overlapping with the current block.
! SUBROUTINE GLOBALRANGE2LOCAL(BI,BJ,RANGE,INBLOCK)
!   as previous, but can compute inblock for any block on any
!   processor. the coordinate range is changed "in place".
! INTEGER FUNCTION GI2BLOCK(GI)   
!   return the block coordinate bi of global point gi
! INTEGER FUNCTION GJ2BLOCK(GJ)   
!   return the block coordinate bj of global point gj
!
! Support routines
!
! subroutine blocks_covering_region(reg, bli1,bli2,blj1,blj2)
! subroutine find_block_ranges(criteria)
! subroutine find_block_neighbours()
! subroutine find_psum_block_ranges()
! subroutine reallocate_sendbuffers
! subroutine sec2ind(n,i,j,k)
! subroutine ind2sec(i,j,k,n)
! real function psum(f,im,jm,kb)
!   find the sum of the elements of the distributed array f, with global dimensions
!   im,jm,kb. psum uses a (parallelized) blocking algorithm to obtain a sum not
!   depending on the number of processors. 
!   the interface "parsum" in norms.f90 should be used instead of this.  
!
  use qmpi
  
  implicit none

  private globval2d, globval3d, genglobval
  
! psum constants
  integer numpbi,numpbj,mblszi,mblszj

! timing
  integer(8) :: commtime1=0, commtime2=0, totcommtime=0, totcopytime=0, totsumtime=0
  

! the grid
  
  integer :: gim, gjm, gkb                    ! global grid dimension (set by init_mpilib) 
  integer :: ghost_wi=2, ghost_wj=2           ! width of ghost zone
  integer, parameter :: default_ghost_wi=2, default_ghost_wj=2
  logical :: omit_dry_blocks=.true.          ! whether or not to include blocks with no ocean

! if true, the east and west edge of the grid is connected wraparound style
  logical :: periodic_ew=.false.

  
! for copy_region
  integer, parameter :: copy_max_bufsize= 100*100*100    ! a typical i_blocksize*j_blocksize*kb


! aliases to get loops right easy. let IM be the maximum global i dimension, JM max in J.
! when we parallelize loops, 1 is always translated to 1, 
! IM to I_eq_IM, JM to J_eq_JM. example: DO I=2,IM is translated to DO I=I_EQ_2,I_eq_IM

  integer :: i_eq_1, i_eq_2, i_eq_3
  integer :: j_eq_1, j_eq_2, j_eq_3
  integer :: i_eq_im, j_eq_jm
  integer :: im_mi_1, im_mi_2, im_mi_3
  integer :: jm_mi_1, jm_mi_2, jm_mi_3
  integer :: im_p_1, jm_p_1
  
  integer :: i_eq_2_rb, i_eq_3_rb
  integer :: j_eq_2_rb, j_eq_3_rb

  
! block data structures 
  integer :: dry_blocks=0
  integer :: numbli=1,numblj=1                ! number of blocks in i and j direction  
  integer i_blocksize, j_blocksize            ! blocksize excluding ghost zones
  integer full_i_blocksize, full_j_blocksize  ! blocksize including ghost zones  
  integer i1,i2,j1,j2                         ! local start and end indices for blocks including ghost
  integer fci1,fci2,fcj1,fcj2                 ! local start and end indices for blocks including ghost, but excluding the area outside the grid.
  integer :: extended_range=0                 ! the halo is currently valid for this extra width.  
 
  type block_info_type
     
     !! global indices for block excluding ghost:
     !! first i, last i, first j last j...
     integer :: fi=1,li=1,fj=1,lj=1 

     integer :: lim=1, ljm=1 ! the block local value of the last indice in i,j direction
     
     integer :: prn=mpi_proc_null      ! processor number     
     integer :: ibn=0,jbn=0            ! i,j block coordinate

     !! the processor number of the neighbours, mpi_proc_null means no neighb.
     integer :: east=mpi_proc_null, west=mpi_proc_null, north=mpi_proc_null, south=mpi_proc_null
     
     integer :: wet=0      ! the number of wet cells
     
     logical :: sbowner=.false.  ! true, if processor contains a psum subblock

     !! the processors first and last psum subblock for i and j direction
     integer :: sbi1=1,sbi2=1,sbj1=1,sbj2=1
  end type block_info_type

  !! the block info for the processor
  type (block_info_type) :: block
  
  !! information about all blocks
  type (block_info_type), allocatable, dimension(:,:) :: bltab

  !! the block coordinate bi,bj of rank; usage  bi=block_of_rank(1,processor),
  !! bj=block_of_rank(2,processor)
  integer, dimension(:,:), allocatable :: block_of_rank


  integer, allocatable :: gridbox(:,:)
  
  interface globval
     module procedure globval2d, globval3d
  end interface

  interface scatterfield
     module procedure scatterfield4, scatterfield8, scatterfield42d, scatterfield82d
  end interface
  
!! ghost update stuff

! ghost update method:  0 send-recv (default), 1 isend-irecv-waitall, 2 sendrecv  
  integer :: mod_mpi_gum=0
  
  interface tag_ghost
     module procedure pack2d, pack3d
  end interface  
  
! send&recv buffers. allocated size = ghost_max_bufsize
  real, dimension(:), allocatable :: sbe, sbw, sbn, sbs, rbe, rbw, rbn, rbs
  
! pointers into send&recv buffers; east,west, north & south:
  integer :: i_sbe=0 , i_sbw=0, i_sbn=0 , i_sbs=0, i_rbe=0 , i_rbw=0, i_rbn=0 , i_rbs=0
 

! maximum size of send buffers to each direction
  integer :: ghost_max_bufsize=-1
  
! counter used to check for full send buffers.  
  integer send_size
  integer, parameter :: tageast=1234,tagwest=2341,tagnorth=3412, tagsouth=4123
  
! the sides argument:
  integer, parameter :: east=10,west=100,north=1000,south=10000,all_sides=east+west+north+south
  
  
  type tabellpeker2d
     real, pointer, dimension(:,:) :: f => null()
     integer width
     logical sendwest, sendeast, sendsouth, sendnorth
     logical recvwest, recveast, recvsouth, recvnorth
  end type tabellpeker2d
  
  type tabellpeker3d
     real, pointer, dimension(:,:,:) :: f => null()
     integer width,kdim
     logical sendwest, sendeast, sendsouth, sendnorth
     logical recvwest, recveast, recvsouth, recvnorth
  end type tabellpeker3d
  
! maximum number of successive pack's before an update_ghost:
  integer, parameter :: updatelist_size=20
  
  type(tabellpeker2d) :: updatelist2d(updatelist_size)
  type(tabellpeker3d) :: updatelist3d(updatelist_size) 
  integer :: ul_ind2d=0, ul_ind3d=0


contains


  subroutine broadcast_blocks(blocks)
!    
! broadcast a 2d array of a derived type using a character buffer and
! qmpi
!    
    implicit none
    type (block_info_type), dimension(:,:) :: blocks
    character, allocatable :: buffer(:)
    integer bufsize
    if(numproc == 1) return
  
    bufsize = size(transfer(blocks,(/'x'/)))
    allocate(buffer(bufsize))
    if(master)then
        buffer = transfer(blocks, buffer)
    end if
    
    call broadcast(buffer)
    
    if(.not.master)then
        blocks = reshape(transfer(buffer, blocks), (/size(blocks,1),size(blocks,2)/))
    end if
    deallocate(buffer)
  end subroutine broadcast_blocks
  
  
subroutine init_mpilib(im,jm,kb,fglobal,criteria)
!
! set up a 2D processor topology, and initialize all necessary data structures for mpilib to work.
! FGLOBAL(IM,JM) is a global array that defines active cells in the geometry. 
! example: ocean modeling: cells where  FGLOBAL(I,J)<CRITERIA
! are land cells, and are neglected.  a block with only dry cells is removed from the
! process topology. fglobal must be defined on all processors.
!
! MPI's cart create etc. could have been used, but dont know how to "delete" dry blocks from
! such a setup, and the performance benefits on modern systems are not clear, probably none.
!
  implicit none

  integer im,jm
  integer, optional :: kb
  real, optional :: fglobal(:,:), criteria

  integer i,j,imt,jmt,ierr1,ierr2
  real, allocatable :: tmpreal(:,:)
  
! store the dimensions of the global grid
  gim = im
  gjm = jm
  if(present(kb)) then
      gkb = kb
  else
      gkb = 1
  end if

  do i=1,updatelist_size
      updatelist2d(i)%f => null()
      updatelist3d(i)%f => null()
  end do
  
! bltab(0,0) is an inactive block by definition. if we have processors
! that should be "idle", they should point to 0,0 - which is why
! block_of_rank is 0 by default:
  
  allocate(bltab(0:numbli,0:numblj), block_of_rank(2,0:numproc-1))
  block_of_rank = 0
  
! find reasonable block dimensions.
! search for global grid dimensions that fits the number of blocks
  
  imt=gim
  DO
      IF(MOD(imt,numbli)== 0) EXIT
      imt=imt+1
  END DO
  jmt=gjm
  DO
      IF(MOD(jmt,numblj)==0) EXIT
      jmt=jmt+1
  END DO
  
  if(jm == 1) ghost_wj=0
  
  i_blocksize=imt/numbli
  full_i_blocksize=i_blocksize + 2*ghost_wi
  i1=1-ghost_wi
  i2=i_blocksize +ghost_wi
  
  j_blocksize=jmt/numblj
  full_j_blocksize=j_blocksize + 2*ghost_wj
  j1=1-ghost_wj
  j2=j_blocksize+ghost_wj

  if(master) then
      write(*,'(/,2x,a)') "init_mpilib: setting up data structures for MPI" 
      write(*,'(2x,a,i0,a,i0,a,i0,a,i0)') 'numbli,numblj=',numbli,',',numblj,&
           ' i_blocksize,j_blocksize=',i_blocksize,',',j_blocksize
  end if
  
! process fglobal (depth matrix) to find bltab
  
  if(master)then
      call find_block_ranges(fglobal, criteria)
  end if

  call broadcast_blocks(bltab)
  call broadcast(block_of_rank)
  
  call find_block_neighbours()

  if(block_of_rank(1,myid) > 0 .and. block_of_rank(2,myid) > 0) then  
      block = bltab(block_of_rank(1,myid),block_of_rank(2,myid))
  end if
  
  call use_default_loop_bounds()

! set up the mini blocks used to compute the parallel, processor
! independent summing routine.
  
  call find_psum_block_ranges()
  
  if(block_of_rank(1,myid) > 0 .and. block_of_rank(2,myid) > 0) then  
      block = bltab(block_of_rank(1,myid),block_of_rank(2,myid))
  end if
!
! allocate space for the exchange buffers
! BOM guess: never more than 6 3d updates so totally 7 should be suff.
!
  if(jm > 1)then
     ghost_max_bufsize = 7*max(full_i_blocksize,full_j_blocksize)*gkb*ghost_wi
  else
     ghost_max_bufsize = 7*gkb*ghost_wi
  end if

  ierr1=0
  ierr2=0
  
  allocate(sbe(ghost_max_bufsize), sbw(ghost_max_bufsize), &
       rbe(ghost_max_bufsize), rbw(ghost_max_bufsize), stat=ierr1)
  
  if(jm > 1) then
      allocate(sbn(ghost_max_bufsize), sbs(ghost_max_bufsize), &
           rbn(ghost_max_bufsize), rbs(ghost_max_bufsize), stat=ierr2)
  end if

! this is 100% equivalent to fsm, but integer

  allocate(gridbox(i1:i2,j1:j2), tmpreal(i1:i2,j1:j2))

  if(present(fglobal)) then
      call scatterfield(fglobal, tmpreal)  
      where(tmpreal > 0.0001) 
          gridbox = 1
      elsewhere
          gridbox = 0
      end where
      deallocate(tmpreal)
  else
      gridbox = 1
  end if
  
  if(ierr1 .NE. 0 .or. ierr2.ne.0)then
      write(*,*) 'ierr1,ierr2=',ierr1,ierr2
      write(*,*) 'init_mpilib: failed to allocate send buffers of size',ghost_max_bufsize,'stopping'
      stop
  end if


  if(viewblocks) then
      if(master) write(*,'(/,2x,a)')'MPI block summary:'
      
      call barrier()
      
      write(*,"(2x, i4.4, ' ibn,jbn:',i3,',',i3,' fi,li:',i5,',',i5,' fj,lj:',i4,',',i4 &
           &,' e:',i4,' w:',i4,' n:',i4,' s:',i4 ,' wetpts:',i8,' on? ',l)") &
           &block%prn, block%ibn, block%jbn, block%fi,block%li, block%fj,block%lj  &
           &,block%east, block%west, block%north, block%south,block%wet, block%prn.ne.mpi_proc_null

      call barrier()        
  end if
  
end subroutine init_mpilib


subroutine deallocate_mod_mpi()
  implicit none
  deallocate(sbe, sbw, rbe, rbw, gridbox)  
  if(gjm > 1) deallocate(sbn, sbs, rbn, rbs)  
  deallocate(bltab, block_of_rank)
end subroutine deallocate_mod_mpi


subroutine use_default_loop_bounds()
!
! set default loop bounds.
!  
  implicit none

  i_eq_1  = 1
  i_eq_2  = 1
  i_eq_3  = 1           
  i_eq_im = block%lim
  im_mi_1 = block%lim
  im_mi_2 = block%lim   
  im_mi_3 = block%lim

! in periodic mode, we pretend the grid edge is just another mpi block
! end
  
  if(.not. periodic_ew) then   
      if(block%ibn == 1)then
          i_eq_2 = 2
          i_eq_3 = 3
      end if
      
      if(block%li >= gim)then
          im_mi_1 = block%lim - 1 
          im_mi_2 = block%lim - 2 
          im_mi_3 = block%lim - 3
      end if
  end if
  
  j_eq_1  = 1            
  j_eq_2  = 1
  j_eq_3  = 1           
  j_eq_jm = block%ljm   
  jm_mi_1 = block%ljm
  jm_mi_2 = block%ljm   
  jm_mi_3 = block%ljm
  
  if(block%jbn == 1)then
      j_eq_2 = 2
      j_eq_3 = 3
  end if
  if(block%lj >= gjm)then
      jm_mi_1 = block%ljm - 1
      jm_mi_2 = block%ljm - 2 
      jm_mi_3 = block%ljm - 3
  end if

  if(gjm == 1)then
      j_eq_1  = 1
      j_eq_2  = 1
      j_eq_3  = 1
      j_eq_jm = 1
      jm_mi_1 = 1
      jm_mi_2 = 1
      jm_mi_3 = 1
  end if

! sometimes it can be useful to have a loop range that cover the full global grid, and all ghost zones
!    do j=fcj1,fcj2
!        do i=fci1,fci2
  fci1 = i1
  fci2 = i2
  fcj1 = j1
  fcj2 = j2
  if(block%ibn == 1) fci1 = 1
  if(block%ibn == numbli) fci2 = i_eq_im
  if(block%jbn == 1) fcj1 = 1
  if(block%jbn == numblj) fcj2 = j_eq_jm

end subroutine use_default_loop_bounds



subroutine find_block_ranges(fglobal,criteria)
!
! on master, count up, and find which blocks are dry (omittable) and
! which blocks to to compute on.
!
  implicit none
  real, optional :: fglobal(:,:), criteria
  integer i,j,gi1,gi2,gj1,gj2,li,lj
  integer target_process
  integer :: wetpts=0
  real crit

  crit=0.0
  if(present(criteria)) crit = criteria
  target_process=0
  dry_blocks=0
  
  do j=1,numblj
      do i=1,numbli
          !! global range for a block
          gi1 = (i-1)*i_blocksize + 1
          gi2 = min(gim, i*i_blocksize)
          gj1 = (j-1)*j_blocksize + 1
          gj2 = min(gjm, j*j_blocksize)
          
          bltab(i,j)%fi = gi1
          bltab(i,j)%li = gi2
          bltab(i,j)%fj = gj1
          bltab(i,j)%lj = gj2
          bltab(i,j)%ibn = i
          bltab(i,j)%jbn = j

          if(present(fglobal)) then
              wetpts = 0
              do lj=gj1,gj2
                  do li=gi1,gi2
                      if(fglobal(li,lj) > crit) wetpts = wetpts + 1
                  end do
              end do
              bltab(i,j)%wet = wetpts
          else
              bltab(i,j)%wet = (gi2-gi1+1)*(gj2-gj1+1)
          end if
          
          
          if(bltab(i,j)%wet == 0) then
              dry_blocks = dry_blocks + 1
          end if

          if(bltab(i,j)%wet == 0 .and. OMIT_DRY_BLOCKS) then
              bltab(i,j)%prn = MPI_PROC_NULL
          else
              if(target_process > numproc-1)then
                  call message("more blocks than processors. adjust NUMBLI,NUMBLJ and/or NUMPROC. &
                       &current ibn,jbn,dry_blocks", 2, ival=(/i,j,dry_blocks/))
              end if
              bltab(i,j)%prn = target_process
              
              block_of_rank(:,target_process) = (/i,j/)
              
              bltab(i,j)%lim = min(i_blocksize, gim-(i-1)*i_blocksize)
              if(bltab(i,j)%lim < 1) bltab(i,j)%lim = i_blocksize
              
              bltab(i,j)%ljm = min(j_blocksize, gjm-(j-1)*j_blocksize)
              if(bltab(i,j)%ljm < 1) bltab(i,j)%ljm = j_blocksize
              
              target_process = target_process + 1
          end if
      end do
  end do
      
  if(dry_blocks > 0 .and. .not. omit_dry_blocks)then
      write(*,'(/,a)') 'PS'
      write(*,'(i0,a)') dry_blocks, ' dry blocks was found. you can restart the model with '
      write(*,'(a,i0,a,i0,",",i0)') 'OMIT_DRY_BLOCKS=.true. (edit setupfile2.dat), ',&
           numproc-dry_blocks,' processors, and NUMBLI,NUMBLJ=',numbli,numblj
      write(*,*)
  end if
  
  if(.not. omit_dry_blocks .and. target_process < numproc) then
      write(*,*)'only found ',target_process,' wet blocks, numprocs=',numproc
      write(*,*)'you should restart with ',target_process,' processors'
      call abort()
  end if
end subroutine find_block_ranges



subroutine find_block_neighbours()
!  
! get to know all the neighbours. give dry blocks process number MPI_PROC_NULL
!
  implicit none
  integer i,j

  do i=1, numbli
      do j=1, numblj
          if(bltab(i,j)%prn .ne. mpi_proc_null) then
              if(i<numbli) then
                  if(bltab(i+1,j)%prn .ne. mpi_proc_null) then
                      bltab(i,j)%east=bltab(i+1,j)%prn
                  else
                      bltab(i,j)%east=mpi_proc_null
                  end if
              else
                  bltab(i,j)%east=mpi_proc_null
              end if

              if(i>1) then
                  if(bltab(i-1,j)%prn .ne. mpi_proc_null) then
                      bltab(i,j)%west=bltab(i-1,j)%prn
                  else
                      bltab(i,j)%west=mpi_proc_null
                  end if
              else
                  bltab(i,j)%west=mpi_proc_null
              end if

              if(j<numblj) then
                  if(bltab(i,j+1)%prn .ne. mpi_proc_null) then
                      bltab(i,j)%north=bltab(i,j+1)%prn
                  else
                      bltab(i,j)%north=mpi_proc_null
                  end if
              else
                  bltab(i,j)%north=mpi_proc_null
              end if

              if(j>1) then
                  if(bltab(i,j-1)%prn .ne. mpi_proc_null) then
                      bltab(i,j)%south=bltab(i,j-1)%prn
                  else
                      bltab(i,j)%south=mpi_proc_null
                  end if
              else
                  bltab(i,j)%south=mpi_proc_null
              end if
          end if
      end do
  end do

  if(periodic_ew) then
      i=1
      do j=1, numblj
          if(bltab(i,j)%prn .ne. mpi_proc_null) then
              bltab(i,j)%west = bltab(numbli,j)%prn
          end if
      end do
      i=numbli
      do j=1, numblj
          if(bltab(i,j)%prn .ne. mpi_proc_null) then
              bltab(i,j)%east = bltab(1,j)%prn
          end if
      end do
  end if
  
end subroutine find_block_neighbours


subroutine get_optimal_block_decomp(ib,jb,im,jm)
!  
! for a given horisontal grid size im,jm, search for a block
! decomposition that gives as quadratic blocks as possible
! by minimizing  f(numbli,numblj) = abs(im/numbli-jm/numblj)
!  
  implicit none
  integer ib,jb,im,jm,i,j
  real diff, mindiff
  
  if(jm == 1) then
      numbli = numproc
      numblj = 1
  else
      mindiff = 1000000
      do j=1,numproc
          do i=1,numproc
              if(i*j == numproc) then
                  diff = abs(real(im)/real(i) - real(jm)/real(j))
                  
                  if(diff < mindiff .and. &
                       .not. (periodic_ew .and. mod(im,i).ne.0)) then !!! if periodic: mod(im,ib) must be 0.
                      ib = i
                      jb = j
                      mindiff = diff
                  end if
              end if
          end do
      end do
  end if
  if(master) then
      write(*,'(/,2x,a,i0,"x",i0)') 'Will choose a recommended block decomposition NUMBLIxNUMBLJ=',ib,jb
      write(*,'(2x,a)') "(if desired, override by setting NUMBLI and NUMBLJ directly)"
  end if
end subroutine get_optimal_block_decomp


subroutine find_psum_block_ranges
!  
! find the range of psum blocks that belong to each processor
!  
  implicit none
  integer bi,bj,ii,jj,ci1,ci2,cj1,cj2,pbi,pbj
  logical, allocatable :: sbfree(:,:)

!! first, the size of a psum block

  mblszi=1+2*default_ghost_wi
  
!! the number of psum blocks  
  numpbi=gim/mblszi
  if(mod(gim,mblszi).ne.0) numpbi=numpbi+1
  
  mblszj=1+2*default_ghost_wj
  numpbj=gjm/mblszj
  if(mod(gjm,mblszj).ne.0) numpbj=numpbj+1
  
  bltab(:,:)%sbi1=numpbi
  bltab(:,:)%sbi2=1
  bltab(:,:)%sbj1=numpbj
  bltab(:,:)%sbj2=1

!! for each mpi block, loop over all psum blocks  
  do bi=1,numbli
      do ii=1,numpbi

          !! global start and end indices for each psum block
          ci1= (ii-1)*mblszi + 1
          ci2= min(ii*mblszi,gim)

          !! if the current psum block is on a mpi block
          if(ci1>=bltab(bi,1)%fi-ghost_wi .and. ci2<=bltab(bi,1)%li+ghost_wi) then

              !! if current psum block pos is before the current first psum block on 
              !! the mpi block, make it the new "current first"
              
              if(ii<bltab(bi,1)%sbi1) bltab(bi,:)%sbi1=ii
              if(ii>bltab(bi,1)%sbi2) bltab(bi,:)%sbi2=ii
          end if
      end do
  end do
  do bj=1,numblj
      do jj=1,numpbj
          cj1= (jj-1)*mblszj + 1
          cj2= min(jj*mblszj,gjm)
          if(cj1>=bltab(1,bj)%fj-ghost_wj .and. cj2<=bltab(1,bj)%lj+ghost_wj) then
              if(jj<bltab(1,bj)%sbj1) bltab(:,bj)%sbj1=jj
              if(jj>bltab(1,bj)%sbj2) bltab(:,bj)%sbj2=jj

              if(gjm==1) then
                  bltab(:,bj)%sbj1=1
                  bltab(:,bj)%sbj2=1
              end if              
          end if
      end do
  end do



! fix for the omit_dry_blocks problem. for most decomps where im is
! not divisible by 5, psum blocks may end up on omitted blocks. we
! increase the psum block range for neighbours of such omitted blocks
! so that only wet blocks have psum blocks.
  
  do bj=1,numblj
      do bi=2,numbli
          if(bltab(bi-1,bj)%prn == mpi_proc_null) then
              
              ! the start index of the first psum block on the current processor
              pbi = (bltab(bi,bj)%sbi1-1)*mblszi + 1
              
              if(pbi > bltab(bi,bj)%fi-ghost_wi) then
                  ! this means the left processor sits on a psum block,
                  ! but is dry. we fix this by including the block on
                  ! the current processor, and add checks to avoid out
                  ! of range.
                  bltab(bi,bj)%sbi1 = bltab(bi,bj)%sbi1 - 1
                  bltab(bi-1,bj)%sbi2 = bltab(bi-1,bj)%sbi2 - 1
              end if
          end if
      end do
  end do
!same logic for j diretion
  do bj=2,numblj
      do bi=1,numbli
          if(bltab(bi,bj-1)%prn == mpi_proc_null) then              
              pbj = (bltab(bi,bj)%sbj1-1)*mblszj + 1
              if(pbj > bltab(bi,bj)%fj-ghost_wj) then
                  bltab(bi,bj)%sbj1 = bltab(bi,bj)%sbj1 - 1
                  bltab(bi,bj-1)%sbj2 = bltab(bi,bj-1)%sbj2 - 1
              end if
          end if
      end do
  end do
! then the east block
  do bj=1,numblj
      do bi=1,numbli-1
          if(bltab(bi+1,bj)%prn == mpi_proc_null) then
              pbi = bltab(bi,bj)%sbi2*mblszi              
              if(pbi < bltab(bi,bj)%li+ghost_wi) then
                  bltab(bi,bj)%sbi2 = bltab(bi,bj)%sbi2 + 1
                  bltab(bi+1,bj)%sbi1 = bltab(bi+1,bj)%sbi1 + 1
              end if
          end if
      end do
  end do
! and the north
  do bj=1,numblj-1
      do bi=1,numbli
          if(bltab(bi,bj+1)%prn == mpi_proc_null) then
              pbj = bltab(bi,bj)%sbj2*mblszj             
              if(pbj < bltab(bi,bj)%lj+ghost_wj) then
                  bltab(bi,bj)%sbj2 = bltab(bi,bj)%sbj2 + 1
                  bltab(bi,bj+1)%sbj1 = bltab(bi,bj+1)%sbj1 + 1
              end if
          end if
      end do
  end do
  

  
! find which processors that are not sitting on a block (sbowner=.false.)

  allocate(sbfree(numpbi,numpbj))
  sbfree(:,:)=.true.
  bltab(:,:)%sbowner=.false.

  do bj=1,numblj
      do bi=1,numbli
          do jj=1,numpbj
              do ii=1,numpbi
                  ci1= (ii-1)*mblszi + 1 ; ci2= min(ii*mblszi,gim)
                  cj1= (jj-1)*mblszj + 1 ; cj2= min(jj*mblszj,gjm)

                  if( ci1 >= bltab(bi,bj)%fi - ghost_wi .and. &
                      ci2 <= bltab(bi,bj)%li + ghost_wi .and. &
                      cj1 >= bltab(bi,bj)%fj - ghost_wj .and. &
                      cj2 <= bltab(bi,bj)%lj + ghost_wj .and. &
                      bltab(bi,bj)%prn .ne. mpi_proc_null .and. &
                      sbfree(ii,jj) ) then

                      bltab(bi,bj)%sbowner=.true.
                      sbfree(ii,jj)=.false.
                  end if
              end do
          end do
      end do
  end do

  deallocate(sbfree)

  if(master) then
      do bj=1,numblj
          do bi=1,numbli
              if(bltab(bi,bj)%prn == mpi_proc_null .and. bltab(bi,bj)%sbowner) &
                   call message("find_psum_block_ranges bug: psum block on omitted mpi block",2)
          end do
      end do
  end if
end subroutine find_psum_block_ranges


subroutine reallocate_sendbuffers
  integer allocate_success

! increase buffsize with 100%? 50?
  ghost_max_bufsize = ghost_max_bufsize + &
       2*max(full_i_blocksize,full_j_blocksize)*gkb*ghost_wi

  write(*,*) 'MPILIB: increasing ghost sendbuffer sizes to ',ghost_max_bufsize

  deallocate(sbe, sbw, rbe, rbw)
  if(gjm > 1) deallocate(sbn, sbs, rbn, rbs)
  
  allocate(sbe(ghost_max_bufsize), sbw(ghost_max_bufsize), rbe(ghost_max_bufsize), rbw(ghost_max_bufsize), &
       stat=allocate_success)
  if(gjm > 1) then
      allocate(sbn(ghost_max_bufsize), sbs(ghost_max_bufsize), rbn(ghost_max_bufsize), rbs(ghost_max_bufsize))
  end if
  
  if(allocate_success /= 0) then
      write(*,*) 'fatal error. could not allocate MPI ghost exchange buffers. size=',&
           ghost_max_bufsize
      stop
  end if
end subroutine reallocate_sendbuffers


integer function gi2block(gi)
  !! return the block coordinate bi of global point gi
  integer gi
  gi2block = (gi - 1)/i_blocksize + 1
end function gi2block


integer function gj2block(gj)
  !! return the block coordinate bj of global point gj
  integer gj
  gj2block = (gj - 1)/j_blocksize + 1
end function gj2block


subroutine blocks_covering_region(reg, bli1,bli2,blj1,blj2)
!
! return block the indices of the corners of region reg
!   (i-1)/i_blocksize+1    gives the block number for indice i
!
  implicit none
  integer reg(4), bli1,bli2,blj1,blj2

  bli1 = (reg(1) - 1)/i_blocksize + 1
  bli2 = (reg(2) - 1)/i_blocksize + 1
  
  if(gjm == 1) return
  
  blj1 = (reg(3) - 1)/j_blocksize + 1
  blj2 = (reg(4) - 1)/j_blocksize + 1
end subroutine blocks_covering_region


subroutine ronblock(r, inblock)
! return inblock=.true. if the horisontal global grid range r is on block
  implicit none
  real r(4),br(4)
  logical inblock
    
! block range in global grid coordinates
  br = (/block%fi-1,block%li,block%fj-1,block%lj/)
  inblock=.true.
    
  if( (r(1)<= br(1) .and. r(2)>=br(2))                    .or. &               ! range fully overlaps block
       (r(1)<= br(1) .and.(r(2)>=br(1)  .and. r(2)<=br(2))).or. &               ! range overlaps left edge of block
       ((r(1)>=br(1) .and. r(1)<=br(2)) .and. r(2)>=br(1)) .or. &               ! range overlaps right edge of block
       ((r(1)>=br(1) .and. r(1)<=br(2)) .and. (r(2)>=br(1) .and. r(2)<=br(2))))then ! range is inside block
  else
      inblock=.false.
      return
  end if
  if( (r(3)<= br(3) .and. r(4)>=br(4))                    .or. &              ! range fully overlaps block
       (r(3)<= br(3) .and.(r(4)>=br(3)  .and. r(4)<=br(4))).or. &              ! range overlaps left edge of block
       ((r(3)>=br(3) .and. r(3)<=br(4)) .and. r(4)>=br(3)) .or. &              ! range overlaps right edge of block
       ((r(3)>=br(3) .and. r(3)<=br(4)) .and. (r(4)>=br(3) .and. r(4)<=br(4))))then! range is inside block
  else
      inblock=.false.
      return
  end if
end subroutine ronblock


subroutine globalcoo2local(gx,gy,inblock)
!
! given the global coos gx,gy return inblock=.true. if in block,
! (excluding ghost cells) and also convert gi,gj into local coordinates.
!
  implicit none
  logical inblock
  real gx,gy

  inblock=.true.

  if(numproc == 1) return
  
  if(gx>=block%fi-1.0 .and. gx<=real(block%li)) then
     gx= gx - block%fi + 1
  else
     inblock=.false.
     return
  end if

  if(gjm == 1) return
  
  if(gy>=block%fj-1.0 .and. gy<=real(block%lj)) then
     gy= gy - block%fj + 1
  else
     inblock=.false.
     return
  end if
end subroutine globalcoo2local


subroutine globalind2local(gi,gj,inblock,ghost)
!
! given the global coos gi,gj return inblock=.true. if in block,
! (including ghost cells if ghost=.true.,the default) and also convert
! gi,gj into local coordinates. a hit in the "real" part of one block,
! will also lead to a hit in the relevant ghost zone(s) of its
! neighbours.
!
  implicit none
  logical inblock
  integer gi,gj,di,dj
  logical, optional :: ghost

  di=ghost_wi
  dj=ghost_wj
  if(present(ghost)) then
      if(.not.ghost)then
          di=0
          dj=0
      end if
  end if
  
  inblock=.true.

  if(numproc == 1) return
  
  if(gi>=block%fi-di .and. gi<=block%li + di) then
     gi= gi - block%fi + 1
  else
     inblock=.false.
     gi=1 ; gj=1 
     return
  end if

  if(gjm == 1) return
  
  if(gj>=block%fj-dj .and. gj<=block%lj + dj) then
     gj= gj - block%fj + 1
  else
     inblock=.false.
     gi=1 ; gj=1 
     return
  end if
end subroutine globalind2local



subroutine globalrect2local(gi1,gi2,gj1,gj2, li1,li2,lj1,lj2, inblock, ghost)
!
! given four global indices defining a rectangle
!
! gi1,gj1 southwest corner,
! gi2,gj2 northeast corner.
!
! return the local coordinates li1,... of the portion that lies
! within the block, inblock=false if none of the coordinates
! are inside the block (the returned coordinates will not
! include ghost cells, unless optional arg ghost is true.)
!
! Helge Avlesen, para//ab <avle@ii.uib.no>
!
  implicit none
  integer, intent(in) :: gi1,gi2,gj1,gj2
  integer, intent(out) :: li1,li2,lj1,lj2
  logical, intent(out) :: inblock
  integer di,dj
  logical, optional :: ghost

  if(gi1>gi2 .or. gj1>gj2) then
      if(master) write(*,*) "gi1,gi2,gj1,gj2=",gi1,gi2,gj1,gj2      
      call abort('globalrect2local: error gi1>gi2 or gj1>gj2')
  end if
  inblock=.true.

! (ghost is false by default)
  di=0
  dj=0
  if(present(ghost)) then
      if(ghost)then
          di = ghost_wi
          dj = ghost_wj
      end if
  end if

  if(gi1<= block%fi-di .and. gi2>=block%li+di) then
     li1=1-di
     li2=block%lim+di
  else if(gi1<= block%fi-di .and. (gi2>=block%fi-di .and. gi2<=block%li+di)) then
     li1=1-di
     li2=gi2-block%fi+1
  else if((gi1>=block%fi-di .and. gi1<=block%li+di) .and. gi2>=block%li+di) then
     li1=gi1-block%fi+1
     li2=block%lim+di
  else if((gi1>=block%fi-di .and. gi1<=block%li+di) .and. &
       (gi2>=block%fi-di .and. gi2<=block%li+di)) then
     li1=gi1-block%fi+1
     li2=gi2-block%fi+1
  else
     inblock=.false.
     return
  end if

  if(gj1<= block%fj-dj .and. gj2>=block%lj+dj) then
     lj1=1-dj
     lj2=block%ljm+dj
  else if(gj1<= block%fj-dj .and. (gj2>=block%fj-dj .and. gj2<=block%lj+dj)) then
     lj1=1-dj
     lj2=gj2-block%fj+1
  else if((gj1>=block%fj-dj .and. gj1<=block%lj+dj) .and. gj2>=block%lj+dj) then
     lj1=gj1-block%fj+1
     lj2=block%ljm+dj

  else if((gj1>=block%fj-dj .and. gj1<=block%lj+dj) .and. &
       (gj2>=block%fj-dj .and. gj2<=block%lj+dj)) then

     lj1=gj1-block%fj+1
     lj2=gj2-block%fj+1
  else
     inblock=.false.
     return
  end if
end subroutine globalrect2local


subroutine globalrange2local(bi,bj,range,inblock)
!
! given four global indices defining a rectangle/range
!
! gi1,gj1 southwest corner, range(1),range(3)
! gi2,gj2 northeast corner. range(2),range(4)
!
! return the local coordinates of the portion on block bi,bj that lies
! within the block, inblock=false if none of the coordinates
! are inside the block (the returned coordinates will not
! include ghost cells)
!
! Helge Avlesen, para//ab <avle@ii.uib.no>
!
  implicit none
  integer gi1,gi2,gj1,gj2, li1,li2,lj1,lj2, bi,bj
  integer range(4)
  logical inblock

  gi1=range(1) ;  gi2=range(2)
  gj1=range(3) ;  gj2=range(4)

  if(gi1>gi2 .or. gj1>gj2) call abort('global2local: error gi1>gi2 or gj1>gj2')
  inblock=.true.

  if(gi1<= bltab(bi,bj)%fi .and. gi2>=bltab(bi,bj)%li) then
     li1=1
     li2=bltab(bi,bj)%li-bltab(bi,bj)%fi+1

  else if(gi1<= bltab(bi,bj)%fi .and. (gi2>=bltab(bi,bj)%fi .and. gi2<=bltab(bi,bj)%li)) then
     li1=1
     li2=gi2-bltab(bi,bj)%fi+1

  else if((gi1>=bltab(bi,bj)%fi .and. gi1<=bltab(bi,bj)%li) .and. gi2>=bltab(bi,bj)%li) then
     li1=gi1-bltab(bi,bj)%fi+1
     li2=bltab(bi,bj)%li-bltab(bi,bj)%fi+1

  else if((gi1>=bltab(bi,bj)%fi .and. gi1<=bltab(bi,bj)%li) .and. &
           (gi2>=bltab(bi,bj)%fi .and. gi2<=bltab(bi,bj)%li)) then

     li1=gi1-bltab(bi,bj)%fi+1
     li2=gi2-bltab(bi,bj)%fi+1
  else
     inblock=.false.
     return
  end if

  if(gj1<= bltab(bi,bj)%fj .and. gj2>=bltab(bi,bj)%lj) then
     lj1=1
     lj2=bltab(bi,bj)%lj-bltab(bi,bj)%fj+1
  else if(gj1<= bltab(bi,bj)%fj .and. (gj2>=bltab(bi,bj)%fj .and. gj2<=bltab(bi,bj)%lj)) then
     lj1=1
     lj2=gj2-bltab(bi,bj)%fj+1

  else if((gj1>=bltab(bi,bj)%fj .and. gj1<=bltab(bi,bj)%lj) .and. gj2>=bltab(bi,bj)%lj) then
     lj1=gj1-bltab(bi,bj)%fj+1
     lj2=bltab(bi,bj)%lj-bltab(bi,bj)%fj+1

  else if((gj1>=bltab(bi,bj)%fj .and. gj1<=bltab(bi,bj)%lj) .and. &
       (gj2>=bltab(bi,bj)%fj .and. gj2<=bltab(bi,bj)%lj)) then

     lj1=gj1-bltab(bi,bj)%fj+1
     lj2=gj2-bltab(bi,bj)%fj+1
  else
     inblock=.false.
     return
  end if

  range(1)=li1  ;  range(2)=li2
  range(3)=lj1  ;  range(4)=lj2
end subroutine globalrange2local



subroutine gatherfield(flocal,fglobal,im,jm,kb,undef)
!
! from the given local field, assemble the global field at the master process
!
  implicit none
  integer,intent(in):: im,jm,kb
  real, intent(in) :: flocal(i1:i2,j1:j2,kb)
  real,intent(out) :: fglobal(im,jm,kb)
  real, optional :: undef
  real given_undef
  integer i,j,k,b_i,b_j, gi1, gi2, gj1, gj2, r_ind, blocksize_3d
  real, dimension(numproc*i_blocksize*j_blocksize*kb) :: recvbuff

  if(numproc==1)then
      fglobal(1:im,1:jm,1:kb) = flocal(1:im,1:jm,1:kb)
      return
  end if

  given_undef = 0.0
  if(present(undef)) given_undef = undef
  
  blocksize_3d = i_blocksize*j_blocksize*kb
  
  call mpi_gather(flocal(1:i_blocksize,1:j_blocksize,1:kb), blocksize_3d, &
       mpi_float_type,  recvbuff, blocksize_3d, mpi_float_type, 0, qmpi_comm_world, ierr)

  if(master)then
      do b_j=1,numblj
          do b_i=1,numbli
              gi1=bltab(b_i,b_j)%fi  ;   gi2=gi1 + i_blocksize -1
              gj1=bltab(b_i,b_j)%fj  ;   gj2=gj1 + j_blocksize -1

              if(bltab(b_i,b_j)%prn .ne. mpi_proc_null) then
                  r_ind= bltab(b_i,b_j)%prn * blocksize_3d + 1

                  do k=1,kb
                      do j=gj1,gj2
                          do i=gi1,gi2
                              if(i <= bltab(b_i,b_j)%li .and. &
                                 j <= bltab(b_i,b_j)%lj) then
                                     fglobal(i,j,k) = recvbuff(r_ind)
                              end if
                              
                              r_ind = r_ind + 1
                          end do
                      end do
                  end do                  
              else
                  gi2 = min(gi2, gim)
                  gj2 = min(gj2, gjm)
                  fglobal(gi1:gi2,gj1:gj2,1:kb) = given_undef
              end if
          end do
      end do
  end if
end subroutine gatherfield




subroutine scatterfield4(fglobal, flocal)
!
! distribute the field fglobal out to the cpu's given by structures in mod_mpi
! and return flocal, also including ghost zones, on each process.
!
  implicit none
  integer, parameter :: localkind=4

  include "scatterfield.f90"
  
! this normally works, if it doesnt use below lines.  
  call mpi_scatter(sendbuf, blocksize_3d, MPI_REAL4, flocal, blocksize_3d, &
       MPI_REAL4, 0, qmpi_comm_world, ierr)

!  call mpi_scatter(sendbuf, blocksize_3d, mpi_REAL4, recvbuf, blocksize_3d, &
!       mpi_real4, 0, qmpi_comm_world, ierr)
  
!  flocal = reshape(recvbuf, (/full_i_blocksize,full_j_blocksize,gkb/))
  deallocate(sendbuf)
end subroutine scatterfield4

subroutine scatterfield8(fglobal, flocal)
!
! distribute the field fglobal out to the cpu's given by structures in mod_mpi
! and return flocal, also including ghost zones, on each process.
!
  implicit none
  integer, parameter :: localkind=8

  include "scatterfield.f90"
  
! this normally works, if it doesnt use below lines.  
  call mpi_scatter(sendbuf, blocksize_3d, MPI_REAL8, flocal, blocksize_3d, &
       mpi_real8, 0, qmpi_comm_world, ierr)

!  call mpi_scatter(sendbuf, blocksize_3d, mpi_REAL8, recvbuf, blocksize_3d, &
!       mpi_real8, 0, qmpi_comm_world, ierr)
  
!  flocal = reshape(recvbuf, (/full_i_blocksize,full_j_blocksize,kb/))
  deallocate(sendbuf)
end subroutine scatterfield8

subroutine scatterfield42d(fglobal, flocal)
  implicit none
  integer, parameter :: localkind=4

  include "scatterfield2d.f90"

  call mpi_scatter(sendbuf, blocksize_3d, MPI_REAL4, flocal, blocksize_3d, &
       mpi_real4, 0, qmpi_comm_world, ierr)
end subroutine scatterfield42d

subroutine scatterfield82d(fglobal, flocal)
  implicit none
  integer, parameter :: localkind=8

  include "scatterfield2d.f90"

  call mpi_scatter(sendbuf, blocksize_3d, MPI_REAL8, flocal, blocksize_3d, &
       mpi_real8, 0, qmpi_comm_world, ierr)
end subroutine scatterfield82d



! comment:
! The following 2 routines rely on the feature of some compilers that you may
! point to arrays that do not have the target attribute, as long as the dummy array in a
! subroutine carries the attribute.
! (this is standard conforming but dependent on implementation. works on Intel,SGI,SUN,IBM)
! (it will be 100% standards conforming if arrays pointed to by packXX get the target attr.)
!
! � Helge Avlesen <avle@ii.uib.no>, para//ab
!


subroutine pack2d(field, sides, width)
!
! add a pointer to "field" to the "updatelist2d". for each field also add info
! on whether to send/recieve over a side or not.
! 
!    sides:   0 none  (makes no sense)
!            10 east
!           100 west
!          1000 north
!         10000 south
!         11110 all_sides
! example:
!  
!   call tag_ghost(eta, width=1)
!   call tag_ghost(u, east+west)
!   call update_ghost
!  
  implicit none
  real, pointer, dimension(:,:) :: field
  integer, optional, intent(in) :: sides, width
  integer given_sides, given_width, lsides
  logical do_north, do_south, do_west, do_east

  if(block%prn == mpi_proc_null) return  
  if(numproc==1 .and. .not.periodic_ew) return
  
  commtime1=rdtsc()

  given_width=ghost_wi
  if(present(width)) given_width = width

  given_sides=all_sides
  if(present(sides)) given_sides = sides
  
  if(given_sides>all_sides .or. given_sides<10) then
      write(*,*)'pack2d error: sides=',given_sides
      return
  end if
  
  if(given_width>ghost_wi) then
      write(*,*)'pack2d error: width>ghost_wi'
      stop
  end if
  
  if(ul_ind2d>updatelist_size) then
      write(*,*)block%prn,' pack2d: updatelist2d full, increase updatelist_size'
      write(*,*)'updatelist:',ul_ind2d, updatelist_size
      stop
  end if

  if(qmpi_debug > 2) then
      if(.not.associated(field)) call message("tag_ghost 2d field not assiciated",2)
  end if
  
  ul_ind2d=ul_ind2d + 1
  updatelist2d(ul_ind2d)%f => field
!  updatelist(ul_ind)%kdim=kb
  updatelist2d(ul_ind2d)%width=given_width

  lsides=given_sides
  
  do_south=.false.    
  if(lsides/south>=1)then
      if(gjm>1) do_south=.true.
      lsides=lsides-south
  end if
  do_north=.false.
  if( lsides/north>=1)then
      if(gjm>1) do_north=.true.
      lsides=lsides-north
  end if
  do_west=.false.
  if( lsides/west>=1)then
      do_west=.true.
      lsides=lsides-west
  end if
  do_east=.false.
  if( lsides==east)then
      do_east=.true.
  end if

  if(gjm > 1)then
     send_size=send_size + given_width*max(full_i_blocksize,full_j_blocksize)
  else
     send_size=send_size + given_width
  end if

  if(send_size > ghost_max_bufsize) then
      call reallocate_sendbuffers
  end if
  
  updatelist2d(ul_ind2d)%sendwest=  do_east
  updatelist2d(ul_ind2d)%sendeast=  do_west
  updatelist2d(ul_ind2d)%recvwest=  do_west .and. block%west .ne. mpi_proc_null
  updatelist2d(ul_ind2d)%recveast=  do_east .and. block%east .ne. mpi_proc_null
  updatelist2d(ul_ind2d)%sendsouth= do_north
  updatelist2d(ul_ind2d)%sendnorth= do_south
  updatelist2d(ul_ind2d)%recvsouth= do_south .and. block%south .ne. mpi_proc_null
  updatelist2d(ul_ind2d)%recvnorth= do_north .and. block%north .ne. mpi_proc_null
  commtime2=rdtsc()
  totcopytime=totcopytime + (commtime2-commtime1)
end subroutine pack2d


subroutine pack3d(field, sides, width)
!
! add a pointer to "field" to the "updatelist2d". for each field also add info
! on whether to send/recieve over a side or not.
!
  implicit none

  real, pointer, dimension(:,:,:) :: field  
  integer, intent(in), optional :: sides, width
  integer lsides, given_width, given_sides
  logical do_north, do_south, do_west, do_east

  if(block%prn == mpi_proc_null) return  
  if(numproc==1 .and. .not.periodic_ew) return
  
  commtime1=rdtsc()
  
  given_width=ghost_wi
  if(present(width))given_width=width

  given_sides=all_sides
  if(present(sides)) given_sides=sides
  
  if(given_sides>all_sides .or. given_sides<10) then
      write(*,*)'pack2d error: sides=',given_sides
      return
  end if  
  if(given_width>ghost_wi) then
      write(*,*)'pack2d error: width>ghost_wi'
      stop
  end if  
  if(ul_ind3d>updatelist_size) then
      write(*,*)block%prn,' pack3d: updatelist3d full, increase updatelist_size!'
      stop
  end if

  ul_ind3d=ul_ind3d + 1

  if(qmpi_debug > 2) then
      if(.not.associated(field)) call message("tag_ghost 3d field not assiciated",2)
  end if
  
  updatelist3d(ul_ind3d)%f => field
  updatelist3d(ul_ind3d)%kdim = size(field,3)
  updatelist3d(ul_ind3d)%width = given_width

  lsides=given_sides

  do_south=.false.    
  if(lsides/south>=1)then
      if(gjm>1) do_south=.true.
      lsides=lsides-south
  end if
  do_north=.false.
  if( lsides/north>=1)then
      if(gjm>1) do_north=.true.
      lsides=lsides-north
  end if
  do_west=.false.
  if( lsides/west>=1)then
      do_west=.true.
      lsides=lsides-west
  end if
  do_east=.false.
  if( lsides==east)then
      do_east=.true.
  end if

  if(gjm > 1) then
     send_size=send_size + given_width*max(full_i_blocksize,full_j_blocksize)*gkb
  else
     send_size=send_size + given_width*gkb
  end if
  if(send_size>ghost_max_bufsize) then
      call reallocate_sendbuffers
  end if
  
  updatelist3d(ul_ind3d)%sendwest=  do_east
  updatelist3d(ul_ind3d)%sendeast=  do_west
  updatelist3d(ul_ind3d)%recvwest=  do_west .and. block%west .ne. mpi_proc_null
  updatelist3d(ul_ind3d)%recveast=  do_east .and. block%east .ne. mpi_proc_null
  updatelist3d(ul_ind3d)%sendsouth= do_north
  updatelist3d(ul_ind3d)%sendnorth= do_south
  updatelist3d(ul_ind3d)%recvsouth= do_south .and. block%south .ne. mpi_proc_null
  updatelist3d(ul_ind3d)%recvnorth= do_north .and. block%north .ne. mpi_proc_null
  commtime2=rdtsc()
  totcopytime=totcopytime + (commtime2-commtime1)
end subroutine pack3d



subroutine update_ghost
!
! iterate through the list of fields to be updated, and build send buffers
! for east and west direction
!
! east-west communication first, then update all fields befor the north
! south communication. finish up by resetting all pointers.
!
  implicit none
  integer, parameter :: maxreq=5
  integer:: reqlist(maxreq), numreq
  integer:: wstat(mpi_status_size,2)
  integer width, kdim, li,i,j,k
! wheter or not to exchange ghost packages over an edge
  logical :: upd_east=.false. , upd_west=.false., upd_north=.false. , upd_south=.false.

  
! reasons to return:  
  if(.not.periodic_ew  .and.  (numproc==1 .or. block%prn==mpi_proc_null)) return
  
  upd_east=.false.  ; upd_west=.false.; upd_north=.false. ; upd_south=.false.
  
  commtime1=rdtsc()

! tag_ghost list:

  do li=1,ul_ind2d
      width=updatelist2d(li)%width

!  register if there is any activity on a side

      upd_east=upd_east .or. updatelist2d(li)%sendwest
      upd_west=upd_west .or. updatelist2d(li)%sendeast

      if(updatelist2d(li)%sendeast)then
          do j=j1,j2
              do i=i_eq_im-width+1,i_eq_im
                  i_sbe=i_sbe+1
                  sbe(i_sbe)=updatelist2d(li)%f(i,j)
              end do
          end do
      end if
      if(updatelist2d(li)%sendwest)then
          do j=j1,j2
              do i=1,width
                  i_sbw=i_sbw+1
                  sbw(i_sbw)=updatelist2d(li)%f(i,j)
              end do
          end do
      end if
  end do

! from tag_ghost:
  do li=1,ul_ind3d
      width=updatelist3d(li)%width
      kdim=updatelist3d(li)%kdim

!  register if there is any activity on a side

      upd_east=upd_east .or. updatelist3d(li)%sendwest
      upd_west=upd_west .or. updatelist3d(li)%sendeast

      if(updatelist3d(li)%sendeast)then
          do k=1,kdim
              do j=j1,j2
                  do i=i_eq_im-width+1,i_eq_im
                      i_sbe=i_sbe+1
                      sbe(i_sbe)=updatelist3d(li)%f(i,j,k)
                  end do
              end do
          end do
      end if
      if(updatelist3d(li)%sendwest)then
          do k=1,kdim
              do j=j1,j2
                  do i=1,width
                      i_sbw=i_sbw+1
                      sbw(i_sbw)=updatelist3d(li)%f(i,j,k)
                  end do
              end do
          end do
      end if
  end do
  commtime2=rdtsc()
  totcopytime=totcopytime + (commtime2-commtime1)

  commtime1=rdtsc()

  
! now do the actual exchange of halo data, for high cpu counts the
! code can easily spend half of the time doing this exchange.  on most
! platforms the fastest method is isend+irecv+waitall, but for some
! reason this is not always stable, e.g. recent mpich on linux. the
! default is therefore the blocking send/recv calls. on most platforms
! mpi_send is implemented as a non blocking call if the message size
! is below the "eager_limit". the below echange depends on this
! assumption. if the code deadlocks below, use mod_mpi_gum=1 or 2.
  
  select case(mod_mpi_gum)
  case(0)

! these calls may deadlock on systems with small eager size.
      if(upd_east.and. block%west.ne.mpi_proc_null)then
          if(block%east .ne. myid) then
              call mpi_send(sbw, i_sbw, mpi_float_type, block%west, tagwest, &
                   qmpi_comm_world, mpistatus, ierr)
          end if
      end if
      
      if(upd_east.and. block%east.ne.mpi_proc_null)then
          if(block%east == myid) then
              rbe = sbw
          else         
              call mpi_recv(rbe, i_sbw, mpi_float_type, block%east, tagwest, &
                   qmpi_comm_world, mpistatus, ierr)
          end if
      end if
      
      if(upd_west .and. block%east.ne.mpi_proc_null)then
          if(block%west .ne. myid) then
              call mpi_send(sbe, i_sbe, mpi_float_type, block%east, tageast, &
                   qmpi_comm_world, mpistatus, ierr)
          end if
      end if
      
      if(upd_west .and. block%west.ne.mpi_proc_null)then
          if(block%west == myid) then
              rbw = sbe
          else
              call mpi_recv(rbw, i_sbe, mpi_float_type, block%west, tageast, &
                   qmpi_comm_world, mpistatus, ierr)
          end if
      end if
      
  case(1)
      numreq=0
      if(upd_east.and. block%east.ne.mpi_proc_null)then
          if(block%east == myid) then
            !! periodic bc on one processor
              rbe = sbw
          else         
              numreq=numreq+1
              call mpi_irecv(rbe, i_sbw, mpi_float_type, block%east, tagwest, &
                   qmpi_comm_world, reqlist(numreq), ierr)
          end if
      end if
      
      if(upd_east.and. block%west.ne.mpi_proc_null)then
          if(block%east .ne. myid) then
              numreq=numreq+1
              call mpi_isend(sbw, i_sbw, mpi_float_type, block%west, tagwest, &
                   qmpi_comm_world, reqlist(numreq), ierr)
          end if
      end if          
      
      if(upd_west .and. block%west.ne.mpi_proc_null)then
          if(block%west == myid) then
            !! receive from ourselves
              rbw = sbe
          else
              numreq=numreq+1
              call mpi_irecv(rbw, i_sbe, mpi_float_type, block%west, tageast, &
                   qmpi_comm_world, reqlist(numreq),ierr)
          end if
      end if
      
      if(upd_west .and. block%east.ne.mpi_proc_null)then
          if(block%west .ne. myid) then
              numreq=numreq+1
              call mpi_isend(sbe, i_sbe, mpi_float_type, block%east, tageast, &
                   qmpi_comm_world, reqlist(numreq),ierr)
          end if
      end if
      
      if(numreq > 0)then
          call mpi_waitall(numreq, reqlist, wstat, ierr)
      end if
      
  case(2)  
      if(upd_west)then
          call mpi_sendrecv(sbe, i_sbe, mpi_float_type, block%east, tageast, &
               rbw, i_sbe, mpi_float_type, block%west, tageast, &
               qmpi_comm_world, mpistatus, ierr)
      end if

      if(upd_east)then
          call mpi_sendrecv(sbw, i_sbw, mpi_float_type, block%west, tagwest, &
               rbe, i_sbw, mpi_float_type, block%east, tagwest, &
               qmpi_comm_world, mpistatus, ierr)
      end if
  
  end select

  
  commtime2=rdtsc()
  totcommtime=totcommtime + (commtime2-commtime1)

  commtime1=rdtsc()
  
  do li=1,ul_ind2d
      width=updatelist2d(li)%width
!      kdim=updatelist(li)%kdim

      if(updatelist2d(li)%recveast)then
!          do k=1,kdim
              do j=j1,j2
                  do i=i_eq_im+1,i_eq_im+width
                      i_rbe=i_rbe+1
                      updatelist2d(li)%f(i,j)=rbe(i_rbe)
                  end do
              end do
!          end do
      end if
      if(updatelist2d(li)%recvwest)then
!          do k=1,kdim
              do j=j1,j2
                  do i=1-width,0
                      i_rbw=i_rbw+1
                      updatelist2d(li)%f(i,j)=rbw(i_rbw)
                  end do
              end do
!          end do
      end if
  end do

  do li=1,ul_ind3d
      width=updatelist3d(li)%width
      kdim=updatelist3d(li)%kdim

      if(updatelist3d(li)%recveast)then  !
          do k=1,kdim
              do j=j1,j2
                  do i=i_eq_im+1,i_eq_im+width
                      i_rbe=i_rbe+1
                      updatelist3d(li)%f(i,j,k)=rbe(i_rbe)
                  end do
              end do
          end do
      end if
      if(updatelist3d(li)%recvwest)then
          do k=1,kdim
              do j=j1,j2
                  do i=1-width,0
                      i_rbw=i_rbw+1
                      updatelist3d(li)%f(i,j,k)=rbw(i_rbw)
                  end do
              end do
          end do
      end if
  end do


! now build buffers for north and south

  if(gjm > 1)then
  
  do li=1,ul_ind2d
      width=updatelist2d(li)%width
!      kdim=updatelist(li)%kdim
      upd_north=upd_north .or. updatelist2d(li)%sendsouth
      upd_south=upd_south .or. updatelist2d(li)%sendnorth

      if(updatelist2d(li)%sendnorth)then
!          do k=1,kdim
              do j=j_eq_jm-width+1,j_eq_jm
                  do i=i1,i2
                      i_sbn=i_sbn+1
                      sbn(i_sbn)=updatelist2d(li)%f(i,j)
                  end do
              end do
!          end do
      end if
      if(updatelist2d(li)%sendsouth)then
!          do k=1,kdim
              do j=1,width
                  do i=i1,i2
                      i_sbs=i_sbs+1
                      sbs(i_sbs)=updatelist2d(li)%f(i,j)
                  end do
              end do
!          end do
      end if
  end do
  do li=1,ul_ind3d
      width=updatelist3d(li)%width
      kdim=updatelist3d(li)%kdim
      upd_north=upd_north .or. updatelist3d(li)%sendsouth
      upd_south=upd_south .or. updatelist3d(li)%sendnorth

      if(updatelist3d(li)%sendnorth)then
          do k=1,kdim
              do j=j_eq_jm-width+1,j_eq_jm
                  do i=i1,i2
                      i_sbn=i_sbn+1
                      sbn(i_sbn)=updatelist3d(li)%f(i,j,k)
                  end do
              end do
          end do
      end if
      if(updatelist3d(li)%sendsouth)then
          do k=1,kdim
              do j=1,width
                  do i=i1,i2
                      i_sbs=i_sbs+1
                      sbs(i_sbs)=updatelist3d(li)%f(i,j,k)
                  end do
              end do
          end do
      end if
  end do
  commtime2=rdtsc()

  totcopytime=totcopytime + (commtime2 - commtime1)

  commtime1=rdtsc()


  select case(mod_mpi_gum)
  case(0)
      if(upd_north .and. block%south .ne. mpi_proc_null)then
          call mpi_send(sbs, i_sbs, mpi_float_type,block%south, tagsouth, qmpi_comm_world, mpistatus, ierr)
      end if
      if(upd_north .and. block%north.ne.mpi_proc_null)then
          call mpi_recv(rbn, i_sbs, mpi_float_type, block%north, tagsouth, qmpi_comm_world, mpistatus, ierr)
      end if
      if(upd_south .and. block%north .ne. mpi_proc_null)then
          call mpi_send(sbn, i_sbn, mpi_float_type,block%north, tagnorth, qmpi_comm_world, mpistatus, ierr)
      end if
      if(upd_south .and. block%south .ne. mpi_proc_null)then
          call mpi_recv(rbs, i_sbn, mpi_float_type,block%south,tagnorth, qmpi_comm_world, mpistatus, ierr)
      end if

  case(1)    
      numreq=0
      if(upd_north .and. block%north.ne.mpi_proc_null)then
          numreq=numreq+1
          call mpi_irecv(rbn, i_sbs, mpi_float_type,block%north, tagsouth,qmpi_comm_world, reqlist(numreq),ierr)          
      end if
      if(upd_north .and. block%south .ne. mpi_proc_null)then
          numreq=numreq+1
          call mpi_isend(sbs, i_sbs, mpi_float_type,block%south, tagsouth, qmpi_comm_world,reqlist(numreq),ierr)
      end if
      if(upd_south .and. block%south .ne. mpi_proc_null)then
          numreq=numreq+1
          call mpi_irecv(rbs, i_sbn, mpi_float_type,block%south,tagnorth, qmpi_comm_world, reqlist(numreq),ierr)
      end if
      if(upd_south .and. block%north .ne. mpi_proc_null)then
          numreq=numreq+1
          call mpi_isend(sbn, i_sbn, mpi_float_type,block%north, tagnorth, qmpi_comm_world,reqlist(numreq),ierr)
      end if

      if(numreq > 0)then
          call mpi_waitall(numreq, reqlist, wstat, ierr)
      end if

case(2)  
      if(upd_south)then
          call mpi_sendrecv(sbn, i_sbn, mpi_float_type, block%north, tagnorth, &
               rbs, i_sbn, mpi_float_type, block%south, tagnorth, &
               qmpi_comm_world, mpistatus, ierr)
      end if

      if(upd_north)then
          call mpi_sendrecv(sbs, i_sbs, mpi_float_type, block%south, tagsouth, &
               rbn, i_sbs, mpi_float_type, block%north, tagsouth, &
               qmpi_comm_world, mpistatus, ierr)
      end if
  end select

  
  commtime2=rdtsc()
  totcommtime=totcommtime + (commtime2-commtime1)
  
  commtime1=rdtsc()

  do li=1,ul_ind2d
      width=updatelist2d(li)%width
!      kdim=updatelist(li)%kdim
      if(updatelist2d(li)%recvsouth)then
!          do k=1,kdim
              do j=1-width,0
                  do i=i1,i2
                      i_rbs=i_rbs+1
                      updatelist2d(li)%f(i,j)=rbs(i_rbs)
                  end do
              end do
!          end do
      end if
      if(updatelist2d(li)%recvnorth)then
!          do k=1,kdim
              do j=j_eq_jm+1,j_eq_jm+width
                  do i=i1,i2
                      i_rbn = i_rbn+1
                      updatelist2d(li)%f(i,j) = rbn(i_rbn)
                  end do
              end do
!          end do
      end if
  end do
  do li=1,ul_ind3d
      width=updatelist3d(li)%width
      kdim=updatelist3d(li)%kdim
      if(updatelist3d(li)%recvsouth)then
          do k=1,kdim
              do j=1-width,0
                  do i=i1,i2
                      i_rbs=i_rbs+1
                      updatelist3d(li)%f(i,j,k)=rbs(i_rbs)
                  end do
              end do
          end do
      end if
      if(updatelist3d(li)%recvnorth)then
          do k=1,kdim
              do j=j_eq_jm+1,j_eq_jm+width
                  do i=i1,i2
                      i_rbn=i_rbn+1
                      updatelist3d(li)%f(i,j,k)=rbn(i_rbn)
                  end do
              end do
          end do
      end if
  end do

  end if ! jm > 1

  commtime2=rdtsc()
  totcopytime=totcopytime + (commtime2-commtime1)
  
!
! now everything should be updated - and we can reset all counters etc.
!
  send_size=0
  ul_ind2d=0
  ul_ind3d=0
  i_sbe=0 ; i_sbw=0; i_sbn=0 ; i_sbs=0
  i_rbe=0 ; i_rbw=0; i_rbn=0 ; i_rbs=0
end subroutine update_ghost



subroutine copy_region(from, to, kdim, f,f2,f3,f4)
!
! given a distributed array f, optionally f2,f3,f4,
! from=(fi1,fi2,fj1,fj2,fk1,fk2), to=(ti1,ti2,tj1,tj2,tk1,tk2)
! defines global ranges for the regions, 
!   call copy_region(from, to, kb, f, f2)   
! perform the equivalent of the sequential calls
!   f(ti1:ti2,tj1:tj2,tk1:tk2)  = f(fi1:fi2,fj1:fj2,fk1:fk2)
!   f2(ti1:ti2,tj1:tj2,tk1:tk2) = f2(fi1:fi2,fj1:fj2,fk1:fk2)
!   ...
!
  implicit none

  integer, dimension(6),intent(in) :: from, to
  integer kdim
  real, dimension(i1:i2,j1:j2,kdim) :: f
  real, dimension(i1:i2,j1:j2,kdim), optional :: f2,f3,f4

  integer, dimension(6) :: tgt, src
  integer, dimension(4) :: block_first, range, offset, ssrc
  integer, dimension(:,:,:), allocatable :: lsrc, ltgt

  integer i,j,bi1,bi2,bj1,bj2, isize, jsize, ksize, size, sendsize,recvsize
  integer recvtag, sendtag, tsum, ssum

  integer, parameter :: maxreq=8
  integer:: reqlist(maxreq), numreq
  integer,dimension(mpi_status_size,2) :: wstat
  logical inblock

  type buftype
     real data(copy_max_bufsize)
  end type buftype

  type(buftype), dimension(:,:), allocatable :: recv_buf, send_buf

  tgt=to
  src=from
  if(tgt(6)-tgt(5)>=kdim .or. src(6)-src(5)>=kdim) then
      write(*,*)'ERROR: (copy_region) vertical target/source region larger than 3d array KB'
      return
  end if

  if(numproc==1)then
      f(tgt(1):tgt(2),tgt(3):tgt(4),tgt(5):tgt(6))=f(src(1):src(2),src(3):src(4),src(5):src(6))
      if(present(f2)) f2(tgt(1):tgt(2),tgt(3):tgt(4),tgt(5):tgt(6)) = &
                      f2(src(1):src(2),src(3):src(4),src(5):src(6))
      if(present(f3)) f3(tgt(1):tgt(2),tgt(3):tgt(4),tgt(5):tgt(6)) = &
                      f3(src(1):src(2),src(3):src(4),src(5):src(6))
      if(present(f4)) f4(tgt(1):tgt(2),tgt(3):tgt(4),tgt(5):tgt(6)) = &
                      f4(src(1):src(2),src(3):src(4),src(5):src(6))
      return
  end if

  block_first = (/ block%fi, block%fi, block%fj, block%fj /)
  offset      = (/tgt(1)-src(1), tgt(1)-src(1), tgt(3)-src(3), tgt(3)-src(3)/)
  tsum        = tgt(2)-tgt(1) + tgt(4)-tgt(3) + tgt(6)-tgt(5)
  ssum        = src(2)-src(1) + src(4)-src(3) + src(6)-src(5)

  if(tsum /= ssum) then
      write(*,*)'ERROR: (copy_region) the target and the source dimensions are different! '
      return
  end if

  numreq=0

! return the local piece of the full target
!
  call globalrange2local(block%ibn, block%jbn, tgt, inblock)
!
! Set up the IRECV's if current processor own a piece of the target
!
  if(inblock) then

     ! global coordinates of the piece in target
      range = tgt(1:4) + block_first -1
      range= range-offset

      call blocks_covering_region(range,bi1,bi2,bj1,bj2)

      allocate(recv_buf(bi1:bi2,bj1:bj2), ltgt(4,bi1:bi2,bj1:bj2), lsrc(4,bi1:bi2,bj1:bj2))

      do j=bj1,bj2
          do i=bi1,bi2
              ltgt(:,i,j) = range
              call globalrange2local(i,j, ltgt(:,i,j), inblock)
              lsrc(:,i,j)=ltgt(:,i,j)

              ltgt(:,i,j)=ltgt(:,i,j)+(/bltab(i,j)%fi,bltab(i,j)%fi,bltab(i,j)%fj,bltab(i,j)%fj/)-1
              ltgt(:,i,j)=ltgt(:,i,j)+offset- block_first+1

              size = (ltgt(2,i,j)-ltgt(1,i,j)+1) * (ltgt(4,i,j)-ltgt(3,i,j)+1) * (tgt(6)-tgt(5)+1)
              recvsize=size
              if(present(f2)) recvsize=2*size
              if(present(f3)) recvsize=3*size
              if(present(f4)) recvsize=4*size

              if(bltab(i,j)%prn .ne. mpi_proc_null .AND. bltab(i,j)%prn /= myid) then

                !  each pair of processors can exchange one message: a unique tag is therefore
                !    source*1000+target

                  recvtag = myid + 1000*bltab(i,j)%prn
                  numreq=numreq+1

                  call mpi_irecv(recv_buf(i,j)%data, recvsize, mpi_float_type, bltab(i,j)%prn, recvtag, &
                       qmpi_comm_world, reqlist(numreq), ierr)
              end if

          end do
      end do
  end if

!
! Set up the ISENDs on the source processor
!
  call globalrange2local(block%ibn, block%jbn, src, inblock)

  if(inblock) then
      range = src(1:4) + block_first -1 + offset
      call blocks_covering_region(range,bi1,bi2,bj1,bj2)

      allocate(send_buf(bi1:bi2,bj1:bj2))

      do j=bj1,bj2
          do i=bi1,bi2

              if(bltab(i,j)%prn /= myid .AND. bltab(i,j)%prn .ne. mpi_proc_null) then

                  ssrc=range
                  call globalrange2local(i,j, ssrc, inblock)

                  ssrc=ssrc+(/bltab(i,j)%fi,bltab(i,j)%fi,bltab(i,j)%fj,bltab(i,j)%fj/)-1
                  ssrc=ssrc - offset - block_first + 1

                  size=(ssrc(2)-ssrc(1)+1) * (ssrc(4)-ssrc(3)+1) *(src(6)-src(5)+1)
                  sendsize=size

                !  hvert prosessorpar kan bare send en melding til hverandre.
                !  en unik tag blir dermed source*1000+target
                  sendtag = myid*1000 + bltab(i,j)%prn

                  send_buf(i,j)%data(1:size) = &
                       reshape(f(ssrc(1):ssrc(2),ssrc(3):ssrc(4),src(5):src(6)),(/size/))

                  if(present(f2)) then
                      send_buf(i,j)%data(size+1:2*size) = &
                           reshape(f2(ssrc(1):ssrc(2),ssrc(3):ssrc(4),src(5):src(6)),(/size/))
                      sendsize=2*size
                  end if
                  if(present(f3)) then
                      send_buf(i,j)%data(2*size+1:3*size) = &
                           reshape(f3(ssrc(1):ssrc(2),ssrc(3):ssrc(4),src(5):src(6)),(/size/))
                      sendsize=3*size
                  end if
                  if(present(f4)) then
                      send_buf(i,j)%data(3*size+1:4*size) = &
                           reshape(f4(ssrc(1):ssrc(2),ssrc(3):ssrc(4),src(5):src(6)),(/size/))
                      sendsize=4*size
                  end if

                  numreq=numreq+1
                  call mpi_isend(send_buf(i,j)%data, sendsize, mpi_float_type, bltab(i,j)%prn, &
                       sendtag, qmpi_comm_world, reqlist(numreq), ierr)
              end if

          end do
      end do
  end if


! we do not start writing anything in f before all messages are sent and received

  if(numreq>0)then
      call mpi_waitall(numreq, reqlist, wstat, ierr)
  end if


  tgt=to
  call globalrange2local(block%ibn, block%jbn, tgt, inblock)

  if(inblock) then
      range = tgt(1:4) + block_first -1 -offset
      call blocks_covering_region(range,bi1,bi2,bj1,bj2)

      do j=bj1,bj2
          do i=bi1,bi2

              if(bltab(i,j)%prn .NE. myid) then
                  isize=ltgt(2,i,j)-ltgt(1,i,j)+1
                  jsize=ltgt(4,i,j)-ltgt(3,i,j)+1
                  ksize=tgt(6)-tgt(5)+1
                  size= isize*jsize*ksize

                  f(ltgt(1,i,j):ltgt(2,i,j), ltgt(3,i,j):ltgt(4,i,j), tgt(5):tgt(6))=&
                       reshape(recv_buf(i,j)%data(1:size), (/isize,jsize,ksize/))
                  if(present(f2))then
                      f2(ltgt(1,i,j):ltgt(2,i,j), ltgt(3,i,j):ltgt(4,i,j), tgt(5):tgt(6))=&
                           reshape(recv_buf(i,j)%data(size+1:2*size), (/isize,jsize,ksize/))
                  end if
                  if(present(f3))then
                      f3(ltgt(1,i,j):ltgt(2,i,j), ltgt(3,i,j):ltgt(4,i,j), tgt(5):tgt(6))=&
                           reshape(recv_buf(i,j)%data(2*size+1:3*size), (/isize,jsize,ksize/))
                  end if
                  if(present(f4))then
                      f4(ltgt(1,i,j):ltgt(2,i,j), ltgt(3,i,j):ltgt(4,i,j), tgt(5):tgt(6))=&
                           reshape(recv_buf(i,j)%data(3*size+1:4*size), (/isize,jsize,ksize/))
                  end if

              else
                  f(ltgt(1,i,j):ltgt(2,i,j), ltgt(3,i,j):ltgt(4,i,j), tgt(5):tgt(6))=&
                       f(lsrc(1,i,j):lsrc(2,i,j), lsrc(3,i,j):lsrc(4,i,j), src(5):src(6))
                  if(present(f2))then
                      f2(ltgt(1,i,j):ltgt(2,i,j), ltgt(3,i,j):ltgt(4,i,j), tgt(5):tgt(6))=&
                           f2(lsrc(1,i,j):lsrc(2,i,j), lsrc(3,i,j):lsrc(4,i,j), src(5):src(6))
                  end if
                  if(present(f3))then
                      f3(ltgt(1,i,j):ltgt(2,i,j), ltgt(3,i,j):ltgt(4,i,j), tgt(5):tgt(6))=&
                           f3(lsrc(1,i,j):lsrc(2,i,j), lsrc(3,i,j):lsrc(4,i,j), src(5):src(6))
                  end if
                  if(present(f4))then
                      f4(ltgt(1,i,j):ltgt(2,i,j), ltgt(3,i,j):ltgt(4,i,j), tgt(5):tgt(6))=&
                           f4(lsrc(1,i,j):lsrc(2,i,j), lsrc(3,i,j):lsrc(4,i,j), src(5):src(6))
                  end if
              end if
          end do
      end do

      deallocate(recv_buf, ltgt, lsrc)
  end if

  if(allocated(send_buf)) deallocate(send_buf)

end subroutine copy_region


subroutine sec2ind(n,i,j,k)
!
! convert from (block) local sequential index n (starting from 0 !) to local i,j,k
!
  integer, intent(in) :: n
  integer, intent(out) :: i,j,k

  i = mod(n,i_blocksize) + 1
  k = n/(i_blocksize*j_blocksize) + 1
  j = ((n+1 - (k-1)*i_blocksize*j_blocksize)-1)/i_blocksize +1
end subroutine sec2ind

subroutine ind2sec(i,j,k,n)
!
! convert from local coordinates i,j,k to the (block)local sequential index n (starting from 0 !)
!
  integer, intent(out) :: n
  integer, intent(in) :: i,j,k

  n = (k-1)*i_blocksize*j_blocksize + (j-1)*i_blocksize + i - 1
end subroutine ind2sec




real function globval2d(f,i,j,k)
  implicit none
  real f(i1:,j1:)
  integer i,j
  integer, optional :: k
  if(present(k)) then
      write(*,*)'globval2d: argument k does not make sense for 2d array f'
      stop
  end if
  globval2d = genglobval(f,1,i,j,1)
  return
end function globval2d

real function globval3d(f,i,j,k)
  implicit none
  real f(i1:,j1:,:)
  integer i,j
  integer, optional :: k
  if(.not.present(k)) then
      write(*,*)'globval3d: argument k not present'
      stop
  end if
  if(k>size(f,3)) then
      write(*,*)'globval3d: argument k>size(f,3)'
      stop
  end if
  globval3d = genglobval(f,size(f,3),i,j,k)
  return
end function globval3d

real function genglobval(f,kb,i,j,k)
  implicit none
  integer kb,i,j,k
  real f(i1:i2,j1:j2,kb)
  real(dp) tmpf
  integer li,lj
  logical inblock

  if(k>kb) then
      write(*,*)'genglobval: k must be in the range 1,kb=',1,kb,' k=',k
      stop
  end if
  if(i>gim .or. i<1) then
      write(*,*)'genglobval: global coordinate out of range 1,im'
      stop
  end if
  if(j>gjm .or. j<1) then
      write(*,*)'genglobval: global coordinate out of range 1,jm'
      stop
  end if
  if(numproc==1)then
      genglobval=f(i,j,k)
      return
  end if

  li=i
  lj=j
  call globalind2local(li, lj, inblock)

  if(inblock)then
      tmpf = real(f(li,lj,k), dp)
  else
      tmpf = 1.0_dp
  end if
  call reduce('mul', tmpf)
  genglobval = tmpf
end function genglobval



subroutine find_max_reals(kb, f, val, loc, root, wet, range)
!
! return the global maximum value and location for the distributed array f.
! if given, root specifies on which processor to return result,
! if root is not given, the result is returned on all processors
! if wet is present and true, only consider wet points
! if range is present, this only find the max in that range.
  implicit none
  integer, intent(in) :: kb
  real, dimension(i1:i2,j1:j2,kb) :: f
  real, intent(out) :: val
  integer, intent(out) :: loc(3)
  integer, optional, intent(in) :: root
  logical, optional, intent(in) :: wet
  integer, optional :: range(6)
  logical checkall, inblock
  real in(2,1), out(2,1)
  integer n, boss, lloc(3), maxrank, m, bi,bj,i,j,k,kmax,r(6)
  real, allocatable :: tmpf(:,:,:)

  loc(:)=0
  lloc(:)=1
  boss=0
  if(present(root))boss=root
  checkall=.true.
  if(present(wet)) then
      if(wet) checkall=.false. ! only find max among wet points
  end if
  if(present(range)) then
      r=range
  else
      if(kb==1) then
          r=(/1,gim,1,gjm,1,1/)
      else
          r=(/1,gim,1,gjm,1,kb-1/)
      end if
  end if
  call globalrange2local(block%ibn, block%jbn, r(1:4), inblock)

  m=1
  in(1,m) = -1e20

  if(inblock)then
      if(checkall)then
          in(1,m) = maxval(f(r(1):r(2),r(3):r(4),r(5):r(6)))
          lloc = maxloc(f(r(1):r(2),r(3):r(4),r(5):r(6)))
      else
          allocate(tmpf(r(1):r(2),r(3):r(4),r(5):r(6)))
          do k=r(5),r(6)
              do j=r(3),r(4)
                  do i=r(1),r(2)
                      if(gridbox(i,j) == 0) then
                          tmpf(i,j,k) = -1e20
                      else
                          tmpf(i,j,k) = f(i,j,k)
                      end if
                  end do
              end do
          end do
          in(1,m) = maxval(tmpf(r(1):r(2),r(3):r(4),r(5):r(6)))
          lloc = maxloc(tmpf(r(1):r(2),r(3):r(4),r(5):r(6)))
          deallocate(tmpf)
      end if

      if(numproc==1)then
          val=in(1,m)
          loc=lloc
          return
      end if
  end if  ! inblock

  lloc(1:2) = lloc(1:2) + (/r(1)-1,r(3)-1/)
  call ind2sec(lloc(1),lloc(2),lloc(3),n)

  in(2,m)= myid*i_blocksize*j_blocksize*kb + n

  if(present(root))then
      call mpi_reduce(in, out, m, mpi_2floats, mpi_maxloc, boss, qmpi_comm_world, ierr)
  else
      call mpi_allreduce(in, out, m, mpi_2floats, mpi_maxloc, qmpi_comm_world, ierr)
  end if

  if(myid==boss .or. .not. present(root))then
      val = out(1,m)

      n= mod(int(out(2,m)),i_blocksize*j_blocksize*kb)
      call sec2ind(n,loc(1),loc(2),loc(3))

      maxrank = out(2,m)/(i_blocksize*j_blocksize*kb)
      bi=block_of_rank(1,maxrank) ; bj=block_of_rank(2,maxrank)

      loc(1:2)=loc(1:2) + (/bltab(bi,bj)%fi,bltab(bi,bj)%fj/) - 1
  end if
end subroutine find_max_reals



subroutine find_min_reals(kb, f, val, loc, root, wet, range)
!
! return the global minimum value and location for the distributed array f
! on processor 0. if given, root specifies on which processor to return result,
! if root is not specified, result will be returned on all processors.
! if wet is present and true, only consider wet points
!  
  implicit none
  integer, intent(in) :: kb
  real, dimension(i1:i2,j1:j2,kb) :: f
  real, intent(out) :: val
  integer, intent(out) :: loc(3)
  integer, optional, intent(in) :: root
  logical, optional :: wet  
  integer, optional, intent(in) :: range(6)
  real in(2,1), out(2,1)
  integer n, boss, lloc(3), maxrank, m, bi,bj,i,j,k,kmax,r(6)
  real, allocatable :: tmpf(:,:,:)
  logical checkall, inblock
  
  loc(:)=0
  lloc(:)=1
  boss=0
  if(present(root))boss=root

  checkall=.true.
  if(present(wet)) then
      if(wet) checkall=.false. ! only find max among wet points
  end if
  if(present(range)) then
      r=range
  else
      if(kb==1) then
          r=(/1,gim,1,gjm,1,1/)
      else
          r=(/1,gim,1,gjm,1,kb-1/)
      end if
  end if
  call globalrange2local(block%ibn, block%jbn, r(1:4), inblock)
  
  m=1
  if(checkall)then
      in(1,m) = minval(f(r(1):r(2),r(3):r(4),r(5):r(6)))
      lloc = minloc(f(r(1):r(2),r(3):r(4),r(5):r(6)))
  else
      allocate(tmpf(i1:i2,j1:j2,kb))
      do k=r(5),r(6)
         do j=r(3),r(4)
            do i=r(1),r(2)
                  if(gridbox(i,j) == 0) then
                      tmpf(i,j,k) = 1e20
                  else
                      tmpf(i,j,k) = f(i,j,k)
                  end if
              end do
          end do
      end do
      in(1,m) = minval(tmpf(r(1):r(2),r(3):r(4),r(5):r(6)))
      lloc = minloc(tmpf(r(1):r(2),r(3):r(4),r(5):r(6)))
      deallocate(tmpf) 
  end if 
 
  if(numproc==1)then
      val=in(1,m)
      loc=lloc
      return
  end if

  lloc(1:2) = lloc(1:2) + (/r(1)-1,r(3)-1/)      
  call ind2sec(lloc(1),lloc(2),lloc(3),n)

  in(2,m)= myid*i_blocksize*j_blocksize*kb + n

  if(present(root))then
      call mpi_reduce(in, out, m, mpi_2floats, mpi_minloc, boss, qmpi_comm_world, ierr)
  else
      call mpi_allreduce(in, out, m, mpi_2floats, mpi_minloc, qmpi_comm_world, ierr)
  end if

  if(myid==boss .or. .not. present(root))then
      val = out(1,m)
      maxrank = out(2,m)/(i_blocksize*j_blocksize*kb)

      n= mod(int(out(2,m)),i_blocksize*j_blocksize*kb)

      call sec2ind(n,loc(1),loc(2),loc(3))
      bi=block_of_rank(1,maxrank) ; bj=block_of_rank(2,maxrank)
      loc(1:2)=loc(1:2) + (/bltab(bi,bj)%fi,bltab(bi,bj)%fj/) - 1
  end if
end subroutine find_min_reals


function qsum(f,kb,wet,r,inblock)
  implicit none
  real qsum
  integer im,jm,kb
  real f(i1:i2,j1:j2,kb)
  integer :: r(4),i,j,k
  logical :: wet, inblock
   
  qsum = 0.0
  if(inblock)then
      do k=1,kb
          do j=r(3),r(4)
              do i=r(1),r(2)
                  qsum = qsum + f(i,j,k)
              end do
          end do
      end do
  end if
  call reduce('sum', qsum)  
end function qsum



function psum(f,im,jm,kb,range)
!
! the sum of f is found by first dividing f up into blocks of size
! mblsz*mblsz*kb, find the sum on each of these blocks, and then in the
! final step sum these togheter in a fixed order. mblsz=1+2*ghost_width,
! so that the block boundary location can be arbitrary.
! IMPORTANT: the routine assumes f to be FULLY UPDATED in the ghost zone.
!
! range is an optional global range, for which to compute the sum. if
! not present, the sum will be computed over
! range=(/2,im-1,2,jm-1,1,kb-1/)
  implicit none

  real psum, fsum
  integer im,jm,kb
  real f(i1:i2,j1:j2,kb)
  integer, optional :: range(6)
  real, allocatable :: gcf(:,:), cf(:,:)
  integer i,j,k,ii,jj,bi,bj
  real, allocatable :: sendbuf(:),recvbuf(:)
  integer displs(numproc),recvcounts(numproc)
  integer sendcount, sbi,sbj, gi,gj,li,lj,currdisp,p
  integer d1,d2,bi1,bi2,bj1,bj2
  integer r(6)

  r = (/2,im-1,2,jm-1,1,max(1,kb-1)/)

  if(present(range)) then
      r = range
      if(r(1) < 1 .or. r(2)>gim .or. r(3)<1 .or. r(4)>gjm .or. r(5)<1 .or. r(6)>kb) then
          if(master)write(*,'(a,i0,":",i0,",",i0,":",i0,",",i0,":",i0)') "The provided range for PSUM is wrong, range=",range
          call abort()
      end if      
  end if
  
  sendcount=(block%sbi2-block%sbi1+1)*(block%sbj2-block%sbj1+1)
  
  if(.not.block%sbowner) sendcount=0
  
  allocate(cf(block%sbi1:block%sbi2,block%sbj1:block%sbj2), sendbuf(sendcount))
  
  allocate(recvbuf(numpbi*numpbj), gcf(numpbi,numpbj))
  
  gcf(:,:)=0.0 
  cf(:,:)=0.0
  
  !! on each processor, loop over sub blocks and find sums.
  if(block%sbowner)then

  if(jm > 1) then
      
  do sbj=block%sbj1,block%sbj2
      do sbi=block%sbi1,block%sbi2
          do k=r(5),r(6)
              do j=1,mblszj
                  do i=1,mblszi
                      gi=(sbi-1)*mblszi+i
                      gj=(sbj-1)*mblszj+j
                      if(gi>=r(1) .and. gj>=r(3) .and. gi<=r(2) .and. gj<=r(4))then
                          ! we're on the model grid
                          li = gi-block%fi+1
                          lj = gj-block%fj+1
                          if(li >= i1 .and. li <=i2 .and. lj >= j1 .and. lj <= j2) then
                              ! we're on the block
                              cf(sbi,sbj) = cf(sbi,sbj) + f(li,lj,k)
                          end if
                      end if
                  end do
              end do
          end do
      end do
  end do
  
  else
! jm==1
     do sbi=block%sbi1,block%sbi2
        do k=r(5),r(6)
           do i=1,mblszi
              gi=(sbi-1)*mblszi+i
              if(gi>=r(1) .and. gi<=r(2))then
                 li = gi-block%fi+1
                 cf(sbi,1) = cf(sbi,1) + f(li,1,k)
              end if
           end do
        end do
     end do
  end if

  if(numproc>1)then
      sendbuf=reshape(cf(block%sbi1:block%sbi2,block%sbj1:block%sbj2), (/sendcount/))
  else
      gcf = cf
  end if
  
  end if

  commtime1 = rdtsc()
  
  if(numproc>1)then  
      currdisp=0
      do p=1,numproc
          bi=block_of_rank(1,p-1) ; bj=block_of_rank(2,p-1)
          displs(p) = currdisp
          if(bltab(bi,bj)%sbowner)then
              recvcounts(p) = (bltab(bi,bj)%sbi2-bltab(bi,bj)%sbi1+1)*(bltab(bi,bj)%sbj2-bltab(bi,bj)%sbj1+1)
          else
              recvcounts(p)=0
          end if
          currdisp = currdisp + recvcounts(p)
      end do
      
      call mpi_gatherv(sendbuf, sendcount, mpi_float_type, &
           recvbuf, recvcounts, displs, mpi_float_type, 0, qmpi_comm_world, ierr)
  
      if(master)then      
          do p=1,numproc
              bi=block_of_rank(1,p-1)
              bj=block_of_rank(2,p-1)
              if(bltab(bi,bj)%sbowner)then
                  bi1=bltab(bi,bj)%sbi1 ; bi2=bltab(bi,bj)%sbi2
                  bj1=bltab(bi,bj)%sbj1 ; bj2=bltab(bi,bj)%sbj2
                  d1=displs(p)+1
                  d2=d1+recvcounts(p)-1
                  gcf(bi1:bi2,bj1:bj2) = reshape(recvbuf(d1:d2), (/bi2-bi1+1,bj2-bj1+1/))
              end if
          end do
      end if
  end if
  
  if(master)then
      fsum=0.0
      do jj=1,numpbj
          do ii=1,numpbi
              fsum=fsum + gcf(ii,jj)
          end do
      end do
  end if

  deallocate(recvbuf,sendbuf,cf,gcf)
  
  if(numproc>1) call broadcast(fsum)
  commtime2 = rdtsc()
  totsumtime = totsumtime + (commtime2-commtime1)
  
  psum=fsum
end function psum


subroutine phsum(f,im,jm,kb,hsum,range)
!
! The sum of each horisontal slice of F is found by first dividing F
! up into blocks of size MBLSZ*MBLSZ*KB, find the sum on each of these
! blocks, and then in the final step sum these togheter in a fixed
! order. MBLSZ=1+2*GHOST_WIDTH, so that the block boundary location
! can be arbitrary. The sums are returned in the HSUM array, 1d, dimension KB.
!  
! IMPORTANT: the routine assumes F to be fully updated in the ghost zone.
!
! RANGE is an optional global range, for which to compute the sum. if
! not present, the sum will be computed over RANGE=(/2,IM-1,2,JM-1,1,KB-1/)
!  
  implicit none
  real hsum(kb), fhsum(kb)
  integer im,jm,kb
  real f(i1:i2,j1:j2,kb)
  integer, optional :: range(6)
  real, allocatable :: gcf(:,:,:), cf(:,:,:)
  integer i,j,k,ii,jj,bi,bj
  real, allocatable :: sendbuf(:),recvbuf(:)
  integer displs(numproc),recvcounts(numproc)
  integer sendcount, sbi,sbj, gi,gj,li,lj,currdisp,p
  integer d1,d2,bi1,bi2,bj1,bj2
  integer r(6),slice

  r = (/2,im-1,2,jm-1,1,max(1,kb-1)/)

  if(present(range)) then
      r = range
      if(r(1) < 1 .or. r(2)>gim .or. r(3)<1 .or. r(4)>gjm .or. r(5)<1 .or. r(6)>kb) then
          if(master)write(*,'(a,i0,":",i0,i0,":",i0,i0,":",i0)') "The provided range for PHSUM is wrong, range=",range
          call abort()
      end if      
  end if
  
  sendcount=(block%sbi2-block%sbi1+1)*(block%sbj2-block%sbj1+1)*kb
  
  if(.not.block%sbowner) sendcount=0
  
  allocate(cf(block%sbi1:block%sbi2,block%sbj1:block%sbj2,kb), sendbuf(sendcount))
  
  allocate(recvbuf(numpbi*numpbj*kb), gcf(numpbi,numpbj,kb))
  
  gcf(:,:,:)=0.0 
  cf(:,:,:)=0.0
  
  !! on each processor, loop over sub blocks and find sums.
  if(block%sbowner)then

  if(jm > 1) then
      
  do sbj=block%sbj1,block%sbj2
      do sbi=block%sbi1,block%sbi2
          do k=r(5),r(6)
              do j=1,mblszj
                  do i=1,mblszi
                      gi=(sbi-1)*mblszi+i
                      gj=(sbj-1)*mblszj+j
                      if(gi>=r(1) .and. gj>=r(3) .and. gi<=r(2) .and. gj<=r(4))then
                          li = gi-block%fi+1
                          lj = gj-block%fj+1
                          cf(sbi,sbj,k) = cf(sbi,sbj,k) + f(li,lj,k)
                      end if
                  end do
              end do
          end do
      end do
  end do
  
  else
     do sbi=block%sbi1,block%sbi2
        do k=r(5),r(6)
           do i=1,mblszi
              gi=(sbi-1)*mblszi+i
              if(gi>=r(1) .and. gi<=r(2))then
                 li = gi-block%fi+1
                 cf(sbi,1,k) = cf(sbi,1,k) + f(li,1,k)
              end if
           end do
        end do
     end do
  end if

  if(numproc > 1)then
      sendbuf = reshape(cf(block%sbi1:block%sbi2,block%sbj1:block%sbj2,1:kb), (/sendcount/))
  else
      gcf = cf
  end if
  
  end if

  commtime1 = rdtsc()
  
  if(numproc>1)then
  
      currdisp=0
      do p=1,numproc
          bi=block_of_rank(1,p-1) ; bj=block_of_rank(2,p-1)
          displs(p) = currdisp
          if(bltab(bi,bj)%sbowner)then
              recvcounts(p) = (bltab(bi,bj)%sbi2-bltab(bi,bj)%sbi1+1)*(bltab(bi,bj)%sbj2-bltab(bi,bj)%sbj1+1)*kb
          else
              recvcounts(p)=0
          end if
          currdisp = currdisp + recvcounts(p)
      end do
      
      call mpi_gatherv(sendbuf, sendcount, mpi_float_type, &
           recvbuf, recvcounts, displs, mpi_float_type, 0, qmpi_comm_world, ierr)
  
      if(master)then      
          do p=1,numproc
              bi=block_of_rank(1,p-1)
              bj=block_of_rank(2,p-1)
              if(bltab(bi,bj)%sbowner)then
                  bi1=bltab(bi,bj)%sbi1 ; bi2=bltab(bi,bj)%sbi2
                  bj1=bltab(bi,bj)%sbj1 ; bj2=bltab(bi,bj)%sbj2
                  d1=displs(p)+1
                  d2=d1+recvcounts(p)-1
                  gcf(bi1:bi2,bj1:bj2,1:kb) = reshape(recvbuf(d1:d2), (/bi2-bi1+1,bj2-bj1+1,kb/))
              end if
          end do
      end if
  end if
  
  if(master)then
      fhsum = 0.0
      do slice=1,kb
          do jj=1,numpbj
              do ii=1,numpbi
                  fhsum(slice) = fhsum(slice) + gcf(ii,jj,slice)
              end do
          end do
      end do
  end if

  deallocate(recvbuf,sendbuf,cf,gcf)
  
  if(numproc > 1) call broadcast(fhsum)

  commtime2 = rdtsc()
  totsumtime = totsumtime + (commtime2-commtime1)
  
  hsum = fhsum
end subroutine phsum


SUBROUTINE PVSUM(F,IM,JM,KB,DIRECTION,VSUM,RANGE)
!
! The sum of each vertical slice of F, where DIRECTION=1 gives slices
! along constant j, 2 for i, is found by first dividing F up into blocks of
! size MBLSZ*MBLSZ*KB, find the sum on each of these blocks, and then
! in the final step sum these togheter in a fixed
! order. MBLSZ=1+2*GHOST_WIDTH, so that the block boundary location
! can be arbitrary. The sums are returned in the 1d array VSUM.
!  
! IMPORTANT: the routine assumes F to be fully updated in the ghost zone.
!
! RANGE is an optional global range, for which to compute the sum. if
! not present, the sum will be computed over RANGE=(/2,IM-1,2,JM-1,1,KB-1/)
!  
  implicit none
  integer im,jm,kb,direction
  real f(i1:i2,j1:j2,kb), vsum(:)
  integer, optional :: range(6)
  real, allocatable :: gcf(:,:,:), cf(:,:,:)
  integer i,j,k,ii,jj,bi,bj
  real, allocatable :: sendbuf(:),recvbuf(:)
  integer displs(numproc),recvcounts(numproc)
  integer sendcount, sbi,sbj, gi,gj,li,lj,currdisp,p
  integer d1,d2,bi1,bi2,bj1,bj2
  integer r(6),kdim,dir,js,je
  
  r = (/2,im-1,2,jm-1,1,max(1,kb-1)/)

! use the k dimension in the summing array to store a vector along the pvsum direction on each psum block.
  select case(direction)
  case(1)
      kdim=mblszi
      if(size(vsum) < jm) call abort("argument VSUM not large anough to hold PVSUM result")
  case(2)
      kdim=mblszj
      if(size(vsum) < im) call abort("argument VSUM not large anough to hold PVSUM result")
  case default
      call abort("invalid direction for subroutine PVSUM")
  end select
  
  if(present(range)) then
      r = range
      if(r(1) < 1 .or. r(2)>gim .or. r(3)<1 .or. r(4)>gjm .or. r(5)<1 .or. r(6)>kb) then
          if(master)write(*,'(a,i0,":",i0,i0,":",i0,i0,":",i0)') "The provided range for PVSUM is wrong, RANGE=",range
          call abort()
      end if      
  end if
  
  sendcount=(block%sbi2-block%sbi1+1)*(block%sbj2-block%sbj1+1)*kdim
  
  if(.not.block%sbowner) sendcount=0
  
  allocate(cf(block%sbi1:block%sbi2,block%sbj1:block%sbj2,kdim), sendbuf(sendcount))
  
  allocate(recvbuf(numpbi*numpbj*kdim), gcf(numpbi,numpbj,kdim))
  
  gcf(:,:,:)=0.0 
  cf(:,:,:)=0.0
  
  !! on each processor, loop over sub blocks and find sums.
  if(block%sbowner)then

  if(jm > 1) then
      
  do sbj=block%sbj1,block%sbj2
      do sbi=block%sbi1,block%sbi2
          do k=r(5),r(6)
              do j=1,mblszj
                  do i=1,mblszi
                      gi=(sbi-1)*mblszi+i
                      gj=(sbj-1)*mblszj+j                      
                      if(gi>=r(1) .and. gj>=r(3) .and. gi<=r(2) .and. gj<=r(4))then
                          if(direction==2) dir=j
                          if(direction==1) dir=i                          
                          li = gi-block%fi+1
                          lj = gj-block%fj+1
                          cf(sbi,sbj,dir) = cf(sbi,sbj,dir) + f(li,lj,k)
                      end if
                  end do
              end do
          end do
      end do
  end do
  
  else
      sbj=1
      do sbi=block%sbi1,block%sbi2
          do k=r(5),r(6)
              j=1
              do i=1,mblszi
                  gi=(sbi-1)*mblszi+i
                  gj = 1
                  if(gi >= r(1) .and. gi <= r(2))then                  
                      dir = i
                      li = gi-block%fi+1
                      lj = 1
                      cf(sbi,sbj,dir) = cf(sbi,sbj,dir) + f(li,lj,k)
                  end if
              end do
          end do
      end do
  end if

  if(numproc > 1)then
      sendbuf = reshape(cf(block%sbi1:block%sbi2,block%sbj1:block%sbj2,1:kdim), (/sendcount/))
  else
      gcf = cf
  end if
  
  end if

  commtime1 = rdtsc()
  
  if(numproc>1)then
  
      currdisp=0
      do p=1,numproc
          bi=block_of_rank(1,p-1) ; bj=block_of_rank(2,p-1)
          displs(p) = currdisp
          if(bltab(bi,bj)%sbowner)then
              recvcounts(p) = (bltab(bi,bj)%sbi2-bltab(bi,bj)%sbi1+1)*(bltab(bi,bj)%sbj2-bltab(bi,bj)%sbj1+1)*kdim
          else
              recvcounts(p)=0
          end if
          currdisp = currdisp + recvcounts(p)
      end do
      
      call mpi_gatherv(sendbuf, sendcount, mpi_float_type, &
           recvbuf, recvcounts, displs, mpi_float_type, 0, qmpi_comm_world, ierr)
  
      if(master)then      
          do p=1,numproc
              bi=block_of_rank(1,p-1)
              bj=block_of_rank(2,p-1)
              if(bltab(bi,bj)%sbowner)then
                  bi1=bltab(bi,bj)%sbi1 ; bi2=bltab(bi,bj)%sbi2
                  bj1=bltab(bi,bj)%sbj1 ; bj2=bltab(bi,bj)%sbj2
                  d1=displs(p)+1
                  d2=d1+recvcounts(p)-1
                  gcf(bi1:bi2,bj1:bj2,1:kdim) = reshape(recvbuf(d1:d2), (/bi2-bi1+1,bj2-bj1+1,kdim/))
              end if
          end do
      end if
  end if
  
  if(master)then
      vsum = 0.0
      do jj=1,numpbj
          do ii=1,numpbi
              if(direction == 1) then
                  js = (ii-1)*kdim+1
                  je = min(im,js+kdim-1)
              else if(direction == 2) then
                  js = (jj-1)*kdim+1
                  je = min(jm,js+kdim-1)
              end if
!              print *,'hello',js,je,kdim
              vsum(js:je) = vsum(js:je) + gcf(ii,jj,1:je-js+1)
          end do
      end do
  end if

  deallocate(recvbuf,sendbuf,cf,gcf)
  
  if(numproc > 1) call broadcast(vsum)

  commtime2 = rdtsc()
  totsumtime = totsumtime + (commtime2-commtime1)  
end subroutine pvsum



SUBROUTINE PZSUM(F,IM,JM,KB,TZ,ZZ,H,ETA,ZSUM,NVALUES,RANGE)
!
! The sum of F is found in the (SIZE(TZ)-1) layers between the
! vertical positions in TZ.  If the complete range is to be covered,
! TZ(1) must be 0, TZ(size(TZ))=-MAXVAL(H).  The sum from each
! layer will be returned in the array ZSUM of dimension SIZE(TZ), and 
! the number of samples in each layer NVALUES
!
! H is the static depth, ETA the elevation, ZZ is the sigma coordinate
! for box centers (as usual in BOM)
!
! RANGE is an optional global range, over which to compute the sum. if
! not present, the sum will be computed over RANGE=(/2,IM-1,2,JM-1,1,KB-1/)
!  
! Important: the routine assumes F to be FULLY UPDATED in the ghost zone.  
!
! It is advantageous that the grid resolution is finer than the
! distance between bin layers.
!  
! The horizontal summation algorithm is as for PSUM.  
!
  implicit none
  integer im,jm,kb
  real f(i1:i2,j1:j2,kb)
  real tz(:), h(i1:i2,j1:j2), eta(i1:i2,j1:j2), zz(kb)
  integer, optional :: range(6)
  real zsum(size(tz))
  integer nvalues(size(tz))
  
  real, allocatable :: gcf(:,:,:), cf(:,:,:)
  integer i,j,k,ii,jj,bi,bj
  real, allocatable :: sendbuf(:),recvbuf(:)
  integer displs(numproc),recvcounts(numproc)
  integer sendcount, sbi,sbj, gi,gj,li,lj,currdisp,p
  integer d1,d2,bi1,bi2,bj1,bj2
  integer r(6),numbin,bin
  real dep
  
  
  numbin = size(tz)-1
  r = (/2,im-1,2,jm-1,1,max(1,kb-1)/)

  if(present(range)) then
      r = range
      if(r(1) < 1 .or. r(2)>gim .or. r(3)<1 .or. r(4)>gjm .or. r(5)<1 .or. r(6)>kb) then
          if(master)write(*,'(a,i0,":",i0,i0,":",i0,i0,":",i0)') "The provided range for PZSUM is wrong, range=",range
          call abort()
      end if      
  end if
  
  sendcount=(block%sbi2-block%sbi1+1)*(block%sbj2-block%sbj1+1)*numbin
  
  if(.not.block%sbowner) sendcount=0
  
  allocate(cf(block%sbi1:block%sbi2,block%sbj1:block%sbj2,numbin), sendbuf(sendcount))
  
  allocate(recvbuf(numpbi*numpbj*numbin), gcf(numpbi,numpbj,numbin))
  
  gcf(:,:,:)=0.0 
  cf(:,:,:)=0.0
  nvalues(:) = 0
  
  !! on each processor, loop over sub blocks and find sums.
  if(block%sbowner)then

  if(jm > 1) then

! loop over psum blocks present on this processor      
  do sbj=block%sbj1,block%sbj2
     do sbi=block%sbi1,block%sbi2
        ! on each box, loop over the local block range
        do k=r(5),r(6)
           do j=1,mblszj
              do i=1,mblszi
                 ! global location of box
                 gi=(sbi-1)*mblszi+i
                 gj=(sbj-1)*mblszj+j
                 ! if we are in the desired glob range, convert to processor local coordinates
                 if(gi>=r(1) .and. gj>=r(3) .and. gi<=r(2) .and. gj<=r(4))then
                    li = gi-block%fi+1
                    lj = gj-block%fj+1
                    if(h(li,lj) > 0.0) then

                       ! Z depth below 0 level for current k (assuming eta point):
                       dep = zz(k)*(h(li,lj)+eta(li,lj)) + eta(li,lj)

                       ! search for bin containing dep
                       do bin=1,numbin
                          if(dep > tz(bin+1) .and. dep <= tz(bin))then
                             nvalues(bin) = nvalues(bin) + 1
                             cf(sbi,sbj,bin) = cf(sbi,sbj,bin) + f(li,lj,k)
                             exit
                          end if
                       end do
                    end if
                 end if
              end do
           end do
        end do
     end do
  end do
  ! yeah. never listen to so called programming experts claiming 10 nested levels is too much :)

  else

! as for jm>1, but no j loops.

     sbj=1
     do sbi=block%sbi1,block%sbi2
        ! on each box, loop over the local block range
        do k=r(5),r(6)
           j=1
           do i=1,mblszi
              ! global location of box
              gi=(sbi-1)*mblszi+i
              gj=1
              ! if we are in the desired glob range, convert to processor local coordinates
              if(gi>=r(1) .and. gi<=r(2))then
                 li = gi-block%fi+1
                 lj=1
                 if(h(li,lj) > 0.0) then
                    ! depth below 0 level for current k (assuming eta point)
                    dep = zz(k)*(h(li,lj)+eta(li,lj)) + eta(li,lj)
                    ! search for bin containing depth
                    do bin=1,numbin
                       if(dep > tz(bin+1) .and. dep <= tz(bin))then
                          nvalues(bin) = nvalues(bin) + 1
                          cf(sbi,sbj,bin) = cf(sbi,sbj,bin) + f(li,lj,k)
                          exit
                       end if
                    end do
                 end if
              end if
           end do
        end do
     end do
  end if

  if(numproc > 1)then
      sendbuf = reshape(cf(block%sbi1:block%sbi2,block%sbj1:block%sbj2,1:numbin), (/sendcount/))
  else
      gcf = cf
  end if
  
  end if

  commtime1 = rdtsc()
  
  if(numproc>1)then
  
      currdisp=0
      do p=1,numproc
          bi=block_of_rank(1,p-1) ; bj=block_of_rank(2,p-1)
          displs(p) = currdisp
          if(bltab(bi,bj)%sbowner)then
              recvcounts(p) = (bltab(bi,bj)%sbi2-bltab(bi,bj)%sbi1+1)*(bltab(bi,bj)%sbj2-bltab(bi,bj)%sbj1+1)*numbin
          else
              recvcounts(p)=0
          end if
          currdisp = currdisp + recvcounts(p)
      end do
      
      call mpi_gatherv(sendbuf, sendcount, mpi_float_type, &
           recvbuf, recvcounts, displs, mpi_float_type, 0, qmpi_comm_world, ierr)
  
      if(master)then      
          do p=1,numproc
              bi=block_of_rank(1,p-1)
              bj=block_of_rank(2,p-1)
              if(bltab(bi,bj)%sbowner)then
                  bi1=bltab(bi,bj)%sbi1 ; bi2=bltab(bi,bj)%sbi2
                  bj1=bltab(bi,bj)%sbj1 ; bj2=bltab(bi,bj)%sbj2
                  d1=displs(p)+1
                  d2=d1+recvcounts(p)-1
                  gcf(bi1:bi2,bj1:bj2,1:numbin) = reshape(recvbuf(d1:d2), (/bi2-bi1+1,bj2-bj1+1,numbin/))
              end if
          end do
      end if
      
      call reduce('sum', nvalues)
  end if
  
  if(master)then
      do bin=1,numbin
          zsum(bin) = 0.0
          do jj=1,numpbj
              do ii=1,numpbi
                  zsum(bin) = zsum(bin) + gcf(ii,jj,bin)
              end do
          end do
      end do
  end if

  deallocate(recvbuf,sendbuf,cf,gcf)
  
  if(numproc > 1) call broadcast(zsum)

  commtime2 = rdtsc()
  totsumtime = totsumtime + (commtime2-commtime1)
  
end subroutine PZSUM


END MODULE MOD_MPI
