module mod_particle
!
! Module that contains necessary routines for running BOM with particle
! tracking.
!
! Tomas Torsvik <tomas.torsvik@uni.no>, Uni Research
! Helge Avlesen <helge.avlesen@uni.no>, Uni Research
! Oyvind Thiem <oyvind.thiem@uni.no>, Uni Research
!
! Usage:
!   the configuration file "particle_setup.dat" will be assumed to be present in the run directory. 
!
! NOTE:
! - Bilinear interpolation should be used for simulations with
!   rugged bathymetry and coast line. With linear interpolation,
!   particles tend to atrificially adhere to the coast if land 
!   cells are located south or west of the particle location.
!
!  
! The particle tracking model is based on theory presented in:
! [1] A. F. Blumberg, D. J. Dunning, H. Li, D. Heimbuch, and W. R. Geyer:
!     "Use of Particle-tracking Model for Predicting Entrainment at
!      Power Plants on the Hudson River",
!      Estuaries, Vol. 27, No. 3, p. 515-526, June 2004.
! [2] K. N. Dimou and E. E. Adams:
!     "A Random-walk, Particle tracking Model for Well-mixed Estuaries 
!      and Coastal Waters",
!     Estuarine, Coastal and Shelf Research, Vol. 37, p. 99-110, 1993.
! [3] A. F. B. Tompson and L. W. Gelhar:
!     "Numerical simulation of Solute Transport in Three-Dimensional,
!      Randomly Heterogeneous Porous Media",
!     Water Resources Research, Vol. 26, No. 10, p. 2541-2562, Oct. 1990.
!
! The particle tracking routine also contains elements from Bjoern 
! Aadlandsvik's Ladim model, and from Oyvind Thiem's part_tracking 
! code for bom.
!
! Particle tracking in 3D for a sigma (S) coordinate model:
!
!   X(n,t+dt) = X(n,t) + [U + (1/D)*(d(AH*D)/dx)]*dt +
!               sqrt(2*AH*dt)*GAMMA
!   Y(n,t+dt) = Y(n,t) + [V + (1/D)*(d(AH*D)/dy)]*dt +
!               sqrt(2*AH*dt)*GAMMA
!   S(n,t+dt) = S(n,t) + [(w/D) + (1/D)*(d(KH/D)/ds)]*dt +
!               (1/D)*sqrt(2*KH*dt)*GAMMA
!
! where (X,Y,S) are sigma coordinates, D is the dynamical depth (h+eta), 
! (U,V) are horizontal velocities, w is the velocity normal to a 
! sigma level, and GAMMA is a random number with zero mean and
! unit variance.
!
! Non-dimensinal index variables (X/dx,Y/dy,S) are used to keep track
! of the particles internally in the module. 
! get_sigma_point  : Convert from physical coordinates to internal
!                    index coordiantes
! get_xyz_point    : Convert from internal index coordinates to
!                    physical coordinates
!
!
! Routines to be called from external routines (user_init,user_step):
! -----------------------------------------------------------------
! setup_particle   : standard setup for particle tracking
! part_initialize  : used to initiate particle tracking routine
! move_particle    : standard routine for moving particles
! part_step        : advection and turbulent dispersion of particles
! part_output      : write output to stdout and to files
! clear_particle   : clear particle arrays, and reset default values
! part_checkpoint  : dump particle data to files
!
! Comment on routines:
! For normal usage, only the standar routines SETUP_PARTICLE and 
! MOVE_PARTICLE need to be called from USER_INIT and USER_STEP,
! respectively. The other routines are available for external 
! routines that build on the particle tracking framework.
!
!
! Keywords:
! - Logical variables are false by default
! - Integer type switches are lowest possible value by default
!
! keyword            value  description
! -----------------------------------------------------------------
! part_activate      <t/f>  Turn particle tracking on/off
! part_restart       <t/f>  True if restarting from a checkpoint
! part_random_step   <t/f>  Random perturbation in advection 
!                           of particles 
! part_random_seed   <t/f>  Set new seed for pseudorandom
!                           number generator
! part_random_type   <int>  Type of random walk model
!                           0=Random walk with horizontal and 
!                             vertical diffusivity from BOM
!                           1=Random walk with constant horizontal 
!                             and vertical diffusivity
! part_ah            <real> Constant horizontal diffusivity
! part_kh            <real> Constant vertical diffusivity 
! part_runge_kutta4  <t/f>  Use trapeziodal method in calculation 
!                           of advection
! part_bilinear      <t/f>  Use bilinear method in calculation 
!                           of advection
! part_h_is_positive <t/f>  Particle depth specified as positive 
!                           numbers
! part_vert_velocity <int>  Type of vertical motion for particles
!                           0=No vertical velocity (passively drifting)
!                           1=Read vertical velocity from 5th 
!                             column in particle init file
!                           2=Read particle class from 5th column, use this to compute velocity
! part_lookup_ws     <t/f>  If true, read particle%ws from table
! part_lookup_sg_and_diameter  <t/f>  If true, read time dependent sg and dia from table
! part_apply_stokes  <t/f>  if true, use the properties given via particle class to compute vert. velocity
! part_add_class <name> <material number> <datafile> <numprop>
!    define time dependent properties for a class of particles
!      <name>              user specified name for the class
!      <material number>   user specified, unique, integer id for the class
!      <datafile>          ascii file, where each line has a time (the particle age) and value
!                          of properties for the given time. properties are interpolated in time
!      <numprop>           the number of properties given at each line in the above file  
! part_capture_land  <t/f>  Particles captured when hitting land
! part_capture_h     <t/f>  Particles captured when hitting bottom
! part_capture_eta   <t/f>  Particles captured when hitting surface
! part_minh          <real> Min distance the particle can be above 
!                           the bottom
! part_mineta        <real> Min distance the particle can be close 
!                           to eta
! part_init_minh     <real> Min initial distance the particle can 
!                           be above the bottom
! part_init_mineta   <real> Min initial distance the particle can
!                           be close to eta
! part_minland       <real> Distance to bump back particles that
!                           enter land cells.
! part_init_data     <int>  <string>  <int>
!                           column 2: 1=data from ascii file
!                                     2=data from binary file
!                                     3=data from binary file,
!                                       stream access
!                           column 3: filename
!                           column 4: number of particles
! part_exchange_size <int>  Max number of particles that can be 
!                           exchanged between subdomains
! part_bound_east    <int>  <real>
!                           Particle boundary condition (east)
!                           column 2: 1=capture particles 
!                                     2=bump back particles 
!                                     3=periodic boundary cond.
!                           column 3: Offset distance (index space)
!                                     between BOM and particle
!                                     boundaries. 
! part_bound_west    <int>  <real>
!                           Particle boundary condition (west)
! part_bound_north   <int>  <real>
!                           Particle boundary condition (north)
! part_bound_south   <int>  <real>
!                           Particle boundary condition (south)
! part_check_status  <t/f>  Check particle status in output routine
! part_enc_dumpint   <int>  How often to dump statistics for captured particles, defaults to twice per day.
! part_outputlevel   <int>  How much writing to stdout:
!                           0=silent 
!                           1=normal 
!                           2=debug
!                           3=detailed debug (not well suited for high number of procs!)  
! part_write_time    <int>  Write output at timestep intervals 
!                           part_write_time
! partplot_binary    <t/f>  Dump distribution and timeseries as binary stream files  
! partplot_timeseries 
!                    <t/f>  Write time series of tracks to file
! partplot_ts_mod    <int>  Only write timeseries for each partplot_ts_mod'th particle
!
! partplot_ts_mod_bsize  <int>   instead of writing timeseries for every partplot_ts_mod'th particle, 
!                                write batches of partplot_ts_mod_bsize particles every 
!                                partplot_ts_mod*partplot_ts_mod_bsize particle. the reduced number of 
!                                timeseries written is the same, just a different way of selecting which 
!                                particles are dumped. default value is 1.
!  
! partplot_ts_write  <int>  Write time series output at timestep 
!                           intervals partplot_ts_write
! partplot_distribution     Write distribution of all active 
!                    <t/f>  particles to file
! partplot_dist_write       Write distribution output at timestep
!                    <int>  intervals partplot_dist_write
! part_use_pgplot    <t/f>  <int>
!                           column 2: Plot particles with pgplot
!                           column 3: Number of plots to be made 
!                                     by pgplot (default = -1)
! pgpart_plot_time   <int>  Time for updating fields used in pgplot
! pgpart_ntime_pts   <int>  Number of time levels in history record
!                           for particle positions
! pgpart_set_plot    <int>  <int>  <int>  <int>  <int>
!                           Define parameters for pgpart_object:
!                           column 2: ind
!                           column 3: type
!                           column 4: style
!                           column 5: color
!                           column 6: line width
! part_boxlogger <name> <x range> <y range> <z range>    
!    log to file, filename constructed from name, whenever a particle is in the provided range
!  
! part_boxlogger_interval   <int>   number of timesteps between box checks. 50 is the default. advection
!                           speed rarely exceeds 1m/s, box size typically tens of meters.
!  
!========================================================================
  use state, only : debug,dp,istep,ndivis,numstep,dt,dx,dy,u,v,w,im,jm,kb,h,d,eta,etap,z,dz,grav,rho0,depmax,depmin,fsm,ah,kh
  use qmpi, only : myid,master,numproc,message,reduce,abort,barrier,communicator_object,mpi_proc_null, &
       broadcast,send,receive,isend,irecv,waitall,mbroadcast,reduce
  use mod_mpi, only : block,bltab,i1,i2,j1,j2,i_blocksize,j_blocksize,globalind2local,gi2block,gj2block, ronblock
  use control_parameters, only : particlets_dir, particledist_dir, particlestat_dir
  
  implicit none

  !!---------------------------------------------------------------------
  !! General variables for particle advection, dispersion and capture.
  !!---------------------------------------------------------------------
  ! Turn particle tracking on/off
  logical :: part_activate = .false.
  ! True if restarting simulation from a checkpoint
  logical :: part_restart = .false.
  ! Random perturbation in advection of particles 
  logical :: part_random_step = .false.
  ! Set new seed for pseudorandom number generator
  logical :: part_random_seed = .false.
  ! Type of random walk model
  !   0 : Random walk with default horizontal and vertical diffusivity
  !   1 : Random walk with constant horizontal and vertical diffusivity
  integer :: part_random_type = 0
  ! Constant horizontal and vertical diffusivity
  real(dp) :: part_ah = 0.
  real(dp) :: part_kh = 0.
  !
  ! get ws from file or compute from properties like sg,dia ... 
  logical :: part_lookup_ws=.false.
  logical :: part_lookup_sg_and_diameter=.false.
  !
  ! use stokes law to compute terminal velocity for a particle given sg,D,...
  logical :: part_apply_stokes=.false.
  !logical :: part_apply_sealice=.false.
  ! 
  ! Use trapeziodal method in calculation of advection
  logical :: part_runge_kutta4 = .false.
  ! Use bilinear method in calculation of advection
  logical :: part_bilinear = .false.
  ! Particle depth is specified as positive numbers
  logical :: part_h_is_positive = .false.
  ! Type of vertical velocity
  !   0 : No vertical velocity
  !   1 : Read vertical velocity from 5th column in particle init file 
  integer :: part_vert_velocity = 0
  ! Particles captured when hitting land
  logical :: part_capture_land = .false.
  ! Particles captured when hitting bottom
  logical :: part_capture_h = .false.
  ! Particles captured when hitting surface
  logical :: part_capture_eta = .false.
  ! Total number of particles defined in particle_init.dat
  integer :: part_total = 0
  ! Min dist the particle can be above the bottom
  real(dp) :: part_minh = -1.
  ! Min dist the particle can be close to eta
  real(dp) :: part_mineta = -1.
  ! Min initial dist the particle can be above the bottom
  real(dp) :: part_init_minh = -1.
  ! Min initial dist the particle can be close to eta
  real(dp) :: part_init_mineta = -1.
  ! Distance to bump back particles that enter land cells.
  real(dp) :: part_minland = 0.5

  !!---------------------------------------------------------------------
  !! Boundary conditions for particles
  !! part_bound_<edge>_type
  !!   1 : capture particles at boundaries
  !!   2 : reflective boundary conditions
  !!   3 : periodic boundary conditions
  !! part_bound_offset_<edge>
  !!   offset between BOM and particle boundaries
  !! part_bound_loc_<edge>
  !!   particle boundary location on block
  !!---------------------------------------------------------------------
  integer :: part_bound_east_type = 1
  integer :: part_bound_west_type = 1
  integer :: part_bound_north_type = 1
  integer :: part_bound_south_type = 1
  real(dp) :: part_bound_offset_east = 1.0
  real(dp) :: part_bound_offset_west = 1.0
  real(dp) :: part_bound_offset_north = 1.0
  real(dp) :: part_bound_offset_south = 1.0
  real(dp) :: part_bound_loc_east, part_bound_loc_west
  real(dp) :: part_bound_loc_north, part_bound_loc_south

  !!---------------------------------------------------------------------
  !! Variables for writing output to stdout and to files
  !!---------------------------------------------------------------------
  ! Check particle status in output routine
  logical :: part_check_status = .false. 
  ! How much writing to stdout: 0=silent; 1=normal; 2=debug
  integer :: part_outputlevel = 0
  ! Write time series of tracks to file
  logical :: partplot_timeseries = .false.
  ! Write time series output at intervals partplot_ts_write
  integer :: partplot_ts_write = 1
  ! Write distribution of all active particles to file
  logical :: partplot_distribution = .false.
  ! Write distribution output at intervals partplot_dist_write
  integer :: partplot_dist_write = 1
  ! Write output at intervals part_write_time
  integer :: part_write_time = 1
  ! Write timeseries of particle mod(p,partplot_ts_mod)
  integer :: partplot_ts_mod = -1
  integer :: partplot_ts_mod_bsize = 1
  ! Write counts of captured particles every mod(istep,part_enc_dumpint) step
  integer :: part_enc_dumpint = -1
  ! Write timeseries and distribution to binary files
  logical :: partplot_binary=.false.
  
  !!---------------------------------------------------------------------
  !! Variables for plotting particles with pgplot
  !!---------------------------------------------------------------------
  !!-- Variables to be aset by user
  ! Plot particles with pgplot
  logical :: part_use_pgplot = .false.
  ! Number of plots to be made by pgplot 
  integer :: pgpart_num_plots = -1
  ! Time for updating fields used in pgplot
  integer :: pgpart_plot_time = 1
  ! Number of time levels in history record for particle positions
  integer :: pgpart_ntime_pts = 1
  ! assume particles read in are already sorted if this is true:
  logical :: part_skip_sort = .false.
  
  !!-- Internal pgpart routines
  ! Counter for pgparts objects
  integer :: pgparts_count = 0
  ! Array for storing history of (x,y,z)-coordinates for all particles
  real(dp), allocatable, dimension(:,:,:) :: pgpart_history
  logical :: pgpart_set_history = .false.
  integer :: pgpart_history_count = 0
  ! Array for storing capture coordinates for all particles
  real(dp), allocatable, dimension(:,:) :: pgpart_capture
  logical :: pgpart_set_capture = .false.
  integer :: pgpart_capture_count = 0
  ! Array of pgplot colors
  integer, allocatable, dimension(:) :: pgpart_color_array

  type pgpart_object
  ! Object for plotting particles with pgplot:
  ! set        : use this object if (set == true)
  ! ind        : Index in pgplots array
  ! type       : 1 : Plot active particles with scatter method
  !              2 : Plot active particles with pldj
  !              3 : Plot captured particles with scatter method
  !              4 : Scatter plot of particles projected in the y direction onto the first section of a vsect
  !              5 : Scatter plot of particles projected in the x direction onto the first section of a vsect
  !              6 : Trajectory of particles projected in the y direction onto the first section of a vsect
  ! style      : type 1,3 : symbol style (-8 -- 31)
  !              type 2   : line style   (1 -- 5)
  ! color      :   -1 : use color_array with gray scale
  !              0-15 : Standard pgplot color styles
  ! line_width : pgplot line width in range (1 -- 201)
     logical :: set = .false.
     integer :: ind
     integer :: type
     integer :: style = 1
     integer :: color = 6
     integer :: line_width = 4
     ! Arrays for sending particle position data to pgplot routines
     real, allocatable, dimension(:) :: x1,x2,y1,y2,z1,z2
  end type pgpart_object
  type(pgpart_object), allocatable :: pgparts(:)


!! array of objects used to register location of particles. each time
!! a particle is registred in one of these boxes, the time is logged
!! to file.  it is recommended to use thin boxes, perhaps 0.5dx thick,
!! to avoid logging too much data.
  
  type particleboxobject
     character(80) :: name=""
     logical :: inblock=.false.   ! true if part of region is on processor
     real :: r(6)                 ! the hit range (recomputed to grid coordinates during setup)
  end type particleboxobject
 
  integer :: numpbox=0, maxnumpbox=50
  integer :: boxloggerintval=3    ! must be small enough to never allow a particle to pass a box unnoticed
  type(particleboxobject), allocatable :: pbox(:)


  !!---------------------------------------------------------------------
  !! 2D array for initial particle property, the shape is (part_total,5)
  !! The 5 numbers are <activation time> <x> <y> <z> <sinking velocity.(positive=rising!)>
  real(dp), dimension (:,:), allocatable :: part_init_data
  
  !! part_init_file      : Name of particle init file
  !! part_init_file_type : 1 = ascii, 2 = binary
  !! Default: Data read from ascii file particle_init.dat
  !!---------------------------------------------------------------------
  character(len=80) :: part_init_file='./particle_init.dat'
  integer :: part_init_file_type = 1

  type particletype
  ! Particle object:
  ! id    : a unique paricle id number
  ! status: myid : Particle active, and belongs to myid
  !           -1 : Particle not yet activated
  !           -2 : Particle activated, but belongs to other block
  !          -10 : Particle inactive, captured by bottom
  !          -20 : Particle inactive, captured at surface
  !          -30 : Particle inactive, captured by land
  !          -40 : Particle inactive, captured at boundary
  !          -50 : Particle inactive, initiation error
  ! block : subregion block on which the particle is located
  ! time  : if status = -1 : time when particle will be activated
  !         if status < -1 : time when particle was captured
  ! pos   : location of particle 
  !            if status  = -1 : in xyz coordinates
  !            if status != -1 : in sigma coordinates
  ! ws    : vertical velocity in m/s : < 0.0 sinking 
  !                                  : > 0.0 going up 
  ! class:  type of particle. used to read properties from file. if class=0, ws is set to 0.
  ! d    :  diameter in meters, used to compute ws
  ! sg   :  specific gravity wrt rho0.
  ! box  :  bit 1-64 is set if particle is in the box
     integer :: id
     integer :: status
     integer :: block
     integer :: class=0
     real(dp) :: time
     real(dp), dimension(3) :: pos
     real :: ws
     real :: d
     real :: sg
     integer(dp) box
  end type particletype

  type(particletype), allocatable:: particle(:)
  integer :: particle_charsize=0  ! the size of a particle object measured as characters


! object for keeping lookup table data for particle material properties
  type matter_type
     character(80) :: name="materialX", file=""
     integer :: numprop=0
     integer :: n=0
     real, allocatable :: time(:), prop1(:), prop2(:), prop3(:), prop4(:)
  end type matter_type

  integer, parameter :: part_maxnumclasses=50
  type(matter_type) matter(part_maxnumclasses)
  integer :: matnumbers(part_maxnumclasses)
  integer :: part_numclasses=0
  
  ! Number of activated particles. index_activated <= part_total
  integer :: index_activated

  ! If needed, used by master to keep track of particle status on all procs
  integer, dimension(:,:), allocatable :: status_data

  ! Exchange particles between subdomains
  type(particletype), allocatable:: put_north(:), put_east(:), put_south(:), put_west(:)
  type(particletype), allocatable:: get_north(:), get_east(:), get_south(:), get_west(:)
  type(communicator_object),save :: part_comm_ew, part_comm_ns
  integer :: iput_east, iput_west, iput_north, iput_south 
  integer :: iget_east, iget_west, iget_north, iget_south
  integer :: eastblock, westblock, northblock, southblock
  integer :: part_exchange_size = -1
  INTEGER, PARAMETER :: MIN_EXCHANGE_SIZE = 10000
  integer :: maxput=0   ! the current maximum number of particles exchanged with another processor
  character, allocatable :: part_buffer_send1(:), part_buffer_send2(:)
  character, allocatable :: part_buffer_recv1(:), part_buffer_recv2(:)
  integer :: part_buffer_size

  character(len=80) :: part_workdir='./'

  integer :: numput=0
  
contains


!========================================================================
! --- setup
!
! Subroutines for setup of the particle tracking model
!
!========================================================================


  subroutine setup_particle(start_time)
  !
  ! Standard setup for particle tracking
  !
    implicit none
    real(dp), optional, intent(in) :: start_time

    if (present(start_time)) then
       call part_initialize(start_time)
    else
       call part_initialize(0.0_8)
    end if
    call part_output

  end subroutine setup_particle


  subroutine part_initialize(start_time)
  !
  ! Routine for setting up particles and reading particle_init.dat
  ! 
    use util
    use pgplot_core
    
    implicit none

    real(dp), intent(in) :: start_time

    integer, parameter :: max_active_lines=100
    character(200) :: lines(max_active_lines)=" "
    integer :: srcline(max_active_lines)
    character(132) particle_splitline(50)
    integer :: ierr=0, kw, keywords,i,j

    logical :: file_is_there
    integer :: loop_rel
    integer, dimension(:), allocatable :: init_status
    real(dp), dimension(3) :: xyz,sig
    integer :: root

    if(master) inquire(file="particle_setup.dat", exist=file_is_there)

    call broadcast(file_is_there)
    
    if(file_is_there)then
        if(master) call get_lines("particle_setup.dat", keywords, lines, srcline)
    else
        call message("particle_setup.dat not found, particle tracking will not be activated!", 0)
        return
    end if


    matnumbers = -1
    
    
    call broadcast(keywords)
    call broadcast(lines)
    call broadcast(srcline)

    do kw=1,keywords
       call split(lines(kw), particle_splitline)

       select case(lower(trim(particle_splitline(1))))          
       case('part_activate')
          read(particle_splitline(2),*,err=101,end=101) part_activate
          if(master) write(*,*) 'part_activate: ',part_activate

       case('part_init_data')
           call read_particles(particle_splitline)
           
       case('part_restart')
          read(particle_splitline(2),*,err=101,end=101) part_restart
          if(master) write(*,*) 'part_restart: ',part_restart

       case('part_skip_sort')
          read(particle_splitline(2),*,err=101,end=101) part_skip_sort
          if(master) write(*,*) 'part_skip_sort: ',part_skip_sort
          
       case('part_random_step')
          read(particle_splitline(2),*,err=101,end=101) part_random_step
          if(master) write(*,*) 'part_random_step: ',part_random_step

       case('part_random_seed')
          read(particle_splitline(2),*,err=101,end=101) part_random_seed
          if(master) write(*,*) 'part_random_seed: ',part_random_seed

       case('part_random_type')
          read(particle_splitline(2),*,err=103,end=103) part_random_type
          if(master) write(*,*) 'part_random_type: ',part_random_type

       case('part_ah')
          read(particle_splitline(2),*,err=102,end=102) part_ah
          if(master) write(*,*) 'part_ah: ',part_ah

       case('part_kh')
          read(particle_splitline(2),*,err=102,end=102) part_kh
          if(master) write(*,*) 'part_kh: ',part_kh

       case('part_lookup_ws')
          read(particle_splitline(2),*,err=101,end=101) part_lookup_ws
          if(master) write(*,*) 'part_lookup_ws: ',part_lookup_ws
          
       case('part_lookup_sg_and_diameter')
          read(particle_splitline(2),*,err=101,end=101) part_lookup_sg_and_diameter
          if(master) write(*,*) 'part_lookup_sg_and_diameter: ',part_lookup_sg_and_diameter
          
       case('part_apply_stokes')
          read(particle_splitline(2),*,err=101,end=101) part_apply_stokes
          if(master) write(*,*) 'part_apply_stokes: ',part_apply_stokes
          
       case('part_bilinear')
          read(particle_splitline(2),*,err=101,end=101) part_bilinear
          if(master) write(*,*) 'part_bilinear: ',part_bilinear

       case('part_runge_kutta4')
          read(particle_splitline(2),*,err=101,end=101) part_runge_kutta4
          if(master) write(*,*) 'part_runge_kutta4: ',part_runge_kutta4

       case('part_h_is_positive')
          read(particle_splitline(2),*,err=101,end=101) part_h_is_positive
          if(master) write(*,*) 'part_h_is_positive: ',part_h_is_positive

       case('part_vert_velocity')
          read(particle_splitline(2),*,err=103,end=103) part_vert_velocity
          if(master) write(*,*) 'part_vert_velocity: ',part_vert_velocity

       case('part_capture_land')
          read(particle_splitline(2),*,err=101,end=101) part_capture_land
          if(master) write(*,*) 'part_capture_land: ',part_capture_land

       case('part_capture_h')
          read(particle_splitline(2),*,err=101,end=101) part_capture_h
          if(master) write(*,*) 'part_capture_h: ',part_capture_h

       case('part_capture_eta')
          read(particle_splitline(2),*,err=101,end=101) part_capture_eta
          if(master) write(*,*) 'part_capture_eta: ',part_capture_eta

       case('part_minh')
          read(particle_splitline(2),*,err=102,end=102) part_minh
          if(master) write(*,*) 'part_minh: ',part_minh

       case('part_mineta')
          read(particle_splitline(2),*,err=102,end=102) part_mineta
          if(master) write(*,*) 'part_mineta: ',part_mineta

       case('part_init_minh')
          read(particle_splitline(2),*,err=102,end=102) part_init_minh
          if(master) write(*,*) 'part_init_minh: ',part_init_minh

       case('part_init_mineta')
          read(particle_splitline(2),*,err=102,end=102) part_init_mineta
          if(master) write(*,*) 'part_init_mineta: ',part_init_mineta

       case('part_minland')
          read(particle_splitline(2),*,err=102,end=102) part_minland
          if(master) write(*,*) 'part_minland: ',part_minland

       case('part_exchange_size')
          read(particle_splitline(2),*,err=102,end=102) part_exchange_size
          if(master) write(*,*) 'part_exchange_size: ',part_exchange_size


      case('part_add_class')
          call add_class(particle_splitline)
          
       !!----------------------------------------------------------------------- 
       !! Boundary conditions
       !!----------------------------------------------------------------------- 

       case('part_bound_east')
          read(particle_splitline(2),*,err=103,end=103) part_bound_east_type
          read(particle_splitline(3),*,err=102,end=102) part_bound_offset_east
          if(master) write(*,*) 'part_bound_east: ',part_bound_east_type,  &
               part_bound_offset_east

       case('part_bound_west')
          read(particle_splitline(2),*,err=103,end=103) part_bound_west_type
          read(particle_splitline(3),*,err=102,end=102) part_bound_offset_west
          if(master) write(*,*) 'part_bound_west: ',part_bound_west_type,  &
               part_bound_offset_west

       case('part_bound_north')
          read(particle_splitline(2),*,err=103,end=103) part_bound_north_type
          read(particle_splitline(3),*,err=102,end=102) part_bound_offset_north
          if(master) write(*,*) 'part_bound_north: ',part_bound_north_type,  &
               part_bound_offset_north

       case('part_bound_south')
          read(particle_splitline(2),*,err=103,end=103) part_bound_south_type
          read(particle_splitline(3),*,err=102,end=102) part_bound_offset_south
          if(master) write(*,*) 'part_bound_south: ',part_bound_south_type,  &
               part_bound_offset_south

       !!----------------------------------------------------------------------- 
       !! Variables for output
       !!----------------------------------------------------------------------- 

       case('part_check_status')
          read(particle_splitline(2),*,err=101,end=101) part_check_status
          if(master) write(*,*) 'part_check_status: ',part_check_status

       case('part_outputlevel')
          read(particle_splitline(2),*,err=103,end=103) part_outputlevel
          if(master) write(*,*) 'part_outputlevel: ',part_outputlevel

      case('partplot_binary')
          read(particle_splitline(2),*,err=101,end=101) partplot_binary
          if(master) write(*,*) 'partplot_binary: ',partplot_binary
          
       case('partplot_timeseries')
          read(particle_splitline(2),*,err=101,end=101) partplot_timeseries
          if(master) write(*,*) 'partplot_timeseries: ',partplot_timeseries

       case('partplot_ts_write')
          read(particle_splitline(2),*,err=103,end=103) partplot_ts_write
          if(master) write(*,*) 'partplot_ts_write: ',partplot_ts_write

       case('partplot_distribution')
          read(particle_splitline(2),*,err=101,end=101) partplot_distribution
          if(master) write(*,*) 'partplot_distribution: ',partplot_distribution

       case('partplot_dist_write')
          read(particle_splitline(2),*,err=103,end=103) partplot_dist_write
          if(master) write(*,*) 'partplot_dist_write: ',partplot_dist_write

       case('part_write_time')
          read(particle_splitline(2),*,err=103,end=103) part_write_time
          if(master) write(*,*) 'part_write_time: ',part_write_time

       case('partplot_ts_mod')
          read(particle_splitline(2),*,err=103,end=103) partplot_ts_mod
          if(master) write(*,*) 'partplot_ts_mod: ',partplot_ts_mod

      case('partplot_ts_mod_bsize')
          read(particle_splitline(2),*,err=103,end=103) partplot_ts_mod_bsize

          if(partplot_ts_mod_bsize < 1 .or. partplot_ts_mod_bsize > partplot_ts_mod) then
              if(master) write(*,*) "confused. got partplot_ts_mod_bsize=",partplot_ts_mod_bsize
              if(master) write(*,*) "reverting it to partplot_ts_mod_bsize=1"
              partplot_ts_mod_bsize = 1
          end if          
          if(master) write(*,*) 'partplot_ts_mod_bsize: ',partplot_ts_mod_bsize

      case('part_enc_dumpint')
          read(particle_splitline(2),*,err=103,end=103) part_enc_dumpint
          if(master) write(*,*) 'part_enc_dumpint: ',part_enc_dumpint
          
      case('part_boxlogger')
          call add_boxlogger(particle_splitline)
      case('part_boxlogger_interval')
          read(particle_splitline(2),*,err=103,end=103) boxloggerintval
          if(master) write(*,*) 'part_boxlogger_interval: ',boxloggerintval

          
       !!----------------------------------------------------------------------- 
       !! Variables for plotting with pgplot
       !!----------------------------------------------------------------------- 

       case('part_use_pgplot')
          read(particle_splitline(2),*,err=101,end=101) part_use_pgplot
          read(particle_splitline(3),*,err=103,end=103) pgpart_num_plots
          if (pgpart_num_plots < 1) pgpart_num_plots = pgplot_maxplots
          if(master) then
             allocate (pgparts(pgpart_num_plots))
             write(*,*) 'part_use_pgplot: ',  &
                  part_use_pgplot, pgpart_num_plots
          end if

       case('pgpart_plot_time')
          read(particle_splitline(2),*,err=103,end=103) pgpart_plot_time
          if(master) write(*,*) 'pgpart_plot_time: ',pgpart_plot_time

       case('pgpart_ntime_pts')
          read(particle_splitline(2),*,err=103,end=103) pgpart_ntime_pts
          if(master) write(*,*) 'pgpart_ntime_pts: ',pgpart_ntime_pts

       case('pgpart_set_plot')
          pgparts_count = pgparts_count + 1
          if (.not.part_use_pgplot) then
             call abort('ERROR: part_use_pgplot is false!')
          else if (pgparts_count > pgpart_num_plots) then
             call abort('ERROR: pgparts_count > pgpart_num_plots')
          end if

          if(master) then
             read(particle_splitline(2),*,err=103,end=103) pgparts(pgparts_count)%ind
             read(particle_splitline(3),*,err=103,end=103) pgparts(pgparts_count)%type
             read(particle_splitline(4),*,err=103,end=103) pgparts(pgparts_count)%style
             read(particle_splitline(5),*,err=103,end=103) pgparts(pgparts_count)%color
             read(particle_splitline(6),*,err=103,end=103) pgparts(pgparts_count)%line_width
             pgparts(pgparts_count)%set = .true.
          end if
           
       case default
          if(master) write(*,*) 'Not a valid keyword: ',  &
               lower(trim(particle_splitline(1)))
       end select
    end do
    
    if (.not.part_activate) return

    if(part_total.le.0) then
       if(master) write(*,*) 'WARNING: part_total is too small.'
       part_activate=.false.
       return
    else if(.not.allocated(part_init_data)) then
       if(master) write(*,*) 'WARNING: part_init_data not allocated.'
       part_activate=.false.
       return
    end if

    if(partplot_timeseries .and. part_total > 1000)then
        if(partplot_ts_mod > 0) then
            if(master .and. part_total/partplot_ts_mod > 1000) then
                write(*,*) "WARNING: there may be more than 1000 particle timeseries, &
                     &it is strongly recommended to increase partplot_ts_mod"
            end if
        end if
    end if
    
! Check parameters, and set default values

    if(part_enc_dumpint < 0) part_enc_dumpint = ndivis*12
    
    if(master) call part_parameter_sanity_check
    call mbroadcast(part_minh,part_mineta)
    call mbroadcast(part_init_minh,part_init_mineta)

    ! Set up boundary conditions
    call part_set_boundary()

    if (.not.part_restart) then
       ! Check that the initial data looks reasonable
       allocate(init_status(part_total))
       if(master) call particle_sanity_check(init_status)
       call broadcast(init_status)
    end if

    ! Set new seed for pseudorandom number generator
    if(part_random_seed) call part_set_random_seed()

    ! Allocate particle exchange arrays
    call allocate_exchange_arrays()

    allocate(particle(part_total))

    particle_charsize = size(transfer(particle(1),(/"c"/)))
    
    if (part_restart) then
       call restart_particles
    else
       ! Initiate particle array
       do loop_rel = 1,part_total
          particle(loop_rel)%time = part_init_data(loop_rel,1)
          particle(loop_rel)%pos(1) = part_init_data(loop_rel,2)
          particle(loop_rel)%pos(2) = part_init_data(loop_rel,3)
          particle(loop_rel)%pos(3) = part_init_data(loop_rel,4)
          if(part_vert_velocity==1) then
              particle(loop_rel)%ws = part_init_data(loop_rel,5)
          end if
          if(part_vert_velocity==2) then  ! interpret 5th col as integer for class
              particle(loop_rel)%class = nint(part_init_data(loop_rel,5))
          end if
          particle(loop_rel)%id = loop_rel
          particle(loop_rel)%status = init_status(loop_rel)
       end do
       
       ! Set up pgpart structures
       if(part_use_pgplot) call pgpart_setup()

       ! Activate particles for t=<start_time>
       index_activated = 0
       call activate_particles(start_time)
    end if
    
    if (allocated(part_init_data)) deallocate(part_init_data)
    if (allocated(init_status))    deallocate(init_status)

    call open_boxlogger_files()
    
    ! check class of all particles:
    ! can be time consuming so only do this for high debug level:
    if(part_vert_velocity == 2 .and. debug > 2) then
        do i=1,part_total
            if(.not. any(matnumbers(1:part_numclasses) == particle(i)%class)) then
                call message("invalid class for particle; particle no, class=",2,ival=(/i,particle(i)%class/))
            end if
        end do
        write(*,*) " no unknown material id's"
    end if

    if(part_apply_stokes .and. .not. part_lookup_sg_and_diameter) &
         call message("part_apply_stokes for the time requires part_lookup_sg_and_diameter=.true.",1)
    
    if(part_apply_stokes .and. part_vert_velocity.ne.2) &
         call message("part_apply_stokes for the time requires part_vert_velocity=2",1)
    
    return

101 if(master) call message('could not read logical from column 2', 2)
102 if(master) call message('could not read real from column 2', 2)
103 if(master) call message('could not read integer from column 2', 2)
    
  end subroutine part_initialize


  subroutine read_particles(splitline)
    
! TODO this routine depends on a few keywords be set first (skip_sort, vert_vel, ...) should check all of those!!
    
    implicit none
    character(*) splitline(:)
    logical :: read_error=.false.
    integer i,j
    
    read(splitline(2),*,err=103,end=103) part_init_file_type
    part_init_file = trim(splitline(3))
    
    if(master) write(*,'(1x,3a,i0)') 'read_particles: file= >', trim(part_init_file), '< type=', part_init_file_type

! try to read part_total from the file name, fail silently
! the file name is supposed to have this format;  ...-NNNNx5.... a number surrounded by a "-" and "x5":

    part_total = -1          
    i = index(trim(part_init_file), "-", back=.true.)
    j = index(trim(part_init_file), "x5", back=.true.)

    if(i > 0 .and. j > 0 .and. i < j)then
        read(part_init_file(i+1:j-1),*,err=106,end=106) part_total
    end if

! if the number of particles was not in filename, or a number was given after the filename, use that
    
106 if(part_total < 0 .or. trim(splitline(4)).ne."") then
        if(master) write(*,*) "could not get the number of particles from filename, trying column 4"
        read(splitline(4),*,err=105,end=105) part_total
    end if

    if (.not.part_activate) then
        if(master) write(*,*) 'INFO: part_activate is false. Will not read particle data.'
        return
    end if

    if(part_total < 0) then
        if(master) write(*,*) 'ERROR: could not get the number of particles, deactivating particle tracking'
        part_activate = .false.
        return
    end if

    if(master) write(*,'(2x,a,i0,a)') "Reading ", part_total," particles from file"

    allocate(part_init_data(part_total,5))

! Read particle init file
    if(master) then
        select case(part_init_file_type)
        case(1)
            call read_part_ascii(trim(part_init_file),read_error)
        case(2, 3)
            call read_part_binary(trim(part_init_file),read_error)
        case default
            write(*,*) 'ERROR: part_init_file_type not a valid type, deactivating particle tracking'
            part_activate=.false.
            return
        end select

        if(read_error) then
            write(*,*) 'ERROR: error during file reading, deactivating particle tracking'
            part_activate=.false.
            return
        end if

! Sort particle array according to release time
        if (.not.part_restart .and. .not. part_skip_sort) call sort_particle_array()
    end if

! Send particle data to all processes
    call broadcast(part_activate)
    call broadcast(part_init_data)
    return
    
103 if(master) call message('could not read integer from column 2', 1, margin=1) ; goto 107
104 if(master) call message('could not read string from column 3', 1, margin=1)  ; goto 107
105 if(master) call message('could not read number of particles from column 4', 1, margin=1)
    
107 if(master) write(*,*) 'disabling particle tracking'
    part_activate=.false.
  end subroutine read_particles
        
  

  subroutine read_part_ascii(file,read_error)
  !
  ! Read init data from ascii file
  !
    implicit none
    character(*) file
    logical :: read_error
    logical is_here
    character(7) read_status
    integer :: i
    
    inquire(file=file, exist=is_here, read=read_status)

    if(.not.is_here) then
        if(master) write(*,*)' file name=',file, ' does not exist'
        read_error = .true.
        return
    end if

    if(.not.(read_status == "YES") .and. .not.(read_status == "UNKNOWN")) then
        if(master) write(*,*)'read status=', read_status
        if(master) write(*,*) 'Unable to read file'
        read_error = .true.
        return
    end if
 
    if(master) then
       open(1213,file=file,status='old')
       do i=1,part_total
           read(1213,*,err=101,end=101) part_init_data(i,1:5)
       end do
       close(1213)
    end if
    return
101 call message("failed reading particle data. expected number of cols, lines, filename:",2,ival=(/5,part_total/),sval=(/trim(file)/))
  end subroutine read_part_ascii


  subroutine read_part_binary(file,read_error)
  !
  ! Read init data from binary file
  !
    implicit none
    character(*) file
    logical :: read_error
    
    if(master)then
        if(part_init_file_type .eq. 2) then
            open(unit=1215, file=file, status="old",  &
                 form="unformatted", err=112)
        else if(part_init_file_type .eq. 3) then
            open(unit=1215, file=file, status="old",  &
                 access="stream", form="unformatted", err=112)
        end if
        read(1215, err=113) part_init_data
        close(1215)
        
        if(debug > 1) then
            write(*,'(2x,a,5(f0.1,1x))') "particle file line 1: ",part_init_data(1,:)
            write(*,'(2x,a,5(f0.1,1x))') "particle file line 'part_total': ",part_init_data(part_total,:)
        end if
    end if

    return

112 if(master) write(*,*) 'part-track: could not open file for unformatted io'
113 if(master) write(*,*) 'part-track: could not read unformatted file'
    read_error = .true.
  end subroutine read_part_binary


  subroutine sort_particle_array()
  !
  ! Sort particle array according to release time
  !
    implicit none

    integer :: loop_rel, loop_search
    integer :: tmp_index
    real(dp) :: tmp_time, tmp_x, tmp_y, tmp_z, tmp_cls
    
    do loop_rel = 1,part_total-1
       tmp_time = part_init_data(loop_rel,1)
       tmp_index = loop_rel
       do loop_search = loop_rel+1,part_total
          if(tmp_time .gt. part_init_data(loop_search,1)) then
             tmp_time = part_init_data(loop_search,1)
             tmp_index = loop_search
          end if
       end do

       if(tmp_time .lt. part_init_data(loop_rel,1)) then
          tmp_x = part_init_data(tmp_index,2)
          tmp_y = part_init_data(tmp_index,3)
          tmp_z = part_init_data(tmp_index,4)
          tmp_cls = part_init_data(tmp_index,5)
          part_init_data(tmp_index,1) = part_init_data(loop_rel,1)
          part_init_data(tmp_index,2) = part_init_data(loop_rel,2)
          part_init_data(tmp_index,3) = part_init_data(loop_rel,3)
          part_init_data(tmp_index,4) = part_init_data(loop_rel,4)
          part_init_data(tmp_index,5) = part_init_data(loop_rel,5)          
          part_init_data(loop_rel,1) = tmp_time
          part_init_data(loop_rel,2) = tmp_x
          part_init_data(loop_rel,3) = tmp_y
          part_init_data(loop_rel,4) = tmp_z
          part_init_data(loop_rel,5) = tmp_cls          
       end if
    end do
  end subroutine sort_particle_array


  subroutine part_parameter_sanity_check
  !
  ! Sanity check for defined distance parameters
  !
    implicit none

    ! Check that distance limits are sensible
    if(part_minh .gt. depmax) then
       write(*,*) 'WARNING: part_minh larger than max(h).'
       part_minh = -1.
    else if(part_mineta .gt. depmax) then
       write(*,*) 'WARNING: part_mineta larger than max(h).'
       part_mineta = -1.
    else if((part_minh+part_mineta) .gt. depmax) then
       write(*,*) 'WARNING: Sum of part_minh and part_mineta larger than max(h).'
       part_minh = -1.
       part_mineta = -1.
    end if

    ! Set distance limits to default values if they are not already defined
    if(part_minh .lt. 0.0) then
       part_minh = dz(kb-1)*depmin/2.0
       write(*,*) 'mod_particle: setting part_minh to default value = ',part_minh
    end if
    if(part_mineta .lt. 0.0) then
       part_mineta = dz(1)*depmin/2.0
       write(*,*) 'mod_particle: setting part_mineta to default value = ',part_mineta
    end if

    ! Check that init distance limits are sensible
    if(part_init_minh .lt. part_minh) then
       write(*,*) 'WARNING: Set part_init_minh to part_minh =',part_minh
       part_init_minh = part_minh
    end if
    if(part_init_mineta .lt. part_mineta) then
       write(*,*) 'WARNING: Set part_init_mineta to part_mineta = ',part_mineta
       part_init_mineta = part_mineta
    end if
    if(part_init_minh .gt. depmax) then
       write(*,*) 'WARNING: part_init_minh=',part_init_minh,' is larger than max(h), changing it to be=',part_minh
       part_init_minh = part_minh
    else if(part_init_mineta .gt. depmax) then
       write(*,*) 'WARNING: part_init_mineta=',part_init_mineta,' larger than max(h). changing it to ',part_mineta
       part_init_mineta = part_mineta
    else if((part_init_minh+part_init_mineta) .gt. depmax) then
       part_init_minh = part_minh
       part_init_mineta = part_mineta
       write(*,*) 'WARNING: Sum of part_init_minh and part_init_mineta larger than max(h). changed to:',part_init_minh,part_init_mineta
    end if

    ! Check that land bump distance is sensible
    if((part_minland.le.0.).or.(part_minland.gt.1.)) then
       write(*,*) 'WARNING: Setting part_minland to 0.5 (default value).'
       part_minland = 0.5
    end if

    ! Check that boundary conditions are reasonable
    select case (part_bound_east_type)
    case(1,2,3)  !! pass test
    case default
       write(*,*) 'ERROR: invalid value of part_bound_east_type:',  &
            part_bound_east_type
       call abort()
    end select
    if ((part_bound_offset_east.lt.0.0).or.  &
         (part_bound_offset_east.gt.real(im)/2.0)) then
       write(*,*) 'ERROR: invalid value of part_bound_offset_east:',  &
            part_bound_offset_east
       call abort()
    end if

    select case (part_bound_west_type)
    case(1,2,3)  !! pass test
    case default
       write(*,*) 'ERROR: invalid value of part_bound_west_type:',  &
            part_bound_west_type
       call abort()
    end select
    if ((part_bound_offset_west.lt.0.0).or.  &
         (part_bound_offset_west.gt.real(im)/2.0)) then
       write(*,*) 'ERROR: invalid value of part_bound_offset_west:',  &
            part_bound_offset_west
       call abort()
    end if

    if (jm.gt.1) then
       select case (part_bound_north_type)
       case(1,2,3)  !! pass test
       case default
          write(*,*) 'ERROR: invalid value of part_bound_north_type:',  &
               part_bound_north_type
          call abort()
       end select
       if ((part_bound_offset_north.lt.0.0).or.  &
            (part_bound_offset_north.gt.real(jm)/2.0)) then
          write(*,*) 'ERROR: invalid value of part_bound_offset_north:',  &
               part_bound_offset_north
          call abort()
       end if

       select case (part_bound_south_type)
       case(1,2,3)  !! pass test
       case default
          write(*,*) 'ERROR: invalid value of part_bound_south_type:',  &
               part_bound_south_type
          call abort()
       end select
       if ((part_bound_offset_south.lt.0.0).or.  &
            (part_bound_offset_south.gt.real(jm)/2.0)) then
          write(*,*) 'ERROR: invalid value of part_bound_offset_south:',  &
               part_bound_offset_south
          call abort()
       end if
    end if

  end subroutine part_parameter_sanity_check


  subroutine part_set_boundary()
  !
  ! Set boundary values
  !
    implicit none

    if(block%east.ne.mpi_proc_null) then
       part_bound_loc_east = real(block%lim)
    else
       part_bound_loc_east = real(block%lim) - part_bound_offset_east
    end if

    if(block%west.ne.mpi_proc_null) then
       part_bound_loc_west = 0.0
    else
       part_bound_loc_west = part_bound_offset_west
    end if

    if (jm.gt.1) then
       if(block%north.ne.mpi_proc_null) then
          part_bound_loc_north = real(block%ljm)
       else
          part_bound_loc_north = real(block%ljm) - part_bound_offset_north
       end if

       if(block%south.ne.mpi_proc_null) then
          part_bound_loc_south = 0.0
       else
          part_bound_loc_south = part_bound_offset_south
       end if
    else
       part_bound_loc_north = 2.0
       part_bound_loc_south = 0.0
    end if

  end subroutine part_set_boundary


  subroutine particle_sanity_check(init_status)
  !
  ! Check that the initial data looks reasonable:
  !   Initial time: t >= 0
  !   x coordinate: 0 <= x <= (im-2)*dx
  !   y coordinate: 0 <= y <= (jm-2)*dx
  !
    implicit none

    integer, dimension(part_total) :: init_status

    integer :: loop_rel
    logical :: warning_time, warning_x, warning_y

    init_status = -1
    warning_time=.false.
    warning_x=.false.
    warning_y=.false.
    do loop_rel = 1,part_total 
       if (part_init_data(loop_rel,1) .lt. 0.) then
          init_status(loop_rel) = -50
          warning_time=.true.
       end if
       if ((part_init_data(loop_rel,2) .lt. 0.) .or.  &
            (part_init_data(loop_rel,2) .gt. float(im-2)*dx)) then
          init_status(loop_rel) = -50
          warning_x=.true.
       end if
       if(jm.ne.1)then
          if ((part_init_data(loop_rel,3) .lt. 0.) .or.  &
               (part_init_data(loop_rel,3) .gt. float(jm-2)*dy)) then
             init_status(loop_rel) = -50
             warning_y=.true.
          end if
       end if
    end do
    if(warning_time) write(*,*) 'WARNING: Particle(s) defined with negative time.'
    if(warning_x) write(*,*) 'WARNING: Particle(s) defined outside x-range of model domain.'
    if(warning_y) write(*,*) 'WARNING: Particle(s) defined outside y-range of model domain.'

  end subroutine particle_sanity_check


  subroutine part_set_random_seed
  !
  ! Set random seed for pseudorandom number generator
  !
    implicit none

    integer :: iseed,random_count,i
    integer, dimension(:), allocatable :: numseed

    call random_seed(size=iseed)
    allocate(numseed(iseed))

    do i=1,iseed
       call system_clock(random_count)
       numseed(i) = random_count*(myid+1)*i
    end do
    call random_seed(put=numseed)

  end subroutine part_set_random_seed


  subroutine allocate_exchange_arrays(exsize)
  !
  ! Allocate/reallocate particle exchange arrays for sending particles between processors
  !
    implicit none
    integer, optional :: exsize
    integer :: intsize, charsize, ierr
    integer :: eastblock_max, northblock_max

    if(present(exsize)) then
      ! if exsize is given, reallocate the exchange buffer size to that number of particles,
      ! but never more than part_total  
        if(exsize < 0 .or. exsize > part_total) exsize = part_total
        
        part_exchange_size = exsize
        
        if(allocated(put_east)) then
            deallocate(put_east, put_west, get_east, get_west)
            deallocate(part_buffer_send1, part_buffer_send2, part_buffer_recv1, part_buffer_recv2)            
        end if
        if(allocated(put_north))then
            deallocate(put_north, put_south, get_north, get_south)
        end if
    else        
       ! if exsize is not given, compute a sensible part_exchange_size, never larger than part_total. 
       ! if the number of particles is small, use that.

        if(part_exchange_size.eq.-1) then
            if(part_total.lt.MIN_EXCHANGE_SIZE)then
                part_exchange_size = part_total
            else
                ! assumed worst case: all particles are active, evenly distributed on one half of a processor 
                ! and we need to exchange all particles inside the cells along one edge
                part_exchange_size = nint(part_total * max(i_blocksize, j_blocksize)/(0.5*i_blocksize*j_blocksize))
                
                if(part_exchange_size < min_exchange_size) part_exchange_size = min_exchange_size
                if(part_exchange_size > part_total)        part_exchange_size = part_total
            end if
        end if
    end if

    allocate(put_east(part_exchange_size), put_west(part_exchange_size),  &
         get_east(part_exchange_size), get_west(part_exchange_size))
    
    if(jm.ne.1) then
       allocate(put_north(part_exchange_size), put_south(part_exchange_size),  &
            get_north(part_exchange_size), get_south(part_exchange_size))
    end if

    ! Create buffers for sending and receiving data
    part_buffer_size = size(transfer(put_east,(/'x'/)))
    allocate(part_buffer_send1(part_buffer_size), part_buffer_send2(part_buffer_size))
    allocate(part_buffer_recv1(part_buffer_size), part_buffer_recv2(part_buffer_size))

    if(master) write(*,*) 'Particle exchange array size:  ', part_exchange_size
    if(master) write(*,*) 'Particle exchange buffer size: ', part_buffer_size

    
    ! find which blocks should receive particles
    if(block%east.ne.mpi_proc_null)then
       eastblock = block%east
       if(part_outputlevel.gt.2) write(*,'(i4,a,i4)') myid,  &
            ': east block: ', eastblock
    else if (part_bound_east_type.eq.3) then  ! periodic boundary conditions
       eastblock = bltab(1,block%jbn)%prn
       if(part_outputlevel.gt.2) write(*,'(i4,a,i4)') myid,  &
            ': east block - periodic boundary: ', eastblock
    else
       eastblock = mpi_proc_null
       if(part_outputlevel.gt.2) write(*,'(i4,a)') myid,  &
            ': east block not found'
    end if

    if(block%west.ne.mpi_proc_null)then
       westblock = block%west
       if(part_outputlevel.gt.2) write(*,'(i4,a,i4)') myid,  &
            ': west block: ', westblock
    else if (part_bound_west_type.eq.3) then  ! periodic boundary conditions
       eastblock_max = gi2block(im)
       westblock = bltab(eastblock_max,block%jbn)%prn
       if(part_outputlevel.gt.2) write(*,'(i4,a,i4)') myid,  &
            ': west block - periodic boundary: ', westblock
    else
       westblock = mpi_proc_null
       if(part_outputlevel.gt.2) write(*,'(i4,a)') myid,  &
            ': west block not found'
    end if

    if(block%north.ne.mpi_proc_null)then
       northblock = block%north
       if(part_outputlevel.gt.2) write(*,'(i4,a,i4)') myid,  &
            ': north block: ', northblock
    else if (part_bound_north_type.eq.3) then  ! periodic boundary conditions
       northblock = bltab(block%ibn,1)%prn
       if(part_outputlevel.gt.2) write(*,'(i4,a,i4)') myid,  &
            ': north block - periodic boundary: ', northblock
    else
       northblock = mpi_proc_null
       if(part_outputlevel.gt.2) write(*,'(i4,a,i4)') myid,  &
            ': north block not found'
    end if
    
    if(block%south.ne.mpi_proc_null)then
       southblock = block%south
       if(part_outputlevel.gt.2) write(*,'(i4,a,i4)') myid,  &
            ': south block: ', southblock
    else if (part_bound_south_type.eq.3) then  ! periodic boundary conditions
       northblock_max = gj2block(jm)
       southblock = bltab(block%ibn,northblock_max)%prn
       if(part_outputlevel.gt.2) write(*,'(i4,a,i4)') myid,  &
            ': south block - periodic boundary: ', southblock
    else
       southblock = mpi_proc_null
       if(part_outputlevel.gt.2) write(*,'(i4,a,i4)') myid,  &
            ': south block not found'
    end if
  end subroutine allocate_exchange_arrays


  subroutine pgpart_setup()
  !
  ! Setup structures for plotting particles with pgplot
  !
    implicit none
    integer :: i

    if (master) then
       do i=1,pgpart_num_plots
          if (pgparts(i)%set) then
             select case(pgparts(i)%type)
             case(1)
                pgpart_set_history = .true.
                allocate(pgparts(i)%x1(part_total))
                allocate(pgparts(i)%y1(part_total))
             case(2)
                pgpart_set_history = .true.
                if (pgpart_ntime_pts < 2) pgpart_ntime_pts = 2
                allocate(pgparts(i)%x1(part_total))
                allocate(pgparts(i)%y1(part_total))
                allocate(pgparts(i)%x2(part_total))
                allocate(pgparts(i)%y2(part_total))
             case(3)
                pgpart_set_capture = .true.
                allocate(pgparts(i)%x1(part_total))
                allocate(pgparts(i)%y1(part_total))
             case(4)
                pgpart_set_history = .true.
                allocate(pgparts(i)%x1(part_total))
                allocate(pgparts(i)%y1(part_total))
             case(5)
                pgpart_set_history = .true.
                allocate(pgparts(i)%x1(part_total))
                allocate(pgparts(i)%y1(part_total))
             case(6)
                pgpart_set_history = .true.
                if (pgpart_ntime_pts < 2) pgpart_ntime_pts = 2
                allocate(pgparts(i)%x1(part_total))
                allocate(pgparts(i)%y1(part_total))
                allocate(pgparts(i)%x2(part_total))
                allocate(pgparts(i)%y2(part_total))
                
             case default
                write(*,*) '(plot; set; type): ', i, pgparts(i)%set, pgparts(i)%type
                call abort('ERROR: Not a valid pgparts(i)%type value!')
             end select
          end if
       end do
    end if

    call broadcast(pgpart_set_history)
    call broadcast(pgpart_set_capture)

    if (master) then
       if (pgpart_set_history) then
          allocate(pgpart_history(pgpart_ntime_pts,3,part_total))
          pgpart_history = -1.
       end if
       if (pgpart_set_capture) then
          allocate(pgpart_capture(4,part_total))
          pgpart_capture = -1.
       end if

       allocate(pgpart_color_array(pgpart_ntime_pts-1))
    end if

  end subroutine pgpart_setup


  subroutine activate_particles(now)
  !
  ! Activate particles if particle(:)%time <= now
  !
    implicit none
  
    real(dp) :: now
    integer :: init_status
    integer :: ti,tj
    integer :: root
    logical :: init_capture_h, init_capture_eta
    real(dp), dimension(3) :: xyz

    do
       if (index_activated.eq.part_total) exit
       if (particle(index_activated+1)%time .gt. now) exit
       index_activated = index_activated+1

       if(particle(index_activated)%status .eq. -1) then
          xyz(1) = particle(index_activated)%pos(1)
          xyz(2) = particle(index_activated)%pos(2)
          xyz(3) = particle(index_activated)%pos(3)
          root = -1
          call get_sigma_point(xyz,particle(index_activated)%pos,root)
          call reduce('max',root)
          particle(index_activated)%block = root
          if(myid.eq.particle(index_activated)%block) then
             init_capture_h = .true.
             init_capture_eta = .true.
             call check_vert_dist(part_init_minh,part_init_mineta,  &
                  init_capture_h,init_capture_eta,  &
                  particle(index_activated)%pos,  &
                  particle(index_activated)%block,  &
                  particle(index_activated)%status, &
                  particle(index_activated)%time)

             ! check if particle is seeded on land
             ti = nint(particle(index_activated)%pos(1) + 0.5)
             tj = nint(particle(index_activated)%pos(2) + 0.5)
             if(jm==1) tj=1

             if (fsm(ti,tj).eq.0.0) then
                particle(index_activated)%status = -30
             end if

             if(particle(index_activated)%status .eq. -10) then
                write(*,*) 'WARNING: Particle defined below allowed limit.'
                particle(index_activated)%status = -50
             else if(particle(index_activated)%status .eq. -20) then
                write(*,*) 'WARNING: Particle defined above allowed limit.'
                particle(index_activated)%status = -50
             else if(particle(index_activated)%status .eq. -30) then
                write(*,'(1x,a,i0,a,3(f0.1,1x))') 'WARNING: Particle defined on land. n=',index_activated,"xyz=",particle(index_activated)%pos(1:3)
                particle(index_activated)%status = -50
             else
                particle(index_activated)%status = myid
             end if
          else
             particle(index_activated)%status = -2
          end if

       end if
    end do

  end subroutine activate_particles


!========================================================================
! --- time stepping
!
! Subroutines called during time stepping: advect particles,  
! activate and deactivate particles
!
!========================================================================


  subroutine move_particle
  !
  ! Standard routine for moving particles
  !
    use checkpointing
    use control_parameters, only : do_checkpoint
    implicit none

    call particle_properties
    call part_step
    call part_output
    if(do_checkpoint) call part_checkpoint
    
  end subroutine move_particle


  subroutine part_step
  !
  ! Move and activate particles on the right timestep
  !
    implicit none

    logical :: part_transfer
    integer :: part, ti, tj
    real(dp) :: pxp, pyp, pzp
    real(dp) :: part_buoyancy
    real(dp) :: xadv, yadv, zadv
    real(dp), dimension(3) :: new_pos
    real(dp) :: in_flight, totdist_moved, avphysdist_moved
    
    if (.not.part_activate) return

    iput_east = 0
    iget_east = 0
    iput_west = 0
    iget_west = 0
    iput_north = 0
    iget_north = 0
    iput_south = 0
    iget_south = 0

    in_flight=0.0
    totdist_moved=0.0
    avphysdist_moved=0.0
    
    ! Move particle on subdomain where it is active.
    do part = 1, index_activated
       if (particle(part)%status.eq.myid) then
          pxp = particle(part)%pos(1)
          pyp = particle(part)%pos(2)
          pzp = particle(part)%pos(3)

          part_buoyancy=particle(part)%ws

          if (part_runge_kutta4) then
             call heun_method(pxp,pyp,pzp,part_buoyancy,  &
                  xadv,yadv,zadv)
          else
             ! Use Euler forward method
             if(part_bilinear) then
                call sample_uvw2d(pxp,pyp,pzp,part_buoyancy,  &
                     xadv,yadv,zadv)
             else
                call sample_uvw1d(pxp,pyp,pzp,part_buoyancy,  &
                     xadv,yadv,zadv)
             end if
          end if

          new_pos(1) = pxp + xadv
          new_pos(2) = pyp + yadv
          new_pos(3) = pzp + zadv

          ! make a record of "global" movement horisontally
          totdist_moved = totdist_moved + sqrt(xadv**2 + yadv**2)
          in_flight=in_flight+1
          
          ! check if new position is on land
          call check_land(part,new_pos)

          ! check if new position is too close to boundaries
          call check_vert_dist(part_minh,part_mineta,  &
               part_capture_h,part_capture_eta,  &
               new_pos, particle(part)%block,  &
               particle(part)%status,  &
               particle(part)%time)
          
          ! Update particle location
          particle(part)%pos(1) = new_pos(1)
          if(jm.ne.1) particle(part)%pos(2) = new_pos(2)
          particle(part)%pos(3) = new_pos(3)

          ! Check if particle is moved out of subdomain
          call check_domain_east_west(part,part_transfer)
          if ((.not.part_transfer) .and. (jm.ne.1)) then
             call check_domain_north_south(part)
          end if
       end if
    end do

    if(debug > 1 .and. mod(istep,10) == 0) then
        call reduce('sum', in_flight, totdist_moved, root=0)
        
        if(in_flight > 0.) avphysdist_moved = totdist_moved*dx/in_flight 
        
        if(master) write(*,'(a,1x,i0,1x, a,1x,f0.7,1x, a,1x,f0.7)') &
             " particles moved:",int(in_flight),"TDMG:",totdist_moved,"ADMR(m)",avphysdist_moved
    end if
    
    ! Exchange particles between subdomains
    call exchange_particles

    ! Activate new particles
    call activate_particles(real(istep,8)*real(dt,8))

  end subroutine part_step


  subroutine heun_method(pxp,pyp,pzp,part_buoyancy,xadv,yadv,zadv)
  !
  ! Heun's method (predictor-corrector method)
  ! also called modified Euler/expkecut trapezoidal rule
  ! Predictor: Euler forward
  ! Corrector: Trapezoidal
  ! 
    implicit none

    real(dp) :: pxp, pyp, pzp
    real(dp) :: part_buoyancy
    real(dp) :: xadv, yadv, zadv
    real(dp) :: pred_pxp, pred_pyp, pred_pzp
    real(dp) :: pred_xadv, pred_yadv, pred_zadv
    real(dp) :: corr_xadv, corr_yadv, corr_zadv

    ! Predictor
    if(part_bilinear) then
       call sample_uvw2d(pxp,pyp,pzp,part_buoyancy,  &
            pred_xadv,pred_yadv,pred_zadv)
    else
       call sample_uvw1d(pxp,pyp,pzp,part_buoyancy,  &
            pred_xadv,pred_yadv,pred_zadv)
    end if

    pred_pxp = pxp + pred_xadv
    pred_pyp = pyp + pred_yadv
    pred_pzp = pzp + pred_zadv

    ! Corrector
    if(part_bilinear) then
       call sample_uvw2d(pred_pxp,pred_pyp,pred_pzp,  &
            part_buoyancy,corr_xadv,corr_yadv,corr_zadv)
    else
       call sample_uvw1d(pred_pxp,pred_pyp,pred_pzp,  &
            part_buoyancy,corr_xadv,corr_yadv,corr_zadv)
    end if

    xadv = 0.5*(pred_xadv + corr_xadv)
    yadv = 0.5*(pred_yadv + corr_yadv)
    zadv = 0.5*(pred_zadv + corr_zadv)

  end subroutine heun_method


  subroutine sample_uvw1d(p_x,p_y,p_z,part_buoyancy,xadv0,yadv0,zadv0)
  !
  ! Subroutine that samples a 3d staggered velocity field
  ! with linear interpolation
  !
    implicit none
    real(dp), intent(in) :: p_x,p_y,p_z
    real(dp), intent(in) :: part_buoyancy
    real(dp), intent(out) :: xadv0, yadv0, zadv0

    real(dp) :: u0, v0, w0
    integer :: k0, i, j, k
    integer :: ci, cj
    real(dp) :: del_x, del_y, del_z
    real(dp) :: dkh
    real(dp) :: xran,yran,zran

    integer :: gi,gj
    integer :: root

    ! closest cell south west
    i=nint(p_x)
    j=nint(p_y)
    if(jm.eq.1) j = 1

    ! Cell center
    ci=nint(p_x+0.5)
    cj=nint(p_y+0.5)
    if(jm.eq.1) cj = 1

    if (fsm(ci,cj).gt.0.) then
       ! Find which vertical grid cell containing the particle
       k0=int(-p_z*float(kb-1))
       k=bisearch(real(-z,dp), kb, -p_z, k0)
       if (k.eq.0) k=1
    
       ! Compute weights for linear interpolation based on particle location
       del_x = p_x - (i - 1.0) 
       del_y = p_y - (j - 1.0) 
       del_z = (z(k) - p_z)/dz(k)

       ! Interpolate x, y and z advection into the particles location
       u0=(1-del_x)*u(i,j,k)+del_x*u(i+1,j,k)
       xadv0 = u0*dt/dx

       if(jm.ne.1) then
          v0=(1-del_y)*v(i,j,k)+del_y*v(i,j+1,k)
          yadv0 = v0*dt/dy
       else
          yadv0 = 0.
       end if
       
       if (k.ge.kb-1) then
          w0=w(ci,cj,kb-1)
       else
          w0=(1-del_z)*w(ci,cj,k)+del_z*w(ci,cj,k+1)
       end if
       zadv0 = w0*dt
       if(part_vert_velocity /= 0) then
          ! particle with vertical self-propulsion
          zadv0 = zadv0 + part_buoyancy*dt/D(ci,cj)
       end if
       
       ! move random step
       if(part_random_step) then
          call move_random(ci,cj,k,real(D(ci,cj),dp),xran,yran,zran)
          
          xadv0 = xadv0 + xran
          yadv0 = yadv0 + yran
          zadv0 = zadv0 + zran
       end if       
    else
       xadv0 = 0.
       yadv0 = 0.
       zadv0 = 0.
    end if

    return

  end subroutine sample_uvw1d


  subroutine sample_uvw2d(p_x,p_y,p_z,part_buoyancy,xadv0,yadv0,zadv0)
  !
  ! Subroutine that samples a 3d staggered velocity field
  ! with linear interpolation
  !
    implicit none
    real(dp), intent(in) :: p_x,p_y,p_z
    real(dp), intent(in) :: part_buoyancy
    real(dp), intent(out) :: xadv0, yadv0, zadv0

    real(dp), dimension(3) :: sig
    real(dp) :: D_interp, eta_interp, etap_interp
    real(dp) :: u0, v0, w0
    integer :: ci,cj,k,k0
    real(dp) :: dkh
    real(dp) :: xran,yran,zran

    sig(1) = p_x
    sig(2) = p_y
    sig(3) = p_z

    call part_bilinear_calc(sig,Dm=D_interp,um=u0,vm=v0,wm=w0)

    if (abs(D_interp) .gt. 0.) then
       xadv0 = u0*dt/dx
       if(jm.ne.1) then
          yadv0 = v0*dt/dy
       else
          yadv0 = 0.
       end if

       zadv0 = w0*dt
       if(part_vert_velocity /= 0) then
          ! particle with vertical self-propulsion
          zadv0 = zadv0 + part_buoyancy*dt/D_interp
       end if

       ! move random step
       if(part_random_step) then
          ci=nint(p_x + 0.5)
          cj=nint(p_y + 0.5)
          k0=int(-p_z*float(kb-1))
          k=bisearch(real(-z,dp),kb,-p_z,k0)
          if (k.eq.0) k=1

          call move_random(ci,cj,k,D_interp,xran,yran,zran)
          
          xadv0 = xadv0 + xran
          yadv0 = yadv0 + yran
          zadv0 = zadv0 + zran
       end if
    else
       xadv0 = 0.
       yadv0 = 0.
       zadv0 = 0.
    end if

    return

  end subroutine sample_uvw2d


  subroutine move_random(i,j,k,D_interp,xran,yran,zran)
  ! Make random step
    implicit none
    integer, intent(in) :: i,j,k
    real(dp), intent(in) :: D_interp
    real(dp), intent(out) :: xran, yran, zran
    real(dp) :: dkh

    xran = 0.
    yran = 0.
    zran = 0.

    if (fsm(i,j) .gt. 0.0) then
       
       select case(part_random_type)
       case(0)
          if ((fsm(i+1,j).gt.0.0).and.(fsm(i-1,j).gt.0.0)) then
             xran = ((dt/D_interp)*  &
                  (ah(i+1,j,k)*D(i+1,j) - ah(i-1,j,k)*D(i-1,j))/(2.*dx) +  &
                  sqrt(2.*ah(i,j,k)*dt) * gran())/dx
          else if (fsm(i+1,j).gt.0.0) then
             xran = ((dt/D_interp)*  &
                  (ah(i+1,j,k)*D(i+1,j) - ah(i,j,k)*D(i,j))/dx +  &
                  sqrt(2.*ah(i,j,k)*dt) * gran())/dx
          else if (fsm(i-1,j).gt.0.0) then
             xran = ((dt/D_interp)*  &
                  (ah(i,j,k)*D(i,j) - ah(i-1,j,k)*D(i-1,j))/dx +  &
                  sqrt(2.*ah(i,j,k)*dt) * gran())/dx
          end if

          if(jm.ne.1) then
             if ((fsm(i,j+1).gt.0.0).and.(fsm(i,j-1).gt.0.0)) then
                yran = ((dt/D_interp)*  &
                     (ah(i,j+1,k)*D(i,j+1) - ah(i,j+1,k)*D(i,j+1))/(2.*dy) +  &
                     sqrt(2.*ah(i,j,k)*dt) * gran())/dy
             else if (fsm(i,j+1).gt.0.0) then
                yran = ((dt/D_interp)*  &
                     (ah(i,j+1,k)*D(i,j+1) - ah(i,j,k)*D(i,j))/dy +  &
                     sqrt(2.*ah(i,j,k)*dt) * gran())/dy
             else if (fsm(i,j-1).gt.0.0) then
                yran = ((dt/D_interp)*  &
                     (ah(i,j,k)*D(i,j) - ah(i,j-1,k)*D(i,j-1))/dy +  &
                     sqrt(2.*ah(i,j,k)*dt) * gran())/dy
             end if
          end if

          if (k.eq.1) then
             dkh = (kh(i,j,k+1) - kh(i,j,k))/(z(k+1) - z(k))
          else if(k.ge.(kb-1)) then
             dkh = (kh(i,j,k) - kh(i,j,k-1))/(z(k) - z(k-1))
          else
             dkh = (kh(i,j,k+1) - kh(i,j,k-1))/(z(k+1) - z(k-1))
          end if
          zran = (dt/(D_interp*D_interp))*dkh +  &
               sqrt(2.*kh(i,j,k)*dt)/D_interp * gran()

       case(1)
          if ((fsm(i+1,j).gt.0.0).and.(fsm(i-1,j).gt.0.0)) then
             xran = ((dt/D_interp)*  &
                  (part_ah*D(i+1,j) - part_ah*D(i-1,j))/(2.*dx) +  &
                  sqrt(2.*part_ah*dt) * gran())/dx
          else if (fsm(i+1,j).gt.0.0) then
             xran = ((dt/D_interp)*  &
                  (part_ah*D(i+1,j) - part_ah*D(i,j))/dx +  &
                  sqrt(2.*part_ah*dt) * gran())/dx
          else if (fsm(i-1,j).gt.0.0) then
             xran = ((dt/D_interp)*  &
                  (part_ah*D(i,j) - part_ah*D(i-1,j))/dx +  &
                  sqrt(2.*part_ah*dt) * gran())/dx
          end if

          if(jm.ne.1) then
             if ((fsm(i,j+1).gt.0.0).and.(fsm(i,j-1).gt.0.0)) then
                yran = ((dt/D_interp)*  &
                     (part_ah*D(i,j+1) - part_ah*D(i,j-1))/(2.*dy) +  &
                     sqrt(2.*part_ah*dt) * gran())/dy
             else if (fsm(i,j+1).gt.0.0) then
                yran = ((dt/D_interp)*  &
                     (part_ah*D(i,j+1) - part_ah*D(i,j))/dy +  &
                     sqrt(2.*part_ah*dt) * gran())/dy
             else if (fsm(i,j-1).gt.0.0) then
                yran = ((dt/D_interp)*  &
                     (part_ah*D(i,j) - part_ah*D(i,j-1))/dy +  &
                     sqrt(2.*part_ah*dt) * gran())/dy
             end if
          end if

          zran = sqrt(2.*part_kh*dt)/D_interp * gran()          

       case default
          call abort('ERROR: Not a valid option for part_random_type')
       end select
    end if

  end subroutine move_random


  subroutine check_domain_east_west(part,part_transfer)
  ! Check if particle is moved out of subdomain in the east-west direction
    implicit none
    integer :: part
    logical :: part_transfer

    part_transfer=.false.

    if(particle(part)%status.ge.0) then
       if(particle(part)%pos(1).lt.part_bound_loc_west) then
          if(part_outputlevel.gt.2) write(*,'(i4,a,i8)') myid,  &
               ': particle exit west boundary: ', particle(part)%id
          if(westblock.ne.mpi_proc_null)then   ! Also works for periodic bc
             iput_west = iput_west+1
             put_west(iput_west) = particle(part)
             put_west(iput_west)%pos(1) = particle(part)%pos(1) - part_bound_loc_west
             particle(part)%status = -2
             part_transfer = .true.
          else
             select case(part_bound_west_type)
             case(1)  ! Capture particle
                particle(part)%status = -40
                particle(part)%time = real(istep,8)*real(dt,8)
             case(2)  ! Reflective boundary condition
                particle(part)%pos(1) = 2*part_bound_loc_west - particle(part)%pos(1)
             case default
                write(*,*) 'ERROR: Invalid part_bound_west_type: ', part_bound_west_type
                call abort()
             end select
          end if
          
       else if(particle(part)%pos(1).gt.part_bound_loc_east) then
          if(part_outputlevel.gt.2) write(*,'(i4,a,i8)') myid,  &
               ': particle exit east boundary: ', particle(part)%id
          if(eastblock.ne.mpi_proc_null)then   ! Also works for periodic bc
             iput_east = iput_east+1
             put_east(iput_east) = particle(part)
             put_east(iput_east)%pos(1) = particle(part)%pos(1) - part_bound_loc_east
             particle(part)%status = -2
             part_transfer = .true.
          else
             select case(part_bound_east_type)
             case(1)  ! Capture particle
                particle(part)%status = -40
                particle(part)%time = real(istep,8)*real(dt,8)
             case(2)  ! Reflective boundary condition
                particle(part)%pos(1) = 2*part_bound_loc_east - particle(part)%pos(1)
            case default
                write(*,*) 'ERROR: Invalid part_bound_east_type: ', part_bound_east_type
                call abort()
             end select
          end if
       end if
    end if

  end subroutine check_domain_east_west


  subroutine check_domain_north_south(part)
  ! Check if particle is moved out of subdomain in the north-south direction
    implicit none
    integer :: part

    if(particle(part)%status.ge.0) then
       if(particle(part)%pos(2).lt.part_bound_loc_south) then
          if(jm.eq.1) call abort('ERROR: Can not move particle south')
          if(part_outputlevel.gt.2) write(*,'(i4,a,i8)') myid,  &
               ': particle exit south boundary: ', particle(part)%id
          if(southblock.ne.mpi_proc_null)then   ! Also works for periodic bc
             iput_south = iput_south+1
             put_south(iput_south) = particle(part)
             put_south(iput_south)%pos(2) = particle(part)%pos(2) - part_bound_loc_south
             particle(part)%status = -2
          else
             select case(part_bound_south_type)
             case(1)  ! Capture particle
                particle(part)%status = -40
                particle(part)%time = real(istep,8)*real(dt,8)
             case(2)  ! Reflective boundary condition
                particle(part)%pos(2) = 2*part_bound_loc_south - particle(part)%pos(2)
             case default
                write(*,*) 'ERROR: Invalid part_bound_south_type: ', part_bound_south_type
                call abort()
             end select
          end if
          
       else if(particle(part)%pos(2).gt.part_bound_loc_north) then
          if(jm.eq.1) call abort('ERROR: Can not move particle north')
          if(part_outputlevel.gt.2) write(*,'(i4,a,i8)') myid,  &
               ': particle exit north boundary: ', particle(part)%id
          if(northblock.ne.mpi_proc_null)then   ! Also works for periodic bc
             iput_north = iput_north+1
             put_north(iput_north) = particle(part)
             put_north(iput_north)%pos(2) = particle(part)%pos(2) - part_bound_loc_north
             particle(part)%status = -2
          else
             select case(part_bound_north_type)
             case(1)  ! Capture particle
                particle(part)%status = -40
                particle(part)%time = real(istep,8)*real(dt,8)
             case(2)  ! Reflective boundary condition
                particle(part)%pos(2) = 2*part_bound_loc_north - particle(part)%pos(2)
             case default
                write(*,*) 'ERROR: Invalid part_bound_north_type: ', part_bound_north_type
                call abort()
             end select
          end if
       end if
    end if

  end subroutine check_domain_north_south


  subroutine exchange_particles
!    
! Exchange particles between subdomains
!    
    implicit none
    integer :: part, ierr, n

    numput = numput + iput_east + iput_west + iput_north + iput_south
    
    if(istep==1 .or. mod(istep,10) == 0) then
       call reduce("sum",numput)
       if(master.and.debug > 1)then
           write(*,*) "particles exchanged last 10 steps:",numput
       end if
       numput=0
    end if

    call isend(iput_east, eastblock)
    call receive(iget_west, westblock)
        
    call isend(iput_west, westblock)
    call receive(iget_east, eastblock)
        
    if(iput_east > 0) then
        part_buffer_send1 = transfer(put_east(1:iput_east), part_buffer_send1)       
        call isend(part_buffer_send1(1:iput_east*particle_charsize), eastblock)
    end if
    if(iput_west > 0) then
        part_buffer_send2 = transfer(put_west(1:iput_west), part_buffer_send2)
        call isend(part_buffer_send2(1:iput_west*particle_charsize), westblock)
    end if
    if(iget_west > 0) then
        call receive(part_buffer_recv1(1:iget_west*particle_charsize), westblock)
    end if
    if(iget_east > 0) then
        call receive(part_buffer_recv2(1:iget_east*particle_charsize), eastblock)
    end if
    call waitall()

    ! special case for periodic bc east-west on one processor:
    if(numproc == 1) then
        if(iput_east > 0) then
            iget_west = iput_east
            part_buffer_recv1 = part_buffer_send1
        end if
        if(iput_west > 0) then
            iget_east = iput_east
            part_buffer_recv2 = part_buffer_send2
        end if
    end if
    
    ! Activate sent particles on new subdomains. 
    ! Check if any new particles must be moved north-south
    if(iget_east > 0) get_east = transfer(part_buffer_recv2(1:iget_east*particle_charsize),get_east)
    
    do part=1,iget_east
       if(part_outputlevel.gt.2) write(*,'(i4,a,i8)') myid,  &
            ': particle enter east boundary: ', get_east(part)%id
       get_east(part)%pos(1) = get_east(part)%pos(1) + part_bound_loc_east
       particle(get_east(part)%id) = get_east(part)
       particle(get_east(part)%id)%status = myid
       particle(get_east(part)%id)%block = myid
       if (jm.ne.1) call check_domain_north_south(get_east(part)%id)
    end do

    if(iget_west.gt.0) get_west = transfer(part_buffer_recv1(1:iget_west*particle_charsize), get_west)
    
    do part=1,iget_west
       if(part_outputlevel.gt.2) write(*,'(i4,a,i8)') myid,  &
            ': particle enter west boundary: ', get_west(part)%id
       get_west(part)%pos(1) = get_west(part)%pos(1) + part_bound_loc_west
       particle(get_west(part)%id) = get_west(part)
       particle(get_west(part)%id)%status = myid
       particle(get_west(part)%id)%block = myid
       if (jm.ne.1) call check_domain_north_south(get_west(part)%id)
    end do

    
    ! Move particles between blocks in north-south direction
    if(jm.ne.1)then
        call isend(iput_north, northblock)
        call receive(iget_south, southblock)
        
        call isend(iput_south, southblock)
        call receive(iget_north, northblock)

        call waitall()

        if(iput_north > 0) then
            part_buffer_send1 = transfer(put_north(1:iput_north), part_buffer_send1)
            call isend(part_buffer_send1(1:iput_north*particle_charsize), northblock)
        end if
        if(iput_south > 0) then
            part_buffer_send2 = transfer(put_south(1:iput_south), part_buffer_send2)

            call isend(part_buffer_send2(1:iput_south*particle_charsize), southblock)
        end if
        if(iget_south > 0) then
            call receive(part_buffer_recv1(1:iget_south*particle_charsize), southblock)
        end if
        if(iget_north > 0) then
            call receive(part_buffer_recv2(1:iget_north*particle_charsize), northblock)
        end if
        call waitall()

     
! Activate sent particles on new subdomains.
        if(iget_north > 0) get_north = transfer(part_buffer_recv2(1:iget_north*particle_charsize), get_north)

     
       do part=1,iget_north
          if(part_outputlevel.gt.2) write(*,'(i4,a,i8)') myid,  &
               ': particle enter north boundary: ', get_north(part)%id
          get_north(part)%pos(2) = get_north(part)%pos(2) + part_bound_loc_north
          particle(get_north(part)%id) = get_north(part)
          particle(get_north(part)%id)%status = myid
          particle(get_north(part)%id)%block = myid
       end do

       if(iget_south > 0) get_south = transfer(part_buffer_recv1(1:iget_south*particle_charsize), get_south)
       
       do part=1,iget_south
          if(part_outputlevel.gt.2) write(*,'(i4,a,i8)') myid,  &
               ': particle enter south boundary: ', get_south(part)%id
          get_south(part)%pos(2) = get_south(part)%pos(2) + part_bound_loc_south
          particle(get_south(part)%id) = get_south(part)
          particle(get_south(part)%id)%status = myid
          particle(get_south(part)%id)%block = myid
       end do
   end if

   
! periodically, globally check if the buffers are big enough
   
   maxput = max(maxput, maxval((/iput_east,iput_west,iput_north,iput_south/)))

   if(istep==1 .or. mod(istep,10) == 0) then
       call reduce("max",maxput)
       if(maxput > nint(0.7*part_exchange_size)) then
           call message("reallocating particle exchange buffer to size ", 1, ival=(/2*maxput/))
           call allocate_exchange_arrays( maxput*2 )
       end if
   end if
   
  end subroutine exchange_particles


!=======================================================================
! Properties

  subroutine lookup(timearr, ptime, property1, val1, property2, val2, property3, val3, property4, val4)
! for a given particle age/time, linearly interpolate value from a property table of dimension n.
    implicit none
    integer n           ! number of items in lookup table(s)
    real timearr(:)     ! lookup table for time/age
    real ptime          ! time/age to look up
    real property1(:)   ! table of property 1 at the time(s) given in timearr
    real val1           ! interpolated value of property 1 at given time
    real, optional :: property2(:), property3(:), property4(:)
    real, optional :: val2, val3, val4
    integer i
    real alfa,beta

    n=size(timearr,1)
    
    if(ptime <= timearr(1)) then
        val1 = property1(1)
        if(present(val2)) val2 = property2(1)
        if(present(val3)) val3 = property3(1)
        if(present(val4)) val4 = property4(1)
        return
    end if
    if(ptime > timearr(n)) then
        val1 = property1(n)
        if(present(val2)) val2 = property2(n)
        if(present(val3)) val3 = property3(n)
        if(present(val4)) val4 = property4(n)        
        return
    end if

! for long arrays binary search is faster, but the overhead of this
! loop is hard to beat for short arrays... improve later if a problem
    i=2
    do
        if(ptime > timearr(i)) then
            i=i+1
            cycle
        end if
        exit
    end do
    
! now timearr(i-1) < ptime < timearr(i)
    alfa = (timearr(i)-ptime)/(timearr(i)-timearr(i-1))
    val1 = property1(i-1)*alfa + property1(i)*(1-alfa)
    if(present(val2)) val2 = property2(i-1)*alfa + property2(i)*beta
    if(present(val3)) val3 = property3(i-1)*alfa + property3(i)*beta
    if(present(val4)) val4 = property4(i-1)*alfa + property4(i)*beta
  end subroutine lookup

  
  subroutine particle_properties()
    implicit none
    integer i, part
    real :: my_h2o = 1.0e-3, r, age, time
    
    do part = 1, index_activated
        if (particle(part)%status == myid) then
            
            if(particle(part)%class == 0) cycle
            
            i = particle(part)%class
            
! this is where we use the class and time to look up particle properties in the provided tables:

            if(part_lookup_ws) then
                call lookup(matter(i)%time,  real(istep*dt-particle(part)%time), &
                            matter(i)%prop1, particle(part)%ws)
            end if
            
            if(part_lookup_sg_and_diameter)then
                ! get sg from property1, diameter from property2
                call lookup(matter(i)%time,  real(istep*dt-particle(part)%time), &
                            matter(i)%prop1, particle(part)%sg, &
                            matter(i)%prop2, particle(part)%d)               
            end if


! now read or compute the vertical velocity
            
!         if(part_use_lice_model)then            
!            particle(part)%ws = salmonlice(time, temperature, salinity, ...)
!         etc.            
            
! stokes law
            if(part_apply_stokes) then
                R = particle(part)%d/2
                particle(part)%ws = -2./9*(particle(part)%sg - 1.0)*rho0*grav*R**2/my_h2o
            end if
            
!print *,part,"ws,d,sg", particle(part)%ws, particle(part)%d, particle(part)%sg
            
        end if
    end do
  end subroutine particle_properties
  

  subroutine add_class(splitline)
! part_add_class <name> <material number> <datafile> <numprop>
!   connect a text file <datafile> with a lookup table to a particle
!   class of given <name>. the materials must be given a <material number>
!   1-50. this class number can be used in the fifth
!   column in the init files for particles to give particles specific
!   behaviour that can be e.g. time dependent or aware of its
!   environment. currently we only support text files with up to five
!   columns; Time(seconds), property1, ..., property4
!   <numprop> is the number of material properties to read, defaults to 2.
    use file_util
    implicit none
    character(*) splitline(:)
    real, dimension(100) :: col1,col2,col3,col4,col5
    integer n,pn,i,numprop

    numprop=2
    part_numclasses = part_numclasses + 1

    read(splitline(3),*,err=101,end=101) pn
    if(pn < 1 .or. pn > part_maxnumclasses) call message("material number in col 3 outside allowed range",2)

    read(splitline(5),*,err=102,end=102) numprop
    if(numprop < 1 .or. numprop > 4) call message("number of properties in col 5 outside allowed range",2)
    
    matnumbers(part_numclasses) = pn  ! bookeeping...
    
    matter(pn)%name = splitline(2)
    matter(pn)%file = splitline(4)
    matter(pn)%numprop = numprop
    
    select case(numprop)
    case(1)
        call get_fields(splitline(4), col1, col2, lines=n)
    case(2)        
        call get_fields(splitline(4), col1, col2, col3, lines=n)
    case(3)
        call get_fields(splitline(4), col1, col2, col3, col4, lines=n)
    case(4)
        call get_fields(splitline(4), col1, col2, col3, col4, col5, lines=n)
    end select
    
    matter(pn)%n = n
    allocate(matter(pn)%time(n))
    matter(pn)%time = col1(1:n)
    
    if(numprop >= 1) allocate(matter(pn)%prop1(n))
    if(numprop >= 1) matter(pn)%prop1 = col2(1:n)
    if(numprop >= 2) allocate(matter(pn)%prop2(n))
    if(numprop >= 2) matter(pn)%prop2 = col3(1:n)    
    if(numprop >= 3) allocate(matter(pn)%prop3(n))
    if(numprop >= 3) matter(pn)%prop3 = col4(1:n)   
    if(numprop >= 4) allocate(matter(pn)%prop4(n))
    if(numprop >= 4) matter(pn)%prop4 = col5(1:n)
    
    if(master) write(*,'(a,a,a,i0,a,i0,a,i0)') ' part_add_class: name="',trim(matter(pn)%name),'" n=',matter(pn)%n,&
         ' material id=',pn,' numprop=',numprop
    
    if(debug > 1 .and. master) then
        select case(numprop)
        case(1)
            write(*,'(2x,4a11)') "Time(s)","Prop1"
            do i=1,n
                write(*,'(3x, f10.1, 1x, f10.5)') matter(pn)%time(i),matter(pn)%prop1(i)
            end do
        case(2)
            write(*,'(2x,4a11)') "Time(s)","Prop1","Prop2"
            do i=1,n
                write(*,'(3x, f10.1, 1x, f10.5, 1x, f10.5)') matter(pn)%time(i),matter(pn)%prop1(i),matter(pn)%prop2(i)
            end do
        case(3)
            write(*,'(2x,4a11)') "Time(s)","Prop1","Prop2","Prop3" 
            do i=1,n
                write(*,'(3x, f10.1, 1x, f10.5, 1x, f10.5, 1x, f10.5)') &
                     matter(pn)%time(i),matter(pn)%prop1(i),matter(pn)%prop2(i),matter(pn)%prop3(i)
            end do
        case(4)
            write(*,'(2x,5a11)') "Time(s)","Prop1","Prop2","Prop3","Prop4" 
            do i=1,n
                write(*,'(3x, f10.1, 1x, f10.5, 1x, f10.5, 1x, f10.5, 1x, f10.5)') &
                     matter(pn)%time(i),matter(pn)%prop1(i),matter(pn)%prop2(i),matter(pn)%prop3(i),matter(pn)%prop4(i)
            end do
        end select
    end if
    return
101 call message("failed to read pn from column 3",2)
102 call message("failed to read numprop from column 5",2)
  end subroutine add_class
  

!========================================================================
! --- output
!
! Subroutines for printing movements of the particles
! and their end location
!
!========================================================================


  subroutine part_output
  !
  ! Manage output for particle tracking

    implicit none
    integer :: part, firstp, lastp, numopen, numclosing, tsunit
    logical :: part_status_updated, file_is_open

    if (.not.part_activate) return

    
! Write particle tracks to files.
! Many file systems do not scale well for thousands of files, we
! should try to keep the number of open files below e.g. 5000.
! For runs with > 10 000 particles, if the user does not reduce the
! number of timeseries using partplot_ts_mod, we restrict the number
! of tracks to the first 999 files.
    
    if ((partplot_timeseries).and.(mod(istep,partplot_ts_write).eq.0)) then


! periodically check for (and close) open particle track files for inactive particles

        if(mod(istep,partplot_ts_write*10).eq.0) then        
            numopen = 0
            numclosing = 0
            do part = 1, index_activated
                tsunit = 2000000+(1000*myid)+part
                inquire(unit=tsunit, opened=file_is_open)
                if(file_is_open) then
                    if (particle(part)%status < 0) then
                        close(tsunit)
                        numclosing = numclosing + 1
                    else
                        numopen = numopen + 1
                    end if
                end if
            end do
            call reduce("sum", numopen, numclosing, root=0)
            if(master) write(*,*) " mod_particle timeseries:  open files:",numopen," closing inactive files:",numclosing
        end if
        
        
        if (partplot_ts_mod > 0) then
            do part = 1, index_activated
                
                if(mod(part-1,partplot_ts_mod*partplot_ts_mod_bsize) == 0) then
                    firstp = part
                    lastp  = part + partplot_ts_mod_bsize - 1
                end if
                
                if(part >= firstp .and. part <= lastp)then
                    call timeserie_part('particle',part)
                end if
            end do
        else
            do part = 1, min(index_activated,999)
                call timeserie_part('particle',part)
            end do
        end if
    end if

    ! Write position of all active particles
    if ((partplot_distribution).and.  &
         (mod(istep,partplot_dist_write).eq.0)) then
       call distribution_part()
    end if

    part_status_updated = .false.
    if(part_check_status) then 
       ! Print a timeserie of the total number of particles 
       ! that have been taken out of the simulation.
       if((mod(istep,part_write_time).eq.0) .or. (istep.eq.numstep)) then
          if (istep.gt.0) then
             call update_status_data
             part_status_updated = .true.             
          end if
       end if

       ! Write encounter locations to file at last time step
       if (istep > 1 .and. mod(istep,part_enc_dumpint).eq.0  .or.  istep.eq.numstep) call encounter_loc
    end if
    
    if(numpbox > 0 .and. mod(istep, boxloggerintval) == 0)  call particleboxlogger()
    
    ! Update fields used by pgplot
    if (part_use_pgplot .and.  &
         (mod(istep,pgpart_plot_time).eq.0)) then
       if(pgpart_set_history) call update_pgpart_history()
       if(pgpart_set_capture .and. (istep.gt.0)) then
          call update_pgpart_capture(part_status_updated)
       end if
    end if

  end subroutine part_output

  
  subroutine open_boxlogger_files()
    implicit none
    integer b, tsunit, ios
    logical file_exist
    character(120) datafile
    
    do b=1,numpbox
        if(pbox(b)%inblock ) then
            tsunit = 4000000 + 1000*myid + b            
            
            write(datafile,'(a,"-",i3.3,"-",i5.5,"x",i5.5,"x",i3.3,"-P",i4.4,".data")')  & 
                 trim(pbox(b)%name),b,im,jm,kb-1,myid
            
            inquire(file=(trim(particlestat_dir)//trim(datafile)), exist=file_exist)

            if(file_exist) then
                open(unit=tsunit, file=(trim(particlestat_dir)//trim(datafile)), status="old", position="append", &
                     err=110, iostat=ios)
            else
                open(unit=tsunit, file=(trim(particlestat_dir)//trim(datafile)), err=110, iostat=ios)
            end if
        end if
    end do
    return
110 call message("error opening boxloggerunit, iostat,myid,file=", 1, ival=(/ios,myid/), sval=(/trim(particlestat_dir)//trim(datafile)/))
  end subroutine open_boxlogger_files


  subroutine close_boxlogger_files()
    implicit none
    integer b, tsunit
    logical file_open
    
    do b=1,numpbox
        if(pbox(b)%inblock ) then
            tsunit = 4000000 + 1000*myid + b            
            inquire(unit=tsunit, opened=file_open)            
            if(file_open) close(unit=tsunit)
        end if
    end do
  end subroutine close_boxlogger_files

  
  subroutine add_boxlogger(splitline)
! usage:
!   part_boxlogger <name> <x range> <y range> <z range>
    implicit none
    character(*) splitline(:)

    numpbox = numpbox + 1
    if(numpbox > maxnumpbox) then
        call message("too many particlebox loggers, ignoring this one with name=",1,sval=(/splitline(2)/),margin=1)
        return
    end if
    if(.not.allocated(pbox)) then
        allocate(pbox(maxnumpbox))
    end if

    pbox(numpbox)%name = splitline(2)
    read(splitline(3:4),*,err=102,end=102) pbox(numpbox)%r(1:2)
    read(splitline(5:6),*,err=103,end=103) pbox(numpbox)%r(3:4)
    read(splitline(7:8),*,err=104,end=104) pbox(numpbox)%r(5:6)
    
! convert ranges to global grid coordinates
    pbox(numpbox)%r(1:2) = (pbox(numpbox)%r(1:2) + dx)/dx
    pbox(numpbox)%r(3:4) = (pbox(numpbox)%r(3:4) + dy)/dy

    call ronblock(pbox(numpbox)%r(1:4), pbox(numpbox)%inblock)
    if(master) write(*,*) "part_boxlogger: ",pbox(numpbox)%r(1:6)," inblock=",pbox(numpbox)%inblock
    return
    
102 call message("could not read x range for particlebox logger=",1,sval=(/splitline(2)/),margin=1)    
    numpbox = numpbox - 1
    return
103 call message("could not read y range for particlebox logger=",1,sval=(/splitline(2)/),margin=1)    
    numpbox = numpbox - 1
    return
104 call message("could not read z range for particlebox logger=",1,sval=(/splitline(2)/),margin=1)    
    numpbox = numpbox - 1
  end subroutine add_boxlogger

  
  subroutine particleboxlogger()
    implicit none
    integer part,i,j,b,ios,tsunit
    real x,y,z, r(4)
    logical unit_is_open, file_exist, inbox
    character(100) datafile
    
    do b=1,numpbox
        if(pbox(b)%inblock ) then
            ! this unit should be open. also make sure the line is
            ! equal to that in the routine OPEN_BOXLOGGER_FILES:
            tsunit = 4000000 + 1000*myid + b
  
            do part = 1, index_activated
                if(particle(part)%status == myid) then
                    x = particle(part)%pos(1)
                    y = particle(part)%pos(2)
                    r = pbox(b)%r(1:4)
                    inbox = .false.
                    
                    if(x >= r(1) .and. x < r(2) .and. y>= r(3) .and. y < r(4)) then
                    ! particle is in horisontal range of box
                        i = int(x + 1.) - block%fi + 1
                        j = int(y + 1.) - block%fj + 1
                        if(jm==1) j=1
                        z = particle(part)%pos(3)*d(i,j) + eta(i,j)

                        if(z >= pbox(b)%r(5) .and. z < pbox(b)%r(6)) then
                            inbox = .true.
                            
                            if(.not. btest(particle(part)%box, b)) then
                              ! particle is entering box. set bit and log                                
                                particle(part)%box = ibset(particle(part)%box, b)
                                write(tsunit,'(f0.1,1x,i7.7,1x,i0)') istep*dt,part,1
                            end if                            
                        end if
                    end if

                    if(.not. inbox) then
                        if(btest(particle(part)%box, b)) then
                            ! particle just left box, clear box bit and log
                            particle(part)%box = ibclr(particle(part)%box, b)
                            write(tsunit,'(f0.1,1x,i7.7,1x,i0)') istep*dt,part,0
                        end if
                    end if
                end if
            end do
        end if
    end do
  end subroutine particleboxlogger


  subroutine timeserie_part(label,partnr)
  !
  ! Timeseries of particle movement.
  ! The routine follows the particle partnr
  ! until it is taken out of the simulation.

    implicit none
    character(len=*), intent(in) :: label
    integer, intent(in) :: partnr
    real(dp), dimension(3) :: xyz

    character(len=60) :: datafile
    logical :: file_is_open, file_exist
    integer :: tsunit, ios, numioerr=0

    
! Check if the particle is active
    if (particle(partnr)%status.eq.myid) then

        tsunit = 2000000 + 1000*myid + partnr            
        inquire(unit=tsunit, opened=file_is_open)
        
        if(.not. partplot_binary)then
            
            write(datafile,'(a,"-",i8.8,"-",i5.5,"x",i5.5,"x",i3.3,"-P",i4.4,".data")')  & 
                 label,partnr,im,jm,kb-1,myid
            
            if (.not.file_is_open) then
                inquire(file=(trim(particlets_dir)//trim(datafile)), exist=file_exist)

                if(file_exist) then
                    open(unit=tsunit, file=(trim(particlets_dir)//trim(datafile)), status="old", position="append", &
                         err=101, iostat=ios)
                else
                    open(unit=tsunit, file=(trim(particlets_dir)//trim(datafile)), err=101, iostat=ios)
                end if
            end if
            
            call get_xyz_point(particle(partnr)%pos,xyz,particle(partnr)%block)
            
            write(tsunit, '((f0.4,3(1x,f0.10)))') float(istep)*dt, xyz(1), xyz(2), xyz(3)
        else

! binary output

            write(datafile,'(a,"-",i8.8,"-",i5.5,"x",i5.5,"x",i3.3,"-P",i4.4,".sr8")')  & 
                 label,partnr,im,jm,kb-1,myid
            
            if(.not.file_is_open) then
                inquire(file=(trim(particlets_dir)//trim(datafile)), exist=file_exist)
                
                if(file_exist) then
                    open(unit=tsunit, file=(trim(particlets_dir)//trim(datafile)), &
                         status="old", position="append", access="stream", form="unformatted", err=101, iostat=ios)
                else
                    open(unit=tsunit, file=(trim(particlets_dir)//trim(datafile)),  &
                         access="stream", form="unformatted", err=101, iostat=ios)
                end if
            end if
            
            call get_xyz_point(particle(partnr)%pos,xyz,particle(partnr)%block)
            
            write(tsunit) real(istep,dp)*real(dt,dp), xyz
        end if
    end if
    return
    
101 numioerr = numioerr + 1
    if(debug > 1 .and. numioerr < 5) then
        write(*,*) "ERROR opening particle timeseries file; myid=",myid," iostat=",ios," part=",partnr,&
             " tsunit=",tsunit," datafile=",trim(datafile)
    end if
  end subroutine timeserie_part


  subroutine distribution_part
  !
  ! Routine for printing an overview over where the active particles 
  ! are located at a specific time.
  !
    implicit none
    character(len=200) :: datafile
    integer :: part,iproc
    integer :: i, j
    real(dp), dimension(:,:), allocatable :: part_data, part_data_dummy
    real(dp), dimension(3) :: xyz

    allocate(part_data(part_total,5))
    part_data = -1

    do part = 1,part_total
       if (particle(part)%status.ge.0) then
          call get_xyz_point(particle(part)%pos,xyz,myid)
          part_data(part,1) = particle(part)%time
          part_data(part,2) = xyz(1)
          part_data(part,3) = xyz(2)
          part_data(part,4) = xyz(3)
          part_data(part,5) = particle(part)%class          
       end if
    end do

    if (master) then
       allocate(part_data_dummy(part_total,5))
       do iproc = 1,(numproc-1)
          call receive(part_data_dummy,iproc)

          do part = 1,index_activated
             if (part_data_dummy(part,1).ne.-1) then
                part_data(part,1) = part_data_dummy(part,1)
                part_data(part,2) = part_data_dummy(part,2)
                part_data(part,3) = part_data_dummy(part,3)
                part_data(part,4) = part_data_dummy(part,4)
                part_data(part,5) = part_data_dummy(part,5)                
             end if
          end do
       end do
    else
       call send(part_data,0)
    end if

! Print to file the overview

    if(master) then
        if(.not. partplot_binary)then
            write(datafile,'("active-particles-",i8.8,".data")') istep
            open(unit=1300, file=(trim(particledist_dir)//trim(datafile)) )
            do part = 1, index_activated
                if (part_data(part,1).ge.0) then
                    write(1300,'(5(f19.10,1x))') part_data(part,1:5)
                end if
            end do
            close(1300)
        else
            j=0
            do part = 1,index_activated
                if (part_data(part,1).ge.0) then
                    j = j + 1
                    part_data_dummy(j,1:5) = part_data(part,1:5)
                end if
            end do

            if(j > 0) then
                write(datafile,'("active-particles-step",i8.8,"-",i0,"x5.sr8")') istep,j
                open(unit=1300, file=(trim(particledist_dir)//trim(datafile)), form="unformatted", access="stream")
                write(1300) part_data_dummy(1:j,1:5)
                close(1300)
            else
                if(debug > 1) write(*,'(2x,a)') "active particles=0, no particle distribution written"
            end if
        end if

        deallocate(part_data_dummy)
    end if
  end subroutine distribution_part


  subroutine update_status_data()
  ! 
  ! keep track of particle status on all procs
  !

    implicit none
    integer :: loop_part,iproc
    integer, dimension(:,:), allocatable :: status_data_dummy
    integer :: dummy_index, dummy_id
    integer :: encland, encbottom, encsurf, encbound, nrpart
    character(len=60) :: datafile
    logical :: file_is_there

    if(.not.allocated(status_data)) then
       allocate(status_data(part_total,2))
       status_data = -1
    end if

    allocate(status_data_dummy(part_total,2))

    encland = 0
    encbottom = 0
    encsurf = 0
    encbound = 0

    if(part_outputlevel.gt.2) then ! debug level
    ! check if particle is claimed by more than one processor
       status_data = -1

       do loop_part = 1,index_activated
          if ((particle(loop_part)%status.ge.-40) .and.  &
               (particle(loop_part)%status.le.-10)) then
             status_data(loop_part,1) = myid
             status_data(loop_part,2) = particle(loop_part)%status
             if (master) then
                if(particle(loop_part)%status.eq.-10) encbottom = encbottom+1
                if(particle(loop_part)%status.eq.-20) encsurf = encsurf+1
                if(particle(loop_part)%status.eq.-30) encland = encland+1
                if(particle(loop_part)%status.eq.-40) encbound = encbound+1
             end if
          else if (particle(loop_part)%status.eq.myid) then
             status_data(loop_part,1) = myid
             status_data(loop_part,2) = particle(loop_part)%block
          else
             status_data(loop_part,:) = -1
          end if
       end do

       if(master) then
          do iproc = 1,(numproc-1)
             status_data_dummy = -1
             call receive(status_data_dummy,iproc)
             
             do loop_part = 1,index_activated
                if (status_data_dummy(loop_part,1).gt.-1) then
                   if(status_data(loop_part,1).ne.-1) then
                      write(*,*) 'ERROR: Particle claimed by more than one processor'
                      write(*,*) 'proc1: ',status_data(loop_part,1),  &
                           status_data(loop_part,2)
                      write(*,*) 'proc2: ',status_data_dummy(loop_part,1),  &
                           status_data_dummy(loop_part,2)
                      call abort()
                   else
                      status_data(loop_part,1) = status_data_dummy(loop_part,1)
                      status_data(loop_part,2) = status_data_dummy(loop_part,2)
                      if(status_data(loop_part,2).eq.-10) encbottom = encbottom+1
                      if(status_data(loop_part,2).eq.-20) encsurf = encsurf+1
                      if(status_data(loop_part,2).eq.-30) encland = encland+1
                      if(status_data(loop_part,2).eq.-40) encbound = encbound+1
                   end if
                end if
             end do
          end do
       else
          call send(status_data,0)
       end if
    else
    ! just update master if something has changed since last update
       dummy_index = 0

       do loop_part = 1,index_activated
          if ((particle(loop_part)%status.ge.-40) .and.  &
               (particle(loop_part)%status.le.-10)) then
             if(status_data(loop_part,2).ne.particle(loop_part)%status) then
                status_data(loop_part,1) = myid
                status_data(loop_part,2) = particle(loop_part)%status
                if (.not.master) then
                   dummy_index = dummy_index + 1
                   status_data_dummy(dummy_index,1) = particle(loop_part)%id
                   status_data_dummy(dummy_index,2) = particle(loop_part)%status
                end if
             end if
          else if (particle(loop_part)%status.eq.myid) then
             status_data(loop_part,1) = myid
             status_data(loop_part,2) = particle(loop_part)%block
             if(status_data(loop_part,2).ne.particle(loop_part)%status) then
                status_data(loop_part,1) = myid
                status_data(loop_part,2) = particle(loop_part)%status
                if (.not.master) then
                   dummy_index = dummy_index + 1
                   status_data_dummy(dummy_index,1) = particle(loop_part)%id
                   status_data_dummy(dummy_index,2) = particle(loop_part)%status
                end if
             end if
          end if
          if(master) then
             if(status_data(loop_part,2).eq.-10) encbottom = encbottom+1
             if(status_data(loop_part,2).eq.-20) encsurf = encsurf+1
             if(status_data(loop_part,2).eq.-30) encland = encland+1
             if(status_data(loop_part,2).eq.-40) encbound = encbound+1
          end if
       end do

       if(master) then
          do iproc = 1,(numproc-1)
             call receive(dummy_index,iproc)
             
             do loop_part = 1,dummy_index
                call receive(status_data_dummy(loop_part,1),iproc)
                call receive(status_data_dummy(loop_part,2),iproc)
                
                dummy_id = status_data_dummy(loop_part,1)
                status_data(dummy_id,1) = iproc
                status_data(dummy_id,2) = status_data_dummy(loop_part,2)
                if(status_data(dummy_id,2).eq.-10) encbottom = encbottom+1
                if(status_data(dummy_id,2).eq.-20) encsurf = encsurf+1
                if(status_data(dummy_id,2).eq.-30) encland = encland+1
                if(status_data(dummy_id,2).eq.-40) encbound = encbound+1
             end do
          end do
       else
          call send(dummy_index,0)
          do loop_part = 1,dummy_index
             call send(status_data_dummy(loop_part,1),0)
             call send(status_data_dummy(loop_part,2),0)
          end do
       end if
    end if

    ! Print particle data to file and screen (optional)
    if(master) then
       if(part_outputlevel.gt.0) then
          write(*,*) 'Activated particles:  ',index_activated
          write(*,*) 'Captured at bottom:   ',encbottom
          write(*,*) 'Captured at surface:  ',encsurf
          write(*,*) 'Captured at land:     ',encland
          write(*,*) 'Captured at boundary: ',encbound
       end if
       nrpart = encland + encbottom + encsurf + encbound

       if(part_check_status) then 
          write(datafile,'("encounter-timeserie-",i5.5,"x",i5.5,".data")') im,jm
          inquire( file=trim(datafile), exist=file_is_there )

          if (.not.file_is_there) then
             open(unit=55555, file=(trim(particledist_dir)//trim(datafile)) )
          end if

          write(55555, '((1x,f12.4,1x,5(i7.7,1x)))') float(istep)*dt, nrpart,  &
               encbottom, encsurf, encland, encbound
               
          if (istep.eq.numstep) then
             close(55555)
          end if
       end if
    end if

    deallocate(status_data_dummy)
 
  end subroutine update_status_data


  subroutine encounter_loc
  !
  ! Routine for printing an overview over
  ! where the particles have been taken out.
  !
  ! Find the location where the particles 
  ! became passive. 
  !
    implicit none
    character(len=60) :: datafile
    integer :: x_loc, y_loc
    integer :: part,iproc
    integer :: i, j
    integer, dimension(:,:), allocatable :: erate_bottom, erate_surface
    integer, dimension(:,:), allocatable :: erate_land, erate_bound
    real(dp), dimension(:,:), allocatable :: encounter_data, encounter_data_dummy
    real(dp), dimension(3) :: xyz
    character(6) classstr
    character(8) stepstr
    integer class,nsurf,nbot,nbound,nland
    
    allocate(encounter_data(part_total,5))
    encounter_data = -1

    do part = 1,part_total
       if (status_data(part,1) .eq. myid) then
          call get_xyz_point(particle(part)%pos,xyz,myid)
          encounter_data(part,1) = particle(part)%time
          encounter_data(part,2) = xyz(1)
          encounter_data(part,3) = xyz(2)
          encounter_data(part,4) = xyz(3)
          encounter_data(part,5) = particle(part)%class
       end if
    end do

    if (master) then
       allocate(encounter_data_dummy(part_total,5))
       do iproc = 1,(numproc-1)
          call receive(encounter_data_dummy,iproc)

          do part = 1,index_activated
             if (status_data(part,1).eq.iproc) then
                encounter_data(part,1) = encounter_data_dummy(part,1)
                encounter_data(part,2) = encounter_data_dummy(part,2)
                encounter_data(part,3) = encounter_data_dummy(part,3)
                encounter_data(part,4) = encounter_data_dummy(part,4)
                encounter_data(part,5) = encounter_data_dummy(part,5)                
             end if
          end do
       end do
       deallocate(encounter_data_dummy)

    else
       call send(encounter_data,0)
    end if

    if(master) then
       allocate( erate_bottom(im,jm), erate_surface(im,jm),  &
            erate_land(im,jm), erate_bound(im,jm) )

       do class=1,max(1,part_numclasses)
           erate_bottom=0
           erate_surface=0
           erate_land=0
           erate_bound=0
           x_loc=0
           y_loc=0

           nsurf=0; nbot=0 ; nbound=0; nland=0
           
           do part = 1, index_activated
               if(part_numclasses==0 .or. particle(part)%class == matnumbers(class)) then
                   if(status_data(part,2).eq.-10) then
                       x_loc = int(encounter_data(part,2)/dx) + 1
                       y_loc = int(encounter_data(part,3)/dy) + 1
                       erate_bottom(x_loc,y_loc)=erate_bottom(x_loc,y_loc)+1
                       nbot=nbot+1
                   else if(status_data(part,2).eq.-20) then
                       x_loc = int(encounter_data(part,2)/dx) + 1
                       y_loc = int(encounter_data(part,3)/dy) + 1
                       erate_surface(x_loc,y_loc)=erate_surface(x_loc,y_loc)+1
                       nsurf=nsurf + 1
                   else if(status_data(part,2).eq.-30) then
                       x_loc = int(encounter_data(part,2)/dx) + 1
                       y_loc = int(encounter_data(part,3)/dy) + 1
                       erate_land(x_loc,y_loc)=erate_land(x_loc,y_loc)+1
                       nland=nland+1
                   else if(status_data(part,2).eq.-40) then
                       x_loc = int(encounter_data(part,2)/dx) + 1
                       y_loc = int(encounter_data(part,3)/dy) + 1
                       erate_bound(x_loc,y_loc)=erate_bound(x_loc,y_loc)+1
                       nbound=nbound+1
                   end if
               end if
           end do

           classstr=""
           if(part_numclasses > 0)then
               write(classstr,'("-c",i4.4)') matnumbers(class)
           end if
           write(stepstr,'(i8.8)') istep
           
! Print to file the overview (if there is something to report)
           
           if(nbot > 0) then
               write(datafile,'("encounter-bottom",a,"-is",a,"-",i5.5,"x",i5.5,".data")') trim(classstr),stepstr,im,jm
               open(unit=1310, file=(trim(particledist_dir)//trim(datafile)) )
               write(1310,'(i5.5,1x,i5.5,1x,i6.6)') ((i, j, erate_bottom(i,j),i=1,im),j=1,jm)
               close(1310)
           end if

           if(nsurf>0)then
               write(datafile,'("encounter-surface",a,"-is",a,"-",i5.5,"x",i5.5,".data")') trim(classstr),stepstr,im,jm
               open(unit=1320, file=(trim(particledist_dir)//trim(datafile)) )
               write(1320,'(i5.5,1x,i5.5,1x,i6.6)') ((i, j, erate_surface(i,j),i=1,im),j=1,jm)
               close(1320)
           end if

           if(nland>0)then
               write(datafile,'("encounter-land",a,"-is",a,"-",i5.5,"x",i5.5,".data")') trim(classstr),stepstr,im,jm
               open(unit=1330, file=(trim(particledist_dir)//trim(datafile)) )
               write(1330,'(i5.5,1x,i5.5,1x,i6.6)') ((i, j, erate_land(i,j),i=1,im),j=1,jm)
               close(1330)
           end if

           if(nbound>0)then
               write(datafile,'("encounter-bound",a,"-is",a,"-",i5.5,"x",i5.5,".data")') trim(classstr),stepstr,im,jm
               open(unit=1340, file=(trim(particledist_dir)//trim(datafile)) )
               write(1340,'(i5.5,1x,i5.5,1x,i6.6)') ((i, j, erate_bound(i,j),i=1,im),j=1,jm)
               close(1340)
           end if
       end do
               
       deallocate( erate_bottom, erate_surface, erate_land, erate_bound )
    end if
  end subroutine encounter_loc


  subroutine update_pgpart_history()
    implicit none
    integer :: part,iproc
    integer :: i, j
    real(dp), dimension(:,:), allocatable :: part_data, part_data_dummy
    real(dp), dimension(3) :: xyz

    pgpart_history_count = index_activated

    allocate(part_data(index_activated,4))
    part_data = -1

    do part = 1,index_activated
       if (particle(part)%status.ge.0) then
          call get_xyz_point(particle(part)%pos,xyz,myid)
          part_data(part,1) = particle(part)%time
          part_data(part,2) = xyz(1)
          part_data(part,3) = xyz(2)
          part_data(part,4) = xyz(3)
      end if
    end do

    if (master) then
       allocate(part_data_dummy(index_activated,4))
       do iproc = 1,(numproc-1)
          call receive(part_data_dummy,iproc)

          do part = 1,index_activated
              if (part_data_dummy(part,1).ne.-1)  part_data(part,1:4) = part_data_dummy(part,1:4)
          end do
       end do
       deallocate(part_data_dummy)
    else
       call send(part_data,0)
    end if

    if (master) then
       do part=1,index_activated
          ! Prepare pgpart_history array for new data
          if (pgpart_history(pgpart_ntime_pts,1,part) .ne. -1) then
             do j=1,3
                do i=1,pgpart_ntime_pts-1
                   if (pgpart_history(i,j,part) .eq. -1) then
                      pgpart_history(i,j,part) =  &
                           pgpart_history(pgpart_ntime_pts,j,part)
                   else
                      pgpart_history(i,j,part) = pgpart_history(i+1,j,part)
                   end if
                end do
            end do
          end if
          ! Update pgpart_history array
          if (part_data(part,1).ne.-1) then
             do j=1,3
                pgpart_history(pgpart_ntime_pts,j,part) = part_data(part,j+1)
            end do

          else if (pgpart_history(pgpart_ntime_pts-1,1,part) .ne. -1) then
             do j=1,3
                do i=1,pgpart_ntime_pts
                   pgpart_history(i,j,part) = -1
                end do
            end do
            
         end if
         
       end do
    end if
    
  end subroutine update_pgpart_history


  subroutine update_pgpart_capture(part_status_updated)
    implicit none
    logical :: part_status_updated
    integer :: part,iproc
    integer :: i, j
    real(dp), dimension(:,:), allocatable :: encounter_data, encounter_data_dummy
    real(dp), dimension(3) :: xyz

    pgpart_capture_count = index_activated

    if (.not.part_status_updated) then
       call update_status_data
       part_status_updated = .true.
    end if

    allocate(encounter_data(part_total,4))
    encounter_data = -1

    do part = 1,index_activated
       if (status_data(part,1) .eq. myid) then
          call get_xyz_point(particle(part)%pos,xyz,myid)
          encounter_data(part,1) = particle(part)%time
          encounter_data(part,2) = xyz(1)
          encounter_data(part,3) = xyz(2)
          encounter_data(part,4) = xyz(3)
       end if
    end do

    if (master) then
       allocate(encounter_data_dummy(part_total,4))
       do iproc = 1,(numproc-1)
          call receive(encounter_data_dummy,iproc)

          do part = 1,index_activated
             if (status_data(part,1).eq.iproc) then
                encounter_data(part,1) = encounter_data_dummy(part,1)
                encounter_data(part,2) = encounter_data_dummy(part,2)
                encounter_data(part,3) = encounter_data_dummy(part,3)
                encounter_data(part,4) = encounter_data_dummy(part,4)
             end if
          end do
       end do
       deallocate(encounter_data_dummy)
    else
       call send(encounter_data,0)
    end if

    if(master) then
       do part = 1, index_activated
          if(status_data(part,2) <= -10) then
             pgpart_capture(1,part) = status_data(part,2)
             pgpart_capture(2,part) = encounter_data(part,2)
             pgpart_capture(3,part) = encounter_data(part,3)
             pgpart_capture(4,part) = encounter_data(part,4)
          end if
       end do
    end if

  end subroutine update_pgpart_capture


!========================================================================
! --- clear particle data
!
! Subroutines for clearing and deallocating particle data
!
!========================================================================
  
  subroutine part_finalize
! shut down the particle model
    call close_boxlogger_files()
    call deallocate_particle()
  end subroutine part_finalize


  subroutine deallocate_particle()
  !
  ! Deallocate all arrays for particle data
  !
    implicit none
    integer i
    
    if (allocated(particle))           deallocate(particle)
    if (allocated(part_init_data))     deallocate(part_init_data)
    if (allocated(status_data))        deallocate(status_data)

    if (allocated(put_north))          deallocate(put_north)
    if (allocated(put_east))           deallocate(put_east)
    if (allocated(put_south))          deallocate(put_south)
    if (allocated(put_west))           deallocate(put_west)
    if (allocated(get_north))          deallocate(get_north)
    if (allocated(get_east))           deallocate(get_east)
    if (allocated(get_south))          deallocate(get_south)
    if (allocated(get_west))           deallocate(get_west)
    if (allocated(part_buffer_send1))  deallocate(part_buffer_send1)
    if (allocated(part_buffer_send2))  deallocate(part_buffer_send2)
    if (allocated(part_buffer_recv1))  deallocate(part_buffer_recv1)
    if (allocated(part_buffer_recv2))  deallocate(part_buffer_recv2)

    if (allocated(pgparts))            deallocate(pgparts)
    if (allocated(pgpart_history))     deallocate(pgpart_history)
    if (allocated(pgpart_capture))     deallocate(pgpart_capture)
    if (allocated(pgpart_color_array)) deallocate(pgpart_color_array)

    if (allocated(pbox)) deallocate(pbox)

    do i=1,part_maxnumclasses
        if(allocated(matter(i)%time))     deallocate(matter(i)%time)
        if(allocated(matter(i)%prop1))    deallocate(matter(i)%prop1)
        if(allocated(matter(i)%prop2))    deallocate(matter(i)%prop2)
        if(allocated(matter(i)%prop3))    deallocate(matter(i)%prop3)
        if(allocated(matter(i)%prop4))    deallocate(matter(i)%prop4)        
    end do    
  end subroutine deallocate_particle


!========================================================================
! --- checkpointing
!
! Subroutines for dumping particle setupfile, position, status, and 
! pgplot data to file 
!
!========================================================================

  subroutine part_checkpoint
  !
  ! Wrapper for file dumping routines
  !
    use checkpointing
    implicit none
    integer i

    if(master) call dump_part_setup

    if (.not.part_activate) return

    call dump_part_data
    
    if (part_use_pgplot) call dump_part_pgplot

    contains

      subroutine dump_part_setup
      !
      ! Recreate a working particle_setup.dat file, and store other 
      ! important parameters in the file particle_parameters.dat
      !
        use util, only : ensureendslash
        implicit none
        character(200) :: part_setup_file, part_param_file
        integer :: i

        call ensureendslash(checkpoint_dir)
        
        part_setup_file = (trim(checkpoint_dir)//"particle_setup.dat")
        
        open(unit=3000001, file=(trim(part_setup_file)))
        write(3000001,'(a)') '! -*-f90-*-'
        write(3000001,'(a)') '!'
        write(3000001,'(a)') '! Setup file for particles'
        write(3000001,'(a)') '!=========================='
        write(3000001,'(a)') '!'
        write(3000001,'(a)') ' '
        write(3000001,'(a)') 'part_restart        T'
        write(3000001,'(a, l2)') 'part_activate      ', part_activate        
        write(3000001,'(a, i0)') 'part_exchange_size ', part_exchange_size
        write(3000001,'(a, l2)') 'part_bilinear      ', part_bilinear
        write(3000001,'(a, l2)') 'part_runge_kutta4  ', part_runge_kutta4  
        write(3000001,'(a, l2)') 'part_h_is_positive ', part_h_is_positive 
        write(3000001,'(a)') ' '
        write(3000001,'(a, l2)') 'part_random_step   ', part_random_step
        write(3000001,'(a, l2)') 'part_random_seed   ', part_random_seed
        write(3000001,'(a, i4)') 'part_random_type   ', part_random_type
        write(3000001,'(a, g12.4)') 'part_ah            ', part_ah
        write(3000001,'(a, g12.4)') 'part_kh            ', part_kh
        write(3000001,'(a)') ' '
        
        write(3000001,'(a, i0)') 'part_vert_velocity          ',part_vert_velocity
        write(3000001,'(a, l2)') 'part_lookup_ws              ',part_lookup_ws
        write(3000001,'(a, l2)') 'part_lookup_sg_and_diameter ',part_lookup_sg_and_diameter
        write(3000001,'(a, l2)') 'part_apply_stokes           ',part_apply_stokes

        do i=1,part_numclasses
            write(3000001,'(a,a,1x,i0,a,a,a,i0 )') 'part_add_class ',&
                 trim(matter(i)%name), i, ' "',trim(matter(i)%file), '" ',matter(i)%numprop
        end do
        
        write(3000001,'(a, i0, a)') 'part_init_data 3 "particledata-',part_total,'x5.sr8"'
        write(3000001,'(a)') ' '
        write(3000001,'(a, l2)') 'part_check_status      ', part_check_status
        write(3000001,'(a, i8)') 'part_outputlevel       ', part_outputlevel
        write(3000001,'(a, i8)') 'part_write_time        ', part_write_time
        write(3000001,'(a, l2)') 'partplot_timeseries    ', partplot_timeseries
        write(3000001,'(a, i8)') 'partplot_ts_write      ', partplot_ts_write
        write(3000001,'(a, i8)') 'partplot_ts_mod        ', partplot_ts_mod
        write(3000001,'(a, i8)') 'partplot_ts_mod_bsize  ', partplot_ts_mod_bsize
        write(3000001,'(a, i8)') 'part_enc_dumpint       ', part_enc_dumpint

        do i=1,numpbox
            write(3000001,*) "part_boxlogger "//trim(pbox(i)%name), pbox(i)%r(1:4)*dx - dx, pbox(i)%r(5:6)
        end do
        write(3000001,'(a, i8)') 'part_boxlogger_interval ', boxloggerintval
        
        write(3000001,'(a, l2)') 'partplot_distribution  ', partplot_distribution
        write(3000001,'(a, i8)') 'partplot_dist_write    ', partplot_dist_write
        write(3000001,'(a, l2)') 'partplot_binary        ', partplot_binary
        write(3000001,'(a)') ' '
        write(3000001,'(a, i4, g12.4)') 'part_bound_east    ', part_bound_east_type,  &
             part_bound_offset_east
        write(3000001,'(a, i4, g12.4)') 'part_bound_west    ', part_bound_west_type,  &
             part_bound_offset_west
        write(3000001,'(a, i4, g12.4)') 'part_bound_north   ', part_bound_north_type,  &
             part_bound_offset_north
        write(3000001,'(a, i4, g12.4)') 'part_bound_south   ', part_bound_south_type,  &
             part_bound_offset_south
        write(3000001,'(a)') ' '
        write(3000001,'(a, g12.4)') 'part_minh          ', part_minh
        write(3000001,'(a, g12.4)') 'part_mineta        ', part_mineta
        write(3000001,'(a, g12.4)') 'part_init_minh     ', part_init_minh
        write(3000001,'(a, g12.4)') 'part_init_mineta   ', part_init_mineta
        write(3000001,'(a, g12.4)') 'part_minland       ', part_minland
        write(3000001,'(a)') ' '
        write(3000001,'(a, l2)') 'part_capture_h     ', part_capture_h
        write(3000001,'(a, l2)') 'part_capture_eta   ', part_capture_eta
        write(3000001,'(a, l2)') 'part_capture_land  ', part_capture_land
        write(3000001,'(a)') ' '
        if (part_use_pgplot) then
           write(3000001,'(a, l2, i8)') 'part_use_pgplot    ', part_use_pgplot,  &
                pgpart_num_plots
           write(3000001,'(a, i8)') 'pgpart_plot_time   ', pgpart_plot_time
           write(3000001,'(a, i8)') 'pgpart_ntime_pts   ', pgpart_ntime_pts
           do i=1,pgpart_num_plots
              if (pgparts(i)%set) then
                 write(3000001,'(a, 5i4)') 'pgpart_set_plot ',  &
                      pgparts(i)%ind, pgparts(i)%type, pgparts(i)%style,  &
                      pgparts(i)%color, pgparts(i)%line_width
              end if
           end do
           write(3000001,'(a)') ' '
        end if
        close(3000001)

        part_param_file = (trim(checkpoint_dir)//"particle_parameters.dat")
        open(unit=3000002, file=(trim(part_param_file)))
        write(3000002,*) index_activated
        write(3000002,*) pgpart_set_history
        write(3000002,*) pgpart_history_count
        write(3000002,*) pgpart_set_capture
        write(3000002,*) pgpart_capture_count
        close(3000002)
        
      end subroutine dump_part_setup


      subroutine dump_part_data
        implicit none

        character(200) :: datafile, statusfile
        integer :: part,iproc
        integer :: i, j
        real(dp), dimension(:,:), allocatable :: part_data, part_data_dummy
        integer, dimension(:,:), allocatable :: part_restart_status, part_restart_status_dummy
        real(dp), dimension(3) :: xyz
        character(10) dimstring
        
        allocate(part_data(part_total,5))
        allocate(part_restart_status(part_total,2))
        part_data = -1.
        part_restart_status = -1

        do part = 1,part_total

! active particles: recompute the position in real coordinates for the restart file
! if the particle is on the current block, give it a status and id            
           if (myid.eq.particle(part)%block) then
              part_restart_status(part,1) = particle(part)%id
              part_restart_status(part,2) = particle(part)%status
              if(particle(part)%status >= 0) then
                  call get_xyz_point(particle(part)%pos,xyz,myid)
                  part_data(part,1) = particle(part)%time
                  part_data(part,2) = xyz(1)
                  part_data(part,3) = xyz(2)
                  part_data(part,4) = xyz(3)
                  part_data(part,5) = particle(part)%class
              end if
          end if

          if(particle(part)%status == -1) then
! these particles have not been started/initialized yet. should save their state, which is on master.
              part_data(part,1) = particle(part)%time
              part_data(part,2:4) = particle(part)%pos
              part_data(part,5) = particle(part)%class              
          end if

! todo: should preserve also the captured particles during cp/restart(?). if so, add more checks for status here.
! these particles have their coordinate in which coordinates?
          if(particle(part)%status < -2) then
              part_restart_status(part,1) = particle(part)%id
              part_restart_status(part,2) = particle(part)%status
              
              part_data(part,1) = particle(part)%time
              part_data(part,2:4) = particle(part)%pos
              part_data(part,5) = particle(part)%class       
          end if
          
        end do

        ! Transfer status data to master, all except from particles that have not been activated
        if (master) then
           allocate(part_restart_status_dummy(part_total,2))
           do iproc = 1,(numproc-1)
              call receive(part_restart_status_dummy,iproc)

              do part = 1,part_total
!                  if (part_restart_status_dummy(part,1).ne.-1) then
! status is -2 if p is active on another proc. status is procid if active.                  
                  if (part_restart_status_dummy(part,2) .gt. -1) then
                      part_restart_status(part,1) = part_restart_status_dummy(part,1)
                      part_restart_status(part,2) = part_restart_status_dummy(part,2)
                  end if
              end do
           end do
           deallocate(part_restart_status_dummy)
        else
           call send(part_restart_status,0)
        end if

        call barrier()

! Transfer time and location data to master for all particles, except inactivate 
        if (master) then
           allocate(part_data_dummy(part_total,5))
           do iproc = 1,(numproc-1)
              call receive(part_data_dummy,iproc)

              do part = 1,part_total
                 if (part_data_dummy(part,1).ne.-1.) then
                    part_data(part,1) = part_data_dummy(part,1)
                    part_data(part,2) = part_data_dummy(part,2)
                    part_data(part,3) = part_data_dummy(part,3)
                    part_data(part,4) = part_data_dummy(part,4)
                    part_data(part,5) = part_data_dummy(part,5)                    
                 end if
              end do
           end do
           deallocate(part_data_dummy)
        else
           call send(part_data,0)
        end if

        ! Print to file the overview
        if(master) then
!            ...-NNNNx5.... a number surrounded by a "-" and "x5"
           write(dimstring,'(i0,"x5")') part_total 
           datafile = (trim(checkpoint_dir)//"particledata-"//trim(dimstring)//".sr8")
           open(unit=3000003, file=(trim(datafile)), form="unformatted", access="stream")
           write(3000003) part_data
           close(3000003)

           statusfile = (trim(checkpoint_dir)//"particle_status.si")
           open(unit=3000004, file=(trim(statusfile)), form="unformatted", access="stream")
           write(3000004) part_restart_status
           close(3000004)
        end if

      end subroutine dump_part_data


      subroutine dump_part_pgplot
        implicit none
        character(200) :: history_file, capture_file

        if (master) then
           if (pgpart_set_history) then
              history_file = (trim(checkpoint_dir)//"particle_pgpart_history.unfr")
              open(unit=3000005, file=(trim(history_file)), form="unformatted")
              write(3000005) pgpart_history
              close(3000005)
           end if

           if (pgpart_set_capture) then
              capture_file = (trim(checkpoint_dir)//"particle_pgpart_capture.unfr")
              open(unit=3000006, file=(trim(capture_file)), form="unformatted")
              write(3000006) pgpart_capture
              close(3000006)
           end if
        end if
      end subroutine dump_part_pgplot


  end subroutine part_checkpoint


  subroutine restart_particles
!
! Restart particles from checkpoint
!
    implicit none
    integer root

    integer :: i
    integer, dimension(:,:), allocatable :: part_restart_status
    real(dp), dimension(3) :: xyz
    character(len=80) :: part_param_file='./particle_parameters.dat'
    character(len=80) :: part_restart_status_file='./particle_status.si'
    character(len=80) :: part_history_file='./particle_pgpart_history.unfr'
    character(len=80) :: part_capture_file='./particle_pgpart_capture.unfr'
    integer numact, numina, numcap

! Load particle parameters
    if(master) then
        open(unit=4000002, file=(trim(part_param_file)), status="old", err=121)
        read(4000002,*) index_activated
        read(4000002,*) pgpart_set_history
        read(4000002,*) pgpart_history_count
        read(4000002,*) pgpart_set_capture
        read(4000002,*) pgpart_capture_count
        close(4000002)
    end if
    call broadcast(index_activated)
    call broadcast(pgpart_set_history)
    call broadcast(pgpart_history_count)
    call broadcast(pgpart_set_capture)
    call broadcast(pgpart_capture_count)

! Load particle status from checkpoint dump
    allocate(part_restart_status(part_total,2))

! Count active
    if(master) then
        open(unit=4000004, file=(trim(part_restart_status_file)), status="old", err=122, form="unformatted", access="stream")
        read(4000004, err=122) part_restart_status
        close(4000004)

        numact=0
        numina=0
        numcap=0
        do i=1,part_total
            if(part_restart_status(i,2) >= 0) numact=numact+1
            if(part_restart_status(i,2) < -2) numcap=numcap+1
            if(part_restart_status(i,2) == -1) numina=numina+1
        end do
        write(*,*) "particle status restart file read with ",numact,"active particles",numina,"not activated",numcap,"captured"
    end if

! Send particle status to all processes
    call broadcast(part_restart_status)

! reset particle data
    allocate(status_data(part_total,2))
    status_data = -1
    do i = 1,part_total
        particle(i)%id = part_restart_status(i,1)

        if(particle(i)%id > part_total .or. particle(i)%id < 1)then
            call abort("particle(i)%id out of valid range on restart")
        end if
        
! set status temporarily, this only shows if a particle is activated or not (the actual proc may be wrong!)
        particle(i)%status = part_restart_status(i,2)        
    end do

! get particle position data from part_init_data (part_init_data is on all processors just after read)
    do i = 1,part_total
        if (particle(i)%status .eq. -1) then
          ! not started, transfer position directly
            particle(i)%pos(1) = part_init_data(i,2)
            particle(i)%pos(2) = part_init_data(i,3)
            particle(i)%pos(3) = part_init_data(i,4)
        else
            xyz(1) = part_init_data(i,2)
            xyz(2) = part_init_data(i,3)
            xyz(3) = part_init_data(i,4)
            particle(i)%pos = 0.
            root = -1

            if((xyz(1) > 0. .and. xyz(1) < im*dx) .and.  &
                 (jm .eq. 1 .or. (xyz(2) > 0. .and. xyz(2) < jm*dy))) then
                call get_sigma_point(xyz,particle(i)%pos,root)
            else
                root = 0
            end if
            call reduce('max',root)
            particle(i)%block = root
        end if

        particle(i)%time = part_init_data(i,1)
        
        if(part_vert_velocity .eq. 1) particle(i)%ws    = part_init_data(i,5)
        if(part_vert_velocity .eq. 2) particle(i)%class = nint(part_init_data(i,5))
    end do

! %block is now up 2 date wrt coordinate of each particle
    do i = 1,part_total
        if (particle(i)%block .eq. myid) then                  ! particle pos from restart is on this processor  
            if (part_restart_status(i,2) > -1) then            ! particle is active
                particle(i)%status = myid                      
            else
                particle(i)%status = part_restart_status(i,2)  ! not active, but position is on processor. save its status
            end if
        else
            if (part_restart_status(i,2) > -1) then            ! partpos from restart not on this proc, but part was active
                particle(i)%status = -2
            else
                particle(i)%status = part_restart_status(i,2)  ! particle not on processor, keep status from restart file.
                status_data(i,2) = particle(i)%status
            end if
        end if
    end do
    deallocate(part_restart_status)


! Load pgplot data, and allocate needed arrays
    if (master .and. part_use_pgplot) then
        do i=1,pgpart_num_plots
            if (pgparts(i)%set) then
                select case(pgparts(i)%type)
                case(1)
                    allocate(pgparts(i)%x1(part_total))
                    allocate(pgparts(i)%y1(part_total))
                case(2)
                    if (pgpart_ntime_pts < 2) pgpart_ntime_pts = 2
                    allocate(pgparts(i)%x1(part_total))
                    allocate(pgparts(i)%y1(part_total))
                    allocate(pgparts(i)%x2(part_total))
                    allocate(pgparts(i)%y2(part_total))
                case(3)
                    allocate(pgparts(i)%x1(part_total))
                    allocate(pgparts(i)%y1(part_total))

                case(4)
                    allocate(pgparts(i)%x1(part_total))
                    allocate(pgparts(i)%y1(part_total))
                case(5)
                    allocate(pgparts(i)%x1(part_total))
                    allocate(pgparts(i)%y1(part_total))
                case(6)
                    if (pgpart_ntime_pts < 2) pgpart_ntime_pts = 2
                    allocate(pgparts(i)%x1(part_total))
                    allocate(pgparts(i)%y1(part_total))
                    allocate(pgparts(i)%x2(part_total))
                    allocate(pgparts(i)%y2(part_total))
                case default
                    write(*,*) '(plot; set; type): ', i, pgparts(i)%set, pgparts(i)%type
                    call abort('ERROR: Not a valid pgparts(i)%type value!')
                end select
            end if
        end do

        if (pgpart_set_history) then
            allocate(pgpart_history(pgpart_ntime_pts,3,part_total))
            open(unit=4000005, file=(trim(part_history_file)), status="old", err=123, form="unformatted")
            read(4000005, err=123) pgpart_history
            close(4000005)
        end if

        if (pgpart_set_capture) then
            allocate(pgpart_capture(4,part_total))
            open(unit=4000006, file=(trim(part_capture_file)), status="old", err=124, form="unformatted")
            read(4000006, err=124) pgpart_capture
            close(4000006)
        end if

        allocate(pgpart_color_array(pgpart_ntime_pts-1))
    end if

    return

121 call abort('part-track: could not open file for formatted io')   
122 call abort("could not read particle status data, restart failed")
    
! not the end of the world if for some reason these files are not in restart dir
    
123 write(*,*) "could not open pgplot particle history file, disabling"
    pgpart_set_history=.false.
    return
124 write(*,*) "could not open pgplot particle capture file, disabling"
    pgpart_set_capture=.false.
  end subroutine restart_particles


!========================================================================
! --- other subroutines
!
! * Check if the particle is within a valid region.
! * Transform between cartesian ('real') and sigma (grid) coordinates
!
!========================================================================


  subroutine check_land(part,sig)
  !
  ! Check if the particle is on land.
  ! Capture particles at land points if particle capturing is turned on.
  ! Bump back particles that are not captured near land.
  !
    implicit none

    integer :: part
    real(dp), dimension(3) :: sig
    integer :: ti,tj,dj
    real(dp) :: xmove,ymove

    ti = nint(sig(1) + 0.5)
    tj = nint(sig(2) + 0.5)
    dj = 1
    if(jm.eq.1) then
       tj = 1
       dj = 0
    end if
    

    if ((ti < i1) .or. (ti > i2) .or. (tj < j1) .or. (tj > j2)) then
        write(*,*) 'Particle, i, j: ', part, ti, tj, 'old pos=', particle(part)%pos(1:3)
        call message("New particle position not on block, particle deactivated",1)
        particle(part)%status = -60
    end if

    if (fsm(ti,tj).eq.0.0) then
!    if ((fsm(ti,tj).eq.0.0) .or. (fsm(ti-1,tj).eq.0.0) .or.  &
!         (fsm(ti,tj-dj).eq.0.0) .or. (fsm(ti-1,tj-dj).eq.0.0)) then
       if(part_capture_land) then
          particle(part)%status = -30
          particle(part)%time = real(istep,8)*real(dt,8)
       else
          xmove = sig(1) - particle(part)%pos(1)
          if(jm.ne.1) ymove = sig(2) - particle(part)%pos(2)

          ! Move East/West
          sig(1) = particle(part)%pos(1) - xmove*part_minland

          ! Move North/South
          if(jm.ne.1) sig(2) = particle(part)%pos(2) - ymove*part_minland

          ! Check if particle is still on land
          ti = nint(sig(1) + 0.5)
          tj = nint(sig(2) + 0.5)
          if (fsm(ti,tj).eq.0.0) then  ! particle bump failed
             particle(part)%status = -30
             particle(part)%time = real(istep,8)*real(dt,8)
          end if
       end if
    end if

  end subroutine check_land


  subroutine check_vert_dist(minh,mineta,capture_h,capture_eta,  &
       sig,root,status,ptime)
  !
  ! Check if the particle is within the valid vertical range.
  ! Capture particles at boundaries if particle capturing is turned on.
  ! Bump back particles that are not captured near boundaries.
  !
    implicit none

    real(dp) :: minh,mineta
    logical :: capture_h,capture_eta
    real(dp), dimension(3) :: sig
    integer :: root,status
    real(dp) :: ptime
    integer :: ti,tj,dj
    real(dp) :: del_x,del_y,h_interp,eta_interp
    real(dp) :: dist_minh, dist_mineta

    dj = 1
    if(jm.eq.1) dj = 0

    if(myid.eq.root)then
       if(part_bilinear) then
          call part_bilinear_calc(sig,hm=h_interp,etam=eta_interp)
       else
          ti = nint(sig(1) + 0.5)
          tj = nint(sig(2) + 0.5)
          if(jm.eq.1) tj = 1
          h_interp = h(ti,tj)
          eta_interp = eta(ti,tj)
       end if

       ! Assume position is on land if: h_interp + eta_interp = 0.
       if ((h_interp + eta_interp) .eq. 0.) then
          if(part_outputlevel.gt.1) write(*,'(i4,a)') myid,  &
               ': total water depth is zero! Remove particle.'
          status = -30
          ptime = float(istep)*dt
       else
          ! Distances in sigma coordinates
          dist_minh = (-(h_interp - minh) - eta_interp)/  &
               (h_interp + eta_interp) 
          dist_mineta = -mineta/(h_interp + eta_interp)

          ! Check bottom
          if(sig(3).lt.dist_minh) then
             if(capture_h) then
                status = -10
                ptime = float(istep)*dt
             else
                sig(3) = dist_minh
             end if
          end if

          ! Check surface
          if(sig(3).gt.dist_mineta) then
             if(capture_eta) then
                status = -20
                ptime = float(istep)*dt
             else
                sig(3) = dist_mineta
             end if
          end if
       end if

    end if

  end subroutine check_vert_dist


  subroutine get_sigma_point(xyz,sig,root)
  !
  ! Get sigma coordinates for point given in pysical coordinates xyz.
  ! z is assumed to be distance from the 0-level (free surface), given 
  ! either as positive or negative real numbers.
  !
    implicit none
    real(dp), dimension(3) :: xyz,sig
    integer :: root
    logical :: inblock,ghost
    integer :: ti,tj,dj
    real(dp) :: del_x,del_y,D_interp,eta_interp

    dj = 1
    if(jm.eq.1) dj = 0

    if (xyz(1).lt.0.) call abort('ERROR: x < 0.')
    if (xyz(2).lt.0.) call abort('ERROR: y < 0.')

    sig(1) = xyz(1)/dx + 1.0
    if(jm.gt.1) then
       sig(2) = xyz(2)/dy + 1.0
    else
       sig(2) = 1.
    end if

    ! Bilinear interpolation
    ti = nint(sig(1)+0.5)
    tj = nint(sig(2)+0.5)
    if(jm.eq.1) tj = 1
    call globalind2local(ti,tj,inblock,ghost=.false.)
    if(inblock)then
       root = myid

       sig(1) = sig(1) - (block%fi-1)
       sig(2) = sig(2) - (block%fj-1)
       sig(3) = -2.
       call part_bilinear_calc(sig,Dm=D_interp,etam=eta_interp)
       if(part_h_is_positive) then
          xyz(3) = -xyz(3)
       end if
       if (abs(D_interp).ne.0) then
          sig(3) = (xyz(3) - eta_interp)/D_interp
       end if
    else
       root = -1
       sig(1) = -1.
       sig(2) = -1.
       sig(3) = -2.
    end if
    
  end subroutine get_sigma_point


  subroutine get_xyz_point(sig,xyz,root)
  !
  ! Get physical coordinates xyz for point given in sigma coordinates.
  !
    implicit none
    real(dp), dimension(3) :: sig,xyz
    integer :: root
    integer :: ti,tj,dj
    real(dp) :: del_x,del_y,D_interp,eta_interp

    dj = 1
    if(jm.eq.1) dj = 0

    xyz = -1.

    ! Bilinear interpolation
    if(myid.eq.root)then
       call part_bilinear_calc(sig,Dm=D_interp,etam=eta_interp)
       
       xyz(1) = (sig(1) + (block%fi-1) - 1.0)*dx
       if(jm.gt.1) then
          xyz(2) = (sig(2) + (block%fj-1) - 1.0)*dy
       else 
          xyz(2) = 0.
       end if
       xyz(3) = sig(3)*D_interp + eta_interp
       if(part_h_is_positive) then
          xyz(3) = -xyz(3)
       end if
    end if
    
  end subroutine get_xyz_point


  subroutine part_bilinear_calc(sig,Dm,hm,etam,etapm,um,vm,wm)
  !
  ! bilinear interpolation along sigma layers.
  !
    implicit none

    real(dp), intent(in), dimension(3) :: sig
    real(dp), optional :: Dm,hm,etam,etapm,um,vm,wm
    
    real(dp) gx,gy
    integer i,j,k,k0, ui,uj, vi,vj, ci, cj
    real(dp), dimension(4) :: c,cu,cv
    real(dp) lx,ly,uly,ulx,vlx,vly
    real(dp) :: del_z
    integer dj

    dj = 1
    if(jm.eq.1) dj = 0

    gx= sig(1)
    gy= sig(2)

    !! Find which vertical grid cell containing the particle
    if((present(um) .and. present(vm)) .or. (present(wm))) then
       k0=int(-sig(3)*float(kb-1))
       k=bisearch(real(-z,dp),kb,-sig(3),k0)
       if (k.eq.0) k=1
       del_z = (z(k) - sig(3))/dz(k)
    end if

    !! closest cell south west
    i=nint(gx)
    j=nint(gy)

    !! cell of coordinate
    ci=nint(gx+0.5)
    cj=nint(gy+0.5)

    lx=gx-(i-0.5)
    ly=gy-(j-0.5)

    if(jm.eq.1) then
       j = 1
       cj = 1
       gy = 0.0
       ly = 0.0
    end if

    if(present(um) .and. present(vm)) call find_ucoeffs()
    call find_coeffs()


    if(present(Dm)) then
       Dm = c(1)*(h(i,j)+eta(i,j)) + c(2)*(h(i+1,j)+eta(i+1,j)) +  &
            c(3)*(h(i+1,j+dj)+eta(i+1,j+dj)) + c(4)*(h(i,j+dj)+eta(i,j+dj))
    end if
    if(present(hm)) then
       hm = c(1)*h(i,j) + c(2)*h(i+1,j) + c(3)*h(i+1,j+dj) + c(4)*h(i,j+dj)
    end if
    if(present(etam)) then
       etam = c(1)*eta(i,j) + c(2)*eta(i+1,j) +  &
            c(3)*eta(i+1,j+dj) + c(4)*eta(i,j+dj)
    end if
    if(present(etapm)) then
       etapm = c(1)*etap(i,j) + c(2)*etap(i+1,j) +  &
            c(3)*etap(i+1,j+dj) + c(4)*etap(i,j+dj)
    end if
    if(present(um) .and. present(vm)) then
       um=cu(1)*u(ui,uj,k)+cu(2)*u(ui+1,uj,k)+cu(3)*u(ui+1,uj+dj,k)+cu(4)*u(ui,uj+dj,k)
       vm=cv(1)*v(vi,vj,k)+cv(2)*v(vi+1,vj,k)+cv(3)*v(vi+1,vj+dj,k)+cv(4)*v(vi,vj+dj,k)
    end if
    if(present(wm)) then
       !! Use trilinear interpolation for vertical velocity
       if (k.ge.kb-1) then
          wm=  c(1)*w(i,j,kb-1) +c(2)*w(i+1,j,kb-1)  &
               +c(3)*w(i+1,j+dj,kb-1) +c(4)*w(i,j+dj,kb-1)
       else
          wm=(1-del_z)*(c(1)*w(i,j,k) +c(2)*w(i+1,j,k)     &
               +c(3)*w(i+1,j+dj,k) +c(4)*w(i,j+dj,k))      &
               +del_z*(c(1)*w(i,j,k+1) +c(2)*w(i+1,j,k+1)  &
               +c(3)*w(i+1,j+dj,k+1) +c(4)*w(i,j+dj,k+1))
       end if   
    end if

  contains

    subroutine find_ucoeffs
      implicit none
      
      ui=int(gx+1.0)
      uj=j
      ulx=gx-(i-1.0)
      uly=ly
        
      vi=i
      vj=int(gy+1.0)
      vlx=lx
      vly=gy-(j-1.0)

      cu(1)=(1.-ulx)*(1.-uly)
      cu(2)=ulx*(1.-uly)
      cu(3)=ulx*uly
      cu(4)=(1.-ulx)*uly

      cv(1)=(1.-vlx)*(1.-vly)
      cv(2)=vlx*(1.-vly)
      cv(3)=vlx*vly
      cv(4)=(1.-vlx)*vly      
    end subroutine find_ucoeffs
    
    subroutine find_coeffs
      implicit none
      
      integer ifsm(4),fsmsum

      !! default coeffs. u and v coeffs are valid also with mixed
      !! ocean/land pts
      
      c(:) = 0.0
      if(fsm(ci,cj) == 0.0) return

      ifsm(1) = nint(fsm(i,j))
      ifsm(2) = nint(fsm(i+1,j))
      ifsm(3) = nint(fsm(i+1,j+dj))
      ifsm(4) = nint(fsm(i,j+dj))
      fsmsum = ifsm(1) + ifsm(2) + ifsm(3) + ifsm(4)
      
      select case(fsmsum)
      case(4)
          !! bilinear interpolation
          c(1) = (1.-lx)*(1.-ly)
          c(2) = lx*(1.-ly)
          c(3) = lx*ly
          c(4) = (1.-lx)*ly
      case(3)
          !! interpolation using plane defined by the 3 wet points

          if(ifsm(1) == 0) then
              c(2) = 1.-ly ; c(3)=lx+ly ; c(4)=-lx
          else if(ifsm(2) == 0) then
              c(1)=1.-ly ; c(3)=lx ; c(4)=ly-lx
          else if(ifsm(3) == 0) then
              c(1)=1-lx-ly ; c(2)=lx ; c(4)=ly
          else if(ifsm(4) == 0) then
              c(1)=1.-lx ; c(2)=lx-ly ; c(3)=ly
          end if
      case(2)
          !! linear interpolation using the 2 wet points

          if(ifsm(1) + ifsm(2) == 2) then
              c(1)=1.-lx ; c(2)=lx
          else if(ifsm(2) + ifsm(3) == 2) then
              c(2)=1.-ly ; c(3)=ly
          else if(ifsm(3) + ifsm(4) == 2) then
              c(3)=lx ; c(4)=1.0-lx
          else if(ifsm(1) + ifsm(4) == 2) then
              c(1)=1.-ly ; c(4)=ly
          else            
              !! f1 and f3 or f2 and f4 active: treat as single point case
              i=ci ; j=cj
              c(1)=1.0
          end if

      case(1)
          !! one active point? extrapolate
          i=ci ; j=cj
          c(1)=1.0

      case default
          write(*,*)'??? sum(fsm)=',fsmsum,' i,j=',i,j          
      end select
    end subroutine find_coeffs

  end subroutine part_bilinear_calc


!========================================================================
! --- subfunctions
!
! * Search for index (k) corresponding to vertical coordinate
! * Get a random number from a Gaussian distribution with
!   zero mean and unit variance
!
!========================================================================

  function bisearch(X, n, x0, k0) result(k)
  ! bisection search
  !
  ! Given an sorted array X with n elements,
  ! find k such that
  !    X(k) <= x0 < X(k+1), k = 1, ..., n-1
  ! if x0 < X(1)  return k = 0
  ! if x0 >= X(n) return k = n
  ! 
  ! k0 is an initial guess for k
  ! 
  ! Modified from the hunt search routine in 
  ! Numerical Recipes (Press et al.)
  !

    integer :: k
    integer, intent(in) :: n
    real(dp), dimension(n), intent(in) :: X
    real(dp), intent(in) :: x0
    integer, intent(in) :: k0
    integer :: m1, m2, m

    if ((k0 <= 0) .or. (k0 > n)) then
       m1 = 0
       m2 = n+1
    else if(X(k0) <= x0) then
       m1 = k0
       m2 = n+1
    else
       m1 = 0
       m2 = k0
    end if

    do while ((m2 - m1) > 1)
       m = (m1 + m2) / 2
       if (X(m) <= x0) then
          m1 = m
       else
          m2 = m
       end if
    end do
    k = m1
    return

  end function bisearch


  function gran() result(random)
  !
  !***BEGIN PROLOGUE GRAN
  !***DATE WRITTEN    25 April 1993
  !***LAST MODIFIED   02 January 1999
  !***AUTHOR
  !            Bjoern AAdlandsvik
  !            Institute of Marine research
  !            Postbox 1870 Nordnes
  !            N-5024 Bergen, Norway
  !            Internet: bjorn@imr.no 
  !
  !***PURPOSE GRAN
  !  Compute a normal distributed random number  
  !  with zero mean and unit variance.
  !
  !
  !***DESCRIPTION GRAN
  !  Uses the Box-Muller transform to compute two gaussian random
  !  numbers from uniformly distributed random numbers generated
  !  by the internal subroutine RANDOM_NUMBER.
  !
  !  One of the numbers is returned, while the other is stored
  !  for the next call to GRAN.
  !  The standard uniform random number generator random_number
  !  is
  !
  !  This routine is coded in standard FORTRAN 77 except for the
  !  use of the non-standard internal function RAND.
  !
  !
  !***REFERENCES    ! Return random value
!
! --- Local variables.
!

  !  Press, Flannery, Teukolsky, Vetterling:
  !    Numerical Recipes,
  !    The Art of Scientific Computing
  !      (FORTRAN Version)
  !    Cambridge University Press 1989
  !
  !
    implicit none
    !
    ! --- Arguments.
    !
    real(dp) :: random   ! Return random value
    !
    ! --- Local variables.
    !
    logical, save :: have_value = .FALSE.
    ! Flag for existence of already computed value
    real(dp), dimension(2) :: v
    ! Uniformly distributed random vector
    real(dp) :: r
    ! Square of the length of the vector v
    real(dp) :: bmfac
    ! Multiplicative factor in Box-Muller transform
    real(dp), save :: stored_value
    ! Storage for extra random number
    !
    !  save have_value, stored_value
    !  data have_value /.FALSE./
    !
    !***END DECLARATIONS GRAN

    if (have_value) then
       have_value = .FALSE.
       random = stored_value
    else
       have_value = .TRUE.

       !
       ! Compute a uniformly distributed random vector
       ! on the punctured unit disc by rejection.
       ! 
       do
          call random_number(v)  ! Uniform on [0, 1] x [0, 1]
          v = 2.0 * v - 1.0      ! Uniform on [-1,1] x [-1,1]
          r = v(1)**2 + v(2)**2
          if ((r < 1.) .AND. (r > 0.)) exit ! in punctured disc 
       end do

       !
       ! --- Box-Muller transform.
       !
       bmfac = sqrt(-2.0*log(r)/r)
       random        = v(1)*bmfac
       stored_value  = v(2)*bmfac

    end if ! (have_value

  end function gran


end module mod_particle
