module mod_rampup
!  
! Ramp-up boundary condition:  
! The boundary condition array will be linearly interpolated from
! state 1 to 2 in the given time interval. Before the interval the BV
! is state 1, after state 2.
!  
  integer :: num_rampup=0
  integer, parameter :: max_rampup=20
  
  type rampup_object
     logical :: inblock
     integer :: range(6), globrange(6)
     real :: rampup_time=0.0
     real :: dx=0, dy=0, dz=0, ofx=0, ofy=0, ofz=0
     integer :: idim=0, jdim=0, kdim=0
     real, allocatable :: x(:), z(:)
     real, allocatable :: s1f2d(:,:), s1f3d(:,:,:)  ! data state 1
     real, allocatable :: s2f2d(:,:), s2f3d(:,:,:)  ! data state 2
     real, pointer :: f2d(:,:) => null(), f3d(:,:,:) => null()
  end type rampup_object

  type(rampup_object), save :: rampup(max_rampup)

  
contains

  
  subroutine apply_rampups()
    use time_util    
    implicit none    
    integer ru, r(6)    
    real weight
    
    do ru=1,num_rampup

        if(rampup(ru)%inblock)then
            r = rampup(ru)%range

            if(time >= rampup(ru)%rampup_time) then
                weight = 1.0
            else if(time>=0.0 .and. time < rampup(ru)%rampup_time) then
                weight = (rampup(ru)%rampup_time - time)/rampup(ru)%rampup_time
            else
                weight = 0.0
            end if
            
            if(associated(rampup(ru)%f2d))then
                rampup(ru)%f2d(r(1):r(2),r(3):r(4)) = weight*rampup(ru)%s1f2d(r(1):r(2),r(3):r(4)) &
                                               + (1.-weight)*rampup(ru)%s2f2d(r(1):r(2),r(3):r(4))
            else if(associated(rampup(ru)%f3d))then
                rampup(ru)%f3d(r(1):r(2),r(3):r(4),r(5):r(6)) = weight*rampup(ru)%s1f3d(r(1):r(2),r(3):r(4),r(5):r(6)) &
                                                         + (1.-weight)*rampup(ru)%s2f3d(r(1):r(2),r(3):r(4),r(5):r(6))
            end if
        end if
    end do
  end subroutine apply_rampups 

  subroutine deallocate_rampups()
    integer ru
    do ru=1,num_rampup
        if(allocated(rampup(ru)%s1f2d)) deallocate(rampup(ru)%s1f2d)
        if(allocated(rampup(ru)%s1f3d)) deallocate(rampup(ru)%s1f3d)
        if(allocated(rampup(ru)%s2f2d)) deallocate(rampup(ru)%s2f2d)
        if(allocated(rampup(ru)%s2f3d)) deallocate(rampup(ru)%s2f3d)
    end do
  end subroutine deallocate_rampups
  
end module mod_rampup
