module mod_transport
!========================================================================
! Module for calculating mass transport across vertical sections,
! using the bomplot vsect routine and tools from timeseries
!
! Tomas Torsvik <tomas.torsvik@uni.no>, CMU, BCCS, Uni Research
!
! The module calculates transport across an area defined by a horizontal
! line segment and the total (dynamic) depth along the segment.
! Transports are calculated using the normal velocity UN from the vsect
! routine in bomplot. The resulting volume flux is plotted as a
! timeseries in a similar way as EKIN and MASS.
!
! The convention used here is that the normal     -----------------------
! velocity is irected to the left relative        |          ^          |
! to the direction of the line segment.           |          |          |
!                                                 | (C) UN <-|   UN (B) |
!      Line segment:       positive UN direction: |        <-|   ^^     |
!      -------------       ---------------------- |          |   ||     |
! (A)  NORTH -> SOUTH  =>  EAST                   |  <-------*------->  |
! (B)  WEST  -> EAST   =>  NORTH                  |     ||   |          |
! (C)  SOUTH -> NORTH  =>  WEST                   |     vv   |->        |
! (D)  EAST  -> WEST   =>  SOUTH                  | (D) UN   |-> UN (A) |
!                                                 |          |          |
!                                                 |          v          |
!                                                 -----------------------
!
!========================================================================

  implicit none

  ! Number of vertical sections  used for transport calculation
  integer :: transport_count = 0

  ! Max number of vertical sections  used for transport calculation
  ! (Increase this value if more sections are needed)
  integer, PARAMETER :: TRANSPORT_MAXCOUNT = 100

  ! Array for storing transport values
  real, dimension(TRANSPORT_MAXCOUNT) :: transport_array

  ! Arrays for storing (x,y) coordinates for start and end points 
  ! of the vertical sections (assume only 2 points are needed)
  real, dimension(TRANSPORT_MAXCOUNT,2) :: transport_px
  real, dimension(TRANSPORT_MAXCOUNT,2) :: transport_py

  ! Arrays for storing vsect grid coordinates
  real, dimension(:), allocatable :: transport_x
  real, dimension(:,:), allocatable :: transport_z

  ! Label for individual transport calculation
  character(12) :: transport_label

contains

  subroutine get_transport
  !
  ! Get vsect values, interpolate to find cell area, and 
  ! calculate the transport (volume flux)
  !
    use state, only : im,jm,kb,u,v,fsm, debug,dx    
    use mod_mpi, only : i1,i2,j1,j2,im_mi_1,jm_mi_1,i_eq_im,j_eq_jm,  &
         master,tag_ghost,update_ghost    
    use bomplot, only : vsect
    use pgplot_core

    implicit none

    integer :: count
    integer :: i,k
    real :: dtranspx

    if(transport_count == 0) return

    ! loop through all vsect transport lines
    do count = 1,transport_count

       ! Get vsect coordinates and normal velocities
       call vsect("UN", transport_px(count,:), transport_py(count,:), pgplot=.true.)

       ! Find cell area
       if(master) then
          transport_x = 0.
          transport_z = 0.

          ! Find vsect grid coordinates
          do i=1,(pgplot_current_totsec-1)
             transport_x(i+1) = 0.5*(pg_vx(i,1) + pg_vx(i+1,1))
             do k = 2,kb-1
                transport_z(i+1,k) = 0.25*(pg_vz(i,k-1) +  &
                     pg_vz(i+1,k-1) + pg_vz(i,k) + pg_vz(i+1,k))
             end do
             ! Surface coordinate
             transport_z(i+1,1) = pg_vz(i,1) + pg_vz(i+1,1) - transport_z(i+1,2)
             ! Bottom coordinate
             transport_z(i+1,kb) = pg_vz(i,kb-1) + pg_vz(i+1,kb-1) - transport_z(i+1,kb-1)
          end do

          ! First horizontal point (x = 0.), and last horizontal point
          transport_x(pgplot_current_totsec+1) =   &
               2*transport_x(pgplot_current_totsec)  &
               - transport_x(pgplot_current_totsec-1)
          do k=1,kb
             transport_z(1,k) = 2*transport_z(2,k) - transport_z(3,k)
             transport_z(pgplot_current_totsec+1,k) =  &
                  2*transport_z(pgplot_current_totsec,k)  &
                  - transport_z(pgplot_current_totsec-1,k)
          end do

          ! Calculate transport
          transport_array(count) = 0.
          do i=1,pgplot_current_totsec
             dtranspx = transport_x(i+1) - transport_x(i)
             do k=1,kb-1
                transport_array(count) = transport_array(count) +  &
                     pg_usec(i,k)*dtranspx*0.5*(  &
                     (transport_z(i,k) - transport_z(i,k+1)) +  &
                     (transport_z(i+1,k) - transport_z(i+1,k+1)))
             end do
          end do
       end if

    end do

  end subroutine get_transport


end module mod_transport
