SUBROUTINE MODESPLIT(IDX)
!
! MODESPLIT propagates the solution of the equations
!
! d(UD)/dt = -g*D*d(eta)/dx + f*V*D + d((KM/D)*(dU/dsigma))/dsigma +
!       d(AM*(d(U*D)/dy))/dy + d(AM*(d(U*D)/dx))/dx + deltaU 
!
! d(VD)/dt = -g*D*d(eta)/dy - f*U*D + d((KM/D)*(dV/dsigma))/dsigma +
!       d(AM*(d(V*D)/dx))/dx + d(AM*(d(V*D)/dy))/dy + deltaV 
!
! d(eta)/dt = -d(UA)/dx -d(VA)/dy
!
! with a time split method.
!
! The Depth integrated versions of these equations are
! propagated with a number of shorter timesteps (DTE).
! The Baroclinic remainders, UBCLIN and VBCLIN, are
! then propagated with a longer time step (DT). 
! U and V are then computed as the sums of the barotropic
! the baroclinic contributions.
!
! deltaU and deltaV are the sums of internal pressure terms and
! non-linear terms computed before the call to MODESPLIT
! by the routines INTERNAL and LAXWUV.
!
! The order in which the U and V equations are propagated
! is symmetric over a double step both in the barotropic
! steps and in the baroclinic steps.
!
! UA and VA are the depth integrated transports in
! x and y direction respectively.
!
!
  USE STATE
  USE MOD_BOUND
  USE CONTROL_PARAMETERS  
  USE MOD_MODESPLIT
  use mod_drag_region
  
  IMPLICIT NONE
  INTEGER IDX
!
! Local variables.
!
  INTEGER I,J,K
  REAL RX,RY,AREA
  INTEGER IS,RIV,DJ
  REAL DTE,FRAC
  logical inblock

  in_modesplit=.true.
!
  DJ=1
  IF(JM==1) DJ=0
  
  DTE = DT/N2D
  RX = DTE/DX
  RY = DTE/DY
  AREA = DX*DY
!
  UCORR = 0.
  VCORR = 0.
  WCORR = 0.
!
!  Depth integrate the corrections DELTAU and DELTAV
!
  AX = 0.
  DO K = 1,KB-1
      DO J = 1,J_EQ_JM
          DO I = 1,I_EQ_IM
              IF (DUM(I,J).NE.0.) THEN
                  AX(I,J) = AX(I,J)+DUM(I,J)*DELTAU(I,J,K)*DZ(K)
              ENDIF
          END DO
      END DO
  END DO

  AY = 0.
  DO K = 1,KB-1
      DO J = 1,J_EQ_JM
          DO I = 1,I_EQ_IM
              IF (DVM(I,J).NE.0.) THEN
                  AY(I,J) = AY(I,J)+DVM(I,J)*DELTAV(I,J,K)*DZ(K)
              ENDIF
          END DO
      END DO
  END DO
!
!
! Depth integrate U and V
!
  UA = 0.
  DO K = 1,KB-1
      DO J = 1,J_EQ_JM
          DO I = 1,I_EQ_IM
              UA(I,J) = UA(I,J) + DUM(I,J)*UN(I,J,K)*DZ(K)
          END DO
      END DO
  END DO
!
  VA = 0.
  DO K = 1,KB-1
      DO J = 1,J_EQ_JM
          DO I = 1,I_EQ_IM
              VA(I,J) = VA(I,J) + DVM(I,J)*VN(I,J,K)*DZ(K)
          END DO
      END DO
  END DO
!
!  Compute the baroclinic parts of U and V
!
  UBCLIN = 0.
  VBCLIN = 0.
  DO K = 1,KB-1
      DO J = 1,J_EQ_JM
          DO I = 1,I_EQ_IM
              UBCLIN(I,J,K) = UN(I,J,K) - UA(I,J)
              VBCLIN(I,J,K) = VN(I,J,K) - VA(I,J)
          END DO
      END DO
  END DO
!
!  Initiate W from step N
!
  W = WN
!
!  2-D loop N2D times
  UASUM = 0.
  VASUM = 0.
  UAB = UA
  VAB = VA
  ETAB = ETA
  UASTAR = UA
  VASTAR = VA
  ETASTAR = ETA
  UAF = UA
  VAF = VA
  ETAF = ETA

  CALL DUMP_STATE("modesplit_pre2d")

  in_2d = .true.
  
  DO IS = 1,N2D
!
!  If TIDE, then update the BCs every 2D time step.
!
      IF (TIDE) THEN
         CALL TIDEFRS(IS)
      ENDIF
!
!  River Specific
!
      UAD = 0.
      VAD = 0.
      IF (RIVLOG) THEN
          DO RIV = 1,NUMC
              i=iriver(riv) ;  j=jriver(riv)
              call globalind2local(i,j,inblock)
              if(inblock) then      
                  IF (CODERIVER(riv).EQ.1) THEN
                      UAD(i,j) = FRIVER(riv)/DY
                  ELSEIF (CODERIVER(riv).EQ.2) THEN
                      VAD(i,j) = FRIVER(riv)/DX
                  ELSEIF (CODERIVER(riv).EQ.3) THEN
                      if(i+1.gt.i2) cycle                      
                      UAD(i+1,j) = -FRIVER(riv)/DY
                  ELSEIF (CODERIVER(riv).EQ.4) THEN
                      if(j+1.gt.j2) cycle                      
                      VAD(i,j+1) = -FRIVER(riv)/DX
                  ELSE
                      WRITE(*,*)'ERROR IN MODESPLIT-1'
                      WRITE(*,*)I,CODERIVER(riv)
                      STOP
                  ENDIF
              end if
          END DO
      ENDIF
  
      call tag_ghost(eta,west+south,1)
      call tag_ghost(ua, width=1)
      call tag_ghost(va, width=1)
      call update_ghost
!
! Compute contributions horizontal viscosity.
!
      IF (SMAG) THEN
          CALL SMAGOR2D
      ENDIF
      
      call tag_ghost(am2d, width=1)
      call update_ghost
      
      IF (VISCPOM) THEN
          CALL HORVISCUV2DPOM(C2U,C2V)
      ELSE
          CALL HORVISCUV2D(C2U,C2V)
      ENDIF
!
!
!  Add effects of wind stress and bottom friction and AX to C2U.
!
      DO J = J_EQ_2,JM_mi_1
          DO I=I_EQ_2,I_EQ_IM
              C2U(I,J) = WUSURF(I,J)+WUBOT(I,J) + AX(I,J) + C2U(I,J)
          END DO
      END DO
!
! Add effects of wind stress and bottom friction and AY to C2V.
!
      DO J=J_EQ_2,J_EQ_JM
          DO I = I_EQ_2,IM_mi_1
              C2V(I,J) = WVSURF(I,J)+WVBOT(I,J) + AY(I,J) + C2V(I,J)
          END DO
      END DO
!
!  Compute VA*D in U-points
!
    VU2D = 0.

    IF(JM > 1)THEN
          

      DO J = J_EQ_2,JM_mi_1
          DO I=I_EQ_2,I_EQ_IM
              IF (DUM(I,J).GT.0.) THEN
                  IF (DVM(I,J+1).GT.0.) THEN
                      VU2D(I,J) = VA(I,J+1)*DV(I,J+1)/HVSQR(I,J+1)
                  ENDIF
                  IF (DVM(I,J).GT.0.) THEN
                      VU2D(I,J) = VU2D(I,J) + VA(I,J)*DV(I,J)/HVSQR(I,J)
                  ENDIF
                  IF (DVM(I-1,J+1).GT.0.) THEN
                      VU2D(I,J) = VU2D(I,J) + VA(I-1,J+1)*DV(I-1,J+1)/HVSQR(I-1,J+1)
                  ENDIF
                  IF (DVM(I-1,J).GT.0.) THEN
                      VU2D(I,J) = VU2D(I,J) + VA(I-1,J)*DV(I-1,J)/HVSQR(I-1,J)
                  ENDIF
                  VU2D(I,J) = 0.25*VU2D(I,J)*HUSQR(I,J)
              ENDIF
          END DO
      END DO
      
    ELSE
      
      J=1
      DO I = i_eq_2,i_eq_IM
         VU2D(I,J) = 0.5*(DV(I,J)*VA(I,J)+DV(I-1,J)*VA(I-1,J))
      END DO
     
    END IF

!
!Define U*D in V-points.
!
    UV2D = 0.
    IF(JM > 1)THEN

      DO J=J_EQ_2,J_EQ_JM
          DO I = I_EQ_2,IM_mi_1
              IF (DVM(I,J).NE.0.) THEN
                  IF (DUM(I+1,J).NE.0.) THEN
                      UV2D(I,J) = UA(I+1,J)*DU(I+1,J)/HUSQR(I+1,J)
                  ENDIF
                  IF (DUM(I,J).NE.0.) THEN
                      UV2D(I,J) = UV2D(I,J) + UA(I,J)*DU(I,J)/HUSQR(I,J)
                  ENDIF
                  IF (DUM(I+1,J-1).NE.0.) THEN
                      UV2D(I,J) = UV2D(I,J) + UA(I+1,J-1)*DU(I+1,J-1)/HUSQR(I+1,J-1)
                  ENDIF
                  IF (DUM(I,J-1).NE.0.) THEN
                      UV2D(I,J) = UV2D(I,J) + UA(I,J-1)*DU(I,J-1)/HUSQR(I,J-1)
                  ENDIF
                  UV2D(I,J) = 0.25*UV2D(I,J)*HVSQR(I,J)
              ENDIF
          END DO
      END DO
    ELSE
      J=1
      DO I = i_eq_2,IM_mi_1
        UV2D(I,J) = 0.5*(DU(I+1,J)*UA(I+1,J)+DU(I,J)*UA(I,J))
      END DO  
    END IF
!
!  IF IS = 1, Compute solution at next time step with a predictor-corrector method similar to the trapzoidal method.
!
      IF (IS.EQ.1) THEN
!
! Propagate the solution taking all terms into account.
!
          DO J = J_EQ_2,JM_mi_1
              DO I=I_EQ_2,I_EQ_IM
                  CORU(I,J) = COR(I,J)*VU2D(I,J) - GRAV*DU(I,J)* &
                       &  (ETA(I,J)-ETA(I-1,J))/DX + C2U(I,J)
              END DO
          END DO
          DO J = J_EQ_2,JM_mi_1
              DO I=I_EQ_2,I_EQ_IM
                  IF (DUM(I,J).NE.0.) THEN
                      UASTAR(I,J) = UA(I,J) + DTE*CORU(I,J)/DU(I,J)
                  ENDIF
              END DO
          END DO
!
!  Then the V-equation to compute VASTAR.
!
!
          DO J=J_EQ_2,J_EQ_JM
              DO I = I_EQ_2,IM_mi_1
                  CORV(I,J) = -COR(I,J)*UV2D(I,J) - GRAV*DV(I,J)* &
                       &  (ETA(I,J)-ETA(I,J-DJ))/DY +  C2V(I,J)
              END DO
          END DO
          DO J=J_EQ_2,J_EQ_JM
              DO I = I_EQ_2,IM_mi_1
                  IF (DVM(I,J).NE.0.) THEN
                      VASTAR(I,J) = VA(I,J) + DTE*CORV(I,J)/DV(I,J)
                  ENDIF
              END DO
          END DO
!
!
!
! Compute horizontal transports
!
          DO J = J_EQ_2,JM_mi_1
              DO I=I_EQ_2,I2 ! I_EQ_IM
                  IF (DUM(I,J).NE.0.) THEN
                      UAD(I,J) = UA(I,J)*DU(I,J)
                  ENDIF
              END DO
          END DO
          DO J=J_EQ_2,J2  ! J_EQ_JM
              DO I = I_EQ_2,IM_mi_1
                  IF (DVM(I,J).NE.0.) THEN
                      VAD(I,J) = VA(I,J)*DV(I,J)
                  ENDIF
              END DO
          END DO
!
!Then  the continuity eqn.
!
! shouldnt be necessary, since ua,va,du,dv are fully updated, and ranges above are extended
!          call tag_ghost(uad, width=1)
!          call tag_ghost(vad, width=1)
!          call update_ghost
          
          DO J = J_EQ_2,JM_mi_1
              DO I = I_EQ_2,IM_mi_1
                  ETASTAR(I,J) = ETA(I,J)-   &
                       &  RX*(UAD(I+1,J)-UAD(I,J))- RY*(VAD(I,J+DJ)-VAD(I,J))
                  ETASTAR(I,J) = ETASTAR(I,J)*FSM(I,J)
              END DO
          END DO
!
!  add possible contr. from dived source
!
          IF (UPWELL) THEN
              DO RIV = 1,NUMWELL
                  i=iwell(riv) ;  j=jwell(riv)
                  call globalind2local(i,j,inblock)
                  if(inblock) then      
                      ETASTAR(I,J) = ETASTAR(I,J) + DTE*FWELL(RIV)/AREA
                  end if
              END DO
          ENDIF

!
      ELSE ! IF IS > 1
!
! Propagate the solution taking all terms into account.
! First a leapfrog predictor step
!
          DO J = J_EQ_2,JM_mi_1
              DO I=I_EQ_2,I_EQ_IM
                  CORU(I,J) = COR(I,J)*VU2D(I,J) - GRAV*DU(I,J)* &
                       &  (ETA(I,J)-ETA(I-1,J))/DX + C2U(I,J)
              END DO
          END DO
          DO J = J_EQ_2,JM_mi_1
              DO I=I_EQ_2,I_EQ_IM
                  IF (DUM(I,J).NE.0.) THEN
                      UASTAR(I,J) = UAB(I,J) + 2*DTE*CORU(I,J)/DU(I,J)
                  ENDIF
              END DO
          END DO
!
!  Then the V-equation to compute VASTAR.
!
!
          DO J=J_EQ_2,J_EQ_JM
              DO I = I_EQ_2,IM_mi_1
                  CORV(I,J) = -COR(I,J)*UV2D(I,J) - GRAV*DV(I,J)* &
                       &  (ETA(I,J)-ETA(I,J-DJ))/DY +  C2V(I,J)
              END DO
          END DO
          DO J=J_EQ_2,J_EQ_JM
              DO I = I_EQ_2,IM_mi_1
                  IF (DVM(I,J).NE.0.) THEN
                      VASTAR(I,J) = VAB(I,J) + 2*DTE*CORV(I,J)/DV(I,J)
                  ENDIF
              END DO
          END DO
!
!
!
! Compute horizontal transports
!
          DO J = J_EQ_2,JM_mi_1
              DO I=I_EQ_2,I2  !I_EQ_IM
                  IF (DUM(I,J).NE.0.) THEN
                      UAD(I,J) = UA(I,J)*DU(I,J)
                  ENDIF
              END DO
          END DO
          DO J=J_EQ_2,J2  !J_EQ_JM
              DO I = I_EQ_2,IM_mi_1
                  IF (DVM(I,J).NE.0.) THEN
                      VAD(I,J) = VA(I,J)*DV(I,J)
                  ENDIF
              END DO
          END DO
!
!Then  the continuity eqn.
!
! same here, should not be necessary
!          call tag_ghost(uad, width=1)
!          call tag_ghost(vad, width=1)
!          call update_ghost
          
          DO J = J_EQ_2,JM_mi_1
              DO I = I_EQ_2,IM_mi_1
                  ETASTAR(I,J) = ETAB(I,J)-   &
                       &  2*RX*(UAD(I+1,J)-UAD(I,J))- 2*RY*(VAD(I,J+DJ)-VAD(I,J))
                  ETASTAR(I,J) = ETASTAR(I,J)*FSM(I,J)
              END DO
          END DO
!
!  add possible contr. from dived source
!
          IF (UPWELL) THEN              
              DO RIV = 1,NUMWELL
                  I=IWELL(RIV) ;  J=JWELL(RIV)
                  CALL GLOBALIND2LOCAL(I,J,INBLOCK)
                  IF(INBLOCK) THEN      
                      ETASTAR(I,J) = ETASTAR(I,J) + DTE*FWELL(RIV)/AREA
                  END IF
              END DO
          ENDIF
      ENDIF
!
!  Boundary conditions for the intermediate values of ETA, UA and VA
!
      CALL BOUND2(7)

      call tag_ghost(uastar, east+south, 1)
      call tag_ghost(vastar, west+north, 1)
      call tag_ghost(etastar, west+south, 1)
      call update_ghost
!
!
!  Then a corrector step with the theta- method
!
!
!  First in UA
!
!
!  Compute VA*D in U-points
!
    VU2D = 0.

    IF(JM > 1)THEN

        DO J = J_EQ_2,JM_mi_1
            DO I=I_EQ_2,I_EQ_IM
                IF (DUM(I,J).GT.0.) THEN
                    IF (DVM(I,J+1).GT.0.) THEN
                        VU2D(I,J) = (THETA2D*VASTAR(I,J+1) + (1-THETA2D)*VA(I,J+1))*DV(I,J+1)/HVSQR(I,J+1)
                    ENDIF
                    IF (DVM(I,J).GT.0.) THEN
                        VU2D(I,J) = VU2D(I,J) + (THETA2D*VASTAR(I,J) + (1-THETA2D)*VA(I,J))*DV(I,J)/HVSQR(I,J)
                    ENDIF
                    IF (DVM(I-1,J+1).GT.0.) THEN
                        VU2D(I,J) = VU2D(I,J) + (THETA2D*VASTAR(I-1,J+1) + (1-THETA2D)*VA(I-1,J+1))*DV(I-1,J+1)/HVSQR(I-1,J+1)
                    ENDIF
                    IF (DVM(I-1,J).GT.0.) THEN
                        VU2D(I,J) = VU2D(I,J) + (THETA2D*VASTAR(I-1,J) + (1-THETA2D)*VA(I-1,J))*DV(I-1,J)/HVSQR(I-1,J)
                    ENDIF
                    VU2D(I,J) = 0.25*VU2D(I,J)*HUSQR(I,J)
                ENDIF
            END DO
        END DO
    ELSE
        J=1
        DO I = i_eq_2,i_eq_IM
            VU2D(I,J) = 0.5*(DV(I,J)*(THETA2D*VASTAR(I,J) + (1-THETA2D)*VA(I,J)) + &
                 &  DV(I-1,J)*(THETA2D*VASTAR(I-1,J) + (1-THETA2D)*VA(I-1,J)))
        END DO
    END IF

!
! Propagate the solution taking all terms into account.
!
      DO J = J_EQ_2,JM_mi_1
        DO I=I_EQ_2,I_EQ_IM
              CORU(I,J) = COR(I,J)*VU2D(I,J) - GRAV*DU(I,J)* &
                   &  (THETA2D*(ETASTAR(I,J)-ETASTAR(I-1,J)) + &
                   &   (1-THETA2D)*(ETA(I,J)-ETA(I-1,J)))/DX + C2U(I,J)
          END DO
      END DO
      DO J = J_EQ_2,JM_mi_1
          DO I=I_EQ_2,I_EQ_IM
              IF (DUM(I,J).NE.0.) THEN
                  UAF(I,J) = UA(I,J) + DTE*CORU(I,J)/DU(I,J)
              ENDIF
          END DO
      END DO
!
!
!  Then the VA-eqn.
!
      UV2D = 0.
      IF(JM>1)THEN
          

      DO J=J_EQ_2,J_EQ_JM
          DO I = I_EQ_2,IM_mi_1
              IF (DVM(I,J).NE.0.) THEN
                  IF (DUM(I+1,J).NE.0.) THEN
                      UV2D(I,J) = (THETA2D*UASTAR(I+1,J) + (1-THETA2D)*UA(I+1,J))*DU(I+1,J)/HUSQR(I+1,J)
                  ENDIF
                  IF (DUM(I,J).NE.0.) THEN
                      UV2D(I,J) = UV2D(I,J) + (THETA2D*UASTAR(I,J) + (1-THETA2D)*UA(I,J))*DU(I,J)/HUSQR(I,J)
                  ENDIF
                  IF (DUM(I+1,J-1).NE.0.) THEN
                      UV2D(I,J) = UV2D(I,J) + (THETA2D*UASTAR(I+1,J-1) + (1-THETA2D)*UA(I+1,J-1))*DU(I+1,J-1)/HUSQR(I+1,J-1)
                  ENDIF
                  IF (DUM(I,J-1).NE.0.) THEN
                      UV2D(I,J) = UV2D(I,J) + (THETA2D*UASTAR(I,J-1) + (1-THETA2D)*UA(I,J-1))*DU(I,J-1)/HUSQR(I,J-1)
                  ENDIF
                  UV2D(I,J) = 0.25*UV2D(I,J)*HVSQR(I,J)
              ENDIF
            END DO
         END DO
      ELSE
         J=1
         DO I = i_eq_2,IM_mi_1
             UV2D(I,J) = 0.5*(DU(I+1,J)*(THETA2D*UASTAR(I+1,J) + (1-THETA2D)*UA(I+1,J)) + &
                & DU(I,J)*(THETA2D*UASTAR(I,J) + (1-THETA2D)*UA(I,J))) 
         END DO
      END IF
!
!
      DO J=J_EQ_2,J_EQ_JM
          DO I = I_EQ_2,IM_mi_1
              CORV(I,J) = -COR(I,J)*UV2D(I,J) - GRAV*DV(I,J)* &
                   &  (THETA2D*(ETASTAR(I,J)-ETASTAR(I,J-DJ)) + &
                   &   (1-THETA2D)*(ETA(I,J)-ETA(I,J-DJ)))/DY +  C2V(I,J)
          END DO
      END DO
      DO J=J_EQ_2,J_EQ_JM
          DO I = I_EQ_2,IM_mi_1
              IF (DVM(I,J).NE.0.) THEN
                  VAF(I,J) = VA(I,J) + DTE*CORV(I,J)/DV(I,J)
              ENDIF
          END DO
      END DO
!
!
!  Finally the ETA-eqn.
!
!
!
! Compute horizontal transports
!
      DO J = J_EQ_2,JM_mi_1
          DO I=I_EQ_2,i2  !I_EQ_IM
              IF (DUM(I,J).NE.0.) THEN
                  UAD(I,J) = (THETA2D*UASTAR(I,J) + (1-THETA2D)*UA(I,J))*DU(I,J)
              ENDIF
          END DO
      END DO
      DO J=J_EQ_2,j2   !J_EQ_JM
          DO I = I_EQ_2,IM_mi_1
              IF (DVM(I,J).NE.0.) THEN
                  VAD(I,J) = (THETA2D*VASTAR(I,J) + (1-THETA2D)*VA(I,J))*DV(I,J)
              ENDIF
          END DO
      END DO
!
!Then  the continuity eqn.
!
      DO J = J_EQ_2,JM_mi_1
          DO I = I_EQ_2,IM_mi_1
              ETAF(I,J) = ETA(I,J)-   &
                   &  RX*(UAD(I+1,J)-UAD(I,J))- RY*(VAD(I,J+DJ)-VAD(I,J))
              ETAF(I,J) = ETAF(I,J)*FSM(I,J)
          END DO
      END DO
!
!  add possible contr. from dived source
!
      IF (UPWELL) THEN
          DO RIV = 1,NUMWELL
              I=IWELL(RIV) ;  J=JWELL(RIV)
              CALL GLOBALIND2LOCAL(I,J,INBLOCK)
              IF(INBLOCK) THEN      
                  ETAF(I,J) = ETAF(I,J) + DTE*FWELL(RIV)/AREA
              END IF
          END DO
      ENDIF
!
!  Boundary values for the final values at time step n+1
!
!      CALL BOUND2D(ETAF,UAF,VAF)
      call bound2(8)
!
      UASUM = UASUM + THETA2D*UASTAR + (1-THETA2D)*UA
      VASUM = VASUM + THETA2D*VASTAR + (1-THETA2D)*VA
      UAB = UA
      UA = UAF
      VAB = VA
      VA = VAF
      ETAB = ETA
      ETA = ETAF

      call user_step2d()
     
!
!  End 2-D loop
!
  END DO

  in_2d=.false.
  
  call dump_state('modesplit_post2d')
!
  UASUM = UASUM/N2D
  VASUM = VASUM/N2D
!
!
! Compute the divergence free velocity fields.
!
  UADV = 0.
  VADV = 0.
  DO K = 1,KB-1
      DO J = 1,J_EQ_JM
          DO I = 1,I_EQ_IM
              UADV(I,J,K) = UBCLIN(I,J,K) + UASUM(I,J)
              VADV(I,J,K) = VBCLIN(I,J,K) + VASUM(I,J)
          END DO
      END DO
  END DO
!
  IF (SMAG) THEN
      CALL SMAGOR
  ENDIF
  
  call tag_ghost( am, width=1 )
  call tag_ghost( km, west+south, 1 )
  call tag_ghost( ubclin, width=1 )
  call tag_ghost( vbclin, width=1 )
  call update_ghost
  
  IF (VISCPOM) THEN
      CALL HORVISCUVPOM(UBCLIN,VBCLIN,C3U,C3V)
  ELSE 
      CALL HORVISCUV(UBCLIN,VBCLIN,C3U,C3V)
  ENDIF
  
  IF (MOD(ISTEP,2).EQ.1) THEN
      call tag_ghost(vbclin, west+north, 1)
      call update_ghost
      CALL STEPU3D(UBCLIN,VBCLIN,C3U,AX)
      call tag_ghost(ubclin, east+south, 1)
      call update_ghost
      CALL STEPV3D(UBCLIN,VBCLIN,C3V,AY)
  ELSE
      call tag_ghost(ubclin, east+south, 1)
      call update_ghost
      CALL STEPV3D(UBCLIN,VBCLIN,C3V,AY)
      call tag_ghost(vbclin, west+north, 1)
      call update_ghost
      CALL STEPU3D(UBCLIN,VBCLIN,C3U,AX)
  END IF
!
!  Include the effects of vertical viscosity on UBCLIN and VBCLIN
!

  call tag_ghost(u,east+south,1)
  call tag_ghost(v,west+north,1)
  call update_ghost
  
  CALL VERTVISCUVB(UBCLIN,VBCLIN)
!  
!  After VERTVISCUVB the vertical integrals of UBCLIN and VBCLIN
!  are expected to be non-zero due to surface and bottom stress.
!  This is corrected for in the following loops.
!
  DO K = 1,KB-1  
      DO J = J_EQ_2,JM_mi_1
          DO I=I_EQ_2,I_EQ_IM
              IF (DUM(I,J).NE.0.) THEN
                  UBCLIN(I,J,K) = UBCLIN(I,J,K) - DT*(WUBOT(I,J)+WUSURF(I,J))/DU(I,J)
                  U(I,J,K) = UBCLIN(I,J,K) + UA(I,J)
                  UADV(I,J,K) = UBCLIN(I,J,K) + UASUM(I,J)
              ENDIF
          END DO
      END DO
  END DO
  DO K = 1,KB-1
      DO J=J_EQ_2,J_EQ_JM
          DO I = I_EQ_2,IM_mi_1
              IF (DVM(I,J).NE.0.) THEN
                  VBCLIN(I,J,K) = VBCLIN(I,J,K) - DT*(WVBOT(I,J)+WVSURF(I,J))/DV(I,J)
                  V(I,J,K) = VBCLIN(I,J,K) + VA(I,J)
                  VADV(I,J,K) = VBCLIN(I,J,K) + VASUM(I,J)
              ENDIF
          END DO
      END DO
  END DO
!
! Add friction to regions specified in the setupfile  
  call apply_regional_drag
!  Adjust ETA so that it is in balance with UADV (and VADV in 3D)
!
!  First the fluxes
!
!
!  River Specific
!
  UV2D = 0.
  VU2D = 0.
  IF (RIVLOG) THEN
      DO RIV = 1,NUMC
          I=IRIVER(RIV) ;  J=JRIVER(RIV)
          CALL GLOBALIND2LOCAL(I,J,INBLOCK)
          IF(INBLOCK) THEN      
              IF (CODERIVER(RIV).EQ.1) THEN
                  UV2D(I,J) = FRIVER(riv)/DY
              ELSEIF (CODERIVER(riv).EQ.2) THEN
                  VU2D(I,J) = FRIVER(riv)/DX
              ELSEIF (CODERIVER(riv).EQ.3) THEN
                  if(i+1.gt.i2) cycle                                        
                  UV2D(I+1,J) = -FRIVER(riv)/DY
              ELSEIF (CODERIVER(riv).EQ.4) THEN
                  if(j+1.gt.j2) cycle                                        
                  VU2D(I,J+1) = -FRIVER(riv)/DX
              ELSE
                  WRITE(*,*)'ERROR IN MODESPLIT-1'
                  WRITE(*,*)I,CODERIVER(riv)
                  STOP
              ENDIF
          END IF
      END DO
  END IF

!
  DO K = 1,KB-1
      DO J = 1,J_EQ_JM
          DO I = 1,I_EQ_IM
              UV2D(I,J) = UV2D(I,J) + DZ(K)*UADV(I,J,K)*DU(I,J)*DUM(I,J)
          END DO
      END DO
  END DO
  UV2D = UV2D*DY
!
  DO K = 1,KB-1
      DO J = 1,J_EQ_JM
          DO I = 1,I_EQ_IM
              VU2D(I,J) = VU2D(I,J) + DZ(K)*VADV(I,J,K)*DV(I,J)*DVM(I,J)
          END DO
      END DO
  END DO
  VU2D = VU2D*DX
  
  call tag_ghost(uv2d, width=1)
  call tag_ghost(vu2d, width=1)
  call tag_ghost(u)
  call tag_ghost(v)
  call tag_ghost(ubclin, width=1)
  call tag_ghost(vbclin, width=1)
  call tag_ghost(uadv)
  call tag_ghost(vadv)
  call update_ghost
!
!
!  Then the water levels
!
!
  AREA = DX*DY
!  recompute ETA
!
!
  DO J = J_EQ_2,JM_mi_1
      DO I = I_EQ_2,IM_mi_1
          ETA(I,J) = ETAP(I,J) - DT*((UV2D(I+1,J)-UV2D(I,J))+(VU2D(I,J+DJ)-VU2D(I,J)))/AREA
      END DO
  END DO
!
!
!  add possible contr. from dived source
!
  IF (UPWELL) THEN
      DO RIV = 1,NUMWELL
          I=IWELL(RIV) ;  J=JWELL(RIV)
          CALL GLOBALIND2LOCAL(I,J,INBLOCK)
          IF(INBLOCK) THEN      
              ETA(I,J) = ETA(I,J) + DT*FWELL(RIV)/AREA
          END IF
      END DO
  ENDIF
!
!  Relax updated ETA in FRS-zone
!
!  CALL BOUNDETA
  call bound2(9)
!
!   compute the sigma-coord. vertical velocity W
!   from the equation
! 
!   d(W)/d(sigma) = -(d(eta)/d(T) + d(U*D)/d(X) + d(V*D)/d(Y)) .
!
!
!
!
!  If NONHYD
!
  IF (NONHYD ) THEN
!
!  Advection and horisontal diffusion of W
!
      CALL SUPERBEEW3D
!
! Vertical diffusion of W
!
      CALL VERTVISCW
!
!  Estimate nonhydrostatic pressure
!
      CALL DYNP()

      call tag_ghost(p, width=1)
      call update_ghost

      UCORR = 0.0

      DO K = 1,KB-1
          DO J = 1,J_EQ_JM
       DO I = I_EQ_2,I_EQ_IM
                  IF (DUM(I,J).NE.0.) THEN
                      UCORR(I,J,K) = -(DT/RHO0)*(P(I,J,K)-P(I-1,J,K))/DX
                  ENDIF
              END DO
          END DO
      END DO
!  WRITE(*,*)'AFTER UCORR',MAXVAL(P),MINVAL(P),MINVAL(UCORR),MAXVAL(UCORR)
      VCORR = 0.0
      IF(JM>1)THEN
      DO K = 1,KB-1
          DO J = J_EQ_2,J_EQ_JM
              DO I = 1,I_EQ_IM
                  IF (DVM(I,J).NE.0.) THEN
                      VCORR(I,J,K) = -(DT/RHO0)*(P(I,J,K)-P(I,J-1,K))/DY
                  ENDIF
              END DO
          END DO
      END DO
      END IF
!   WRITE(*,*)'AFTER VCORR',MINVAL(VCORR),MAXVAL(VCORR)
!  
      WCORR = 0.0
      DO K = 2,KB-1
          DO J = 1,J_EQ_JM
              DO I = 1,I_EQ_IM
                  IF (FSM(I,J).NE.0.) THEN
                      WCORR(I,J,K) = -(DT/RHO0)*(P(I,J,K-1)-P(I,J,K))/(D(I,J)*DZZ(K-1))
                  ENDIF
              END DO
          END DO
      END DO
!   WRITE(*,*)'AFTER WCORR',MINVAL(WCORR),MAXVAL(WCORR)
!
!  
!  add corrections to U, V and W
!   
      DO K = 1,KB-1
          DO J = 1,J_EQ_JM
              DO I = 1,I_EQ_IM
                  IF (DUM(I,J).NE.0.0) THEN
                      U(I,J,K) = U(I,J,K) + UCORR(I,J,K)
                      UADV(I,J,K) = UADV(I,J,K) + UCORR(I,J,K)
                  ENDIF
              END DO
          END DO
      END DO
      DO K = 1,KB-1
          DO J = 1,J_EQ_JM
              DO I = 1,I_EQ_IM
                  IF (DVM(I,J).NE.0.0) THEN
                      V(I,J,K) = V(I,J,K) + VCORR(I,J,K)
                      VADV(I,J,K) = VADV(I,J,K) + VCORR(I,J,K)
                  ENDIF
              END DO
          END DO
      END DO
      DO K = 1,KB-1
          DO J = 1,J_EQ_JM
              DO I = 1,I_EQ_IM
                  IF (FSM(I,J).NE.0.0) THEN
                      W(I,J,K) = W(I,J,K) + WCORR(I,J,K)
                  ENDIF
              END DO
          END DO
      END DO
!   WRITE(*,*)'NONHYDEND',MAXVAL(UADV),MAXVAL(VADV),MAXVAL(W)
!

      if(cdw > 0.0) then
          CALL WALL_DRAG_W()          
      end if

      
  ENDIF
!
!
!   compute the sigma-coord. vertical velocity W
!   from the equation
! 
!   d(W)/d(sigma) = -(d(eta)/d(T) + d(U*D)/d(X) + d(V*D)/d(Y)) .
!
!
  AREA = DX*DY
!
!   Compute the horisontal fluxes.
!
!   d(U*D)/d(X) times DX*DY
!
  UV = 0.
  VU = 0.
!
  DO J = 1,J_EQ_JM
      DO I = 1,I_EQ_IM
          UV(I,J,1) = DZ(1)*UADV(I,J,1)
          DO K = 2,KB-1
              UV(I,J,K) = UV(I,J,K-1) + DZ(K)*UADV(I,J,K)
          END DO
      END DO
  END DO
  UV = UV*DY
!
  DO K = 1,KB-1
      DO J = 1,J_EQ_JM
          DO I = 1,I_EQ_IM
              UV(I,J,K) = UV(I,J,K)*DU(I,J)
          END DO
      END DO
  END DO
!
!   d(V*D)/d(Y) times DX*DY
!
  DO J = 1,J_EQ_JM
      DO I = 1,I_EQ_IM
          VU(I,J,1) = DZ(1)*VADV(I,J,1)
          DO K = 2,KB-1
              VU(I,J,K) = VU(I,J,K-1) + DZ(K)*VADV(I,J,K)
          END DO
      END DO
  END DO
  VU = VU*DX
  DO K = 1,KB-1
      DO J = 1,J_EQ_JM
          DO I = 1,I_EQ_IM
              VU(I,J,K) = VU(I,J,K)*DV(I,J)
          END DO
      END DO
  END DO
!
!  River spesific
!
  IF (RIVLOG) THEN
      DO RIV = 1,NUMC
          i=iriver(riv) ;  j=jriver(riv)
          call globalind2local(i,j,inblock)
          if(inblock) then      
              IF (CODERIVER(riv).EQ.1) THEN
                  IF (KRIVER1(riv).GT.1) THEN
                      DO K = 1,KRIVER1(riv)-1
                        UV(i,j,K) = 0.0
                      END DO
                  ENDIF
                  FRAC = DZ(KRIVER1(riv))/DZRIVER(riv)
                  UV(i,j,KRIVER1(riv)) = FRAC*FRIVER(riv)
                  IF (KRIVER2(riv).GT.KRIVER1(riv)) THEN
                      DO K = KRIVER1(riv)+1,KRIVER2(riv)
                          FRAC = DZ(K)/DZRIVER(riv)
                          UV(i,j,K) = UV(i,j,K-1) + FRAC*FRIVER(riv)
                      END DO
                  ENDIF
                  IF (KRIVER2(riv).LT.KB-1) THEN
                      DO K = KRIVER2(riv)+1,KB-1
                          UV(i,j,K) = UV(i,j,KRIVER2(riv))
                      END DO
                  ENDIF
              ELSEIF (CODERIVER(riv).EQ.2) THEN
                  IF (KRIVER1(riv).GT.1) THEN
                      DO K = 1,KRIVER1(riv)-1
                        VU(i,j,K) = 0.0
                      END DO
                  ENDIF
                  FRAC = DZ(KRIVER1(riv))/DZRIVER(riv)
                  VU(i,j,KRIVER1(riv)) = FRAC*FRIVER(riv)
                  IF (KRIVER2(riv).GT.KRIVER1(riv)) THEN
                      DO K = KRIVER1(riv)+1,KRIVER2(riv)
                          FRAC = DZ(K)/DZRIVER(riv)
                          VU(i,j,K) = VU(i,j,K-1) + FRAC*FRIVER(riv)
                      END DO
                  ENDIF
                  IF (KRIVER2(riv).LT.KB-1) THEN
                      DO K = KRIVER2(riv)+1,KB-1
                          VU(i,j,K) = VU(i,j,KRIVER2(riv))
                      END DO
                  ENDIF
              ENDIF
          end if
          i=iriver(riv)+1 ;  j=jriver(riv)
          call globalind2local(i,j,inblock)
          if(inblock) then      
              IF (CODERIVER(riv).EQ.3) THEN
                  IF (KRIVER1(riv).GT.1) THEN
                      DO K = 1,KRIVER1(riv)-1
                        UV(i,j,K) = 0.0
                      END DO
                  ENDIF
                  FRAC = DZ(KRIVER1(riv))/DZRIVER(riv)
                  UV(i,j,KRIVER1(riv)) = -FRAC*FRIVER(riv)
                  IF (KRIVER2(riv).GT.KRIVER1(riv)) THEN
                       DO K = KRIVER1(riv)+1,KRIVER2(riv)
                          FRAC = DZ(K)/DZRIVER(riv)
                          UV(i,j,K) = UV(i,j,K-1) - FRAC*FRIVER(riv)
                      END DO
                  ENDIF
                  IF (KRIVER2(riv).LT.KB-1) THEN
                      DO K = KRIVER2(riv)+1,KB-1
                          UV(i,j,K) = UV(i,j,KRIVER2(riv))
                      END DO
                  ENDIF
              ENDIF
          end if
          i=iriver(riv) ;  j=jriver(riv)+1
          call globalind2local(i,j,inblock)
          if(inblock) then      
              IF (CODERIVER(riv).EQ.4) THEN
                  IF (KRIVER1(riv).GT.1) THEN
                      DO K = 1,KRIVER1(riv)-1
                        VU(i,j,K) = 0.0
                      END DO
                  ENDIF
                  FRAC = DZ(KRIVER1(riv))/DZRIVER(riv)
                  VU(i,j,KRIVER1(riv)) = -FRAC*FRIVER(riv)
                  IF (KRIVER2(riv).GT.KRIVER1(riv)) THEN
                      DO K = KRIVER1(riv)+1,KRIVER2(riv)
                          FRAC = DZ(K)/DZRIVER(riv)
                          VU(i,j,K) = VU(i,j,K-1) - FRAC*FRIVER(riv)
                      END DO
                  ENDIF
                  IF (KRIVER2(riv).LT.KB-1) THEN
                      DO K = KRIVER2(riv)+1,KB-1
                          VU(i,j,K) = VU(i,j,KRIVER2(riv))
                      END DO
                  ENDIF
              ENDIF
          end if
      END DO
  ENDIF
!
!  Recompute ETA after P adjustments
!
  call tag_ghost(uv,east,1)
  call tag_ghost(vu,north,1)
  call update_ghost
!
  DO I = i_eq_2,IM_mi_1
      DO J = j_eq_2,JM_mi_1
          ETA(I,J) = ETAP(I,J) - DT*FSM(I,J)*((UV(I+1,J,KB-1)-UV(I,J,KB-1))+  &
               &  (VU(I,J+DJ,KB-1)-VU(I,J,KB-1)))/AREA
      END DO
  END DO
!
  IF (UPWELL) THEN
      DO RIV = 1,NUMWELL
          I=IWELL(RIV) ;  J=JWELL(RIV)
          CALL GLOBALIND2LOCAL(I,J,INBLOCK)
          IF(INBLOCK) THEN      
              ETA(I,J) = ETA(I,J) + DT*FWELL(RIV)/AREA
          END IF
      END DO
  ENDIF

!
!  Again adjust ETA in FRS-zone.
!
!  CALL BOUNDETA
  call bound2(9)

  eta = eta*fsm
  
!   W = 0 for sigma = 0.
!
  W(:,:,1) = 0.
  WADV(:,:,1) = 0.
!
!   Compute W from the equation
!   d(w)/d(sigma) = -(d(eta)/d(T) + d(U*D)/d(X) + d(V*D)/d(Y)) .
!
  DO K = 1,KB-1
      DO J = 1,JM_mi_1
          DO I = 1,IM_mi_1
              WADV(I,J,K+1) = (UV(I+1,J,K) - UV(I,J,K) + &
                   &             (VU(I,J+DJ,K)-VU(I,J,K)))/AREA + &
                   &             ((-Z(K+1))*(ETA(I,J)-ETAP(I,J)))/DT
              WADV(I,J,K+1) = WADV(I,J,K+1)*FSM(I,J)
          END DO
      END DO
  END DO
  IF (.NOT.NONHYD) THEN
      W = WADV
  ENDIF
!
  W(:,:,KB) = 0.
  WADV(:,:,KB) = 0.
!
!  IF UPWELL W THROUGH SEA BED
!
  IF (UPWELL) THEN
      DO RIV = 1,NUMWELL
          I=IWELL(RIV) ;  J=JWELL(RIV)
          CALL GLOBALIND2LOCAL(I,J,INBLOCK)
          IF(INBLOCK) THEN      
              W(I,J,KB) = FWELL(RIV)/(DX*DY)
              WADV(I,J,KB) = FWELL(RIV)/(DX*DY)
          END IF
      END DO
  ENDIF
!
!  save corrections.
!
  IF (IDX.EQ.1) THEN
!
!  version 4.0
!
      DELUN = U - UN
      U = UNM1 + 2.0*DELUN
      DELVN = V - VN
      V = VNM1 + 2.0*DELVN
      DELWN = W - WN
      W = WNM1 + 2.0*DELWN
  ELSEIF (IDX.EQ.2) THEN
!
! version 4.0
!
      IF (AM2) THEN
          DELUSTAR = U - UN
          DELVSTAR = V - VN
          DELWSTAR = W - WN
          U = UN + (5.0/12.0)*DELUSTAR + (2.0/3.0)*DELUN - (1.0/12.0)*DELUNM1
          V = VN + (5.0/12.0)*DELVSTAR + (2.0/3.0)*DELVN - (1.0/12.0)*DELVNM1
          W = WN + (5.0/12.0)*DELWSTAR + (2.0/3.0)*DELWN - (1.0/12.0)*DELWNM1
      ENDIF
  ELSE
      WRITE(*,*)'ERROR IN MODESPLIT',IDX
      STOP
  ENDIF

  CALL DUMP_STATE("modesplit")

  in_modesplit=.false.
  
END SUBROUTINE MODESPLIT
