SUBROUTINE NCALFA
!***BEGIN PROLOGUE NCALFA
!
!          This routine was originally written by Harald
!          Engedal, The Norwegian Meteorological Institute,
!          Box 320, 0314 Oslo 3, NORWAY.
!
!          NCALFA is later revised by
!            Jarle Berntsen 
!          Latest revision 140799
!
!
! ******************************************************************
!
!-----------------------------------------------------------------------
!  COMPUTES THE RELAXATION PARAMETERS USED BY THE FLOW RELAXATION SCHEME IN THE
!  BOUNDARY ZONES.
!  HERE : *** ALFA-2 (HYPERBOLIC REPR.) ***
!-----------------------------------------------------------------------
!
!***END PROLOGUE NCALFA
!
!   Common variables
!
  USE STATE
  USE MOD_BOUND
  IMPLICIT NONE
!
!   Local variables
!
  INTEGER I,JJ
  REAL X
  REAL DTE0,DTE
  REAL KAPPA(LB)
  logical doeastwest, donorthsouth
!
!***********************************************************************
!**                  THE RELAXATION PARAMETER                         **
!***********************************************************************
!
!   COMPUTES THE RELAXATION FACTOR "ALFA" FOR ONE COLUMN IN THE
!   I (X) -DIRECTION AND FOR ONE COLUMN IN THE J (Y) -DIRECTION. THEN,
!   THESE VALUES ARE FILLED INTO THE REST OF THE ARRAY "ALFA".
!

  alpha = 0.0
  alphae = 0.0

  DO I = 1,LB
      ALPHA(I) = 1.0 - TANH(3.2*REAL(I-1)/(LB-1))
  END DO
  
  ALPHAE(1)=1.0
  ALPHAE(LB)=0.
!JB
!   We assume that these alfas are appropriate for DT = DTE0
!   and compute corresponding kappas.
!JB
!  Approx. DTE0 = DX/SQRT(2*G*DEPMAX)
!
  if(master)write(*,*)
  if(ncalfa_dte0 > 0.0) then
      dte0 = ncalfa_dte0
      if(master.and.debug>0) write(*,*) 'NCALFA: dte0 from setup file=', dte0
  else
      dte0 = dx/sqrt(2*9.8*depmax)
      if(master.and.debug>0) write(*,*) 'NCALFA: use DTE0 = DX/SQRT(2*G*DEPMAX) =', dte0
  end if
!
  DO I = 2,LB-1
      KAPPA(I) = ALPHA(I)/(DTE0*(1.0-ALPHA(I)))
  END DO
!JB
!   Then recompute the alfas for the actual value of DT
!   First ALPHA that is to be used for 3-D variables.
!JB
  DO I = 2,LB-1
      ALPHA(I) = KAPPA(I)*DT/(1.+KAPPA(I)*DT)
  END DO
!
!  Then ALPHAE to be used in 2-D external time steps.
!
  DTE = DT/N2D
  DO I = 2,LB-1
      ALPHAE(I) = KAPPA(I)*DTE/(1.+KAPPA(I)*DTE)
  END DO

  if(master)then
      write(*,*)
      WRITE(*,*)'NCALFA: The alphas used for relaxation in FRS'
      DO I = 1,LB
          WRITE(*,*)I,ALPHA(I),ALPHAE(I)
      END DO
  end if


! check if there is enough room for frs zones
  doeastwest=.true.
  donorthsouth=.true.

  if(num_zones_east==0 .and. num_zones_west==0) doeastwest=.false.
  if(num_zones_north==0 .and. num_zones_south==0) donorthsouth=.false.  
  
  if(2*lb > im  .and. (num_zones_east>0 .and. num_zones_west>0)) then
      if(master) write(*,'(1x,a,i0,a,i0,a)') 'NCALFA: Warning, lb=',lb,' im=',im,&
           ' cannot compute relaxation arrays for both eastern and western edge'
      doeastwest=.false.
  end if
  if(2*lb > jm .and. jm>1 .and. (num_zones_south>0 .and. num_zones_north>0)) then
      if(master) write(*,'(1x,a,i0,a,i0,a)') 'NCALFA: Warning, lb=',lb,' jm=',jm,&
           ' cannot compute relaxation arrays for both northern and southern edge'
      donorthsouth=.false.
  end if

  if(lb > im .and. (num_zones_east>0 .or. num_zones_west>0)) then
      if(master) write(*,'(1x,a,i0,a,i0,a)') 'NCALFA: Warning, lb=',lb,' im=',im,&
           ' cannot compute relaxation arrays for eastern and/or western edge'
      doeastwest=.false.
  end if

  if(lb > jm .and. (num_zones_north>0 .or. num_zones_south>0)) then
      if(master) write(*,'(1x,a,i0,a,i0,a)') 'NCALFA: Warning, lb=',lb,' im=',im,&
           ' cannot compute relaxation arrays for northern and/or southern edge'
      donorthsouth=.false.
  end if 
  
  if(jm==1) donorthsouth=.false.

  
! Precompute global relaxation arrays for u, v and eta points

  alfa_u=0.0
  alfa_v=0.0
  alfa_eta=0.0
  alfae_u=0.0
  alfae_v=0.0
  alfae_eta=0.0
  
  call getalfa(u_point, alfa_u, alpha) 
  call getalfa(v_point, alfa_v, alpha)
  call getalfa(eta_point, alfa_eta, alpha)
  call getalfa(u_point, alfae_u, alphae) 
  call getalfa(v_point, alfae_v, alphae)
  call getalfa(eta_point, alfae_eta, alphae)

  if(master)write(*,*)

contains
                  

  subroutine getalfa(ix, alfa, alf)
    implicit none
    integer ix
    real alf(:)
    real, pointer :: alfa(:,:)
    integer i,j,ifrs
    integer li1,li2,lj1,lj2,gi,gj
    logical inblock
    
!! BOUNDARY ZONE WEST 

    if(ix.eq.2) then

!! u-points:

        call globalrect2local(2,lb,1,jm, li1,li2,lj1,lj2, inblock)
        if(inblock.and.doeastwest)then            
            do j=lj1,lj2
                do i = li1,li2                
                    if(block%ibn==1) alfa(1,j)=1.0
                    gi=block%fi+i-1  ; gj=block%fj+j-1
                    
                    if (gj.le.lb.and.jm>1) then
                        if(sw_connected) then
                            ifrs = min(gi-1,gj)
                        else
                            ifrs = gi-1
                        end if
                    elseif (gj.ge.jm-lb.and.jm>1) then
                        if(nw_connected) then                    
                            ifrs = min(gi-1,gjm+1-gj)
                        else
                            ifrs = gi-1
                        end if
                    else
                        ifrs = gi-1
                    endif
                    alfa(i,j) = alf(ifrs)
                end do
            end do
        end if
    else

!! v-points / el-points
        call globalrect2local(1,lb,1,jm, li1,li2,lj1,lj2, inblock)
        if(inblock.and.doeastwest)then
            do j=lj1,lj2
                do i = li1,li2
                    gi=block%fi+i-1  ; gj=block%fj+j-1
                    
                    if (gj.le.lb.and.jm>1) then
                        if(sw_connected)then
                            ifrs = min(gi,gj)
                        else
                            ifrs = gi
                        end if
                    elseif (gj.ge.jm-lb.and.jm>1) then
                        if(nw_connected)then
                            ifrs = min(gi,jm+1-gj)
                        else
                            ifrs = gi
                        end if
                    else
                        ifrs = gi
                    endif
                    alfa(i,j) = alf(ifrs)
                end do
            end do
        end if ! inblock
    end if

!!  boundary zone east

!! u- or v- or el-points
    
    call globalrect2local(im-lb+1,im,1,jm, li1,li2,lj1,lj2, inblock)
    
    if(inblock.and.doeastwest)then
        do j=lj1,lj2
            do i = li1,li2
                gi=block%fi+i-1  ; gj=block%fj+j-1
                
                if (gj.le.lb.and.jm>1) then
                    if(se_connected) then
                        ifrs = min(im+1-gi,gj)
                    else
                        ifrs = im+1-gi
                    end if
                elseif (gj.ge.jm-lb.and.jm>1) then
                    if(ne_connected) then
                        ifrs = min(im+1-gi,jm+1-gj)
                    else
                        ifrs = im+1-gi
                    end if
                else
                    ifrs = im+1-gi
                endif
                alfa(i,j) = alf(ifrs)
            end do
        end do
    end if

    if(jm > 1)then
!!  boundary zone south

    if (ix == 3) then
!! v-points:
        if(doeastwest)then
            call globalrect2local(lb+1,im-lb, 2,lb, li1,li2,lj1,lj2, inblock)
        else
            call globalrect2local(1,im, 2,lb, li1,li2,lj1,lj2, inblock)
        end if
        
        if(inblock.and.donorthsouth)then
            do j=lj1,lj2
                do i = li1,li2
                    gi=block%fi+i-1  ; gj=block%fj+j-1
                    if(block%ibn == 1) alfa(i,1)=1.0
                    ifrs = gj-1
                    alfa(i,j) = alf(ifrs)
                end do
            end do
        end if
    else
!! u-points / el-points
        if(doeastwest)then
            call globalrect2local(lb+1,im-lb, 1,lb, li1,li2,lj1,lj2, inblock)
        else
            call globalrect2local(1,im, 1,lb, li1,li2,lj1,lj2, inblock)
        end if
        
        if(inblock.and.donorthsouth)then
            do j=lj1,lj2
                do i = li1,li2
                    gi=block%fi+i-1  ; gj=block%fj+j-1
                    ifrs = gj
                    alfa(i,j) = alf(ifrs)
                end do
            end do
        end if
    end if

!!   boundary zone north
!!
!! u- or v- or el-points

    if(doeastwest)then
        call globalrect2local(lb+1,im-lb, jm-lb+1,jm, li1,li2,lj1,lj2, inblock)
    else
        call globalrect2local(1,im, jm-lb+1,jm, li1,li2,lj1,lj2, inblock)
    end if
    
    if(inblock.and.donorthsouth)then
        do j=lj1,lj2
            do i = li1,li2
                gi=block%fi+i-1  ; gj=block%fj+j-1
                ifrs = jm+1-gj
                alfa(i,j) = alf(ifrs)
            end do
        end do
    end if
    
    end if  ! jm > 1

    call tag_ghost(alfa)
    call update_ghost()
    
  end subroutine getalfa


end subroutine ncalfa
