module norms
!
! various utility functions for the dump_state.f90 subroutine.
!  
! subroutine l6n2d(label,f1,[f2,f3,f4,f5,f6])
!   prints a label, as well as 1-6 sums of the given real, 1d, arrays fX.
!   the type of sum which is computed is given by "debugnorm"
!  
! subroutine l6n3d(label,f1,[f2,f3,f4,f5,f6])
!   same as above, but for 2d arrays.
!
! real function parsum(f)
!   uses the parallel sum psum to get a processor number independent sum of  
!   the real array f
!
! integer function linorm(f)
!   from 2d or 3d real array f finds an integer checksum by converting 
!   reals to char arrays and summing their ichar value we get a endian
!   independent sum
!
! subroutine find_norm(f)
!   called from dump_state. uses "debugnorm" to decide on what norm to print
!   given the list assembled in dump_state.
!
! subroutine print_norms
!   called from dump_state, pretty prints all norms specified in dump_state
!     
!  
!  Helge Avlesen <avle@ii.uib.no>  para//ab
! 


!! MPI bom
  use qmpi, only : dbl, dp, long, short, sp, master, myid, numproc, reduce, abort
  
  use mod_mpi, only : ghost_wi, ghost_wj, i_eq_im, j_eq_jm, gim, gjm, gkb, gridbox, &
       i1,i2,j1,j2, find_max_reals, tag_ghost, update_ghost, psum, block, omit_dry_blocks,&
       globalrect2local, i_eq_2, im_mi_1, j_eq_2, jm_mi_1
  
  implicit none
 
  
! What to print from dump_state:  
!   debuglevel:  0=zero output, 1=print label only, 2=maxmin and selected 3d, 3=also selected 2d
  integer :: debugnormlevel=0

! which norm to use:
!   debugnorm:  0=sum  1=max(abs()), 2=integer checksum  
  integer :: debugnorm=0   
  
  
  interface linorm
! linorm returns a long integer
!     
! Naive Integer checksum routine:
! the idea is to use transfer to convert the floats in each array to integers.
! the integers of each array are added together into a integer kind=8 checksum,
! so that it is possible to compare the checksum of each array from different runs. 
! the ints from a float can max be 2^32/2, so from e.g. 800x600x100 fields the checksum is
! < 103079215104000000. the headroom from the integer kind=8 sums is 2^64/2
! so (2^64/2)/(2^32/2*800*600*100) = 89 of these checksums can again be added together
! without danger of overflow.
! using short ints this headroom can be expanded
!
     module procedure li3d, dli3d, li2d, dli2d, li1d, lireal4, lireal8, liinteger4
  end interface

  interface charnorm
     module procedure char3d4, char3d8,char2d4,char2d8,char1d4,char1d8,charreal4,charreal8,charinteger4
  end interface
  
  interface parsum
     module procedure ps3d, ps2d, dp1d     
  end interface

  interface find_norm
     module procedure comp3d, comp2d, comp1d, compreal4, compreal8, compinteger
  end interface

  integer :: first_dump_state=0
  logical :: dump_state_barrier=.false.
  integer, parameter :: max_num_norms=200
  integer :: norm_num=0  
  real computed_norms(max_num_norms)
  integer(8) computed_linorms(max_num_norms)
  character(len=10) norm_labels(max_num_norms)
  

contains


  function kahansum(f,im,jm,kb,range)
!    
! Compute the block local sum of f in active points. optionally an
! array range can be provided. (range must be a block local range for
! a distributed array)    
! This routine should preferrably be compiled with no optimization,
! otherwise most compilers will rearrange operations of carefully
! orchestrated floating point operations.
! 
    implicit none
    integer im,jm,kb
    real f(i1:i2,j1:j2,kb)
    integer, optional :: range(6)
    integer r(6),i,j,k
    real(8) kahansum,c,y,t

    r = (/i_eq_2, im_mi_1, j_eq_2, jm_mi_1, 1, max(1,kb-1)/)
    if(present(range)) r=range
    
    kahansum = 0._8
    c = 0._8
    do k = r(5),r(6)
        do j = r(3),r(4)
            do i = r(1),r(2)
                if(gridbox(i,j) == 1) then
                    y = real(f(i,j,k),8) - c
                    t = kahansum + y
                    c = (t - kahansum) - y
                    kahansum = t
                end if
            end do
        end do
    end do
  end function kahansum

  
  subroutine l6n2d(label,f1,f2,f3,f4,f5,f6)
    implicit none
    character(len=*) label
    character(len=10) l2
    real, optional, pointer, dimension(:,:) :: f1,f2,f3,f4,f5,f6
    real nrm(6)
    integer(8) linrm(6)
    integer n
    
     l2=" "
    if(len(label)>10) then
        l2=label(:10)
    else        
        l2=label
    end if    
   
!    l2=label
    n=0
    if(present(f1)) call eval(f1)
    if(present(f2)) call eval(f2)
    if(present(f3)) call eval(f3)
    if(present(f4)) call eval(f4)
    if(present(f5)) call eval(f5)
    if(present(f6)) call eval(f6)
    if(master) then
        select case(debugnorm)
        case(0)
            write(*,*)l2,nrm(1:n)
        case(1)
            write(*,*)l2,nrm(1:n)
        case(2,3)
            write(*,'(a,6(1x,i13))') l2,linrm(1:n)
        end select
    end if
    return
  contains
    subroutine eval(f)
      integer loc(3)      
      real, pointer :: f(:,:)
      n=n+1
      select case(debugnorm)
      case(0)
          nrm(n)=parsum(f)
      case(1)
          call find_max_reals(1, f, nrm(n), loc)
      case(2)
          linrm(n)=linorm(f)
      case(3)
          linrm(n)=charnorm(f)
      end select
    end subroutine eval
  end subroutine l6n2d
  
  subroutine l6n3d(label,f1,f2,f3,f4,f5,f6)
    implicit none
    character(len=*) label
    character(len=10) l2    
    real, optional, pointer, dimension(:,:,:) :: f1,f2,f3,f4,f5,f6
    real nrm(6)
    integer(8)linrm(6)
    integer n
    
    n=0
    l2=" "
    if(len(label)>10) then
        l2=label(:10)
    else        
        l2=label
    end if
    
    if(present(f1)) call eval(f1)
    if(present(f2)) call eval(f2)
    if(present(f3)) call eval(f3)
    if(present(f4)) call eval(f4)
    if(present(f5)) call eval(f5)
    if(present(f6)) call eval(f6)
    if(master) then
        select case(debugnorm)
        case(0)
            write(*,*)l2,nrm(1:n)
        case(1)
            write(*,*)l2,nrm(1:n)
        case(2,3)
            write(*,'(a,6(1x,i13))') l2,linrm(1:n)
        end select
    end if
  contains
    subroutine eval(f)
      integer loc(3)
      real, pointer :: f(:,:,:)
      n=n+1
      select case(debugnorm)
      case(0)
          nrm(n)=parsum(f)
      case(1)
          call find_max_reals(size(f,3), f, nrm(n), loc)          
      case(2)
          linrm(n)=linorm(f)
      case(3)
          linrm(n)=charnorm(f)
      end select
    end subroutine eval    
    
  end subroutine l6n3d

  
  function ps3d(f,r)
    implicit none
    real ps3d
    real, pointer :: f(:,:,:), tmpf(:,:,:)
    integer, optional :: r(:)
    integer r2(6)
    logical inblock
    real(8) tmpsum
    
    !! we do this copy and full update because psum uses data in
    !! the ghost zone, and we want the sum to be correct even if a
    !! variable is not (yet) fully up to date in the ghost.

    allocate( tmpf(i1:i2,j1:j2,size(f,3)) )
    tmpf=f
    call tag_ghost(tmpf)
    call update_ghost
    if(present(r))then
        ps3d = psum(tmpf,gim,gjm,size(f,3), r)
    else
        ps3d = psum(tmpf,gim,gjm,size(f,3))
    end if
    deallocate( tmpf )
  end function ps3d
  
  function ps2d(f,r)
    implicit none
    real ps2d
    real, pointer :: f(:,:)
    integer, optional :: r(:)
    real, pointer :: tmpf(:,:,:)
    integer r2(6)
    logical inblock
    real(8) tmpsum
    
    allocate( tmpf(i1:i2,j1:j2,1) )    
    tmpf(:,:,1) = f
    
    call tag_ghost(tmpf)
    call update_ghost
    
    ps2d = psum(tmpf,gim,gjm,1,r)
    deallocate(tmpf)
  end function ps2d 
  
  function dp1d(f,r)
    implicit none
    integer, parameter :: dbl=8
    real(dbl) dp1d
    real, dimension(:) :: f
    integer, optional :: r(:)
    real(dbl) :: sum
    integer i

    sum=0.0_dbl
    if(present(r))then
        if(r(2) > size(f).or.r(1)<1) call abort("norms.f90, dp1d error: invalid range")
        do i=r(1),r(2)
            sum=sum+real( f(i),dbl)
        end do
    else
        do i=1,size(f)
            sum=sum+real( f(i),dbl)
        end do
    end if
    dp1d=sum
  end function dp1d
  
!--------------------  
  
  function char3d8(f)
    implicit none
    integer(long) char3d8
    real(8) f(1-ghost_wi:,1-ghost_wj:,:)
    real(8) ftmp(I_eq_IM*J_eq_JM*size(f,3))
    character mold, shorts(8*I_eq_IM*J_eq_JM*size(f,3) )
    integer(long) sum
    integer i,j,k,l

    k=0
    if(block%wet > 0) then
        do l=1,size(f,3)
            do j=1,j_eq_jm
                do i=1,i_eq_im
                    if(gridbox(i,j) == 1) then
                        k = k + 1
                        ftmp(k) = f(i,j,l)
                    end if
                end do
            end do
        end do
    end if

    sum=0_long
    
    if(k > 0) then
        shorts(1:8*k) = transfer(ftmp(1:k), mold, 8*k)
        do i=1,8*k
            sum = sum + ichar(shorts(i))
        end do
    end if

    if(numproc > 1) call reduce('sum',sum)
    char3d8=sum
  end function char3d8
  
  function char3d4(f)
    implicit none
    integer(long) char3d4
    real(4) f(1-ghost_wi:,1-ghost_wj:,:)
    real(4) ftmp(I_eq_IM*J_eq_JM*size(f,3))
    character mold, shorts( 4*I_eq_IM*J_eq_JM*size(f,3) )
    integer(long) sum
    integer i,j,k,l

    k=0
    if(block%wet > 0) then
        do l=1,size(f,3)
            do j=1,j_eq_jm
                do i=1,i_eq_im
                    if(gridbox(i,j) == 1) then
                        k = k + 1
                        ftmp(k) = f(i,j,l)
                    end if
                end do
            end do
        end do
    end if

    sum=0_long
    
    if(k > 0) then
        shorts(1:4*k) = transfer(ftmp(1:k), mold, 4*k)
        do i=1,4*k
            sum = sum + ichar(shorts(i))
        end do
    end if

    if(numproc > 1) call reduce('sum',sum)
    char3d4=sum
  end function char3d4

  function char2d8(f)
    implicit none
    integer(long) char2d8
    real(8) f(1-ghost_wi:,1-ghost_wj:)
    real(8) ftmp(I_eq_IM*J_eq_JM)
    character mold, shorts(8*I_eq_IM*J_eq_JM)
    integer(long) sum
    integer i,j,k
    
    k=0
    if(block%wet > 0) then
        do j=1,j_eq_jm
            do i=1,i_eq_im
                if(gridbox(i,j) == 1) then
                    k = k + 1
                    ftmp(k) = f(i,j)
                end if
            end do
        end do
    end if

    sum=0_long
    
    if(k > 0) then
        shorts(1:8*k) = transfer(ftmp(1:k), mold, 8*k)
        do i=1,8*k
            sum = sum + ichar(shorts(i))
        end do
    end if

    if(numproc > 1) call reduce('sum',sum)
    
    char2d8=sum
  end function char2d8
  
  function char2d4(f)
    implicit none
    integer(long) char2d4
    real(4) f(1-ghost_wi:,1-ghost_wj:)
    real(4) ftmp(I_eq_IM*J_eq_JM)
    character mold, shorts(4*I_eq_IM*J_eq_JM)
    integer(long) sum
    integer i,j,k

    k=0
    if(block%wet > 0) then
        do j=1,j_eq_jm
            do i=1,i_eq_im
                if(gridbox(i,j) == 1) then
                    k = k + 1
                    ftmp(k) = f(i,j)
                end if
            end do
        end do
    end if

    sum=0_long
    
    if(k > 0) then
        shorts(1:4*k) = transfer(ftmp(1:k), mold, 4*k)
        do i=1,4*k
            sum = sum + ichar(shorts(i))
        end do
    end if

    if(numproc > 1) call reduce('sum',sum)

    char2d4=sum
  end function char2d4

  function char1d8(f)
    implicit none
    integer(long) char1d8
    real(8) f(:)
    character mold, shorts(8*size(f,1))
    integer(long) sum
    integer i

    shorts = transfer(f, mold, 8*size(f,1)) 
    sum=0_long
    do i=1,8*size(f)
        sum=sum + ichar(shorts(i))
    end do
    char1d8=sum
  end function char1d8
  
  function char1d4(f)
    implicit none
    integer(long) char1d4
    real(4) f(:)
    character mold, shorts(4*size(f))
    integer(long) sum
    integer i

    shorts=transfer(f, mold, 4*size(f)) 
    sum=0_long
    do i=1,4*size(f)
        sum=sum+ichar(shorts(i))
    end do
    char1d4=sum
  end function char1d4

  function charreal4(f)
    implicit none
    integer(long) charreal4
    character mold, shorts(4)
    real(4) f
    integer i
    
    shorts=transfer(f, mold, 4)
    charreal4=0_long
    do i=1,4
        charreal4=charreal4+ichar(shorts(i))
    end do
  end function charreal4
  
  function charreal8(f)
    implicit none
    integer(long) charreal8
    real(8) f
    character mold, shorts(8)
    integer i
    
    shorts=transfer(f, mold, 8)
    charreal8=0_long
    do i=1,8
        charreal8=charreal8+ichar(shorts(i))
    end do
  end function charreal8
  
  function charinteger4(f)
    implicit none
    integer(long) charinteger4
    integer(4) f
    character mold, shorts(4)
    integer i
    
    shorts=transfer(f, mold, 4)
    charinteger4=0_long
    do i=1,4
        charinteger4=charinteger4+ichar(shorts(i))
    end do    
  end function charinteger4

  
!---------------------
  
  function li3d(f)
    implicit none
    integer(long) li3d
    real(4) f(1-ghost_wi:,1-ghost_wj:,:)
    real(4) ftmp(I_eq_IM*J_eq_JM*size(f,3))
    integer(short) mold, shorts( 2*I_eq_IM*J_eq_JM*size(f,3) )
    integer(long) sum
    integer i

    ftmp=reshape( f(1:I_eq_IM,1:J_eq_JM,:), (/I_eq_IM*J_eq_JM*size(f,3)/) ) 
    shorts=transfer( ftmp, mold, 2*I_eq_IM*J_eq_JM*size(f,3) ) 
    sum=0_long
    do i=1,2*I_eq_IM*J_eq_JM*size(f,3)
        sum=sum+int( shorts(i),long)
    end do
    if(numproc>1)call reduce('sum',sum)
    li3d=sum
  end function li3d

  function dli3d(f)
    implicit none
    integer(long) dli3d
    real(8) f(1-ghost_wi:,1-ghost_wj:,:)
    real(8) ftmp(I_eq_IM*J_eq_JM*size(f,3))
    integer(short) mold, shorts( 4*I_eq_IM*J_eq_JM*size(f,3) )
    integer(long) sum
    integer i

    ftmp=reshape( f(1:I_eq_IM,1:J_eq_JM,:), (/I_eq_IM*J_eq_JM*size(f,3)/) ) 
    shorts=transfer( ftmp, mold, 4*I_eq_IM*J_eq_JM*size(f,3) ) 
    sum=0_long
    do i=1,4*I_eq_IM*J_eq_JM*size(f,3)
        sum=sum+int( shorts(i),long)
    end do
    if(numproc>1)call reduce('sum',sum)
    dli3d=sum
  end function dli3d

  function li2d(f)
    implicit none
    integer(long) li2d
    real(4) f(1-ghost_wi:,1-ghost_wj:)
    real(4) ftmp(I_eq_IM*J_eq_JM)
    integer(short) mold, shorts( 2*I_eq_IM*J_eq_JM )
    integer(long) sum
    integer i

    ftmp=reshape( f(1:I_eq_IM,1:J_eq_JM), (/I_eq_IM*J_eq_JM/) )
    shorts=transfer( ftmp, mold, 2*I_eq_IM*J_eq_JM )

    sum=0_long
    do i=1,2*I_eq_IM*J_eq_JM
        sum=sum+int( shorts(i),long)
    end do
    if(numproc>1)call reduce('sum',sum)
    li2d=sum
  end function li2d

  function dli2d(f)
    implicit none
    integer(long) dli2d
    real(dbl) f(1-ghost_wi:,1-ghost_wj:)
    real(dbl) ftmp(I_eq_IM*J_eq_JM)
    integer(short) mold, shorts( 4*I_eq_IM*J_eq_JM )
    integer(long) sum
    integer i

    ftmp=reshape( f(1:I_eq_IM,1:J_eq_JM), (/I_eq_IM*J_eq_JM/) )
    shorts=transfer( ftmp, mold, 4*I_eq_IM*J_eq_JM )

    sum=0_long
    do i=1,4*I_eq_IM*J_eq_JM
        sum=sum+int( shorts(i),long)
    end do
    if(numproc>1)call reduce('sum',sum)
    dli2d=sum
  end function dli2d

  function li1d(f)
! f assumed to be defined in the vertical direction  
    implicit none
    integer, parameter :: long=8, short=2
    integer(long) li1d
    real f(:)
    integer(short) mold, shorts(2*size(f))
    integer(long) sum
    integer i

    shorts=transfer( f, mold, 2*size(f))
    sum=0_long
    do i=1,2*size(f)
        sum=sum+int( shorts(i),long)
    end do
    li1d=sum
  end function li1d

  function lireal4(f)
    implicit none
    integer, parameter :: long=8, short=2
    integer(long) lireal4
    real(4) f
    integer(short) mold, shorts(2)
    integer(long) sum
    integer i

    shorts=transfer( f, mold, 2)
    sum=0_long
    do i=1,2
        sum=sum+int( shorts(i),long)
    end do
    lireal4=sum
  end function lireal4
  
  function lireal8(f)
    implicit none
    integer, parameter :: long=8, short=2
    integer(long) lireal8
    real(8) f
    integer(short) mold, shorts(4)
    integer(long) sum
    integer i

    shorts=transfer( f, mold, 4)
    sum=0_long
    do i=1,4
        sum=sum+int( shorts(i),long)
    end do
    lireal8=sum
  end function lireal8  

  function liinteger4(f)
    implicit none
    integer, parameter :: long=8, short=2
    integer(long) liinteger4
    integer(4) f
    integer(short) mold, shorts(2)
    integer(long) sum
    integer i

    shorts=transfer( f, mold, 2)
    sum=0_long
    do i=1,2
        sum=sum+int( shorts(i),long)
    end do
    liinteger4=sum
  end function liinteger4  
  
  subroutine comp3d(label,f)
    real, pointer :: f(:,:,:)
    character(len=*) label
    integer loc(3)
    
    norm_num=norm_num+1
    norm_labels(norm_num)=label    

    select case(debugnorm)
    case(0)
        computed_norms(norm_num)=parsum(f)
    case(1)
        !computed_norms(norm_num)=maxval(abs(f))
        call find_max_reals(size(f,3), f, computed_norms(norm_num), loc)          
    case(2)
        computed_linorms(norm_num)=linorm(f)
    case(3)
        computed_linorms(norm_num)=charnorm(f)
    end select
  end subroutine comp3d

  subroutine comp2d(label,f)
    real, pointer :: f(:,:)
    integer loc(3)
    
    character(len=*) label
    norm_num=norm_num+1

    norm_labels(norm_num)=label    

    select case(debugnorm)
    case(0)
        computed_norms(norm_num)=parsum(f)
    case(1)
        !computed_norms(norm_num)=maxval(abs(f))
        call find_max_reals(1, f, computed_norms(norm_num), loc)          
    case(2)
        computed_linorms(norm_num)=linorm(f)
    case(3)
        computed_linorms(norm_num)=charnorm(f)
    end select
  end subroutine comp2d  

  subroutine comp1d(label,f)
    real f(:)
    character(len=*) label
    norm_num=norm_num+1

    norm_labels(norm_num)=label    

    select case(debugnorm)
    case(0)
        computed_norms(norm_num)=parsum(f)
    case(1)
        computed_norms(norm_num)=maxval(abs(f))
    case(2)
        computed_linorms(norm_num)=linorm(f)
    case(3)
        computed_linorms(norm_num)=charnorm(f)
    end select
  end subroutine comp1d

  subroutine compreal4(label,f)
    real(4) f
    character(*) label
    norm_num=norm_num+1
    norm_labels(norm_num)=label    
    select case(debugnorm)
    case(0)
        computed_norms(norm_num)=f
    case(1)
        computed_norms(norm_num)=f
    case(2)
        computed_linorms(norm_num)=linorm(f)
    case(3)
        computed_linorms(norm_num)=charnorm(f)
    end select
  end subroutine compreal4  

  subroutine compreal8(label,f)
    real(8) f
    character(*) label
    norm_num=norm_num+1
    norm_labels(norm_num)=label    
    select case(debugnorm)
    case(0)
        computed_norms(norm_num)=f
    case(1)
        computed_norms(norm_num)=f
    case(2)
        computed_linorms(norm_num)=linorm(f)
    case(3)
        computed_linorms(norm_num)=charnorm(f)
    end select
  end subroutine compreal8
  
  subroutine compinteger(label,f)
    integer f
    character(*) label
    norm_num=norm_num+1
    norm_labels(norm_num)=label    
    select case(debugnorm)
    case(0)
        computed_norms(norm_num)=f
    case(1)
        computed_norms(norm_num)=f
    case(2)
        computed_linorms(norm_num)=linorm(f)
    case(3)
        computed_linorms(norm_num)=charnorm(f)
    end select
  end subroutine compinteger

  
  subroutine print_norms
!    
! pretty print norms
!
    implicit none
    character(len=10) tmplabel(4)
    character(len=40) flabel
    character(200) long_line
    real tmpnorms(4), totnorm
    integer n,count,i
    integer(long) checksum, tmplinorms(4)

    count = 0
    totnorm = 0.0_dbl
    checksum = 0

    do n=1,norm_num
        count=count+1
        
        tmplabel(count) = "          "
        tmplabel(count) = norm_labels(n)

        select case(debugnorm)
        case(0:1)
            tmpnorms(count)=computed_norms(n)
        case(2,3)
            tmplinorms(count) = computed_linorms(n)
        end select
        
        if(debugnorm==0)then
            totnorm = totnorm + tmpnorms(count)
        end if
        if(debugnorm==2.or.debugnorm==3)then
            checksum = checksum + tmplinorms(count)
        end if
        
        if(master)then
            if(count==4) then
                write(flabel,'(3(a,","),a)') (trim(tmplabel(i)),i=1,4)
                select case(debugnorm)
                case(0)
                    write(long_line,*) flabel, tmpnorms(1:4)
                    write(*,'(a)') trim(long_line)
                case(1)
                    write(*,'(a,4(1x,g12.5))') flabel, (real(tmpnorms(i)),i=1,4)
                case(2,3)
                    write(*,10) sum(tmplinorms(1:4)), flabel, (tmplinorms(i),i=1,4)
                end select
            else if(n==norm_num .and. count>0) then
                write(flabel,'(3(a,","),a)') (trim(tmplabel(i)),i=1,count)
                select case(debugnorm)
                case(0)
                    write(long_line,*) flabel, tmpnorms(1:count)
                    write(*,'(a)') trim(long_line)
                case(1)
                    write(*,'(a,4(1x,g12.5))') flabel, (real(tmpnorms(i)),i=1,count)
                case(2,3)
                    write(*,10) sum(tmplinorms(1:count)), flabel, (tmplinorms(i),i=1,count)
                end select
            end if
        end if
        if(count==4) count=0
    end do
    
!10 format(i14,1x,a,4(1x,i0))
10  format(i12,1x,a,4(1x,i12))  
    
    if(master)then
!      if(debugnorm==0)then
!          write(*,'(g14.8,1x,a)') totnorm, 'Sum of all norms'
!      end if
        if(debugnorm==2.or.debugnorm==3)then
            write(*,'(i14,1x,a)') checksum, 'Sum of all norms'          
        end if
!        write(*,'(a)') "______________________________________________________________________________"
        write(*,*)
    end if
    
    norm_num=0
  end subroutine print_norms
  
end module norms
