! A module containing BOM specific quick interfaces to the pgplot library.
! In the code, make the interface available via
!    ...
!    use pgplot
!    ...
! In the Makefile, add $(PGPLOT) to the link command, where PGPLOT
! can be something like this (linux example)
!    ...
!    PGPLOT= -L$(HOME)/linux/pgplot-5.2 -lpgplot -lpng -lz \
!        -L/usr/X11R6/lib -lX11 -Wl,-rpath $(HOME)/linux/pgplot-5.2
!    ...
!
! Usage: 2 ways
! 1) by specifying predifined keywords in the setup file
!
! See documentation for keywords starting with "pgplot_" in setup2.f90
! or via running bomhelp.py
!
! Basically each plotting keyword specified, adds a plot object to the
! list "pgplots" below. the routine batch_pgplot (called once per
! timestep) loops over the list and does whatever is specified in the
! object.
!
! 2) calling the interface routines directly
!
! The interface; (optional arguments in [])
!
! subroutine init_pgplot()
!
! pgplot will only run on master, and rely on data beeing present in
! its own global arrays. we use bomplot with a special flag to fetch
! the data.
!
module pgplot

  use pgplot_core
  implicit none
  integer, parameter :: sp=4
  logical :: pgplot_do_init=.true.
  character(22) extratitle

  integer :: pgplot_numdumploc=0
  character(50) pgplot_dump_location(20)
  
contains
  
  subroutine batch_pgplot(fromwhere)    
!! iterate through list of pgplots and see if we should do something.
    use state, only : istep, dt, master
    use time_util, only : ndate
    use user_pgplot
    implicit none
    character(*), optional :: fromwhere
    integer i
    logical gotloc

! reasons to do nothing
    if(pgplot_numplots == 0) return

! if we are going to do pgplot_timeseries, fetch as much timeseries
! data as possible from timeseries arrays before it reallocates:
    call pgplot_gatherts()
    
    if(istep < pgplot_interval(1) .or. istep > pgplot_interval(2)) return    
    if(.not. mod(istep, pgplot_interval(3)) == 0) return

    extratitle=" "
    select case(pgplot_title_time)
    case(1)
        write(extratitle,'(1x,i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2)') ndate
    case(2)
        write(extratitle,'(1x,"t=",f0.2,"s")') istep*dt        
    end select
    
    if(dump_state_pgplot > 0) then
        if(present(fromwhere))then
            extratitle = fromwhere

            gotloc=.false.
            do i=1,pgplot_numdumploc
                if(fromwhere == pgplot_dump_location(i)) gotloc=.true.
            end do           
            if(.not.gotloc) return
        end if
    end if

    if(pgplot_do_init) call init_pgplot()

    if(pgplot_cache_results > 0 .and. master) then
        if(mod(istep, pgplot_cache_interval)==0) call pgplot_write_cache()
    end if

    
    do i=1,max(pgplot_numplots, pgplot_layout(1)*pgplot_layout(2))

!! (re)set default viewport for each plot, as many routines use pgwnad        
        if(master) call pgsvp(pgplot_viewport(1), pgplot_viewport(2), pgplot_viewport(3), pgplot_viewport(4))

        select case(pgplots(i)%type)
        case(0)
            if(master) call pgpage()           
        case(1)
            call pgquick2d(pgplots(i))
            if(master) call pghsect_particles(pgplots(i))
            call user_pgquick2d(pgplots(i))
        case(2)
            call pghsect(pgplots(i))
            if(master) call pghsect_particles(pgplots(i))
            call user_pghsect(pgplots(i))            
        case(3)
            call pgvsect(pgplots(i))
            if(master) call pgvsect_particles(pgplots(i))
            call user_pgvsect(pgplots(i))           
        case(4)
            call pgtimeseries(pgplots(i))
        case(5)
            call pgvprof(pgplots(i))            
        case(6,7)
            call pgslice(pgplots(i))
        case(8)
            call pgatmos(pgplots(i))
        case(9)
            call pggeo(pgplots(i))
            call user_pggeo(pgplots(i))
        case default
            write(*,*)'batch pgplot error: unknown plot type=', pgplots(i)%type
        end select
    end do
  end subroutine batch_pgplot


  subroutine init_pgplot()
!
! set reasonable defaults for pgplot
! 
    use state, only : im, jm, h, dx, dy
    use mod_mpi, only : master, myid, gatherfield
    implicit none    

    if(pgplot_core_do_init) call init_pgplot_core(procid=myid)
    
    call allocate_pgplot()

    if(master) call get_cached_results()
    
! transform used by pcolor(i.e.pgimag) to map from i,j to x,y
    pgplot_transform = (/-1.5*dx, dx, 0.0, -1.5*dy, 0.0, dy/)
    
! initialize the land mask used by hsect    
    call gatherfield(h, pg_f, im,jm,1)
    if(master)then
        where(pg_f > 0.0001)
            pgmask = 1
        elsewhere
            pgmask = 0
        end where
    end if

    pgplot_do_init = .false.
  end subroutine init_pgplot


  subroutine pgplot_gatherts
!    
! calls gatherts if BATCH_PGPLOT will be called later, and we have at
! least one pgplot timeseries specified.
!    
    use qmpi, only : numproc, master
    use state, only : istep, numstep, debug
    use control_parameters, only : do_checkpoint, restart, first_step
    use timeseries, only : ts, timeseries_dump_start, gatherts, timeseries_interval
    implicit none
    logical time_to_dump
    integer numsamp, i, step1, step2
   
    if(pgplot_numts == 0) return
    if(pgplot_numplots == 0) return
    if(restart .and. istep < first_step) return
    
! if this is true, timeseries will dump data later this timestep and reallocate the
! arrays, i.e. we must copy that data before timeseries is called.
    time_to_dump = istep > 0 .and. (mod(istep,timeseries_interval)==0  .or. istep==numstep  .or.  do_checkpoint)

! if we're out of the pgplot interval it is safe to return   
    if(.not.time_to_dump) then
        if(istep < pgplot_interval(1) .or. istep > pgplot_interval(2)) return    
        if(.not. mod(istep, pgplot_interval(3)) == 0) return
    end if

! this routine must be called before the timeseries reallocation 
! happen. fetch as much as possible; try to fetch pgplot_inteval
! values back, but a reallocation may have happended during the
! interval, if so fetch from timeseries_dump_start.

    step1 = max(istep - pgplot_interval(3) + 1, timeseries_dump_start)
    step2 = istep
    if(numproc > 1) call gatherts(step1, step2)

! save complete timeseries data in the pgplot object

    if(master)then
        do i=1, pgplot_numplots
            if(pgplots(i)%type==4)then
                pgplots(i)%data1d(step1:step2) = ts(pgplots(i)%ts_i)%data(step1:step2)                
            end if
        end do
    end if
  end subroutine pgplot_gatherts


  subroutine get_cached_results()
    use state, only : numstep, debug, dt
    implicit none
    character(200) filename
    integer i
    logical got_you
    real dummy
    
    do i=1, pgplot_numplots
        if(pgplots(i)%type==4)then
            write(filename,'(".cachedpgts-",i0,".sr",i0)') i, kind(dummy)
            inquire(file=filename, exist=got_you)
            if(got_you) then
                if(debug > 1) write(*,'(1x,a,a)') "Found pgplot timeseries cache ",trim(filename)
                open(1000007, file=filename, access="stream", form='unformatted')
                if(.not.allocated(pgplots(i)%cache1d)) allocate(pgplots(i)%cache1d(0:numstep))
                pgplots(i)%cache1d(0:numstep) = 0.0
             ! read as much as possible, be silent on end of file
                read(1000007, end=101, err=102) pgplots(i)%cache1d(0:numstep)
101             close(1000007)
            end if
        end if
        cycle
102     write(*,*) " There was a non critical error reading pgplot cache file "//trim(filename)
    end do

    write(filename,'(".cachedpgt.sr",i0)') kind(dummy)
    inquire(file=filename, exist=got_you)
    if(got_you) then
        open(1000007, file=filename, access="stream", form='unformatted')
        read(1000007, end=103, err=103) pgplot_cachedt
        close(1000007)
        goto 104
103     write(*,*) " could not read DT for cached timeseries"    
    end if
! overwrite with the new timestep length
104 if(pgplot_cache_results>0)then
        open(1000007, file=filename, access="stream", form='unformatted')
        write(1000007) dt
        close(1000007)
    end if
  end subroutine get_cached_results

  
  subroutine pgplot_write_cache()
    use state, only : debug, istep
    implicit none
    character(200) filename
    integer i
    logical got_you
    real dummy

    if(pgplot_cache_results==0) return
    
    do i=1, pgplot_numplots
        if(pgplots(i)%type==4)then
            write(filename,'(".cachedpgts-",i0,".sr",i0)') i, kind(dummy)
            if(debug > 2) write(*,'(1x,a,a)') "Updating pgplot timeseries cache ",trim(filename)
            open(1000007, file=filename, access="stream", form='unformatted')           
            write(1000007, err=102) pgplots(i)%data1d(0:istep)
            close(1000007)
        end if
        cycle
102     write(*,*) " failed writing to pgplot cache file "//trim(filename)
    end do
  end subroutine pgplot_write_cache
  

  subroutine pg_vimag(f, x, z, fmin, fmax, grid, colorbar)
!    
! draw "vertical pcolor" using filled polygons.
! speedwise, it is a "must" to use pgbbuf and pgebuf with this routine
!    
    use state, only : dz,dx,dy
    implicit none
    real, intent(in) :: f(:,:), x(:,:), z(:,:)
    real, intent(in), optional ::  fmin, fmax
    logical, optional :: grid
    logical, optional :: colorbar
    integer :: pn=4
    integer :: i,j, ie, je, ci, cmin, cmax
    real(sp) px(4),py(4)
    real(sp) displ, width,gfmax,gfmin
    real(sp), allocatable :: xg(:,:), yg(:,:)

    ie=size(f,1) 
    je=size(f,2)

    call pgsfs(1)  ! fill style solid

    if(present(fmin)) then
        gfmin=fmin
    else
        gfmin=minval(f)
    end if
    if(present(fmax)) then
        gfmax=fmax
    else
        gfmax=maxval(f)
    end if

! available color range (0-15 is predefined)
    call pgqcol (cmin, cmax)    
    cmin=16

    if(gfmax == gfmin)then
        write(*,*)'pg_vimag: fmax=fmin, '
        gfmax = gfmax + 0.01*abs(gfmax)
        gfmin = gfmin - 0.01*abs(gfmax)
        if(gfmax == 0.0)then
            gfmax = 0.01 ; gfmin=-0.01
        end if
    end if

    allocate(xg(ie+1,je+1), yg(ie+1,je+1))

! get the coordinates xg,yg of the corners of the "cells" we plot in the
! vertical:

    do j=1,je+1
        do i=1,ie+1
            if(i>1 .and. i<ie+1)then
                if(j<je+1)then
                    xg(i,j) = 0.5*(x(i,j) + x(i-1,j))                    
                    yg(i,j) = 0.5*(z(i,j) + z(i-1,j)) - 0.25*dz(j)*(z(i,je)+z(i-1,je))
                else
                    xg(i,j) = 0.5*(x(i,j-1) + x(i-1,j-1))                    
                    yg(i,j) = 0.5*(z(i,je) + z(i-1,je)) + 0.25*dz(j)*(z(i,je)+z(i-1,je))
                end if
            else if(i==1)then
                if(j<je+1)then
                    xg(i,j) = x(i,j) - 0.5*(x(i+1,j)-x(i,j))
                    yg(i,j) = z(i,j) - 0.5*dz(j)*z(i,je)
                else
                    xg(i,j) = x(i,j-1) - 0.5*(x(i+1,j-1)-x(i,j-1))                    
                    yg(i,j) = z(i,je) + 0.5*dz(j)*z(i,je)
                end if
            else if(i==ie+1) then
                if(j<je+1)then
                    xg(i,j) = x(i-1,j) + 0.5*(x(i-1,j)-x(i-2,j))
                    yg(i,j) = z(i-1,j) - 0.5*dz(j)*z(i-1,je)
                else
                    xg(i,j) = x(i-1,j-1) + 0.5*(x(i-1,j-1)-x(i-2,j-1))                    
                    yg(i,j) = z(i-1,je) + 0.5*dz(j)*z(i-1,je)
                end if
            end if
        end do
    end do

    call pgbbuf

    do j=1,je-1
        do i=1,ie
!! vertices: sw,se,ne,nw
            px=(/xg(i,j+1),xg(i+1,j+1),xg(i+1,j),xg(i,j)/)
            py=(/yg(i,j+1),yg(i+1,j+1),yg(i+1,j),yg(i,j)/)

            ci = cmin + nint((cmax-cmin)*(f(i,j)-gfmin)/(gfmax-gfmin))
            if(ci < cmin) ci=0
            if(ci > cmax) ci=cmax

            call pgsci(ci)
            call pgpoly(pn, px, py)

            if(present(grid)) then
                if(grid) then
                    call pgsci(pgplot_text_color)
                    call pgline(pn,px,py)
                end if
            end if
        end do
    end do

    call pgebuf
    deallocate(xg, yg)

    if(present(colorbar)) then
        if(colorbar)then
            displ = 2.0_4  
            width = 4.0_4  
            call pgsci(pgplot_grid_color)  
            call pgsch(pgplot_character_height*0.7_4)  
            call pgwedg('RI', displ, width, gfmin, gfmax ,' ')  
        end if
    end if
  end subroutine pg_vimag


  subroutine draw_coastline(color)
!    
! draw the "ragged" coastline defined by a masking array pgmask, a
! global 2d array which is the integer version of fsm; 0 means land
! and 1 means water.
!
    use state, only : im,jm,dx,dy
    
    implicit none
    integer, optional :: color
    integer i,j,n
    real(4) x,y
    
    if(present(color))then
        call pgsci(color)
    else
        call pgsci(pgplot_text_color)
    end if

    n = pgplot_coastline_segments
    if(n > 0)then
        call pldj(pgcl(1,1:n), pgcl(2,1:n), pgcl(3,1:n), pgcl(4,1:n), color=color)
        return
    end if

    
! horisontal line segments

    call pgbbuf()

    do j=1,jm-1
        do i=1,im
            if(pgmask(i,j) .ne. pgmask(i,j+1)) then
                x=dx*(i-2)
                y=dy*(j-1)
                call pgmove(x,y)
                x=dx*(i-1)
                call pgdraw(x,y)
            end if
        end do
    end do

! vertical segs

    do j=1,jm
        do i=1,im-1
            if(pgmask(i,j) .ne. pgmask(i+1,j)) then
                x = dx*(i-1)
                y = dy*(j-2)
                call pgmove(x,y)
                y = dy*(j-1)
                call pgdraw(x,y)
            end if
        end do
    end do

! boundary 

    do j=1,jm
        i=1
        if(pgmask(i,j) == 0) then
            x = dx*(i-2)
            y = dy*(j-2)
            call pgmove(x,y)
            y = dy*(j-1)
            call pgdraw(x,y)
        end if
        i=im
        if(pgmask(i,j) == 0) then
            x = dx*(i-1)
            y = dy*(j-2)
            call pgmove(x,y)
            y = dy*(j-1)
            call pgdraw(x,y)
        end if
    end do
    do i=1,im
        j=1
        if(pgmask(i,j) == 0)then
            x = dx*(i-2)
            y = dy*(j-2)
            call pgmove(x,y)
            x = dx*(i-1)
            call pgdraw(x,y)
        end if
        j=jm
        if(pgmask(i,j) == 0)then
            x = dx*(i-2)
            y = dy*(j-1)
            call pgmove(x,y)
            x = dx*(i-1)
            call pgdraw(x,y)
        end if
    end do

    call pgebuf()
  end subroutine draw_coastline


  subroutine pgslice(slice)
!    
! This routine plots a slice of an array specified in the pgplot object "slice".
! Pretty close to all distributed arrays in BOM are supported.
! In setup2 the keyword line is
!    pgplot_slice <label> <what> <slice> <ind> <range>
! To call this routine directly, initialize the object like below to plot U in the surface layer
!    use setup2
!    type(pgplot_object) s
!    ...    
!    s%type = 6              ! 6 for 3d, 7 for 2d.
!    s%f3d => u              ! makes s%f3d point to u
!    s%slice = "k"
!    s%k = 1    
!    s%range = (/1,im,1,jm/)
!    s%title = "U(:,:,1) (m/s)"
!    call pgslice(s)
!
    use state, only : master, im, jm, kb, i1, j1, gatherfield, abort
    implicit none
    type(pgplot_object) slice
    integer r(4)
    real backup_transform(6)

! transform used by pcolor(i.e.pgimag) to map from i,j to x,y by
! default we use real world coordinates, so temporarily revert to
! index space for slicing:
    backup_transform = pgplot_transform
    pgplot_transform = (/-0.5, 1.0, 0.0, -0.5, 0.0, 1.0/)
    
    r = slice%range

    if(associated(slice%f2d)) then
        call pgslice2d(slice%f2d)
    else if(associated(slice%f3d)) then        
        call pgslice3d(slice%f3d)
    else
        if(master) write(*,'(a)') "pgplot title=", trim(slice%title)
        call abort("no array associated to f2d or f3d, cannot plot")
    end if

    pgplot_transform=backup_transform
    
  contains

    subroutine pgslice3d(f)
! gather a 3d distributed array and send the right slice from the
! global 3d array to the plotting routine
      implicit none
      real f(i1:,j1:,:), gf(im,jm,kb)

      if(slice%type.ne.6) then
          if(master) write(*,*) "what=", trim(slice%what), " title=", trim(slice%title)
          call abort(" pgplot_slice2d probably called with 3d array as what?")
      end if

      call gatherfield(f, gf, im,jm,kb)

      if(master)then
          select case(slice%slice)
          case("i")
              call plotf2d(gf(slice%i,:,:))
          case("j")
              call plotf2d(gf(:,slice%j,:))
          case("k")
              call plotf2d(gf(:,:,slice%k))
          end select
      end if
    end subroutine pgslice3d

    subroutine pgslice2d(f)
! gather and send 2d distributed array to the plotting routine      
      implicit none
      real f(i1:,j1:), gf(im,jm)

      if(slice%type.ne.7) then
          if(master) write(*,*) "what=", trim(slice%what), " title=", trim(slice%title)
          call abort(" pgplot_slice3d probably called with 2d array as what?")
      end if
      call gatherfield(f, gf, im,jm,1)
      if(master) call plotf2d(gf)
    end subroutine pgslice2d

    subroutine plotf2d(f)
! plot 2d array f.      
! f is a non distributed array.
      implicit none
      real f(:,:)
      real gfmax,gfmin,backuptr(6)
      character(100) title

      call pgpage()

      if(slice%fmax > -1e30)then
          gfmax = slice%fmax
      else
          gfmax = maxval(f(r(1):r(2),r(3):r(4)))
      end if
      if(slice%fmin < 1e30)then
          gfmin = slice%fmin
      else
          gfmin = minval(f(r(1):r(2),r(3):r(4)))
      end if

! reverse k axis for i or j slices, seems to work with the std transform
      select case(slice%slice)
      case("k")
          call pgwnad(r(1)-1._4, r(2)*1._4, r(3)-1._4, r(4)*1._4)
      case("i","j")
          call pgswin(r(1)-1._4, r(2)*1._4, r(4)*1._4, r(3)-1._4)
      end select

      call pcolor(f, fmin=gfmin, fmax=gfmax, range=r, colorbar=.true., batch=.true.)

! put on correct axis labels and a title containing slice index and from where we're called      
      select case(slice%slice)
      case("i")
          write(title,'(a,1x,"(",a,1x,"i=",i0,")")') trim(slice%title), trim(extratitle), slice%i
          call qaxes("j","k",trim(title),skipframe=.true., grid=.true.)
      case("j")
          write(title,'(a,1x,"(",a,1x,"j=",i0,")")') trim(slice%title), trim(extratitle), slice%j
          call qaxes("i","k",trim(title),skipframe=.true., grid=.true.)
      case("k")
          write(title,'(a,1x,"(",a,1x,"k=",i0,")")') trim(slice%title), trim(extratitle), slice%k
          call qaxes("i","j",trim(title),skipframe=.true., grid=.true.)
      end select
    end subroutine plotf2d

  end subroutine pgslice


  subroutine pgquick2d(pgobj)
!    
! plot ETA with UA,VA overlayed as vectors with properties from pgplot_core.f90.
! the setup2 line is
!   pgplot_2d <label> <i1> <i2> <j1> <j2>
! where <label> is in PGOBJ%TITLE, <i1> <i2> <j1> <j2> is in PGOBJ%RANGE.
! if JM is 1, plot a graph of ETA and U vs i instead.
!
    use state, only : eta, ua, va, im, jm, i1, i2, j1, j2, i_eq_im, j_eq_jm, gridbox, master, gatherfield
   
    implicit none
    type(pgplot_object) pgobj
    integer r(4),li,lj,dj,k,n
    real gfmax, gfmin, delta
    real, allocatable :: utmp(:,:), vtmp(:,:)
    real(4) trr(4)
    
    dj=1
    if(jm==1)dj=0

!! 2d, first find ETA
    call gatherfield(eta, pg_f, im,jm,1)

    r = pgobj%range

    if((pgobj%fmax > -1e30) .and. (pgobj%fmin < 1e30))then
        gfmax = pgobj%fmax
        gfmin = pgobj%fmin
        where(pgmask == 0) pg_f = -pgundef
    else
        if(master)then
            if (pgobj%fmin < 1e30) then
                gfmin = pgobj%fmin
            else
                where(pgmask == 0) pg_f = pgundef
                gfmin = minval(pg_f(r(1):r(2),r(3):r(4)))
            end if
            if (pgobj%fmax > -1e30) then
                gfmax = pgobj%fmax
            else
                where(pgmask == 0) pg_f = -pgundef
                gfmax = maxval(pg_f(r(1):r(2),r(3):r(4)))
            end if

            if(jm==1) then
                delta = abs(gfmax-gfmin)
                if(delta > 0) then
                    gfmin = gfmin - 0.1*delta
                    gfmax = gfmax + 0.1*delta
                else
                    gfmin = gfmin - 1e-5
                    gfmax = gfmax + 1e-5
                end if
            end if
        end if
    end if

!! then velocities
    allocate(utmp(i1:i2,j1:j2), vtmp(i1:i2,j1:j2))
    utmp = 0.0

    do lj=1,j_eq_jm
        do li=1,i_eq_im
            if(li<im .and. gridbox(li,lj)==1) then
                utmp(li,lj) = 0.5*(ua(li,lj)+ua(li+1,lj))
            else
                if(jm>1) utmp(li,lj) = -pgundef
            end if
            if(lj<jm .and. gridbox(li,lj)==1) then
                vtmp(li,lj) = 0.5*(va(li,lj)+va(li,lj+dj))
            else
                vtmp(li,lj) = -pgundef
            end if
        end do
    end do
    call gatherfield(utmp, pg_u, im,jm,1)
    if(jm > 1) call gatherfield(vtmp, pg_v, im,jm,1)

    deallocate(utmp, vtmp)

    if(jm == 1) then
        if(master) call q2dslice(pgobj, gfmin, gfmax)
        return
    end if

!! plot ETA with pcolor
    if(master)then

!! adjust viewport according to range, and enforce quadratic cells
        trr = pgobj%true_range
        call pgwnad(trr(1), trr(2), trr(3), trr(4))
        call pgpage()
        call pcolor(pg_f, fmax=gfmax, fmin=gfmin, range=r, colorbar=.true., batch=.true., mask=.true.)

        if(pgplot_2d_show_vsects) call overlay_vsects()

        call quiver(pg_u, pg_v, range=r, scale=pgplot_quiver_scale, blank=-pgundef, dens=pgplot_quiver_dens)
        call qaxes("x(m)","y(m)",grid=.true., title="ETA (m) and [UA,VA]"//trim(extratitle))
        if(pgplot_coastline) call draw_coastline()
    end if
  end subroutine pgquick2d


  subroutine pggeo(pgobj)
! plot geometry, i.e. a horisontal plot from given region with optional bathymetry, coastline.
! meant to be used as the foundation for e.g. particle trajectories
    use state, only : h,im,jm,master,depmax,depmin,gatherfield
    implicit none
    type(pgplot_object) pgobj
    integer r(4)
    real(4) trr(4)
    logical includegrid
    
    if(pgobj%i == 1) then
        call gatherfield(h, pg_f, im,jm,1)
    end if

    r = pgobj%range
    
    if(master)then
        trr = pgobj%true_range
        call pgwnad(trr(1), trr(2), trr(3), trr(4))
        call pgpage()
        
        if(pgobj%i == 1) then
            ! plot bath
            call pcolor(pg_f, fmax=depmax, fmin=depmin, range=r, colorbar=.true., batch=.true., mask=.true.)
        end if
        if(pgobj%j == 1) then
            call draw_coastline()
        end if        
        includegrid=.false.
        if(pgobj%k == 1) includegrid = .true.
        
        call qaxes("x(m)","y(m)",grid=includegrid, title=trim(pgobj%title)//" "//trim(extratitle))
    end if
  end subroutine pggeo


  subroutine pgatmos(pgobj)
!    
! plot PATM with WSPEEDU,WSPEEDV overlayed as vectors with properties from pgplot_core.f90.
! the setup2 line is
!   pgplot_atmos <label> <i1> <i2> <j1> <j2>
! where <label> is in PGOBJ%TITLE, <i1> <i2> <j1> <j2> is in PGOBJ%RANGE.
! if JM is 1, plot a graph of ETA and U vs i instead.
!
    use state, only : patm, im, jm, i1, i2, j1, j2, i_eq_im, j_eq_jm, gridbox, master, gatherfield
    use time_util, only : ndate
    use mod_bound, only : wspeedu, wspeedv
    implicit none
    type(pgplot_object) pgobj
    integer r(4),li,lj,dj,k,n
    real gfmax, gfmin, delta
    real, allocatable :: utmp(:,:), vtmp(:,:)
    real(4) trr(4)
    character(19) isodate, fart    
    real maxspeed
    
    dj=1
    if(jm==1)dj=0

!! 2d, first find PATM
    call gatherfield(patm, pg_f, im,jm,1)

    r = pgobj%range

    if((pgobj%fmax > -1e30) .and. (pgobj%fmin < 1e30))then
        gfmax = pgobj%fmax
        gfmin = pgobj%fmin
    else
        if(master)then
            if (pgobj%fmax > -1e30) then
                gfmax = pgobj%fmax
            else
                gfmax = maxval(pg_f(r(1):r(2),r(3):r(4)))
            end if
            if (pgobj%fmin < 1e30) then
                gfmin = pgobj%fmin
            else
                gfmin = minval(pg_f(r(1):r(2),r(3):r(4)))
                gfmin = gfmin - 0.005*abs(gfmax-gfmin) ! avoid black
            end if

            if(jm==1) then
                delta = abs(gfmax-gfmin)
                if(delta > 0) then
                    gfmin = gfmin - 0.1*delta
                    gfmax = gfmax + 0.1*delta
                else
                    gfmin = gfmin - 1e-5
                    gfmax = gfmax + 1e-5
                end if
            end if
        end if
    end if

!! then velocities
    allocate(utmp(i1:i2,j1:j2), vtmp(i1:i2,j1:j2))
    utmp = 0.0

    do lj=1,j_eq_jm
        do li=1,i_eq_im
            if(li < im) then
                utmp(li,lj) = 0.5*(wspeedu(li,lj)+wspeedv(li+1,lj))
            else
                if(jm > 1) utmp(li,lj) = -pgundef
            end if
            if(lj < jm) then
                vtmp(li,lj) = 0.5*(wspeedv(li,lj)+wspeedv(li,lj+dj))
            else
                vtmp(li,lj) = -pgundef
            end if
        end do
    end do
    call gatherfield(utmp, pg_u, im,jm,1)
    if(jm > 1) call gatherfield(vtmp, pg_v, im,jm,1)

    deallocate(utmp, vtmp)

    if(jm == 1) then
        if(master) call q2dslice(pgobj, gfmin, gfmax)
        return
    end if

!! plot ETA with pcolor
    if(master)then

!! adjust viewport according to range, and enforce quadratic cells
        trr = pgobj%true_range      
        call pgwnad(trr(1), trr(2), trr(3), trr(4))       
        call pgpage()
        call pcolor(pg_f, fmax=gfmax, fmin=gfmin, range=r, colorbar=.true., batch=.true.)
    
        if(pgplot_2d_show_vsects) call overlay_vsects()

        call quiver(pg_u, pg_v, range=r, scale=pgplot_quiver_scale, blank=-pgundef, dens=pgplot_quiver_dens, maxv=maxspeed)

        write(isodate,'(i4.4,"-",i2.2,"-",i2.2,"T",i2.2,":",i2.2,":",i2.2,5x,a,5x,a,5x,a)') ndate
        write(fart,'(a,f0.1,a)') "(max=",maxspeed,"m/s)"
        call qaxes("x(m)","y(m)",grid=.true., title="PATM (Pa), Wind speed "//trim(fart)//" "//isodate//" "//trim(extratitle))
        if(pgplot_coastline) call draw_coastline()
    end if
  end subroutine pgatmos

  

  subroutine overlay_vsects()
! draw a line along each defined vsect in the current (horizontal) plot
    use state, only : dx,dy
    implicit none
    integer k,n

    do k=1,pgplot_numplots
        if(pgplots(k)%type == 3)then
            n = pgplots(k)%vsect_np
            call pgsci(pgplot_text_color)
            call pgline(n, real(dx*(pgplots(k)%px(1:n)-1),4), real(dy*(pgplots(k)%py(1:n)-1),4))
        end if
    end do
  end subroutine overlay_vsects

  
  subroutine overlay_processors()
! draw the processor boundaries on top of the current hor. plot
    use state
    implicit none
    integer i,j,n
    real(kind=4) x(5), y(5)
    
    do j=1,numblj
        do i=1,numbli
            x = ((/bltab(i,j)%fi, bltab(i,j)%li, bltab(i,j)%li, bltab(i,j)%fi, bltab(i,j)%fi/)-0.5)*dx
            y = ((/bltab(i,j)%fj, bltab(i,j)%fj, bltab(i,j)%lj, bltab(i,j)%lj, bltab(i,j)%fj/)-0.5)*dy            
            call pgsci(pgplot_text_color)
            call pgline(5, x, y)
        end do
    end do
  end subroutine overlay_processors
  

  subroutine q2dslice(pgobj, gfmin, gfmax)
! runs on master, plot a trend plot of eta and ua if quick2d is called for jm=1
    use state, only : eta,ua,im
    implicit none
    type(pgplot_object) pgobj
    real gfmin, gfmax, xg(im)
    real umax,umin,delta
    real(4) vp(4)
    integer ii,is,ie

    is = pgobj%range(1) ; ie=pgobj%range(2)

    do ii=is,ie
        xg(ii) = ii-0.5
    end do

    call pgsch(pgplot_character_height)     
    call pgsci(pgplot_grid_color)

    vp = pgplot_viewport
    call pgsvp(vp(1),vp(2),vp(3),vp(4))
    call pgpage()

! eta range      
    call pgswin(pgobj%range(1)-1.0_4, 1.0_4*pgobj%range(2),real(gfmin,4),real(gfmax,4))      
    call pgslw(pgplot_grid_lw)                
    call pgbox('BCTSI',0.0,0,'BTSI',0.0,0)          ! left, bot and top
    call pgslw(pgplot_numbering_lw)
    call pgbox('N',0.0,0,'N',0.0,0)
    call trend(pg_f(is:ie,1), xg(is:ie))
    call pgslw(pgplot_label_lw)
    call pglab("", "ETA (m)", "")

! ua only on right, first find a usable y range
    umax = maxval(pg_u(is:ie,1))
    umin = minval(pg_u(is:ie,1))      
    delta = abs(umax-umin)      
    if(delta > 0) then
        umin = umin - 0.1*delta
        umax = umax + 0.1*delta
    else
        umin = umin - 1e-5
        umax = umax + 1e-5
    end if

    call pgswin(pgobj%range(1)-1.0_4, 1.0_4*pgobj%range(2), real(umin,4), real(umax,4))
    call pgsci(pgplot_grid_color)
    call pgslw(pgplot_grid_lw)                
    call pgbox('',0.0,0,'CTSI',0.0,0)
    call pgslw(pgplot_grid_lw)
    call pgbox('',0.0,0,'M',0.0,0)
    call trend(pg_u(is:ie,1), xg(is:ie), 3)
    call pgslw(pgplot_label_lw)
    call pgmtxt("R", 2.0_4, 0.5_4, 0.5_4, "UA (m/s)")      

! white labels  
    call pgsci(pgplot_text_color)
    call pgslw(pgplot_label_lw)
    call pglab("X grid pos", "", pgobj%title)      
  end subroutine q2dslice


  subroutine pghsect(pgobj)
!
! pgplot_hsect <label> <what> <depth>  <i1> <i2> <j1> <j2> <speed>
!
!  <what> is one of s, t, rho, wr, eta, h, p, u, v, w, speed, curl
!  <speed> .true. overlays velocity
!
    use state, only : im,jm,kb,u,v,s,t,rho,w,wr,p,eta,h,rhoaver,dx,dy
    use mod_mpi, only : i1,i2,j1,j2,im_mi_1,jm_mi_1,master,tag_ghost, update_ghost, gatherfield    
    use bomplot, only : hsect, daver
    use mod_concentration, only : co
    
    implicit none
    type(pgplot_object) pgobj
    real, allocatable :: utmp(:,:), vtmp(:,:)
    real, pointer :: speed(:,:,:)
    integer r(4),k,li,lj,dj,i,j
    real gfmin, gfmax, vel_plus, vel_minus, tmpdx
    real(4) trr(4)

    pg_f = 0.0
    r = pgobj%range

    dj=1
    if(jm==1)dj=0

    select case(trim(pgobj%what))
    case("speed")
        allocate(speed(i1:i2,j1:j2,kb))
        speed = 0.0
        do k=1,kb-1
            do lj=1,jm_mi_1
                do li=1,im_mi_1
                    speed(li,lj,k) = sqrt(0.25*(u(li,lj,k)+u(li+1,lj,k))**2 + &
                         0.25*(v(li,lj,k)+v(li,lj+dj,k))**2 + wr(li,lj,k)**2)
                end do
            end do
        end do
        call tag_ghost(speed)
        call update_ghost()
        call hsect("yo1", field=speed, position=pgobj%depth, range=pgobj%range, output="pgplot")
        deallocate(speed)
        
    case("curl")
! use hsect to find u,v in eta points, store in global arrays pg_u,pg_v.
        
! alternatively we could find u,v in u,v points, and then use central
! differencing on those, but also that approach requires
! e.g. averaging of u into v points, so what is best numerically is
! not clear.
        call hsect("yocurl", position=pgobj%depth, range=pgobj%range, output="pgplot")
        if(master)then
            pg_f = 0.0
            ! shrink range for curl due to stencil size, r is used to find max/min of field.
            r = (/max(2,r(1)), min(im-1,r(2)), max(2,r(3)), min(jm-1,r(4))/)

            do j=2,jm-1
                do i=2,im-1
                    ! central difference by default, one sided
                    ! difference next to land.  dV/dx-dU/dy
                    tmpdx = 2*dx
                    vel_plus = pg_v(i+1,j)
                    vel_minus = pg_v(i-1,j)                    
                    if(pgmask(i+1,j) == 0) then
                        tmpdx = dx
                        vel_plus = pg_v(i,j)
                    end if
                    if(pgmask(i-1,j) == 0) then
                        tmpdx = dx
                        vel_minus = pg_v(i,j)
                    end if
                    pg_f(i,j) = (vel_plus - vel_minus)/tmpdx
                end do
            end do
            do j=2, jm-1
                do i=2, im-1
                    tmpdx = 2*dy
                    vel_plus = pg_u(i,j+1)
                    vel_minus = pg_u(i,j-1)
                    if(pgmask(i,j+1) == 0) then
                        tmpdx = dy
                        vel_plus = pg_u(i,j)
                    end if
                    if(pgmask(i,j-1) == 0) then
                        tmpdx = dy
                        vel_minus = pg_u(i,j)
                    end if
                    pg_f(i,j) = pg_f(i,j) -  (vel_plus - vel_minus)/tmpdx
                    if(pgmask(i,j) == 0) pg_f(i,j) = -pgundef
                end do
            end do
        end if
    case("s")
        if(associated(s))call hsect("yo1", field=s, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("t")
        if(associated(t))call hsect("yo1", field=t, position=pgobj%depth, range=pgobj%range, output="pgplot")
        
    case("conc1")
        if(co(1)%allocated)call hsect("yo1", field=co(1)%conc, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("conc2")
        if(co(2)%allocated)call hsect("yo1", field=co(2)%conc, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("conc3")
        if(co(3)%allocated)call hsect("yo1", field=co(3)%conc, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("conc4")
        if(co(4)%allocated)call hsect("yo1", field=co(4)%conc, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("conc5")
        if(co(5)%allocated)call hsect("yo1", field=co(5)%conc, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("conc6")
        if(co(6)%allocated)call hsect("yo1", field=co(6)%conc, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("conc7")
        if(co(7)%allocated)call hsect("yo1", field=co(7)%conc, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("conc8")
        if(co(8)%allocated)call hsect("yo1", field=co(8)%conc, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("conc9")
        if(co(9)%allocated)call hsect("yo1", field=co(9)%conc, position=pgobj%depth, range=pgobj%range, output="pgplot")
        
! depth averaged 
    case("conc1_av")
        if(co(1)%allocated) call daver(co(1)%name, field=co(1)%conc, range=pgobj%range, output="pgplot")
    case("conc2_av")
        if(co(2)%allocated) call daver(co(2)%name, field=co(2)%conc, range=pgobj%range, output="pgplot")
    case("conc3_av")
        if(co(3)%allocated) call daver(co(3)%name, field=co(3)%conc, range=pgobj%range, output="pgplot")
    case("conc4_av")
        if(co(4)%allocated) call daver(co(4)%name, field=co(4)%conc, range=pgobj%range, output="pgplot")
    case("conc5_av")
        if(co(5)%allocated) call daver(co(5)%name, field=co(5)%conc, range=pgobj%range, output="pgplot")
        
    case("rho")
        call hsect("yo1", field=rho+1000+rhoaver, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("u")
        call hsect("yo1", field=u, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("v")
        call hsect("yo1", field=v, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("w")
        call hsect("yo1", field=w, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("wr")
        call hsect("yo1", field=wr, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("p")
        if(associated(p)) call hsect("yo1", field=p, position=pgobj%depth, range=pgobj%range, output="pgplot")
    case("eta")
        call fetch2d(eta)
    case("h")
        call fetch2d(h)
    case default
        write(*,*)'unknown "what" to plot for hsect, what=', trim(pgobj%what)
    end select

    if(master)then
        trr = pgobj%true_range
        call pgwnad(trr(1), trr(2), trr(3), trr(4))
        call pgpage

        if(pgobj%fmax > -1e30)then
            gfmax = pgobj%fmax
        else
            where(pg_f > 1e19) pg_f = -1e20
            gfmax = maxval(pg_f(r(1):r(2),r(3):r(4)))            
            where(pg_f == -1e20) pg_f = 2e19
        end if
        if(pgobj%fmin < 1e30)then
            gfmin = pgobj%fmin
        else
            where(pg_f == -pgundef) pg_f = 1e20
            gfmin = minval(pg_f(r(1):r(2),r(3):r(4)))
            where(pg_f == 1e20) pg_f = -pgundef
        end if
        call pcolor(pg_f, fmin=gfmin, fmax=gfmax, range=pgobj%range, colorbar=.true., batch=.true., mask=.true.)
    end if

    if(pgobj%quiv) then
! use hsect to fetch u and v for section into pg_u,pg_v
        call hsect("yo2", position=pgobj%depth, range=pgobj%range, output="pgplot")

        if(master)then
            r = pgobj%range
            allocate(utmp(im,jm), vtmp(im,jm))
            do lj=r(3),r(4)
                do li=r(1),r(2)
                    utmp(li,lj) = -pgundef
                    vtmp(li,lj) = -pgundef
                    if(pg_u(li,lj) .ne. -pgundef) utmp(li,lj) = pg_u(li,lj)
                    if(pg_v(li,lj) .ne. -pgundef) vtmp(li,lj) = pg_v(li,lj)
                end do
            end do
            call quiver(utmp, vtmp, range=r, scale=pgplot_quiver_scale, blank=-pgundef, dens=pgplot_quiver_dens)
            deallocate(utmp,vtmp)
        end if
    end if

    if(pgplot_coastline.and.master) call draw_coastline()
    
    if(master) call qaxes("x(m)","y(m)",grid=.true., title=trim(pgobj%title)//trim(extratitle))

  contains

    subroutine fetch2d(field)
      real field(i1:,j1:)
      real, allocatable :: tmpreal(:,:)

      allocate(tmpreal(im,jm))      
      call gatherfield(field, tmpreal, im,jm,1)
      pg_f = tmpreal
      deallocate(tmpreal)
      if(master) where(pgmask == 0) pg_f = -pgundef
    end subroutine fetch2d
  end subroutine pghsect


  subroutine pgvsect(pgobj)
!
! vsect:
!   pgplot_vsect label <what>  N x1 y1 ... xN yN   vsect
! <what> is one of speed, un, ut, wr, s, t, rho
!
    use state, only : im,jm,kb,u,v,s,t,rho,wr,w,rhoaver,fsm, debug,dx    
    use mod_mpi, only : i1,i2,j1,j2,im_mi_1,jm_mi_1,i_eq_im,j_eq_jm,master,tag_ghost,update_ghost    
    use bomplot, only : vsect
    
    implicit none
    type(pgplot_object) pgobj
    integer n,k,li,lj,dj
    real, pointer :: speed(:,:,:)
    real gfmin,gfmax
    real(4) zmin, zmax
    
    dj=1
    if(jm==1)dj=0

    n = pgobj%vsect_np
    
    if(master)call pgsvp(pgplot_viewport(1), pgplot_viewport(2), pgplot_viewport(3), pgplot_viewport(4))

    select case(pgobj%what)
    case("speed")
        allocate(speed(i1:i2,j1:j2,kb))
        do k=1,kb-1
            do lj=1,jm_mi_1
                do li=1,im_mi_1
                    speed(li,lj,k) = sqrt(0.25*(u(li,lj,k)+u(li+1,lj,k))**2 + &
                         0.25*(v(li,lj,k)+v(li,lj+dj,k))**2 + wr(li,lj,k)**2)
                end do
            end do
        end do
        call tag_ghost(speed)
        call update_ghost()
        call do_vsect("speed", speed)
        deallocate(speed)
    case("un")
        call vsect("UN", pgobj%px(1:n), pgobj%py(1:n), pgplot=.true.)
        call get_zminzmax()
        if(master)then
            call pgswin(0.0_4, &
                 real(pg_vx(pgplot_current_totsec,1)+0.5*dx,4), &
                 zmin, &
                 zmax)
            if(pgobj%fmax > -1e30)then
                gfmax = pgobj%fmax
            else
                gfmax = maxval(pg_usec(1:pgplot_current_totsec,1:kb-1))
            end if
            if(pgobj%fmin < 1e30)then
                gfmin = pgobj%fmin
            else
                gfmin = minval(pg_usec(1:pgplot_current_totsec,1:kb-1))
            end if
            call pgpage()
            call pg_vimag(pg_usec(1:pgplot_current_totsec,1:kb-1), &
                 pg_vx(1:pgplot_current_totsec,1:kb-1), &
                 pg_vz(1:pgplot_current_totsec,1:kb-1), &
                 fmin=gfmin, fmax=gfmax, colorbar=.true.)
        end if
    case("ut")
        call vsect("UT", pgobj%px(1:n), pgobj%py(1:n), pgplot=.true.)        
        call get_zminzmax()
        if(master)then
            call pgswin(0.0_4, &
                 real(pg_vx(pgplot_current_totsec,1)+0.5*dx,kind=4), &
                 zmin, &
                 zmax)
            if(pgobj%fmax > -1e30)then
                gfmax = pgobj%fmax
            else
                gfmax = maxval(pg_utsec(1:pgplot_current_totsec,1:kb-1))
            end if
            if(pgobj%fmin < 1e30)then
                gfmin = pgobj%fmin
            else
                gfmin = minval(pg_utsec(1:pgplot_current_totsec,1:kb-1))
            end if
            call pgpage()
            call pg_vimag(pg_utsec(1:pgplot_current_totsec,1:kb-1), &
                 pg_vx(1:pgplot_current_totsec,1:kb-1), &
                 pg_vz(1:pgplot_current_totsec,1:kb-1), &
                 fmin=gfmin, fmax=gfmax, colorbar=.true.)
        end if
    case("wr")
        call tag_ghost(wr)
        call update_ghost()
        call do_vsect("WR", wr)
    case("w")
        call tag_ghost(w)
        call update_ghost()
        call do_vsect("W", w)
    case("s")
        if(associated(s)) then
            call do_vsect("S", s)
        else
            call pgpage()
        end if
    case("t")
        if(associated(t)) then
            call do_vsect("T", t)
        else
            call pgpage()
        end if
    case("rho")
        allocate(speed(i1:i2,j1:j2,kb))                
        do k=1,kb-1
            do lj=1,j_eq_jm
                do li=1,i_eq_im
                    speed(li,lj,k) = (rho(li,lj,k) + rhoaver + 1000)*fsm(li,lj)
                end do
            end do
        end do
        call tag_ghost(speed)
        call update_ghost()
        call do_vsect("RHO", speed)
        deallocate(speed)
    case("conc1")
        call conc_vsect(1)
    case("conc2")
        call conc_vsect(2)
    case("conc3")
        call conc_vsect(3)
    case("conc4")
        call conc_vsect(4)
    case("conc5")
        call conc_vsect(5)
    case("conc6")
        call conc_vsect(6)
    case("conc7")
        call conc_vsect(7)
    case("conc8")
        call conc_vsect(8)
    case("conc9")
        call conc_vsect(9)
    case default
        write(*,*)'error pgplot.f90 batch_pgplot, unknown what=', pgobj%what
        if(master)call pgpage()
        return
    end select

    if(master) call qaxes("Sec. pos. (m)","Z [m]", grid=.true., title=trim(pgobj%title)//trim(extratitle))

  contains

    subroutine get_zminzmax()
      ! set the vrange vertical range.
      ! must be called after vsect
      
      if(pgobj%zmin < 1e30) then
          zmin = pgobj%zmin
      else
          zmin = 1.02*real(minval(pg_vz(1:pgplot_current_totsec,kb-1)),kind=4)
      end if
      if(pgobj%zmax > -1e30) then
          zmax = pgobj%zmax
      else
          zmax = 0 !real(maxval(pg_vz(1:pgplot_current_totsec,kb-1)),kind=4)
      end if
      if(master.and.debug > 3) write(*,'(1x,a,f0.7,1x,f0.7,a,2(1x,e14.7))') "pgvsect zmin,zmax=", zmin, zmax,&
           " pgobj%zmin and zmax=",pgobj%zmin, pgobj%zmax
    end subroutine get_zminzmax
    
    subroutine conc_vsect(id)
      use mod_concentration, only : co
      integer id
      if(co(id)%allocated) then
          call do_vsect(trim(co(id)%name), co(id)%conc)
      else
          call pgpage()
      end if
    end subroutine conc_vsect
    
    subroutine do_vsect(lab, field)
      implicit none
      character(*) lab
      real field(i1:,j1:,:), fieldmax, fieldmin

!! store field in  pg_vx, pg_vz, pg_ssect:

      call vsect(lab, pgobj%px(1:n), pgobj%py(1:n), field1=field, pgplot=.true.)      
      call get_zminzmax()
      
      if(master)then
          if(pgplot_current_totsec == 0) then
              write(*,*) "pgplot, do_vsect error: bomplot returned with empty section for what=",trim(pgobj%what)
              if(debug > 0) then
                  write(*,'(a,20(f0.1,1x))') "segpoints X(:)=",pgobj%px(1:n)
                  write(*,'(a,20(f0.1,1x))') "segpoints Y(:)=",pgobj%py(1:n)
              end if
              call pgpage()
              return
          end if
          
          call pgswin(0.0_4, real(pg_vx(pgplot_current_totsec,1)+0.5*dx,kind=4), &
               zmin, zmax)
          call pgpage()

          if(pgobj%fmax > -1e30)then
              fieldmax = pgobj%fmax
          else
              fieldmax = maxval(pg_ssec(1:pgplot_current_totsec,1:kb-1))
          end if
          if(pgobj%fmin < 1e30)then
              fieldmin = pgobj%fmin
          else
              fieldmin = minval(pg_ssec(1:pgplot_current_totsec,1:kb-1))
          end if

          call pg_vimag(pg_ssec(1:pgplot_current_totsec,1:kb-1), pg_vx(1:pgplot_current_totsec,1:kb-1), &
               pg_vz(1:pgplot_current_totsec,1:kb-1), fmin=fieldmin, fmax=fieldmax, &
               colorbar=.true., grid=pgplot_vsect_grid)      
      end if
    end subroutine do_vsect
  end subroutine pgvsect


  subroutine pgtimeseries(pgobj)
!    
! pgplot_timeseries <what>      label    i j k            time series from grid point
!   ekin,speed,u,v,wr,ua,va,s,t,rho
!    
    use state, only : istep, master, dt
    use timeseries, only : ts, timeseries_dump_start
    use qmpi, only : numproc
    implicit none
    type(pgplot_object) pgobj
    real maxy,miny
    real(4) scale
    integer :: firststep

    scale = pgplot_cachedt/dt
    
    if(istep < 1) then
        if(master) call pgpage()
        return
    end if

    if(master)then
        
        ! the time step index range to view
        if (pgplot_timeseries_window > 0) then
           firststep = max(0, istep-pgplot_timeseries_window)
        else
           firststep = 0
        end if
       
        if(pgobj%fmax > -1e30)then
            maxy = pgobj%fmax
        else
            maxy = maxval(pgobj%data1d(firststep:istep))
        end if
        if(pgobj%fmin < 1e30)then
            miny = pgobj%fmin
        else
            miny = minval(pgobj%data1d(firststep:istep))
        end if

        if(maxy-miny < 1e-20)then
            if(abs(maxy) > 1e-18)then
                maxy = maxy + 0.1*abs(maxy)
                miny = miny - 0.1*abs(miny)
            else
                maxy = 1e-20
                miny = -1e-20
            end if
        end if

        if(miny < -1e20 .or. maxy > 1e20) then
            call pgpage()
            write(*,*) " PGTIMESERIES: y out of range, cannot plot. miny=",miny,"maxy=",maxy
            return
        end if

        call pgsvp(pgplot_viewport(1), pgplot_viewport(2), pgplot_viewport(3), pgplot_viewport(4))

        call pgswin(real(timex(firststep),4), real(timex(istep),kind=4), real(miny,4), real(maxy,4))
        call pgpage()

        if(allocated(pgobj%cache1d)) then
            call trend(pgobj%cache1d(firststep:istep), timex(firststep:istep)*scale, col=3)
        end if
        call trend(pgobj%data1d(firststep:istep), timex(firststep:istep))
        call qaxes("Time (hours)", trim(pgobj%title), grid=.true.)
    end if
  end subroutine pgtimeseries


  subroutine pgvprof(pgobj)
!
! pgplot_vprof <what> <label> <x> [<y> <z>]
!   what u,v,wr,s,t,rho,speed,...
!
    use state, only : im,jm,kb,h,zz,z,u,v,wr,s,t,rho,am,km,kh,rhoaver
    use mod_mpi, only : master,myid, i1,j1,bltab, globalind2local, gi2block, gj2block, send, receive
    use mod_concentration, only : co
    implicit none
    type(pgplot_object) pgobj
    integer k,dj,li,lj,processor
    real speed1d(kb)
    logical inblock

    dj=1
    if(jm==1)dj=0

    if(master) call pgsvp(pgplot_viewport(1), pgplot_viewport(2), pgplot_viewport(3), pgplot_viewport(4))

    select case(pgobj%what)
    case("u")
        call do_vprof(u, pgobj%i, pgobj%j)
    case("v")
        call do_vprof(v, pgobj%i, pgobj%j)
    case("wr")
        call do_vprof(wr, pgobj%i, pgobj%j)
    case("s")
        if(associated(s)) call do_vprof(s, pgobj%i, pgobj%j)
    case("t")
        if(associated(t)) call do_vprof(t, pgobj%i, pgobj%j)
    case("conc1")
        if(co(1)%allocated) call do_vprof(co(1)%conc, pgobj%i, pgobj%j)
    case("conc2")
        if(co(2)%allocated) call do_vprof(co(2)%conc, pgobj%i, pgobj%j)
    case("conc3")
        if(co(3)%allocated) call do_vprof(co(3)%conc, pgobj%i, pgobj%j)
    case("conc4")
        if(co(4)%allocated) call do_vprof(co(4)%conc, pgobj%i, pgobj%j)
    case("conc5")
        if(co(5)%allocated) call do_vprof(co(5)%conc, pgobj%i, pgobj%j)
    case("conc6")
        if(co(6)%allocated) call do_vprof(co(6)%conc, pgobj%i, pgobj%j)
    case("conc7")
        if(co(7)%allocated) call do_vprof(co(7)%conc, pgobj%i, pgobj%j)
    case("conc8")
        if(co(8)%allocated) call do_vprof(co(8)%conc, pgobj%i, pgobj%j)
    case("conc9")
        if(co(9)%allocated) call do_vprof(co(9)%conc, pgobj%i, pgobj%j)
    case("rho")
        call do_vprof(rho+1000+rhoaver, pgobj%i, pgobj%j)
    case("speed")
        li=pgobj%i ; lj=pgobj%j
        call globalind2local(li,lj,inblock)
        if(inblock)then
            do k=1,kb-1
                speed1d(k) = sqrt(0.25*(u(li,lj,k)+u(li+1,lj,k))**2 + &
                     0.25*(v(li,lj,k)+v(li,lj+dj,k))**2 + wr(li,lj,k)**2)
            end do
            call send(speed1d, 0)
        end if
        if(master)then
            processor = bltab(gi2block(pgobj%i), gj2block(pgobj%j)) % prn
            call receive(speed1d, processor)
        end if
        call do_vprof(u,li,lj, field1d=speed1d)
    case("km")
        call do_vprof(km, pgobj%i, pgobj%j)
    case("kh")
        call do_vprof(kh, pgobj%i, pgobj%j)
    case("am")
        call do_vprof(am, pgobj%i, pgobj%j)
    end select

    if(master) call qaxes(trim(pgobj%title)//trim(extratitle), "Z [m]", grid=.true.)

  contains

    subroutine do_vprof(field,ci,cj,field1d)
!        
! draw trend of f(ci,cj,:) versus zz(:)*h(ci,cj), or field1d versus the same.
! (field1d is assumed to be on master already)
!        
      implicit none
      real field(i1:,j1:,:)
      real, optional :: field1d(:)
      integer ci,cj,li,lj,loc(3)      
      real(kind=4) xmin,xmax,ymin,ymax,delta
      real zcoo(kb),f(kb)
      logical inblock
      integer pointrank

      pointrank = bltab(gi2block(ci),gj2block(cj))%prn     

      if(myid == pointrank) then
          li=ci ; lj=cj
          call globalind2local(li,lj,inblock)
          zcoo = h(li,lj)*zz
          ymin = h(li,lj)*z(kb)
          if(.not.present(field1d))then
              f = field(li,lj,:)
              if(pointrank.ne.0) call send(f, 0, tag=1)
          end if
          if(pointrank.ne.0) call send(zcoo, 0, tag=2)
          if(pointrank.ne.0) call send(ymin, 0, tag=3)
      end if

      if(master) then
          ymax = 0.0
          if(.not.present(field1d))then
              if(pointrank.ne.0) call receive(f, pointrank, tag=1)
          end if
          if(pointrank.ne.0) call receive(zcoo, pointrank, tag=2)
          if(pointrank.ne.0) call receive(ymin, pointrank, tag=3)

          if(pgobj%fmax > -1e30)then
              xmax = pgobj%fmax
          else            
              if(present(field1d))then
                  xmax = maxval(field1d(1:kb-1))
              else
                  xmax = maxval(f(1:kb-1))
              end if
          end if
          if(pgobj%fmin < 1e30)then
              xmin = pgobj%fmin
          else            
              if(present(field1d))then
                  xmin = minval(field1d(1:kb-1))
              else
                  xmin = minval(f(1:kb-1))
              end if
          end if

          delta = abs(xmax - xmin)

          if(delta > 0) then
              xmin = xmin - 0.1*delta
              xmax = xmax + 0.1*delta
          else
              xmin = xmin - 1e-5
              xmax = xmax + 1e-5
          end if

          call pgswin(xmin,xmax,ymin,ymax)      
          call pgpage()
          if(present(field1d))then
              call trend(zcoo(1:kb-1), field1d(1:kb-1))
          else
              call trend(zcoo(1:kb-1), f(1:kb-1))
          end if
      end if
    end subroutine do_vprof
  end subroutine pgvprof


  subroutine get_pgplot_coastline(splitline)
    use qmpi, only : master, message
    use state, only : debug, dx, dy, xpole, ypole, ylong, minlon, maxlon
    use grid_util, only : sph2gr
    implicit none
    character(*) splitline(:)
    integer i,j, ierr
    real glon, glat
    
    read(splitline(2),*,err=101,end=101) pgplot_coastline_file
    read(splitline(3),*,err=102,end=102) pgplot_coastline_segments

    if(.not.master) return
    
    allocate(pgcl(4,pgplot_coastline_segments))
    
    open(1000010, file=pgplot_coastline_file, access="stream", form="unformatted", err=103)
    read(1000010,err=104,end=104) pgcl
    close(1000010)

    if(debug > 1.and. master) then
        write(*,*) "  got pgplot_coastline_file name=",trim(pgplot_coastline_file)
        write(*,*) "  expected number of segments=",pgplot_coastline_segments
        write(*,*) "  expected kind of reals in file=",kind(pgcl(1,1))
    end if
    
    j=0
    do i=1, pgplot_coastline_segments
        glon=pgcl(1,i)
        glat=pgcl(2,i)
        if(minlon<glon.and.glon<maxlon) then
             call sph2gr(glon, glat, (xpole+dx)/dx, (ypole+dy)/dy, dx*0.001, dy*0.001, ylong, pgcl(1,i), pgcl(2,i), ierr)
             glon=pgcl(3,i)
             glat=pgcl(4,i)
             call sph2gr(glon, glat, (xpole+dx)/dx, (ypole+dy)/dy, dx*0.001, dy*0.001, ylong, pgcl(3,i), pgcl(4,i), ierr)
             j=j+1
             pgcl(:,j) = (pgcl(:,i)-1)*dx
         end if        
    end do

    pgplot_coastline_segments = j
    
    if(debug>1) write(*,*) "  number of segments on the grid=",j
    
    return
101 call message("could not read file name from column 2",2)
102 call message("could not get the number of coastline segments from column 3",2)
103 call message("could not open unformatted stream access file with line segments",2)
104 call message("error while reading unformatted stream access file with line segments, check kind and number of segs.",2)
  end subroutine get_pgplot_coastline

  
  subroutine set_pgplot_colorscale(splitline,f_ind)
    use state, only : debug
    use qmpi, only : master, message
    implicit none
    character(*) splitline(:)
    character(4), optional, intent(in) :: f_ind
    integer plotind

    read(splitline(2),*) plotind
    if(plotind > pgplot_maxplots) call message("plotind > pgplot_maxplots",2)
! otherwise trust the user to keep track of the plot numbers...
    if (present(f_ind)) then
       select case(f_ind)
       case("fmin")
          read(splitline(3),*) pgplots(plotind)%fmin
          if(debug > 0 .and. master) then
             write(*,'(2x,a,i0,a,f0.3,",",f0.3)') "manually set the min range for the color scale on plot ",plotind,&
                  " to fmin=",pgplots(plotind)%fmin
          end if
       case("fmax")
          read(splitline(3),*) pgplots(plotind)%fmax
          if(debug > 0 .and. master) then
             write(*,'(2x,a,i0,a,f0.3,",",f0.3)') "manually set the max range for the color scale on plot ",plotind,&
                  " to fmax=",pgplots(plotind)%fmax
          end if
       case default
          call message("Do not recognize colorscale limiter")
       end select
    else
       read(splitline(3:4),*) pgplots(plotind)%fmin, pgplots(plotind)%fmax
       if(debug > 0 .and. master) then
          write(*,'(2x,a,i0,a,f0.3,",",f0.3)') "manually set the max and min range for the color scale on plot ",plotind,&
               " to fmax,fmin=",pgplots(plotind)%fmin, pgplots(plotind)%fmax
       end if
    end if
  end subroutine set_pgplot_colorscale

  
  subroutine pgvsect_zrange(splitline)
!  pgplot_vsect_zrange <label> <z1> <z2>
    use qmpi, only : message
    implicit none
    character(*) splitline(:)
    integer i
    logical found

    found = .false.
   
    do i=1, pgplot_numplots
        if(pgplots(i)%title == splitline(2))then
            found = .true.
            read(splitline(3:4),*, err=1, end=1) pgplots(i)%zmin, pgplots(i)%zmax
        end if
    end do
    if(.not.found) call message("the provided label to pgplot_vsect_zrange was not found",2)
    return
1   call message("could not read zmin,zmax from column 3 and 4",2)
  end subroutine pgvsect_zrange

  
  subroutine add_pgplot(type, splitline)
    use state
    use mod_bound
    use mod_concentration, only : co
    use control_parameters
    use grid_util
    use util, only : lower   
    use timeseries
    implicit none
    character(*) type, splitline(:)
!
! pgplot_2d              label    i1 i2 j1 j2      eta with 2d velocity on top
! pgplot_hsect label <what> depth  i1 i2 j1 j2 <speed>
!  <what> is one of s, t, rho, wr
!  <speed> .true. overlays velocity
! pgplot_vsect label <what>       N x1 y1 ... xN yN   vsect
!  <what> is one of speed, un, wr, s, t, rho
! pgplot_timeseries <what>      label    i j k            time series from grid point
!   speed, u,v,wr,ua,va,s,t,rho

    character(20) label
    integer i,j,k,gi,gj,n,r(4)
    logical inblock,logvar
    real reals(4)

    select case(type)

    case('pgplot_layout')
        read(splitline(2:3),*,err=101,end=101) pgplot_layout
        if(debug > 0.and.master) write(*,*)' value=', pgplot_layout

    case('pgplot_interval')
        read(splitline(2:4),*,err=110,end=110) pgplot_interval
        if(debug > 0.and.master) write(*,*)' value=', pgplot_interval

    case('pgplot_2d')
        pgplot_numplots = pgplot_numplots + 1
        pgplots(pgplot_numplots)%ind = pgplot_numplots
        pgplots(pgplot_numplots)%title = splitline(2)

        pgplots(pgplot_numplots)%type = 1
        read(splitline(3:6),*,err=103,end=103) pgplots(pgplot_numplots)%true_range
        call apply0000convention(pgplots(pgplot_numplots)%true_range)
        pgplots(pgplot_numplots)%range = true2globrange(pgplots(pgplot_numplots)%true_range)
        call check_range(pgplots(pgplot_numplots)%range)

        if(debug > 0 .and.master) write(*,*)' title=',trim(pgplots(pgplot_numplots)%title),' range=',pgplots(pgplot_numplots)%range

    case('pgplot_atmos')
        pgplot_numplots = pgplot_numplots + 1
        pgplots(pgplot_numplots)%ind = pgplot_numplots
        pgplots(pgplot_numplots)%title = splitline(2)

        pgplots(pgplot_numplots)%type = 8
        read(splitline(3:6),*,err=103,end=103) pgplots(pgplot_numplots)%true_range
        call apply0000convention(pgplots(pgplot_numplots)%true_range)
        pgplots(pgplot_numplots)%range = true2globrange(pgplots(pgplot_numplots)%true_range)
        call check_range(pgplots(pgplot_numplots)%range)

        if(debug > 0 .and.master) write(*,*)' title=',trim(pgplots(pgplot_numplots)%title),' range=',pgplots(pgplot_numplots)%range

    case('pgplot_geo')
        pgplot_numplots = pgplot_numplots + 1
        pgplots(pgplot_numplots)%ind = pgplot_numplots
        pgplots(pgplot_numplots)%title = splitline(2)

        pgplots(pgplot_numplots)%type = 9

        read(splitline(3),*,err=118,end=118) logvar        
        if(logvar) pgplots(pgplot_numplots)%i = 1
        
        read(splitline(4),*,err=118,end=118) logvar        
        if(logvar) pgplots(pgplot_numplots)%j = 1

        read(splitline(5),*,err=118,end=118) logvar        
        if(logvar) pgplots(pgplot_numplots)%k = 1
        
        read(splitline(6:9),*,err=117,end=117) pgplots(pgplot_numplots)%true_range
        call apply0000convention(pgplots(pgplot_numplots)%true_range)
        pgplots(pgplot_numplots)%range = true2globrange(pgplots(pgplot_numplots)%true_range)
        call check_range(pgplots(pgplot_numplots)%range)

        if(debug > 0 .and.master) write(*,*)' title=',trim(pgplots(pgplot_numplots)%title),' range=',pgplots(pgplot_numplots)%range

        
    case('pgplot_hsect')
        pgplot_numplots = pgplot_numplots + 1
        pgplots(pgplot_numplots)%ind = pgplot_numplots        
        pgplots(pgplot_numplots)%type = 2
        if(pgplot_numplots > pgplot_maxplots) &
             call message('too many pgplots. reduce, or increase pgplot_maxplots and recompile',2)
        pgplots(pgplot_numplots)%title = splitline(2)
        pgplots(pgplot_numplots)%what = lower(splitline(3))

        read(splitline(4),*,err=105,end=105) pgplots(pgplot_numplots)%depth
        read(splitline(5:8),*,err=112,end=112) pgplots(pgplot_numplots)%true_range

        call apply0000convention(pgplots(pgplot_numplots)%true_range)
        pgplots(pgplot_numplots)%range = true2globrange(pgplots(pgplot_numplots)%true_range)
        call check_range(pgplots(pgplot_numplots)%range)

        read(splitline(9),*,err=106,end=106) pgplots(pgplot_numplots)%quiv

        if(debug > 0.and.master) write(*,*)' what=',trim(pgplots(pgplot_numplots)%what),&
             ' depth=',pgplots(pgplot_numplots)%depth, ' overlay velocity=',pgplots(pgplot_numplots)%quiv

    case('pgplot_vsect')
        pgplot_numplots = pgplot_numplots + 1
        pgplots(pgplot_numplots)%ind = pgplot_numplots        
        pgplots(pgplot_numplots)%type = 3

        if(pgplot_numplots > pgplot_maxplots) &
             call message('too many pgplots. reduce, or increase pgplot_maxplots and recompile',2)
        pgplots(pgplot_numplots)%title = splitline(2)
        pgplots(pgplot_numplots)%what = lower(splitline(3))

        read(splitline(4),*,err=108,end=108) n
        pgplots(pgplot_numplots)%vsect_np = n
        read(splitline(5:),*,err=109,end=109) &
             (pgplots(pgplot_numplots)%px(i), pgplots(pgplot_numplots)%py(i),i=1,n)

        if(jm>1.and.any(pgplots(pgplot_numplots)%py(1:n) >= ly)) then
            if(master.and.debug>0) write(*,*) "  ly=",ly," y's=",pgplots(pgplot_numplots)%py(1:n)
            call message("1 or more invalid Y's for vsect, all Y must be < LY",2)
        end if
        if(any(pgplots(pgplot_numplots)%px(1:n) >= lx)) then
            if(master.and.debug>0) write(*,*) "  lx=",lx," x's=",pgplots(pgplot_numplots)%px(1:n)
            call message("1 or more invalid X's for vsect, all X must be < LX",2)
        end if
        if(jm>1.and.any(pgplots(pgplot_numplots)%py(1:n) <= 0.0)) &
             call message("1 or more invalid Y's for vsect, all Y must be > 0",2)
        if(any(pgplots(pgplot_numplots)%px(1:n) <= 0.0)) &
             call message("1 or more invalid X's for vsect, all X must be > 0",2)
        
        pgplots(pgplot_numplots)%px(1:n) = pgplots(pgplot_numplots)%px(1:n)/dx + 1.0        
        if(jm > 1) then
            pgplots(pgplot_numplots)%py(1:n) = pgplots(pgplot_numplots)%py(1:n)/dy + 1.0
        else
            pgplots(pgplot_numplots)%py(1:n) = 0.0
        end if

        if(debug>0.and.master) then
            write(*,*)' adding vsect title=',trim(pgplots(pgplot_numplots)%title),' what=',trim(pgplots(pgplot_numplots)%what)
            write(*,*)' with these grid coordinates:'
            do i=1, pgplots(pgplot_numplots)%vsect_np
                write(*,'(2x,a,i0,a,i0,a,f0.1,1x,f0.1)') "x(",i,"),y(",i,")=", &
                     pgplots(pgplot_numplots)%px(i), pgplots(pgplot_numplots)%py(i)
            end do
        end if

        if(jm == 1 .and. any(pgplots(pgplot_numplots)%py(1:n) .ne. 0.0)) then
            call message("ignoring vsect, when JM=1, segment y points must all be 0",1)
            pgplots(pgplot_numplots)%type = 0
        end if
        
    case('pgplot_vprof')
        pgplot_numplots = pgplot_numplots + 1
        pgplots(pgplot_numplots)%ind = pgplot_numplots        
        pgplots(pgplot_numplots)%type = 5

        if(pgplot_numplots > pgplot_maxplots) &
             call message('too many pgplots. reduce, or increase pgplot_maxplots and recompile',2)
        pgplots(pgplot_numplots)%title = splitline(2)
        pgplots(pgplot_numplots)%what = lower(splitline(3))

        read(splitline(4:5),*,err=111,end=111) reals(1:2)

        call xyz2gind(reals(1), reals(2), 0.0, pgplots(pgplot_numplots)%i, pgplots(pgplot_numplots)%j, k)
        
        if(debug>0.and.master) then
            write(*,*)' adding vertical profile under i,j=',pgplots(pgplot_numplots)%i, pgplots(pgplot_numplots)%j
            write(*,*)' for <what>=',trim(pgplots(pgplot_numplots)%what)
        end if

    case('pgplot_timeseries')
        pgplot_numplots = pgplot_numplots + 1
        pgplot_numts = pgplot_numts + 1
        pgplots(pgplot_numplots)%ind = pgplot_numplots        
        pgplots(pgplot_numplots)%type = 4

        if(pgplot_numplots > pgplot_maxplots) &
             call message('too many pgplots. reduce, or increase pgplot_maxplots and recompile',2)

        pgplots(pgplot_numplots)%title = splitline(2)
        pgplots(pgplot_numplots)%what = lower(splitline(3))

        if(debug>0.and.master) then
            write(*,*)' adding pgplot timeseries, title=',trim(pgplots(pgplot_numplots)%title),' what=',&
                 trim(pgplots(pgplot_numplots)%what)
        end if

        gi=-1 ; gj=-1

        if(master) allocate(pgplots(pgplot_numplots)%data1d(0:numstep))
        
! decide whether to get coordinates or not, 2d or 3d etc.
        
        select case(pgplots(pgplot_numplots)%what)
        case("ua ","va ","eta","ua_bound ","va_bound ","eta_bound")
         ! add all supported 2d fields here
            read(splitline(4:6),*,err=107,end=107) reals(1:2)
            call checkpos(reals(1:2))
            call xyz2gind(reals(1),reals(2),reals(3),gi,gj,k)
        case("ekin","wcourant","mass")
         ! nada. (don't try to read coordinates for scalars)
        case default
         ! by default, we assume a 3d variable (the unknown "what" is handled further below) 
            read(splitline(4:7),*,err=104,end=104) reals(1:3) ! gi,gj,k
            call checkpos(reals)
            call xyz2gind(reals(1),reals(2),reals(3),gi,gj,k)                
            pgplots(pgplot_numplots)%k = k
        end select

        i=gi
        j=gj
        call globalind2local(i,j,inblock,ghost=.false.)

        pgplots(pgplot_numplots)%i = i
        pgplots(pgplot_numplots)%j = j

        if(debug>0.and.master) then
            write(*,*)' horisontal position=',gi,gj
        end if
        
        label = " "

! this is where we actually add support for plotting a variable:        
        select case(pgplots(pgplot_numplots)%what)
        case("ekin")
            call add_ts_scalar("Kinetic energy",ekin)
        case("wcourant")
            call add_ts_scalar("Courant number for W",wcourant)
        case("mass")
            call add_ts_scalar("Mass",mass)
        case("speed")
            call add_ts_speed("Speed", gi,gj,k)
        case("u")
            write(label,'(a,i0,a,i0,a,i0,a)') "U(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, u(i,j,k))
        case("u_bound")
            write(label,'(a,i0,a,i0,a,i0,a)') "U_BOUND(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, u_bound(i,j,k))
        case("v")
            write(label,'(a,i0,a,i0,a,i0,a)') "V(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, v(i,j,k))
        case("v_bound")
            write(label,'(a,i0,a,i0,a,i0,a)') "V_BOUND(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, v_bound(i,j,k))
        case("w")
            write(label,'(a,i0,a,i0,a,i0,a)') "W(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, w(i,j,k))            
        case("wr")
            write(label,'(a,i0,a,i0,a,i0,a)') "WR(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, wr(i,j,k))
        case("s")
            if(advect_density) call message('advect_density is .true., not possible to plot S',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "S(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, s(i,j,k))
        case("s_bound")
            if(advect_density) call message('advect_density is .true., not possible to plot S',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "S_BOUND(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, s_bound(i,j,k))
        case("t")
            if(advect_density) call message('advect_density is .true., not possible to plot T',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "T(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, t(i,j,k))
        case("t_bound")
            if(advect_density) call message('advect_density is .true., not possible to plot T',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "T_BOUND(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, t_bound(i,j,k))
        case("rho")
            write(label,'(a,i0,a,i0,a,i0,a)') "RHO(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, rho(i,j,k))
            ts(num_ts)%density = 1.0
        case("eta")
            write(label,'(a,i0,a,i0,a)') "ETA(",gi,",",gj,")"
            call add_ts_scalar(label, eta(i,j))
        case("eta_bound")
            write(label,'(a,i0,a,i0,a)') "ETA_BOUND(",gi,",",gj,")"
            call add_ts_scalar(label, eta_bound(i,j))
        case("ua")
            write(label,'(a,i0,a,i0,a)') "UA(",gi,",",gj,")"
            call add_ts_scalar(label, ua(i,j))
        case("ua_bound")
            write(label,'(a,i0,a,i0,a)') "UA_BOUND(",gi,",",gj,")"
            call add_ts_scalar(label, ua_bound(i,j))
        case("va")
            write(label,'(a,i0,a,i0,a)') "VA(",gi,",",gj,")"
            call add_ts_scalar(label, va(i,j))
        case("va_bound")
            write(label,'(a,i0,a,i0,a)') "VA_BOUND(",gi,",",gj,")"
            call add_ts_scalar(label, va_bound(i,j))
        case("conc1")
            if(.not. co(1)%ALLOCATED) call message('concentration field not allocated, cannot plot',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "CONC1(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, co(1)%conc(i,j,k))
        case("conc2")
            if(.not. co(2)%ALLOCATED) call message('concentration field not allocated, cannot plot',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "CONC2(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, co(2)%conc(i,j,k))
        case("conc3")
            if(.not. co(3)%ALLOCATED) call message('concentration field not allocated, cannot plot',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "CONC3(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, co(3)%conc(i,j,k))
        case("conc4")
            if(.not. co(4)%ALLOCATED) call message('concentration field not allocated, cannot plot',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "CONC4(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, co(4)%conc(i,j,k))
        case("conc5")
            if(.not. co(5)%ALLOCATED) call message('concentration field not allocated, cannot plot',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "CONC5(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, co(5)%conc(i,j,k))
        case("conc6")
            if(.not. co(6)%ALLOCATED) call message('concentration field not allocated, cannot plot',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "CONC6(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, co(6)%conc(i,j,k))
        case("conc7")
            if(.not. co(7)%ALLOCATED) call message('concentration field not allocated, cannot plot',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "CONC7(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, co(7)%conc(i,j,k))
        case("conc8")
            if(.not. co(8)%ALLOCATED) call message('concentration field not allocated, cannot plot',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "CONC8(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, co(8)%conc(i,j,k))
        case("conc9")
            if(.not. co(9)%ALLOCATED) call message('concentration field not allocated, cannot plot',2)
            write(label,'(a,i0,a,i0,a,i0,a)') "CONC9(",gi,",",gj,",",k,")"
            call add_ts_scalar(label, co(9)%conc(i,j,k))
        case default
            if(master)write(*,*)'what=',trim(pgplots(pgplot_numplots)%what)
            call message("uknown 'what' for pgplot",2)
        end select

        pgplots(pgplot_numplots)%ts_i = num_ts
        
! scalars like ekin,mass etc are always kept on master. other
! scalars, e.g. pointers into arrays are not.
        if(gi>0 .and.gj>0)then
            ts(num_ts)%processor = bltab(gi2block(gi),gj2block(gj))%prn
            ts(num_ts)%inblock = .false.
            if(ts(num_ts)%processor == myid) ts(num_ts)%inblock = .true.
        end if

    case("pgplot_slice3d")
!   pgplot_slice3d <label> <what> <slice> <ind> <range>
        pgplot_numplots = pgplot_numplots + 1
        pgplots(pgplot_numplots)%ind = pgplot_numplots        
        pgplots(pgplot_numplots)%type = 6

        if(pgplot_numplots > pgplot_maxplots) &
             call message('too many pgplots. reduce, or increase pgplot_maxplots and recompile',2)

        pgplots(pgplot_numplots)%title = splitline(2)
        pgplots(pgplot_numplots)%what = lower(splitline(3))

        if(debug>0.and.master) then
            write(*,*)' adding pgplot 3d slice, title=',trim(pgplots(pgplot_numplots)%title),' what=',&
                 trim(pgplots(pgplot_numplots)%what)
        end if

        call assignslice(pgplots(pgplot_numplots))
        
        read(splitline(4),'(a)',err=113,end=113) pgplots(pgplot_numplots)%slice
        pgplots(pgplot_numplots)%slice = lower(pgplots(pgplot_numplots)%slice)
        
        read(splitline(5),*,err=115,end=115) i       
        read(splitline(6:9),*,err=114,end=114) r

        if(all(r==0)) then ! set range to full range
            select case(pgplots(pgplot_numplots)%slice)
            case("i")            
                r=(/1,jm,1,kb/)
            case("j")            
                r=(/1,im,1,kb/)
            case("k")            
                r=(/1,im,1,jm/)
            end select
        end if
        
        pgplots(pgplot_numplots)%range = r
        
        select case(pgplots(pgplot_numplots)%slice)
        case("i")
            pgplots(pgplot_numplots)%i = i
            if(pgplots(pgplot_numplots)%i <1 .or. pgplots(pgplot_numplots)%i>im) call message("invalid i slice",2)
            if(r(1)<1.or.r(2)>jm) call message("invalid j range for i slice",2)
            if(r(3)<1.or.r(4)>kb) call message("invalid k range for i slice",2)            
        case("j")
            pgplots(pgplot_numplots)%j = i
            if(r(1)<1.or.r(2)>im) call message("invalid i range for j slice",2)
            if(pgplots(pgplot_numplots)%j <1 .or. pgplots(pgplot_numplots)%j>jm) call message("invalid j slice",2)
            if(r(3)<1.or.r(4)>kb) call message("invalid k range for j slice",2)
        case("k")
            pgplots(pgplot_numplots)%k = i            
            if(r(1)<1.or.r(2)>im) call message("invalid i range for k slice",2)
            if(r(3)<1.or.r(4)>jm) call message("invalid j range for k slice",2)
            if(pgplots(pgplot_numplots)%k <1 .or. pgplots(pgplot_numplots)%k>kb) call message("invalid k slice",2)            
        case default
            call message("invalid slice. check config line column 4",2)
        end select
        
    case("pgplot_slice2d")
!   pgplot_slice2d <label> <what> <range>
        pgplot_numplots = pgplot_numplots + 1
        pgplots(pgplot_numplots)%ind = pgplot_numplots        
        pgplots(pgplot_numplots)%type = 7

        if(pgplot_numplots > pgplot_maxplots) &
             call message('too many pgplots. reduce, or increase pgplot_maxplots and recompile',2)

        pgplots(pgplot_numplots)%title = splitline(2)
        pgplots(pgplot_numplots)%what = lower(splitline(3))
        pgplots(pgplot_numplots)%slice = "k"
        
        if(debug>0.and.master) then
            write(*,*)' adding pgplot 2d slice, title=',trim(pgplots(pgplot_numplots)%title),' what=',&
                 trim(pgplots(pgplot_numplots)%what)
        end if

        call assignslice(pgplots(pgplot_numplots))
        
        read(splitline(4:7),*,err=116,end=116) r

        if(all(r==0)) then ! set range to full range
            r=(/1,im,1,jm/)
        end if
        
        pgplots(pgplot_numplots)%range = r        
        if(r(1)<1.or.r(2)>im) call message("invalid i range for k slice",2)
        if(r(3)<1.or.r(4)>jm) call message("invalid j range for k slice",2)
        
    case default
        call message('unknown pgplot type',2)
    end select

    return
101 call message('could not read 2 integers for layout from column 2 and 3',2)
102 call message('could not read integer for pgplot frequency from column 2',2)
103 call message('could not read 4 integers for pgplot_2d range from col 3 to 6',2)
104 call message('could not read 3 integers for timeseries grid position from column 4 to 7',2)
105 call message('could not read pgplot hsect depth from column 4',2)
106 call message('could not read pgplot hsect quiv keyword from column 9',2)
107 call message('could not read 2 integers for timeseries grid position from column 4 to 6',2)
108 call message('could not read integer for the number of vsect points',2)
109 call message('could not read N pairs of floats for vsect coordinates',2)
110 call message('could not read 3 integers for pgplot_interval: n1 n2 nd',2)
111 call message('could not read 2 integers i,j for horisontal grid box from columns 4,5',2)
112 call message('could not read 4 integers for pgplot_hsect range from col 5 to 8',2)
113 call message('could not get single char for slice from column 4',2)
114 call message('could not read 4 integers for pgplot_slice3d range from col 6 to 9',2)
115 call message('could not get integer slice position from column 5',2)
116 call message('could not read 4 integers for pgplot_slice2d range from col 4 to 7',2)   
117 call message('could not read 4 integers for pgplot_geo range from col 6 to 9',2)
118 call message('could not read a logical value from one of the columns 3-5',2)
  end subroutine add_pgplot
  
  
  subroutine assignslice(slice)
! given a pgplot_object SLICE this routine connects the 2d or 3d array
! pointer SLICE%F2D, SLICE%F3D, to the array specified by the string
! SLICE%WHAT.
    use pgplot_core
    use state
    use mod_bound
    use mod_modesplit
    
    implicit none
    type(pgplot_object) slice
    
    select case(slice%what)
    case("u")
        slice%f3d => u
    case("v")
        slice%f3d => v
    case("w")
        slice%f3d => w
    case("wr")
        slice%f3d => wr
    case("s")
        slice%f3d => s
    case("t")
        slice%f3d => t
    case("rho")
        slice%f3d => rho
    case("q2")
        slice%f3d => q2
    case("q2l")
        slice%f3d => q2l
    case("am")
        slice%f3d => am
    case("ah")
        slice%f3d => ah
    case("km")
        slice%f3d => km
    case("kh")
        slice%f3d => kh
    case("p")
        slice%f3d => p
    case("uadv")
        slice%f3d => uadv
    case("vadv")
        slice%f3d => vadv
    case("wadv")
        slice%f3d => wadv
    case("deltau")
        slice%f3d => deltau
    case("deltav")
        slice%f3d => deltav
    case("cor")
        slice%f2d => cor
    case("fsm")
        slice%f2d => fsm
    case("dum")
        slice%f2d => dum
    case("dvm")
        slice%f2d => dvm
    case("eta")
        slice%f2d => eta
    case("etap")
        slice%f2d => etap
    case("h")
        slice%f2d => h
    case("hu")
        slice%f2d => hu
    case("hv")
        slice%f2d => hv
    case("d")
        slice%f2d => d
    case("du")
        slice%f2d => du
    case("dv")
        slice%f2d => dv
    case("wusurf")
        slice%f2d => wusurf
    case("wvsurf")
        slice%f2d => wvsurf
    case("ssurf")
        slice%f2d => ssurf
    case("tsurf")
        slice%f2d => tsurf
    case("rhosurf")
        slice%f2d => rhosurf
    case("wubot")
        slice%f2d => wubot
    case("wvbot")
        slice%f2d => wvbot
    case("cbc")
        slice%f2d => cbc
    case("patm")
        slice%f2d => patm
    case("am2d")
        slice%f2d => am2d
    case("ua")
        slice%f2d => ua
    case("va")
        slice%f2d => va
    case("husqr")
        slice%f2d => husqr
    case("hvsqr")
        slice%f2d => hvsqr
    case("sponge_cm2d")
        slice%f2d => sponge_cm2d
    case("sponge_cm")
        slice%f2d => sponge_cm
    case("sponge_km")
        slice%f2d => sponge_km
    case("sponge_kh")
        slice%f2d => sponge_kh
    case("sponge_kh_c1")
        slice%f2d => sponge_kh_c1
        
! from mod_bound        
    case("u_bound")
        slice%f3d => u_bound
    case("v_bound")
        slice%f3d => v_bound
    case("w_bound")
        slice%f3d => w_bound
    case("s_bound")
        slice%f3d => s_bound
    case("t_bound")
        slice%f3d => t_bound
    case("q2_bound")
        slice%f3d => q2_bound
    case("q2l_bound")
        slice%f3d => q2l_bound
    case("rho_bound")
        slice%f3d => rho_bound
    case("s_initial")
        slice%f3d => s_initial
    case("t_initial")
        slice%f3d => t_initial
    case("rho_initial")
        slice%f3d => rho_initial

    case("alfa_u")
        slice%f2d => alfa_u
    case("alfa_v")
        slice%f2d => alfa_v
    case("alfa_eta")
        slice%f2d => alfa_eta
    case("alfae_u")
        slice%f2d => alfae_u
    case("alfae_v")
        slice%f2d => alfae_v
    case("alfae_eta")
        slice%f2d => alfae_eta
    case("ua_bound")
        slice%f2d => ua_bound
    case("va_bound")
        slice%f2d => va_bound
    case("eta_bound")
        slice%f2d => eta_bound
    case("etabfrs")
        slice%f2d => etabfrs
    case("etatfrs")
        slice%f2d => etatfrs
    case("ua_tide")
        slice%f2d => ua_tide
    case("va_tide")
        slice%f2d => va_tide
    case("eta_tide")
        slice%f2d => eta_tide
    case("wspeedu")
        slice%f2d => wspeedu
    case("wspeedv")
        slice%f2d => wspeedv
    case("etatamp")
        slice%f3d => etatamp
    case("etatpha")
        slice%f3d => etatpha
    case("uatamp")
        slice%f3d => uatamp
    case("uatpha")
        slice%f3d => uatpha
    case("vatamp")
        slice%f3d => vatamp
    case("vatpha")
        slice%f3d => vatpha

        
! from mod_modesplit
    case("uv")
        slice%f3d => uv
    case("vu")
        slice%f3d => vu
    case("ubclin")
        slice%f3d => ubclin
    case("vbclin")
        slice%f3d => vbclin
    case("c3u")
        slice%f3d => c3u
    case("c3v")
        slice%f3d => c3v
    case("ucorr")
        slice%f3d => ucorr
    case("vcorr")
        slice%f3d => vcorr
    case("wcorr")
        slice%f3d => wcorr
    case("ax")
        slice%f2d => ax
    case("ay")
        slice%f2d => ay
    case("uv2d")
        slice%f2d => uv2d
    case("vu2d")
        slice%f2d => vu2d
    case("uasum")
        slice%f2d => uasum
    case("vasum")
        slice%f2d => vasum
    case("coru")
        slice%f2d => coru
    case("corv")
        slice%f2d => corv
    case("c2u")
        slice%f2d => c2u
    case("c2v")
        slice%f2d => c2v
    case("uad")
        slice%f2d => uad
    case("vad")
        slice%f2d => vad
    case("uastar")
        slice%f2d => uastar
    case("vastar")
        slice%f2d => vastar
    case("etastar")
        slice%f2d => etastar
    case("uaf")
        slice%f2d => uaf
    case("uab")
        slice%f2d => uab
    case("vaf")
        slice%f2d => vaf
    case("vab")
        slice%f2d => vab
    case("etab")
        slice%f2d => etab
    case("etaf")
        slice%f2d => etaf

! from mod_concentration
    case("conc1")
        slice%f3d => co(1)%conc
    case("conc2")
        slice%f3d => co(2)%conc
    case("conc3")
        slice%f3d => co(3)%conc
    case("conc4")
        slice%f3d => co(4)%conc
    case("conc5")
        slice%f3d => co(5)%conc
    case("conc6")
        slice%f3d => co(6)%conc
    case("conc7")
        slice%f3d => co(7)%conc
    case("conc8")
        slice%f3d => co(8)%conc
    case("conc9")
        slice%f3d => co(9)%conc
        
    case default
        call message("not yet taught how to add this 'what'",2)
    end select    
  end subroutine assignslice


  subroutine pghsect_particles(pg)
    use mod_particle
    implicit none
    type(pgplot_object) pg
    integer :: i, j, k, plot
    integer :: use_color

    integer :: color_scale, color_offset
    real :: col_r, col_g, col_b, col_r0, col_g0, col_b0
    real :: gray

    if (.not. part_activate) return
    
    color_scale = 8
    color_offset = 6
    col_r0 = 1.0
    col_g0 = 0.0
    col_b0 = 1.0

    do i=1,color_scale
       col_r = col_r0*real(i-1)/real(color_scale-1)
       col_g = col_g0*real(i-1)/real(color_scale-1)
       col_b = col_b0*real(i-1)/real(color_scale-1)
       if(master) call pgscr(i+color_offset, col_r, col_g, col_b)
!       gray = real(i-1)/real(color_scale-1)
!       call pgscr(i+color_offset, gray, gray, gray)
    end do
    do i=1,(pgpart_ntime_pts-1)
       pgpart_color_array(i) = color_offset + 1 +  &
            nint(real((i-1)*(color_scale-1))/real(pgpart_ntime_pts-2))
    end do

    do plot = 1,pgpart_num_plots
       if (pgparts(plot)%ind .eq. pg%ind) then
          select case (pgparts(plot)%type)
          case(1)
             j = 0
             do i=1,pgpart_history_count
                j = j+1
                if (pgpart_history(pgpart_ntime_pts,1,i) .ne. -1) then
                   pgparts(plot)%x1(j) = pgpart_history(pgpart_ntime_pts,1,i)
                   pgparts(plot)%y1(j) = pgpart_history(pgpart_ntime_pts,2,i)
                else
                   j = j-1
                end if
             end do
             call scatter(pgparts(plot)%x1(1:j),pgparts(plot)%y1(1:j),  &
                  style=pgparts(plot)%style,color=pgparts(plot)%color,  &
                  line_width=pgparts(plot)%line_width)
          case(2)
             j = 0
             do i=1,pgpart_history_count
                j = j+1
                if (pgpart_history(pgpart_ntime_pts,1,i) .ne. -1) then
                   pgparts(plot)%x1(j) = max(pgpart_history(2,1,i),0.)
                   pgparts(plot)%y1(j) = max(pgpart_history(2,2,i),0.)
                   pgparts(plot)%x2(j) = max(pgpart_history(1,1,i),0.)
                   pgparts(plot)%y2(j) = max(pgpart_history(1,2,i),0.)
                else
                   j = j-1
                end if
             end do
             if (pgparts(plot)%color .eq. -1) then
                use_color = pgpart_color_array(1)
             else
                use_color = pgparts(plot)%color
             end if
             call pldj(pgparts(plot)%x1(1:j),pgparts(plot)%y1(1:j),  &
                  pgparts(plot)%x2(1:j),pgparts(plot)%y2(1:j),  &
                  style=pgparts(plot)%style,color=use_color,  &
                  line_width=pgparts(plot)%line_width)
             
             do k=3,pgpart_ntime_pts
                   pgparts(plot)%x2 = pgparts(plot)%x1
                   pgparts(plot)%y2 = pgparts(plot)%y1
                   j = 0
                   do i=1,pgpart_history_count
                      j = j+1
                      if (pgpart_history(pgpart_ntime_pts,1,i) .ne. -1) then
                         pgparts(plot)%x1(j) = max(pgpart_history(k,1,i),0.)
                         pgparts(plot)%y1(j) = max(pgpart_history(k,2,i),0.)
                      else
                         j = j-1
                      end if
                   end do
                   if (pgparts(plot)%color .eq. -1) then
                      use_color = pgpart_color_array(k-1)
                   else
                      use_color = pgparts(plot)%color
                   end if

                   call pldj(pgparts(plot)%x1(1:j),pgparts(plot)%y1(1:j),  &
                        pgparts(plot)%x2(1:j),pgparts(plot)%y2(1:j),  &
                        style=pgparts(plot)%style,color=use_color,  &
                        line_width=pgparts(plot)%line_width)
                end do
          case(3)
             j = 0
             do i=1,pgpart_capture_count
                j = j+1
                if (pgpart_capture(1,i) .ne. -1) then
                   pgparts(plot)%x1(j) = pgpart_capture(2,i)
                   pgparts(plot)%y1(j) = pgpart_capture(3,i)
                else
                   j = j-1
                end if
             end do
             call scatter(pgparts(plot)%x1(1:j),pgparts(plot)%y1(1:j),  &
                  style=pgparts(plot)%style,color=pgparts(plot)%color,  &
                  line_width=pgparts(plot)%line_width)
          case default
             call abort('ERROR: Not a valid pgparts(plot)%type value!')
          end select
       end if
    end do     
  end subroutine pghsect_particles


  subroutine pgvsect_particles(pg)
    use mod_particle
    use state, only : lx
    implicit none
    type(pgplot_object) pg
    integer :: i, j, k, plot
    integer :: use_color

    integer :: color_scale, color_offset
    real :: col_r, col_g, col_b, col_r0, col_g0, col_b0
    real :: gray, icosa, delx,dely,x_part

    if (.not. part_activate) return

    color_scale = 8
    color_offset = 6
    col_r0 = 1.0
    col_g0 = 0.0
    col_b0 = 1.0

    do i=1,color_scale
        col_r = col_r0*real(i-1)/real(color_scale-1)
        col_g = col_g0*real(i-1)/real(color_scale-1)
        col_b = col_b0*real(i-1)/real(color_scale-1)
        if(master) call pgscr(i+color_offset, col_r, col_g, col_b)
!       gray = real(i-1)/real(color_scale-1)
!       call pgscr(i+color_offset, gray, gray, gray)
    end do
    do i=1,(pgpart_ntime_pts-1)
        pgpart_color_array(i) = color_offset + 1 +  &
             nint(real((i-1)*(color_scale-1))/real(pgpart_ntime_pts-2))
    end do

    do plot = 1,pgpart_num_plots
        if (pgparts(plot)%ind .eq. pg%ind) then
            select case (pgparts(plot)%type)
            case(4)

! find the inverse of cos(section angle wrt x axis)
                delx = pg%px(2) - pg%px(1)
                dely = pg%py(2) - pg%py(1)
                icosa = sqrt(delx**2 + dely**2)/delx   

                j = 0
                do i=1,pgpart_history_count
                    x_part = pgpart_history(pgpart_ntime_pts,1,i)
                    if(x_part/dx+1 < min(pg%px(1),pg%px(2)) .or. x_part/dx+1 > max(pg%px(1),pg%px(2))) cycle

                    j = j+1
                    if (pgpart_history(pgpart_ntime_pts,1,i) .ne. -1) then
! the along sect coordinate. for now we only allow
! one sect for such overlays and project the x or
! y coordinate of the particles onto that
                        pgparts(plot)%x1(j) = (x_part - (pg%px(1)-1)*dx)*icosa
!!  print *,"HMx",pgparts(plot)%x1(j)                
                        pgparts(plot)%y1(j) = pgpart_history(pgpart_ntime_pts,3,i)
                    else
                        j = j-1
                    end if
                end do

                call scatter(pgparts(plot)%x1(1:j),pgparts(plot)%y1(1:j),  &
                     style=pgparts(plot)%style,color=pgparts(plot)%color,  &
                     line_width=pgparts(plot)%line_width)
            case(5)
! like 4, but project the y coordinate onto the section (icos is now isin etc)
                delx = pg%px(2) - pg%px(1)
                dely = pg%py(2) - pg%py(1)
                icosa = sqrt(delx**2 + dely**2)/dely   

                j = 0
                do i=1,pgpart_history_count
                    x_part = pgpart_history(pgpart_ntime_pts,2,i)
                    if(x_part/dx+1 < min(pg%py(1),pg%py(2)) .or. x_part/dx+1 > max(pg%py(1),pg%py(2))) cycle

                    j = j+1
                    if (pgpart_history(pgpart_ntime_pts,2,i) .ne. -1) then
                        pgparts(plot)%x1(j) = (x_part - (pg%py(1)-1)*dx)*icosa
                        pgparts(plot)%y1(j) = pgpart_history(pgpart_ntime_pts,3,i)
                    else
                        j = j-1
                    end if
                end do

                call scatter(pgparts(plot)%x1(1:j),pgparts(plot)%y1(1:j),  &
                     style=pgparts(plot)%style,color=pgparts(plot)%color,  &
                     line_width=pgparts(plot)%line_width)

            case(6)
                delx = pg%px(2) - pg%px(1)
                dely = pg%py(2) - pg%py(1)
                icosa = sqrt(delx**2 + dely**2)/delx
                
                j = 0
                do i=1,pgpart_history_count
                    x_part = pgpart_history(2,1,i)
                    if(x_part/dx+1 < min(pg%px(1),pg%px(2)) .or. x_part/dx+1 > max(pg%px(1),pg%px(2))) cycle
                    
                    j = j+1
                    if (pgpart_history(pgpart_ntime_pts,1,i) .ne. -1) then
                        pgparts(plot)%x1(j) = max((x_part - (pg%px(1)-1)*dx)*icosa,0.)!max(pgpart_history(2,1,i),0.)
                        pgparts(plot)%y1(j) = pgpart_history(2,3,i)

                        x_part = pgpart_history(1,1,i)
                        
                        pgparts(plot)%x2(j) = max((x_part - (pg%px(1)-1)*dx)*icosa,0.) !max(pgpart_history(1,1,i),0.)
                        pgparts(plot)%y2(j) = pgpart_history(1,3,i)

                        ! if a line segment goes across the domain in the x direction, ignore (periodic bnd artifact)
                        if((pgparts(plot)%x1(j) < 0.25*lx .and. pgparts(plot)%x2(j) > 0.75*lx) .or. &
                           (pgparts(plot)%x1(j) > 0.75*lx .and. pgparts(plot)%x2(j) < 0.25*lx)) then
                            j=j-1
                        end if
                    else
                        j = j-1
                    end if
                end do
                if (pgparts(plot)%color .eq. -1) then
                    use_color = pgpart_color_array(1)
                else
                    use_color = pgparts(plot)%color
                end if
                call pldj(pgparts(plot)%x1(1:j),pgparts(plot)%y1(1:j),  &
                     pgparts(plot)%x2(1:j),pgparts(plot)%y2(1:j),  &
                     style=pgparts(plot)%style,color=use_color,  &
                     line_width=pgparts(plot)%line_width)
                
                do k=3,pgpart_ntime_pts
                    pgparts(plot)%x2 = pgparts(plot)%x1
                    pgparts(plot)%y2 = pgparts(plot)%y1
                    
                    j = 0
                    do i=1,pgpart_history_count
                        x_part = pgpart_history(k,1,i)
                        if(x_part/dx+1 < min(pg%px(1),pg%px(2)) .or. x_part/dx+1 > max(pg%px(1),pg%px(2))) cycle
                        
                        j = j+1
                        if (pgpart_history(pgpart_ntime_pts,1,i) .ne. -1) then
                            pgparts(plot)%x1(j) = max((x_part - (pg%px(1)-1)*dx)*icosa, 0.)  !max(pgpart_history(k,1,i),0.)
                            pgparts(plot)%y1(j) = pgpart_history(k,3,i)

                            ! if a line segment goes across the domain in the x direction, ignore (periodic bnd artifact)
                            if((pgparts(plot)%x1(j) < 0.25*lx .and. pgparts(plot)%x2(j) > 0.75*lx) .or. &
                               (pgparts(plot)%x1(j) > 0.75*lx .and. pgparts(plot)%x2(j) < 0.25*lx)) then
                                j=j-1
                            end if
                        else
                            j = j-1
                        end if
                    end do
                    
                    if (pgparts(plot)%color .eq. -1) then
                        use_color = pgpart_color_array(k-1)
                    else
                        use_color = pgparts(plot)%color
                    end if
                    
                    if(j >= 1)then
                        call pldj(pgparts(plot)%x1(1:j),pgparts(plot)%y1(1:j),  &
                             pgparts(plot)%x2(1:j),pgparts(plot)%y2(1:j),  &
                             style=pgparts(plot)%style,color=use_color,  &
                             line_width=pgparts(plot)%line_width)
                    end if
                end do
            case default
                call message('pgvsect invalid pgparts(plot)%type',2,ival=(/pgparts(plot)%type/))
            end select
        end if
    end do
  end subroutine pgvsect_particles

  
end module pgplot
