module pgplot_core
!
! stand alone wrapper on top of pgplot which offers a matlab like
! interface to some common plotting operations.
!
! SUBROUTINE INIT_PGPLOT_CORE([VIEWPORT, WIDTH, ASPECT, TRANSFORM, DEVICE, LAYOUT, PALETTE])
!
!  some devices (see drivers.list in pgplot directory for all)
!   /NULL      (Null device, no output)
!   /GIF       (Graphics Interchange Format file, landscape orientation)
!   /PNG       (Portable network graphics)
!   /PS        (PostScript file, landscape orientation)
!   /VPS       (PostScript file, portrait orientation)
!   /CPS       (Colour PostScript file, landscape orientation)
!   /VCPS      (Colour PostScript file, portrait orientation)
!  
! SUBROUTINE PCOLOR(F [,FMIN, FMAX, RANGE, PALETTE, COLORBAR, AXES])
!   color map of 2d array f  
! SUBROUTINE QUIVER(A, B [,RANGE, SCALE, BLANK, DENS, HEADSIZE, AXES])
!   vector plot of vector field with components (a,b)
! SUBROUTINE CONTOUR(F [,FMIN, FMAX, FILLED, LABELS, RANGE, PALETTE, COLORBAR, AXES])
!   contour lines for 2d array f  
! SUBROUTINE PLOT(F, X, [FMAX, FMIN, XMIN, XMAX, XLABEL, YLABEL, TITLE])
!   trend plot for f(x), where f is 1d array, x positions, with options for plot title,axis labels etc.
! SUBROUTINE TREND(F, X [, COL, AXES])
!   same as "plot" but just the line.  
! SUBROUTINE SCATTER(X, Y [, STYLE, COLOR, LINE_WIDTH])
!   plot scatter of points in vectors x,y with given style and color. STYLE is as follows:
!                   -1, -2  : a single dot (diameter = current
!                             line width).
!                   -3..-31 : a regular polygon with ABS(SYMBOL)
!                             edges (style set by current fill style).
!                   0..31   : standard marker symbols.
!                   32..127 : ASCII characters (in current font).
!                             e.g. to use letter F as a marker, let
!                             STYLE = ICHAR('F'). 
!                   > 127  :  a Hershey symbol number.  
! SUBROUTINE PLDJ(X1, Y1, X2, Y2 [, STYLE, COLOR, COLOR_ARRAY, LINE_WIDTH])
!   plot disjoint lines between points in vectors x1,y1 to x2,y2 with given style  
! SUBROUTINE QAXES([XLABEL, YLABEL, TITLE])
!   add label to plot  
! SUBROUTINE SETVP()
!   set at default viewport
! SUBROUTINE PALETT(TYPE)
!   pick one of the predefined palettes.

  
! data with value pgundef, which should be positive, default real, should be ignored in plots
  real, parameter :: pgundef = real(30000)
  
! plot from step=pgplot_interval(1), until pgplot_interval(2), with
! increment pgplot_interval(3)
  integer :: pgplot_interval(3)=(/0,200000,1/)

  integer :: pgplot_cache_results=0, pgplot_cache_interval=100
  real :: pgplot_cachedt=0.0
  
! if 0, dont call pgplot in dump_state. if 1 dump state in 3d routines, if 2 also in 2d routines.  
  integer :: dump_state_pgplot=0
  
  integer, parameter, private :: sp=4
  
! default rotation angle
  real(sp) :: pgplot_theta = 0.0
  
! transform: f(i,j) are located as   (tr(3) and tr(5) normally 0)
!  x = tr(1) + tr(2)*i + tr(3)*j
!  y = tr(4) + tr(5)*i + tr(6)*j
  
  real(sp) :: pgplot_transform(6)
  logical :: pgplot_core_do_init=.true., pgplot_debug=.false., pgplot_wait_after_plot=.false.
  logical :: pgplot_2d_show_vsects=.false., pgplot_coastline=.false., pgplot_vsect_grid=.false.
  logical :: pgplot_allocated=.false.
  character(30) :: pgplot_device = "?"
  
  character(200) :: pgplot_coastline_file=" "
  integer :: pgplot_coastline_segments=0
  real, allocatable :: pgcl(:,:)
  
  real(sp) :: pgplot_character_height = 1.0, pgplot_aspect = 1.0, pgplot_width = 5.0, &
       pgplot_viewport(4) = (/0.12,0.88,0.12,0.88/), pgplot_range(4) = (/0.0, 1.0, 0.0, 1.0/)

  logical :: pgplot_width_set=.false., pgplot_aspect_set=.false.
  integer :: pgplot_title_time=1
  integer :: pgplot_layout(2) = (/1,1/), pgplot_palette = 2, pgplot_contour_levels=10
! plot an arrow avery dens'th cell, default 2:
  integer :: pgplot_quiver_dens=2

! colors
  integer :: pgplot_text_color=1, pgplot_background=0, pgplot_grid_color=2, pgplot_curve_color=36

! line thickness
  integer :: pgplot_numbering_lw=1, pgplot_label_lw=1, pgplot_grid_lw=4, pgplot_curve_lw=5, pgplot_quiver_lw=2

  real(4) :: pgplot_headsize=0.3
  real :: pgplot_quiver_scale=2.0
  integer :: pgplot_quiver_color=1
  
! support for bomplot vsect plots using the normal vsect routine and
! the "pgplot=.true."optional argument.  these arrays should be
! allocated by calling subroutine allocate_pgvsect()  
  integer :: pgplot_maxsec=-1, pgplot_current_totsec  
  real, dimension(:,:), allocatable :: pg_vx, pg_vz, pg_usec, pg_utsec, pg_wsec, pg_ssec, pg_tsec
  
! hsect dumps fields here when called from pgplot  
  real, dimension(:,:), allocatable :: pg_u, pg_v, pg_f

! the integer land mask used by pgplot hsect
  integer, allocatable :: pgmask(:,:)
  
  real, allocatable :: timex(:)
  
  
! count the number of plots, and the number of pgplot timeseries plots
  integer :: pgplot_numplots=0
  integer :: pgplot_numts=0
  
! timeseries window
  integer :: pgplot_timeseries_window = 0
  
  type pgplot_object   
     integer :: ind=0                                         ! index in the pgplots array for this object     
     character(100) :: title=" ", xlabel=" ", ylabel=" "      ! plot title and labels if applicable     
     character(20) :: what=" "                                ! what variable to plot     
     character(1) :: slice=" "                                ! which slice? "i","j" or "k", only relevant for 3d pgslice.     
     integer :: type=0                                        ! the type of plot, 0 means blank. 1 is 2d, 2 hsect, 3 vsect, 4
                                                              ! timeseries, 5 vprof, 6 3d slice, 7 2d slice.
     integer :: i=0, j=0, k=0                                 ! gridpt. to fetch time series, or slice number for 3d pgslice.
     real :: true_range(4) = 0.0                              ! the horisontal grid range to plot (meters), if applicable
     integer :: range(4)=0                                    ! the horisontal, global array range from which to plot hsect, slice
     integer :: range3d(6)=0                                  ! full 3d range used by maxmin.f90 and maybe others.
     real, pointer:: f2d(:,:) => null(), f3d(:,:,:) => null() ! pointers to arrays to slice
     real :: fmax=-1e30, fmin=1e30                            ! span color map between these values, for trends it is the y range.
     real :: zmax=-1e30, zmin=1e30                            ! vertical range to plot, if applicable
     real :: depth=0.0                                        ! hsect depth     
     logical :: quiv=.true.                                   ! overlay velocity on hsect?     
     integer :: ts_i=0                                        ! index in timeseries for which series to view
     integer :: vsect_np=0                                    ! vsect points
     real :: px(10), py(10)
     real, allocatable :: data1d(:), cache1d(:)               ! storage for e.g. timeseries.
     real, allocatable :: data2d(:,:), data3d(:,:,:)
  end type pgplot_object
    
  integer, parameter :: pgplot_maxplots=30
  type(pgplot_object), save :: pgplots(pgplot_maxplots)

contains

  subroutine init_pgplot_core(width, aspect, transform, device, layout, palette, procid)
    implicit none
    integer pgbeg
    real(sp), optional :: width, aspect, transform(6)
    character(*), optional :: device
    integer, optional :: layout(2), palette
    integer, optional :: procid
    logical master
    integer mi(1)

    master=.true.
    if(present(procid)) then
        master=.false.
        if(procid == 0) master=.true.
    end if

    if(present(layout)) then
        pgplot_layout=layout
    end if

! The tr transform. gives the mapping between indices i,j and screen coordinates.
! by default, i is "x", j is "y", and 1:1 scale.

    if(present(transform)) then
        pgplot_transform = transform
    else
! rotation tensor; pgplot_theta = radians (counter clock wise)
        pgplot_transform(1) = -0.5
        pgplot_transform(2) = cos(pgplot_theta)
        pgplot_transform(3) = -sin(pgplot_theta)
        pgplot_transform(4) = -0.5
        pgplot_transform(5) = sin(pgplot_theta)
        pgplot_transform(6) = cos(pgplot_theta)
    end if

! open graphics device

    if(present(device)) pgplot_device = device

    if (master)then
        if(pgbeg(0, trim(pgplot_device), pgplot_layout(1), pgplot_layout(2)) .ne. 1) then
            write(*,*)'ERROR subroutine init_pgplot_core: could not open device=',trim(pgplot_device)
            call abort()
        end if
    end if

    if(present(aspect))then
        pgplot_aspect=aspect
    end if

    if(present(width)) pgplot_width=width

    if(master)then
        if(pgplot_width_set .or. pgplot_aspect_set)then
            call pgpap (pgplot_width , pgplot_aspect)
        end if

        if(pgplot_wait_after_plot)then
            call pgask(.true.)
        else
            call pgask(.false.)
        end if
    end if

!  color map. 1=gråtone, 2=flammetoner, 3=annenflammetoneosv
!             7:invgrå, 10=viridis
    if(present(palette)) pgplot_palette = palette

    if(master)then
        call palett(pgplot_palette)
        call tweak_color_range("extend_down")   
    end if
    
    pgplot_core_do_init=.false.
  end subroutine init_pgplot_core
 
  
  subroutine qaxes(xlabel, ylabel, title, grid, skipframe)
! some reasonable default axes
    character(*), optional :: xlabel, ylabel, title
    character(100) :: gxlabel, gylabel, gtitle
    logical, optional :: grid, skipframe   
    logical drawframe

    drawframe=.true.
    if(present(skipframe)) drawframe=.not.skipframe

    call pgsci(pgplot_grid_color)
    call pgslw(pgplot_grid_lw)    
    
! draw the fat red plot box
    if(drawframe)then
        call pgbox('BCTSI',0.0,0,'BCTSI',0.0,0)
    else
        call pgbox('TSI',0.0,0,'TSI',0.0,0)
    end if
        
! numeric labels, thickness, height
    call pgslw(pgplot_numbering_lw)
    call pgsch(pgplot_character_height)
    call pgbox('N',0.0,0,'N',0.0,0)

! white labels  
    call pgsci(pgplot_text_color)
    call pgslw(pgplot_label_lw)

    gxlabel='x'
    gylabel='y'
    gtitle=' '
    if(present(xlabel)) gxlabel=xlabel
    if(present(ylabel)) gylabel=ylabel    
    if(present(title)) gtitle=title
    call pglab(trim(gxlabel), trim(gylabel), trim(gtitle))

    if(present(grid)) then
        if(grid)then
            call pgslw(1)   ! narrow 
            call pgsci(pgplot_grid_color)   ! red          
            call pgsls(4)   ! dots
            call pgbox('G',0.0,0,'G',0.0,0)  ! grid lines
            call pgsls(1)   ! dots
        end if
    end if
  end subroutine qaxes


  subroutine pcolor(f, fmin, fmax, range, palette, colorbar, axes, batch, mask)
!    
! paint a (pixel) color image of the REAL, 2D array f 
! the other arguments are all optional;
!   fmin:   the first color of the palette starts for this value of f
!           default fmin=minval(f)    
!   fmax:   the last color of the palette. fmax=maxval(f) default.
!   range = (i1,i2,j1,j2). only plot f in the range f(i1:i2,j1:j2),
!           default range=(1,size(f,1),1,size(f,2))    
!   file    a file name, if present also dump the image to a file.
!           defaults to not present.
!   palette: an integer from 1-9 selecting a predefined palette.
!   colorbar: logical, if .true. add a colorbar. defaults to .false.
!   batch:  logical, if true, pcolor has been called from setup2 batch mode
!   mask:   logical, if true, give fmin the background color to mask out e.g. land points.
!    
    implicit none
    real f(:,:)
    real, optional :: fmin, fmax
    integer, optional :: range(4), palette
    logical, optional :: colorbar, axes
    logical, optional :: batch, mask
    real(sp) gfmax, gfmin
    logical :: draw_cbar, set_window, masking
    integer is,ie,js,je
    real(sp) displ, width, ofset

    set_window = .true.
    if(present(batch)) then
        if(batch) set_window = .false.
    end if

    masking = .false.
    if(present(mask)) masking=mask 
    
    if(pgplot_core_do_init) call init_pgplot_core()

    if(present(fmin))then
        if(abs(fmin) > 1e-20) then
        gfmin=fmin
        else
            gfmin=0.0
        end if
    else
        gfmin=minval(real(f,4))
    end if
    if(present(fmax))then
        gfmax=fmax
    else
        gfmax=maxval(f)
    end if

! if f is constant, give it a color in the lower range
    ofset=0.01*gfmax
    if(ofset==0.) ofset=0.01

    if(gfmax == gfmin) then
        gfmax = gfmax + ofset
        gfmin = gfmin - ofset
    end if

    if(gfmin > 1e20 .or. gfmax > 1e20 .or. gfmin<-1e20 .or. gfmax<-1e20) then
        write(*,*) "subroutine pcolor: I don't want plots with gfmin, gfmax=:", gfmin, gfmax
        return
    end if
    
    draw_cbar=.false.
    if(present(colorbar))then
        if(colorbar) draw_cbar=.true.
    end if

    if(present(range))then
        if(size(range) .ne. 4) then
            write(*,*)'ERROR subroutine pcolor: range not an array of length 4'
            return
        end if
        is=range(1) ; ie=range(2) ; js=range(3) ; je=range(4)
    else
        is=1 ; ie=size(f,1) ; js=1 ; je=size(f,2)
    end if

    if(present(palette))then
        call palett(palette)
    end if
    
!    if(masking) call tweak_color_range("extend_down")   
    
    if(set_window) call pgwnad( real(is-1,4), real(ie,4), real(js-1,4), real(je,4))

! std colors: 0=background 1=white 2=red 3=green 4=blue 5  6  7 .. 15
    call pgsci(pgplot_grid_color)

    call pgimag(real(f,4), size(f,1), size(f,2), is, ie, js,je, gfmin, gfmax, pgplot_transform)

    if(draw_cbar)then
        displ=2  
        width=4.0  
        call pgsci(pgplot_grid_color)  
        call pgsch(pgplot_character_height*0.7_4)  
        call pgwedg('RI', displ, width, gfmin, gfmax ,' ')  
    end if

!    if(masking) call tweak_color_range("reset")
    
    if(present(axes)) then
        if(axes) call qaxes
    end if
  end subroutine pcolor

  
  subroutine tweak_color_range(action)
    implicit none
    character(*) action
    real(sp) :: cr=0.,cg=0.,cb=0.,bg_cr=0.,bg_cg=0.,bg_cb=0.
    integer :: newcol1=0, col1=0, col2=0

    select case(action)
    case("extend_down")
! tweak the palettes slightly: we want the first color in the palette
! to be the same as the background, in order to give e.g. land points the
! background color. if it isnt, expand the color range by one
! (downwards). matlab does it differently; with a palette with the bg color in
! the middle. we could do that, but it would require to detect which
! level has the middle color for each plot.

! this trick must be done before pgimag, and only some devices allow
! to use e.g.pgscr after the first call to pgimag. /xs seems ok, /gif not.
        
! find the background color of the current device (and palette?)
        call pgqcr (0, bg_cr, bg_cg, bg_cb)
! get the current color range for pgimag
        call pgqcir(col1,col2)
! get the first color        
        call pgqcr (col1, cr, cg, cb)        
! if the first color is not equal to background, extend range
! downwards into the colors reserved for curves,text etc;
        if(cr.ne.bg_cr .or. cg.ne.bg_cg .or. cb.ne.bg_cb)then
            if(col1 > 0) then
                call pgqcr (col1-1, cr, cg, cb)           ! store the old color
                call pgscr (col1-1, bg_cr, bg_cg, bg_cb)  ! set col-1 to bg color
                call pgscir(col1-1, col2-15)              ! set color range to be col1-1 to col2
            else
                write(*,*) "INFO: subroutine pcolor: could not extend color range"
            end if                
        end if
    case("reset")
! set the color range back to what it was
        call pgscr (col1-1, cr, cg, cb)
        call pgscir(col1, col2)
    case default
        write(*,*) "pgplot_core tweak_color_range ignoring unknown argument=",trim(action)
    end select
  end subroutine tweak_color_range

  

  subroutine quiver(a, b, range, scale, blank, dens, headsize, axes, maxv)
!
! QUIVER -- add vector map of a 2D data array, with blanking
!
    real, intent(in) :: a(:,:), b(:,:)
    integer, intent(in), optional :: range(4), dens 
    real, intent(in), optional :: scale, blank, headsize
    real, intent(out), optional :: maxv
    logical, optional :: axes

    integer idim, jdim, is, ie, js, je    
    integer  i, j, gdens
    real(sp) x1, y1, x2, y2
    real(sp) cc
    real gscale
    intrinsic sqrt, max, min
    logical blanking

    if(pgplot_core_do_init) call init_pgplot_core

    is=1 ; ie=size(a,1)
    js=1 ; je=size(a,2)
    if(present(range))then
        if(size(range).ne.4) then
            write(*,*)'ERROR in subroutine quiver: size(range) .ne. 4'
            return
        end if
        is=range(1) ; ie=range(2) ; js=range(3) ; je=range(4)
    end if
    idim=size(a,1) ; jdim=size(a,2)
    if(idim.ne.size(b,1) .or. jdim.ne.size(b,2))then
        write(*,*)'ERROR in subroutine quiver: a,b dimensions not equal!'        
        write(*,*)'  size(a)=',size(a),'size(b)=',size(b)
        return
    end if
    if (is.lt.1 .or. ie.gt.idim .or. is.ge.ie .or. &
         js.lt.1 .or. je.gt.jdim .or. js.ge.je) then
        write(*,*)'ERROR in subroutine quiver: range=',is,ie,js,je
        write(*,*)'  does not correspond to a,b dimension=',idim,jdim
        return
    end if

    blanking=.false.
    if(present(blank)) then
        blanking=.true.
    end if

    gdens=1
    if(present(dens)) gdens=dens

! scaling of arrow lengths
    cc=0.0
    do  j=js,je
        do i=is,ie
            if(blanking) then
                if (a(i,j).ne.blank .and. b(i,j).ne.blank) &
                     cc = max(cc, a(i,j)**2+b(i,j)**2)
            else
                cc = max(cc, a(i,j)**2+b(i,j)**2)
            end if
        end do
    end do
    if (cc.eq.0.0) return

    if(present(maxv)) maxv=sqrt(cc)
    
    gscale=pgplot_quiver_scale
    if(present(scale)) gscale=scale
    cc = 1.0/sqrt(cc) * gscale * gdens

    if(present(headsize)) then
        call pgsch(real(headsize,4))
    else
        call pgsch(real(pgplot_headsize,4))
    end if

    call pgsci(pgplot_quiver_color)
    call pgslw(pgplot_quiver_lw)

    call pgbbuf()

    do j=js,je,gdens
        do i=is,ie,gdens
            if(blanking) then
                if (a(i,j) == blank .or. b(i,j) == blank) cycle
            end if
            x2=pgplot_transform(1) + pgplot_transform(2)*i + pgplot_transform(3)*j
            y2=pgplot_transform(4) + pgplot_transform(5)*i + pgplot_transform(6)*j
            x1=x2 - cc*( pgplot_transform(2)*a(i,j) + pgplot_transform(3)*b(i,j) )
            y1=y2 - cc*( pgplot_transform(5)*a(i,j) + pgplot_transform(6)*b(i,j) )
            call pgarro(x1, y1, x2, y2)            
        end do
    end do

    call pgebuf()

    if(present(headsize)) then
        call pgsch(pgplot_character_height)
    end if

    if(present(axes)) then
        if(axes) call qaxes
    end if
  end subroutine quiver


  subroutine plot(f, x, fmax, fmin, xmin, xmax, xlabel, ylabel, title, style)
!    
! plot the graph of f versus x, optionally with a title, xlabel and y label.
! the optional fmax and fmin gives a fixed range for f. same for x.
!    
    implicit none
    real f(:), x(:)
    real, optional :: fmax,fmin,xmin,xmax
    character(*), optional :: xlabel,ylabel,title
    integer, optional :: style
    character(100) :: cxlabel="x",cylabel="f(x)",ctitle=""
    real(4) :: cfmax,cfmin,cxmin,cxmax
    integer cstyle
    
    if(pgplot_core_do_init) call init_pgplot_core()
    if(present(xlabel)) cxlabel=xlabel
    if(present(ylabel)) cylabel=ylabel
    if(present(title)) ctitle=title
    cfmax = maxval(f) ; if(present(fmax)) cfmax = fmax
    cfmin = minval(f) ; if(present(fmin)) cfmin = fmin
    cxmax = maxval(x) ; if(present(xmax)) cxmax = xmax
    cxmin = minval(x) ; if(present(xmin)) cxmin = xmin

    cstyle = 1
    if(present(style)) cstyle=style
    
! make sure the range is sane    
    if( abs(cfmax-cfmin) < 1e-20) then
        if(abs(cfmax) > 1e-19) then
            cfmax = cfmax + 0.1*abs(cfmax)
            cfmin = cfmin - 0.1*abs(cfmax)
        else
            cfmax = 1e-20
            cfmin = 1e-20
        end if
    end if
    
    if(cfmax > 1e20 .or. cfmin < -1e20) return
    
    if( abs(cxmax-cxmin) < 1e-20) then
        if(abs(cxmax) > 1e-19) then
            cxmax = cxmax + 0.1*abs(cxmax)
            cxmin = cxmin - 0.1*abs(cxmax)
        else
            cxmax = 1e-20
            cxmin = 1e-20
        end if
    end if
    
    if(cxmax > 1e20 .or. cxmin < -1e20) return
    
! use the (default) viewport from pgplot_core    
    call pgsvp(pgplot_viewport(1), pgplot_viewport(2), pgplot_viewport(3), pgplot_viewport(4))
    call pgswin(cxmin, cxmax, cfmin, cfmax)
    call trend(f, x, style=cstyle)
    call qaxes(trim(cxlabel), trim(cylabel), trim(ctitle), grid=.true.)
  end subroutine plot


  subroutine text(x, y, string, angle, current, col, size, weight)
!
! draw text at the given x,y position, optionally with given angle,
! color, size and weight. if current is true, use the current
! properties, otherwise, if no argument properties are given, use
! pgplot_core defaults.
!
    implicit none
    character(*) string
    real x,y
    real, optional :: angle
    logical, optional :: current
    integer, optional :: col, weight
    real, optional :: size
    real(4) :: aangle, asize, fjust=0.0  ! fjust range 0-1, 0=left,1=right.
    integer :: acol, aweight
    logical acurrent

    aangle = 0.0
    if(present(angle)) aangle=angle
    if(present(current)) acurrent=current
    if(.not.acurrent)then
        acol = pgplot_text_color
        asize = pgplot_character_height    
        aweight = pgplot_label_lw
        if(present(col)) acol = col
        if(present(size)) asize = size
        if(present(weight)) aweight = weight
        call pgsci(acol)
        call pgsch(asize)
        call pgslw(aweight)
    end if
    
    call pgptxt(real(x,4), real(y,4), aangle, fjust, trim(string))
  end subroutine text


  subroutine trend(f,x,col,style)
!    
! plot the graph of f versus x with color col in the current "scene", i.e. no labels, box etc.
!    
    implicit none
    real f(:),x(:)
    integer, optional :: col, style
    integer im, gcol, gstyle

    im = size(f)
    if(im .ne. size(x)) then
        write(*,*)"ERROR in subroutine trend: size(f).NE.size(x)"
        return
    end if

    if(pgplot_core_do_init) call init_pgplot_core

    gcol=pgplot_curve_color
    if(present(col)) gcol=col
    call pgslw(pgplot_curve_lw)
    call pgsci(gcol)

    gstyle=1
    if(present(style)) gstyle=style
    call pgsls(gstyle)
    
    call pgline(im, real(x,kind=4), real(f,kind=4))
    call pgsls(1)
  end subroutine trend


  subroutine scatter(x, y, style, color, line_width)
!
! Scatter plot of points located at (x(:),y(:))
!  
    implicit none
    real x(:),y(:)
    integer, optional :: style, color, line_width
    integer :: num_pts = 0
    integer :: symbol = -1
    integer :: ci,lw
!
!-- Get array size   
    num_pts = size(x)
    if (num_pts.ne.size(y)) then
       write(*,*) 'ERROR in subroutine scatter: x and y do not match!'
       call abort()
    end if
!-- Set point symbol
    if (present(style)) symbol = style
!-- Set color index
    if (present(color)) then
       call pgqci(ci)
       call pgsci(color)
    end if
!-- Set line width
    if (present(line_width)) then
       call pgqlw(lw)
       if ((line_width >= 1) .and. (line_width <= 201)) call pgslw(line_width)
    end if
!    
    call pgpt(num_pts,real(x,4),real(y,4),symbol)
!
!-- Reset original color index and line width
    if (present(color)) call pgsci(ci)
    if (present(line_width)) call pgslw(lw)
  end subroutine scatter


  subroutine pldj(x1, y1, x2, y2, style, color, color_array, line_width)
!
! Plot lines between points
!
    implicit none
    real :: x1(:),y1(:),x2(:),y2(:)
    integer, optional :: color_array(:)
    integer, optional :: style, color, line_width
    logical :: use_color_array
    integer :: num_pts = 0
    integer :: ci,ls,lw
    integer :: i,j
!
!-- Get array size   
    num_pts = size(x1)
    if ((num_pts.ne.size(y1)) .or. (num_pts.ne.size(x2)) .or.  &
         (num_pts.ne.size(y2))) then
       write(*,*) 'ERROR in subroutine pldj: points in x and y do not match'
       call abort()
    end if
!-- Save old color index
    use_color_array = .false.
    if (present(color_array)) then
        if (num_pts.ne.size(color_array)) then
          write(*,*) 'ERROR in subroutine pldj: color array does not match data'
          call abort()
       end if
       use_color_array = .true.
    end if
    if (present(color) .or. use_color_array) call pgqci(ci)
!-- Set line style
    if (present(style)) then
       call pgqls(ls)
       if ((style >= 1) .and. (style <= 5)) call pgsls(style)
    end if
!-- Set line width
    if (present(line_width)) then
       call pgqlw(lw)
       if ((line_width >= 1) .and. (line_width <= 201)) call pgslw(line_width)
    end if
!
    if (present(color)) call pgsci(color)
!   
    call pgbbuf()
!
    do i=1,num_pts
       if (use_color_array) call pgsci(color_array(i))
       call pgmove(real(x1(i),4),real(y1(i),4))
       call pgdraw(real(x2(i),4),real(y2(i),4))
    end do
!
    call pgebuf()
!
!-- Reset original color index, line style, and line width
    if (present(color) .or. use_color_array) call pgsci(ci)
    if (present(style)) call pgsls(ls)
    if (present(line_width)) call pgslw(lw)
  end subroutine pldj


  subroutine setvp  
!
! Set the viewport, allowing margins around the edge for annotation.
! (This is similar in effect to PGVSTD, but has different margins.)
! The routine determines the view-surface size and allocates margins
! as fractions of the minimum of width and height.
!
    real(sp) d, vpx1, vpx2, vpy1, vpy2  
!
    if(pgplot_core_do_init) call init_pgplot_core

    call pgsvp(0.0, 1.0, 0.0, 1.0)  
    call pgqvp(1, vpx1, vpx2, vpy1, vpy2)  
    d = min(vpx2-vpx1, vpy2-vpy1)/40.0  
    vpx1 = vpx1 + 5.0*d  
    vpx2 = vpx2 - 2.0*d  
    vpy1 = vpy1 + 8.0*d  
    vpy2 = vpy2 - 2.0*d  
    call pgvsiz(vpx1, vpx2, vpy1, vpy2)  
  end subroutine setvp


  subroutine palett(type)  
!
! Set a "palette" of colors in the range of color indices used by
! PGIMAG.
! 
    implicit none
    integer type  
!
    real(sp) gl(2), gr(2), gg(2), gb(2)
    real(sp) gl2(2), gr2(2), gg2(2), gb2(2)  
    real(sp) rl(9), rr(9), rg(9), rb(9)  
    real(sp) hl(5), hr(5), hg(5), hb(5)  
    real(sp) wl(10), wr(10), wg(10), wb(10)  
    real(sp) al(20), ar(20), ag(20), ab(20)  
    real(sp) tl(4), tr(4), tg(4), tb(4)

    real(sp), dimension(64):: m1l,m1r,m1g,m1b
    real(sp), dimension(64):: m2l,m2r,m2g,m2b  

    integer dim,i

! viridis from matplotlib. not sure how vl should be set, but try linear from 0 to 1.
    real(sp), parameter, dimension(*) :: &
         vr = (/0.267004, 0.268510, 0.269944, 0.271305, 0.272594, 0.273809,   &
         0.274952, 0.276022, 0.277018, 0.277941, 0.278791, 0.279566, 0.280267,&
         0.280894, 0.281446, 0.281924, 0.282327, 0.282656, 0.282910, 0.283091,&
         0.283197, 0.283229, 0.283187, 0.283072, 0.282884, 0.282623, 0.282290,&
         0.281887, 0.281412, 0.280868, 0.280255, 0.279574, 0.278826, 0.278012,&
         0.277134, 0.276194, 0.275191, 0.274128, 0.273006, 0.271828, 0.270595,&
         0.269308, 0.267968, 0.266580, 0.265145, 0.263663, 0.262138, 0.260571,&
         0.258965, 0.257322, 0.255645, 0.253935, 0.252194, 0.250425, 0.248629,&
         0.246811, 0.244972, 0.243113, 0.241237, 0.239346, 0.237441, 0.235526,&
         0.233603, 0.231674, 0.229739, 0.227802, 0.225863, 0.223925, 0.221989,&
         0.220057, 0.218130, 0.216210, 0.214298, 0.212395, 0.210503, 0.208623,&
         0.206756, 0.204903, 0.203063, 0.201239, 0.199430, 0.197636, 0.195860,&
         0.194100, 0.192357, 0.190631, 0.188923, 0.187231, 0.185556, 0.183898,&
         0.182256, 0.180629, 0.179019, 0.177423, 0.175841, 0.174274, 0.172719,&
         0.171176, 0.169646, 0.168126, 0.166617, 0.165117, 0.163625, 0.162142,&
         0.160665, 0.159194, 0.157729, 0.156270, 0.154815, 0.153364, 0.151918,&
         0.150476, 0.149039, 0.147607, 0.146180, 0.144759, 0.143343, 0.141935,&
         0.140536, 0.139147, 0.137770, 0.136408, 0.135066, 0.133743, 0.132444,&
         0.131172, 0.129933, 0.128729, 0.127568, 0.126453, 0.125394, 0.124395,&
         0.123463, 0.122606, 0.121831, 0.121148, 0.120565, 0.120092, 0.119738,&
         0.119512, 0.119423, 0.119483, 0.119699, 0.120081, 0.120638, 0.121380,&
         0.122312, 0.123444, 0.124780, 0.126326, 0.128087, 0.130067, 0.132268,&
         0.134692, 0.137339, 0.140210, 0.143303, 0.146616, 0.150148, 0.153894,&
         0.157851, 0.162016, 0.166383, 0.170948, 0.175707, 0.180653, 0.185783,&
         0.191090, 0.196571, 0.202219, 0.208030, 0.214000, 0.220124, 0.226397,&
         0.232815, 0.239374, 0.246070, 0.252899, 0.259857, 0.266941, 0.274149,&
         0.281477, 0.288921, 0.296479, 0.304148, 0.311925, 0.319809, 0.327796,&
         0.335885, 0.344074, 0.352360, 0.360741, 0.369214, 0.377779, 0.386433,&
         0.395174, 0.404001, 0.412913, 0.421908, 0.430983, 0.440137, 0.449368,&
         0.458674, 0.468053, 0.477504, 0.487026, 0.496615, 0.506271, 0.515992,&
         0.525776, 0.535621, 0.545524, 0.555484, 0.565498, 0.575563, 0.585678,&
         0.595839, 0.606045, 0.616293, 0.626579, 0.636902, 0.647257, 0.657642,&
         0.668054, 0.678489, 0.688944, 0.699415, 0.709898, 0.720391, 0.730889,&
         0.741388, 0.751884, 0.762373, 0.772852, 0.783315, 0.793760, 0.804182,&
         0.814576, 0.824940, 0.835270, 0.845561, 0.855810, 0.866013, 0.876168,&
         0.886271, 0.896320, 0.906311, 0.916242, 0.926106, 0.935904, 0.945636,&
         0.955300, 0.964894, 0.974417, 0.983868, 0.993248/)

    real(sp), parameter, dimension(*) :: &
         vg = (/ 0.004874, 0.009605, 0.014625, 0.019942, 0.025563, 0.031497,  &
         0.037752, 0.044167, 0.050344, 0.056324, 0.062145, 0.067836, 0.073417,&
         0.078907, 0.084320, 0.089666, 0.094955, 0.100196, 0.105393, 0.110553,&
         0.115680, 0.120777, 0.125848, 0.130895, 0.135920, 0.140926, 0.145912,&
         0.150881, 0.155834, 0.160771, 0.165693, 0.170599, 0.175490, 0.180367,&
         0.185228, 0.190074, 0.194905, 0.199721, 0.204520, 0.209303, 0.214069,&
         0.218818, 0.223549, 0.228262, 0.232956, 0.237631, 0.242286, 0.246922,&
         0.251537, 0.256130, 0.260703, 0.265254, 0.269783, 0.274290, 0.278775,&
         0.283237, 0.287675, 0.292092, 0.296485, 0.300855, 0.305202, 0.309527,&
         0.313828, 0.318106, 0.322361, 0.326594, 0.330805, 0.334994, 0.339161,&
         0.343307, 0.347432, 0.351535, 0.355619, 0.359683, 0.363727, 0.367752,&
         0.371758, 0.375746, 0.379716, 0.383670, 0.387607, 0.391528, 0.395433,&
         0.399323, 0.403199, 0.407061, 0.410910, 0.414746, 0.418570, 0.422383,&
         0.426184, 0.429975, 0.433756, 0.437527, 0.441290, 0.445044, 0.448791,&
         0.452530, 0.456262, 0.459988, 0.463708, 0.467423, 0.471133, 0.474838,&
         0.478540, 0.482237, 0.485932, 0.489624, 0.493313, 0.497000, 0.500685,&
         0.504369, 0.508051, 0.511733, 0.515413, 0.519093, 0.522773, 0.526453,&
         0.530132, 0.533812, 0.537492, 0.541173, 0.544853, 0.548535, 0.552216,&
         0.555899, 0.559582, 0.563265, 0.566949, 0.570633, 0.574318, 0.578002,&
         0.581687, 0.585371, 0.589055, 0.592739, 0.596422, 0.600104, 0.603785,&
         0.607464, 0.611141, 0.614817, 0.618490, 0.622161, 0.625828, 0.629492,&
         0.633153, 0.636809, 0.640461, 0.644107, 0.647749, 0.651384, 0.655014,&
         0.658636, 0.662252, 0.665859, 0.669459, 0.673050, 0.676631, 0.680203,&
         0.683765, 0.687316, 0.690856, 0.694384, 0.697900, 0.701402, 0.704891,&
         0.708366, 0.711827, 0.715272, 0.718701, 0.722114, 0.725509, 0.728888,&
         0.732247, 0.735588, 0.738910, 0.742211, 0.745492, 0.748751, 0.751988,&
         0.755203, 0.758394, 0.761561, 0.764704, 0.767822, 0.770914, 0.773980,&
         0.777018, 0.780029, 0.783011, 0.785964, 0.788888, 0.791781, 0.794644,&
         0.797475, 0.800275, 0.803041, 0.805774, 0.808473, 0.811138, 0.813768,&
         0.816363, 0.818921, 0.821444, 0.823929, 0.826376, 0.828786, 0.831158,&
         0.833491, 0.835785, 0.838039, 0.840254, 0.842430, 0.844566, 0.846661,&
         0.848717, 0.850733, 0.852709, 0.854645, 0.856542, 0.858400, 0.860219,&
         0.861999, 0.863742, 0.865448, 0.867117, 0.868751, 0.870350, 0.871916,&
         0.873449, 0.874951, 0.876424, 0.877868, 0.879285, 0.880678, 0.882046,&
         0.883393, 0.884720, 0.886029, 0.887322, 0.888601, 0.889868, 0.891125,&
         0.892374, 0.893616, 0.894855, 0.896091, 0.897330, 0.898570, 0.899815,&
         0.901065, 0.902323, 0.903590, 0.904867, 0.906157/)

    real(sp), parameter, dimension(*) :: &
         vb = (/0.329415, 0.335427, 0.341379, 0.347269, 0.353093, 0.358853,   &
         0.364543, 0.370164, 0.375715, 0.381191, 0.386592, 0.391917, 0.397163,&
         0.402329, 0.407414, 0.412415, 0.417331, 0.422160, 0.426902, 0.431554,&
         0.436115, 0.440584, 0.444960, 0.449241, 0.453427, 0.457517, 0.461510,&
         0.465405, 0.469201, 0.472899, 0.476498, 0.479997, 0.483397, 0.486697,&
         0.489898, 0.493001, 0.496005, 0.498911, 0.501721, 0.504434, 0.507052,&
         0.509577, 0.512008, 0.514349, 0.516599, 0.518762, 0.520837, 0.522828,&
         0.524736, 0.526563, 0.528312, 0.529983, 0.531579, 0.533103, 0.534556,&
         0.535941, 0.537260, 0.538516, 0.539709, 0.540844, 0.541921, 0.542944,&
         0.543914, 0.544834, 0.545706, 0.546532, 0.547314, 0.548053, 0.548752,&
         0.549413, 0.550038, 0.550627, 0.551184, 0.551710, 0.552206, 0.552675,&
         0.553117, 0.553533, 0.553925, 0.554294, 0.554642, 0.554969, 0.555276,&
         0.555565, 0.555836, 0.556089, 0.556326, 0.556547, 0.556753, 0.556944,&
         0.557120, 0.557282, 0.557430, 0.557565, 0.557685, 0.557792, 0.557885,&
         0.557965, 0.558030, 0.558082, 0.558119, 0.558141, 0.558148, 0.558140,&
         0.558115, 0.558073, 0.558013, 0.557936, 0.557840, 0.557724, 0.557587,&
         0.557430, 0.557250, 0.557049, 0.556823, 0.556572, 0.556295, 0.555991,&
         0.555659, 0.555298, 0.554906, 0.554483, 0.554029, 0.553541, 0.553018,&
         0.552459, 0.551864, 0.551229, 0.550556, 0.549841, 0.549086, 0.548287,&
         0.547445, 0.546557, 0.545623, 0.544641, 0.543611, 0.542530, 0.541400,&
         0.540218, 0.538982, 0.537692, 0.536347, 0.534946, 0.533488, 0.531973,&
         0.530398, 0.528763, 0.527068, 0.525311, 0.523491, 0.521608, 0.519661,&
         0.517649, 0.515571, 0.513427, 0.511215, 0.508936, 0.506589, 0.504172,&
         0.501686, 0.499129, 0.496502, 0.493803, 0.491033, 0.488189, 0.485273,&
         0.482284, 0.479221, 0.476084, 0.472873, 0.469588, 0.466226, 0.462789,&
         0.459277, 0.455688, 0.452024, 0.448284, 0.444467, 0.440573, 0.436601,&
         0.432552, 0.428426, 0.424223, 0.419943, 0.415586, 0.411152, 0.406640,&
         0.402049, 0.397381, 0.392636, 0.387814, 0.382914, 0.377939, 0.372886,&
         0.367757, 0.362552, 0.357269, 0.351910, 0.346476, 0.340967, 0.335384,&
         0.329727, 0.323998, 0.318195, 0.312321, 0.306377, 0.300362, 0.294279,&
         0.288127, 0.281908, 0.275626, 0.269281, 0.262877, 0.256415, 0.249897,&
         0.243329, 0.236712, 0.230052, 0.223353, 0.216620, 0.209861, 0.203082,&
         0.196293, 0.189503, 0.182725, 0.175971, 0.169257, 0.162603, 0.156029,&
         0.149561, 0.143228, 0.137064, 0.131109, 0.125405, 0.120005, 0.114965,&
         0.110347, 0.106217, 0.102646, 0.099702, 0.097452, 0.095953, 0.095250,&
         0.095374, 0.096335, 0.098125, 0.100717, 0.104071, 0.108131, 0.112838,&
         0.118128, 0.123941, 0.130215, 0.136897, 0.143936/)

    real(sp), dimension(size(vb)) :: vl

    data m2r /  0  ,  0  ,  0  ,  0  ,   0  ,     0  ,     0  ,     0,  &
         0  ,     0  ,     0  ,     0  ,     0  ,     0  ,     0  ,     0 , &
         0  ,     0  ,     0  ,     0  ,     0  ,     0  ,     0  ,     0,  &
         0.0625  ,0.1250  ,0.1875  ,0.2500  ,0.3125  ,0.3750  ,0.4375  ,0.5000,  &
         0.5625  ,0.6250  ,0.6875  ,0.7500  ,0.8125  ,0.8750  ,0.9375  ,1.0000,  &
         1.0000  ,1.0000  ,1.0000  ,1.0000  ,1.0000  ,1.0000  ,1.0000  ,1.0000,  &
         1.0000  ,1.0000  ,1.0000  ,1.0000  ,1.0000  ,1.0000  ,1.0000  ,1.0000,  &
         1.0000  ,0.9375  ,0.8750  ,0.8125  ,0.7500  ,0.6875  ,0.6250  ,0.5625&
         /

    data m2g/    0,    0,     0,         0,         0,         0,&
         0,         0,    0.0625,    0.1250,    0.1875,    0.2500,&
         0.3125,    0.3750,    0.4375,    0.5000,    0.5625,    0.6250,&
         0.6875,    0.7500,    0.8125,    0.8750,    0.9375,    1.0000,&
         1.0000,    1.0000,    1.0000,    1.0000,    1.0000,    1.0000,&
         1.0000,    1.0000,    1.0000,    1.0000,    1.0000,    1.0000,&
         1.0000,    1.0000,    1.0000,    1.0000,    1.0000,    0.9375,&
         0.8750,    0.8125,    0.7500,    0.6875,    0.6250,    0.5625,&
         0.5000,    0.4375,    0.3750,    0.3125,    0.2500,    0.1875,&
         0.1250,    0.0625,         0,         0,         0,         0,&
         0,         0,         0,         0/

    data m2b/0.5625,0.6250, 0.6875, 0.7500, 0.8125, 0.8750, 0.9375,  1.0000,&
         1.0000,  1.0000,  1.0000,  1.0000,  1.0000,  1.0000,  1.0000,  1.0000,&
         1.0000,  1.0000,  1.0000,  1.0000,  1.0000,  1.0000,  1.0000,  1.0000,&
         1.0000,  0.9375,  0.8750,  0.8125,  0.7500,  0.6875,  0.6250,  0.5625,&
         0.5000,  0.4375,  0.3750,  0.3125,  0.2500,  0.1875,  0.1250,  0.0625,&
         0,       0,       0,       0,       0,       0,       0,       0,&
         0,       0,       0,       0,       0,       0,       0,       0,&
         0,       0,       0,       0,       0,       0,       0,      &
         0/

    data m1r/0.0417,0.0833,0.1250 ,0.1667,0.2083,0.2500,0.2917,0.3333,   &
         0.3750,0.4167,0.4583,0.5000,0.5417,0.5833,0.6250,0.6667,0.7083,&
         0.7500,0.7917,0.8333,0.8750,0.9167,0.9583,1.0000,1.0000,1.0000,&
         1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,&
         1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,&
         1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,&
         1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,&
         1.0000,1.0000/

    data m1g/0,   0,   0,    0,       0,       0,       0,     0,&
         0,       0,       0,       0,       0,       0,       0,     0,&
         0,       0,       0,       0,       0,       0,       0,     0,&
         0.0417,0.0833,0.1250,0.1667,0.2083,0.2500,0.2917,0.3333,0.3750,0.4167,&
         0.4583,0.5000,0.5417,0.5833,0.6250,0.6667,0.7083,0.7500,0.7917,0.8333,&
         0.8750,0.9167,0.9583,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,&
         1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000,1.0000/

    data m1b/  0, 0,      0,      0,      0,      0,      0,      0,      0,&
         0,      0,      0,      0,      0,      0,      0,      0,      0,&
         0,      0,      0,      0,      0,      0,      0,      0,      0,&
         0,      0,      0,      0,      0,      0,      0,      0,      0,&
         0,      0,      0,      0,      0,      0,      0,      0,      0,&
         0,      0,      0, 0.0625, 0.1250, 0.1875, 0.2500, 0.3125, 0.3750,&
         0.4375, 0.5000, 0.5625, 0.6250, 0.6875, 0.7500, 0.8125, 0.8750, 0.9375,&
         1.0000/


!
    data gl /0.0, 1.0/  
    data gr /0.0, 1.0/  
    data gg /0.0, 1.0/  
    data gb /0.0, 1.0/
!
    data gl2 /1.0, 0.0/  
    data gr2 /1.0, 0.0/  
    data gg2 /1.0, 0.0/  
    data gb2 /1.0, 0.0/  


!
    data rl /-0.5, 0.0, 0.17, 0.33, 0.50, 0.67, 0.83, 1.0, 1.7/  
    data rr / 0.0, 0.0,  0.0,  0.0,  0.6,  1.0,  1.0, 1.0, 1.0/  
    data rg / 0.0, 0.0,  0.0,  1.0,  1.0,  1.0,  0.6, 0.0, 1.0/  
    data rb / 0.0, 0.3,  0.8,  1.0,  0.3,  0.0,  0.0, 0.0, 1.0/  
!
    data hl /0.0, 0.2, 0.4, 0.6, 1.0/  
    data hr /0.0, 0.5, 1.0, 1.0, 1.0/  
    data hg /0.0, 0.0, 0.5, 1.0, 1.0/  
    data hb /0.0, 0.0, 0.0, 0.3, 1.0/  
!
    data wl /0.0, 0.5, 0.5, 0.7, 0.7, 0.85, 0.85, 0.95, 0.95, 1.0/  
    data wr /0.0, 1.0, 0.0, 0.0, 0.3,  0.8,  0.3,  1.0,  1.0, 1.0/  
    data wg /0.0, 0.5, 0.4, 1.0, 0.0,  0.0,  0.2,  0.7,  1.0, 1.0/  
    data wb /0.0, 0.0, 0.0, 0.0, 0.4,  1.0,  0.0,  0.0, 0.95, 1.0/  
!
    data al /0.0, 0.1, 0.1, 0.2, 0.2, 0.3, 0.3, 0.4, 0.4, 0.5, &
         &         0.5, 0.6, 0.6, 0.7, 0.7, 0.8, 0.8, 0.9, 0.9, 1.0/
    data ar /0.0, 0.0, 0.3, 0.3, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0, &
         &         0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0/
    data ag /0.0, 0.0, 0.3, 0.3, 0.0, 0.0, 0.0, 0.0, 0.8, 0.8, &
         &         0.6, 0.6, 1.0, 1.0, 1.0, 1.0, 0.8, 0.8, 0.0, 0.0/
    data ab /0.0, 0.0, 0.3, 0.3, 0.7, 0.7, 0.7, 0.7, 0.9, 0.9, &
         &         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0/
!
    data tl /0.0, 0.5, 0.5, 1.0/  
    data tr /0.2, 0.6, 0.6, 1.0/  
    data tg /0.0, 0.0, 0.5, 1.0/  
    data tb /1.0, 0.0, 0.0, 0.0/  
!

    if (type.eq.1) then  
!        -- gray scale
        call pgctab(gl, gr, gg, gb, 2, 1.0_4, 0.5_4)  
    else if (type.eq.2) then  
!        -- rainbow
        call pgctab(rl, rr, rg, rb, 9, 1.0_4, 0.5_4)  
    else if (type.eq.3) then  
!        -- heat
        call pgctab(hl, hr, hg, hb, 5, 1.0_4, 0.5_4)  
    else if (type.eq.4) then  
!        -- weird iraf
        call pgctab(wl, wr, wg, wb, 10, 1.0_4, 0.5_4)

    else if (type.eq.5) then  
!        -- aips
        call pgctab(al, ar, ag, ab, 20, 1.0_4, 0.5_4)

    else if (type.eq.6) then  
!        -- tjp
        call pgctab(tl, tr, tg, tb, 4, 1.0_4, 0.5_4)

    else if (type.eq.7) then  
!        -- inverse gray scale
        call pgctab(gl2, gr2, gg2, gb2, 2, 1.0_4, 0.5_4)

    else if (type.eq. 8) then
! hot     
        dim=size(m1g)
        do i=1,dim
            m1l(i)=real(i,sp)/dim
        end do
        call pgctab(m1l, m1r, m1g, m1b, dim, 1.0_4, 0.5_4)

    else if (type.eq. 9) then
! jet     
        dim=size(m2g)
        do i=1,dim
            m2l(i)=real(i,sp)/dim
        end do
        call pgctab(m2l, m2r, m2g, m2b, dim, 1.0_4, 0.5_4)

    else if (type.eq.10) then
! the new viridis from matplotlib        
        dim = size(vg)
        vl = (/(real(i-1,sp)/(dim-1), i=1, dim)/)
        call pgctab(vl, vr, vg, vb, dim, 1.0_4, 0.5_4)

    end if
  end subroutine palett


  subroutine contour(f, fmin, fmax, nlevels, levels, labels, range, palette, colorbar, axes, blank, labelling)
! 
! paint a contour map, optionally filled, of the REAL, 2D array f 
! the other arguments are all optional;
!   fmin:   only draw contours between fmin and fmax.
!           default fmin=minval(f)    
!   fmax:   fmax=maxval(f) per default.
!   levels  the contour levels to draw. if not present draw 10 equally spaced levels.
!   labels: if present, is a 1d real array containing contour levels to be labelled.
!   range = (i1,i2,j1,j2). only plot f in the range f(i1:i2,j1:j2),
!           default range=(1,size(f,1),1,size(f,2))    
!   palette: an integer from 1-9 selecting a predefined palette for filling.
!   colorbar: logical, if .true. add a colorbar. defaults to .false.
!
    real(sp) f(:,:)
    real(sp), optional :: fmin, fmax, blank
    logical, optional :: colorbar
    real(sp), optional :: labels(:), levels(:)
    integer, optional :: range(4), palette, nlevels, labelling
    logical, optional :: axes
    real(sp) glevels(200)
    integer nglevels,i
    logical draw_cbar
    real gfmin, gfmax, displ, width
    integer is,ie,js,je

    if((present(fmin) .or. present(fmax)) .and. present(levels)) then
        write(*,*)"ERROR in subroutine contour: don't know what to do; both fmin/fmax and levels are specified"
        return
    end if

    nglevels = pgplot_contour_levels
    if(present(nlevels)) nglevels = nlevels

    if(pgplot_core_do_init) call init_pgplot_core

    if(present(fmin))then
        gfmin = fmin
    else
        gfmin = minval(f)
    end if
    if(present(fmax))then
        gfmax = fmax
    else
        gfmax = maxval(f)
    end if

    if(present(levels)) then
        nglevels = size(levels)
        glevels(1:nglevels) = levels
    else
        do i=1,nglevels+1
            glevels(i) = gfmin + (gfmax-gfmin)/nglevels*(i-1)
        end do
    end if

    draw_cbar = .false.
    if(present(colorbar))then
        if(colorbar) draw_cbar = .true.
    end if

    if(present(range))then
        if(size(range) .ne. 4) then
            write(*,*)'ERROR in subroutine contour: range not an array of length 4'
            return
        end if
        is=range(1) ; ie=range(2) ; js=range(3) ; je=range(4)
    else
        is=1 ; ie=size(f,1) ; js=1 ; je=size(f,2)
    end if

    if(present(palette))then
        call palett(palette)
    end if

!  farger: 1 hvit  3 grønn 4 blå 2 rød  5  6  7 .. 15      
    call pgsci(pgplot_grid_color)

    if(present(labelling))then
        if(labelling == 0) then
            if(.not.present(labels)) then
                write(*,*)'ERROR in subroutine contour: labelling method=0, but labels are not given'
                return
            end if
        else if(labelling == 1) then
!            if(present(label_interval))then

!            else

!            end if            
        else if(labelling==2) then

        end if

! set color invisible (background)
        call pgsci(pgplot_background)
!      SUBROUTINE PGCONT (A, IDIM, JDIM, I1, I2, J1, J2, C, NC, PGPLOT_TRANSFORM)
    end if

    if(present(blank)) then
        call pgconb (f, size(f,1), size(f,2), is, ie, js, je, glevels, nglevels, pgplot_transform, blank)
    else
        call pgcons (f, size(f,1), size(f,2), is, ie, js, je, glevels, nglevels, pgplot_transform)
    end if

    if(present(labelling)) then
!              SUBROUTINE PGCONL (A, IDIM, JDIM, I1, I2, J1, J2, C, PGPLOT_TRANSFORM,
!     1                   LABEL, INTVAL, MININT)
        write(*,*) 'WARNING subroutine contour: labelling is on the TODO'
    end if

    if(draw_cbar)then
        displ=2  
        width=4.0  
        call pgsci(pgplot_grid_color)  
        call pgsch(pgplot_character_height*0.7)  
        call pgwedg('RI', displ, width, gfmin, gfmax ,' ')  
    end if

    if(present(axes)) then
        if(axes) call qaxes
    end if
  end subroutine contour
  
end module pgplot_core
