module qmpi
!  
! A module defining a minimalist interface to a subset of MPI.
! The first five primitives can in theory be used to parallelize
! any program. The module hides type specification, communicators,
! explicit error handling, the need to give explicit buffer size etc.
! Also provided are a few interfaces for often used broadcast and 
! reduction operations
!
! � Helge Avlesen <avle@ii.uib.no>, para//ab
!
! primitives: (optional arguments in brackets)
!
!   subroutine start_qmpi()
!      starts the mpi subsystem. all processesors are assigned a number (myid).
!      the number of processors is numproc.
!   subroutine stop_qmpi()
!      stops the mpi subsystem
!   subroutine barrier([label])
!      syncronization point for all processors. optionally prints a label on
!      the master processor (0).
!   subroutine send(data, target [,tag])
!      send object data to processor number target, tag is an optional integer
!      that defaults to 0. (if multiple messages are exchanged between a
!      pair of processors, a unique tag MUST be used for each exhange)
!      Generally it is recommended to use the tag.  
!   subroutine receive(data, source [,tag])
!      get object data from processor source, tag is optional and as for send
!      MPI will fail if the size of the object received is different from what
!      was sent.
!  
! The rest of the routines are included for convenience, they can be
! also be implemented using the above subroutines.
!
!   subroutine broadcast(data [,root])
!      broadcast data (any type) from processor root (default=0) to all
!      other processors.
!   subroutine mbroadcast(data [,data2,data3,data4,data5,data6] [,root])
!      broadcast up to 6 scalar variables of the same type, to all processors
!      from processor root (default=0)
!   subroutine reduce(type, data [,data2,data3,data4,data5,data6] [,root] )  
!      reduce data, optionally also data2-data6, return result on all
!      processes.
!        if data is an array, reduction is done element by element on
!      each array. currently only supports 1d real4 and real8 arrays,
!      as well as 1d and 2d integer arrays.
!        the operation can currently be of type 'sum', 'mul', 'min' or
!      'max' i.e. a sum or a product. data-data6 must be of the same
!      type. if integer root is present, only return result on that
!      processor (faster)  
!   function gather(data [,root])  
!      on all processors, return an array of dimension(qmpi_num_procs)  
!      containing the value of DATA from each processor. if ROOT is
!      present, only return result on that processor (somewhat faster)
!
!   subroutine isend(data, target [,tag])
!      as send, but immediate return, finalize all pending isend/irecv's with 
!      one waitall. make sure not to touch data before the waitall is posted.
!   subroutine irecv(data, source [,tag])
!      as receive, and see note for isend.
!   subroutine waitall()
!      finalize pending isend, irecv's.
!  
!  
! Example: a program that sends a real from processor 0 to processor 1
!   use qmpi
!   real data
!   call start_mpi
!   data=myid
!   if(myid==0) call send(data, 1)
!   if(myid==1) then
!      call receive(data, 0)
!      write(*,*)'hello, I am',myid,'got ',data,'from process 0'
!   end if
!   call stop_mpi
!   end
!
! More advanced usage example: to send a derived type from 0 to 1; 
! pack it in a string (could be packed into any array), send, receive, unpack.
! 
! type(any_type) var1
! character, allocatable :: buffer(:)
! ...
! N=size(transfer(var1,(/'x'/))))   !! compute size of type once
! allocate(buffer(N))
! if(myid==0)then
!     buffer = transfer(var1,buffer)
!     call send(buffer,1)
! end if
! if(myid==1)then
!     call receive(buffer,0)
!     var1 = transfer(buffer,var1)
! end if
! ...
!

  implicit none
  
  include 'mpif.h'
!  integer, parameter :: MPI_REAL=-1, MPI_2REAL=-1, mpi_status_size=1
!  integer, parameter :: mpi_proc_null=-1

  
  integer :: qmpi_proc_num=0, qmpi_num_proc=1, ierr=0, errorcode=0, mpistatus(mpi_status_size)  
  integer :: myid=0, numproc=1
  integer :: qmpi_comm_world=0
  
  logical :: master=.true., slave=.false.

  integer :: qmpi_debug=0       ! if 0 no extra checks or debug output
  
! some kinds. could use selected_real_kind(..) for this instead of hard coding

  integer, parameter :: dp=8, dbl=dp, sp=4, long=8, short=2, dr=kind(1.0)

  integer :: mpi_float_type = MPI_REAL, mpi_2floats = MPI_2REAL
  
  interface send
     module procedure            &
          qmpi_send_real4,       &
          qmpi_send_real4_1d,    &
          qmpi_send_real4_2d,    &
          qmpi_send_real4_3d,    &
          qmpi_send_real4_4d,    &
          qmpi_send_real8,       &
          qmpi_send_real8_1d,    &
          qmpi_send_real8_2d,    &
          qmpi_send_real8_3d,    &
          qmpi_send_real8_4d,    &
          qmpi_send_integer4,    &
          qmpi_send_integer4_1d, &
          qmpi_send_integer4_2d, &
          qmpi_send_integer4_3d, &
          qmpi_send_integer4_4d, &
          qmpi_send_integer8,    &
          qmpi_send_integer8_1d, &
          qmpi_send_integer8_2d, &
          qmpi_send_integer8_3d, &
          qmpi_send_integer8_4d, &
          qmpi_send_string,      &
          qmpi_send_character_1d,&
          qmpi_send_logical
  end interface

  interface receive
     module procedure &
          qmpi_recv_real4,       &
          qmpi_recv_real4_1d,    &
          qmpi_recv_real4_2d,    &
          qmpi_recv_real4_3d,    &
          qmpi_recv_real4_4d,    &
          qmpi_recv_real8,       &
          qmpi_recv_real8_1d,    &
          qmpi_recv_real8_2d,    &
          qmpi_recv_real8_3d,    &
          qmpi_recv_real8_4d,    &
          qmpi_recv_integer4,    & 
          qmpi_recv_integer4_1d, &
          qmpi_recv_integer4_2d, &
          qmpi_recv_integer4_3d, &
          qmpi_recv_integer4_4d, &
          qmpi_recv_integer8,    &
          qmpi_recv_integer8_1d, &
          qmpi_recv_integer8_2d, &
          qmpi_recv_integer8_3d, &
          qmpi_recv_integer8_4d, &
          qmpi_recv_string,      &
          qmpi_recv_character_1d,&
          qmpi_recv_logical
  end interface

  interface reduce
     module procedure &
          qmpi_integer_reduction, &
          qmpi_integer1d_reduction, &          
          qmpi_integer2d_reduction, &
          qmpi_integer4d_reduction, &
          qmpi_integer8_reduction,&
          qmpi_real_reduction,    &
          qmpi_real8_reduction, &
          qmpi_1dreal4_reduction, &
          qmpi_2dreal4_reduction, &
          qmpi_1dreal8_reduction
  end interface

  interface gather
     module procedure &
          qmpi_gather_integer4, &
          qmpi_gather_integer8, &
          qmpi_gather_real4,    &
          qmpi_gather_real8
     ! ...
  end interface
  
  interface broadcast
     module procedure &
          qmpi_broadcast_logical,  &
          qmpi_broadcast_string,   &
          qmpi_broadcast_stringarr,&
          qmpi_broadcast_integer4, &
          qmpi_broadcast_integer4_array1d,  &
          qmpi_broadcast_integer4_array2d,  &
          qmpi_broadcast_integer8, &
          qmpi_broadcast_integer8_array1d, &
          qmpi_broadcast_integer8_array2d, &
          qmpi_broadcast_real4, &
          qmpi_broadcast_real4_array1d, &
          qmpi_broadcast_real4_array2d, &
          qmpi_broadcast_real4_array3d, &
          qmpi_broadcast_real4_array4d, &
          qmpi_broadcast_real8, &
          qmpi_broadcast_real8_array1d, &
          qmpi_broadcast_real8_array2d, &
          qmpi_broadcast_real8_array3d, &
          qmpi_broadcast_real8_array4d
  end interface

  interface mbroadcast
     module procedure &
          qmpi_broadcast_logicals, &
          qmpi_broadcast_real4s, &
          qmpi_broadcast_real8s, &
          qmpi_broadcast_integer4s, &
          qmpi_broadcast_integer8s
  end interface

  interface isend
     module procedure &
          qmpi_isend_integer4, &
          qmpi_isend_integer4_1d, &
          qmpi_isend_integer4_2d, &
          qmpi_isend_character_1d, &
          qmpi_isend_real4, &
          qmpi_isend_real8, &
          qmpi_isend_real4_1d, &
          qmpi_isend_real8_1d, &
          qmpi_isend_real4_2d, &
          qmpi_isend_real8_2d, &
          qmpi_isend_real4_3d, &
          qmpi_isend_real8_3d, &
          qmpi_isend_real4_4d, &
          qmpi_isend_real8_4d
  end interface
  
  interface irecv
     module procedure &
          qmpi_irecv_integer4_1d, &
          qmpi_irecv_real4_1d, &
          qmpi_irecv_real8_1d, &          
          qmpi_irecv_real4_2d, &
          qmpi_irecv_real8_2d, &
          qmpi_irecv_real8_3d     
  end interface

! the maximum number of dangling requests
  integer, parameter :: maxnumrequests=500
  integer:: request_list(maxnumrequests), numrequests=0
  integer:: wait_stat(mpi_status_size,2)


  integer, parameter :: max_members=50
  
  type communicator_object
     integer :: size=0                                    ! the number of processors in the comm
     integer :: comm=mpi_comm_null                        ! the communicator
     integer :: group=mpi_group_null                      ! the group
     integer :: myid=mpi_proc_null                        ! the local rank in the communicator
     integer :: global_rank(max_members)=mpi_proc_null    ! the list of global ranks
  end type communicator_object


! the host processor clock frequency to be used by the high res timer  
  integer(kind=8) :: clock_frequency = 1000000000   !  for mpi_wtime 
  
! uncomment these 2 lines to link in rdtsc, or use the function below (default)
!  integer(8) rdtsc
!  external rdtsc 
  

! BOM specific, setup2.f90: "current" data while parsing
  logical :: parsing_done=.false.
  integer :: line_number = 0
  character(len=200) :: file_name=" ", keyword=" ", current_setup_line
  public parsing_done, line_number, file_name, keyword, current_setup_line
  logical :: viewblocks=.true.
  
contains

  
  integer(8) function rdtsc()
! poor mans (read time stamp counter) rdtsc, if the assembly snippet is not available
    rdtsc = mpi_wtime()*clock_frequency
  end function rdtsc

  
  subroutine start_qmpi()
!  
! initialize the core MPI subsystem
! this routine should be called as the first statement in the program.
! MPI does not specify what happen before MPI_init and after mpi_finalize
!  
    implicit none
    real :: default_real=3.14
    
    call mpi_init(ierr)
    call mpi_comm_size(mpi_comm_world, numproc, ierr)
    call mpi_comm_rank(mpi_comm_world, myid, ierr)
    
    qmpi_comm_world = mpi_comm_world
    qmpi_num_proc=numproc
    qmpi_proc_num=myid

  
    !! call set_stack_space(hard, soft, ierr)
    master=.false.
    if(myid==0) master=.true.
    if(myid>0) slave=.true.
    
    if(master) then
        write(*,'(/,a,1x,i0,a,/)') 'MPI found ',numproc,' processors'
    end if

! make sure we call mpi routines with the right mpi types if we decide to
! compile with -r8 (-qrealsize=8 on IBM)
! this check will fail if the kind and the number of bytes used for a float
! are not the same, but most compilers follow this scheme

    if(kind(default_real)==8) then
        mpi_float_type = MPI_DOUBLE_PRECISION
        mpi_2floats = MPI_2DOUBLE_PRECISION
    end if
  end subroutine start_qmpi

  subroutine stop_qmpi()
    implicit none
    call mpi_finalize(ierr)
  end subroutine stop_qmpi

  
  subroutine barrier(label)
! makes all processes sync at this point, optionally print a label
    implicit none
    character(*), optional :: label
    call mpi_barrier(qmpi_comm_world, ierr)
    if(master.and.present(label)) write(*,'(/,a,a,/)')'B:',label
  end subroutine barrier

  
  subroutine abort(message, vals)
!
! write message, some info and exit mpi as clean as possible
!
    character(len=*), optional :: message
    real, optional :: vals(:)

    if(present(message)) then
        write(*,*) 'BOM abort called:',message,' MYID=',myid
        if(present(vals)) then
            write(*,'(a,10(f0.2,1x))') "Values= ",vals
        end if
    else
        if(master) write(*,*) 'BOM abort called, ending execution'
    end if
    call mpi_abort(qmpi_comm_world, errorcode, ierr)
    stop
  end subroutine abort


  subroutine message(text, severity, rval, rval4, rval8, lval, sval, ival, margin, advance, sync)
!
! print a message on screen, severity; 0=info, 1=warning, 2=fatal => call abort()
! if one of the real arrays RVAL, integer array IVAL,
! etc. is present, append values at end of line.  MARGIN is the number
! of spaces at the beginning of the line, by default 0.
! if called from setup2, also print current location in setup file.
! if SYNC, insert barrier before writing text.
! example:
!    call message("Flux F(i,j) out of range. F,I,J,IM,JM,KB = ", 2, rval=(/f(i,j)/), ival=(/i,j,im,jm,kb/))

    implicit none
    character(*), optional :: text
    integer, optional :: severity
    real, optional :: rval(:)
    real(4), optional :: rval4(:)
    real(8), optional :: rval8(:)
    logical, optional :: lval(:)
    character(*), optional :: sval(:)
    integer, optional :: ival(:), margin
    character(3), optional :: advance
    logical, optional :: sync
    character(10) header
    character(3) adv
    character(3) marg
    integer i
    logical val

    if(present(sync)) then
        if(sync) call barrier()
    end if
    
    val = present(rval).or.present(ival).or.present(lval).or.present(sval).or.present(rval4).or.present(rval8)
    
    header=' '
    if(present(severity)) then
        select case(severity)
        case(0)
            header='! INFO:'
        case(1)
            header='! WARNING:'
        case(2)
            header='! ERROR:'
        end select
    end if

    marg=""
    if(present(margin)) write(marg,'(i1,"x,")') margin

    adv="yes"
    if(val) adv="no"
    if(present(advance)) adv=advance
    
    if(master)then
        if(present(text).and. header.ne."") then
            write(*,'('//marg//'a,1x,a)', advance=adv) trim(header),trim(text)
        else if(present(text)) then
            write(*,'('//marg//'a)', advance=adv) trim(text)
        else if(header.ne."") then
            write(*,'(a)', advance=adv) trim(header)
        end if
    end if

    if(val.and.master) then
        ! up to 10 of each val shoould be sufficient
        if(present(rval)) write(*,'(10(1x,f0.7))',advance="no") (rval(i),i=1,size(rval))
        if(present(rval4)) write(*,'(10(1x,f0.7))',advance="no") (rval4(i),i=1,size(rval4))
        if(present(rval8)) write(*,'(10(1x,f0.7))',advance="no") (rval8(i),i=1,size(rval8))
        if(present(ival)) write(*,'(10(1x,i0))',advance="no") (ival(i),i=1,size(ival))
        if(present(lval)) write(*,'(10(1x,l1))',advance="no") (lval(i),i=1,size(lval))
        if(present(sval)) write(*,'(10(1x,a))',advance="no") (sval(i),i=1,size(sval))        
        write(*,*)
    end if
    
! BOM specific:
    if(present(severity).and..not.parsing_done)then
        if(severity==2.and.master) then
            write(*,'(2x,a,a,a,i0,a,a)')'Keyword=',trim(keyword), ', line=', line_number,', in file=', trim(file_name)
            write(*,'(2x,a,//,2x,a,/)') 'Config line in question:',trim(current_setup_line)
        end if
    end if
    
    if(present(severity)) then
        if(severity .eq. 2) call abort()
    end if
  end subroutine message

  
  subroutine waitall()
    implicit none
    if(numrequests > 0)then
      call mpi_waitall(numrequests, request_list, wait_stat, ierr)
    end if
    numrequests = 0
  end subroutine waitall

  
  subroutine qmpi_send_logical(data, target, tag)
    implicit none
    logical data
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1
    call mpi_send(data, count, mpi_logical, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_logical count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_logical

  subroutine qmpi_send_string(data, target, tag)
    implicit none
    character(*) data
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=len(data)
    call mpi_send(data, count, mpi_character, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_string count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_string

  subroutine qmpi_send_character_1d(data, target, tag)
    implicit none
    character data(:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)
    call mpi_send(data, count, mpi_character, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_character_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_character_1d
  
  subroutine qmpi_recv_character_1d(data, target, tag)
    implicit none
    character data(:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)
    call mpi_recv(data, count, mpi_character, target, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_character_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_character_1d
    
  subroutine qmpi_send_integer4(data, target, tag)
    implicit none
    integer(sp) data
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1
    call mpi_send(data, count, mpi_integer, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_integer4 count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_integer4

  subroutine qmpi_isend_integer4(data, target, tag)
    implicit none
    integer(sp) data
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_integer4, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_integer, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_integer4 count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_integer4

  subroutine qmpi_send_integer4_1d(data, target, tag)
    implicit none
    integer(sp) data(:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)
    call mpi_send(data, count, mpi_integer, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_integer4_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_integer4_1d

  subroutine qmpi_isend_integer4_1d(data, target, tag)
    implicit none
    integer(sp) data(:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_integer4_1d, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_integer, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_integer4_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_integer4_1d

  subroutine qmpi_isend_character_1d(data, target, tag)
    implicit none
    character data(:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_integer4_1d, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_character, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_character_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_character_1d
  
  subroutine qmpi_send_integer4_2d(data, target, tag)
    implicit none
    integer(sp) data(:,:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)
    call mpi_send(data, count, mpi_integer, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_integer4_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_integer4_2d

  subroutine qmpi_isend_integer4_2d(data, target, tag)
    implicit none
    integer(sp) data(:,:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_integer4_2d, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_integer, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_integer4_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_integer4_2d

  subroutine qmpi_send_integer4_3d(data, target, tag)
    implicit none
    integer(sp) data(:,:,:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)
    call mpi_send(data, count, mpi_integer, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_integer4_3d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_integer4_3d

  subroutine qmpi_send_integer4_4d(data, target, tag)
    implicit none
    integer(sp) data(:,:,:,:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)
    call mpi_send(data, count, mpi_integer, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_integer4_4d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_integer4_4d

  subroutine qmpi_send_integer8(data, target, tag)
    implicit none
    integer(long) data
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1
    call mpi_send(data, count, mpi_integer8, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_integer8 count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_integer8

  subroutine qmpi_send_integer8_1d(data, target, tag)
    implicit none
    integer(long) data(:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)
    call mpi_send(data, count, mpi_integer8, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_integer8_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_integer8_1d

  subroutine qmpi_send_integer8_2d(data, target, tag)
    implicit none
    integer(long) data(:,:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)
    call mpi_send(data, count, mpi_integer8, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_integer8_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_integer8_2d

  subroutine qmpi_send_integer8_3d(data, target, tag)
    implicit none
    integer(8) data(:,:,:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)
    call mpi_send(data, count, mpi_integer8, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_integer8_3d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_integer8_3d

  subroutine qmpi_send_integer8_4d(data, target, tag)
    implicit none
    integer(8) data(:,:,:,:)
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)
    call mpi_send(data, count, mpi_integer8, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_integer8_4d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_integer8_4d

  subroutine qmpi_send_real4(data, target, tag)
    implicit none
    real(sp) data
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1
    call mpi_send(data, count, mpi_real, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_real4 count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_real4

  subroutine qmpi_isend_real4(data, target, tag)
    implicit none
    real(sp) data
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_real4, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_real, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_real4 count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_real4

  subroutine qmpi_send_real8(data, target, tag)
    implicit none
    real(dp) data
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1
    call mpi_send(data, count, mpi_double_precision, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_real8 count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_real8

  subroutine qmpi_isend_real8(data, target, tag)
    implicit none
    real(dp) data
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_real8, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_double_precision, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_real8 count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_real8

  subroutine qmpi_send_real4_1d(data, target, tag)
    implicit none
    real(sp) data(:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)
    call mpi_send(data, count, mpi_real, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_real4_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_real4_1d

  subroutine qmpi_isend_real4_1d(data, target, tag)
    implicit none
    real(sp) data(:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_real4_1d, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_real, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_real4_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_real4_1d

  subroutine qmpi_send_real8_1d(data, target, tag)
    implicit none
    real(dp) data(:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)
    call mpi_send(data, count, mpi_double_precision, target, given_tag, qmpi_comm_world, ierr)
    if(ierr.ne.0)then
        write(*,*)'error send_real8_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_real8_1d

  subroutine qmpi_isend_real8_1d(data, target, tag)
    implicit none
    real(dp) data(:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_real8_1d, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_real8, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_real8_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_real8_1d

  subroutine qmpi_send_real4_2d(data, target, tag)
    implicit none
    real(sp) data(:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)

    call mpi_send(data, count, mpi_real, target, given_tag, qmpi_comm_world, ierr)

    if(ierr.ne.0)then
        write(*,*)'error send_real4_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_real4_2d

  subroutine qmpi_isend_real4_2d(data, target, tag)
    implicit none
    real(sp) data(:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_real4_2d, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_real, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_real4_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_real4_2d

  
  subroutine qmpi_send_real8_2d(data, target, tag)
    implicit none
    real(dp) data(:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)

    call mpi_send(data, count, mpi_double_precision, target, given_tag, qmpi_comm_world, ierr)

    if(ierr.ne.0)then
        write(*,*)'error send_real8_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_real8_2d

  subroutine qmpi_isend_real8_2d(data, target, tag)
    implicit none
    real(dp) data(:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_real8_2d, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_real8, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_real8_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_real8_2d
  
  subroutine qmpi_send_real4_3d(data, target, tag)
    implicit none
    real(sp) data(:,:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)

    call mpi_send(data, count, mpi_real, target, given_tag, qmpi_comm_world, ierr)

    if(ierr.ne.0)then
        write(*,*)'error send_real4_3d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_real4_3d

  subroutine qmpi_isend_real4_3d(data, target, tag)
    implicit none
    real(sp) data(:,:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_real4_3d, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_real, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_real4_3d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_real4_3d

  
  subroutine qmpi_isend_real4_4d(data, target, tag)
    implicit none
    real(sp) data(:,:,:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_real4_3d, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_real, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_real4_4d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_real4_4d
  
  
  subroutine qmpi_send_real8_3d(data, target, tag)
    implicit none
    real(dp) data(:,:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)

    call mpi_send(data, count, mpi_double_precision, target, given_tag, qmpi_comm_world, ierr)

    if(ierr.ne.0)then
        write(*,*)'error send_real8_3d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_real8_3d

  subroutine qmpi_isend_real8_3d(data, target, tag)
    implicit none
    real(dp) data(:,:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_real8_3d, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_real8, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_real8_3d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_real8_3d

  
  subroutine qmpi_send_real4_4d(data, target, tag)
    implicit none
    real(sp) data(:,:,:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)

    call mpi_send(data, count, mpi_real, target, given_tag, qmpi_comm_world, ierr)

    if(ierr.ne.0)then
        write(*,*)'error send_real4_4d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_real4_4d

  
  subroutine qmpi_isend_real8_4d(data, target, tag)
    implicit none
    real(dp) data(:,:,:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error isend_real8_3d, too many isend/irecv requests on this processor'
        call abort()
    end if
    
    call mpi_isend(data, count, mpi_real8, target, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error isend_real8_4d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_isend_real8_4d

  
  subroutine qmpi_send_real8_4d(data, target, tag)
    implicit none
    real(dp) data(:,:,:,:)
    integer target
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)

    call mpi_send(data, count, mpi_double_precision, target, given_tag, qmpi_comm_world, ierr)

    if(ierr.ne.0)then
        write(*,*)'error send_real8_4d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_send_real8_4d

  subroutine qmpi_recv_integer4(data, source, tag)
    implicit none
    integer(sp) data
    integer source, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1
    call mpi_recv(data, count, mpi_integer, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_integer4_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_integer4

  subroutine qmpi_recv_integer4_1d(data, source, tag)
    implicit none
    integer(sp) data(:)
    integer source, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)
    call mpi_recv(data, count, mpi_integer, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_integer4_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_integer4_1d

  
  subroutine qmpi_irecv_integer4_1d(data, source, tag)
    implicit none
    integer(sp) data(:)
    integer source, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error irecv_integer4_1d, too many isend/irecv requests on this processor'
        call abort()
    end if
   
    call mpi_irecv(data, count, mpi_integer, source, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    if(ierr.ne.0)then
        write(*,*)'error irecv_integer4_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_irecv_integer4_1d

  
  subroutine qmpi_recv_integer4_2d(data, source, tag)
    implicit none
    integer(sp) data(:,:)
    integer source, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)
    call mpi_recv(data, count, mpi_integer, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_integer4_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_integer4_2d

  subroutine qmpi_recv_integer4_3d(data, source, tag)
    implicit none
    integer(sp) data(:,:,:)
    integer source, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)
    call mpi_recv(data, count, mpi_integer, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_integer4_3d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_integer4_3d

  subroutine qmpi_recv_integer4_4d(data, source, tag)
    implicit none
    integer(sp) data(:,:,:,:)
    integer source, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)
    call mpi_recv(data, count, mpi_integer, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_integer4_4d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_integer4_4d

  subroutine qmpi_recv_integer8(data, source, tag)
    implicit none
    integer(long) data
    integer source, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1
    call mpi_recv(data, count, mpi_integer8, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_integer8 count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_integer8

  subroutine qmpi_recv_integer8_1d(data, source, tag)
    implicit none
    integer(long) data(:)
    integer source, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)
    call mpi_recv(data, count, mpi_integer8, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_integer8_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_integer8_1d

  subroutine qmpi_recv_integer8_2d(data, source, tag)
    implicit none
    integer(long) data(:,:)
    integer source, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)
    call mpi_recv(data, count, mpi_integer8, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_integer8_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_integer8_2d

  subroutine qmpi_recv_integer8_3d(data, source, tag)
    implicit none
    integer(8) data(:,:,:)
    integer source, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)
    call mpi_recv(data, count, mpi_integer8, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_integer8_3d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_integer8_3d

  subroutine qmpi_recv_integer8_4d(data, source, tag)
    implicit none
    integer(8) data(:,:,:,:)
    integer source, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)
    call mpi_recv(data, count, mpi_integer8, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_integer8_4d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_integer8_4d

  subroutine qmpi_recv_real4(data, source, tag)
    implicit none
    real(sp) data
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1
    call mpi_recv(data, count, mpi_real, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_real4 count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_real4

  subroutine qmpi_recv_real8(data, source, tag)
    implicit none
    real(dp) data
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1
    call mpi_recv(data, count, mpi_double_precision, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_real8 count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_real8

  subroutine qmpi_recv_real4_1d(data, source, tag)
    implicit none
    real(sp) data(:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)
    call mpi_recv(data, count, mpi_real, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_real4_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_real4_1d

  
  subroutine qmpi_irecv_real4_1d(data, source, tag)
    implicit none
    real(sp) data(:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error irecv_integer4_1d, too many isend/irecv requests on this processor'
        call abort()
    end if
       
    call mpi_irecv(data, count, mpi_real, source, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    
    if(ierr.ne.0)then
        write(*,*)'error irecv_real4_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_irecv_real4_1d

  
  subroutine qmpi_irecv_real4_2d(data, source, tag)
    implicit none
    real(sp) data(:,:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error irecv_real4_2d, too many isend/irecv requests on this processor'
        call abort()
    end if
       
    call mpi_irecv(data, count, mpi_real, source, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    
    if(ierr.ne.0)then
        write(*,*)'error irecv_real4_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_irecv_real4_2d

  
  subroutine qmpi_recv_real8_1d(data, source, tag)
    implicit none
    real(dp) data(:)
    integer source
    integer, optional :: tag
    integer count, given_tag
    
    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)
    call mpi_recv(data, count, mpi_double_precision, source, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_real8_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_real8_1d

  subroutine qmpi_irecv_real8_1d(data, source, tag)
    implicit none
    real(dp) data(:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error irecv_real8_1d, too many isend/irecv requests on this processor'
        call abort()
    end if
       
    call mpi_irecv(data, count, mpi_real8, source, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    
    if(ierr.ne.0)then
        write(*,*)'error irecv_real8_1d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_irecv_real8_1d
  subroutine qmpi_irecv_real8_2d(data, source, tag)
    implicit none
    real(dp) data(:,:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error irecv_real8_2d, too many isend/irecv requests on this processor'
        call abort()
    end if
       
    call mpi_irecv(data, count, mpi_real8, source, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    
    if(ierr.ne.0)then
        write(*,*)'error irecv_real8_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_irecv_real8_2d

  subroutine qmpi_irecv_real8_3d(data, source, tag)
    implicit none
    real(dp) data(:,:,:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)

    numrequests=numrequests + 1

    if(numrequests > maxnumrequests) then
        write(*,*)myid,'error irecv_real8_3d, too many isend/irecv requests on this processor'
        call abort()
    end if
       
    call mpi_irecv(data, count, mpi_real8, source, given_tag, qmpi_comm_world, request_list(numrequests), ierr)
    
    if(ierr.ne.0)then
        write(*,*)'error irecv_real8_3d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_irecv_real8_3d

  
  subroutine qmpi_recv_real4_2d(data, source, tag)
    implicit none
    real(sp) data(:,:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)

    call mpi_recv(data, count, mpi_real, source, given_tag, qmpi_comm_world, mpistatus, ierr)

    if(ierr.ne.0)then
        write(*,*)'error recv_real4_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_real4_2d

  subroutine qmpi_recv_real8_2d(data, source, tag)
    implicit none
    real(dp) data(:,:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)

    call mpi_recv(data, count, mpi_double_precision, source, given_tag, qmpi_comm_world, mpistatus, ierr)

    if(ierr.ne.0)then
        write(*,*)'error recv_real8_2d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_real8_2d

  subroutine qmpi_recv_real4_3d(data, source, tag)
    implicit none
    real(sp) data(:,:,:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)

    call mpi_recv(data, count, mpi_real, source, given_tag, qmpi_comm_world, mpistatus, ierr)

    if(ierr.ne.0)then
        write(*,*)'error recv_real4_3d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_real4_3d

  subroutine qmpi_recv_real8_3d(data, source, tag)
    implicit none
    real(dp) data(:,:,:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)

    call mpi_recv(data, count, mpi_double_precision, source, given_tag, qmpi_comm_world, mpistatus, ierr)

    if(ierr.ne.0)then
        write(*,*)'error recv_real8_3d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_real8_3d

  subroutine qmpi_recv_real4_4d(data, source, tag)
    implicit none
    real(sp) data(:,:,:,:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)

    call mpi_recv(data, count, mpi_real, source, given_tag, qmpi_comm_world, mpistatus, ierr)

    if(ierr.ne.0)then
        write(*,*)'error recv_real4_4d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_real4_4d

  subroutine qmpi_recv_real8_4d(data, source, tag)
    implicit none
    real(dp) data(:,:,:,:)
    integer source
    integer, optional :: tag
    integer count, given_tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)

    call mpi_recv(data, count, mpi_double_precision, source, given_tag, qmpi_comm_world, mpistatus, ierr)

    if(ierr.ne.0)then
        write(*,*)'error recv_real8_4d count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_real8_4d

  subroutine qmpi_recv_logical(data, target, tag)
    implicit none
    logical data
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=1
    call mpi_recv(data, count, mpi_logical, target, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_logical count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_logical

  subroutine qmpi_recv_string(data, target, tag)
    implicit none
    character(*) data
    integer target, count, given_tag
    integer, optional :: tag

    if(numproc == 1) return
    given_tag=0
    if(present(tag)) given_tag=tag
    count=len(data)
    call mpi_recv(data, count, mpi_character, target, given_tag, qmpi_comm_world, mpistatus, ierr)
    if(ierr.ne.0)then
        write(*,*)'error recv_string count=',count,'tag=',given_tag
        stop
    end if
  end subroutine qmpi_recv_string

  subroutine qmpi_broadcast_string(string,root,comm)
!
! send string out to all processes. if not given
! process 0 will be used as the sender - root otherwise.
!
    implicit none
    character(len=*) string
    integer, optional :: root,comm
    integer count,boss,communicator
    if(numproc == 1) return
    
    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return

    count=len(string)

    boss=0
    if(present(root)) then
        boss=root
    end if

    call mpi_bcast(string , count, mpi_character, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_string

  subroutine qmpi_broadcast_stringarr(data,root,comm)
    implicit none
    character(len=*) data(:)
    integer, optional :: root,comm
    integer count, boss,communicator

    if(numproc == 1) return
    
    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=len(data(1))*size(data)

    boss=0
    if(present(root)) then
        boss=root
    end if

    call mpi_bcast(data, count, mpi_character, boss, communicator ,ierr)
  end subroutine qmpi_broadcast_stringarr

  subroutine qmpi_broadcast_real4(data,root,comm)
    implicit none
    real(4) data
    integer, optional :: root,comm
    integer count,boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=1
    boss=0
    if(present(root)) boss=root
    call mpi_bcast(data , count, mpi_real, boss, communicator, ierr)
  end subroutine qmpi_broadcast_real4

  subroutine qmpi_broadcast_real8(data,root,comm)
    implicit none
    real(8) data
    integer, optional :: root,comm
    integer count,boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=1
    boss=0
    if(present(root)) boss=root
    call mpi_bcast(data , count, mpi_double_precision, boss, communicator, ierr)
  end subroutine qmpi_broadcast_real8

  subroutine qmpi_broadcast_integer4(data,root,comm)
    implicit none
    integer(4) data
    integer, optional :: root,comm
    integer count,boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=1
    boss=0
    if(present(root)) boss=root
    call mpi_bcast(data , count, mpi_integer, boss, communicator, ierr)
  end subroutine qmpi_broadcast_integer4

  subroutine qmpi_broadcast_integer8(data,root,comm)
    implicit none
    integer(8) data
    integer, optional :: root,comm
    integer count,boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=1
    boss=0
    if(present(root)) boss=root
    call mpi_bcast(data , count, mpi_integer8, boss, communicator, ierr)
  end subroutine qmpi_broadcast_integer8

  subroutine qmpi_broadcast_logical(data, root, comm)
    implicit none
    logical data
    integer, optional :: root,comm
    integer count,boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=1
    boss=0
    if(present(root)) boss=root
    call mpi_bcast(data , count, mpi_logical, boss, communicator, ierr)
  end subroutine qmpi_broadcast_logical


  subroutine qmpi_broadcast_integer4_array1d(data,root,comm)
    implicit none
    integer(sp) data(:)
    integer, optional :: root,comm
    integer count,boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data , count, mpi_integer, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_integer4_array1d
  
  subroutine qmpi_broadcast_integer8_array1d(data,root,comm)
    implicit none
    integer(long) data(:)
    integer, optional :: root,comm
    integer count,boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data , count, mpi_integer8, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_integer8_array1d

  subroutine qmpi_broadcast_integer4_array2d(data,root,comm)
    implicit none
    integer(sp) data(:,:)
    integer, optional :: root,comm
    integer count,boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data,1)*size(data,2)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data , count, mpi_integer, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_integer4_array2d
    
  subroutine qmpi_broadcast_integer8_array2d(data,root,comm)
    implicit none
    integer(long) data(:,:)
    integer, optional :: root,comm
    integer count,boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data,1)*size(data,2)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data , count, mpi_integer8, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_integer8_array2d

  subroutine qmpi_broadcast_real4_array1d(data,root,comm)
    implicit none
    real(sp) data(:)
    integer, optional :: root,comm
    integer count, boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data , count, mpi_real, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_real4_array1d

  subroutine qmpi_broadcast_real8_array1d(data,root,comm)
    implicit none
    real(dp) data(:)
    integer, optional :: root,comm
    integer count, boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data , count, mpi_double_precision, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_real8_array1d

  subroutine qmpi_broadcast_real4_array2d(data,root,comm)
    implicit none
    real(sp) data(:,:)
    integer, optional :: root,comm
    integer count, boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data,1)*size(data,2)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data, count, mpi_real, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_real4_array2d

  subroutine qmpi_broadcast_real8_array2d(data,root,comm)
    implicit none
    real(dp) data(:,:)
    integer, optional :: root,comm
    integer count, boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data,1)*size(data,2)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data, count, mpi_double_precision, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_real8_array2d

  subroutine qmpi_broadcast_real4_array3d(data,root,comm)
    implicit none
    real(sp) data(:,:,:)
    integer, optional :: root,comm
    integer count, boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data,1)*size(data,2)*size(data,3)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data , count, mpi_real, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_real4_array3d

  subroutine qmpi_broadcast_real8_array3d(data,root,comm)
    implicit none
    real(dp) data(:,:,:)
    integer, optional :: root,comm
    integer count, boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data,1)*size(data,2)*size(data,3)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data , count, mpi_double_precision, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_real8_array3d

  subroutine qmpi_broadcast_real4_array4d(data,root,comm)
    implicit none
    real(sp) data(:,:,:,:)
    integer, optional :: root,comm
    integer count, boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data , count, mpi_real, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_real4_array4d

  subroutine qmpi_broadcast_real8_array4d(data,root,comm)
    implicit none
    real(dp) data(:,:,:,:)
    integer, optional :: root,comm
    integer count, boss,communicator

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=size(data,1)*size(data,2)*size(data,3)*size(data,4)
    boss=0
    if(present(root)) then
        boss=root
    end if
    call mpi_bcast(data , count, mpi_double_precision, boss, communicator  ,ierr)
  end subroutine qmpi_broadcast_real8_array4d

  
  function qmpi_gather_integer4(data,root,comm)
    implicit none
    integer(4) data, qmpi_gather_integer4(qmpi_num_proc)
    integer, optional :: root,comm
    integer count,boss,communicator

    qmpi_gather_integer4=data
    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=1
    boss=0
    if(present(root)) then
        boss=root
        call mpi_gather(data,                 count, mpi_integer, &
                        qmpi_gather_integer4, count, mpi_integer, &
                        boss, communicator,   ierr)
    else
        call mpi_allgather(data,                 count, mpi_integer, &
                           qmpi_gather_integer4, count, mpi_integer, &
                           communicator, ierr)        
    end if
  end function qmpi_gather_integer4

  function qmpi_gather_integer8(data,root,comm)
    implicit none
    integer(8) data, qmpi_gather_integer8(qmpi_num_proc)
    integer, optional :: root,comm
    integer count,boss,communicator

    qmpi_gather_integer8=data
    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=1
    boss=0
    if(present(root)) then
        boss=root
        call mpi_gather(data,                 count, mpi_integer8, &
                        qmpi_gather_integer8, count, mpi_integer8, &
                        boss, communicator, ierr)
    else
        call mpi_allgather(data,                 count, mpi_integer8, &
                           qmpi_gather_integer8, count, mpi_integer8, &
                           communicator, ierr)        
    end if
  end function qmpi_gather_integer8

  function qmpi_gather_real4(data,root,comm)
    implicit none
    real(4) data, qmpi_gather_real4(qmpi_num_proc)
    integer, optional :: root,comm
    integer count,boss,communicator

    qmpi_gather_real4=data
    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=1
    boss=0
    if(present(root)) then
        boss=root
        call mpi_gather(data,                 count, mpi_real, &
                        qmpi_gather_real4,    count, mpi_real, &
                        boss, communicator, ierr)
    else
        call mpi_allgather(data,              count, mpi_real, &
                           qmpi_gather_real4, count, mpi_real, &
                           communicator, ierr)        
    end if
  end function qmpi_gather_real4

  function qmpi_gather_real8(data,root,comm)
    implicit none
    real(8) data, qmpi_gather_real8(qmpi_num_proc)
    integer, optional :: root,comm
    integer count,boss,communicator
    
    qmpi_gather_real8=data
    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=1
    boss=0
    if(present(root)) then
        boss=root
        call mpi_gather(data,                 count, mpi_real8, &
                        qmpi_gather_real8,    count, mpi_real8, &
                        boss, communicator, ierr)
    else
        call mpi_allgather(data,              count, mpi_real8, &
                           qmpi_gather_real8, count, mpi_real8, &
                           communicator, ierr)        
    end if
  end function qmpi_gather_real8
  
  
  subroutine qmpi_broadcast_real4s(a,b,c,d,e,f,root,comm)
!
! send a,b,c,d,e,f out to all processes. if not given
! process 0 will be used as the sender - root otherwise.
!
    implicit none
    real(sp) a
    real(sp), optional :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    real(sp) rbuff(6)

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    count=0   ;  boss=0
    if(present(root)) then
        boss=root
    end if
!    if(present(a)) then
        count=count+1
        rbuff(count)=a
!    end if
    if(present(b)) then
        count=count+1
        rbuff(count)=b
    end if
    if(present(c)) then
        count=count+1
        rbuff(count)=c
    end if
    if(present(d)) then
        count=count+1
        rbuff(count)=d
    end if
    if(present(e)) then
        count=count+1
        rbuff(count)=e
    end if
    if(present(f)) then
        count=count+1
        rbuff(count)=f
    end if

    call mpi_bcast(rbuff , count, mpi_real, boss, communicator  ,ierr)

    count=1
    a=rbuff(count)
    if(present(b)) then
        count=count+1
        b=rbuff(count)
    end if
    if(present(c)) then
        count=count+1
        c=rbuff(count)
    end if
    if(present(d)) then
        count=count+1
        d=rbuff(count)
    end if
    if(present(e)) then
        count=count+1
        e=rbuff(count)
    end if
    if(present(f)) then
        count=count+1
        f=rbuff(count)
    end if
  end subroutine qmpi_broadcast_real4s

  subroutine qmpi_broadcast_real8s(a,b,c,d,e,f,root,comm)
!
! send a,b,c,d,e,f out to all processes. if not given
! process 0 will be used as the sender - root otherwise.
!
    implicit none
    real(dp) a
    real(dp), optional :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    real(kind=8) rbuff(6)

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    boss=0
    if(present(root)) then
        boss=root
    end if
    count=1
    rbuff(count)=a
    if(present(b)) then
        count=count+1
        rbuff(count)=b
    end if
    if(present(c)) then
        count=count+1
        rbuff(count)=c
    end if
    if(present(d)) then
        count=count+1
        rbuff(count)=d
    end if
    if(present(e)) then
        count=count+1
        rbuff(count)=e
    end if
    if(present(f)) then
        count=count+1
        rbuff(count)=f
    end if

    call mpi_bcast(rbuff , count, mpi_double_precision, boss, communicator  ,ierr)

    count=1
    a=rbuff(count)
    if(present(b)) then
        count=count+1
        b=rbuff(count)
    end if
    if(present(c)) then
        count=count+1
        c=rbuff(count)
    end if
    if(present(d)) then
        count=count+1
        d=rbuff(count)
    end if
    if(present(e)) then
        count=count+1
        e=rbuff(count)
    end if
    if(present(f)) then
        count=count+1
        f=rbuff(count)
    end if
  end subroutine qmpi_broadcast_real8s
  
  subroutine qmpi_broadcast_logicals(a,b,c,d,e,f,root,comm)
!
! send a,b,c,d,e,f out to all processes. if not given
! process 0 will be used as the sender - root otherwise.
!
    implicit none
    logical a
    logical, optional :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    logical lbuff(6)

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    boss=0
    if(present(root)) then
        boss=root
    end if

    count=1
    lbuff(count)=a
    if(present(b)) then
        count=count+1
        lbuff(count)=b
    end if
    if(present(c)) then
        count=count+1
        lbuff(count)=c
    end if
    if(present(d)) then
        count=count+1
        lbuff(count)=d
    end if
    if(present(e)) then
        count=count+1
        lbuff(count)=e
    end if
    if(present(f)) then
        count=count+1
        lbuff(count)=f
    end if

    call mpi_bcast(lbuff , count, mpi_logical, boss, communicator  ,ierr)

    count=1
    a=lbuff(count)

    if(present(b)) then
        count=count+1
        b=lbuff(count)
    end if
    if(present(c)) then
        count=count+1
        c=lbuff(count)
    end if
    if(present(d)) then
        count=count+1
        d=lbuff(count)
    end if
    if(present(e)) then
        count=count+1
        e=lbuff(count)
    end if
    if(present(f)) then
        count=count+1
        f=lbuff(count)
    end if
  end subroutine qmpi_broadcast_logicals

  subroutine qmpi_broadcast_integer4s(a,b,c,d,e,f,root,comm)
!
! send a,b,c,d,e,f out to all processes. if not given
! process 0 will be used as the sender - root otherwise.
!
    implicit none
    integer(sp) a
    integer(sp), optional :: b,c,d,e,f,root,comm
    integer count,boss,communicator
    integer ibuff(6)

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    boss=0
    if(present(root)) then
        boss=root
    end if

    count=1
!    if(present(a)) then
!        count=count+1
        ibuff(count)=a
!    end if
    if(present(b)) then
        count=count+1
        ibuff(count)=b
    end if
    if(present(c)) then
        count=count+1
        ibuff(count)=c
    end if
    if(present(d)) then
        count=count+1
        ibuff(count)=d
    end if
    if(present(e)) then
        count=count+1
        ibuff(count)=e
    end if
    if(present(f)) then
        count=count+1
        ibuff(count)=f
    end if

    call mpi_bcast(ibuff , count, mpi_integer, boss, communicator  ,ierr)

    count=1
    a=ibuff(count)

    if(present(b)) then
        count=count+1
        b=ibuff(count)
    end if
    if(present(c)) then
        count=count+1
        c=ibuff(count)
    end if
    if(present(d)) then
        count=count+1
        d=ibuff(count)
    end if
    if(present(e)) then
        count=count+1
        e=ibuff(count)
    end if
    if(present(f)) then
        count=count+1
        f=ibuff(count)
    end if
  end subroutine qmpi_broadcast_integer4s

  subroutine qmpi_broadcast_integer8s(a,b,c,d,e,f,root,comm)
!
! send a,b,c,d,e,f out to all processes. if not given
! process 0 will be used as the sender - root otherwise.
!
    implicit none
    integer(long) a
    integer(long), optional :: b,c,d,e,f,root,comm
    integer count,boss,communicator
    integer ibuff(6)

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    boss=0
    if(present(root)) then
        boss=root
    end if

    count=1
!    if(present(a)) then
!        count=count+1
        ibuff(count)=a
!    end if
    if(present(b)) then
        count=count+1
        ibuff(count)=b
    end if
    if(present(c)) then
        count=count+1
        ibuff(count)=c
    end if
    if(present(d)) then
        count=count+1
        ibuff(count)=d
    end if
    if(present(e)) then
        count=count+1
        ibuff(count)=e
    end if
    if(present(f)) then
        count=count+1
        ibuff(count)=f
    end if

    call mpi_bcast(ibuff , count, mpi_integer8, boss, communicator  ,ierr)

    count=1
    a=ibuff(count)

    if(present(b)) then
        count=count+1
        b=ibuff(count)
    end if
    if(present(c)) then
        count=count+1
        c=ibuff(count)
    end if
    if(present(d)) then
        count=count+1
        d=ibuff(count)
    end if
    if(present(e)) then
        count=count+1
        e=ibuff(count)
    end if
    if(present(f)) then
        count=count+1
        f=ibuff(count)
    end if
  end subroutine qmpi_broadcast_integer8s

  subroutine qmpi_real_reduction(type,a,b,c,d,e,f,root,comm)
!
! perform a reduction of 'type' on each of the given arguments a - f. 
! if type is:
!  'sum': for each argument, return the sum of the argument over all processors
!  'mul': the product
!  'min': the minimum value
!  'max': the maximum value
! root is an optional argument, if given only return the result on that processor (reduce)
!  the default is to return the result on all processors (allreduce)
!    
    implicit none
    character(3) type
    real(sp) a
    real(sp), optional, intent(inout) :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    integer, parameter :: dp=8
    real(dp) rbuff(6),globrbuff(6)

    if( trim(type).ne.'sum' .and. trim(type).ne.'mul' .and. trim(type).ne.'min' .and. trim(type).ne.'max')then
        write(*,*)'qmpi.f90 reduce error: reduction of type ',type,'not supported'
        stop
    end if
    
    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    boss=0
    if(present(root)) boss=root

    globrbuff(:)=0.0
    count=0
!    if(present(a)) then
        count=count+1
        rbuff(count)=real(a,dp)
!    end if
    if(present(b)) then
        count=count+1
        rbuff(count)=real(b,dp)
    end if
    if(present(c)) then
        count=count+1
        rbuff(count)=real(c,dp)
    end if
    if(present(d)) then
        count=count+1
        rbuff(count)=real(d,dp)
    end if
    if(present(e)) then
        count=count+1
        rbuff(count)=real(e,dp)
    end if
    if(present(f)) then
        count=count+1
        rbuff(count)=real(f,dp)
    end if

    select case(type)
    case('sum')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count,mpi_double_precision,mpi_sum,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count,mpi_double_precision,mpi_sum,communicator,ierr)
        end if
    case('mul')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count,mpi_double_precision,mpi_prod,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count,mpi_double_precision,mpi_prod,communicator,ierr)
        end if
    case('min')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count,mpi_double_precision,mpi_min,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count,mpi_double_precision,mpi_min,communicator,ierr)
        end if
    case('max')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count,mpi_double_precision,mpi_max,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count,mpi_double_precision,mpi_max,communicator,ierr)
        end if
    end select

    count=0
!    if(present(a)) then
        count=count+1
        a=globrbuff(count)
!    end if
    if(present(b)) then
        count=count+1
        b=globrbuff(count)
    end if
    if(present(c)) then
        count=count+1
        c=globrbuff(count)
    end if
    if(present(d)) then
        count=count+1
        d=globrbuff(count)
    end if
    if(present(e)) then
        count=count+1
        e=globrbuff(count)
    end if
    if(present(f)) then
        count=count+1
        f=globrbuff(count)
    end if
  end subroutine qmpi_real_reduction

  subroutine qmpi_real8_reduction(type,a,b,c,d,e,f,root,comm)
!
! perform a reduction of 'type' on each of the given arguments a - f. 
! if type is:
!  'sum': for each argument, return the sum of the argument over all processors
!  'mul': the product
!  'min': the minimum value
!  'max': the maximum value
! root is an optional argument, if given only return the result on that processor (reduce)
!  the default is to return the result on all processors (allreduce)
!    
    implicit none
    integer, parameter :: dp=8
    character(3) type
    real(dp) a
    real(dp), optional, intent(inout) :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    real(dp) rbuff(6),globrbuff(6)

    if( trim(type).ne.'sum' .and. trim(type).ne.'mul' .and. trim(type).ne.'min' .and. trim(type).ne.'max')then
        write(*,*)'qmpi.f90 reduce error: reduction of type ',type,'not supported'
        stop
    end if

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    boss=0
    if(present(root))boss=root

    globrbuff(:)=0.0
    count=0
!    if(present(a)) then
        count=count+1
        rbuff(count)=a
!    end if
    if(present(b)) then
        count=count+1
        rbuff(count)=b
    end if
    if(present(c)) then
        count=count+1
        rbuff(count)=c
    end if
    if(present(d)) then
        count=count+1
        rbuff(count)=d
    end if
    if(present(e)) then
        count=count+1
        rbuff(count)=e
    end if
    if(present(f)) then
        count=count+1
        rbuff(count)=f
    end if
    
    select case(type)
    case('sum')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count,mpi_double_precision,mpi_sum,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count,mpi_double_precision,mpi_sum,communicator,ierr)
        end if
    case('mul')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count,mpi_double_precision,mpi_prod,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count,mpi_double_precision,mpi_prod,communicator,ierr)
        end if
    case('min')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count,mpi_double_precision,mpi_min,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count,mpi_double_precision,mpi_min,communicator,ierr)
        end if
    case('max')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count,mpi_double_precision,mpi_max,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count,mpi_double_precision,mpi_max,communicator,ierr)
        end if
    end select

    count=0
!    if(present(a)) then
        count=count+1
        a=globrbuff(count)
!    end if
    if(present(b)) then
        count=count+1
        b=globrbuff(count)
    end if
    if(present(c)) then
        count=count+1
        c=globrbuff(count)
    end if
    if(present(d)) then
        count=count+1
        d=globrbuff(count)
    end if
    if(present(e)) then
        count=count+1
        e=globrbuff(count)
    end if
    if(present(f)) then
        count=count+1
        f=globrbuff(count)
    end if
  end subroutine qmpi_real8_reduction

  subroutine qmpi_1dreal4_reduction(type,a,b,c,d,e,f,root,comm)
!
! perform a reduction of 'type' on each of the given arguments a - f. 
! if type is:
!  'sum': for each argument, return the sum of the argument over all processors
!  'mul': the product
!  'min': the minimum value
!  'max': the maximum value
! root is an optional argument, if given only return the result on that processor (reduce)
!  the default is to return the result on all processors (allreduce)
!    
    implicit none
    character(3) type
    real(sp) a(:)
    real(sp), dimension(:), optional, intent(inout) :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    real(sp), allocatable :: rbuff(:),globrbuff(:)
    integer sizea,sizeb,sizec,sized,sizee,sizef,maxsize
    
    if( trim(type).ne.'sum' .and. trim(type).ne.'mul' .and. trim(type).ne.'min' .and. trim(type).ne.'max')then
        write(*,*)'qmpi.f90 reduce error: reduction of type ',type,'not supported'
        stop
    end if

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return

    sizea=size(a)
    sizeb=0 ; sizec=0 ; sized=0 ; sizee=0 ; sizef=0
    if(present(b)) sizeb=size(b)
    if(present(c)) sizec=size(c)
    if(present(d)) sized=size(d)
    if(present(e)) sizee=size(e)
    if(present(f)) sizef=size(f)
    maxsize = maxval((/sizea,sizeb,sizec,sized,sizee,sizef/))
    allocate(rbuff(6*maxsize), globrbuff(6*maxsize))
    
    boss=0
    if(present(root))boss=root

    globrbuff(:)=0
    
! count should from now on be the beginning of the "next" argument
! array in rbuff:
    count=1
    
    rbuff(count:count+sizea-1) = a
    count = count + sizea
    
    if(present(b)) then
        rbuff(count:count+sizeb-1)=b
        count=count+sizeb
    end if
    if(present(c)) then
        rbuff(count:count+sizec-1)=c
        count=count+sizec
    end if
    if(present(d)) then
        rbuff(count:count+sized-1)=d
        count=count+sized
    end if
    if(present(e)) then
        rbuff(count:count+sizee-1)=e
        count=count+sizee
    end if
    if(present(f)) then
        rbuff(count:count+sizef-1)=f
        count=count+sizef
    end if
    
! new role for count: now as the last item of our big array to be reduced
    count = count -1
    
    select case(type)
    case('sum')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL4, mpi_sum,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL4, mpi_sum,communicator,ierr)
        end if
    case('mul')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL4, mpi_prod,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL4, mpi_prod,communicator,ierr)
        end if
    case('min')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL4, mpi_min,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL4, mpi_min,communicator,ierr)
        end if
    case('max')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL4, mpi_max,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL4, mpi_max,communicator,ierr)
        end if
    end select

! then copy back results:    
    count=1
    a = globrbuff(count:count+sizea-1)
    count = count + sizea
    
    if(present(b)) then
        b=globrbuff(count:count+sizeb-1)
        count=count+sizeb
    end if
    if(present(c)) then
        c=globrbuff(count:count+sizec-1)
        count=count+sizec
    end if
    if(present(d)) then
        d=globrbuff(count:count+sized-1)
        count=count+sized
    end if
    if(present(e)) then
        e=globrbuff(count:count+sizee-1)
        count=count+sizee
    end if
    if(present(f)) then
        f=globrbuff(count:count+sizef-1)
        count=count+sizef
    end if
    deallocate(rbuff, globrbuff)    
  end subroutine qmpi_1dreal4_reduction
  
  subroutine qmpi_1dreal8_reduction(type,a,b,c,d,e,f,root,comm)
!
! perform a reduction of 'type' on each of the given arguments a - f. 
! if type is:
!  'sum': for each argument, return the sum of the argument over all processors
!  'mul': the product
!  'min': the minimum value
!  'max': the maximum value
! root is an optional argument, if given only return the result on that processor (reduce)
!  the default is to return the result on all processors (allreduce)
!    
    implicit none
    character(3) type
    real(dp) a(:)
    real(dp), dimension(:), optional, intent(inout) :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    real(dp), allocatable :: rbuff(:),globrbuff(:)

    integer sizea,sizeb,sizec,sized,sizee,sizef,maxsize
    
    if( trim(type).ne.'sum' .and. trim(type).ne.'mul' .and. trim(type).ne.'min' .and. trim(type).ne.'max')then
        write(*,*)'qmpi.f90 reduce error: reduction of type ',type,'not supported'
        stop
    end if

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    sizea=size(a)
    sizeb=0 ; sizec=0 ; sized=0 ; sizee=0 ; sizef=0
    if(present(b)) sizeb=size(b)
    if(present(c)) sizec=size(c)
    if(present(d)) sized=size(d)
    if(present(e)) sizee=size(e)
    if(present(f)) sizef=size(f)
    maxsize = maxval((/sizea,sizeb,sizec,sized,sizee,sizef/))
    allocate(rbuff(6*maxsize), globrbuff(6*maxsize))
    
    boss=0
    if(present(root))boss=root

    globrbuff(:)=0
    
! count should from now on be the beginning of the "next" argument
! array in rbuff:
    count=1
    
    rbuff(count:count+sizea-1) = a
    count = count + sizea
    
    if(present(b)) then
        rbuff(count:count+sizeb-1)=b
        count=count+sizeb
    end if
    if(present(c)) then
        rbuff(count:count+sizec-1)=c
        count=count+sizec
    end if
    if(present(d)) then
        rbuff(count:count+sized-1)=d
        count=count+sized
    end if
    if(present(e)) then
        rbuff(count:count+sizee-1)=e
        count=count+sizee
    end if
    if(present(f)) then
        rbuff(count:count+sizef-1)=f
        count=count+sizef
    end if
    
! new role for count: now as the last item of our big array to be reduced
    count = count -1
    
    select case(type)
    case('sum')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL8, mpi_sum,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL8, mpi_sum,communicator,ierr)
        end if
    case('mul')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL8, mpi_prod,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL8, mpi_prod,communicator,ierr)
        end if
    case('min')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL8, mpi_min,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL8, mpi_min,communicator,ierr)
        end if
    case('max')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL8, mpi_max,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL8, mpi_max,communicator,ierr)
        end if
    end select

! then copy back results:    
    count=1
    a = globrbuff(count:count+sizea-1)
    count = count + sizea
    
    if(present(b)) then
        b=globrbuff(count:count+sizeb-1)
        count=count+sizeb
    end if
    if(present(c)) then
        c=globrbuff(count:count+sizec-1)
        count=count+sizec
    end if
    if(present(d)) then
        d=globrbuff(count:count+sized-1)
        count=count+sized
    end if
    if(present(e)) then
        e=globrbuff(count:count+sizee-1)
        count=count+sizee
    end if
    if(present(f)) then
        f=globrbuff(count:count+sizef-1)
        count=count+sizef
    end if
    deallocate(rbuff, globrbuff)    
  end subroutine qmpi_1dreal8_reduction

  subroutine qmpi_integer_reduction(type,a,b,c,d,e,f,root,comm)
!
! perform a reduction of 'type' on each of the given arguments a - f. 
! if type is:
!  'sum': for each argument, return the sum of the argument over all processors
!  'mul': the product
!  'min': the minimum value
!  'max': the maximum value
! root is an optional argument, if given only return the result on that processor (reduce)
!  the default is to return the result on all processors (allreduce)
!    
    implicit none
    character(3) type
    integer(sp) a
    integer(sp), optional, intent(inout) :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    integer rbuff(6),globrbuff(6)

    if( trim(type).ne.'sum' .and. trim(type).ne.'mul' .and. trim(type).ne.'min' .and. trim(type).ne.'max')then
        write(*,*)'qmpi.f90 reduce error: reduction of type ',type,'not supported'
        stop
    end if

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    boss=0
    if(present(root))boss=root

    globrbuff(:)=0
    count=0
    !if(present(a)) then
        count=count+1
        rbuff(count)=a
    !end if
    if(present(b)) then
        count=count+1
        rbuff(count)=b
    end if
    if(present(c)) then
        count=count+1
        rbuff(count)=c
    end if
    if(present(d)) then
        count=count+1
        rbuff(count)=d
    end if
    if(present(e)) then
        count=count+1
        rbuff(count)=e
    end if
    if(present(f)) then
        count=count+1
        rbuff(count)=f
    end if

    select case(type)
    case('sum')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_sum,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_sum,communicator,ierr)
        end if
    case('mul')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_prod,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_prod,communicator,ierr)
        end if
    case('min')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_min,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_min,communicator,ierr)
        end if
    case('max')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_max,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_max,communicator,ierr)
        end if
    end select

    count=0
!    if(present(a)) then
        count=count+1
        a=globrbuff(count)
!    end if
    if(present(b)) then
        count=count+1
        b=globrbuff(count)
    end if
    if(present(c)) then
        count=count+1
        c=globrbuff(count)
    end if
    if(present(d)) then
        count=count+1
        d=globrbuff(count)
    end if
    if(present(e)) then
        count=count+1
        e=globrbuff(count)
    end if
    if(present(f)) then
        count=count+1
        f=globrbuff(count)
    end if
  end subroutine qmpi_integer_reduction


  subroutine qmpi_integer1d_reduction(type,a,b,c,d,e,f,root,comm)
!
! perform a reduction of 'type' on each of the given arguments a - f. 
! if type is:
!  'sum': for each argument, return the sum of the argument over all processors
!  'mul': the product
!  'min': the minimum value
!  'max': the maximum value
! root is an optional argument, if given only return the result on that processor (reduce)
!  the default is to return the result on all processors (allreduce)
!    
    implicit none
    character(3) type
    integer(sp) a(:)
    integer(sp), dimension(:), optional, intent(inout) :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    integer, allocatable :: rbuff(:),globrbuff(:)

    integer sizea,sizeb,sizec,sized,sizee,sizef,maxsize
    
    if( trim(type).ne.'sum' .and. trim(type).ne.'mul' .and. trim(type).ne.'min' .and. trim(type).ne.'max')then
        write(*,*)'qmpi.f90 reduce error: reduction of type ',type,'not supported'
        stop
    end if

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    sizea=size(a)
    sizeb=0 ; sizec=0 ; sized=0 ; sizee=0 ; sizef=0
    if(present(b)) sizeb=size(b)
    if(present(c)) sizec=size(c)
    if(present(d)) sized=size(d)
    if(present(e)) sizee=size(e)
    if(present(f)) sizef=size(f)
    maxsize = maxval((/sizea,sizeb,sizec,sized,sizee,sizef/))
    allocate(rbuff(6*maxsize), globrbuff(6*maxsize))
    
    boss=0
    if(present(root))boss=root

    globrbuff(:)=0
    
! count should from now on be the beginning of the "next" argument
! array in rbuff:
    count=1
    
    rbuff(count:count+sizea-1) = a
    count = count + sizea
    
    if(present(b)) then
        rbuff(count:count+sizeb-1)=b
        count=count+sizeb
    end if
    if(present(c)) then
        rbuff(count:count+sizec-1)=c
        count=count+sizec
    end if
    if(present(d)) then
        rbuff(count:count+sized-1)=d
        count=count+sized
    end if
    if(present(e)) then
        rbuff(count:count+sizee-1)=e
        count=count+sizee
    end if
    if(present(f)) then
        rbuff(count:count+sizef-1)=f
        count=count+sizef
    end if
    
! new role for count: now as the last item of our big array to be reduced
    count = count -1
    
    select case(type)
    case('sum')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_sum,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_sum,communicator,ierr)
        end if
    case('mul')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_prod,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_prod,communicator,ierr)
        end if
    case('min')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_min,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_min,communicator,ierr)
        end if
    case('max')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_max,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_max,communicator,ierr)
        end if
    end select

! then copy back results:    
    count=1
    a = globrbuff(count:count+sizea-1)
    count = count + sizea
    
    if(present(b)) then
        b=globrbuff(count:count+sizeb-1)
        count=count+sizeb
    end if
    if(present(c)) then
        c=globrbuff(count:count+sizec-1)
        count=count+sizec
    end if
    if(present(d)) then
        d=globrbuff(count:count+sized-1)
        count=count+sized
    end if
    if(present(e)) then
        e=globrbuff(count:count+sizee-1)
        count=count+sizee
    end if
    if(present(f)) then
        f=globrbuff(count:count+sizef-1)
        count=count+sizef
    end if
    deallocate(rbuff, globrbuff)    
  end subroutine qmpi_integer1d_reduction

  subroutine qmpi_integer2d_reduction(type,a,b,c,d,e,f,root,comm)
!
! perform a reduction of 'type' on each of the given arguments a - f. 
! if type is:
!  'sum': for each argument, return the sum of the argument over all processors
!  'mul': the product
!  'min': the minimum value
!  'max': the maximum value
! root is an optional argument, if given only return the result on that processor (reduce)
!  the default is to return the result on all processors (allreduce)
!    
    implicit none
    character(3) type
    integer(sp) a(:,:)
    integer(sp), dimension(:,:), optional, intent(inout) :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    integer, allocatable :: rbuff(:),globrbuff(:)

    integer sizea,sizeb,sizec,sized,sizee,sizef,maxsize
    
    if( trim(type).ne.'sum' .and. trim(type).ne.'mul' .and. trim(type).ne.'min' .and. trim(type).ne.'max')then
        write(*,*)'qmpi.f90 reduce error: reduction of type ',type,'not supported'
        stop
    end if

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    sizea=size(a,1)*size(a,2)
    sizeb=0 ; sizec=0 ; sized=0 ; sizee=0 ; sizef=0
    if(present(b)) sizeb=size(b,1)*size(b,2)
    if(present(c)) sizec=size(c,1)*size(c,2)
    if(present(d)) sized=size(d,1)*size(d,2)
    if(present(e)) sizee=size(e,1)*size(e,2)
    if(present(f)) sizef=size(f,1)*size(f,2)
    maxsize = maxval((/sizea,sizeb,sizec,sized,sizee,sizef/))
    allocate(rbuff(6*maxsize), globrbuff(6*maxsize))
    
    boss=0
    if(present(root))boss=root

    globrbuff(:)=0
    
! count should from now on be the beginning of the "next" argument
! array in rbuff:
    count=1
    
    rbuff(count:count+sizea-1) = reshape(a, (/sizea/))
    count = count + sizea
    
    if(present(b)) then
        rbuff(count:count+sizeb-1) = reshape(b, (/sizeb/))
        count=count+sizeb
    end if
    if(present(c)) then
        rbuff(count:count+sizec-1) = reshape(c, (/sizec/))
        count=count+sizec
    end if
    if(present(d)) then
        rbuff(count:count+sized-1) = reshape(d, (/sized/))
        count=count+sized
    end if
    if(present(e)) then
        rbuff(count:count+sizee-1) = reshape(e, (/sizee/))
        count=count+sizee
    end if
    if(present(f)) then
        rbuff(count:count+sizef-1) = reshape(f, (/sizef/))
        count=count+sizef
    end if
    
! new role for count: now as the last item of our big array to be reduced
    count = count -1
    
    select case(type)
    case('sum')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_sum,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_sum,communicator,ierr)
        end if
    case('mul')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_prod,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_prod,communicator,ierr)
        end if
    case('min')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_min,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_min,communicator,ierr)
        end if
    case('max')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_max,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_max,communicator,ierr)
        end if
    end select

! then copy back results:    
    count=1
    a = reshape(globrbuff(count:count+sizea-1), shape(a))
    count = count + sizea
    
    if(present(b)) then
        b = reshape(globrbuff(count:count+sizeb-1), shape(b))
        count=count+sizeb
    end if
    if(present(c)) then
        c = reshape(globrbuff(count:count+sizec-1), shape(c))
        count=count+sizec
    end if
    if(present(d)) then
        d = reshape(globrbuff(count:count+sized-1), shape(d))
        count=count+sized
    end if
    if(present(e)) then
        e = reshape(globrbuff(count:count+sizee-1), shape(e))
        count=count+sizee
    end if
    if(present(f)) then
        f = reshape(globrbuff(count:count+sizef-1), shape(f))
        count=count+sizef
    end if
    deallocate(rbuff, globrbuff)    
  end subroutine qmpi_integer2d_reduction


  subroutine qmpi_integer4d_reduction(type,a,b,c,d,e,f,root,comm)
!
! perform a reduction of 'type' on each of the given arguments a - f. 
! if type is:
!  'sum': for each argument, return the sum of the argument over all processors
!  'mul': the product
!  'min': the minimum value
!  'max': the maximum value
! root is an optional argument, if given only return the result on that processor (reduce)
!  the default is to return the result on all processors (allreduce)
!    
    implicit none
    character(3) type
    integer(sp) a(:,:,:,:)
    integer(sp), dimension(:,:,:,:), optional, intent(inout) :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    integer, allocatable :: rbuff(:),globrbuff(:)

    integer sizea,sizeb,sizec,sized,sizee,sizef,maxsize
    
    if( trim(type).ne.'sum' .and. trim(type).ne.'mul' .and. trim(type).ne.'min' .and. trim(type).ne.'max')then
        write(*,*)'qmpi.f90 reduce error: reduction of type ',type,'not supported'
        stop
    end if

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    sizea=size(a,1)*size(a,2)*size(a,3)*size(a,4)
    sizeb=0 ; sizec=0 ; sized=0 ; sizee=0 ; sizef=0
    if(present(b)) sizeb=size(b,1)*size(b,2)*size(b,3)*size(b,4)
    if(present(c)) sizec=size(c,1)*size(c,2)*size(c,3)*size(c,4)
    if(present(d)) sized=size(d,1)*size(d,2)*size(d,3)*size(d,4)
    if(present(e)) sizee=size(e,1)*size(e,2)*size(e,3)*size(e,4)
    if(present(f)) sizef=size(f,1)*size(f,2)*size(f,3)*size(f,4)
    maxsize = maxval((/sizea,sizeb,sizec,sized,sizee,sizef/))
    allocate(rbuff(6*maxsize), globrbuff(6*maxsize))
    
    boss=0
    if(present(root))boss=root

    globrbuff(:)=0
    
! count should from now on be the beginning of the "next" argument
! array in rbuff:
    count=1
    
    rbuff(count:count+sizea-1) = reshape(a, (/sizea/))
    count = count + sizea
    
    if(present(b)) then
        rbuff(count:count+sizeb-1) = reshape(b, (/sizeb/))
        count=count+sizeb
    end if
    if(present(c)) then
        rbuff(count:count+sizec-1) = reshape(c, (/sizec/))
        count=count+sizec
    end if
    if(present(d)) then
        rbuff(count:count+sized-1) = reshape(d, (/sized/))
        count=count+sized
    end if
    if(present(e)) then
        rbuff(count:count+sizee-1) = reshape(e, (/sizee/))
        count=count+sizee
    end if
    if(present(f)) then
        rbuff(count:count+sizef-1) = reshape(f, (/sizef/))
        count=count+sizef
    end if
    
! new role for count: now as the last item of our big array to be reduced
    count = count -1
    
    select case(type)
    case('sum')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_sum,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_sum,communicator,ierr)
        end if
    case('mul')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_prod,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_prod,communicator,ierr)
        end if
    case('min')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_min,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_min,communicator,ierr)
        end if
    case('max')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_max,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER, mpi_max,communicator,ierr)
        end if
    end select

! then copy back results:    
    count=1
    a = reshape(globrbuff(count:count+sizea-1), shape(a))
    count = count + sizea
    
    if(present(b)) then
        b = reshape(globrbuff(count:count+sizeb-1), shape(b))
        count=count+sizeb
    end if
    if(present(c)) then
        c = reshape(globrbuff(count:count+sizec-1), shape(c))
        count=count+sizec
    end if
    if(present(d)) then
        d = reshape(globrbuff(count:count+sized-1), shape(d))
        count=count+sized
    end if
    if(present(e)) then
        e = reshape(globrbuff(count:count+sizee-1), shape(e))
        count=count+sizee
    end if
    if(present(f)) then
        f = reshape(globrbuff(count:count+sizef-1), shape(f))
        count=count+sizef
    end if
    deallocate(rbuff, globrbuff)    
  end subroutine qmpi_integer4d_reduction

  
  
  subroutine qmpi_integer8_reduction(type,a,b,c,d,e,f,root,comm)
!
! perform a reduction of 'type' on each of the given arguments a - f. 
! if type is:
!  'sum': for each argument, return the sum of the argument over all processors
!  'mul': the product
!  'min': the minimum value
!  'max': the maximum value
! root is an optional argument, if given only return the result on that processor (reduce)
!  the default is to return the result on all processors (allreduce)
!    
    implicit none
    character(3) type
    integer(long) a
    integer(long), optional, intent(inout) :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    integer(long) rbuff(6),globrbuff(6)

    if(len(type).ne.3)then
        write(*,*)'qmpi.f90 reduce error: type must be one of "mul","sum","min" or "max"'
        stop
    end if
    if( trim(type).ne.'sum' .and. trim(type).ne.'mul' .and. trim(type).ne.'min' .and. trim(type).ne.'max')then
        write(*,*)'qmpi.f90 reduce error: reduction of type ',type,'not supported'
        stop
    end if

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return
    
    boss=0
    if(present(root))boss=root

    globrbuff(:)=0_dp
    count=0
!    if(present(a)) then
        count=count+1
        rbuff(count)=a
!    end if
    if(present(b)) then
        count=count+1
        rbuff(count)=b
    end if
    if(present(c)) then
        count=count+1
        rbuff(count)=c
    end if
    if(present(d)) then
        count=count+1
        rbuff(count)=d
    end if
    if(present(e)) then
        count=count+1
        rbuff(count)=e
    end if
    if(present(f)) then
        count=count+1
        rbuff(count)=f
    end if

    select case(type)
    case('sum')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER8, mpi_sum,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER8, mpi_sum,communicator,ierr)
        end if
    case('mul')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER8, mpi_prod,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER8, mpi_prod,communicator,ierr)
        end if
    case('min')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER8, mpi_min,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER8, mpi_min,communicator,ierr)
        end if
    case('max')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_INTEGER8, mpi_max,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_INTEGER8, mpi_max,communicator,ierr)
        end if
    end select

    count=1
    a=globrbuff(count)
    
    if(present(b)) then
        count=count+1
        b=globrbuff(count)
    end if
    if(present(c)) then
        count=count+1
        c=globrbuff(count)
    end if
    if(present(d)) then
        count=count+1
        d=globrbuff(count)
    end if
    if(present(e)) then
        count=count+1
        e=globrbuff(count)
    end if
    if(present(f)) then
        count=count+1
        f=globrbuff(count)
    end if
  end subroutine qmpi_integer8_reduction

  subroutine qmpi_2dreal4_reduction(type,a,b,c,d,e,f,root,comm)
!
! perform a reduction of 'type' on each of the given arguments a - f. 
! if type is:
!  'sum': for each argument, return the sum of the argument over all processors
!  'mul': the product
!  'min': the minimum value
!  'max': the maximum value
! root is an optional argument, if given only return the result on that processor (reduce)
!  the default is to return the result on all processors (allreduce)
!    
    implicit none
    character(3) type
    real(sp) a(:,:)
    real(sp), dimension(:,:), optional, intent(inout) :: b,c,d,e,f
    integer, optional :: root,comm
    integer count,boss,communicator
    real(sp), allocatable :: rbuff(:),globrbuff(:)

    integer sizea,sizeb,sizec,sized,sizee,sizef,maxsize
    
    if( trim(type).ne.'sum' .and. trim(type).ne.'mul' .and. trim(type).ne.'min' .and. trim(type).ne.'max')then
        write(*,*)'qmpi.f90 reduce error: reduction of type ',type,'not supported'
        stop
    end if

    if(numproc == 1) return

    communicator=qmpi_comm_world
    if(present(comm)) communicator=comm
    if(communicator.eq.mpi_comm_null) return    

    sizea=size(a,1)*size(a,2)
    sizeb=0 ; sizec=0 ; sized=0 ; sizee=0 ; sizef=0
    if(present(b)) sizeb=size(b,1)*size(b,2)
    if(present(c)) sizec=size(c,1)*size(c,2)
    if(present(d)) sized=size(d,1)*size(d,2)
    if(present(e)) sizee=size(e,1)*size(e,2)
    if(present(f)) sizef=size(f,1)*size(f,2)
    maxsize = maxval((/sizea,sizeb,sizec,sized,sizee,sizef/))
    allocate(rbuff(6*maxsize), globrbuff(6*maxsize))
    
    boss=0
    if(present(root))boss=root

    globrbuff(:)=0
    
! count should from now on be the beginning of the "next" argument
! array in rbuff:
    count=1
    
    rbuff(count:count+sizea-1) = reshape(a, (/sizea/))
    count = count + sizea
    
    if(present(b)) then
        rbuff(count:count+sizeb-1) = reshape(b, (/sizeb/))
        count=count+sizeb
    end if
    if(present(c)) then
        rbuff(count:count+sizec-1) = reshape(c, (/sizec/))
        count=count+sizec
    end if
    if(present(d)) then
        rbuff(count:count+sized-1) = reshape(d, (/sized/))
        count=count+sized
    end if
    if(present(e)) then
        rbuff(count:count+sizee-1) = reshape(e, (/sizee/))
        count=count+sizee
    end if
    if(present(f)) then
        rbuff(count:count+sizef-1) = reshape(f, (/sizef/))
        count=count+sizef
    end if
    
! new role for count: now as the last item of our big array to be reduced
    count = count -1
    
    select case(type)
    case('sum')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL4, mpi_sum,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL4, mpi_sum,communicator,ierr)
        end if
    case('mul')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL4, mpi_prod,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL4, mpi_prod,communicator,ierr)
        end if
    case('min')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL4, mpi_min,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL4, mpi_min,communicator,ierr)
        end if
    case('max')
        if(present(root))then
            call mpi_reduce(rbuff,globrbuff,count, MPI_REAL4, mpi_max,boss,communicator,ierr)
        else
            call mpi_allreduce(rbuff,globrbuff,count, MPI_REAL4, mpi_max,communicator,ierr)
        end if
    end select

! then copy back results:    
    count=1
    a = reshape(globrbuff(count:count+sizea-1), shape(a))
    count = count + sizea
    
    if(present(b)) then
        b = reshape(globrbuff(count:count+sizeb-1), shape(b))
        count=count+sizeb
    end if
    if(present(c)) then
        c = reshape(globrbuff(count:count+sizec-1), shape(c))
        count=count+sizec
    end if
    if(present(d)) then
        d = reshape(globrbuff(count:count+sized-1), shape(d))
        count=count+sized
    end if
    if(present(e)) then
        e = reshape(globrbuff(count:count+sizee-1), shape(e))
        count=count+sizee
    end if
    if(present(f)) then
        f = reshape(globrbuff(count:count+sizef-1), shape(f))
        count=count+sizef
    end if
    deallocate(rbuff, globrbuff)    
  end subroutine qmpi_2dreal4_reduction


  
  subroutine initialize_communicator(ranks, c)
!    
! given a list of processors, set up a communicator. usage ala
!    
!   type(communicator_object) c
!   ...    
!   call initialize_communicator((/5,2,1/), c)
!   call broadcast(..., comm=c%comm)
!    
    implicit none
    integer ranks(:)
    type(communicator_object) c
    integer orig_group,i,ranks1(numproc), ranks2(numproc)

    c%size = size(ranks)
    c%global_rank(1:c%size) = ranks
    
    call mpi_comm_group(qmpi_comm_world, orig_group, ierr)
    call mpi_group_incl(orig_group, c%size, ranks, c%group, ierr)
    call mpi_comm_create(qmpi_comm_world, c%group, c%comm, ierr)
    call mpi_group_rank(c%group, c%myid, ierr)

    ranks1 = (/(i-1, i=1, numproc)/)
        
    call mpi_group_translate_ranks (orig_group, numproc, ranks1, c%group, ranks2, ierr)
        
    if(c%myid == 0) then
        write(*,*) "global RANK, local RANK"
        do i=1,numproc
            if(ranks2(i) .ne. MPI_UNDEFINED) write(*,*) ranks1(i), ranks2(i)
        end do
    end if
  end subroutine initialize_communicator


  subroutine free_communicator(c)
    implicit none
    type(communicator_object) c

    call mpi_comm_free(c%comm, ierr)
    call mpi_group_free(c%group, ierr)
  end subroutine free_communicator


  subroutine decomp1d(n, n1, n2)
!    
! given a workload N, return the start N1 and end N2 of a equally
! distributed workload for the current processor
!    
    integer n,n1,n2,mn

! if we have less work than processors, return n2<n1 (empty loop) on
! idle processors
    if(n < numproc) then
        if(myid < n) then
            n1 = myid + 1
            n2 = myid + 1
        else
            n1 = 2
            n2 = 1
        end if
        return
    end if
    
    mn = n
    do
        if(mod(mn, numproc) == 0) exit
        mn = mn + 1
    end do
    
! if we have sligthly more work than processors, the last processors
! will be idle with the above algorithm, return "fortran null" range
! also there
    n1 = min(n+1, myid*mn/numproc + 1)
    n2 = min(n,   (myid+1)*mn/numproc)   
  end subroutine decomp1d
  
end module qmpi
