module soriter
  
contains

SUBROUTINE SORITER2D(B,X,ITER,RESID,INFO,IM,KB , &
     EAST,WEST,UP,DOWN,TOTSUM,FSM)
! 
!  -- Iterative routine --
!    Gauss-seidel
!    Written by Jarle Berntsen

  use mod_mpi, direast=>east, dirwest=>west
  use control_parameters
  use state, only : debug
  
  IMPLICIT NONE
  INTEGER IM,KB, ITER, INFO 
  REAL   RESID
  REAL, POINTER ::  X(:,:,:), B(:,:,:),  &
       &  EAST(:,:,:),WEST(:,:,:),  &
       &  UP(:,:,:),DOWN(:,:,:), &
       &  TOTSUM(:,:,:),FSM(:,:)
! 
!  Purpose
!  =======
! 
!  GAUSSITER solves the linear system A!x = b using the
!  Gauss Seidel  iterative method 
! 
!  Convergence test: ( norm( b - A!x ) / norm( b ) ) < TOL.
!  For other measures, see the above reference.
! 
!  Arguments
!  =========
! 
! 
!  B       (input) REAL array, dimension N.
!          On entry, right hand side vector B.
!          Unchanged on exit.
! 
!  X       (input/output) REAL array, dimension N.
!          On input, the initial guess. This is commonly set to 
!          the zero vector. 
!          On exit, if INFO = 0, the iterated approximate solution.
! 
!  WORK    (workspace) REAL array, dimension (LDW,7)
!          Workspace for residual, direction vector, etc.
!          Note that vectors R and S shared the same workspace.
!          (defineres her lokalt)
! 
!  ITER    (input/output) INTEGER
!          On input, the maximum iterations to be performed.
!          On output, actual number of iterations performed.
! 
!  RESID   (input/output) REAL
!          On input, the allowable convergence measure for
!          norm( b - A!x ) / norm( b ).
!          On output, the final value of this measure.
! 
!  INFO    (output) INTEGER
! 
!          =  0: Successful exit. Iterated approximate solution returned.
! 
!          >  0: Convergence to tolerance not achieved. This will be 
!                set to the number of iterations performed.
! 
! 

!  Local variables
  INTEGER IT,I,J,K
  REAL NORM1,NORM2,ERROR
  REAL, POINTER :: TMP1(:,:,:), TMP2(:,:,:), INVTOTSUM(:,:,:)
  REAL OMEGA
  logical :: sorinit=.true.
  integer dk
! 
!  OMEGA is the relax. param. in SOR
!  With OMEGA = 1 , it is Gauss-Seidel
!  Use OMEGA = 1 to estimate the convergence rate RHO for Gauss-seidel,
!  and then estimate the optimal omega, which typically become for large problems
!  around 1.8-1.9
! 
! Test on convergence costs, and is here done every 10th step (May easily be modified)
! 
  !PARAMETER (OMEGA = 1.0)
  PARAMETER (OMEGA = 1.8866)
  !PARAMETER (OMEGA = 1.7694)


  allocate(tmp1(i1:i2,j1:j2,kb), tmp2(i1:i2,j1:j2,kb), invtotsum(i1:i2,j1:j2,kb))
  
  J=1

  if(sorinit)then
!      
! to make the red black loops traverse exact the same numbers as 
! the sequential version, 
!   do i=2,im-1,2    2,4,6,8,...
!   do i=3,im-1,2    3,5,7,..
! we precompute these starting indices for
! each block:
      i_eq_2_rb=1    ; i_eq_3_rb=1
      if(block%ibn==1 .and. .not. periodic_ew)then
          i_eq_2_rb=2 ; i_eq_3_rb=3
      else
          if(mod(block%fi,2) == 0) then
              i_eq_3_rb = i_eq_3_rb + 1
          else
              i_eq_2_rb = i_eq_2_rb + 1        
          end if
      end if
      sorinit=.false.
  end if
 

  call tag_ghost(east)
  call tag_ghost(west)
  call tag_ghost(up)
  call tag_ghost(down)
  call tag_ghost(b)
  call tag_ghost(totsum)
  call tag_ghost(x)
  call update_ghost
  

  INFO = 1

  do k=1,kb
      do i=1,i_eq_im
          if( totsum(i,j,k) .ne. 0.0) then
              invtotsum(i,j,k) = 1.0_8/totsum(i,j,k)
          else            
              invtotsum(i,j,k) = 0.0_8
          end if
          tmp2(i,j,k) = B(I,j,K)**2        
      end do
  end do

  call tag_ghost(tmp2)
  call tag_ghost(invtotsum)
  call update_ghost
  
  norm2 = psum(tmp2,im,1,kb)
  
  tmp1 = 0.0

  DO IT = 1,ITER
      
! Red black ordered SOR
          

      DO K = 1,KB-1,2
          dk=1
          if(k==1) dk=0
          DO I = i_eq_2_rb,IM_mi_1,2
            IF (FSM(I,J).NE.0.0) THEN
              X(I,J,K) = X(I,J,K) + OMEGA*(B(I,J,K) - (TOTSUM(I,J,K)*X(I,J,K) + &
                   &   EAST(I,J,K)*X(I+1,J,K) + WEST(I,J,K)*X(I-1,J,K) + &
                   &   UP(I,J,K)*X(I,J,K-dk) + DOWN(I,J,K)*X(I,J,K+1)  &
                   &   ))*(invTOTSUM(I,J,K))
            END IF
          END DO
      END DO
      

      DO K = 2,KB-1,2
          DO I = I_eq_3_rb,IM_mi_1,2
            IF (FSM(I,J).NE.0.0) THEN
              X(I,J,K) = X(I,J,K) + OMEGA*(B(I,J,K) - (TOTSUM(I,J,K)*X(I,J,K) + &
                   &   EAST(I,J,K)*X(I+1,J,K) + WEST(I,J,K)*X(I-1,J,K) + &
                   &   UP(I,J,K)*X(I,J,K-1) + DOWN(I,J,K)*X(I,J,K+1)  &
                   &   ))*(invTOTSUM(I,J,K))
            END IF
          END DO
      END DO
      
      call tag_ghost(x, width=1)
      call update_ghost
          

      DO K = 2,KB-1,2
          DO I = I_eq_2_rb,IM_mi_1,2
            IF (FSM(I,J).NE.0.0) THEN
              X(I,J,K) = X(I,J,K) + OMEGA*(B(I,J,K) - (TOTSUM(I,J,K)*X(I,J,K) + &
                   &   EAST(I,J,K)*X(I+1,J,K) + WEST(I,J,K)*X(I-1,J,K) + &
                   &   UP(I,J,K)*X(I,J,K-1) + DOWN(I,J,K)*X(I,J,K+1)  &
                   &   ))*(invTOTSUM(I,J,K))
            END IF
          END DO
      END DO
      

      DO K = 1,KB-1,2
          dk=1
          if(k==1) dk=0
          DO I = i_eq_3_rb,IM_mi_1,2
            IF (FSM(I,J).NE.0.0) THEN
              X(I,J,K) = X(I,J,K) + OMEGA*(B(I,J,K) - (TOTSUM(I,J,K)*X(I,J,K) + &
                   &   EAST(I,J,K)*X(I+1,J,K) + WEST(I,J,K)*X(I-1,J,K) + &
                   &   UP(I,J,K)*X(I,J,K-dk) + DOWN(I,J,K)*X(I,J,K+1)  &
                   &   ))*(invTOTSUM(I,J,K))
            END IF
          END DO
      END DO
      
      call tag_ghost(x, width=1)
      call update_ghost

! 
!  test error
! 
      IF (MOD(IT,SOR_CHECK_RESID_INTERVAL).EQ.0 .OR. IT==ITER) THEN
          tmp1 = 0.0
          DO K = 1,KB-1
              dk=1
              if(k==1) dk=0              
              DO I = i_eq_2,IM_mi_1
                IF (FSM(I,J).NE.0.0) THEN
                  tmp1(i,j,k) = (TOTSUM(I,J,K)*X(I,J,K) +  &
                   &  EAST(I,J,K)*X(I+1,J,K) + &
                   &  WEST(I,J,K)*X(I-1,J,K) + &
                   &  UP(I,J,K)*X(I,J,K-dk) + &
                   &  DOWN(I,J,K)*X(I,J,K+1) - B(I,J,K))**2
                END IF
              END DO
          END DO

          call tag_ghost(tmp1)
          call update_ghost
          norm1 = psum(tmp1,im,1,kb)

          IF (NORM2.NE.0.) THEN
              ERROR = sqrt(NORM1/NORM2)
          ELSE
              ERROR = 0.
          ENDIF
          IF (ERROR.LT.RESID) THEN
              INFO = 0
              EXIT
          ENDIF
      ENDIF
  END DO

  x(:,:,kb)=0.0 !1e36 !x(:,kb-1)

  if(debug > 1 .and. master) write(*,*) "soriter2d resid=", error," iter=",it
  deallocate(tmp1, tmp2, invtotsum)
END SUBROUTINE SORITER2D


SUBROUTINE SORITER3D(B,X,ITER,RESID,INFO,IM,JM,KB , &
     EAST,WEST,SOUTH,NORTH,UP,DOWN,TOTSUM,FSM)
! 
!  -- Iterative routine --
!    Gauss-seidel
!    Written by Jarle Berntsen

  use mod_mpi, direast=>east, dirwest=>west, dirsouth=>south, dirnorth=>north
  use control_parameters
  use state, only : debug
  
  IMPLICIT NONE
  INTEGER            IM,JM,KB, ITER, INFO 
  REAL   RESID
  REAL, pointer ::   X(:,:,:), B(:,:,:),  &
       &  EAST(:,:,:),WEST(:,:,:),  &
       &  UP(:,:,:),DOWN(:,:,:), &
       &  SOUTH(:,:,:),NORTH(:,:,:), &
       &  TOTSUM(:,:,:),FSM(:,:)
! 
!  Purpose
!  =======
! 
!  GAUSSITER solves the linear system A!x = b using the
!  Gauss Seidel  iterative method 
! 
!  Convergence test: ( norm( b - A!x ) / norm( b ) ) < TOL.
!  For other measures, see the above reference.
! 
!  Arguments
!  =========
! 
! 
!  B       (input) REAL array, dimension N.
!          On entry, right hand side vector B.
!          Unchanged on exit.
! 
!  X       (input/output) REAL array, dimension N.
!          On input, the initial guess. This is commonly set to 
!          the zero vector. 
!          On exit, if INFO = 0, the iterated approximate solution.
! 
!  WORK    (workspace) REAL array, dimension (LDW,7)
!          Workspace for residual, direction vector, etc.
!          Note that vectors R and S shared the same workspace.
!          (defineres her lokalt)
! 
!  ITER    (input/output) INTEGER
!          On input, the maximum iterations to be performed.
!          On output, actual number of iterations performed.
! 
!  RESID   (input/output) REAL
!          On input, the allowable convergence measure for
!          norm( b - A!x ) / norm( b ).
!          On output, the final value of this measure.
! 
!  INFO    (output) INTEGER
! 
!          =  0: Successful exit. Iterated approximate solution returned.
! 
!          >  0: Convergence to tolerance not achieved. This will be 
!                set to the number of iterations performed.
! 
! 

!  Local variables
  INTEGER IT,I,J,K,DK
  REAL NORM1,NORM2,ERROR
  REAL, POINTER :: TMP1(:,:,:), TMP2(:,:,:) 
  REAL, POINTER :: INVTOTSUM(:,:,:)
  REAL OMEGA
  logical :: sorinit=.true.
! 
!  OMEGA is the relax. param. in SOR
!  With OMEGA = 1 , it is Gauss-Seidel
!  Use OMEGA = 1 to estimate the convergence rate RHO for Gauss-seidel,
!  and then estimate the optimal omega, which typically become for large problems
!  around 1.8-1.9
! 
! Test on convergence costs, and is here done every 10th step (May easily be modified)
! 
  !PARAMETER (OMEGA = 1.0)
  PARAMETER (OMEGA = 1.8866)
  !PARAMETER (OMEGA = 1.7694)

  allocate(tmp1(i1:i2,j1:j2,kb), tmp2(i1:i2,j1:j2,kb), invtotsum(i1:i2,j1:j2,kb))
  
!      
! to make the red black loops traverse exact the same numbers as 
! the sequential version, 
!   do i=2,im-1,2    2,4,6,8,...
!   do i=3,im-1,2    3,5,7,..
! we precompute these starting indices for
! each block:
  
  if(sorinit)then
    ! Is the first x-value even or odd?
    i_eq_2_rb=1    ; i_eq_3_rb=1
    if(block%ibn==1 .and. .not. periodic_ew)then
      i_eq_2_rb=2 ; i_eq_3_rb=3
    else
      if(mod(block%fi,2) == 0) then
          i_eq_3_rb = i_eq_3_rb + 1   ! first even, second odd
      else
          i_eq_2_rb = i_eq_2_rb + 1   ! first odd, second even
      end if
    end if
    ! Is the first y-value even or odd?
    j_eq_2_rb=1    ; j_eq_3_rb=1
    if(block%jbn==1)then
      j_eq_2_rb=2 ; j_eq_3_rb=3
    else
      if(mod(block%fj,2) == 0) then
          j_eq_3_rb = j_eq_3_rb + 1   ! first even, second odd
      else
          j_eq_2_rb = j_eq_2_rb + 1   ! first odd, second even
      end if
    end if
  end if

  call tag_ghost(east)
  call tag_ghost(west)
  call tag_ghost(south)
  call tag_ghost(north)
  call update_ghost
  
  call tag_ghost(up)
  call tag_ghost(down)
  call tag_ghost(b)
  call tag_ghost(totsum)
  call tag_ghost(x)
  call update_ghost
  
  INFO = 1

  do k=1,kb
    do j=1,j_eq_jm 
      do i=1,i_eq_im
          if( totsum(i,j,k) .ne. 0.0) then
              invtotsum(i,j,k) = 1.0_8/totsum(i,j,k)
          else            
              invtotsum(i,j,k) = 0.0_8
          end if
      end do
    end do
  end do

!  call tag_ghost(invtotsum)
!  call update_ghost

  DO IT = 1,ITER
!
! Red black ordered SOR
!          

      DO K = 1,KB-1
       dk=1
       if(k==1) dk=0
          
        DO J = j_eq_2_rb,JM_mi_1,2
          DO I = i_eq_2_rb,IM_mi_1,2
            IF (FSM(I,J).NE.0.0) THEN
              X(I,J,K) = X(I,J,K) + OMEGA*(B(I,J,K) -  &
                   &   (TOTSUM(I,J,K)*X(I,J,K) + &
                   &   EAST(I,J,K)*X(I+1,J,K) + WEST(I,J,K)*X(I-1,J,K) + &
                   &   SOUTH(I,J,K)*X(I,J-1,K) + NORTH(I,J,K)*X(I,J+1,K) + &
                   &   UP(I,J,K)*X(I,J,K-DK) + DOWN(I,J,K)*X(I,J,K+1)  &
                   &   ))*(invTOTSUM(I,J,K))
            END IF
          END DO
        END DO
        DO J = j_eq_3_rb,JM_mi_1,2
          DO I = i_eq_3_rb,IM_mi_1,2
            IF (FSM(I,J).NE.0.0) THEN
              X(I,J,K) = X(I,J,K) + OMEGA*(B(I,J,K) -  &
                   &   (TOTSUM(I,J,K)*X(I,J,K) + &
                   &   EAST(I,J,K)*X(I+1,J,K) + WEST(I,J,K)*X(I-1,J,K) + &
                   &   SOUTH(I,J,K)*X(I,J-1,K) + NORTH(I,J,K)*X(I,J+1,K) + &
                   &   UP(I,J,K)*X(I,J,K-DK) + DOWN(I,J,K)*X(I,J,K+1)  &
                   &   ))*(invTOTSUM(I,J,K))
            END IF
          END DO
        END DO
      END DO
      call tag_ghost(x, width=1)
      call update_ghost
      

      DO K = 1,KB-1
        dk=1
        if(k==1) dk=0
        DO J = j_eq_2_rb,JM_mi_1,2
          DO I = i_eq_3_rb,IM_mi_1,2
            IF (FSM(I,J).NE.0.0) THEN
              X(I,J,K) = X(I,J,K) + OMEGA*(B(I,J,K) -  &
                   &   (TOTSUM(I,J,K)*X(I,J,K) + &
                   &   EAST(I,J,K)*X(I+1,J,K) + WEST(I,J,K)*X(I-1,J,K) + &
                   &   SOUTH(I,J,K)*X(I,J-1,K) + NORTH(I,J,K)*X(I,J+1,K) + &
                   &   UP(I,J,K)*X(I,J,K-DK) + DOWN(I,J,K)*X(I,J,K+1)  &
                   &   ))*(invTOTSUM(I,J,K))
            END IF
          END DO
        END DO
        DO J = j_eq_3_rb,JM_mi_1,2
          DO I = i_eq_2_rb,IM_mi_1,2
            IF (FSM(I,J).NE.0.0) THEN
              X(I,J,K) = X(I,J,K) + OMEGA*(B(I,J,K) -  &
                   &   (TOTSUM(I,J,K)*X(I,J,K) + &
                   &   EAST(I,J,K)*X(I+1,J,K) + WEST(I,J,K)*X(I-1,J,K) + &
                   &   SOUTH(I,J,K)*X(I,J-1,K) + NORTH(I,J,K)*X(I,J+1,K) + &
                   &   UP(I,J,K)*X(I,J,K-DK) + DOWN(I,J,K)*X(I,J,K+1)  &
                   &   ))*(invTOTSUM(I,J,K))
            END IF
          END DO
        END DO
      END DO
      
      call tag_ghost(x, width=1)
      call update_ghost          

! 
!  test error
! 
      IF (MOD(IT,SOR_CHECK_RESID_INTERVAL).EQ.0 .OR. IT==ITER) THEN
          tmp1 = 0.0
          tmp2 = 0.0
          DO K = 1,KB-1
            dk=1
            if(k==1) dk=0
            DO J = j_eq_2,JM_mi_1
              DO I = i_eq_2,IM_mi_1
                IF (FSM(I,J).NE.0.0) THEN
                  tmp1(i,j,k) = (TOTSUM(I,J,K)*X(I,J,K) +  &
                   &  EAST(I,J,K)*X(I+1,J,K) + &
                   &  WEST(I,J,K)*X(I-1,J,K) + &
                   &  SOUTH(I,J,K)*X(I,J-1,K) + &
                   &  NORTH(I,J,K)*X(I-1,J+1,K) + &
                   &  UP(I,J,K)*X(I,J,K-DK) + &
                   &  DOWN(I,J,K)*X(I,J,K+1) - B(I,J,K))**2
                  tmp2(i,j,k) = B(I,J,K)**2
                END IF
              END DO
            END DO
          END DO

          call tag_ghost(tmp1)
          call tag_ghost(tmp2)
          call update_ghost
          norm1 = psum(tmp1,im,jm,kb)
          norm2 = psum(tmp2,im,jm,kb)

          IF (NORM2.NE.0.) THEN
              ERROR = sqrt(NORM1/NORM2)
          ELSE
              ERROR = 0.
          ENDIF
          IF (ERROR.LT.RESID) THEN
              INFO = 0
              EXIT
          ENDIF
      ENDIF
  END DO

  x(:,:,kb)=0.0 !1e36 !x(:,kb-1)

  deallocate(tmp1, tmp2, invtotsum)
  
  if(debug > 1 .and. master) write(*,*) "soriter3d resid=", error, " iter=",it
END SUBROUTINE SORITER3D


end module soriter
