SUBROUTINE SUPERBEEF(FIELDOLD,FIELD,DELFIELD,FAVER,ICODE)
!
!           SUPERBEEF integrates the equation
!
!           d(FIELD*D)/d(T) + d(FIELD*U*D)/d(X) + d(FIELD*V*D)/d(Y) +
!       
!           d(FIELD*W)/d(SIGMA) = d(AH*d(FIELD*D)/d(X))/d(X) +
!                                 d(AH*d(FIELD*D)/d(Y))/d(Y)
!
!           from timestep n to timestep n+1
!           with a superbee limiter TVD scheme.
!
!           d(*) is the differential operator,
!           FIELD is the scalar field to be integrated,
!           D is the dynamic depth,
!           T is time,
!           X and Y are the horisontal coordinates,
!           U is the horisontal velocity in X-direction,
!           V is the horisontal velocity in Y-direction,
!           SIGMA is the vertical coordinate,
!           W is the vertical velocity. 
!           
!           FAVER is the average value of the scalar field which is 
!                 subtracted from FIELD before differences are 
!                 computed to reduce the cancellation errors.
!           ICODE is an integer flag to signal which scalar field
!                 that is advected. ICODE is needed when river-runoff
!                 is included.
!
!***ROUTINES CALLED-NONE
!***END PROLOGUE
!
!   Global variables.
!
  USE STATE
  USE CONTROL_PARAMETERS
  USE MOD_BOUND
  USE ETAIMPLICIT
  USE MOD_CONCENTRATION, ONLY : CO
  IMPLICIT NONE
  REAL,DIMENSION(I1:I2,J1:J2,KB):: FIELDOLD, FIELD, DELFIELD
  REAL FAVER
  INTEGER ICODE
!
!   Local variables.
!
  REAL, DIMENSION(:,:,:), POINTER :: XFLUX,YFLUX,ZFLUX
  REAL :: XFLUXUP,YFLUXUP,ZFLUXUP
  REAL :: XFLUXLW,YFLUXLW,ZFLUXLW
  REAL :: XFLUXD,YFLUXD
  REAL AREA,NETFLUX,VOLUME,VOLUMEP,LAMBDAX,LAMBDAY
  REAL R,PHIR,NUMER,DENUM,EPSILON,AREAT,AREATU,AREATV,AREATW
  INTEGER I,J,K,NSIG
  REAL FRAC,FLUX,FIELDRIVER,FIELDWELL
  LOGICAL INBLOCK
  INTEGER RIV,GI,GJ,IW,DJ
  real, pointer :: tmp3d(:,:,:)
!
  ALLOCATE(XFLUX(I1:I2,J1:J2,KB),YFLUX(I1:I2,J1:J2,KB),ZFLUX(I1:I2,J1:J2,KB))

! use separate AH to advect concentration
  if(icode > 2) then
      if(associated(co(icode-2)%ah))then
          tmp3d => ah      
          ah => co(icode-2)%ah
      end if
  end if
  
  dj=1
  if(jm == 1)dj=0

  LAMBDAX = DT/DX
  LAMBDAY = DT/DY
  EPSILON = 1.E-10
!
!   Define FIELD(KB) = FIELD(KB-1)
!
  FIELD(:,:,KB) = FIELD(:,:,KB-1)
  IF(.NOT.ADVECT_DENSITY)THEN
      FIELD = FIELD - FAVER
      IF(.NOT.IMPLICIT_ETA) FIELDOLD = FIELDOLD - FAVER
  END IF
  IF(.NOT.IMPLICIT_ETA) DELFIELD = 0.
  IF(IMPLICIT_ETA) FIELDOLD = FIELD  
!
  XFLUX = 0.
  XFLUXUP = 0.
  XFLUXLW = 0.
  XFLUXD = 0.
  YFLUX = 0.
  YFLUXUP = 0.
  YFLUXLW = 0.
  YFLUXD = 0.
  ZFLUX = 0.
  ZFLUXUP = 0.
  ZFLUXLW = 0.
!
!   Compute F*U*D in U-points of the
!   Arakawa C-grid.
!
!   Compute the sum of the advective and diffusive
!   fluxes  in X-direction.
!
  IF (RIVLOG) THEN
     DO RIV = 1,NUMC
         I = IRIVER(RIV) ;  J = JRIVER(RIV)    
         CALL GLOBALIND2LOCAL(I,J,INBLOCK)     
         IF(INBLOCK) THEN         
             SELECT CASE(ICODE)
             CASE(1)
                 FIELDRIVER = SRIVER(riv)
             CASE(2)
                 FIELDRIVER = TRIVER(riv)
             CASE(3:MAXNUMCONC+2)
                 FIELDRIVER = CRIVER(ICODE-2,RIV)
             CASE DEFAULT
                 WRITE(*,*)'ERROR IN SUPERBEEF'
                 STOP
             END SELECT
             
             FIELDRIVER = FIELDRIVER - FAVER
             
             DO K = KRIVER1(riv),KRIVER2(riv)
                 FRAC = DZ(K)/DZRIVER(riv)
                 FLUX = FRAC*FRIVER(riv)*FIELDRIVER 
                 IF (CODERIVER(riv).EQ.1) THEN
                     XFLUX(i,j,K) = FLUX
                 ELSEIF (CODERIVER(riv).EQ.2) THEN
                     YFLUX(i,j,K) = FLUX
                 ELSEIF (CODERIVER(riv).EQ.3) THEN
                     if(i+1.gt.i2) cycle                                        
                     XFLUX(i+1,j,K) = -FLUX
                 ELSEIF (CODERIVER(riv).EQ.4) THEN
                     if(j+1.gt.j2) cycle                                        
                     YFLUX(i,j+1,K) = -FLUX
              ELSE
                  WRITE(*,*)'ERROR IN SUPERBEEF-2'
                  STOP
              ENDIF
          END DO
         end if ! inblock
      END DO
  ENDIF
  

  DO K = 1,KB-1
      DO J=J_EQ_2,JM_MI_1
          DO I=I_EQ_2,I_EQ_IM
              IF (DUM(I,J).NE.0.) THEN
                  AREAT = DY*DUDZ(I,J,K)
                  AREATU = AREAT*UADV(I,J,K)
                  XFLUXD = AH(I,J,K)*AREAT*(FIELD(I,J,K)-FIELD(I-1,J,K))/DX
                  IF (UADV(I,J,K).GE.0.) THEN
                      XFLUXUP = AREATU*FIELD(I-1,J,K)         
                      NSIG = 1
                  ELSE
                      XFLUXUP = AREATU*FIELD(I,J,K)         
                      NSIG = -1
                  ENDIF
                  XFLUXLW = 0.5*AREATU*(FIELD(I,J,K)+FIELD(I-1,J,K)) - &
                       & 0.5*LAMBDAX*UADV(I,J,K)**2*(FIELD(I,J,K)-FIELD(I-1,J,K))*AREAT
                  gi=block%fi+i-1  !; gj=block%fj+j-1
        
                  IF ((GI.GT.2 .AND. GI.LE.IM-1) .or. periodic_ew) THEN
                      NUMER = (ABS(UADV(I-NSIG,J,K)) - LAMBDAX*UADV(I-NSIG,J,K)**2)*  &
                           &         DY*(DUDZ(I-NSIG,J,K)*(FIELD(I-NSIG,J,K)-FIELD(I-1-NSIG,J,K)))
                      DENUM = (ABS(UADV(I,J,K)) - LAMBDAX*UADV(I,J,K)**2)*  &
                           &         DY*(DUDZ(I,J,K)*(FIELD(I,J,K)-FIELD(I-1,J,K)))
                      IF (ABS(DENUM).GT.EPSILON) THEN
                          R = NUMER/DENUM
                          PHIR = MAX(0., MIN(2.*R,1.) , MIN(R,2.))
                      ELSE 
                          PHIR = 0.
                      ENDIF
                  ELSE
                      PHIR = 0.
                  ENDIF
                  XFLUX(I,J,K) = XFLUXUP + PHIR*(XFLUXLW-XFLUXUP)
                  XFLUX(I,J,K) = XFLUX(I,J,K) - XFLUXD
              ENDIF
          END DO
      END DO
  END DO
!
!   Compute the sum of the advective and diffusive
!   fluxes  in Y-direction.

IF(JM > 1)THEN
!
  DO K = 1,KB-1
      DO J=J_EQ_2,J_EQ_JM
          DO I=I_EQ_2,IM_MI_1
              IF (DVM(I,J).NE.0.) THEN
                  AREAT = DX*DVDZ(I,J,K)
                  AREATV = AREAT*VADV(I,J,K)
                  YFLUXD = AREAT*AH(I,J,K)*((FIELD(I,J,K)-FIELD(I,J-1,K))/DY)
                  IF (VADV(I,J,K).GE.0.) THEN
                      YFLUXUP = AREATV*FIELD(I,J-1,K)
                      NSIG = 1
                  ELSE
                      YFLUXUP = AREATV*FIELD(I,J,K)
                      NSIG = -1
                  ENDIF
                  YFLUXLW = 0.5*AREATV*(FIELD(I,J,K)+FIELD(I,J-1,K)) - &
                       & 0.5*LAMBDAY*VADV(I,J,K)**2*(FIELD(I,J,K)-FIELD(I,J-1,K))*AREAT
                  gj=block%fj+j-1
        
                  IF (GJ.GT.2 .AND. GJ.LE.JM-1) THEN
                      NUMER = (ABS(VADV(I,J-NSIG,K)) - LAMBDAY*VADV(I,J-NSIG,K)**2)*  &
                           &         DX*(DVDZ(I,J-NSIG,K)*(FIELD(I,J-NSIG,K)-FIELD(I,J-1-NSIG,K)))
                      DENUM = (ABS(VADV(I,J,K)) - LAMBDAY*VADV(I,J,K)**2)*  &
                           &         DX*(DVDZ(I,J,K)*(FIELD(I,J,K)-FIELD(I,J-1,K)))
                      IF (ABS(DENUM).GT.EPSILON) THEN
                          R = NUMER/DENUM
                          PHIR = MAX(0., MIN(2.*R,1.) , MIN(R,2.))
                      ELSE
                          PHIR = 0.
                      ENDIF
                  ELSE
                      PHIR = 0.
                  ENDIF
                  YFLUX(I,J,K) = YFLUXUP + PHIR*(YFLUXLW-YFLUXUP)
                  YFLUX(I,J,K) = YFLUX(I,J,K) - YFLUXD
              ENDIF
          END DO
      END DO
  END DO
  
  END IF
  
!
!   Compute the advective fluxes in sigma-direction.
!
  AREA = DX*DY

  DO K = 2,KB
      DO J=J_EQ_2,JM_MI_1
          DO I=I_EQ_2,IM_MI_1
              IF(FSM(I,J).NE.0.) THEN

                  AREATW = AREA*WADV(I,J,K)
                  IF (WADV(I,J,K).GE.0.) THEN
                      ZFLUXUP = AREATW*FIELD(I,J,K)
                      NSIG = -1
                  ELSE
                      ZFLUXUP = AREATW*FIELD(I,J,K-1)
                      NSIG = 1
                  ENDIF
                  ZFLUXLW = 0.5*AREATW*(FIELD(I,J,K-1)+FIELD(I,J,K)) - &
                       &    0.5*AREA*DT*WADV(I,J,K)**2*(FIELD(I,J,K-1)-FIELD(I,J,K))/(DZZ(K-1)*D(I,J))
                  IF ( (K.EQ.2 .AND. NSIG.EQ.-1) .OR. &
                       &      (K.EQ.KB .AND. NSIG.EQ.1) .OR. &
                       &      (K.GE.3 .AND. K.LE.KB-1)) THEN
                      NUMER = (ABS(WADV(I,J,K-NSIG)) - (DT/(DZZ(K-1-NSIG)*D(I,J)))* &
                           &         WADV(I,J,K-NSIG)**2)*  &
                           &         (FIELD(I,J,K-1-NSIG)-FIELD(I,J,K-NSIG))
                      DENUM = (ABS(WADV(I,J,K)) - (DT/(DZZ(K-1)*D(I,J)))* &
                           &         WADV(I,J,K)**2)*  &
                           &         (FIELD(I,J,K-1)-FIELD(I,J,K))
                      IF (ABS(DENUM).GT.EPSILON) THEN
                          R = NUMER/DENUM
                          PHIR = MAX(0., MIN(2.*R,1.) , MIN(R,2.))
                      ELSE
                          PHIR = 1.
                      ENDIF
                  ELSE
                      PHIR = 1.
                  ENDIF
!
                  ZFLUX(I,J,K) = ZFLUXUP + PHIR*(ZFLUXLW-ZFLUXUP)
              ENDIF
          END DO
      END DO
  END DO
!
!  If UPWELL, then also specify a flux through the sea bed.
!
  IF (UPWELL) THEN
      DO IW = 1,NUMWELL
          I=IWELL(IW) ;  J=JWELL(IW)
          CALL GLOBALIND2LOCAL(I,J,INBLOCK)          
          IF(INBLOCK) THEN
              SELECT CASE(ICODE)
              CASE(1)
                  FIELDWELL = SWELL(IW)
              CASE(2)
                  FIELDWELL = TWELL(IW)
              CASE(3:MAXNUMCONC+2)
                  FIELDWELL = CWELL(ICODE-2,IW)
              CASE DEFAULT
                  WRITE(*,*)'ERROR IN SUPERBEEF'
                  STOP
              END SELECT
              FIELDWELL = FIELDWELL - FAVER
              ZFLUX(I,J,KB) = FWELL(IW)*FIELDWELL
          END IF
      END DO
  ENDIF
!
!
!   Take a step to timestep n+1  by adding up the fluxes in and out of each cell.
!
  call tag_ghost(xflux, east, 1)
  call tag_ghost(yflux, north, 1)
  call update_ghost

  DO K = 1,KB-1
      DO J=J_EQ_2,JM_MI_1
          DO I=I_EQ_2,IM_MI_1
              IF (FSM(I,J).GT.0.0) THEN
                  NETFLUX = (ZFLUX(I,J,K)-ZFLUX(I,J,K+1)) + &
                       &            XFLUX(I+1,J,K) - XFLUX(I,J,K) + &
                       &            YFLUX(I,J+DJ,K) - YFLUX(I,J,K)
                  VOLUME = DX*DY*DZ(K)*(H(I,J)+ETA(I,J))
                  VOLUMEP = DX*DY*DZ(K)*(H(I,J)+ETAP(I,J))
                  FIELD(I,J,K) = (FIELDOLD(I,J,K)*VOLUMEP - DT*NETFLUX)/VOLUME
              ENDIF
          END DO
      END DO
  END DO
!
  IF(.NOT. IMPLICIT_ETA) DELFIELD = FIELD - FIELDOLD
  
  IF(.NOT.ADVECT_DENSITY)THEN
      FIELD = FIELD + FAVER
      IF(.NOT. IMPLICIT_ETA) FIELDOLD = FIELDOLD + FAVER
  END IF

! restore AH
  if(icode > 2) then
      if(associated(co(icode-2)%ah)) then
          ah => tmp3d
      end if
  end if
  
  CALL DUMP_STATE("superbeef")

  DEALLOCATE(XFLUX,YFLUX,ZFLUX)
  
END SUBROUTINE SUPERBEEF
