SUBROUTINE SUPERBEEUV
!
!           SUPERBEEUV computes the following terms from
!           the moment equations on flux form:
!
!           - d(U*U*D)/d(X) - d(U*V*D)/d(Y) -  d(U*W)/d(SIGMA)
!  and
!           - d(V*U*D)/d(X) - d(V*V*D)/d(Y) -  d(V*W)/d(SIGMA)
!
!           from timestep n to timestep n+1
!           with a superbee limiter TVD scheme. (Roe and Sweby)
!
!           d(*) is the differential operator,
!           D is the dynamic depth,
!           T is time,
!           X and Y are the horisontal coordinates,
!           U is the horisontal velocity in X-direction,
!           V is the horisontal velocity in Y-direction,
!           SIGMA is the vertical coordinate,
!           AM is the horizontal eddy viscosity,
!           W is the vertical velocity. 
! The contributions from these terms are added to the
! contributions from the internal pressure terms in the
! arrays DELTAU and DELTAV.
! This means that SUPERBEEUV must be called after INTERNAL
! and before MODESPLIT.
!
!***ROUTINES CALLED-NONE
!***END PROLOGUE
!
!   Global variables.
!
  USE STATE
  USE MOD_BOUND
  USE ETAIMPLICIT
  IMPLICIT NONE
!
!   Local variables.
!
!   XFLUXUP,YFLUXUP,ZFLUXUP are used to store upstream advective fluxes.
!   XFLUXLW,YFLUXLW,ZFLUXLW are used to store lax-wendroff advective fluxes.
!   XFLUX,YFLUX,ZFLUX are used to store the final fluxes in
!                     the three space directions
!
! ..
! .. Local Arrays ..
REAL, POINTER :: XFLUX(:,:,:), YFLUX(:,:,:)
REAL :: XFLUXLW, XFLUXUP, &
        YFLUXLW, YFLUXUP, &
        ZFLUX(I1:I2,J1:J2,KB), ZFLUXLW, ZFLUXUP, &
        FIELDD(I1:I2,J1:J2,KB) !, DELTAI(I1:I2,J1:J2,KB)
!
  REAL AREA,NETFLUX,LAMBDAX,LAMBDAY
  REAL R,PHIR,NUMER,DENUM,EPSILON,AREATU,AREATV,AREATW
  REAL UP,UPSIG
  INTEGER I,J,K,NSIG,DJ
  REAL SPEED
  INTEGER RIV,GI,GJ
  LOGICAL INBLOCK
!
  ALLOCATE(XFLUX(I1:I2,J1:J2,KB), YFLUX(I1:I2,J1:J2,KB))
  LAMBDAX = DT/DX
  LAMBDAY = DT/DY
  EPSILON = 1.E-10
  
  DJ=1
  IF(JM == 1) DJ=0
  
!
!
!  First the advective and diffusive fluxes for the U equation in X-direction.
!
!  River specific
!
  IF (RIVLOG) THEN
  DO RIV = 1,NUMC
      
     i=iriver(riv)  ;  j=jriver(riv)

     call globalind2local(i,j,inblock)

     if(inblock) then         
         IF (CODERIVER(RIV).EQ.1) THEN
             SPEED = FRIVER(RIV)/(DY*DZRIVER(RIV)*H(I,J))
             DO K = KRIVER1(RIV),KRIVER2(RIV)
                 U(I,J,K) = SPEED
                 UADV(I,J,K) = SPEED
              END DO
         ELSEIF (CODERIVER(RIV).EQ.2) THEN
             SPEED = FRIVER(RIV)/(DX*DZRIVER(RIV)*H(I,J))
             DO K = KRIVER1(RIV),KRIVER2(RIV)
                 V(I,J,K) = SPEED
                 VADV(I,J,K) = SPEED
              END DO
          ELSEIF (CODERIVER(RIV).EQ.3) THEN
             if(i+1.gt.i2) cycle
             SPEED = FRIVER(RIV)/(DY*DZRIVER(RIV)*H(I,J))
             DO K = KRIVER1(RIV),KRIVER2(RIV)
                 U(I+1,J,K) = -SPEED
                 UADV(I+1,J,K) = -SPEED
              END DO
         ELSEIF (CODERIVER(RIV).EQ.4) THEN
             if(j+1.gt.j2) cycle
             SPEED = FRIVER(RIV)/(DX*DZRIVER(RIV)*H(I,J))
             DO K = KRIVER1(RIV),KRIVER2(RIV)
                 V(I,J+1,K) = -SPEED
                 VADV(I,J+1,K) = -SPEED
              END DO
          ELSE
              WRITE(*,*)'ERROR IN SUPERBEEUV-1'
              STOP
          ENDIF
    end if ! inblock
    
      END DO
  ENDIF


  XFLUX = 0.
  XFLUXUP = 0.
  XFLUXLW = 0.
  FIELDD = 0.

  DO K = 1,KB-1
      DO J=J1,J2
          DO I=I1,I2
              FIELDD(I,J,K) = U(I,J,K)*DU(I,J)
          END DO
      END DO
  END DO


  DO K = 1,KB-1          
      DO J=J_EQ_2,JM_MI_1
          DO I=1,IM_MI_1
              IF (FSM(I,J).NE.0.) THEN
                  gi=block%fi+i-1  !; gj=block%fj+j-1       
        
                  UP = 0.5*(UADV(I,J,K)+UADV(I+1,J,K))
                  AREATU = D(I,J)*UP
                  IF (UP.GE.0.) THEN
                      XFLUXUP = AREATU*U(I,J,K)
                      IF (GI.GE.2 .or. periodic_ew) THEN
                          UPSIG = 0.5*(UADV(I-1,J,K)+UADV(I,J,K))
                          NSIG = 1
                      ENDIF
                  ELSE
                      XFLUXUP = AREATU*U(I+1,J,K)
                      IF (GI.LE.IM-2 .or. periodic_ew) THEN
                          UPSIG = 0.5*(UADV(I+1,J,K)+UADV(I+2,J,K))
                          NSIG = -1
                      ENDIF
                  ENDIF
                  XFLUXLW = AREATU*UP - &
                       & 0.5*LAMBDAX*UP*UP*(fieldd(i+1,j,k)-fieldd(i,j,k) ) !DELTAI(I+1,J,K)
                  IF ((GI.GE.2 .AND. GI.LE.IM-2) .or. periodic_ew) THEN
                      NUMER = (ABS(UPSIG) - LAMBDAX*UPSIG**2)*  &
                           &         (D(I-NSIG,J)*(U(I+1-NSIG,J,K)-U(I-NSIG,J,K)))
                      DENUM = (ABS(UP) - LAMBDAX*UP**2)*  &
                           &         (D(I,J)*(U(I+1,J,K)-U(I,J,K)))
                      IF (ABS(DENUM).GT.EPSILON) THEN
                          R = NUMER/DENUM
                          PHIR = MAX(0., MIN(2.*R,1.) , MIN(R,2.))
                      ELSE
                          PHIR = 0.
                      ENDIF
                  ELSE
                      PHIR = 0.
                  ENDIF
                  XFLUX(I,J,K) = XFLUXUP + PHIR*(XFLUXLW-XFLUXUP)
                  XFLUX(I,J,K) = XFLUX(I,J,K)*DY
              ENDIF
          END DO
      END DO
  END DO
!
!  Compute the advective and diffusive fluxes for the U-equation in Y-direction.
!
  YFLUX = 0.
  YFLUXUP = 0.
  YFLUXLW = 0.
  
  IF(JM > 1)THEN 

  DO K = 1,KB-1
      DO J=1,JM_MI_1
          DO I=I_EQ_2,IM_MI_1        
              IF (DVM(I-1,J+1).NE.0. .AND. DVM(I,J+1).NE.0.) THEN
                  gj=block%fj+j-1       
                  
                  UP = 0.5*(VADV(I-1,J+1,K)+VADV(I,J+1,K))
                  AREATV = 0.5*(DV(I-1,J+1)+DV(I,J+1))*UP
                  IF (UP.GE.0.) THEN
                      YFLUXUP = AREATV*U(I,J,K)
                      UPSIG = 0.5*(VADV(I-1,J,K)+VADV(I,J,K))
                      NSIG = 1
                  ELSE
                      YFLUXUP = AREATV*U(I,J+1,K)
                      IF (GJ.LE.JM-2) THEN
                          UPSIG = 0.5*(VADV(I-1,J+2,K)+VADV(I,J+2,K))
                          NSIG = -1
                      ENDIF
                  ENDIF
                  YFLUXLW = 0.5*AREATV*(U(I,J,K)+U(I,J+1,K)) - &
                       & 0.5*LAMBDAY*UP*UP*(fieldd(i,j+1,k)-fieldd(i,j,k))   !DELTAI(I,J+1,K)
                  IF (GJ.GE.2 .AND. GJ.LE.JM-2) THEN
                      NUMER = (ABS(UPSIG) - LAMBDAY*UPSIG**2)*  &
                           &         0.5*(DV(I-1,J+1-NSIG)+DV(I,J+1-NSIG))* &
                           &         (U(I,J+1-NSIG,K) - U(I,J-NSIG,K))
                      DENUM = (ABS(UP) - LAMBDAY*UP**2)*  &
                           &         0.5*(DV(I-1,J+1)+DV(I,J+1))*(U(I,J+1,K)-U(I,J,K))
                      IF (ABS(DENUM).GT.EPSILON) THEN
                          R = NUMER/DENUM
                          PHIR = MAX(0., MIN(2.*R,1.) , MIN(R,2.))
                      ELSE
                          PHIR = 0.
                      ENDIF
                  ELSE
                      PHIR = 0.
                  ENDIF
                  YFLUX(I,J,K) = YFLUXUP + PHIR*(YFLUXLW-YFLUXUP)
                  YFLUX(I,J,K) = YFLUX(I,J,K)*DX
              ENDIF
          END DO
      END DO
  END DO
  
  END IF
  
!
!  Compute the advective fluxes in sigma-direction for the U-equation.
!
  ZFLUX = 0.
  ZFLUXUP = 0.
  ZFLUXLW = 0.
  AREA = DX*DY
  if(jm==1) area=dx*dx
  
  DO K=2,KB-1
      DO J=J_EQ_2,JM_MI_1
          DO I=I_EQ_2,IM_MI_1
              IF (DUM(I,J).NE.0.) THEN
                  UP = 0.5*(WADV(I,J,K)+WADV(I-1,J,K))
                  AREATW = AREA*UP
                  IF (UP.GE.0.) THEN
                      ZFLUXUP = AREATW*U(I,J,K)
                      UPSIG =  0.5*(WADV(I,J,K+1)+WADV(I-1,J,K+1))
                      NSIG = -1
                  ELSE
                      ZFLUXUP = AREATW*U(I,J,K-1)
                      UPSIG =  0.5*(WADV(I,J,K-1)+WADV(I-1,J,K-1))
                      NSIG = 1
                  ENDIF
                  ZFLUXLW = 0.5*AREATW*(U(I,J,K-1)+U(I,J,K)) - &
                       &    0.5*AREA*DT*UP*UP*(U(I,J,K-1)-U(I,J,K))/(DZZ(K-1)*DU(I,J))
                  IF ( (K.EQ.2 .AND. NSIG.EQ.-1) .OR. &
                       &      (K.GE.3 .AND. K.LE.KB-1)) THEN
                      NUMER = (ABS(UPSIG) - (DT/(DZZ(K-1-NSIG)*DU(I,J)))* &
                           &         UPSIG**2)*  &
                           &         (U(I,J,K-1-NSIG)-U(I,J,K-NSIG))
                      DENUM = (ABS(UP) - (DT/(DZZ(K-1)*DU(I,J)))* &
                           &         UP**2)*  &
                           &         (U(I,J,K-1)-U(I,J,K))
                      IF (ABS(DENUM).GT.EPSILON) THEN
                          R = NUMER/DENUM
                          PHIR = MAX(0., MIN(2.*R,1.) , MIN(R,2.))
                      ELSE
                          PHIR = 0.
                      ENDIF
                  ELSE
                      PHIR = 1.
                  ENDIF

                  ZFLUX(I,J,K) = ZFLUXUP + PHIR*(ZFLUXLW-ZFLUXUP)
              ENDIF
          END DO
      END DO
  END DO

!
!   Take a step to timestep n+1 for the U-equation
!   by adding up the fluxes in and out of the cell.
!
call tag_ghost(xflux, west, 1)
call tag_ghost(yflux, south,1)
call update_ghost


DO K = 1,KB-1
    DO J=J_EQ_2,JM_MI_1
        DO I=I_EQ_2,IM_MI_1
            IF (DUM(I,J).GT.0.0) THEN
                NETFLUX = XFLUX(I,J,K) - XFLUX(I-1,J,K) + &
                     &             YFLUX(I,J,K) - YFLUX(I,J-DJ,K) + &
                     &             (ZFLUX(I,J,K) - ZFLUX(I,J,K+1))/DZ(K)
                DELTAU(I,J,K) =  DELTAU(I,J,K) - NETFLUX/AREA
            ENDIF
        END DO
    END DO
END DO
!
!
!   Then the momentum equation in V.
!
!   First the sum of the advective and diffusive fluxes in X-direction.
!
  XFLUX = 0.
  XFLUXUP = 0.
  XFLUXLW = 0.
  FIELDD = 0.

  DO K = 1,KB-1
      DO J=J1,J2
          DO I=I1,I2
              FIELDD(I,J,K) = V(I,J,K)*DV(I,J)
          END DO
      END DO
  END DO
!

  DO K = 1,KB-1
      DO J=J_EQ_2,JM_MI_1
          DO I=1,IM_MI_1
              IF (DUM(I+1,J-dj).NE.0. .AND. DUM(I+1,J).NE.0.) THEN
                  gi=block%fi+i-1  !; gj=block%fj+j-1       
                  
                  UP = 0.5*(UADV(I+1,J-dj,K)+UADV(I+1,J,K))
                  AREATU = 0.5*(DU(I+1,J-dj)+DU(I+1,J))*UP
                  IF (UP.GE.0.) THEN
                      XFLUXUP = AREATU*V(I,J,K)
                      UPSIG = 0.5*(UADV(I,J-dj,K)+UADV(I,J,K))
                      NSIG = 1
                  ELSE
                      XFLUXUP = AREATU*V(I+1,J,K)
                      IF (GI.LE.IM-2 .or. periodic_ew) THEN
                          UPSIG = 0.5*(UADV(I+2,J-dj,K)+UADV(I+2,J,K))
                          NSIG = -1
                      ENDIF
                  ENDIF
                  XFLUXLW = 0.5*AREATU*(V(I,J,K)+V(I+1,J,K)) - &
                       &   0.5*LAMBDAX*UP*UP*(fieldd(i+1,j,k)-fieldd(i,j,k))  !DELTAI(I+1,J,K)
                  IF ((GI.GE.2 .AND. GI.LE.IM-2) .or. periodic_ew) THEN
                      NUMER = (ABS(UPSIG) - LAMBDAX*UPSIG**2)*  &
                           &         0.5*(DU(I+1-NSIG,J-dj)+DU(I+1-NSIG,J))* &
                           &         (V(I+1-NSIG,J,K) - V(I-NSIG,J,K))
                      DENUM = (ABS(UP) - LAMBDAX*UP**2)*  &
                           &         0.5*(DU(I+1,J-dj)+DU(I+1,J))*(V(I+1,J,K)-V(I,J,K))
                      IF (ABS(DENUM).GT.EPSILON) THEN
                          R = NUMER/DENUM
                          PHIR = MAX(0., MIN(2.*R,1.) , MIN(R,2.))
                      ELSE
                          PHIR = 0.
                      ENDIF
                  ELSE
                      PHIR = 0.
                  ENDIF
                  XFLUX(I,J,K) = XFLUXUP + PHIR*(XFLUXLW-XFLUXUP)
                  XFLUX(I,J,K) = XFLUX(I,J,K)*DY
              ENDIF
          END DO
      END DO
  END DO
!
!  Compute the advective and diffusive fluxes for the V-equation in Y-direction.
!
  YFLUX = 0.
  YFLUXUP = 0.
  YFLUXLW = 0.

 IF(JM > 1)THEN

  DO K = 1,KB-1
      DO J=1,JM_MI_1
          DO I=I_EQ_2,IM_MI_1
              IF (FSM(I,J).NE.0.) THEN
                  gj=block%fj+j-1
        
                  UP = 0.5*(VADV(I,J,K)+VADV(I,J+1,K))
                  AREATV = D(I,J)*UP
                  IF (UP.GE.0.) THEN
                      YFLUXUP = AREATV*V(I,J,K)
                      IF (GJ.GE.2) THEN
                          UPSIG = 0.5*(VADV(I,J-1,K)+VADV(I,J,K))
                          NSIG = 1
                      ENDIF
                  ELSE
                      YFLUXUP = AREATV*V(I,J+1,K)
                      IF (GJ.LE.JM-2) THEN
                          UPSIG = 0.5*(VADV(I,J+1,K)+VADV(I,J+2,K))
                          NSIG = -1
                      ENDIF
                  ENDIF
                  YFLUXLW = AREATV*UP - &
                       & 0.5*LAMBDAY*UP*UP*(FIELDD(I,J+1,K)-FIELDD(I,J,K))   !DELTAI(I,J+1,K)
                  IF (GJ.GE.2 .AND. GJ.LE.JM-2) THEN
                      NUMER = (ABS(UPSIG) - LAMBDAY*UPSIG**2)*  &
                           &         (D(I,J-NSIG)*(V(I,J+1-NSIG,K)-V(I,J-NSIG,K)))
                      DENUM = (ABS(UP) - LAMBDAY*UP**2)*  &
                           &         (D(I,J)*(V(I,J+1,K)-V(I,J,K)))
                      IF (ABS(DENUM).GT.EPSILON) THEN
                          R = NUMER/DENUM
                          PHIR = MAX(0., MIN(2.*R,1.) , MIN(R,2.))
                      ELSE
                          PHIR = 0.
                      ENDIF
                  ELSE
                      PHIR = 0.
                  ENDIF
                  YFLUX(I,J,K) = YFLUXUP + PHIR*(YFLUXLW-YFLUXUP)
                  YFLUX(I,J,K) = YFLUX(I,J,K)*DX
              ENDIF
          END DO
      END DO
  END DO
  
  END IF
!
!  Compute the advective fluxes for the V-equation in sigma-direction.
!
  ZFLUX = 0.
  ZFLUXUP = 0.
  ZFLUXLW = 0.
  AREA = DX*DY
  IF(JM==1) AREA=DX*DX

  DO K = 2,KB-1
      DO J=J_EQ_2,JM_MI_1
          DO I=I_EQ_2,IM_MI_1
              IF (DVM(I,J).NE.0.) THEN
                  UP = 0.5*(WADV(I,J,K)+WADV(I,J-DJ,K))
                  AREATW = AREA*UP
                  IF (UP.GE.0.) THEN
                      ZFLUXUP = AREATW*V(I,J,K)
                      UPSIG =  0.5*(WADV(I,J,K+1)+WADV(I,J-DJ,K+1))
                      NSIG = -1
                  ELSE
                      ZFLUXUP = AREATW*V(I,J,K-1)
                      UPSIG =  0.5*(WADV(I,J,K-1)+WADV(I,J-DJ,K-1))
                      NSIG = 1
                  ENDIF
                  ZFLUXLW = 0.5*AREATW*(V(I,J,K-1)+V(I,J,K)) - &
                       &    0.5*AREA*DT*UP*UP*(V(I,J,K-1)-V(I,J,K))/(DZZ(K-1)*DV(I,J))
                  IF ( (K.EQ.2 .AND. NSIG.EQ.-1) .OR. &
                       &      (K.GE.3 .AND. K.LE.KB-1)) THEN
                      NUMER = (ABS(UPSIG) - (DT/(DZZ(K-1-NSIG)*DV(I,J)))* &
                           &         UPSIG**2)*  &
                           &         (V(I,J,K-1-NSIG)-V(I,J,K-NSIG))
                      DENUM = (ABS(UP) - (DT/(DZZ(K-1)*DV(I,J)))* &
                           &         UP**2)*  &
                           &         (V(I,J,K-1)-V(I,J,K))
                      IF (ABS(DENUM).GT.EPSILON) THEN
                          R = NUMER/DENUM
                          PHIR = MAX(0., MIN(2.*R,1.) , MIN(R,2.))
                      ELSE
                          PHIR = 0.
                      ENDIF
                  ELSE
                      PHIR = 1.
                  ENDIF
!
                  ZFLUX(I,J,K) = ZFLUXUP + PHIR*(ZFLUXLW-ZFLUXUP)
              ENDIF
          END DO
      END DO
  END DO

!
!   Take a step to timestep n+1 for the V-equation by adding up the
!   fluxes in and out of the cell.
!
  call tag_ghost(xflux,west,1)
  call tag_ghost(yflux,south,1)
  call update_ghost


  DO K = 1,KB-1
      DO J = J_eq_2,JM_mi_1
          DO I =I_eq_2,IM_mi_1
              IF (DVM(I,J).GT.0.0) THEN
                  NETFLUX = XFLUX(I,J,K) - XFLUX(I-1,J,K) + &
                       &             YFLUX(I,J,K) - YFLUX(I,J-DJ,K) + &
                       &             (ZFLUX(I,J,K) - ZFLUX(I,J,K+1))/DZ(K)
                  DELTAV(I,J,K) =  DELTAV(I,J,K) - NETFLUX/AREA
              ENDIF
          END DO
      END DO
  END DO


  IF(IMPLICIT_ETA) THEN
! we assume deltau and v both were 0 before the assignment above. 
      DO K = 1,KB-1
          DO J = J_eq_2,JM_mi_1
              DO I =I_eq_2,IM_mi_1
                  U(I,J,K) = U(I,J,K) + DT*DELTAU(I,J,K)
                  V(I,J,K) = V(I,J,K) + DT*DELTAV(I,J,K)
              END DO
          END DO
      END DO
  END IF
  
!
!  River specific
!
  IF (RIVLOG) THEN
      DO RIV = 1,NUMC
          i=iriver(riv) ;  j=jriver(riv)
          
          call globalind2local(i,j,inblock)
          
          if(inblock) then         
              IF (CODERIVER(riv).EQ.1) THEN
                  DO K = KRIVER1(riv),KRIVER2(riv)
                      U(i,j,K) = 0.
                      UADV(i,j,K) = 0.
                  END DO
              ELSEIF (CODERIVER(riv).EQ.2) THEN
                  DO K = KRIVER1(riv),KRIVER2(riv)
                      V(i,j,K) = 0.
                      VADV(i,j,K) = 0.
                  END DO
              ELSEIF (CODERIVER(riv).EQ.3) THEN
                  if(i+1.gt.i2) cycle                  
                  DO K = KRIVER1(riv),KRIVER2(riv)
                      U(i+1,j,K) = 0.
                      UADV(i+1,j,K) = 0.
                  END DO
              ELSEIF (CODERIVER(riv).EQ.4) THEN
                  if(j+1.gt.j2) cycle                  
                  DO K = KRIVER1(riv),KRIVER2(riv)
                      V(i,j+1,K) = 0.
                      VADV(i,j+1,K) = 0.
                  END DO
              ELSE
                  WRITE(*,*)'ERROR IN SUPERBEEUV-2'
                  STOP
              ENDIF
          end if
      END DO
  ENDIF
  
  CALL DUMP_STATE("superbeeuv")
  DEALLOCATE(XFLUX, YFLUX)
END SUBROUTINE SUPERBEEUV
