module time_util
  use, intrinsic :: iso_c_binding, only: c_int
  
  implicit none
  
! ndate = year month day hour minute second  
  integer ndate(6)

! timezone in range [-12, +12] hours
  real timezone

! the initial time of the simulation, if not given in the setup, it
! will be 0
  integer :: initial_time(6) = 0
  
! the model time in seconds since the beginning of the simulation
  real(8) time
  
! the tictac clock use integer seconds as the counter, while time has full precision.
! store the additional sub-second time from the previous count here.
  real(8) prev_hundreds


!  seconds should be unsigned int, but Fortran only has signed
!  ones. seems to work for what we need it for, which is waiting a
!  second or two.
!
!  interface
!     integer(c_int) function csleep(seconds) bind(C, name="sleep")   ! linux
!     integer(c_int) function wsleep(seconds) bind(C, name="Sleep")   ! windows
!       integer (c_int), intent (in), value :: seconds
!     end function
!  end interface

  
contains


  subroutine fsleep(s)
    integer s, ierr
!    write(*,*) " sleeping",s,"seconds"
!    ierr = wsleep(1000*s)
!    ierr = csleep(s)
  end subroutine fsleep
  

  function clock()
!  
! return the "wall clock time", since some system dependent time, in seconds.
!
    integer, parameter :: dp=8
    real(kind=dp) clock
    integer cnt,cnt_rate

    call system_clock(cnt, cnt_rate)
    clock = real(cnt,dp)/real(cnt_rate,dp)
  end function clock


  function timediff(date1, date2)
!    
! given two gregorian dates, return the elapsed time from date1 to
! date2 in seconds, by converting the dates to a julian date and
! subtract. not very accurate numerically, but probably sufficient for
! our needs. (when we deal with dates, sub second accuracy may not be an issue?)
!
    real timediff
    integer date1(6), date2(6)
    real(8) julian1, julian2

    call greg2julian(date1, julian1)
    call greg2julian(date2, julian2)
    timediff = (julian2-julian1)*86400.0
  end function timediff

  
  subroutine greg2julian(date, julian)
!    
! given a gregorian date, return the julian.
! equations from http://www.astro.uu.nl/~strous/AA/en/reken/juliaansedag.html
! the date 2005-01-01T15:00:00 should give 2453372.125
!    
    implicit none
    integer date(6), j,m,d,h,min,sec
    real(8) c, day, julian
    
    j = date(1)
    m = date(2)
    d = date(3)
    h = date(4)
    min = date(5)
    sec = date(6)

    if(m < 3) then
        m = m+12
        j = j-1
    end if
    
    c = 2.0_8 - floor(real(j,8)/100.0_8) + floor(real(j,8)/400.0_8)
    
! add hours/24 to the day    
    day = real(d,8) + (real(h,8) + real(min,8)/60.0_8 + real(sec,8)/3600.0_8)/24.0_8
    
    julian = floor(1461.0_8*(real(j,8) + 4716.0_8)/4.0_8) + floor(153.0_8*(real(m,8)+1.0_8)/5.0_8) + day + c - 1524.5_8
  end subroutine greg2julian
  
  
  subroutine get_iso8601(string, datearr)
!
! string has a date in iso format, e.g.
!   2009-08-27T20:23:45.2
! read it into integer array "datearr"
!
    implicit none
    character(*) string
    integer datearr(6)
    character(len(string)) stripped_string

    stripped_string = adjustl(string)

    read(stripped_string(1:4),*) datearr(1)
    read(stripped_string(6:7),*) datearr(2)
    read(stripped_string(9:10),*) datearr(3)
    read(stripped_string(12:13),*) datearr(4)
    read(stripped_string(15:16),*) datearr(5)
! allow skipping seconds
    if(len(trim(stripped_string)) > 17) then
        read(stripped_string(18:19),*) datearr(6)
    else
        datearr(6) = 0
    end if
  end subroutine get_iso8601


  function is_leap(year)
!   
! The Gregorian calendar adds a 29th day to February in all years evenly
! divisible by 4, except for centennial years (those ending in -00),
! which receive the extra day only if they are evenly divisible by 400.
! There has been an additional modification of the rule so that years
! divisible by 4000 are not a leap year.
!
    logical is_leap
    integer year

    is_leap=.false.

    if(mod(year,4).ne.0) return
    if(mod(year,100).ne.0) then
        is_leap=.true.
        return
    end if
    if(mod(year,400).ne.0) return
    if(mod(year,4000).ne.0) then
        is_leap=.true.
        return
    else
        return
    end if
  end function is_leap
  

  subroutine tictac(increment, date, simsecond)
!
! increment the time given as date = year-month-day-hour-minute-second by
! increment seconds. simsecond is the accumulated time
! in seconds relative to its initial value since first call.
!
! non active args:    
!  if a new year, month, day etc. has been
!  reached, each of the five last arguments respectively
!  will be set to .true..
!    
    implicit none
    real(8) increment, simsecond, hundreds
    integer date(6), tictac_year,tictac_month,tictac_day,tictac_hour,tictac_minute,tictac_second
    
    integer :: nd(12) = (/31,28,31,30,31,30,31,31,30,31,30,31/)
    logical error
    integer s, intincrement

    tictac_year=   date(1)
    tictac_month=  date(2)
    tictac_day=    date(3)
    tictac_hour=   date(4)
    tictac_minute= date(5)
    tictac_second= date(6)
    
    if(increment <= 0)then
        write(*,*)'subroutine tictac error: increment <= 0'
        stop
    end if

    error = tictac_month<1 .or.tictac_month>12 .or. &
         tictac_day > nd(tictac_month) .or. tictac_day<1 .or.&
         tictac_hour<0 .or. tictac_hour>23 .or. &
         tictac_minute<0 .or. tictac_minute>59 .or. &
         tictac_second<0 .or. tictac_second>59

    if(error)then
        write(*,*)'subroutine tictac error: this time is not possible:'
        write(*,'(i4,"-",i2.2,"-",i2.2,1x,i2.2,":",i2.2,":",i2.2)') &
             tictac_year, tictac_month, tictac_day, tictac_hour, tictac_minute, tictac_second
        stop
    end if

    simsecond = simsecond + increment
    intincrement = floor(increment)
    hundreds = increment - intincrement
    
    if(hundreds + prev_hundreds >= 1.0_8) then    
        intincrement = intincrement + 1
        prev_hundreds = prev_hundreds + hundreds - 1.0_8
    else
        prev_hundreds = prev_hundreds + hundreds
    end if
    
    do s=1,intincrement
        if(tictac_second.eq.59) then
            tictac_second=0

            if(tictac_minute.eq.59) then
                tictac_minute=0

                if(tictac_hour.eq.23) then
                    tictac_hour=0

                    !! uncomment to take leap years into account
                    ! nd(2)=28
                    ! if(is_leap(year)) nd(2)=29

                    if(tictac_day.eq.nd(tictac_month)) then
                        tictac_day=1

                        if(tictac_month.eq.12)then
                            tictac_month=1
                            tictac_year=tictac_year+1
                        else
                            tictac_month=tictac_month+1
                        end if
                    else
                        tictac_day=tictac_day+1
                    end if
                else
                    tictac_hour=tictac_hour+1
                end if
            else
                tictac_minute = tictac_minute + 1
            end if
        else
            tictac_second=tictac_second + 1
        end if
    end do

    date = (/tictac_year, tictac_month, tictac_day, tictac_hour, tictac_minute, tictac_second/)
  end subroutine tictac

end module time_util


