!
! usage: 2 ways
!
! 1) Using keywords starting with "timeseries_" in the setup file
!
! See documentation for keywords starting with "timeseries_" in
! setup2.f90 or via running bomhelp.py
!
! the timeseries_ keywords adds timeseries objects to the list "ts"
! below. at the end of a run, or at specified intervals, all
! timeseries are dumped into its column in a single file, that can be
! "cat'ed" together in the post processing to create a complete time
! series file for the run.
!
! 2) Calling the routines directly
!
! Initialization; build a list of points in which a field should be sampled.
! we use bilinear interpolation in the horisontal, and nearest point in the vertical, 
! (since points normally are very close there.)
!
!   use timeseries
!   ...
!   if( istep==1) then  ! label    field  type    xpos     ypos  depth   gridcoos?
!      call add_ts_3dpos('U punkt1', u,   u_point, 100.0,   50.0,  -20.0, .true. )
!      call add_ts_3dpos('U punkt2', u,   u_point, 100.0,   50.0, -120.0, .true. )
!      call add_ts_3dpos('V',        v,   v_point, 200.0,  150.0, -220.0, .true. )
!      call add_ts_3dpos('U',        u,   u_point, 100.0,   50.0,   20.0, .true. )
!      call add_ts_2dpos('ETA',    eta, eta_point, 200.0,  150.0,         .true. )
!      call add_ts_scalar('Ekin:',ekin )
!      call add_ts_scalar('Mass', mass )
!      call add_ts_1dind(description, field, i)
!      call add_ts_2dind(description, field, i,j)
!      call add_ts_3dind(description, field, i,j,k)
!      ....
!   end if
!
! now, in the time stepping a loop a single
!
!   ...
!   call sample_timeseries
!   ...
!
! each timestep, is enough to do all interpolation and storage
! automatically. the actual data for timeseries N, timestep I is
! stored in TS(N)%DATA(I). at any time only a part of TS(N)%DATA(:) is
! actually allocated. the timeseries data is written to files in batches each
! checkpoint_interval'th timestep. since also other output files are written 
! separately, no data from before a restart will be touched, and it will
! after a complete run be possible to join the timeseries files perfectly.
!
! example: use "cat" to join the ts-batchXXX.dat files:
!    cat ts-batch???.dat > timeseries.dat
!

module timeseries
 
  use state
  use control_parameters, only : first_step,do_checkpoint,restart,checkpoint_dir,advect_density,nonhyd,nonhyd_allocated
  use mod_mpi, only : i1,j1,master,myid,numproc,mpi_proc_null,gather,send,receive,&
       globalind2local,isend,irecv,waitall,mpi_integer,qmpi_comm_world,abort,barrier,message,&
       ghost_wi,ghost_wj
  use util, only : replace, binsearch2
  use grid_util, only : sph2gr
  implicit none

  private
  
  public ts, add_ts_scalar ,add_ts_speed ,add_ts_3dind, add_ts_2dind &
       ,add_ts_1dind ,add_ts_3dpos ,find_hor_pos, add_ts_2dpos &
       ,sample_timeseries, timeseries_interval, num_ts, local_num_ts, gatherts,&
       timeseries_fullprec, init_timeseries_density, store_timeseries, &
       timeseries_file_number, timeseries_dump_start, &
       deallocate_timeseries, timeseries_unformatted, timeseries_formatted, &
       add_timeseries, set_timeseries_format, timeseries_from_file, numseries, allocate_timeseries, &
       timeseries_dir

  character(100) :: timeseries_dir = "TIMESERIES/"
  
  logical :: timeseries_unformatted=.false. ! true if the timeseries files should written as binary streams (access="stream")
  logical :: timeseries_fullprec=.true.     ! true if the time series should be written with full precision.
  logical :: timeseries_formatted=.false.   ! true if the time series should be written with user defined format strings
  
  integer :: numseries=650                  ! the maximum number of timeseries
  
  integer :: timeseries_interval=10000      ! the number of timesteps per interval to dump to file
  integer :: timeseries_file_number=0       ! which file in the sequence to write to  
  integer :: timeseries_dump_start=0        ! the start timestep of the current interval to dump

  
  type tstype
     logical :: inblock=.false.             ! true if sample point is on block
     integer :: processor = mpi_proc_null   ! the processor holding the time series     
     character(80) :: label=" "             ! descriptive label for timeseries
     character(20) :: format="(f0.6)"       ! default edit descriptor for timeseries_formatted
     real, allocatable :: data(:)           ! the timeseries itself
     logical :: is_scalar=.false.           ! is it a scalar value?
     logical :: is_speed=.false.            ! true if speed in i,j,k is to be computed
     real, pointer :: scalar => null()      ! pointer to the actual variable
     integer :: pointtype=0                 ! u-v or eta (centered) point?
     real, pointer, dimension(:,:) :: f2d => null()   ! pointer to 2d field from which to interpolate
     real, pointer, dimension(:,:,:) :: f3d => null() ! pointer to 3d field from which to interpolate
     real xpos,ypos,sampledepth             ! the position of the sampling
     real x,y                               ! box local coordinate w.r. sw corner; range [0.0,1.0]
     integer i,j,k, ci,cj
     logical :: grid_coo=.true.             ! coordinate of point not given as lon-lat
     real :: density=0.0                    ! = RHOAVER if we are plotting density (we then have to add RHOAVER+1000)
  end type tstype
  
  type(tstype), allocatable :: ts(:)        ! the timeseries
  integer :: num_ts=0                       ! the global number of timeseries
  integer :: local_num_ts=0                 ! the local
  integer, allocatable :: localseries(:)                 ! the indices for the local series

  
contains

  
  subroutine allocate_timeseries()
    if(master.and. debug > 1) write(*,'(2x,a,i0,a)') "allocating ",numseries," timeseries"
    if(allocated(ts)) then
        call deallocate_timeseries()
        deallocate(ts, localseries)
    end if
    allocate(ts(numseries), localseries(numseries))
  end subroutine allocate_timeseries
  
  
  subroutine init_timeseries_density(rhoaver)
    real rhoaver
    integer i
    
    do i=1,num_ts
        if(ts(i)%density > 0.0) ts(i)%density = rhoaver
    end do
  end subroutine init_timeseries_density

  
  subroutine deallocate_timeseries()
    integer i
    do i=1,numseries
        if(allocated(ts(i)%data)) deallocate(ts(i)%data)
    end do
  end subroutine deallocate_timeseries


  subroutine reallocate_timeseries(index1, index2)
    integer i, index1, index2
    
    do i=1,num_ts
        if(allocated(ts(i)%data)) then           
            deallocate(ts(i)%data)
            if(index2-index1 < 100000 .and. index1 <= index2) then
                allocate(ts(i)%data(index1:index2))
                if(debug > 2) write(*,'(a,i0,a,i0)') " Reallocating timeseries from ",index1," to ",index2
            else
                write(*,*) "WARNING: reallocation of timeseries not done, range looks wrong:",index1, index2
            end if
        end if
    end do    
  end subroutine reallocate_timeseries
  
  
  subroutine add_ts_scalar( description, scalar, on_block )
    implicit none
    character(len=*) description
    real, target :: scalar
    logical, optional :: on_block
    
    if(num_ts >= numseries ) call abort('ERROR: too many timeseries! increase parameter NUMSERIES')        

    num_ts = num_ts + 1

    if(numstep >= 0 .and. .not. allocated(ts(num_ts)%data)) then
        allocate(ts(num_ts)%data(timeseries_dump_start:timeseries_dump_start + timeseries_interval))
    else
        call message("could not allocate timeseries, check if numstep is set? numstep=", severity=2, ival=(/numstep/))
    end if
    
    ts(num_ts)%label(:len(description)) = description
    ts(num_ts)%is_scalar =.true.
    ts(num_ts)%scalar => scalar

    if(master)then
        ts(num_ts)%inblock = .true.
    else
        ts(num_ts)%inblock = .false.
    end if
    
! override inblock if desired
    if(present(on_block)) ts(num_ts)%inblock = on_block
  end subroutine add_ts_scalar

  
  subroutine add_ts_speed(description, i,j,k)
    implicit none
    character(len=*) description
    integer i,j,k
    
    if(num_ts >= numseries ) call abort('ERROR: too many timeseries! increase parameter NUMSERIES')

    num_ts=num_ts + 1
    if(numstep >= 0 .and. .not. allocated(ts(num_ts)%data)) & ! allocate(ts(num_ts)%data(numstep))
         allocate(ts(num_ts)%data(timeseries_dump_start:timeseries_dump_start+timeseries_interval))
    ts(num_ts)%label(:len(description))=description
    ts(num_ts)%is_speed =.true.
    ts(num_ts)%i = i
    ts(num_ts)%j = j
    ts(num_ts)%k = k

    call globalind2local(ts(num_ts)%i, ts(num_ts)%j, ts(num_ts)%inblock,ghost=.false.)
  end subroutine add_ts_speed

  
  subroutine add_ts_3dind(description, field, i,j,k)
    implicit none
    character(len=*) description  
    real, pointer, dimension(:,:,:) :: field
    integer i,j,k,li,lj

    if(num_ts >= numseries) call abort('ERROR: too many timeseries! increase parameter NUMSERIES')

    num_ts=num_ts + 1
    if(numstep >= 0) &!allocate(ts(num_ts)%data(numstep))
         allocate(ts(num_ts)%data(timeseries_dump_start:timeseries_dump_start+timeseries_interval))
    ts(num_ts)%label(:len(description)) = description
    ts(num_ts)%is_scalar = .true.

    li=i ; lj=j
    call globalind2local(li,lj, ts(num_ts)%inblock, ghost=.false.)
    
    if(ts(num_ts)%inblock) ts(num_ts)%scalar => field(li,lj,k)
  end subroutine add_ts_3dind


  subroutine add_ts_2dind(description, field, i,j)
    implicit none
    character(len=*) description  
    real, pointer, dimension(:,:) :: field
    integer i,j,li,lj

    if(num_ts >= numseries) call abort('ERROR: too many timeseries! increase parameter NUMSERIES')

    num_ts=num_ts + 1
    if(numstep >= 0) &!allocate(ts(num_ts)%data(numstep))
         allocate(ts(num_ts)%data(timeseries_dump_start:timeseries_dump_start+timeseries_interval))
    ts(num_ts)%label(:len(description)) = description
    ts(num_ts)%is_scalar = .true.
    
    li=i ; lj=j
    call globalind2local(li,lj, ts(num_ts)%inblock,ghost=.false.)
    
    if(ts(num_ts)%inblock) ts(num_ts)%scalar => field(li,lj)    
  end subroutine add_ts_2dind


  subroutine add_ts_1dind(description, field, i)
    implicit none
    character(len=*) description  
    real, pointer, dimension(:) :: field
    integer i

    if(num_ts >= numseries) call abort('ERROR: too many timeseries! increase parameter NUMSERIES')

    num_ts=num_ts + 1
    if(numstep >= 0) &!allocate(ts(num_ts)%data(numstep))
         allocate(ts(num_ts)%data(timeseries_dump_start:timeseries_dump_start+timeseries_interval))
    ts(num_ts)%label(:len(description)) = description
    ts(num_ts)%is_scalar = .true.
    ts(num_ts)%scalar => field(i)
  end subroutine add_ts_1dind



  subroutine add_ts_3dpos(description, field, point, x,y,depth, grid_coordinate )
!  
! description: a string that describes the timeseries, location, etc.
! field: sample this field in 
! point: with value  u_point, v_point, eta_point
! x,y:   horisontal position, 
! depth: vertical position: if depth<0, position is |depth| m above bottom
!        otherwise position is depth m below zero level.  
! grid_coordinate: true if x,y is given as a grid coordinate, otherwise
!                  geographical coordinate is assumed, and the xpole etc params of 
!                  do_sample must be specified correctly
    implicit none
    character(len=*) description  
    real, pointer, dimension(:,:,:) :: field
    integer point,li,lj
    real x,y,depth, tmpdepth
    logical grid_coordinate, inblock

    if(num_ts >= numseries) call abort('ERROR: too many timeseries! increase parameter NUMSERIES')

    num_ts=num_ts + 1  
    if(numstep >= 0) &
         allocate(ts(num_ts)%data(timeseries_dump_start:timeseries_dump_start+timeseries_interval))
    
    ts(num_ts)%label(:len(description))=description  
    ts(num_ts)%is_scalar = .false.
    ts(num_ts)%f3d => field
    ts(num_ts)%pointtype = point
    ts(num_ts)%xpos = x
    ts(num_ts)%ypos = y

    call find_hor_pos(point, num_ts)

    ts(num_ts)%sampledepth = depth
    ts(num_ts)%grid_coo = grid_coordinate

    if(.not.ts(num_ts)%inblock) return
    
    li=ts(num_ts)%ci
    lj=ts(num_ts)%cj

! the following tests for depth are not accurate, as we have neglected
! ETA, but good enough for now
    
    if(depth > 0)then
        ! point is |depth| meters above bottom
        tmpdepth = -(h(li,lj) - depth)
    else
        ! point is |depth|meters below surface
        tmpdepth = depth
    end if

!    if(tmpdepth > zz(1)*h(li,lj) .or. &
    if(tmpdepth > 0.0 .or. &    
        tmpdepth < -h(li,lj) .and. associated( ts(num_ts)%f3d ) ) then
        
        write(*,*)'warning: check series ',num_ts,'point probably not in water column'
        write(*,*)'sample depth=',ts(num_ts)%sampledepth
        write(*,*)'static depth h=',h(li,lj)
        write(*,*)'x,y=',x,y,'i,j=',ts(num_ts)%ci,ts(num_ts)%cj
        call abort()
    end if

    if(debug > 1 .and. master) write(*,'(2x,i0,a,3(f0.1,1x),a,f0.1)') num_ts,trim(ts(num_ts)%label)//". 3dpos, sample pos=",&
         x,y,ts(num_ts)%sampledepth," H=",h(li,lj)
    
  end subroutine add_ts_3dpos


  subroutine find_hor_pos(point, k)  
    integer point, k
    logical tmplog

! the grid box containing the samle point
    ts(k)%ci = NINT(ts(k)%xpos + 0.5) ! index of closest cell point
    ts(k)%cj = NINT(ts(k)%ypos + 0.5)
    
    if( point==eta_point )then
        ts(k)%i= INT( ts(k)%xpos + 0.5)   ! index of southwest cell of the 4 intpol cells
        ts(k)%j= INT( ts(k)%ypos + 0.5)

! local interpolation coordinate:
        ts(k)%x= ts(k)%xpos - (ts(k)%i -0.5)
        ts(k)%y= ts(k)%ypos - (ts(k)%j -0.5)
    end if

    if( point==u_point )then 
        ts(k)%i= INT( ts(k)%xpos)+1   ! index of southwest cell of the 4 intpol cells
        ts(k)%j= INT( ts(k)%ypos + 0.5)

!  the local, horisontal "interpolation coordinate"          
        ts(k)%x= ts(k)%xpos - (ts(k)%i -1.0)
        ts(k)%y= ts(k)%ypos - (ts(k)%j -0.5)
    end if

    if( point==v_point )then
        ts(k)%i= INT( ts(k)%xpos + 0.5)   ! index of southwest cell of the 4 intpol cells
        ts(k)%j= INT( ts(k)%ypos) + 1
!  the local, horisontal "interpolation coordinate"          
        ts(k)%x= ts(k)%xpos - (ts(k)%i -0.5)
        ts(k)%y= ts(k)%ypos - (ts(k)%j -1.0)
    end if

! convert to block local coordinates

    call globalind2local(ts(k)%ci, ts(k)%cj, ts(k)%inblock,ghost=.false.)
    call globalind2local(ts(k)%i, ts(k)%j, tmplog, ghost=.false.)
    
  end subroutine find_hor_pos


  subroutine add_ts_2dpos( description, field, point, x,y, grid_coordinate )
!    
! field: sample this field in 
! point: with value  u_point, v_point, eta_point
! x,y:   horisontal position, 
! depth: vertical position: if depth<0, position is |depth| m above bottom
!        otherwise position is depth m below zero level.  
! kdim: k dimension of field
! grid_coordinate: true if x,y is given as a grid coordinate in meters, otherwise
!                  geographical coordinate is assumed, and the xpole etc params of 
!                  do_sample must be specified correctly
    implicit none
    character(len=*) description    
    real, pointer, dimension(:,:) :: field
    integer point
    real x,y
    logical grid_coordinate

    if(num_ts >= numseries) call abort('ERROR: too many timeseries! increase parameter NUMSERIES')

    num_ts=num_ts + 1
    if(numstep >= 0)then
        allocate(ts(num_ts)%data(timeseries_dump_start:timeseries_dump_start+timeseries_interval))
    else
        call abort("NUMSTEP must be set before timeseries are initialized")
    end if
    ts(num_ts)%label(:len(description))=description    
    ts(num_ts)%is_scalar = .false.
    ts(num_ts)%f2d => field
    ts(num_ts)%pointtype = point
    ts(num_ts)%xpos = x
    ts(num_ts)%ypos = y

    call find_hor_pos(point, num_ts)

    if(debug > 1 .and. master) write(*,'(2x,i0,a,2(f0.1,1x))') num_ts,". 2dpos, sample pos=",x,y
    
    ts(num_ts)%grid_coo = grid_coordinate
  end subroutine add_ts_2dpos


  subroutine sample_timeseries(timestep)
!  
! go through the list of timeseries objects and get a sample at the given location
!  
    implicit none
    integer timestep, kz
    real slevel
    integer k,k_slev,ind
    real, dimension(kb) :: zlev
    integer, dimension(2) :: ne, nw, sw, se
    real x,y,depth,a(4)
    integer numsamp
    integer i,j,ci,cj
    logical time_to_dump
    real fdown,fup,ipcoeff
    
! a couple of reasons to do nothing    
    if(num_ts == 0) return

! Initialize communcication stuff:
! get the processor number for each time series
! in the beginning, also if we are restarting
    if(timestep < first_step) then
        if(debug>2) call message("getting processor numbers for timeseries")
        call get_ts_processors()

        if(master)then
            
        do k=1,num_ts
            if( ts(k)%is_scalar ) then
                write(*,'(2x,"Timeseries ",i3,2x,a," processor=",i0)') k,trim(ts(k)%label),ts(k)%processor
            else if( ts(k)%is_speed ) then
                write(*,'(2x,"Timeseries ",i3,2x,a,a,i0,",",i0,",",i0," processor=",i0)') &
                     k,trim(ts(k)%label)," i,j,k=",ts(k)%i,ts(k)%j,ts(k)%k,ts(k)%processor
            else if( associated( ts(k)%f3d ) ) then
                write(*,'(2x,"Timeseries ",i3,2x,a,a,1x,f0.1,1x,f0.1,1x,f0.1," processor=",i0)') &                   
                     k,trim(ts(k)%label),': x,y,z =',&
                     ts(k)%xpos,ts(k)%ypos,ts(k)%sampledepth,ts(k)%processor
            else if( associated( ts(k)%f2d ) ) then
                write(*,'(2x,"Timeseries ",i3,2x,a,a,1x,f0.1,1x,f0.1," processor=",i0)') &
                     k,trim(ts(k)%label),': x,y =',ts(k)%xpos,ts(k)%ypos,ts(k)%processor
            end if
        end do

        end if
        
        if(restart) return  ! should be no need to sample before timestep loop on restart.
    end if

    
! in this loop we do the actual sampling
    
    do k=1,num_ts
        if(.not. ts(k)%inblock ) cycle

        if( ts(k)%is_scalar ) then
            ts(k)%data(timestep)= ts(k)%scalar
            if(ts(k)%density > 0.0) ts(k)%data(timestep) = ts(k)%data(timestep) + ts(k)%density
            cycle
        end if

       ! define some shortcuts        
        i  = ts(k)%i
        j  = ts(k)%j 
        kz = ts(k)%k
        ci = ts(k)%ci   ! the cell (center) closest to the given point.
        cj = ts(k)%cj

        if(ts(k)%is_speed) then           
            ts(k)%data(timestep) = sqrt( 0.25*(u(i,j,kz)+u(i+1,j,kz))**2 + 0.25*(v(i,j,kz)+v(i+1,j,kz))**2 + wr(i,j,kz)**2)
            cycle
        end if

        sw=(/i,j/)
        se=(/i+1,j/)
        nw=(/i,j+1/)
        ne=(/i+1,j+1/)
        x=ts(k)%x  ; y=ts(k)%y

       ! if the current object has a pointer to a 2d array assume we should
       ! do a 2d interpolation

        if( associated( ts(k)%f2d ) ) then
            
            if( ts(k)%pointtype==eta_point )then            
                if( fsm(i,j)<=0.0    ) sw=(/ci,cj/)
                if( fsm(i+1,j)<=0.0  ) se=(/ci,cj/)
                if( fsm(i,j+1)<=0.0  ) nw=(/ci,cj/)
                if( fsm(i+1,j+1)<=0.0) ne=(/ci,cj/)
            end if

            a(1)=ts(k)%f2d(sw(1),sw(2))
            a(2)=ts(k)%f2d(se(1),se(2))-a(1)
            a(3)=ts(k)%f2d(nw(1),nw(2))-a(1)
            a(4)=ts(k)%f2d(ne(1),ne(2))-(a(1)+a(2)+a(3))

            ts(k)%data(timestep)=a(1) + a(2)*x + a(3)*y + a(4)*x*y          
            cycle

        else if( associated( ts(k)%f3d ) ) then

            if(ts(k)%pointtype == eta_point)then

! accurate depth of point:

                a(1)=d(i,j) ; a(2)=d(i+1,j)-a(1) ;  a(3)=d(i,j+1)-a(1)
                a(4)=d(i+1,j+1)-(a(1)+a(2)+a(3))
                depth=a(1) + a(2)*x + a(3)*y + a(4)*x*y
                
! slevel is the negative distance from the surface
                if(ts(k)%sampledepth < 0) then
                    slevel= -eta(ci,cj) + ts(k)%sampledepth
                else
                    slevel= ts(k)%sampledepth - depth
                end if

! special case for eta points: extrapolate to avoid 0's in the bilin intpol stencil

                if(fsm(i,j)<=0.0) sw=(/ci,cj/)
                if(fsm(i+1,j)<=0.0) se=(/ci,cj/)
                if(fsm(i,j+1)<=0.0) nw=(/ci,cj/)
                if(fsm(i+1,j+1)<=0.0) ne=(/ci,cj/)        
            end if

! dont extrapolate for u and v points -we assume only velocities to
! be estimated there, and 0 on land is correct

            if(ts(k)%pointtype == u_point)then
                a(1)=du(i,j)   ;   a(2)=du(i+1,j)-a(1)  ;  a(3)=du(i,j+1)-a(1)
                a(4)=du(i+1,j+1)-(a(1)+a(2)+a(3))
                depth=a(1) + a(2)*x + a(3)*y + a(4)*x*y

                if(ts(k)%sampledepth < 0) then
                    slevel = ts(k)%sampledepth - eta(ci,cj)
                else
                    slevel= ts(k)%sampledepth - depth
                end if
            end if

            if(ts(k)%pointtype == v_point)then
                a(1)=dv(i,j)   ;   a(2)=dv(i+1,j)-a(1)  ;   a(3)=dv(i,j+1)-a(1)
                a(4)=dv(i+1,j+1)-(a(1)+a(2)+a(3))
                depth=a(1) + a(2)*x + a(3)*y + a(4)*x*y

                if(ts(k)%sampledepth < 0) then
                    slevel= ts(k)%sampledepth - eta(ci,cj)
                else
                    slevel= ts(k)%sampledepth - depth
                end if
            end if

            zlev = zz*depth          
            k_slev = binsearch2(slevel, zlev, last=kb-1)
            
            if(k_slev <= 1) then
                ts(k)%data(timestep) = ts(k)%f3d(ci,cj,1)
                goto 1
            end if

            if(k_slev == kb) then
                ts(k)%data(timestep)=ts(k)%f3d(ci,cj,kb-1)              
                goto 1
            end if
            
! bilinear interpolation horisontally
            a(1)=ts(k)%f3d(sw(1), sw(2), k_slev)
            a(2)=ts(k)%f3d(se(1), se(2), k_slev)-a(1)
            a(3)=ts(k)%f3d(nw(1), nw(2), k_slev)-a(1)
            a(4)=ts(k)%f3d(ne(1), ne(2), k_slev)-(a(1)+a(2)+a(3))
            fdown = a(1) + a(2)*x + a(3)*y + a(4)*x*y

            a(1)=ts(k)%f3d(sw(1), sw(2), k_slev-1)
            a(2)=ts(k)%f3d(se(1), se(2), k_slev-1)-a(1)
            a(3)=ts(k)%f3d(nw(1), nw(2), k_slev-1)-a(1)
            a(4)=ts(k)%f3d(ne(1), ne(2), k_slev-1)-(a(1)+a(2)+a(3))
            fup = a(1) + a(2)*x + a(3)*y + a(4)*x*y
            
! linear interpolation vertically
            ipcoeff = (slevel-zlev(k_slev))/(zlev(k_slev-1)-zlev(k_slev))
            ts(k)%data(timestep) = fup*ipcoeff + fdown*(1.0 - ipcoeff)           
1           if(ts(k)%density > 0.0) ts(k)%data(timestep) = ts(k)%data(timestep) + ts(k)%density
        end if
    end do
  end subroutine sample_timeseries



  subroutine store_timeseries(timestep, cpdir)
!
! write a batch of timeseries
! the tsXX...-batchXXX....dat files may be cat'ed together in order to get a complete
! timeseries file:  cat ts-batch???.dat > timeseries.dat
!
    implicit none
    integer timestep
    character(*) cpdir
    logical time
    real(4), allocatable :: timedata(:)
    integer i, ind
    character(len=200) :: tsfilename

! first, some reasons to not store timeseries...
    
    if(.not. (mod(timestep,timeseries_interval)==0  .or. timestep==numstep  .or.  do_checkpoint)) then
        return
    end if
    
! each batch's file name get this number baked in to indicate the
! position in the sequence
    timeseries_file_number = timeseries_file_number + 1

    
! dump the time in hours for each step
    
    if(master .and. num_ts > 0) then
        allocate(timedata(timeseries_dump_start:timestep))
        do i=timeseries_dump_start,timestep
            timedata(i) = i*dt/3600.
        end do
        if(timeseries_unformatted)then
            write(tsfilename,'("ts-",a,"-batch",i3.3,"-step",i7.7,"-step",i7.7,".sr4")') &
                 "time", timeseries_file_number, timeseries_dump_start, timestep
            open(unit=1000007, file=trim(timeseries_dir)//tsfilename, access="stream", form="unformatted", err=101)
            write(1000007) timedata
        end if

        if(timeseries_fullprec .or. timeseries_formatted)then
            write(tsfilename,'("ts-",a,"-batch",i3.3,"-step",i7.7,"-step",i7.7,".dat")') &
                 "time", timeseries_file_number, timeseries_dump_start, timestep
            open(unit=1000007, file=trim(timeseries_dir)//tsfilename, err=101)
            if(timeseries_fullprec) then
                do i=timeseries_dump_start,timestep
                    write(1000007,*) timedata(i)
                end do
            else if(timeseries_formatted) then
                write(1000007,'(f0.5)') timedata(timeseries_dump_start:timestep)
            end if
        end if
        deallocate(timedata)
        close(1000007)
    end if

! then dump the data
    
    if(local_num_ts > 0) then
        do ind=1,num_ts
            if(ts(ind)%inblock) then
                if(timeseries_unformatted)then                    
                    write(tsfilename,'("ts-",a,i3.3,"-batch",i3.3,"-step",i7.7,"-step",i7.7,".sr4")') &
                         trim(ts(ind)%label), ind, timeseries_file_number, timeseries_dump_start, timestep

                    call replace(tsfilename, "(),", "_")
                    
                    open(unit=1000007, file=trim(timeseries_dir)//tsfilename, access="stream", form="unformatted", err=101)
                    if(debug > 1) write(*,*) "Timeseries batch written to file ",trim(tsfilename)

                    write(1000007) real(ts(ind)%data(timeseries_dump_start:timestep), kind=4)
                    close(1000007)
                end if

                if(timeseries_fullprec .or. timeseries_formatted)then
                    write(tsfilename,'("ts-",a,i3.3,"-batch",i3.3,"-step",i7.7,"-step",i7.7,".dat")') &
                         trim(ts(ind)%label), ind, timeseries_file_number, timeseries_dump_start, timestep
                    
                    call replace(tsfilename, "(),", "_")

                    open(unit=1000007, file=trim(timeseries_dir)//tsfilename, err=101)
                    if(debug > 1) write(*,*) "Timeseries batch written to file ",trim(tsfilename)

                    if(timeseries_fullprec) then
                        do i=timeseries_dump_start,timestep
                            write(1000007,*) ts(ind)%data(i)
                        end do
                    else if(timeseries_formatted) then
                        write(1000007,ts(ind)%format) ts(ind)%data(timeseries_dump_start:timestep)
                    end if
                    close(1000007)
                end if
            end if
        end do
    end if

    if(timestep == numstep) return
    
    timeseries_dump_start = timestep + 1
    call reallocate_timeseries(timeseries_dump_start, timeseries_dump_start+timeseries_interval)

    return
101 call abort('could not open file')    
  end subroutine store_timeseries

  
  subroutine get_ts_processors()
!    
! first fetch the number of timeseries on each processor to master,
! then fetch a list of timeseries from each processor. finally update
! the processor variable in ts(:) on master.
!
    implicit none
    integer i,j, displs(numproc), recvbuf(numproc*num_ts),currdisp,n,tser,ierr
    integer ts_counts(numproc)
    
    if(numproc == 1)then
        do i=1,num_ts
            ts(i)%processor = myid
        end do
        local_num_ts = num_ts
        return
    end if
    
    local_num_ts=0
    do i=1,num_ts
        if(ts(i)%inblock)then
            local_num_ts = local_num_ts + 1
            localseries(local_num_ts) = i
        end if
    end do

! store the number of timeseries on each processor in ts_counts
    ts_counts = gather(local_num_ts)

    if(master) then
        currdisp=0
        do i=1,numproc
            displs(i) = currdisp
            currdisp = currdisp + ts_counts(i)
        end do
    end if
    
    call mpi_gatherv(localseries(1:local_num_ts), local_num_ts, mpi_integer, &
         recvbuf, ts_counts, displs, mpi_integer, 0, qmpi_comm_world, ierr)

    if(master)then
        do i=1,numproc
            !! ts_counts is the number of timeseries on processor i            
            if(ts_counts(i) > 0)then
                do j=1,ts_counts(i)
                    ! n is the first index in recvbuf with data from processor i
                    n = displs(i) + 1   
                    tser = recvbuf(n+j-1)
                    ts(tser)%processor = i-1
                end do
            end if
        end do
    end if
  end subroutine get_ts_processors
    
  
  subroutine gatherts(first, last)
!
! transfer the given interval of sample points from their processors to master
!
    implicit none
    integer first, last, j
    
    if(master .and. debug > 2) then
        write(*,*) "GATHERTS; getting data range ",first, last
    end if
    
    do j=1,num_ts
        if(master)then
            if(.not. ts(j)%inblock)then
                call receive(ts(j)%data(first:last), ts(j)%processor, tag=j)
            end if
        else
            if(ts(j)%inblock)then
                call isend(ts(j)%data(first:last), 0, tag=j)                
            end if
        end if
    end do
    call waitall()
  end subroutine gatherts
  

  subroutine add_timeseries(splitline)
    use mod_transport
    use mod_concentration, only : co
    implicit none
    character(*) splitline(:)
    integer i,j,k
    real gx,gy,gz
    
    select case(splitline(1))
        
    case('timeseries_quick2d')
! timeseries_quick2d <x> <y>
        read(splitline(2:3),*,err=101,end=101) gx,gy
        gx = gx/dx + 1.0
        gy = gy/dy + 1.0
        call add_ts_2dpos("ETA", eta, eta_point, gx, gy, .true.)
        call add_ts_2dpos("UA", ua, u_point, gx, gy, .true.)
        call add_ts_2dpos("VA", va, v_point, gx, gy, .true.)
        if(debug>0.and.master) write(*,*)' adding timeseries for ETA,UA,VA in x,y=',gx,gy
        
    case('timeseries_quick3d')
! timeseries_quick3d <x> <y> <z>
        read(splitline(2:4),*,err=102,end=102) gx,gy,gz
        gx = gx/dx + 1.0
        gy = gy/dy + 1.0
        call add_ts_3dpos("U", u, u_point, gx, gy, gz, .true.)
        call add_ts_3dpos("V", v, v_point, gx, gy, gz, .true.)
        call add_ts_3dpos("WR", wr, eta_point, gx, gy, gz, .true.)
        if(advect_density)then
            call add_ts_3dpos("RHO", rho, eta_point, gx, gy, gz, .true.)
            ts(num_ts)%density = 1.0
        else
            call add_ts_3dpos("S", s, eta_point, gx, gy, gz, .true.)
            call add_ts_3dpos("T", t, eta_point, gx, gy, gz, .true.)
        end if
        if(debug>0.and.master) write(*,*)' adding timeseries for U,V,WR,S,T in x,y,z=',gx,gy,gz
        
    case('timeseries')
! timeseries u <type=0> <x> <y> <z>
! timeseries u <type=1> <i> <j> <k>
    select case(splitline(2))
    case('u')
        call fetch_ts3d("U",u,u_point)
    case('v')
        call fetch_ts3d("V",v,v_point)
    case('wr')
        call fetch_ts3d("WR",wr,eta_point)
    case('s')
        if(advect_density) call message('advect_density is .true., not possible to plot S',2)
        call fetch_ts3d("S",s,eta_point)
    case('t')
        if(advect_density) call message('advect_density is .true., not possible to plot T',2)
        call fetch_ts3d("T",t,eta_point)
    case('rho')
        call fetch_ts3d("RHO",rho,eta_point)
        ts(num_ts)%density = 1.0
    case('eta')
        call fetch_ts2d("ETA",eta,eta_point)
    case('ua')
        call fetch_ts2d("UA",ua,u_point)
    case('va')
        call fetch_ts2d("VA",va,v_point)
    case('ekin')
        call add_ts_scalar('Ekin',ekin)
    case('mass')
        call add_ts_scalar('Mass',mass)
    case('volume')
        call add_ts_scalar('Volume',vol)
    case('eeta')
        call add_ts_scalar('EEta',eeta)
    case('wcourant')
        call add_ts_scalar('WCourant',wcourant)        
    case('p')
        if(.not.nonhyd) call message("NONHYD is false, then you don't want a timeseries for P?",2)
        call fetch_ts3d('P',p, eta_point)
    case('conc1')
        if(.not.co(1)%allocated) call message("concentration field not allocated, cannot add timeseries",2)
        call fetch_ts3d('conc1', co(1)%conc, eta_point)
    case('conc2')
        if(.not.co(2)%allocated) call message("concentration field not allocated, cannot add timeseries",2)
        call fetch_ts3d('conc2', co(2)%conc, eta_point)
    case('conc3')
        if(.not.co(3)%allocated) call message("concentration field not allocated, cannot add timeseries",2)
        call fetch_ts3d('conc3', co(3)%conc, eta_point)
    case('conc4')
        if(.not.co(4)%allocated) call message("concentration field not allocated, cannot add timeseries",2)
        call fetch_ts3d('conc4', co(4)%conc, eta_point)
    case('conc5')
        if(.not.co(5)%allocated) call message("concentration field not allocated, cannot add timeseries",2)
        call fetch_ts3d('conc5', co(5)%conc, eta_point)
    case('conc6')
        if(.not.co(6)%allocated) call message("concentration field not allocated, cannot add timeseries",2)
        call fetch_ts3d('conc6', co(6)%conc, eta_point)
    case('conc7')
        if(.not.co(7)%allocated) call message("concentration field not allocated, cannot add timeseries",2)
        call fetch_ts3d('conc7', co(7)%conc, eta_point)
    case('conc8')
        if(.not.co(8)%allocated) call message("concentration field not allocated, cannot add timeseries",2)
        call fetch_ts3d('conc8', co(8)%conc, eta_point)
    case('conc9')
        if(.not.co(9)%allocated) call message("concentration field not allocated, cannot add timeseries",2)
        call fetch_ts3d('conc9', co(9)%conc, eta_point)
    case('transport')
        call allocate_pgplot()
        call add_transport_vsect(splitline)
        call add_ts_scalar(trim(transport_label),transport_array(transport_count))
    case default
        call message('unknown <what> in column 2',1)
    end select ! timeseries
    
    case default
        call message('unknown type in column 1',1)
    end select
    return
101 call message('could not get 2 floats for x,y for quick2d time series from column 2 and 3',2)
102 call message('could not get 3 floats for x,y,z for quick3d time series from column 2,3,4',2)

  contains

    subroutine fetch_ts3d(label,f,pt_type)
      implicit none
      character(*) label
      real, pointer :: f(:,:,:)
      integer pt_type, ptype
      real glon,glat
      
      read(splitline(3),*,err=103,end=103) ptype

      select case(ptype)
      case(0)
          read(splitline(4:6),*,err=104,end=104) gx,gy,gz
          gx = gx/dx + 1.0
          gy = gy/dy + 1.0
          call add_ts_3dpos(label, f, pt_type, gx, gy, gz, .true.)
      case(1)
          read(splitline(4:6),*,err=105,end=105) i,j,k
          call add_ts_3dind(label, f, i,j,k)
      case(2)
          read(splitline(4:6),*,err=104,end=104) glon,glat,gz
          call sph2gr(glon, glat, (xpole+dx)/dx, (ypole+dy)/dy, dx*0.001, dy*0.001, ylong, gx, gy, ierr)
          call add_ts_3dpos(label, f, pt_type, gx, gy, gz, .true.)         
      case default
          if(master)write(*,*)'ptype=',ptype
          call message('unknown coordinate type, try 0 or 1',2)
      end select
      return
103   call message('could not get coordinate type from column 3',2)
104   call message('could not get float for x,y,z from column 4 to 6',2)
105   call message('could not get indices for i,j,k from column 4 to 6',2)
    end subroutine fetch_ts3d

    subroutine fetch_ts2d(label,f,pt_type)
      implicit none
      character(*) label
      real, pointer :: f(:,:)
      integer pt_type,ptype
      real glon,glat

      read(splitline(3),*,err=103,end=103) ptype
      select case(ptype)
      case(0)
          read(splitline(4:5),*,err=104,end=104) gx,gy
          gx = gx/dx + 1.0
          gy = gy/dy + 1.0
          call add_ts_2dpos(label, f, pt_type, gx, gy, .true.)
      case(1)
          read(splitline(4:5),*,err=105,end=105) i,j
          call add_ts_2dind(label, f, i,j)
      case(2)
          read(splitline(4:5),*,err=104,end=104) glon,glat
          call sph2gr(glon, glat, (xpole+dx)/dx, (ypole+dy)/dy, dx*0.001, dy*0.001, ylong, gx, gy, ierr)
          call add_ts_2dpos(label, f, pt_type, gx, gy, .true.)         
      case default
          if(master)write(*,*)'ptype=',ptype
          call message('unknown coordinate type, try 0 or 1',2)
      end select
      return
103   call message('could not get coordinate type from column 3',2)
104   call message('could not get float for x,y from column 4 to 5',2)
105   call message('could not get indices for i,j from column 4 to 5',2)
    end subroutine fetch_ts2d

  subroutine add_transport_vsect(splitline)
  !
  ! Add a new transport vsect line
  !
    implicit none
    character(*) splitline(:)

    if(master) then
       if(.not.allocated(transport_x)) then
          allocate(transport_x(im*jm))
          allocate(transport_z(im*jm,kb))
       end if
    end if

    ! increment counter for transport calculations
    transport_count = transport_count+1
    
    ! start points and end points
    read(splitline(3:6),*,err=104,end=104)   &
         transport_px(transport_count,1), transport_py(transport_count,1),  &
         transport_px(transport_count,2), transport_py(transport_count,2)

    write(transport_label,'(a,i3.3)') 'transport', transport_count

    if(master) then
       write(*,*) 'transport: (x1,y1)=',  &
            transport_px(transport_count,1), transport_py(transport_count,1)
       write(*,*) '           (x2,y2)=',  &
         transport_px(transport_count,2), transport_py(transport_count,2)
    end if

    ! Convert to grid coordinates
    transport_px(transport_count,1) = transport_px(transport_count,1)/dx + 1.0
    transport_px(transport_count,2) = transport_px(transport_count,2)/dx + 1.0
    transport_py(transport_count,1) = transport_py(transport_count,1)/dy + 1.0
    transport_py(transport_count,2) = transport_py(transport_count,2)/dy + 1.0

    return
104   call message('could not get float for x1,y1,x2,y2 from column 3 to 6',2)
  end subroutine add_transport_vsect

  end subroutine add_timeseries


  subroutine set_timeseries_format(splitline)
! example
!   timeseries_format_string "(f0.1)"  3
! set format for series number 3.
    implicit none
    character(*) splitline(:)
    integer ind
    read(splitline(3),*,err=1,end=1) ind
    if(ind < 1 .or. ind > num_ts) call message("invalid timeseries index",2)
    ts(ind)%format = trim(splitline(2))
    if(ts(ind)%format(1:1) .ne. "(") call message("valid format strings must start with a left paren",2)
    if(.not. any(ts(ind)%format(2:2) == (/"g","f","e","G","F","E"/))) call message("edit descriptor should be E, G or F",2)
    if(.not. any(ts(ind)%format(3:3) == (/"0","1","2","3","4","5","6","7","8","9"/))) &
         call message("the first character after the edit descriptor must be a digit",2)
    return
1   call message("could not read time series index from column 3",2)
  end subroutine set_timeseries_format



  subroutine timeseries_from_file(splitline)
!    
! timeseries_file <file name>
!
    use util
    implicit none
    character(*) splitline(:)
    character(20), allocatable :: tsline(:,:)
    integer i,numlines, n
    
    allocate(tsline(6,numseries))
    call get_splitlines(splitline(2), tsline, numlines=n)
    if(debug>1) write(*,*) "got timeseries data file with ",n," timeseries, now initializing those" 
    do i=1,n
        call add_timeseries(tsline(:,i))
    end do
    deallocate(tsline)
    
    return
101 call message("could not open file with timeseries called ",2, sval=(/trim(splitline(2))/))
  end subroutine timeseries_from_file

end module timeseries


