SUBROUTINE VERTDIFF(F,WFSURF)
!
!  VERTDIFF propagates the solution of the equations
!
!  d(FD)/dt = d((KH/D)*d(F)/d(sigma))/d(sigma)
!
!  with boundary conditions:
!
!  (KH/D)*d(F)/d(sigma) = WFSURF  at sigma=0
!  (KH/D)*d(F)/d(sigma) = 0 at sigma=-1
!
!  The solutions are propagated with an fully implicit method.
!
  USE STATE
  USE MOD_CONCENTRATION, ONLY : CO, MAXNUMCONC
  IMPLICIT NONE
  REAL, POINTER :: F(:,:,:), WFSURF(:,:), TMP3D(:,:,:)
!
! Local variables.
!
  INTEGER I,J,K
  REAL DSQ
  REAL, DIMENSION(kb-1) :: RHS,A,B,C,G, alf, bet
!
  INTEGER CONCI,NUMASS
! use separate KH for concentration
  CONCI=0
  NUMASS=0
  DO I=1,MAXNUMCONC
      !! check if F is CONCX
      IF(ASSOCIATED(F, TARGET=CO(I)%CONC)) THEN
          IF(ASSOCIATED(CO(I)%KH)) THEN
              NUMASS = NUMASS + 1
            !! it's got a KH, store the old and use it
              TMP3D => KH
              KH => CO(I)%KH
              CONCI = I
              IF(MASTER.AND.DEBUG>3) WRITE(*,*)"Using KH from CONC",i
          END IF
      END IF
  END DO
  IF(NUMASS > 1) CALL MESSAGE("MORE THAN ONE KH ASSIGNMENT!",2)
!
!  First propagate the F-equation over all internal sea points.
!
DO J=J_EQ_2,JM_MI_1
  DO I=I_EQ_2,IM_MI_1
          IF (FSM(I,J).NE.0.) THEN
              DSQ = D(I,J)*D(I,J)
!
! Define the right hand side of the solution matrix.
!
              RHS(1) = F(I,J,1) + DT*WFSURF(I,J)/(D(I,J)*DZ(1))
              DO K = 2,KB-1
                  RHS(K) = F(I,J,K)
              END DO
!
! Define the upper diagonal elements of the tridiagonal solution matrix.
!
              DO K = 1,KB-2
                  B(K) = -DT*KH(I,J,K+1)/(DSQ*DZ(K)*DZZ(K))
              END DO
!
! Define the lower diagonal elements.
!
              DO K = 2,KB-1
                  C(K) = -DT*KH(I,J,K)/(DSQ*DZ(K)*DZZ(K-1))
              END DO
!
! Define the diagonal elements.
!
              A(1) = 1. - B(1)
              DO K = 2,KB-2
                  A(K) = 1. - B(K) - C(K)
              END DO
              A(KB-1) = 1. - C(KB-1)
!
! Solve the system of equations.
!
!
!   LU-decompose the linear system.
!
              ALF(1) = A(1)
              DO K = 2,KB-1
                  ALF(K) = A(K) - C(K)*B(K-1)/ALF(K-1)
              END DO
              DO K = 2,KB-1
                  BET(K) = C(K)/ALF(K-1)
              END DO
!
!   Forward substitute the right hand side.
!
              G(1) = RHS(1)
              DO K = 2,KB-1
                  G(K) = RHS(K) - BET(K)*G(K-1)
              END DO
!
!   Back substitute to compute the solution F.
!
              F(i,j,KB-1) = G(KB-1)/ALF(KB-1)
              DO K = KB-2,1,-1
                  F(i,j,K) = (G(K) - B(K)*F(i,j,K+1))/ALF(K)
              END DO
          ENDIF
      END DO
  END DO

! restore KH  
  IF(CONCI > 0) THEN
      KH => TMP3D
  END IF
  CALL DUMP_STATE("vertdiff")
END SUBROUTINE VERTDIFF
