SUBROUTINE VERTVISCUVB(UBCLIN,VBCLIN)
!
!  VERTVISCUV propagates the solution of the equations
!
!  d(UD)/dt = d((KM/D)*d(U)/d(sigma))/d(sigma)
!
!  d(VD)/dt = d((KM/D)*d(V)/d(sigma))/d(sigma)
!
!  with boundary conditions:
!
!  (KM/D)*d(U)/d(sigma) = WSURFX  at sigma=0
!  (KM/D)*d(V)/d(sigma) = WSURFY  at sigma=0
!  (KM/D)*d(U)/d(sigma) = CBC*ABS((U,V))*U at sigma=-1
!  (KM/D)*d(V)/d(sigma) = CBC*ABS((U,V))*V at sigma=-1
!
!  where (WSURFX,WSURFY) are the surface wind stresses in X and Y directions
!  and D the bottom drag coefficient.
!
!  The solutions are propagated with an fully implicit method.
!
USE STATE
USE ETAIMPLICIT  
IMPLICIT NONE
!
! Local variables.
!
INTEGER I,J,K,gi,gj
REAL DUSQ,FRICT,DVSQ,SPEED
REAL, DIMENSION(KB-1)  :: RHS,A,B,C,ALF,BET,G
REAL, DIMENSION(I1:i2,J1:j2,KB) :: KMU,KMV,UBCLIN,VBCLIN
!
!  First propagate the U-equation over all internal sea points.
!
!  Average vertical viscosity coefficients to U-points.
!
DO K=2,KB
    DO J=1,J_eq_JM
        DO I=1,I_eq_IM
            
           !local_i=global_i - block%fi + 1
            gi=i+block%fi-1
            gj=j+block%fj-1
            
            IF( GI>1)THEN
                KMU(I,J,K) = 0.5*(KM(I-1,J,K)+KM(I,J,K))
            ELSE
                KMU(I,J,K) = KM(I,J,K)
            END IF
            IF( GJ>1)THEN
                KMV(I,J,K) = 0.5*(KM(I,J-1,K)+KM(I,J,K))
            ELSE
                KMV(I,J,K) = KM(I,J,K)
            END IF            
        END DO
    END DO
END DO


! for the implicit eta solver we need UA,VA to be 0 in this routine
! it is not used outside modesplit, so this should be ok.
if(implicit_eta) then
    ua = 0.0
    va = 0.0
end if


DO J=J_EQ_2,JM_MI_1
  DO I=I_EQ_2,I_EQ_IM
    IF (DUM(I,J).NE.0.) THEN
      DUSQ = DU(I,J)*DU(I,J)
      IF(JM>1)THEN
         SPEED = U(I,J,KB-1)**2 + (0.25*(V(I,J,KB-1)+V(I,J+1,KB-1)+V(I-1,J,KB-1)+V(I-1,J+1,KB-1)))**2
         FRICT = 0.5*(CBC(I,J)+CBC(I-1,J))*SQRT(SPEED)
      ELSE
         FRICT = 0.5*(CBC(I,J)+CBC(I-1,J))*SQRT(U(I,J,KB-1)**2 &
           + (0.5*(V(I,J,KB-1) + V(I-1,J,KB-1)))**2)
      END IF
!
! Define the right hand side of the solution matrix.
!
      RHS(1) = Ubclin(I,J,1) + DT*WUSURF(I,J)/(DU(I,J)*DZ(1))
      DO K = 2,KB-2
        RHS(K) = Ubclin(I,J,K)
      END DO
      RHS(KB-1) = UBCLIN(I,J,KB-1) - DT*FRICT*UA(I,J)/(DU(I,J)*DZ(KB-1))
      
! Define the upper diagonal elements of the tridiagonal solution matrix.
!
      DO K = 1,KB-2
        B(K) = -DT*KMU(I,J,K+1)/(DUSQ*DZ(K)*DZZ(K))
      END DO
!
! Define the lower diagonal elements.
!
      DO K = 2,KB-1
        C(K) = -DT*KMU(I,J,K)/(DUSQ*DZ(K)*DZZ(K-1))
      END DO
!
! Compute bottom friction.
!
!
! Define the diagonal elements.
!
      A(1) = 1. - B(1)
      DO K = 2,KB-2
        A(K) = 1. - B(K) - C(K)
      END DO
      A(KB-1) = 1. - C(KB-1) + DT*FRICT/(DU(I,J)*DZ(KB-1))
!
! Solve the system of equations.
!
!
!   LU-decompose the linear system.
!
      ALF(1) = A(1)
      DO K = 2,KB-1
        ALF(K) = A(K) - C(K)*B(K-1)/ALF(K-1)
      END DO
      DO K = 2,KB-1
        BET(K) = C(K)/ALF(K-1)
      END DO
!
!   Forward substitute the right hand side.
!
      G(1) = RHS(1)
      DO K = 2,KB-1
        G(K) = RHS(K) - BET(K)*G(K-1)
      END DO
!
!   Back substitute to compute the solution U.
!
      Ubclin(I,J,KB-1) = G(KB-1)/ALF(KB-1)
      DO K = KB-2,1,-1
        Ubclin(I,J,K) = (G(K) - B(K)*Ubclin(I,J,K+1))/ALF(K)
      END DO
      WUBOT(I,J) = -FRICT*(Ubclin(I,J,KB-1)+ua(i,j))
    ENDIF
  END DO
END DO
!
!  Then propagate the V-equation over all internal sea points.
!
!
DO J=J_EQ_2,J_eq_JM
  DO I=I_EQ_2,IM_mi_1
    IF (DVM(I,J).NE.0.) THEN

      DVSQ = DV(I,J)*DV(I,J)
!
! Compute bottom friction.
!
     IF(JM>1)THEN
       SPEED = V(I,J,KB-1)**2 + (0.25*(U(I,J,KB-1)+U(I+1,J,KB-1)+U(I,J-1,KB-1)+U(I+1,J-1,KB-1)))**2
       FRICT = 0.5*(CBC(I,J)+CBC(I,J-1))*SQRT(SPEED)
     ELSE
       SPEED = V(I,J,KB-1)**2 + 0.5*(U(I,J,KB-1) + U(I+1,J,KB-1))**2
       FRICT = CBC(I,J)*SQRT(SPEED)
     END IF
!
! Define the right hand side of the solution matrix.
!
      RHS(1) = Vbclin(I,J,1) + DT*WVSURF(I,J)/(DV(I,J)*DZ(1))
      DO K = 2,KB-2
        RHS(K) = Vbclin(I,J,K)
      END DO
      RHS(KB-1) = VBCLIN(I,J,KB-1) - DT*FRICT*VA(I,J)/(DV(I,J)*DZ(KB-1))!
! Define the upper diagonal elements of the tridiagonal solution matrix.
!
      DO K = 1,KB-2
        B(K) = -DT*KMV(I,J,K+1)/(DVSQ*DZ(K)*DZZ(K))
      END DO
!
! Define the lower diagonal elements.
!
      DO K = 2,KB-1
        C(K) = -DT*KMV(I,J,K)/(DVSQ*DZ(K)*DZZ(K-1))
      END DO
!
! Define the diagonal elements.
!
      A(1) = 1. - B(1)
      DO K = 2,KB-2
        A(K) = 1. - B(K) - C(K)
      END DO
      A(KB-1) = 1. - C(KB-1) + DT*FRICT/(DV(I,J)*DZ(KB-1))
!
! Solve the system of equations.
!
!
!   LU-decompose the linear system.
!
      ALF(1) = A(1)
      DO K = 2,KB-1
        ALF(K) = A(K) - C(K)*B(K-1)/ALF(K-1)
      END DO
      DO K = 2,KB-1
        BET(K) = C(K)/ALF(K-1)
      END DO
!
!   Forward substitute the right hand side.
!
      G(1) = RHS(1)
      DO K = 2,KB-1
        G(K) = RHS(K) - BET(K)*G(K-1)
      END DO
!
!   Back substitute to compute the solution V.
!
      Vbclin(I,J,KB-1) = G(KB-1)/ALF(KB-1)
      DO K = KB-2,1,-1
        Vbclin(I,J,K) = (G(K) - B(K)*Vbclin(I,J,K+1))/ALF(K)
      END DO
      WVBOT(I,J) = -FRICT*(Vbclin(I,J,KB-1)+va(i,j))
    ENDIF
  END DO
END DO

CALL DUMP_STATE("vertviscuvb")
END
