!  SUBROUTINE Z2SIGMA(ICODE,FELT,FINN3D,ZINN,LINN,UNDEF)
!    USE STATE
!    IMPLICIT NONE
!    INTEGER, PARAMETER :: LOCALKIND=4
!    
! Note: the below z point bug is not at all "fixed" for u and v points. and should be tested more for P points.    
!
!***BEGIN PROLOGUE Z2SIGMA
!***DATE WRITTEN   23 JULY 1991
!***AUTHOR
!           Bj{\o}rn {\AA}dlandsvik
!           Institute of Marine Research
!           P.O.Box 1870 Nordnes
!           N-5024 Bergen, Norway
!           Internet:  bjorn@imr.no
!
!           Revised by J.Berntsen 27 august 1991 by
!           introducing ICODE and placing all parameters
!           in the parameter list.
!
!***PURPOSE
!  Input a 3D field given as a set of horizontal fields in 
!  real depth to the Princeton sigma-coordinate model (ECOM3D).
!  This routine should be compiled together with the model.
!
!***PARAMETERS
!***ON ENTRY
!
!  ICODE      Integer.
!             If ICODE = 1 on entry, the data to be transformed
!             are in S and T points of the Arakawa C-grid.
!             If ICODE = 2 on entry, they are in U-points.
!             If ICODE = 3 on entry, they are in V-points.
!  FINN3D     Real array of dimension (IM,JM,LINN).
!             Working array used to store the 3-D field
!             in real depths.
!  ZINN       Real array of dimension LINN.
!             Working array used to store the real depths. Should be in increasing order.
!  LINN       Integer.
!             The number of real depths.
!  
!***ON RETURN
!  FELT       real array of dimension (IM, JM, KB),
!             The 3D field converted to sigma-coordinates.
!
!***LOCAL VARIABLES
!  LBOT       integer,
!     Local number of z-levels downto bottom.
!  ZDYP       real,
!     Local real depth of midpoint in sigma-level.
!  I          integer,
!     Index in X-direction.
!  J          integer,
!     Index in Y-direction.
!  K          integer,
!     Vertical index for sigma-coordinates.
!  LL         integer,
!     Vertical index for z-coordinates.
!
!***END PROLOGUE Z2SIGMA
!
!
    integer icode,linn
    real(localkind), intent(in) :: finn3d(i1:i2,j1:j2,linn), zinn(linn) ! NB LOCALKIND comes from the file that includes this file
    real(localkind), intent(out):: felt(i1:i2,j1:j2,kb)
    real(localkind) :: undef    ! disregard all FINN3D(I,J,K) == UNDEF
    real zdata(linn)
    
    INTEGER LBOT
    REAL ZDYP, DYP
    INTEGER I,J,K,LL
    
! neighbour offsets
    INTEGER NO(2,8), NUMVALID, N, NEWi,NEWJ,NUMEXTRA
    REAL(LOCALKIND) :: UPPER, LOWER, EPS=1e-6

    ! offsets to neighbour points. the 4 first are the closest. 
    no = reshape((/-1,0, 0,-1, 1,0, 0,1,  -2,0, 0,-2, 2,0, 0,2/),(/2,8/))

    numextra = 0
    felt = 0.0
    
    if (icode.eq.1) then
        do j = 1, j_eq_jm
            do  i = 1, i_eq_im
                if (h(i,j) > 0.0) then

                    zdata = finn3d(i,j,:)
                    
                   ! find last Z data point LBOT above the sea floor with valid data
                    lbot = -1
                    do ll = 1, linn
                        if (zinn(ll) < h(i,j) .and. abs(zdata(ll)-undef) > eps)  lbot = ll
                    end do

                    if(lbot == -1) then
                        write(*,*) myid, "ERROR: NO, as in 0, useful Z data points found in water column. Congratulations."
                        write(*,*) myid, "gi,gj=", block%fi+i-1,block%fj+j-1, "H(i,j)=", h(i,j)
                        write(*,*) myid, "finn(i,j,:)=", zdata
                        write(*,*) myid, "zinn=",zinn
                        call abort()
                    end if
                    
                    do k = 1, kb-1
                        zdyp = - zz(k) * h(i,j)
                        
                        if (zdyp <= zinn(1)) then
                            ! extrapolate upwards   
                            if(abs(zdata(1)-undef) > eps)then
                                felt(i,j,k) = zdata(1)                                
                                cycle
                            else
                            ! yes, give up. this should not happen. 
                                write(*,*) myid, "gi,gj=", block%fi+i-1,block%fj+j-1, "H(i,j)=", h(i,j)
                                write(*,*) myid, "finn(i,j,:)=", zdata
                                write(*,*) myid, "zinn=",zinn
                                call abort("cannot extrapolate upwards, finn3d(...,1)=undef, no neighbours")
                            end if
                        end if
                            
                        if (zdyp > zinn(lbot)) then
                            ! problem... the sigma layer is below the last z data point.
                            
                            ! Jarle found downward extrapolation from lbot(makes water homogeneous below last z data point)
                            ! to sometimes give artificial horizontal gradients near bottom, because
                            ! the neighbour "column" may be stratified. so we should avoid that.
                            
                            ! we should first check if we have a valid data point below LBOT. if we do, no problem.
                            if(lbot < linn)then
                                if(zdata(lbot+1) > undef) then
                                   ! found valid data below sea floor :) interpolate!
                                    felt(i,j,k) = zdata(lbot) + (zdata(lbot+1)-zdata(lbot)) &
                                         &      * (zdyp-zinn(lbot)) / (zinn(lbot+1)-zinn(lbot))
                                    cycle
                                end if
                            else
                                ! lbot=linn? no need to search neighbours for data points below lbot. should be
                                ! safe to extrapolate since none of the neighbours can be interpolated either.
                                felt(i,j,k) = zdata(linn)
                                cycle
                            end if

                            ! the below attempts still leaves us with a lot of crappy data. count and warn
                            felt(i,j,k) = zdata(lbot)
                            numextra = numextra + 1
                            cycle                               
                            
                            ! lbot < linn, but finn3d(lbot+1) is invalid. look for valid data w,s,e,n
                            !numvalid=0 ;  upper = 0.0  ;  lower = 0.0
                            !do n=1,4
                            !    if(finn3d(i+no(1,n),j+no(2,n),lbot+1) > undef .and.  &
                            !       finn3d(i+no(1,n),j+no(2,n),lbot)   > undef) then
                            !        numvalid = numvalid + 1
                            !        upper = upper + finn3d(i+no(1,n),j+no(2,n),lbot)
                            !        lower = lower + finn3d(i+no(1,n),j+no(2,n),lbot+1)
                            !    end if
                            !end do

                            !if(numvalid == 0) then
                               ! still 0? try even harder. now look 2 points away
                             !  do n=5,8
                             !     if(finn3d(i+no(1,n),j+no(2,n),lbot+1) > undef .and.  &
                             !          finn3d(i+no(1,n),j+no(2,n),lbot)   > undef) then
                             !        numvalid = numvalid + 1
                             !        upper = upper + finn3d(i+no(1,n),j+no(2,n),lbot)
                             !        lower = lower + finn3d(i+no(1,n),j+no(2,n),lbot+1)
                             !     end if
                             !  end do
                            !end if
                            
                            !if(numvalid > 0)then
                               ! find the average of the valid, surrounding z points
                             !   upper = upper/numvalid
                             !   lower = lower/numvalid
                             !   felt(i,j,k) = upper + (lower-upper)* (zdyp-zinn(lbot)) / (zinn(lbot+1)-zinn(lbot))
                             !   cycle
                            !else
                               ! the nearest neighbours did not have valid z points below lbot either. give up?
                               ! write(*,*) myid, "gi,gj=", block%fi+i-1,block%fj+j-1, "H(i,j)=", h(i,j)
                               ! write(*,*) myid, "LBOT=",LBOT," FINN3D(i,j,:)=", zdata(:)
                               ! write(*,*) myid, "ZINN=",zinn
                               ! call abort("target sigma depth below last valid Z data point")

                               ! nah. fall back to old behaviour. if the z data is coarse and we want to run with high KB, it can be
                               ! impractical to fix the input all the time. extrapolate, and warn instead.
                             !  felt(i,j,k) = zdata(lbot)
                             !  numextra = numextra + 1
                             !  cycle
                            !end if
                        end if
                        
                      ! most of the points are in this category
                        do ll = 1, lbot-1
                            if ((zdyp > zinn(ll)) .and. (zdyp <= zinn(ll+1))) then
                                if(zdata(ll) > undef .and. zdata(ll+1) > undef) then
                                    felt(i,j,k) = zdata(ll) + (zdata(ll+1)-zdata(ll)) * &
                                         (zdyp-zinn(ll))/(zinn(ll+1)-zinn(ll))
                                else
                                    write(*,*) myid, "gi,gj=", block%fi+i-1,block%fj+j-1, "H(i,j)=", h(i,j)
                                    write(*,*) myid, "LBOT=",LBOT," FINN3D(i,j,:)=", zdata(:)
                                    write(*,*) myid, "ZINN=",zinn                                    
                                    call abort("undefined point")
                                end if
                            endif
                        end do
                        
                    end do
                endif
            end do
        end do
    elseif (icode.eq.2) then
        do j = 1, j_eq_jm
            do  i = 1, i_eq_im
                if (dum(i,j) .ne. 0.0) then
                    do ll = 1, linn
                        if (i.eq.1) then
                            dyp = h(i,j)
                        else
                            dyp = 0.5*(h(i-1,j) + h(i,j))
                        endif
                        if (zinn(ll) .lt. dyp) then
                            lbot = ll
                        endif
                    end do

                    do k = 1, kb-1
                        if (i.eq.1) then
                            dyp = h(i,j)
                        else
                            dyp = 0.5*(h(i-1,j) + h(i,j))
                        endif
                        zdyp = - zz(k) * dyp
                        if (zdyp .lt. zinn(1)) then
                            felt(i,j,k) = finn3d(i,j,1)
                        elseif (zdyp .ge. zinn(lbot)) then
                            felt(i,j,k) = finn3d(i,j,lbot)
                        else
                            do ll = 1, lbot-1
                                if ((zdyp .ge. zinn(ll)) .and. (zdyp .lt. zinn(ll+1))) then
                                    felt(i,j,k) = finn3d(i,j,ll) + (finn3d(i,j,ll+1)-finn3d(i,j,ll)) &
                                         &             * (zdyp-zinn(ll)) / (zinn(ll+1)-zinn(ll))
                                endif
                            end do
                        endif
                    end do
                endif
            end do
        end do
    elseif (icode.eq.3) then
        do j = 1, j_eq_jm
            do  i = 1, i_eq_im
                if (dvm(i,j) .ne. 0.0) then
                    do ll = 1, linn
                        if (j.eq.1) then
                            dyp = h(i,j)
                        else
                            dyp = 0.5*(h(i,j-1) + h(i,j))
                        endif
                        if (zinn(ll) .lt. dyp) then
                            lbot = ll
                        endif
                    end do

                    do k = 1, kb-1
                        if (j.eq.1) then
                            dyp = h(i,j)
                        else
                            dyp = 0.5*(h(i,j-1) + h(i,j))
                        endif
                        zdyp = - zz(k) * dyp
                        if (zdyp .lt. zinn(1)) then
                            felt(i,j,k) = finn3d(i,j,1)
                        elseif (zdyp .ge. zinn(lbot)) then
                            felt(i,j,k) = finn3d(i,j,lbot)
                        else
                            do ll = 1, lbot-1
                                if ((zdyp .ge. zinn(ll)) .and. (zdyp .lt. zinn(ll+1))) then
                                    felt(i,j,k) = finn3d(i,j,ll) + (finn3d(i,j,ll+1)-finn3d(i,j,ll)) &
                                         &             * (zdyp-zinn(ll)) / (zinn(ll+1)-zinn(ll))
                                endif
                            end do
                        endif
                    end do
                endif
            end do
        end do
    endif

    call reduce("sum", numextra)
    if(numextra > 1 .and. master) then
       write(*,*) "WARNING potentially inaccurate extrapolation downward has been performed by Z2SIGMA"
       write(*,*) "        this may lead to unphysical horizontal gradients near the bottom"
       write(*,*) "        NUMEXTRA=",numextra," (globally)"
       write(*,*) "        MAXVAL(ZDATA)=",maxval(zinn)
       write(*,*) "        DEPMAX=",depmax
       
    end if
!  END SUBROUTINE Z2SIGMA
