module setup2
!
! utilities for initializing bom variables and features. replaces
! setup.f90, a lot of dypgrin.f90 and fieldinit.f90. Most of the code
! in this file is "select case" statements, acting on "keywords"
! provided in the setup file. The top level routines comes first, see
! e.g. read_init_file which calls parse_setup_line for each keyword in
! the config file.
!
! usage:
!
!   use setup2
!   ...
!   call read_init_file('setupfile2.dat')
!   ...
!   call sanity_check
!
! just before the model starts, one should call "sanity_check" or
! similar to see whether the initialization looks reasonable.
! 
! Helge Avlesen
!
  use mod_bound
  use control_parameters
  use state
  use divinterfaces
  use mod_concentration
  use util
  use grid_util
  use norms
  use time_util
  use intpol
  use file_util
  use bomplot
  use pgplot

  use timeseries
  use mod_forcing
  use checkpointing
  use etaimplicit
  use mod_averaging
  
  implicit none

  private

  public read_init_file, initialize_parallelization


  INTERFACE SET_VAR
!
!  SUBROUTINE SET_VAR(F, SPLITLINE [,RANGE])     
!    general "setter" routine, that assign values to the variable F
!    according to the content of SPLITLINE. see the list of module
!    procedures a few lines ahead for supported types. RANGE, if
!    present, is the Fortran range (global indices) in which to set F.
!
!    scalars and 1d arrays (which are assumed to be vertical only for
!    now) are typically read from splitline in parallel, or read from
!    file and broadcast.
!
     module procedure &
          set_real4, &
          set_real8, &
          set_integer4, &
          set_integer8, &
          set_string, &
          set_logical, &
          set_1dinteger4, &
          set_1dreal4, &
          set_1dreal8, &
          set_2dreal4, &
          set_2dreal8, &
          set_3dreal4, &
          set_3dreal8
  end interface

  interface simple_forcing
     module procedure add_3d_simple_forcing, add_2d_simple_forcing
  end interface

  interface set_bc
     module procedure set_bc2d, set_bc3d
  end interface
 

  logical :: setlonlat=.true.             ! if true, compute lon,lat for each grid point with the given polar st.transform.
  logical :: advect_density_set=.false.
  logical :: adjust_bz=.true.
  logical :: close_h=.false., edge_closed(4)=.false.


! how many times to run the shapiro filter to smooth the sponge fields:
  integer :: sponge_shapiro = 10
  logical :: sponge_set=.false.
  character(2) :: smagorinsky_sponge_type="am"

  
! a couple of variables as book-keeping for which variables are set or not

  logical, public :: &
       advect_s_set=.false., advect_t_set=.false., advturb_set=.false.,&
       ah_set=.false., am_set=.false., am2_set=.false., am2d_set=.false.,&
       atmpress_set=.false., cbc_set=.false., cd_set=.false., ch_set=.false.,&
       cm_set=.false., cm2d_set=.false., cor_set=.false.,&
       debug_set=.false., debugnorm_set=.false., debugnormlevel_set=.false.,&
       dodumpmask_set=.false., store_sigma_layers_set=.false., &
       dx_set=.false., dy_set=.false.,&
       eta_set=.false., eta_bound_set=.false., fourth_set=.false.,&
       grav_set=.false., h_set=.false., &
       im_set=.false., intp4_set=.false., jm_set=.false., kb_set=.false.,&
       kh_set=.false., khmin_set=.false., km_set=.false., kmmin_set=.false.,&
       kq_set=.false., lb_set=.false., n2d_set=.false.,&
       ncalfa_dte0_set=.false., ndate_set=.false., ndivis_set=.false.,&
       nonhyd_set=.false., numbli_set=.false., numblj_set=.false.,&
       numstep_set=.false., patm_set=.false., pmean_set=.false.,&
       q2_set=.false., q2_bound_set=.false., q2l_set=.false.,&
       q2l_bound_set=.false., rho_set=.false., rho0_set=.false., rhoback_set=.false., &
       rhosurf_set=.false., river_set=.false., rivlog_set=.false.,&
       well_conc_set=.false., river_conc_set=.false.,&
       roair_set=.false., s_set=.false., s_bound_set=.false.,&
       smag_set=.false.,&
       sor_check_resid_interval_set=.false., sor_numiter_set=.false.,&
       sor_resid_set=.false., ssurf_set=.false., t_set=.false.,&
       t_bound_set=.false., theta2d_set=.false., tide_set=.false.,&
       tide_east_set=.false., tide_north_set=.false., tide_south_set=.false.,&
       tide_west_set=.false., tsurf_set=.false.,&
       u_set=.false., ua_bound_set=.false., u_bound_set=.false.,&
       unesco_set=.false., upwell_set=.false., v_set=.false.,&
       va_bound_set=.false., v_bound_set=.false.,&
       vertical_closure_set=.false., viscpom_set=.false., w_set=.false.,&
       w_bound_set=.false., well_set=.false.,&
       wind_forcing_set=.false.,wind_set=.false., z_set=.false.,&
       z0_set=.false., p_set=.false.,&
       lx_set=.false., ly_set=.false., un_set=.false., delun_set=.false.,&
       vn_set=.false., delvn_set=.false., wn_set=.false., delwn_set=.false.,&
       sn_set=.false., delsn_set=.false., tn_set=.false., deltn_set=.false.,&
       rhon_set=.false., delrhon_set=.false., unm1_set=.false.,&
       vnm1_set=.false., wnm1_set=.false., rhonm1_set=.false.,&
       snm1_set=.false., tnm1_set=.false., s_initial_set=.false.,rho_initial_set=.false.,&
       t_initial_set=.false., wadv_set=.false., etabfrs_set=.false., etatfrs_set=.false., &
       rho_bound_set=.false., uadv_set=.false., vadv_set=.false., wubot_set=.false., &
       wvbot_set=.false., timezone_set=.false., internal_pressure_set=.false.,&
       xpole_set=.false., ypole_set=.false.,ylong_set=.false., eos_set=.false.,&
       rhobackfromrho=.false.

  
! this switch changes the behaviour of SET_VAR. if true, set_var
! assumes the f argument to be a normal "non distributed" array.
! activate this behaviour by calling "activate_std_array_setting()",
! deactivate by calling "deactivate_std_array_setting()"  
  logical :: init_std_array=.false.
  
! temporary ghost width storage used by the above switch
  integer :: backup_std_array_gwi=0, backup_std_array_gwj=0

! during init it can be handy to have the global bathymetry on all or at least the master processor.
  real, allocatable :: gh(:,:)
  
  
contains


  subroutine read_init_file(file)
!
! this is a simple file parsing routine for BOM init files.
! the file format for the different init files for BOM follow
! the same basic rules:
!
!   -blank lines are allowed
!   -the order of lines normally do not matter (unless clearly stated)
!   -line content after expected data is ignored    
!   -Fortran 90 style comment lines are allowed
!   -each "valid" line in the file is a command to the initialization system. one command
!    cannot cross several lines.
!   -each line starts with a string followed by some data. generally, the string will be
!    used in "select" statements together with the data to tell the initialization
!    system what to initialize and how. the detailed rules should be documented in setupfile.dat.
!
! this routine reads line by line of "file". after ignoring blank lines and
! comments, the routine sends the "valid" lines, split into an array
! of strings, to the separate initialization routines (specified with "type")
!
    implicit none
    character(*) file
    integer :: srcline(max_active_lines)
    character(132) splitline(50), name
    integer :: ierr=0, i


! if we need to give variables a default, which is different from the
! default given in the declaration, do it here:
    call set_pre_setup2_defaults()

!
! Read setup file on the master processor. We have to do this in two
! steps in the mpi version, first read the lines into an array, which
! is then broadcast, and then each processor parses the lines. this
! way as much as possible is done in parallel.
!
    if(master)then
        file_name = file
        call get_lines(file, keywords, lines, srcline)
    end if

!
! Then, on all processors, parse each of the setup lines
!
    call broadcast(keywords)
    call broadcast(lines)
    call broadcast(srcline)

    do i=1,keywords
        current_setup_line = lines(i)
        line_number = srcline(i)
      ! convert line to array of strings
        call split(lines(i), splitline)
        call parse_setup_line(lower(trim(splitline(1))), splitline, ierr)
    end do

    if(master) then
        write(*,'(/,1x,a,i0,a,i0,a,a,///)') 'setup2.f90 done. read ', line_number, ' lines and ',keywords,&
             ' keywords from init file=', trim(file)
    end if

    parsing_done=.true.
    
    if(.not.h_set) then
        write(*,*) "H is not specified, cannot continue"
        call abort()
    end if
    
    deallocate(gh)
    
    if(.not.bom_allocated) then
        write(*,*)'Arrays are not allocated, cannot continue'
        write(*,*)'IM=',im,' JM=',jm, ' KB=',kb
        call abort()
    end if
    
    if(numstep < 0) numstep=0


    if(num_zones_east+num_zones_west+num_zones_north+num_zones_south > 0) then
        call message("Initializing the averaging 'bin' regions for open boundaries:")
        
        call initialize_bin_regions(bin_east, boeast, num_zones_east)
        call initialize_bin_regions(bin_west, bowest, num_zones_west)
        call initialize_bin_regions(bin_north, bonorth, num_zones_north)
        call initialize_bin_regions(bin_south, bosouth, num_zones_south)
    end if   

    
! allocate s_initial etc.
    call allocate_binreg()
    
    
! important variables that are not set by the user should be detected
! and warned for

    call set_defaults()

! apply shapiro filter to sponge arrays if set    
    if(sponge_set) call smooth_sponges()
    
! deallocate unused arrays
    
    if(vertical_closure .ne. 1) deallocate(q2, q2l, q2l_bound, q2_bound)
    
! if jm=1, set the y arrays used in vsect to 0.

    if(jm == 1)then
        do i=1, bomplot_numplots
            bomplots(i)%py = 0.0
        end do
        do i=1, pgplot_numplots
            pgplots(i)%py = 0.0
        end do
    end if

    if(debug > 4 .and. num_zones_west+num_zones_east+num_zones_north+num_zones_south>0) then
        write(name,'(a,i4.4,a)') "bcinfo-", myid, ".dat"
        open(unit=1000010, file=name)
        if(num_zones_west > 0) call dump_bc_summary('west', bowest(1:num_zones_west))
        if(num_zones_east > 0) call dump_bc_summary('east', boeast(1:num_zones_east))
        if(num_zones_north > 0) call dump_bc_summary('north', bonorth(1:num_zones_north))
        if(num_zones_south > 0) call dump_bc_summary('south', bosouth(1:num_zones_south))
        close(1000010)
    end if

! try to create directories for the output files. if fail, the dir name will be "./"
    if(master) then
        call createncheckdir(hsect_dir, patience=0, endslash=.true.,comment="directory for horizontal sections from bomplot.f90")
        call createncheckdir(vsect_dir, patience=0, endslash=.true.,comment="directory for vertical sections from bomplot.f90")
        call createncheckdir(vprof_dir, patience=0, endslash=.true.,comment="directory for vertical profiles from bomplot.f90")
        call createncheckdir(slice_dir, patience=0, endslash=.true.,comment="directory for array slices from bomplot.f90")
        call createncheckdir(timeseries_dir, patience=0, endslash=.true.,comment="directory for timeseries from timeseries.f90")
        call createncheckdir(average_dir, patience=0, endslash=.true.,comment="directory for time averaged data from mod_average.f90")
        call createncheckdir(particlets_dir, patience=0, endslash=.true.,comment="directory for particle trajectories from mod_particle.f90")
        call createncheckdir(particledist_dir, patience=0, endslash=.true.,comment="directory for particle distribution files from mod_particle.f90")
        call createncheckdir(particlestat_dir, patience=0, endslash=.true.,comment="directory for various particle counters in mod_particle.f90")
        
        if(debug > 1) then
            write(*,'(/,a)') "Created the following directories for output:"
            write(*,'(a)') "bomplot:"
            write(*,'(3a)') "  horisontal plots         >",trim(hsect_dir),"<"
            write(*,'(3a)') "  vertical cross sections  >",trim(vsect_dir),"<"
            write(*,'(3a)') "  vertical profiles        >",trim(vprof_dir),"<"
            write(*,'(3a)') "  array slices             >",trim(slice_dir),"<"
            write(*,'(3a)') "timeseries                 >",trim(timeseries_dir),"<"
            write(*,'(3a)') "time averaged fields       >",trim(average_dir),"<"
            write(*,'(3a)') "particles:"
            write(*,'(3a)') "  timeseries               >",trim(particlets_dir),"<"
            write(*,'(3a)') "  counters/statistics      >",trim(particlestat_dir),"<"
            write(*,'(3a,/)') "  spatial distribution     >",trim(particledist_dir),"<"
        end if
    end if
    
    call dump_state('read_init_file')
  end subroutine read_init_file
  

  subroutine parse_setup_line(name, splitline, ierr)
!
! Each "splitline" in the config file begins with a keyword, which is
! a hint for how to interpret the rest of the line. typically a
! keyword equal to a variable name will be followed by info on how to
! initialize that variable. other keywords will set boundary
! conditions or forcing.
!
! The main "doer" is the SET_VAR interface. the compiler will select
! the routine listed in the interface declaration that matches the
! type of the first argument of set_var. Example: if IM is an integer
! of kind 4, the statement "CALL SET_VAR(IM,...)" will be translated
! by the compiler to "CALL SET_INTEGER4(IM,...)"
!    
! Variables can be initialized directly from the value in the second
! column of setupfile.dat, non scalars like arrays can also be
! initialized several ways according to the value of the "code" in the
! second column. see example setupfile2.dat for documentation.
!
! All known keywords should have a "case" in this routine, with
! documentation on the lines just below the case statement.
!
    use maxmin
    use mod_drag_region
    use mod_cruise
    use mod_averaging
    implicit none
    character(*) name, splitline(:)
    integer ierr, zone, tmpint

    ierr=0
    keyword = name

    if(master .and. debug > 0) write(*,'(a,":")') name

    select case(name)


    case('<variable>')
! THE MEANING OF KEYWORDS IN THE INIT SYSTEM
!
! Each line in the setup file generally has the following format        
!   <keyword> <specification>        
! <keyword> is most often a variable name, but can also denote another
! action to be taken.  <specification> can both say what to do, or it can be data.
!      
! First, about the initialization of model variables. If <keyword> below is the name of
! a model variable, such as U, ETA, IM or NONHYD, then for
!      
! 1) scalars or strings: the value is read directly from column 2.
!        <keyword> <value>
!     Example:
!         IM 22
!     will set the first grid dimension to 22.
!        
! 2) arrays; the syntax is
!        <keyword> <code> <data>
!     <code> is an integer, which indicates how to parse the rest of the line (<data>)
!        
!     A common sub keyword of <data> is <range>, which is the
!     horisontal grid range (in meters) where the assignment should be
!     done.  If all values for <range> is 0, e.g. 0,0,0,0 , the whole
!     array will be assigned the given value.
!        
!     Valid values for code is
!        
!        <keyword>  0 <value>
!           assign scalar <value> to the whole array. Example
!             eta 0 0
!           would set elevation to 0 for the whole domain.
!        
!        <keyword>  1  <f(1,1,1)>,...,<f(im,jm,kb)>
!           read directly from rest of line, not likely, but if thats whats asked for...
!        
!        <keyword>  2 <file> <col>
!           read full array from spreadsheet with filename <file>, and column <col>;
!           the given column contains the array data stored in column major order (first indice varies fastest),
!           as usual for Fortran.
!        
!        <keyword>  3 <file> <col1> <col2> <range>
!           read vertical depths and values from column <col1> and <col2> respectively, in the        
!           "spreadsheet" type ascii file named <file>, optionally only assign the values to the 
!           horisontal region given in <range>. set range to 0,0,0,0 for the full range as usual.
!        
!        <keyword>  4 <file> <weighting> <range>
!           interpolation between multiple vertical z profiles.
!        
!           the WEIGHTING argument, by default -1.0 (inverse distance squared),
!           is the exponent of the square of the distance between the current
!           point and a interpolated profile value, used in the weighting. -0.5
!           gives 1/distance as weight, -0.25 gives 1/sqrt(distance) as weight. read code for
!           further info :)
!        
!        <keyword>  5  ... horisontal interpolation from scattered points
!           not yet implemented. only for 2d arrays.
!        
!        <keyword>  6  <fmax> <fmin> <deltaz> <zsymm>
!           horisontal extrapolation from tanh profile
!        
!        <keyword>  7  <value> <range> <k1> <k2>
!           assign value to array region and a specific sigma layer range k1 to k2.
!        
!        <keyword>  8  <fmax> <fmin> <deltaz>
!           horisontal extrapolation from exp profile
!        
!        <keyword>  9  <file>
!           read global array directly from unformatted file (binary io)
!        
!        <keyword>  10  <value> <range> <Z1> <Z2>
!           assign value to a horisontal array region, and from depth z1 to z2.        
!           z1,z2 must be specified as negative depths, i.e. a correct
!           spec is from z1=-10 to z2=-5, so that z2>z1.
!        
!        <keyword>  11   <file> <numlevels> <level file>        
!           read global Z coordinate array from single precision
!           float, unformatted stream file (binary io), and interpolate to
!           sigma.
!
!        <keyword>  12  <file>
!           read global array directly from stream access unformatted file (binary io)
!        
! So as you see, the setup system requires intimate knowledge of core
! BOM variable names and algorithms, which is by design. The overall
! goal behind the new setup system has been to minimize the number of
! code changes needed to run a large variety of cases, as well as
! avoiding unessarey recompilation. the cost is of course some code
! complexity and bloat... but we try to keep it simple.
!
!        
! Now we, list the name and options for all valid keywords
!        
        call set_var(tmpint, (/"tmpint","0     "/))!(dummy assignment that is here to mark the end of the doc string.)
        
    case('restart')
! restart <logical value>        
!   If true, the current setup file should be handled as a restart
!   file. This logical should always be the first in the setup file,
!   if present.
        call set_var(restart, splitline)
        
     case('data_path')
! data_path <string with path>
!   By default "" (empty). If set, prepend the path to all file names for
!   input data listed in the setupfile. This must typically be used on
!   architectures that cannot follow symbolic links.
        call set_var(data_path, splitline)

        !! append / to end of string if not already there
        if(trim(data_path) .ne. "") then
            tmpint = len_trim(data_path)
            if(data_path(tmpint:tmpint) .ne. "/") then
                data_path(tmpint+1:tmpint+1) = "/"
            end if
        end if

    case('im')
! im <integer value>
!   the number of grid boxes in the x direction
        if(.not.bom_allocated)then
            call set_var(im, splitline)
        else
            call message('Arrays already allocated, too late to set this variable',1)
        end if
        im_set=.true.

    case('jm')
! jm <integer value>
!   the number of grid boxes in the y direction
        if(.not.bom_allocated)then
            call set_var(jm, splitline)
        else
            call message('Arrays already allocated, too late to set this variable',1)
        end if
        jm_set=.true.

    case('kb')
! kb <integer value>
!   the number of grid boxes in the sigma (z) direction
        if(.not.bom_allocated)then
            call set_var(kb, splitline)
        else
            call message('Arrays already allocated, too late to set this variable',1)
        end if
        kb_set=.true.
        
    case('omit_dry_blocks')
! omit_dry_blocks <logical value>
!   if true, blocks with only land cells are omitted from the decomposition.
!   BOM must then be started with the correct number of processors and a given block 
!   structure NUMBLI,NUMBLJ
!   must be specified before H                
        if(h_set) call message("this variable must be set before H to have effect",2)
        call set_var(omit_dry_blocks, splitline)
        
    case('h')
! h <integer code> <data for initialization of 2d arrays>
!   define the bathymetry.        
!   H is a 2d array of dimension im,jm specifying the static depth (m) in the center of each grid box.
!   this is an important point in the setup process, it triggers allocation of most variables, and the initialization
!   of the parallelization. some keywords are required to be set before H, e.g. if they affect allocation.
        if(.not.lb_set) call message("LB not set, will use default value=",1,ival=(/lb/))
        call initialize_depth(splitline)
        h_set=.true.

    case('close_h')
! close_h <logical value>
!   if true, close all open boundaries in the provided grid. must be given before h is specified.        
        if(h_set) call message("this keyword must be set before H",2)
        call set_var(close_h, splitline)
        call message("Closing open boundaries",1)

    case('close_boundary')
! close_boundary <[north] [south] [east] [west]>
!   close the boundary on all edges in the list, example, close east and north:
!     close_boundary east north. 
!   must be specified before H
        if(h_set) call message("this keyword must be specified before H",2)
        call set_edge_closed(splitline)

    case('default_boundary_condition')
! default_boundary_condition <integer value>
!   specifies which open boundary condition to use if the BC is not specified for a variable. by default, i.e.
!   if this variable is not set, the default is 0, i.e. FRS. must be specified before H.

        if(h_set) call message("This keyword should be specified early, at least before H",2)        
        call set_var(default_boundary_condition, splitline)
        call change_all_bc_type(bowest, default_boundary_condition)
        call change_all_bc_type(boeast, default_boundary_condition)
        call change_all_bc_type(bonorth, default_boundary_condition)
        call change_all_bc_type(bosouth, default_boundary_condition)

    case('lb')
! lb <integer value>
!   the number of boxes across the frs zone. must be specified before H
        if(h_set) call message("LB must be defined before H",2)
        call set_var(lb, splitline)
        lb_set=.true.

    case('ncalfa_dte0')
! ncalfa_dte0 <real value>
!   the dte0 parameter of the alpha computation for frs can be overridden here. if not set
!   the default value will be SQRT(2*G*MAX(H))
        call set_var(ncalfa_dte0, splitline)
        ncalfa_dte0_set=.true.

    case('lx')
! lx <real value>
!   the full extent of the grid in the x direction in meters. if this is set, do not set dx.
!   lx is the distance in meters from the right edge of grid cell 1,1 to the left edge of cell im,1.
        if(dx_set) call message("dx already set, you should set lx OR dx",2)
        call set_var(lx, splitline)
        lx_set=.true.

    case('ly')
! ly <real value>
!   the full extent of the grid in the y direction in meters. if this is set, do not set dy
        if(dy_set) call message("dy already set, you should set ly OR dy",2)
        call set_var(ly, splitline)
        ly_set=.true.

    case('dx')
! dx <real value>
!   The grid spacing in the x-direction in meters
        if(lx_set) call message("lx already set, you should set lx OR dx",2)
        call set_var(dx, splitline)
        dx_set=.true.

    case('dy')
! dy <real value>
!   The grid spacing in y-direction in meters
        if(ly_set) call message("ly already set, you should set ly OR dy",2)
        call set_var(dy, splitline)
        dy_set=.true.

    case('xpole')
! xpole <real value>
!   The x coordinate of the North pole in meters if we use a polar stereographic grid.
        call set_var(xpole, splitline)
        xpole_set=.true.
        if(all((/xpole_set,ypole_set,ylong_set/))) call init_lonlat()
        
    case('ypole')
! ypole <real value>
!   The y coordinate of the North pole in meters if we use a polar stereographic grid.
        call set_var(ypole, splitline)
        ypole_set=.true.
        if(all((/xpole_set,ypole_set,ylong_set/))) call init_lonlat()

    case('ylong')
! ylong <real value>        
!   The meridian parallel to the y axis if we are in a polar stereographic grid. 
!   by default 58 degrees.
        call set_var(ylong, splitline)
        ylong_set=.true.
        if(all((/xpole_set,ypole_set,ylong_set/))) call init_lonlat()
        
    case('numbli')
! numbli <integer value>
!   Explicitly set the number of MPI blocks in the x-direction. must be specified before H.
        if(h_set) call message("this variable must be set before H to have effect",2)
        call set_var(numbli, splitline)
        numbli_set=.true.

    case('numblj')
! numblj  <integer value>
!   Explicitly set the number of MPI blocks in y-direction. must be specified before H.
        if(h_set) call message("this variable must be set before H to have effect",2)        
        call set_var(numblj, splitline)
        numblj_set=.true.

    case('mod_mpi_gum')
! mod_mpi_gum <integer value>
!   ghost update message passing method, 0 (default) = mpi_send,recv, 1=mpi=isend,irecv,2=mpi_sendrecv.
!   1 is normally faster, but not all implementations are stable. 0 is slowest but normally robust.
!   2 can be fast on some systems, e.g. with grid type networks with wrapped ring topology and fast shift operations.
        call set_var(mod_mpi_gum, splitline)

    case('z')
! z <integer code> <data for initialization of 1d arrays>
!   1d array of dimension KB containing (sigma) coordinates of the
!   cell interfaces, starting with the surface position. this keyword is used to
!   set z directly, alternatively the keyword sigma_layers can be used
!   to choose a predefined distribution.

        if(z_set) call message("the sigma layer distribution has already been specified",2)
        call set_var(z, splitline)

        call depth(kb,z,dz,dzr,dzz,zz)
        z_set=.true.

    case('sigma_layers')
! sigma_layers <integer code> <parameters>
!   The code specifies the distribution of sigma layers:
!
!   code=0: Gives a uniform spacing
!   code=1: Dan Lynch tanh distribution, where column 3 is the thickness of
!           the bottom layer (DSBOT), e.g.
!
!              sigma_layers 1 0.01
!
!           The tanh profile has thin layers near surface and bottom,
!           gradually coarser towards the interior.
!   code=2: Initalize from filename in column 3, e.g.
!        
!              sigma_layers 2 sigmadepth.dat
!        
!   code=3: Use geometrical series with stretching parameter r from column 3, e.g.        
!
!              sigma_layers 3 <r>
!        
!           The stretching parameter r has the following function:
!              r = 1.0  uniform distribution
!              r < 1.0  z points are denser near Z=0 (surface)
!              r > 1.0  z points are denser near Z=-1 (sea bottom)
!           r is normally in the range 0.7 to 1.3        
!
        if(z_set) call message("the sigma layer distribution has already been specified",2)
        call initialize_sigma_layers(splitline)
        z_set=.true.

    case('open_boundary')
! open_boundary <edge> <start> <end>
!   Specify an open boundary region on the given <edge>, which can be one of east, west, north, south.
!   <start>,<end> are integer values for the start and end index respectively.
!   Using this keyword ONCE on an edge, will turn off autodetection of open boundaries on this edge.
!   This means, if at least *one* open region is specified manually on an edge, *all* open regions
!   that are wanted on this edge, must be specified. Example, define two openings on the northern edge:
!       ...
!       open_boundary north 4  45
!       open_boundary north 55 101
!       ...
!   The open_boundary keyword are only allowed to be used before the H keyword.
!
        if(h_set) call message("open_boundary is only allowed to be used before H is specified",2)
        call add_open_boundary(lower(trim(splitline(2))), splitline)

    case('ndivis')
! ndivis <integer value>
!   The Number of time steps per hour. This is the keyword to use to set the 3D time step length.
        call set_var(ndivis, splitline)
        dt = 3600./ndivis
        if(master.and.debug>0) write(*,*)' dt =',dt
        ndivis_set=.true.

    case('n2d')
! n2d <integer value>
!   The number of 2D steps per 3D step. This keyword specifies the 2D time step length.
        call set_var(n2d, splitline)
        n2d_set=.true.

    case('sanity_check_fatal')
! sanity_check_fatal <logical value>
!   If .true., the sanity check is fatal, i.e. we abort the run after
!   all tests are performed if at least one error turn up.        
!   The default value is .true.
        call set_var(sanity_check_fatal, splitline)
        
    case('adjust_bz')
! adjust_bz <logical value>
!   set this to false to turn off the automatic adjustment of boundary zone extents.
!   some bathymetries can make the algorithm go bananas...
        call set_var(adjust_bz, splitline)

    case('numstep')
! numstep <integer value>
!   The total number of (3D) time steps.
        call set_var(tmpint, splitline)
        numstep_set=.true.
        if(num_ts > 0 .and. tmpint > numstep) &
             call message("timeseries has been allocated based on an earlier NUMSTEP, increasing NUMSTEP is not allowed",2)
        numstep = tmpint
        
    case('sor_numiter')
! sor_numiter <integer value>
!   The maximum number of iterations in the SOR solver for non hydrostatic pressure
!   If not set, the default value is 20.
        call set_var(sor_numiter, splitline)
        sor_numiter_set=.true.

    case('sor_resid')
! sor_resid <real value>
!   Relative residual, stopping criteria for the SOR solver.
!   If not set, the default is sor_resid = 1.0e-2. This is usually
!   sufficient to see the effect of non hydrostatic pressure and will
!   give very little computational overhead.
        call set_var(sor_resid, splitline)
        sor_resid_set=.true.

    case('sor_check_resid_interval')
! sor_check_resid_interval <integer value>
!   The number of SOR iterations between each stopping criteria test. By Default 20.
        call set_var(sor_check_resid_interval, splitline)
        sor_check_resid_interval_set=.true.

    case('debug')
! debug <integer value>
!   General debug output level, 0=nothing, 1=informational, 2=very detailed, for debugging.
        if(splitline(2) == "0" .and. master) write(*,'(a,/)')" setting debug level to 0"
        debug = 0
        call set_var(debug, splitline)
        debug_set=.true.

    case('qmpi_debug')
! qmpi_debug <integer value>
!   Specific debug flag for the parallelization layer. 0=nothing, 1=informational, 2=very detailed
        call set_var(qmpi_debug, splitline)
        
    case('array_default_init_value')
! array_default_init_value <real value>
!   By default 0.0
!   Most allocatable arrays will be set to this value just after allocation in allocate_bom.f90.
!   Meant as an aid to hunt for code that is using uninitialized data. Set it to a crazy value e.g. -1e-32
!   and the code may blow up quickly. must be specified before H.
        if(h_set)then
            call message(" IGNORED. Must come before H, as H trigger the allocation and default initialization of arrays",1)
        else            
            call set_var(adiv, splitline)
        end if
        
    case('viewblocks')
! viewblocks <logical value>       
!    list block setup during init. must be specified before H. defaults to true.
        call set_var(viewblocks, splitline)

    case('first_dump_state')
! First_dump_state <integer value>
!   The first timestep on which to evaluate norms. By default 0.
        call set_var(first_dump_state, splitline)

    case('dump_state_barrier')
! dump_state_barrier <logical value>       
!    Call barrier each time a dump_state label is written. By default false.
        call set_var(dump_state_barrier, splitline)

    case('debugnormlevel')
! debugnormlevel <integer value>
!   How much to print at each dump_state call:
!
!   debugnormlevel=0  complete silence
!   debugnormlevel=1  just print the label and the wall time since last call to observe progress
!   debugnormlevel=2  print all selected maxmin's
!   debugnormlevel=3  print all selected maxmin's also in 2d loop.
!   debugnormlevel=4  print selected 3d norms and "wet" maxmin values
!   debugnormlevel=5  also print selected norms in 2d loop (see modesplit.f90)

        call set_var(debugnormlevel, splitline)
        debugnormlevel_set=.true.

    case('debugnorm')
! debugnorm <integer value>
!   Choose type of norm/checksum the dump_state routine computes for a set of arrays
!   debugnorm=0: by default "Kahan" sum, if OMIT_DRYBLOCKS=false, gives the parallel, &
!                MPI decomposition independent, sum
!   debugnorm=1: Gives the MAXVAL of the ABS of each array
!   debugnorm=2: Converts floats to INTEGER KIND 2 (short int) and sum up. (used to check for bit exactness)
!   debugnorm=3: Take ICHAR of each byte in the array and sum. Similar to 2, but since ICHAR always >0 safer.

        call set_var(debugnorm, splitline)
        debugnorm_set=.true.

    case('ndate')
! ndate <integers; year month day hour minute second>
!   Start time for the simulator clock
        call set_var(ndate, splitline)
        ndate_set=.true.
        if(.not.restart) then
           initial_time = ndate
        end if
        
    case('timezone')
! timezone <real; hour>
!   Global time zone offset. Necessary for computing correct astronomical arguments for tides.
!   Example; If the model clock, initialized via NDATE, is in Norwegian time (CET = Central Europe Time)
!      timezone 1.0
        call set_var(timezone, splitline)
        if(timezone > 14. .or. timezone < -12.) call message("invalid value for timezone",2)
        timezone_set=.true.
        
    case('theta2d')
! theta2d <real value>        
!   The degree of implicitness for the 2D mode "theta method" corrector step (predictor=leapfrog). 
!   For stability it is required that 0.5 <= theta2d <= 1.
!   theta2d=1: fully implicit, theta2d=0.5 trapezoidal method, 0 explicit.

        call set_var(theta2d, splitline)
        theta2d_set=.true.

    case('am2')
! am2 <logical value>
!   The corrector scheme for 3d time stepping; t=Adams-Moulton, f=Euler backward
        call set_var(am2, splitline)
        am2_set=.true.

    case('vertical_closure')
! vertical_closure <integer value>
!   Turbulence scheme used in the vertical direction; 0=none,
!   i.e.constant eddy viscosity, 1=Mellor-Yamada.
        call set_var(vertical_closure, splitline)
        vertical_closure_set=.true.

    case('implicit_eta')
! implicit_eta <logical value>      
!   if true, use the original BOM implicit solver for ETA, and a
!   operator splitting technique for the rest.
        call set_var(implicit_eta, splitline)

    case('eta_numiter')
! eta_numiter <integer value>      
!   if implicit_eta is true, use this number of iterations for the implicit SHW solver.
        call set_var(eta_numiter, splitline)
        
    case('nonhyd')
! nonhyd <logical value>
!   f=hydrostatic, t=non hydrostatic (turn on the 3d pressure solver)
!   must be specified before H        
        if(h_set) call message("nonhyd must be specified before H",2)
        call set_var(nonhyd, splitline)
        nonhyd_set=.true.

    case('concentration')
        call message("obsolete variable. if one of conc1 to conc9 is specified, they are advected",2)
    case('conc')
        call message("obsolete variable, use conc1 to conc9 instead",2)
    case('conc15')
        call message("obsolete variable, use conc1 to conc9 instead",2)

    case('conc1')
! conc<N> <integer code> <data for initialization of 3d arrays>
!   add and initialize another 3d scalar field to be advected
!   <N>             integer id for field. we support up to MAXNUMCONC fields. see mod_concentration.f90        
!   <integer code>  the usual code for initializing 3d arrays.
!
        call set_concentration(1, splitline)
    case('conc2')
        call set_concentration(2, splitline)
    case('conc3')
        call set_concentration(3, splitline)
    case('conc4')
        call set_concentration(4, splitline)
    case('conc5')
        call set_concentration(5, splitline)
    case('conc6')
        call set_concentration(6, splitline)
    case('conc7')
        call set_concentration(7, splitline)
    case('conc8')
        call set_concentration(8, splitline)
    case('conc9')
        call set_concentration(9, splitline)
!   ....
        
     case('conc_time')
! conc_time <concentration id number> <time>
!    Time for activation of tracer field <id>. Time can be given in seconds since sim. start,
!    or as a ISO date (must have the T to be recognized as such)
        if(.not. ndate_set) call message("NDATE must be given before this varaible is set",2)
        call add_concentration_time(splitline)

    case('conc_ah')
! conc_ah <concentration id number> <data for initialization of 3d array>
!   Let concentration with given id be advected with its own diffusion prescribed in a 3D array.
        call set_conc_ah(splitline)
    case('conc_kh')
! conc_kh <concentration id number> <data for initialization of 2d array>
!   Let concentration with given id be advected with its own diffusion prescribed in a 2D array.
!   the KH for the given CONC will be constant in the vertical but allowed to vary horizontally.
!   CONC_KH can be applied multiple times for different regions, the resulting 2D field will be 
!   shapiro filtered, before the 2D field is extrapolated onto the final 3D KH field for the given CONC.
        call set_conc_kh(splitline)
        
    case('conc_name')
! conc_name <concentration id number> <name>
!   give concentration a useful name. if not given, it defaults to concN
        read(splitline(2),*) tmpint
        if(tmpint < 1 .or. tmpint > maxnumconc) call message("conc number out of range",2)
        co(tmpint)%name = splitline(3)
        
    case('drag_region')
! drag_region <drag id> <cd region> <x1> <x2> <y1> <y2> <z1> <z2>
!   add drag with coeff CD to region <drag id> spanned by given coordinate range
        if(.not. h_set .or. .not. z_set) call message('Too early to specify drag_regions, specify z (or sigma_layers) and h first!',2)
        call initialize_drag_region(splitline)

    case('upwell')
! upwell <logical value>
!   if true, include the effect of bottom point sources defined with
!   the well keyword, by default false (f)
        call set_var(upwell, splitline)
        upwell_set=.true.

    case('smag')
! smag <logical value>
!   If true use the Smagorinsky mixing length model for horisontal eddy viscosity.
        call set_var(smag, splitline)
        smag_set=.true.

    case('viscpom')
! viscpom <logical value>
!   if true, use POM scheme for horizontal viscosity
        call set_var(viscpom, splitline)
        viscpom_set=.true.

    case('unesco')
! unesco <logical value>
!   if true use UNESCO formulation of Gill to compute density, if false (the default) use less accurate
!   formulation by Wang (JPO)
        if(.not. eos_set) call message("This is the old way to choose EOS, use the EOS keyword instead!",1)
        if(eos_set) call message("Warning: overriding the new EOS keyword that has been set earlier",1)
        call set_var(unesco, splitline)
        if(unesco) then
            eos="unesco"
        else
            eos="wang"
        end if
        eos_set = .true.
        unesco_set=.true.

    case('eos')
! eos <string>
!   choose the equation of state to use to compute density. valid options are:
!     "wang"     the default, same as POM
!     "unesco"   the Unesco formulation from Gill.
!     "user"     call user/case specific EOS subroutine USER_EOS().
        if(s_set) call message("S must not be specified before this keyword",2)
        call set_var(eos, splitline)
        eos_set = .true.
        
    case('advturb')
! advturb <logical value>
!   if true, advect (horisontally) the Mellor Yamada variables Q2 and Q2L
        call set_var(advturb, splitline)
        advturb_set=.true.

    case('khmin')
! khmin <real value>
!   Minimum vertical diffusivity (m^2s^(-1)).
        if(sponge_set) call message("khmin should be set before the add_sponge keyword",2)
        call set_var(khmin, splitline)
        khmin_set=.true.

    case('kmmin')
! kmmin <real value>
!   Minimum vertical viscosity (m^2s^(-1)).
        if(sponge_set) call message("kmmin should be set before the add_sponge keyword",2)
        call set_var(kmmin, splitline)
        kmmin_set=.true.

    case('cm')
! cm <real value>
!   The horizontal 3d mixing coeff for the Smagorinsky scheme
        if(sponge_set) call message("cm must be set before the add_sponge keyword",2)
        call set_var(cm, splitline)
        cm_set=.true.

    case('cm2d')
! cm2d <real value>
!   The horizontal 2D mixing coeff. for the Smagorinsky scheme
        if(sponge_set) call message("cm2d must be set before the add_sponge keyword",2)
        call set_var(cm2d, splitline)
        cm2d_set=.true.

    case('cor')
! cor <integer code> <data for initialization of 2d real array>
!   A 2d grid array with the Coriolis parameter f (s^(-1))
        call set_var(cor, splitline)
        cor_set=.true.

    case('cbc')
! cbc <integer code> <data for initialization of 2d real array>
!   bottom drag coefficient. if not set, it defaults to
!   MAX(CD,.16/LOG((ZZ(KB-1)-Z(KB))*H(I,J)/Z0)**2)
        call set_var(cbc, splitline)
        cbc_set=.true.

    case('cdw')
! cdw <real value>        
!   Use this to specify the lateral (vertical) drag coefficient for
!   W. If not specified, it's default value is 0 (full slip). If set
!   to a value < 0 it will be computed using the standard formulae
!      zb = 0.5*dx
!      cdw = max(0.0025, 0.4**2/(log(zb/z0))**2)
        call set_var(cdw, splitline)
        call message("TODO")

    case('cdland')
! cdland <real value>        
!   Use this to specify the lateral drag coefficient for U,V. If not
!   specified, it's default value is 0 (full slip). If set to a value
!   < 0 it will be computed using the standard formulae        
!      zb = 0.5*dx
!      cdland = max(0.0025, 0.4**2/(log(zb/z0))**2)
        call set_var(cdland, splitline)
        call message("TODO")
        
    case('ch')
! ch <real value>
!   The horizontal mixing coefficient for smagorinsky
        call set_var(ch, splitline)
        ch_set=.true.

    case('internal_pressure')
! internal_pressure <string>
!   A string specifying which scheme to use for the computation of the internal pressure gradient.
!   Valid entries are:
!     "pom"               The standard 2nd order scheme from POM
!     "fourth"            The fourth order scheme of Mccalpin
!     "balanced"          The perfectly balanced scheme of Jarle        
        call set_var(internal_pressure, splitline)
        internal_pressure = lower(internal_pressure)
        internal_pressure_set = .true.
        
    case('intp4')
! intp4 <logical value>
!   (obsoleted by new keyword "internal_pressure")
!   If true (the default) use fourth order accurate internal pressure scheme
        call set_var(intp4, splitline)
        if(intp4)then
            internal_pressure="fourth"
        end if
        call message("This keyword is obsolete. Use keyword INTERNAL_PRESSURE instead",1)
        intp4_set=.true.

    case('fourth')
! fourth <logical value>
!   If true use fourth order accurate advection scheme
        call set_var(fourth, splitline)
        fourth_set=.true.

    case('lag_forcing')
! lag_forcing <integer forcing id>
!   this keyword sets the logical LAG of a forcing object to .TRUE.
!   the result is that the coefficient ALF used to interpolate in time
!   between the previous forcing and the next forcing is always set to
!   0, which means we only use the previous forcing field to compute
!   the boundary value, so called "lagging".
!
!   by using linear interpolation in time between forcing fields, we
!   introduce a discontinuity in the first derivative, that may lead
!   to instability. using "lagging" seems to work more robust, but is
!   of course slightly less accurate
        call lag_forcing(splitline)

    case('wcourantfix')
! wcourantfix <integer value>
!   if we have very thin layers, the vertical courant number can give instabilities/overshooting.
!   one remedy is to increase the viscosity in all cells where WADV is too large.
!   if wcourantfix is > 0, we multiply KM by this factor in such cells (see my2halv.f90)
!   the default value for this variable is 0 (no "fix")        
        call set_var(wcourantfix, splitline)
        
    case('smagorinsky_sponge_type')
! smagorinsky_sponge_type <2 character code>
!   when using sponge regions one can use this keyword to choose what the <value> parameter
!   of the ADD_SPONGE keyword should override;
!     -if smagorinsky_sponge_type is "am", the <value> given via the
!      ADD_SPONGE keyword will override the viscosity AM and/or AM2D
!      directly in the given sponge region. this is the default.        
!     -if smagorinsky_sponge_type is "cm", the sponge layer <value> will
!      instead override CM
!        
        if(sponge_set) call message("if you want to set this, do it before ADD_SPONGE",2)
        call set_var(smagorinsky_sponge_type, splitline)
        if(.not.any(smagorinsky_sponge_type == (/"cm","am"/))) call message("invalid sponge type",2)

    case('sponge_shapiro')
! sponge_shapiro <integer value>
!   the number of times to apply a shapiro filter to the
!   final sponge fields. for a sharp jump, ten iterations will
!   typically produce a gradual transition over 10 grid boxes.
        call set_var(sponge_shapiro, splitline)
        if(sponge_shapiro < 0 .or. sponge_shapiro > im) call message("this looks wrong",2)
        
    case('add_sponge')
! add_sponge <variable> <value> <range>
!   
!   For a given viscosity/diffusion <variable> (AM,AM2D,KM,KH) this
!   keyword let the user override the computed viscosity value with
!   the given <value> for a given horizontal <range>. (<range> is in
!   meters as usual.) This is a common trick to stabilize problematic
!   open boundaries; by dramatically increasing the
!   viscosity/diffusion in a given region, the solution becomes more
!   smooth.
!
!   In the setup file add_sponge can be used multiple times for the
!   same variable, with different values for different regions. A
!   default background value is given for the areas not covered. Note
!   that if regions overlap, the latter keyword will overwrite the
!   former regions, so that it may be an advantage to specify the
!   keywords with <value> in increasing order.
!        
!   The default background value for the sponge fields are:
!     KMMIN and KHMIN for KM,KH respectively, and
!     AMMIN and AM2DMIN for AM and AM2D.
!   If smagorinsky_sponge_type is "cm", the background values are CM and CM2D
!       
        sponge_set = .true.
        call add_sponge(splitline)
        
    case('rivlog')
! rivlog  <logical value>
!   If true include effect of river inflow in all specified cells.
        call set_var(rivlog, splitline)
        rivlog_set=.true.

    case('river')
! river <name> <x> <y> <code> <depth> <format> <data>
!   Keyword used to set up river inflow
!
!   <name>: River name
!   <x>: the x position of the river inlet        
!        if <x> is given as "lon5.34" it is interpreted as a longitude
!        and converted to grid coordinates using the current
!        transform.        
!   <y>: the y position of the river inlet
!        if <y> is given as "lat60.0" it is interpreted as a latitude.
!        
!   <code>: The cell side where the river enter
!           1, western side
!           2, southern side
!           3, eastern side
!           4, northern side
!   <depth>: Depth of the river inlet in meters (m)
!        
!   <format> 0 
!         if we advect S and T        
!               <data> is <friver> <triver> <sriver>               
!               where: <friver>: Water flux (m^3s^(-1))
!                      <triver>: Temperature (C)
!                      <sriver>: Salinity (p.s.u.)
!
!         if we advect RHO (density)
!               <data> is <friver> <rhoriver>
!               where: <friver>: Water flux (m^3s^(-1))
!                      <rhoriver>: Density (kg/m^3)
!        
!   <format> 1; <data> = the name of the file with the runoff data.        
!                       The format of such files resembles other
!                       forcing files. First column is the time in
!                       seconds or iso 8601 (yyyy-mm-ddThh:mm:ss)        
!                       second, third and fourth columns are the flux
!                       in cubic meters per second, the salinity and        
!                       temperature in celsius. the get_and_apply
!                       forcing routine will interpolate between these
!                       values in time.
!        
!   example:
!      river "Storelva"      8      2       2       2.0     0        10.0    10.0     0.0
!
        call add_river(splitline)
        river_set=.true.

    case('river_cell')
! river_cell <name> <iriver> <jriver>        
!   In some cases the position given for the river makes the river end
!   up in the wrong cell. this keyword set the cell position of a
!   river manually. the river name must exist, all other river properties are kept.
!
!   <name>:   The preexisting river name.        
!   <iriver>: The index in x-direction where the river flows into a sea cell
!   <jriver>: The index in y-direction where the river flows into a sea cell
        call override_river_position(splitline)

    case('river_geometry')
! river_geometry <name> <iriver> <jriver> <kriver1> <kriver2> <coderiver>
!   In some cases it may be desired to take full control over which cells to use to let a river out.
!        
!   this keyword keeps the flux,t,s,rho and forcing type of a river, but
!   changes IRIVER,JRIVER,KRIVER1,KRIVER2 and CODERIVER
!
!   <name>:      The preexisting river name.
!   <iriver>:    The index in x-direction where the river flows into a cell
!   <jriver>:    The index in y-direction where the river flows into a cell       
!   <kriver1>:   The first index in k-direction where the river flows into a cell
!   <kriver2>:   The last index in k-direction where the river flows into a cell
!   <coderiver>: code that says which edge of the cell the river enter: 1=u(i,j), 2=v(i,j), 3=u(i+1,j), 4=v(i,j+1)
        call override_river_geometry(splitline)
        
    case('river_conc')
! river_conc <string with river name> <integer value for conc field> <real value for concentration>
!   If we advect passive trace fields, let the river with the given name have a specified concentration.
!   If the river forcing is not from file, use the given value, otherwise read value from column
!   1-5 AFTER salinity and temperature in the river data file.
        call set_river_conc(splitline)
        river_conc_set=.true.

    case('river_submerge')
! river_submerge <string with river name> <z depth of top of river outlet>
!   by default the river outlet top is at the ocean surface. use this keyword to dive a river.
!   the lower outlet depth is already given via RIVDEP with the RIVER keyword.
!   example; dive river called "Arna" to -20m depth:
!      river_submerge Arna -20
        call submerge_river(splitline)


    case('line_source')
! line_source <name> <concX> <x> <y> <n> <z level file> <forcing file>
!
!   Used to set up a concentration field source, which is a vertical
!   line.  In boxes intersected by the line, the specified values
!   along the line will be interpolated onto the desired concentration
!   field.
!
!   <name>:            Name
!   <concX>:           integer for concentration id (1 = conc1, 2=conc2 etc)        
!   <x>:               The x position of source (meters). 
!   <y>:               The y position of source (meters).        
!   <z level file>:    File containing a 1d array with z levels of the
!                      data. the file name is significant, as the
!                      dimension of the field is assumed to be just
!                      before the period.  the file suffix dictates
!                      the expected data format.
!                      ".sr4"=stream real kind 4. ".ur4", unformatted fortran real 4
!                      anything else ascii.        
!   <forcing file>:    File with time (format as for rivers,wells), and
!                      file name for 1d source. suffix for these files
!                      as for the z levels.        
!   Example:
!      line_source "merd2"   1      5000,6000    z-levels-12.sr4  merd2-2012.list
!
        if(.not. bom_allocated) call message("arrays not allocated, initialize line_source after im,jm,kb are set",2)        
        call add_line_source(splitline)

        
    case('well')
! well <name> <x> <y> <format> <data>
!
!   Used to set up the bottom cell water influx.
!
!   <name>:    Well name
!   <x>:       The x position of well (meters). 
!   <y>:       The y position of well (meters). 
!   <format>:  0
!                <data> = <fwell> <twell> <swell>
!                where: <fwell>: Water flux (m^3s^(-1))
!                       <twell>: Temperature (C)
!                       <swell>: Salinity (p.s.u.)
!              1
!                the next column is the file name with the data
!                   <data>=<file name>        
!                where the file is in the simimar format as rivers,
!                where the first column is the time, either as seconds
!                since beginning of simulation, or in ISO9601.        
!                the three next columns are Flux, S and T. If more
!                columns are found, they will be assumed to be
!                concentrations.
!        
!   Example:
!      well "Arnavaaen"   9      9      0      1.0   10.   0.0
!
        call add_well(splitline)
        well_set=.true.

    case('well_conc')
! well_conc <string with well name> <integer for conc field> <real value for concentration>
!   If we advect passive tracer fields, let the well with the given name have a specified concentration.
!   If the well values for S and T are not read from file, use the given value, otherwise read value from column
!   1-5 AFTER S and T in the well data file.
        call set_well_conc(splitline)
        well_conc_set=.true.
        
    case('wind_forcing')
! wind_forcing <logical value>
!    keyword that switches wind forcing on or off. the default is off
        call set_var(wind_forcing, splitline)
        wind_forcing_set = .true.

    case('wind')
! wind <code> <data>
!
!   Used to set up the wind forcing
!
!   <code>: 1, Uniform wind.
!
!             <data>=<amplitude> <period> <initial_direction> <rotation_speed> <ramp_up_time>
!
!             where: <amplitude>:         Maximum magnitude of the wind (m/s)
!                    <period>:            If larger than 0, the wind strength varies sinusoidally
!                                         with this period in hours
!                    <initial_direction>: The counter clockwise angle w.r. the x-axis (degrees)
!                    <rotation_speed>:    If not 0, the wind direction changes by this number
!                                         of degrees per hour
!                    <ramp_up_time>:      If larger then 0, the wind strength will ramp up linearly
!                                         towards its max in this time in hours.
!
!           2, Ideal, passing low pressure system.
!
!             <data>=<strength> <radius> <startx> <starty> <center_speed> <center_direction> <ramp_up_time>
!
!             where: <strength>:         Strength of pressure (Pa)
!                    <radius>:           The radius of the given isobar (m)
!                    <startx>:           The x-coordinate of the low pressure center (m)
!                    <starty>:           The y-coordinate of the low pressure center (m)
!                    <center_speed>:     The initial speed of the low pressure center (ms^(-1))
!                    <center_direction>: The direction the low pressure is traveling. Given by the
!                                        angle taken counter-clockwise from the x-axis (degrees)
!                    <ramp_up_time>:     If larger than 0, the wind strength will ramp up linearly
!                                        towards its max in this time in hours (hr)
!
!          3, Read fields or uniform wind strength, direction from file. <data> is the file name
!          4, Read fields from NCEP type netcdf files, <data> is the header file with columns of
!               <iso time> <netcdf file with uwnd> <netcdf file vwnd> <netcdf file pressure> <field index in netcdf file>
!               ..         ..                      ..                      ..
!             the field index is the index in each netcdf file's time dimension that correspond to the iso time
!             in the first column. this method is not fully tested.
!          5. wind 5 <header file> <lon dimension> <lat dimension> <1D lon file> <1D lat file>
!             Read fields that are gridded on lon lat grids, and
!             stored in unformatted binary stream files, with one file
!             for each date.        
!             the header file is formed as follows
!               <iso time> <file with uwnd> <file vwnd> <file pressure>
!             the method uses a fast bilinear interpolation.
!          6. wind 6 <header file> <lon dimension> <lat dimension> <2D lon file> <2D lat file>        
!             Read fields that on general lon lat grids, stored in
!             unformatted binary stream files, with one file for each
!             date. the data is expected as speed(m/s), direction(north=0,east=90) and pressure (Pa).
!             the header file is formed as follows
!               <iso time> <file with speed> <file with direction> <file with pressure>        
!             this is the most general wind format for now, but because of the
!             generic grid, the interpolation (linear via triangulation) is slower and slightly
!             less accurate than wind 5.
!          7. wind 7 <header file>
!             read global fields for wind speed in x and y direction, gridded on the model grid.
!             the header file is formed as follows
!               <iso time> <file with windspeed x dir> <file with windspeed y direction>                     
        call initialize_wind(splitline)
        wind_set=.true.


    case('atmpress')
! atmpress <logical value>
!   if true, include the effect of atmospheric pressure
        call set_var(atmpress, splitline)
        atmpress_set=.true.

    case('cd')
! cd <real value>
!   Minimum value of the bottom drag coefficient cbc if cbc is not explicitly initialized.
        call set_var(cd, splitline)
        cd_set=.true.

    case('z0')
! z0 <real value>
!   Bottom roughness
        call set_var(z0, splitline)
        z0_set=.true.

    case('pmean')
! pmean <real value>
!   Mean atmospheric pressure (Pa)
        call set_var(pmean, splitline)
        pmean_set=.true.

    case('roair')
! roair <real value>
!   Density of air (kgm^(-3))
        call set_var(roair, splitline)
        roair_set=.true.

    case('grav')
! grav <real value>
!   Acceleration of gravity (ms^(-2))
        call set_var(grav, splitline)
        grav_set=.true.

    case('rho0')
! rho0 <real value>
!   Reference density (kgm^(-3))
        call set_var(rho0, splitline)
        rho0_set=.true.

    case('rhoback')
! rhoback <integer code> <data for initialization of 3d array> the
!   Background stratification to be used to compute optimal sigma
!   levels for the new perfectly balanced method for the computation
!   of internal pressure gradients. Normally initialized from a profile 
!   so that RHOBACK = F(Z)
        call allocate_balanced()
        call set_var(rhoback, splitline)
        rhoback_set = .true.

    case('rhoback_from_initialdensity')
! rhobackfrom_initialdensity <logical>
!   set RHOBACK to the initial density        
        call allocate_balanced()
        rhobackfromrho = .true.
        
    case('ah')
! ah <integer code> <data for initialization of 3d array>
!   Horizontal diffusivity (m^2s^(-1)), 3D array
        call set_var(ah, splitline)
        ah_set=.true.

    case('am')
! am <integer code> <data for initialization of 3d array>
!   Horizontal viscosity (m^2s^(-1)), 3D array
        call set_var(am, splitline)
        am_set=.true.

    case('am2d')
! am2d <integer code> <data for initialization of 2d array>
!   The vertical average of the horizontal viscosity, am (m^2s^(-1)), 2D array
        call set_var(am2d, splitline)
        am2d_set=.true.

    case('ammin')
! ammin <real value>
!   Minimum value of am, by default 0.0
        if(sponge_set) call message("ammin should be set before the add_sponge keyword",2)
        call set_var(ammin, splitline)
        
    case('am2dmin')
! am2dmin <real value>
!   Minimum value of am2d, by default 0.0
        if(sponge_set) call message("am2dmin should be set before the add_sponge keyword",2)
        call set_var(am2dmin, splitline)
        
    case('eta')
! eta <integer code> <data for initialization of 2d array>
!   The surface elevation w.r. local static depth (m), 2D array
        call set_var(eta, splitline)
        eta_set=.true.

    case('kh')
! kh <integer code> <data for initialization of 3d array>
!   Vertical diffusivity (m^2s^(-1)), 3D array
        call set_var(kh, splitline)
        kh_set=.true.

    case('km')
! km <integer code> <data for initialization of 3d array>
!   Vertical viscosity (m^2s^(-1)), 3D array
        call set_var(km, splitline)
        km_set=.true.

    case('kq')
! kq <integer code> <data for initialization of 3d array>
!   Vertical viscosity used for mixing of q2 and q2l (m^2s^(-1)), 3D array
        call set_var(kq, splitline)
        kq_set=.true.

    case('p')
! p <integer code> <data for initialization of 3d array>
!   The pressure used to compute the contribution of non hydrostatic effects
        call set_var(p, splitline)
        p_set=.true.

    case('patm')
! patm <integer code> <data for initialization of 2d array>
!   Atmospheric pressure (kgs^(-2)m^(-1)), 2D array
        call set_var(patm, splitline)
        patm_set=.true.

    case('q2')
! q2 <integer code> <data for initialization of 3d array>
!   Turbulent kinetic energy (m^2s^(-2)), 3D array
        call set_var(q2, splitline)
        q2_set=.true.

    case('q2l')
! q2l <integer code> <data for initialization of 3d array>
!   Turbulent kinetic energy times length scale (m^3s^(-2)), 3D array
        call set_var(q2l, splitline)
        q2l_set=.true.

    case('advect_density')
! advect_density <logical value>
!   if false, advect salinity and temperature, use equation of state
!   to find rho. this is the default.  if true, advect rho, arrays for
!   salinity and temperature not allocated. must be specified before H.      
        if(h_set) call message("advect_density must be specified before H",2)
        call set_var(advect_density, splitline)
        advect_density_set=.true.
        if(advect_density)then
            if(debug>0) call message('setting advect_s and advect_t to false since advect_density=true was requested',0)
            advect_s = .false.
            advect_t = .false.
        end if

    case('advect_s')
! advect_s <logical value>
!   if true, advect salinity. only allowed to be set if advect_density is false.
        call set_var(advect_s, splitline)
        if(advect_density_set .and. advect_density.and.advect_s) &
             call message("advect_density already is true, advection of S and RHO not supported",2)
        advect_s_set=.true.

    case('advect_t')
! advect_t <logical value>
!   if true, advect temperature. only allowed to be set if advect_density is false.
        call set_var(advect_t, splitline)
        if(advect_density_set .and. advect_density.and.advect_t) &
             call message("advect_density already is true, advection of T and RHO not supported",2)
        advect_t_set=.true.

    case('rho')
! rho <integer code> <data for initializing 3d array>
!   Density (kgm^(-3)), 3D array.
        if(.not.advect_density) call message('advect_density is .false. (also by default) &
             &please set it to true before this point, or drop initializing rho',2)
        call set_var(rho, splitline)
        rho_set=.true.

        if(.not.restart)then
            rho = rho - 1000.0
            call message("rho initialized, subtracting 1000")
        end if
    case('rhosurf')
! rhosurf <integer code> <data for initializing 2d array>
!   Surface boundary condition for density (kgs^(-1)m^(-2)), 2D array
!     (KH/D)*d(RHO)/d(sigma) = RHOSURF
!
        if(.not.advect_density) call message('advect_density is .false. (also by default) &
             &please set it to true before this point, or drop initializing rhosurf',2)
        call set_var(rhosurf, splitline)
        rhosurf_set=.true.

    case('s')
! s <integer code> <data for initializing 3d array>
!   initial salinity (p.s.u.), 3D array
        if(advect_density) &
             call message('advect_density is .true., s should not be initialized',2)
        call set_var(s, splitline)
        s_set=.true.
        
    case('ssurf')
! ssurf <integer code> <data for initializing 2d array>
!   Surface boundary condition for salinity at the surface (ms^(-1)p.s.u.), 2D array
!     (KH/D)*d(S)/d(sigma) = SSURF
!
        if(advect_density) &
             call message('advect_density is .true., ssurf should not be initialized',2)
        call set_var(ssurf, splitline)
        ssurf_set=.true.

    case('t')
! t <integer code> <data for initializing 3d array>
!   Initial temperature (C), 3D array
        if(advect_density) &
             call message('advect_density is .true., t should not be initialized',2)
        call set_var(t, splitline)
        t_set=.true.

    case('tsurf')
! tsurf <integer code> <data for initializing 2d array>
!   Surface boundary condition for temperature at the surface (ms^(-1)C), 2D array
!     (KH/D)*d(T)/d(sigma) = TSURF
!
        if(advect_density) &
             call message('advect_density is .true., tsurf should not be initialized',2)
        call set_var(tsurf, splitline)
        tsurf_set=.true.

    case('relax_tsurf')
! relax_tsurf <time constant C1> <forcing file>
!   adds forcing that relaxes surface temperature towards climatology as in Oey&Chen 1992
!   TSURF is computed as  C1*(TCLIMATE - TMODEL), e.g.
!            c1 = 1.736E-5
!            wtsurf = -c1*(  - t(:,:,1))        
!   <forcing file> has two columns, first is the time as iso or        
!   seconds as usual, second is the 2d climatology field for T as a
!   stream binary file. (global dimensions)        
        call relaxsurftemp(splitline, 0)

    case('relax_tsurf_scalar')
! relax_tsurf_scalar <time constant C1> <forcing file with scalar values>
!   adds forcing that relaxes surface temperature towards climatology as in Oey&Chen 1992
!   TSURF is computed as  C1*(TCLIMATE - TMODEL), e.g.
!            c1 = 1.736E-5
!            wtsurf = -c1*(  - t(:,:,1))        
!   <forcing file> has two columns, first is the time as iso or        
!   seconds as usual, second is the scalar temperature to relax towards
        call relaxsurftemp(splitline, 1)
        
    case('u')
! u <integer code> <data for initializing 3d array>
!   Horizontal velocity component in the x-direction (ms^(-1)), 3D array
        call set_var(u, splitline)
        u_set=.true.

    case('v')
! v <integer code> <data for initializing 3d array>
!   Horizontal velocity in the y-direction (ms^(-1)), 3D array
        call set_var(v, splitline)
        v_set=.true.

    case('w')
! w <integer code> <data for initializing 3d array>
!   Sigma coordinate vertical velocity (ms^(-1)), 3D array
        call set_var(w, splitline)
        w_set=.true.

    case('binreg_west')
! binreg_west <zone> <numlevels> <x1> <x2> <y1> <y2>
!   On outflow through frs zones, a horizontally averaged transport
!   from a region insided each frs zone is enforced in the frs zone.
!   <numlevels> is the number of z levels for which horisontal averages
!   are computed, by default each layer is 1m thick.
!   <x1> <x2> <y1> <y2> defines the region for which horisontal averages in bin_levels
!   levels should be computed for the given edge (here west) and the
!   given boundary zone.
        call add_binreg('west', bin_west, splitline)

    case('binreg_east')
! binreg_east <zone> <numlevels> <x1> <x2> <y1> <y2>
!   see binreg_west
        call add_binreg('east', bin_east, splitline)

    case('binreg_south')
! binreg_south <zone> <numlevels> <x1> <x2> <y1> <y2>
!   see binreg_west
        call add_binreg('south', bin_south, splitline)

    case('binreg_north')
! binreg_north <zone> <numlevels> <x1> <x2> <y1> <y2>
!   see binreg_west
        call add_binreg('north', bin_north, splitline)

    case('bc_u_west')
! bc_<variable>_<edge> <zone> <type> <setting_method> [<data>]
!
!   The keywords starting with "bc_" does two things:
!     1) specify the type of boundary condition to apply to a variable in a certain zone
!     2) set the initial value of the boundary value
!
!   <variable> is one of u,v,s,t,ua,va,eta,s,t,rho,q2,q2l,conc1-conc9.
!   <edge> is one of east,west,north,south.
!
!   <zone>: we only set boundary values for open boundaries. each
!           piece of the boundary which is "open", is called a
!           zone. the open boundaries are numbered along their edge,
!           on each edge the number ing starts from 1, and counting in
!           the same direction as the coordinate system.
!
!           During initialization, BOM will report how many open
!           boundaries that are found in the provided bathymetry.
!
!   <type>: Type of boundary condition: 0=FRS
!                                       1=Neumann
!                                       2=Periodic
!                                       3=Dirichlet
!                                       4=FRS with average value from "binreg"
!        
!           If <type>=2 then the fourth column says which zone on the
!           other side to communicate with.  <setting_mothod>=zone on
!           opposite boundary
!
!   <setting_method>: each variable that has a boundary condition, has
!           a special array, "a boundary array", containing the
!           boundary value to be applied via the routine "bound2". the
!           <setting_method> is a code that determines how the
!           boundary array should be initialized. the name of the
!           boundary array in the code is the variable name followed
!           by "_bound", e.g. u_bound, eta_bound. the <setting_method>
!           is mostly similar to the codes to initialize normal
!           arrays, and valid values are as follows;
!
!    -1 None. Use default initialization, which means from the matching field.
!     0 Initialize array directly from a single scalar value in the fourth column.
!       <data>=Scalar
!     1 If <setting_method> is 1, <data> is a file name, where the
!       file is a forcing file with the quick forcing format. the
!       first column of the file is either the (model) time in seconds since
!       the beginning of the simulation, or the time given as a ISO8601 string, e.g.
!       2008-08-02T04:34:10 (The "T" in the string is significant for auto detection!)
!       The second column is a scalar value which is given to the field to be forced at the
!       specified time.
!     2 Initialize real variables point by point from a spreadsheet type file,
!       where arrays are stored in columns, and in column major order.
!       <data>=<file> <column>
!       <file>=the file name,
!       <column>=which column in the file to use
!     3 Horizontal extrapolation from vertical profile in the file in
!       the fourth column.  3d fields only. Read x z from columns
!       spec. in col 5 and 6, the horis. range to insert into in
!       column 7-10.  <data>=<file> <column with x> <column with z> <x1> <x2> <y1> <y2>
!     4 Horizontal interpolation between multiple vertical profiles in
!       the file in the fourth column real 3d fields only
!     5 Horizontal interpolation between x,y,z points in the file in
!       the fourth column.  real 2d fields only
!     6 Initialize from tanhprofile. Real 3D fields only.
!       f=0.5(fmax+fmin)-0.5(fmax-fmin)tanh((depth(m)-zsymm)/deltaz)
!       <data>=<fmax> <fmin> <deltaz> <zsymm>
!       Where: <fmax>=The maximum initial value of field
!              <fmin>=The minimum initial value of field
!              <deltaz>=Height of slope (m)
!              <zsymm>=Distance from surface to point of symmetry (m)
!     8 Initialize from exponential profile. Real 3D fields only.
!       f=fmax-(fmax-fmin)exp(depth(m)/deltaz)
!       <data>=<fmax> <fmin> <deltaz>
!       Where: <fmax>=The maximum initial value of field
!              <fmin>=The minimum initial value of field
!              <deltaz>=Decides the height of the slope (m)
!     9 Read global array directly from unformatted file (binary io)
!       <data>=the file name
!    10 Assign value to a horisontal array region, and from depth z1 to z2.
!       <keyword> <code> <value> <x1> <x2> <y1> <y2> <Z1> <Z2>
!       z1,z2 must be specified as negative depths, i.e. a correct spec is from z1=-10 to z2=-5, so that z2>z1.
!    11 Read global Z coordinate array from unformatted stream file (binary io), interpolate to sigma
!       <keyword>  11   <file> <numlevels> <level file>    
!    12 Ramp up forcing: initially during <ramp up time(s)>, the bc for the variable is interpolated between state 1 and state 2.
!         <keyword> <zone> 12 <ramp up time (s)> <type state 1> <data state 1> <type state 2> <data state 2>        
!    13 Time dependent boundary condition:
!       At given times, read one slice of data for a model edge from 2D binary file (1D files for 2D model arrays).
!         <keyword> <zone> 13 <header file> <offset> <resolution> <z level file> <dim1> <dim2>        
!       the <header file> has two columns: time data-file-name. the time can be model seconds or iso9601. the data files are binary
!       unformatted stream access files with one field of dimension (dim1,dim2)
!       <offset>       distance in meters from the first slice grid point to the model origo
!       <resolution>   distance in meters between slice grid points (should be constant)
!       <z level file> name of binary file with <dim2> z levels for each layer of the data. z levels are positive, and increasing downwards.
!       <dim1>         the number of slice grid points along the edge
!       <dim2>         the number of z levels
!    14 UA,VA only: set UA_BOUND and VA_BOUND in region to depth average of U_BOUND, V_BOUND respectively.
!    15 Read boundary zone from 3D binary file. The horisontal resolution is the same as the model, the coordinate in the vertical is 
!         <keyword> <zone> 15 <header file> <i offset> <j offset> <z level file> <dim1> <dim2> <dim3>
!       the header file has two columns:  time  data-file-name, where time is in seconds or iso9601
        zone = getzone()
        call set_bc(zone, bowest(zone)%u_type, bowest(zone)%u_set, 'west', u_bound, kb, num_zones_west, splitline)
    case('bc_v_west')
        zone = getzone()
        call set_bc(zone, bowest(zone)%v_type, bowest(zone)%v_set,'west',v_bound,kb,num_zones_west,splitline)
    case('bc_w_west')
        zone = getzone()
        call set_bc(zone, bowest(zone)%w_type, bowest(zone)%w_set,'west',w_bound,kb,num_zones_west,splitline)
    case('bc_s_west')
        zone = getzone()
        if(advect_density) call message('advect density=.true., setting bc for s makes no sense',2)
        call set_bc(zone, bowest(zone)%s_type,bowest(zone)%s_set,'west',s_bound,kb,num_zones_west,splitline)
    case('bc_t_west')
        zone = getzone()
        if(advect_density) call message('advect density=.true., setting bc for t makes no sense',2)
        call set_bc(zone, bowest(zone)%t_type,bowest(zone)%t_set,'west',t_bound,kb,num_zones_west,splitline)
    case('bc_rho_west')
        zone = getzone()
        if(.not.advect_density) call message('advect density=.false., setting bc for rho makes no sense',2)
        call set_bc(zone, bowest(zone)%rho_type,bowest(zone)%rho_set,'west',rho_bound,kb,num_zones_west,splitline)
    case('bc_ua_west')
        zone = getzone()
        call set_bc(zone, bowest(zone)%ua_type,bowest(zone)%ua_set,'west',ua_bound,1,num_zones_west,splitline)
    case('bc_va_west')
        zone = getzone()
        call set_bc(zone, bowest(zone)%va_type,bowest(zone)%va_set,'west',va_bound,1,num_zones_west,splitline)
    case('bc_eta_west')
        zone = getzone()
        call set_bc(zone, bowest(zone)%eta_type,bowest(zone)%eta_set,'west',eta_bound,1,num_zones_west,splitline)
    case('bc_q2_west')
        zone = getzone()
        call set_bc(zone, bowest(zone)%q2_type,bowest(zone)%q2_set,'west',q2_bound,kb,num_zones_west,splitline)
    case('bc_q2l_west')
        zone = getzone()
        call set_bc(zone, bowest(zone)%q2l_type,bowest(zone)%q2l_set,'west',q2l_bound,kb,num_zones_west,splitline)

    case('bc_u_east')
        zone = getzone()
        call set_bc(zone, boeast(zone)%u_type,boeast(zone)%u_set,'east',u_bound,kb,num_zones_east,splitline)
    case('bc_v_east')
        zone = getzone()
        call set_bc(zone, boeast(zone)%v_type,boeast(zone)%v_set,'east',v_bound,kb,num_zones_east,splitline)
    case('bc_w_east')
        zone = getzone()
        call set_bc(zone, boeast(zone)%w_type,boeast(zone)%w_set,'east',w_bound,kb,num_zones_east,splitline)
    case('bc_s_east')
        zone = getzone()
        if(advect_density) call message('advect density=.true., setting bc for s makes no sense',2)
        call set_bc(zone, boeast(zone)%s_type,boeast(zone)%s_set,'east',s_bound,kb,num_zones_east,splitline)
    case('bc_t_east')
        zone = getzone()
        if(advect_density) call message('advect density=.true., setting bc for t makes no sense',2)
        call set_bc(zone, boeast(zone)%t_type,boeast(zone)%t_set,'east',t_bound,kb,num_zones_east,splitline)
    case('bc_rho_east')
        zone = getzone()
        if(.not.advect_density) call message('advect density=.false., setting bc for rho makes no sense',2)
        call set_bc(zone, boeast(zone)%rho_type,boeast(zone)%rho_set,'east',rho_bound,kb,num_zones_east,splitline)
    case('bc_ua_east')
        zone = getzone()
        call set_bc(zone, boeast(zone)%ua_type,boeast(zone)%ua_set,'east',ua_bound,1,num_zones_east,splitline)
    case('bc_va_east')
        zone = getzone()
        call set_bc(zone, boeast(zone)%va_type,boeast(zone)%va_set,'east',va_bound,1,num_zones_east,splitline)
    case('bc_eta_east')
        zone = getzone()
        call set_bc(zone, boeast(zone)%eta_type,boeast(zone)%eta_set,'east',eta_bound,1,num_zones_east,splitline)
    case('bc_q2_east')
        zone = getzone()
        call set_bc(zone, boeast(zone)%q2_type,boeast(zone)%q2_set,'east',q2_bound,kb,num_zones_east,splitline)
    case('bc_q2l_east')
        zone = getzone()
        call set_bc(zone, boeast(zone)%q2l_type,boeast(zone)%q2l_set,'east',q2l_bound,kb,num_zones_east,splitline)

    case('bc_u_north')
        zone = getzone()
        call set_bc(zone, bonorth(zone)%u_type,bonorth(zone)%u_set,'north',u_bound,kb,num_zones_north,splitline)
    case('bc_v_north')
        zone = getzone()
        call set_bc(zone, bonorth(zone)%v_type,bonorth(zone)%v_set,'north',v_bound,kb,num_zones_north,splitline)
    case('bc_w_north')
        zone = getzone()
        call set_bc(zone, bonorth(zone)%w_type,bonorth(zone)%w_set,'north',w_bound,kb,num_zones_north,splitline)
    case('bc_s_north')
        zone = getzone()
        if(advect_density) call message('advect density=.true., setting bc for s makes no sense',2)
        call set_bc(zone, bonorth(zone)%s_type,bonorth(zone)%s_set,'north',s_bound,kb,num_zones_north,splitline)
    case('bc_t_north')
        zone = getzone()
        if(advect_density) call message('advect density=.true., setting bc for t makes no sense',2)
        call set_bc(zone, bonorth(zone)%t_type,bonorth(zone)%t_set,'north',t_bound,kb,num_zones_north,splitline)
    case('bc_rho_north')
        zone = getzone()
        if(.not.advect_density) call message('advect density=.false., setting bc for rho makes no sense',2)
        call set_bc(zone, bonorth(zone)%rho_type,bonorth(zone)%rho_set,'north',rho_bound,kb,num_zones_north,splitline)
    case('bc_ua_north')
        zone = getzone()
        call set_bc(zone, bonorth(zone)%ua_type,bonorth(zone)%ua_set,'north',ua_bound,1,num_zones_north,splitline)
    case('bc_va_north')
        zone = getzone()
        call set_bc(zone, bonorth(zone)%va_type,bonorth(zone)%va_set,'north',va_bound,1,num_zones_north,splitline)
    case('bc_eta_north')
        zone = getzone()
        call set_bc(zone, bonorth(zone)%eta_type,bonorth(zone)%eta_set,'north',eta_bound,1,num_zones_north,splitline)
    case('bc_q2_north')
        zone = getzone()
        call set_bc(zone, bonorth(zone)%q2_type,bonorth(zone)%q2_set,'north',q2_bound,kb,num_zones_north,splitline)
    case('bc_q2l_north')
        zone = getzone()
        call set_bc(zone, bonorth(zone)%q2l_type,bonorth(zone)%q2l_set,'north',q2l_bound,kb,num_zones_north,splitline)

    case('bc_u_south')
        zone = getzone()
        call set_bc(zone, bosouth(zone)%u_type,bosouth(zone)%u_set,'south',u_bound,kb,num_zones_south,splitline)
    case('bc_v_south')
        zone = getzone()
        call set_bc(zone, bosouth(zone)%v_type,bosouth(zone)%v_set,'south',v_bound,kb,num_zones_south,splitline)
    case('bc_w_south')
        zone = getzone()
        call set_bc(zone, bosouth(zone)%w_type,bosouth(zone)%w_set,'south',w_bound,kb,num_zones_south,splitline)
    case('bc_s_south')
        zone = getzone()
        if(advect_density) call message('advect density=.true., setting bc for s makes no sense',2)
        call set_bc(zone, bosouth(zone)%s_type,bosouth(zone)%s_set,'south',s_bound,kb,num_zones_south,splitline)
    case('bc_t_south')
        zone = getzone()
        if(advect_density) call message('advect density=.true., setting bc for t makes no sense',2)
        call set_bc(zone, bosouth(zone)%t_type,bosouth(zone)%t_set,'south',t_bound,kb,num_zones_south,splitline)
    case('bc_rho_south')
        zone = getzone()
        if(.not.advect_density) call message('advect density=.false., setting bc for rho makes no sense',2)
        call set_bc(zone, bosouth(zone)%rho_type,bosouth(zone)%rho_set,'south',rho_bound,kb,num_zones_south,splitline)
    case('bc_ua_south')
        zone = getzone()
        call set_bc(zone, bosouth(zone)%ua_type,bosouth(zone)%ua_set,'south',ua_bound,1,num_zones_south,splitline)
    case('bc_va_south')
        zone = getzone()
        call set_bc(zone, bosouth(zone)%va_type,bosouth(zone)%va_set,'south',va_bound,1,num_zones_south,splitline)
    case('bc_eta_south')
        zone = getzone()
        call set_bc(zone, bosouth(zone)%eta_type,bosouth(zone)%eta_set,'south',eta_bound,1,num_zones_south,splitline)
    case('bc_q2_south')
        zone = getzone()
        call set_bc(zone, bosouth(zone)%q2_type,bosouth(zone)%q2_set,'south',q2_bound,kb,num_zones_south,splitline)
    case('bc_q2l_south')
        zone = getzone()
        call set_bc(zone, bosouth(zone)%q2l_type,bosouth(zone)%q2l_set,'south',q2l_bound,kb,num_zones_south,splitline)

    case('bc_conc1_east')
        zone = getzone()
        call allocate_concentration(1)
        call set_bc(zone,boeast(zone)%conc_type(1),boeast(zone)%conc_set(1),'east',co(1)%conc_bound,kb,num_zones_east,splitline)
    case('bc_conc2_east')
        zone = getzone()
        call allocate_concentration(2)
        call set_bc(zone,boeast(zone)%conc_type(2),boeast(zone)%conc_set(2),'east',co(2)%conc_bound,kb,num_zones_east,splitline)
    case('bc_conc3_east')
        zone = getzone()
        call allocate_concentration(3)
        call set_bc(zone,boeast(zone)%conc_type(3),boeast(zone)%conc_set(3),'east',co(3)%conc_bound,kb,num_zones_east,splitline)
    case('bc_conc4_east')
        zone = getzone()
        call allocate_concentration(4)
        call set_bc(zone,boeast(zone)%conc_type(4),boeast(zone)%conc_set(4),'east',co(4)%conc_bound,kb,num_zones_east,splitline)
    case('bc_conc5_east')
        zone = getzone()
        call allocate_concentration(5)
        call set_bc(zone,boeast(zone)%conc_type(5),boeast(zone)%conc_set(5),'east',co(5)%conc_bound,kb,num_zones_east,splitline)
    case('bc_conc6_east')
        zone = getzone()
        call allocate_concentration(6)
        call set_bc(zone,boeast(zone)%conc_type(6),boeast(zone)%conc_set(6),'east',co(6)%conc_bound,kb,num_zones_east,splitline)
    case('bc_conc7_east')
        zone = getzone()
        call allocate_concentration(7)
        call set_bc(zone,boeast(zone)%conc_type(7),boeast(zone)%conc_set(7),'east',co(7)%conc_bound,kb,num_zones_east,splitline)
    case('bc_conc8_east')
        zone = getzone()
        call allocate_concentration(8)
        call set_bc(zone,boeast(zone)%conc_type(8),boeast(zone)%conc_set(8),'east',co(8)%conc_bound,kb,num_zones_east,splitline)
    case('bc_conc9_east')
        zone = getzone()
        call allocate_concentration(9)
        call set_bc(zone,boeast(zone)%conc_type(9),boeast(zone)%conc_set(9),'east',co(9)%conc_bound,kb,num_zones_east,splitline)
        
    case('bc_conc1_west')
        zone = getzone()
        call allocate_concentration(1)
        call set_bc(zone,bowest(zone)%conc_type(1),bowest(zone)%conc_set(1),'west',co(1)%conc_bound,kb,num_zones_west,splitline)
    case('bc_conc2_west')
        zone = getzone()
        call allocate_concentration(2)
        call set_bc(zone,bowest(zone)%conc_type(2),bowest(zone)%conc_set(2),'west',co(2)%conc_bound,kb,num_zones_west,splitline)
    case('bc_conc3_west')
        zone = getzone()
        call allocate_concentration(3)
        call set_bc(zone,bowest(zone)%conc_type(3),bowest(zone)%conc_set(3),'west',co(3)%conc_bound,kb,num_zones_west,splitline)
    case('bc_conc4_west')
        zone = getzone()
        call allocate_concentration(4)
        call set_bc(zone,bowest(zone)%conc_type(4),bowest(zone)%conc_set(4),'west',co(4)%conc_bound,kb,num_zones_west,splitline)
    case('bc_conc5_west')
        zone = getzone()
        call allocate_concentration(5)
        call set_bc(zone,bowest(zone)%conc_type(5),bowest(zone)%conc_set(5),'west',co(5)%conc_bound,kb,num_zones_west,splitline)
    case('bc_conc6_west')
        zone = getzone()
        call allocate_concentration(6)
        call set_bc(zone,bowest(zone)%conc_type(6),bowest(zone)%conc_set(6),'west',co(6)%conc_bound,kb,num_zones_west,splitline)
    case('bc_conc7_west')
        zone = getzone()
        call allocate_concentration(7)
        call set_bc(zone,bowest(zone)%conc_type(7),bowest(zone)%conc_set(7),'west',co(7)%conc_bound,kb,num_zones_west,splitline)
    case('bc_conc8_west')
        zone = getzone()
        call allocate_concentration(8)
        call set_bc(zone,bowest(zone)%conc_type(8),bowest(zone)%conc_set(8),'west',co(8)%conc_bound,kb,num_zones_west,splitline)
    case('bc_conc9_west')
        zone = getzone()
        call allocate_concentration(9)
        call set_bc(zone,bowest(zone)%conc_type(9),bowest(zone)%conc_set(9),'west',co(9)%conc_bound,kb,num_zones_west,splitline)

    case('bc_conc1_north')
        zone = getzone()
        call allocate_concentration(1)
        call set_bc(zone,bonorth(zone)%conc_type(1),bonorth(zone)%conc_set(1),'north',co(1)%conc_bound,kb,num_zones_north,splitline)
    case('bc_conc2_north')
        zone = getzone()
        call allocate_concentration(2)
        call set_bc(zone,bonorth(zone)%conc_type(2),bonorth(zone)%conc_set(2),'north',co(2)%conc_bound,kb,num_zones_north,splitline)
    case('bc_conc3_north')
        zone = getzone()
        call allocate_concentration(3)
        call set_bc(zone,bonorth(zone)%conc_type(3),bonorth(zone)%conc_set(3),'north',co(3)%conc_bound,kb,num_zones_north,splitline)
    case('bc_conc4_north')
        zone = getzone()
        call allocate_concentration(4)
        call set_bc(zone,bonorth(zone)%conc_type(4),bonorth(zone)%conc_set(4),'north',co(4)%conc_bound,kb,num_zones_north,splitline)
    case('bc_conc5_north')
        zone = getzone()
        call allocate_concentration(5)
        call set_bc(zone,bonorth(zone)%conc_type(5),bonorth(zone)%conc_set(5),'north',co(5)%conc_bound,kb,num_zones_north,splitline)
    case('bc_conc6_north')
        zone = getzone()
        call allocate_concentration(6)
        call set_bc(zone,bonorth(zone)%conc_type(6),bonorth(zone)%conc_set(6),'north',co(6)%conc_bound,kb,num_zones_north,splitline)
    case('bc_conc7_north')
        zone = getzone()
        call allocate_concentration(7)
        call set_bc(zone,bonorth(zone)%conc_type(7),bonorth(zone)%conc_set(7),'north',co(7)%conc_bound,kb,num_zones_north,splitline)
    case('bc_conc8_north')
        zone = getzone()
        call allocate_concentration(8)
        call set_bc(zone,bonorth(zone)%conc_type(8),bonorth(zone)%conc_set(8),'north',co(8)%conc_bound,kb,num_zones_north,splitline)
    case('bc_conc9_north')
        zone = getzone()
        call allocate_concentration(9)
        call set_bc(zone,bonorth(zone)%conc_type(9),bonorth(zone)%conc_set(9),'north',co(9)%conc_bound,kb,num_zones_north,splitline)

    case('bc_conc1_south')
        zone = getzone()
        call allocate_concentration(1)
        call set_bc(zone,bosouth(zone)%conc_type(1),bosouth(zone)%conc_set(1),'south',co(1)%conc_bound,kb,num_zones_south,splitline)
    case('bc_conc2_south')
        zone = getzone()
        call allocate_concentration(2)
        call set_bc(zone,bosouth(zone)%conc_type(2),bosouth(zone)%conc_set(2),'south',co(2)%conc_bound,kb,num_zones_south,splitline)
    case('bc_conc3_south')
        zone = getzone()
        call allocate_concentration(3)
        call set_bc(zone,bosouth(zone)%conc_type(3),bosouth(zone)%conc_set(3),'south',co(3)%conc_bound,kb,num_zones_south,splitline)
    case('bc_conc4_south')
        zone = getzone()
        call allocate_concentration(4)
        call set_bc(zone,bosouth(zone)%conc_type(4),bosouth(zone)%conc_set(4),'south',co(4)%conc_bound,kb,num_zones_south,splitline)
    case('bc_conc5_south')
        zone = getzone()
        call allocate_concentration(5)
        call set_bc(zone,bosouth(zone)%conc_type(5),bosouth(zone)%conc_set(5),'south',co(5)%conc_bound,kb,num_zones_north,splitline)
    case('bc_conc6_south')
        zone = getzone()
        call allocate_concentration(6)
        call set_bc(zone,bosouth(zone)%conc_type(6),bosouth(zone)%conc_set(6),'south',co(6)%conc_bound,kb,num_zones_south,splitline)
    case('bc_conc7_south')
        zone = getzone()
        call allocate_concentration(7)
        call set_bc(zone,bosouth(zone)%conc_type(7),bosouth(zone)%conc_set(7),'south',co(7)%conc_bound,kb,num_zones_south,splitline)
    case('bc_conc8_south')
        zone = getzone()
        call allocate_concentration(8)
        call set_bc(zone,bosouth(zone)%conc_type(8),bosouth(zone)%conc_set(8),'south',co(8)%conc_bound,kb,num_zones_south,splitline)
    case('bc_conc9_south')
        zone = getzone()
        call allocate_concentration(9)
        call set_bc(zone,bosouth(zone)%conc_type(9),bosouth(zone)%conc_set(9),'south',co(9)%conc_bound,kb,num_zones_north,splitline)

        
    case('tide')
! tide <logical value>
!   if true, include tidal forcing

        call set_var(tide, splitline)
        tide_set=.true.

    case('tide_fac')
! tide_fac <logical value>
!   if true, include astronomical tidal factors from adcirc_tide_fac       
        call set_var(tide_fac, splitline)

        if(tide_fac)then
            if(.not.ndate_set .or. .not.timezone_set)  &
                 call message("ndate and timezone must both be set before tide_fac",2)
            if(any((/tide_west_set, tide_east_set, tide_south_set, tide_north_set/))) &
                 call message("tide_fac must be set before specification of tidal constituents",2)
        end if

    case('tide_west')
! tide_west <zone> <type> <data>
!   set up the tidal forcing for a zone on the western edge
!
!   <zone>: on each edge, open boundary regions are assigned an integer id
!           number from 1 and upwards. we count in the same direction
!           as the axes.
!   <type>: an integer that specifies the forcing format;
!           0: add specification of a tidal constituent directly, where
!                <data> = <frequency> <etaamp> <etapha> <uaamp> <uapha> <vaamp> <vapha>
!              where
!                <frequency> is the constituents frequency in degrees per hour
!                <etaamp> is the elevation amplitude in meters,
!                <etapha> is the elevation phase in degrees,
!                <uaamp>,<vaamp> are the amplitudes for ua and va in meters per second,
!                <uapha>,<vapha> are the phases for ua and va in degrees!
!           1: add specification of a tidal constituent directly, just
!              as for <type>=0, but the unit for the constituent
!              <frequency> is instead the period in hours.
        call set_tide('west',num_zones_west,splitline)
        tide_west_set=.true.

    case('tide_east')
        call set_tide('east',num_zones_east,splitline)
        tide_east_set=.true.

    case('tide_north')
        call set_tide('north',num_zones_north,splitline)
        tide_north_set=.true.

    case('tide_south')
        call set_tide('south',num_zones_south,splitline)
        tide_south_set=.true.

    case('tide_from_chart')
! tide_from_chart <period> <file name>        
!   on all open boundaries, specify the period in hours for a
!   constituent, and a file name containing full 2d chart of amplitude
!   and phase for elevation and velocity; ETAAMP ETAPHASE UAAMP UAPHASE VAAMP VAPHASE.
!   The data is stored in columns of 4 byte reals (std Fortran 2d column major order)
        call set_bintide(splitline)        
        
    case('bomplot_workdir')
! bomplot_workdir <string>
!   The directory where the output files from bomplot are saved.  If
!   this keyword is provided, it will override all previously declared
!   output directories for bomplot
        call set_var(bomplot_workdir, splitline)
        call ensureendslash(bomplot_workdir)
        hsect_dir = bomplot_workdir
        vsect_dir = bomplot_workdir
        vprof_dir = bomplot_workdir
        slice_dir = bomplot_workdir
        if(debug > 1) then
            write(*,'(2x,a,a)') "hsect_dir is now ",trim(hsect_dir)
            write(*,'(2x,a,a)') "vsect_dir is now ",trim(vsect_dir)
            write(*,'(2x,a,a)') "vprof_dir is now ",trim(vprof_dir)
            write(*,'(2x,a,a)') "slice_dir is now ",trim(slice_dir)
        end if

    case('hsect_dir')
! hsect_dir <string>
!   bomplot will dump hsect files to this directory
        call set_var(hsect_dir, splitline)
        
    case('vsect_dir')
! vsect_dir <string>
!   bomplot will dump vsect files to this directory
        call set_var(vsect_dir, splitline)
        
    case('vprof_dir')
! vprof_dir <string>
!   bomplot will dump (ts)vprof files to this directory
        call set_var(vprof_dir, splitline)
        
    case('slice_dir')
! slice_dir <string>
!   bomplot will dump slice files to this directory
        call set_var(slice_dir, splitline)

    case('particlets_dir')
! particlets_dir <string>
!   the particle routines will dump files to this directory
        call set_var(particlets_dir, splitline)
        
    case('particledist_dir')
! particledist_dir <string>
!   the particle routines will dump files to this directory
        call set_var(particledist_dir, splitline)

        
    case('bomplot_depth_average')
! bomplot_depth_avlerage <label> <3D field>  <x1> <x2> <y1> <y2>  <plotdt> <start> <end>
!
!   dump the depth average of a 3D field, for a given range, to file at the given interval
!
!   <label>:               First part of data file names
!   <x1> <x2> <y1> <y2>    The horisontal coordinate range for which to dump the field, if 0,0,0,0 dump all.
!   <plotdt>:              Minimum time between data dumps(s)
!   <start>:               Start time for data dumps(s)
!   <end>:                 End time for data dumps(s)
!        
        call add_bomplot(name, splitline)
    case('bomplot_hsect_depth')
! bomplot_hsect_depth <label> <depth> <plotdt> <start> <end>
!
!   Keyword used to set up dump of data from a horizontal section.
!
!   u(ms^(-1)), v(ms^(-1)), s(p.s.u.), t(C) are stored for a specified depth (m).
!
!   If depth<0, sample the section at |depth| meters above bottom, otherwise the depth is in
!   meters below static level.
!
!
!   <label>:    First part of data file names
!   <depth>     Meters below static 0 level(m), (or if < 0, meters above seabed)
!   <plotdt>:   Minimum time between data dumps(s)
!   <start>:    Start time for data dumps(s)
!   <end>:      End time for data dumps(s)
        call add_bomplot(name, splitline)


    case('bomplot_hsect_level')
! bomplot_hsect_level <label> <level> <plotdt> <start> <end>
!
!   Keyword used to set up dumps of data from horizontal sigma layer.
!   the fields u(ms^(-1)), v(ms^(-1)), s(p.s.u.), t(C) for a specified
!   sigma layer are stored.
!
!   <label>:    First part of data file names
!   <level>:    Index of sigma level/layer
!   <plotdt>:   Minimum time between data dumps(s)
!   <start>:    Start time for data dumps(s)
!   <end>:      End time for data dumps(s)
        call add_bomplot(name, splitline)


    case('bomplot_hsect_depth_range')
! bomplot_hsect_depth_range <label> <depth> <x1> <x2> <y1> <y2> <plotdt> <start> <end>
!
!   Same as bomplot_hsect_depth, but includes a range specification,
!   so that data is only dumped in the coordinate range [x1,x2] and [y1,y2] (meters)
!
        call add_bomplot(name, splitline)

    case('bomplot_genhsect')
! bomplot_genhsect <label> <what> <how> <depth> <x1> <x2> <y1> <y2> <plotdt> <start> <end>
!
!    Dump individial field to file        
!      <label> first part of file names dumped
!      <what>  which field to dump, one of Q,Q2L,CONC1..CONC9,AM,AH,KM,KH,RHO,WR,P
!      <how>   how depth is specified, one of "above_bottom" "below_surf" "level"
!      <depth> the actual desired depth, either meters above bottom or below surface, or a K index
!        
!      <x1> <x2> <y1> <y2>       the horisontal coordinate range for which to dump the field, if 0,0,0,0 dump all.
!      <plotdt> <start> <end>    the time between dumps and the time range for which to plot in seconds
!
        call add_bomplot(name, splitline)
        
    case('bomplot_vsect')
! bomplot_vsect <label> N <x1> <y1> ... <xN> <yN> <plotdt> <start> <end>
!
!   Keyword used to set up dumps from a vertical cross section in the 3d
!   grid. The section is defined as the data under the straight
!   lines connecting a sequence of horisontal x,y grid positions given
!   by the user. The variables stored are un (normal velocity) ,
!   ut(tangential velocity), wr(vertical velocity), s(salinity) and
!   t(temperature) or rho(density).
!
!   <label>:    First part of data file names
!   <N>:        Number of the points that define the vertical sections
!   <x1>:       x-coordinate for the first point (meters)
!   <y1>:       y-coordinate for the first point
!   ...
!   <xN>:       x-coordinate for the last point
!   <yN>:       y-coordinate for the last point
!   <plotdt>:   Minimum time between data dumps(s)
!   <start>:    Start time for data dumps(s)
!   <stop>:     End time for data dumps(s)
!
        call add_bomplot(name, splitline)


    case('bomplot_hsect2d')
! bomplot_hsect2d <label> <plotdt> <start> <end>
!
!   Keyword used to set up dumps of horisontal fields for eta(m), ua(ms^(-1)), va(ms^(-1)).
!   <label>:    Name of the new datafile
!   <plotdt>:   Minimum time between data dumps(s)
!   <start>:    Start time for data dumps(s)
!   <stop>:     End time for data dumps(s)
!
        call add_bomplot(name, splitline)


    case('bomplot_hsect2d_range')
! bomplot_hsect2d_range <label> <x1> <x2> <y1> <y2> <plotdt> <start> <end>
!   same as bomplot_hsect2d, but possible to also provide a range as for bomplot_hsect.
!
        call add_bomplot(name, splitline)


    case('bomplot_isosurf')
! bomplot_isosurf <label> <rhoisolevel> <x1> <x2> <y1> <y2> <plotdt> <start> <end>
!
!   Keyword used to set up dumps of the isosurface of rho(kgm^(-3)).
!
!   Given the 3D-array rho(kgm^(-3)), isosurf finds the isosurface
!   with value ISOVAL, as well as the velocity components U,V and WR
!   on the surface. The isolevel is found by searching each water
!   column from the top.
!   <label>:    First part of data file names
!   <x1>:       western x coordinate (meters) for region to dump
!   <x2>:       eastern x coordinate (meters) for region to dump
!   <y1>:       southern x coordinate (meters) for region to dump
!   <y2>:       northern x coordinate (meters) for region to dump
!   <plotdt>:   Minimum time between data dumps(s)
!   <start>:    Start time for data dumps(s)
!   <stop>:     End time for data dumps(s)
!
        call add_bomplot(name, splitline)


    case('bomplot_vprof_ts')
!
! bomplot_vprof_ts <label> <x> <y> <plotdt> <start> <end>
!
!   Keyword used to set up dumps of u(ms^(-1)),v(ms^(-1)),s(p.s.u.)
!   and t(C) under a point (x,y).  the profile is stored every time
!   called, and files are written with a sequence number in the name to
!   make it possible to "cat" them into a complete vertical profile
!   time series.
!   <label>:    Name of the new datafile
!   <x>:        the x-coordinate for the point over the vertical profile
!   <y>:        the y-coordinate for the point over the vertical profile
!   <plotdt>:   Interval in s between dumps        
!   <start>:    Starting time for data dumping(s)
!   <stop>:     Ending time for data dumping(s)
!
        call add_bomplot(name, splitline)


    case('bomplot_vprof_interval')
! bomplot_vprof_interval <integer value>
!
!   The number of vertical profiles from bomplot_vprof_ts per
!   file. Data is written in batches to avoid data loss on crashes.
!
        call set_var(bomplot_vprof_interval, splitline)


    case('bomplot_slice')
! bomplot_slice <label> <plotdt> <start> <stop> <what> <i1> <i2> <j1> <j2> [<k1> <k2>]
!
!   Writes an array region directly to file, if <what> is the 3d array F,
!   the region F(<i1>:<i2>,<j1>:<j2>,<k1>:<k2>) is written.
!
!   <label>:    Name of the new datafile
!   <plotdt>:   Minimum time between data dumping(s)
!   <start>:    Start time for data dumps(s)
!   <stop>:     End time for data dumps(s)
!   <what>:     The field to be plotted. u,v,wr,s,t,rho,ua,va,eta,km,kh,am,ah,conc1-conc9
!   <i1>:       Start index in x-direction
!   <i2>:       End index in x-direction
!   <j1>:       Start index in y-direction
!   <j2>:       End index in y-direction
!   <k1>:       If in 3D, start index in sigma-direction
!   <k2>:       If in 3D, end index in sigma-direction
        call add_bomplot(name, splitline)


    case('bomplot_binslice')
! bomplot_binslice <label> <plotdt> <start> <stop> <what> <i1> <i2> <j1> <j2> [<k1> <k2>]
!   identical to bomplot_slice, but dumps to Fortran unformatted binary files.
        call add_bomplot(name, splitline)


    case('bomplot_accumslice')
! bomplot_accumslice <label> <accum interval> <start> <stop> <what> <i1> <i2> <j1> <j2> [<k1> <k2>]
!   like bomplot_slice, dumps a range of an array to Fortran unformatted binary files, the difference
!   is that this routine dumps the field every time step in the specified interval. <accum interval>
!   is the number of fields(timesteps) in each file dumped.
        call add_bomplot(name, splitline)
        

    case('bomplot_dumping_method')
! bomplot_dumping_method <integer value>
!
!   Integer used to specify how the fields are stored to disk;
!
!	  0 = Use gather to store global fields on master.
!	  1 = Each processor stores its block,
!	  2 = (TODO) Each processor stores netcdf,
!	  3 = (TODO) Use gather to store global netcdf, etc.
!
        call set_var(bomplot_dumping_method, splitline)

    case('bomplot_gnuplot')
! bomplot_gnuplot <logical value>
!
!   when true, dump horisontal fields with a blank line between rows,
!   so that files can be plotted directly with the gnuplot splot
!   command. this variable is only active if bomplot_compress is
!   false. (the default)
        call set_var(bomplot_gnuplot, splitline)

        if(bomplot_compress) then
            call message("setting bomplot_compress to false",0)
            bomplot_compress = .false.
        end if

    case('bomplot_compress')
! bomplot_compress <logical value>
!   when true, dumps horisontal fields in ascii with
!   minimal precision and width in order to reduce file size.
        call set_var(bomplot_compress, splitline)
        if(bomplot_compress.and.bomplot_gnuplot) then
            call message("bomplot_gnuplot will be ignored with bomplot_compress=.true.",0)
        end if
        
    case('numseries')
! numseries <integer value>
!   the maximum number of timeseries. defaults to 500.
        call set_var(numseries, splitline)
        call allocate_timeseries()

    case('timeseries_dir')
! timeseries_dir <string>
!   timeseries will dump files to this directory
        call set_var(timeseries_dir, splitline)
        
    case('timeseries_unformatted')
! timeseries_unformatted <logical value>
!   store timeseries to unformatted (binary) files with "stream"
!   access format. as for the ascii time series format, these files
!   can be cat'ed together to form a complete series. the batch number
!   is stored in each file name.        
        call set_var(timeseries_unformatted, splitline)
        
    case('timeseries_fullprec')
! timeseries_fullprec <logical value>
!   when true (default), dump time series in ascii with full precision
        call set_var(timeseries_fullprec, splitline)
        
    case('timeseries_formatted')
! timeseries_formatted <logical value>
!   when true (default), dump time series in ascii with given format
!   string, which by default is "(f0.7)". use the
!   timeseries_format_string keyword to override the default string.
        call set_var(timeseries_fullprec, splitline)

    case('timeseries_format_string')
! timeseries_format_string <edit descriptor string>  <time series index>        
!    if timeseries_formatted is true, use this to override the default
!    edit descriptor (format) for the given timeseries.
        call set_timeseries_format(splitline)
        
    case('timeseries_interval')
! timeseries_interval <integer value>
!   The number of time steps stored in each data file. To avoid data loss,
!   timeseries are dumped to files in batches, and numbered in a
!   manner that makes it trivial to create a complete time series file
!   using e.g. the unix cat command. Default value 1000.
        call set_var(timeseries_interval, splitline)


    case('timeseries_quick2d')
! timeseries_quick2d <x> <y>
!   Shortcut-keyword used to dump the time series of
!   eta(kgm^(-3)),ua(ms^(-1)),va(ms^(-1)) in a point (x,y).
!
!   <x>:       x-coordinate of point
!   <y>:       y-coordinate of point
!
        call add_timeseries(splitline)


    case('timeseries_quick3d')
!
! timeseries_quick3d <x> <y> <z>
!
!   Shortcut used to dump the time series of
!   u(ms^(-1)),v(ms^(-1)),wr(ms^(-1)) and one of s(p.s.u.),t(C) or
!   rho(kgm^(-3)) in a point (x,y,z).
!
!   <x>:       x-coordinate of point
!   <y>:       y-coordinate of point
!   <z>:       z-coordinate of point(m). if z<0, the position is |z|m below
!              surface, otherwise it is |z|m above the sea floor.
!
        if(.not.z_set) call message("sigma layer distribution must be specified to do this",2)        
        call add_timeseries(splitline)


    case('timeseries')
!
! timeseries <what> <type=0> <x> <y> <z>
!     or
! timeseries <what> <type=1> <i> <j> <k>
!     or
! timeseries <what> <type=2> <longitude> <latitude> <z>
!     or
! timeseries transport <x1> <y1> <x2> <y2>
!
!   Dumps the time series of <what> (u,v,s,t,rho,...) in a point
!   (x,y,z) to file. also scalars ekin, mass,volume, eeta but then
!   the coordinates are of course ignored as they are ... scalar.
!   If used for calculation of transports (<what> = transport), 
!   start points (x1,y1) and end points (x2,y2) of a line segment 
!   should be specified, over which to calculate the transport.
!
!   <type>: Integer
!              0 = Give position with (x,y,z)-coordinates
!              1 = Give position with indices (i,j,k)
!              2 = Give horizontal position as longitude,latitude, depth as cartesian Z
!
!   <x>:       x-coordinate of point(grid coordinates)
!   <y>:       y-coordinate of point(grid coordinates)
!   <z>:       z-coordinate of point(m). if z<0, the position is |z|m below
!              surface, otherwise it is |z|m above the sea floor.
!
!   <i>:       Index in x-direction
!   <j>:       Index in y-direction
!   <k>:       Index in sigma-direction
!
        call check_allocation()        
        if(.not.z_set) call message("sigma layer distribution must be specified to do this",2)        
        call add_timeseries(splitline)

    case('timeseries_file')
!
! timeseries_file <file name>
!
!    read the time series keywords from a separate file. format is just like in the setup file.
        call timeseries_from_file(splitline)

        
        
    case('dump_state_pgplot')
! dump_state_pgplot <integer value> <label>
!    Plot all pgplot's defined in the setup file each time dump_state is called (often...)
!      if 0, never call pgplot from dump_state
!         1, call from 3d routines only
!         2, call also from 2d.
!         3, call only for locations equal to <label>        
!    Used together with e.g. the slice plotting this makes it possible
!    to see what goes on in the model during a time step.
        call set_var(dump_state_pgplot, splitline)
        if(dump_state_pgplot == 3) then
            pgplot_numdumploc = pgplot_numdumploc + 1
            if(pgplot_numdumploc > 20) call message("only 20 dump state locations allowed. tweak pgplot.f90!",2)
            pgplot_dump_location(pgplot_numdumploc) = trim(splitline(3))
            if(debug>0.and.master) write(*,*) " from which dump_state? ",trim(pgplot_dump_location(pgplot_numdumploc))
        end if
        
    case('pgplot_cache_results')
! pgplot_cache_results <integer value>        
!   if > 0, store pgplot timeseries to disk, and view the previous
!   trend with the new one when found.
        call set_var(pgplot_cache_results, splitline)

    case('pgplot_cache_interval')
! pgplot_cache_interval <integer value>
!   pgplot timeseries are cached to disk every interval'th time step.        
        call set_var(pgplot_cache_interval, splitline)
        
    case('pgplot_wait_after_plot')
! pgplot_wait_after_plot <logical value>
!    if true, pgplot will wait for key press after each plot. by default false.
        call set_var(pgplot_wait_after_plot, splitline)
        
    case('pgplot_device')
! pgplot_device <string>
!
!   Keyword that tells pgplot to which device (file, window,...)to plot while running.
!
!   Example: pgplot_device "hei#.gif/gif"
!
!          - The filename is hei#.gif , where  # is a number from 1 to the number of plots made.
!          - The filetype is gif
!
!   Some devices (see drivers.list in pgplot directory for all):
!
!     /NULL      (Null device, no output)
!     /GIF       (Graphics Interchange Format file, landscape orientation)
!     /PNG       (PNG file, landscape)
!     /PS        (PostScript file, landscape orientation)
!     /VPS       (PostScript file, portrait orientation)
!     /CPS       (Colour PostScript file, landscape orientation)
!     /VCPS      (Colour PostScript file, portrait orientation)
!
        call set_var(pgplot_device, splitline)

    case('pgplot_f_scale')
! pgplot_f_scale <integer plot id> <fmin> <fmax>
!   When a plot is defined in the setup file, it is assigned a integer
!   <plot id>, starting from 1.  This keyword lets the user define a
!   fixed max and min for the range of the function that is plotted.
!   For color plots, this will fix the color scale on the color bar,
!   for 2d line plots the keyword will fix the range on the axis of
!   the function value.
        call set_pgplot_colorscale(splitline)

    case('pgplot_f_max')
! pgplot_f_max <integer plot id> <fmax>
        call set_pgplot_colorscale(splitline,'fmax')

    case('pgplot_f_min')
! pgplot_f_min <integer plot id> <fmin>
        call set_pgplot_colorscale(splitline,'fmin')
        
    case('pgplot_width')
! pgplot_width <real value>
!   the width in inches of the whole pgplot window
        call set_var(pgplot_width, splitline)
        pgplot_width_set=.true.
        
    case('pgplot_aspect')
! pgplot_aspect <real value>
!   aspect ratio (height/width) for the whole pgplot window
        call set_var(pgplot_aspect, splitline)
        pgplot_aspect_set=.true.
        
    case('pgplot_palette')
! pgplot_palette <integer code>
!   specification of the color palette to use; 1=gray scale,
!   2=rainbow, 3=heat, 4=iraf, 5=aips, 6=tjp, 7=inverse gray scale, 8=hot, 9=jet
        call set_var(pgplot_palette, splitline)

    case('pgplot_background')
! pgplot_background <integer value>
!   pgplot color index for background
        call set_var(pgplot_background, splitline)
    case('pgplot_grid_color')
! pgplot_grid_color <integer value>
!   pgplot color index for grid
        call set_var(pgplot_grid_color, splitline)
    case('pgplot_text_color')
! pgplot_text_color <integer value>
!   pgplot color index for most text
        call set_var(pgplot_text_color, splitline)
    case('pgplot_curve_color')
! pgplot_curve_color <integer value>
!   pgplot color index for curves
        call set_var(pgplot_curve_color, splitline)
    case('pgplot_numbering_lw')
! pgplot_numbering_lw <integer value>
!   pgplot line width for axis numbers
        call set_var(pgplot_numbering_lw, splitline)
    case('pgplot_label_lw')
! pgplot_label_lw <integer value>
!   pgplot line width for axis labels
        call set_var(pgplot_label_lw, splitline)
    case('pgplot_grid_lw')
! pgplot_grid_lw <integer value>
!   pgplot line width for grid box
        call set_var(pgplot_grid_lw, splitline)
    case('pgplot_curve_lw')
! pgplot_curve_lw <integer value>
!   pgplot line width for curves
        call set_var(pgplot_curve_lw, splitline)
    case('pgplot_character_height')
! pgplot_character_height <real value>
!   The keyword sets the height of the characters in the plot.
!   Example: pgplot_character_height 1.0
        call set_var(pgplot_character_height, splitline)

    case('pgplot_viewport')
! pgplot_viewport 1 <4 real values>
!    specify the pgplot viewport using init syntax for 1d arrays.
!    example:
!       pgplot_viewport 1 0.1, 0.9, 0.1, 0.9        
!    would set the pgplot viewport to its default value
        call set_var(pgplot_viewport, splitline)

    case('pgplot_headsize')
! pgplot_headsize <real value>
!   use this to adjust the size of arrow heads. default value is 0.3
        call set_var(pgplot_headsize, splitline)
        
    case('pgplot_quiver_color')
! pgplot_quiver_color <integer value>
!    color indice to use for vectors        
        call set_var(pgplot_quiver_color, splitline)

    case('pgplot_quiver_lw')
! pgplot_quiver_lw <integer value>        
!    vector line weight in units of 0.13mm
        call set_var(pgplot_quiver_lw, splitline)
        
    case('pgplot_quiver_dens')
!
! pgplot_quiver_dens <integer value>
!
! set the density of arrows in a vector map of a 2D data array. 
! default value is 2.
! Example: to plot an arrow at every fourth grid point use
!    pgplot_quiver_dens 4
!
        call set_var(pgplot_quiver_dens, splitline)

    case('pgplot_quiver_scale')
!
! pgplot_quiver_scale <real value>
!
! adjust the length of quiver arrows. default value is 2.0
!
        call set_var(pgplot_quiver_scale, splitline)

    case('pgplot_layout')
!
! pgplot_layout <hnum> <vnum>
!
!   Specify the number of subplots in the horisontal and vertical direction.
!   <hnum>:      Number of plots in the horizontal
!   <vnum>:      Number of plots in the vertical
!
!   Example: pgplot_layout 1 2
!
        call add_pgplot(name, splitline)

    case('pgplot_interval')
!
! pgplot_interval <start> <end> <interval>
!
!   The interval in timesteps between each time a timeseries is plotted.
!   <start>:     Start timestep
!   <end>:       End   timestep
!   <interval>:  Interval between plots, in timesteps
!
        call add_pgplot(name, splitline)

    case('pgplot_geo')
! pgplot_geo <label> <showbath> <showcoast> <grid> <range>
!   Create an empty horizontal plot, <label> is the plot title, if
!   <showbath> is .true./t the bathymetry is colored by depth, if
!   <showcoast> is true the (grid) coastline is plotted.
!   if <grid> is true, overlay grid at major axis ticks.        
        call add_pgplot(name, splitline)

        
    case('pgplot_2d')
! pgplot_2d <label> <range>
!
!   Similar to bomplot_2d. Plots eta(m) with 2D velocity on top for a given region.
!
!   <label>: PLot title
!   <range>: Horisontal range (meters). If 0,0,0,0 plot full
!            range. the range is x1,x2,y1,y2 in meters, where point1
!            and 2 are SW and NE corner of a rectangle.
!
        call add_pgplot(name, splitline)

    case('pgplot_atmos')
! pgplot_atmos <label> <i1> <i2> <j1> <j2>
!   Plot PATM with WSPEEDU,WSPEEDV overlayed as vectors
!   <label>: Plot title
!   <range>: Horisontal range (meters). If 0,0,0,0 plot full
!            range. the range is x1,x2,y1,y2 in meters, where point1
!            and 2 are SW and NE corner of a rectangle.
        call add_pgplot(name, splitline)
        
    case('pgplot_vsect_grid')
!  pgplot_vsect_grid <logical value>
!    show box outline for vertical sections
        call set_var(pgplot_vsect_grid, splitline)

    case('pgplot_vsect_zrange')
!  pgplot_vsect_zrange <label> <z1> <z2>
!    for the vsect with the given <label> override the vertical range plotted
        call pgvsect_zrange(splitline)
        
    case('pgplot_2d_show_vsects')
! pgplot_2d_show_vsects <logical value>
!   scalar logical, if true, overlay all vsect trajectories in the 2d plot of eta,ua,va.
!   NB. vsects that are defined after this variable is set in the setup file will NOT be shown.
        call set_var(pgplot_2d_show_vsects, splitline)

    case('pgplot_title_time')
! pgplot_title_time <integer value>        
!   by default 1, which gives iso time in the title of pgplot sub windows
!   a value of 0 gives no time, 2 gives time in seconds        
        call set_var(pgplot_title_time, splitline)
        
    case('pgplot_coastline')
! pgplot_coastline <logical value>
!   if true, draw (box) coastline in all horisontal plots
        call set_var(pgplot_coastline, splitline)

    case('pgplot_coastline_file')
! pgplot_coastline_file <filename> <n>
!   if specified, draw this coastline instead of the grid coastline.        
!   the file is a binary stream file with N line segments given as
!   longitude,latitude coordinate pairs on the form
!   lon1,lat1,lon2,lat2.  it is necessary to also provide the correct
!   transform parameters, xpole, ypole etc for this to work.
        call get_pgplot_coastline(splitline)
        
    case('pgplot_timeseries')
! pgplot_timeseries <label> <what> <x> <y> <z>
!   Plot timeseries for one of the fields (speed (absolute value of velocity vector),u(ms^1),
!   v(ms^1),wr(ms^1),ua(ms^1),va(ms^1),s(p.s.u.),t(C),rho(kgm^3)
!
!   <what>:      The field to be plotted, e.g. speed,u,v,wr,ua,va,s,t,rho
!   <label>:     Name of the new plot
!   <x> <y> <z>: Global coordinate. The coordinate is converted into a constant i,j,k during initialization
!                assuming ETA=0, so it will not be 100% accurate during simulation. Fix later.
!
       if(.not.z_set) call message("sigma layer distribution must be specified to do this",2)
       call add_pgplot(name, splitline)

    case('pgplot_timeseries_window')
! pgplot_timeseries_window <integer value>        
!   pgplot will only show the pgplot_timeseries_window last
!   timesteps. defaults to 0 (show timeseries from t=0s)
        call set_var(pgplot_timeseries_window, splitline)        
        
    case('pgplot_hsect')
! pgplot_hsect <label> <what> <depth> <range> <speed>
!
!   Plot horizontal section of one of speed (absolute value of velocity vector), u(ms^1),v(ms^1),
!   wr(ms^1), s(p.s.u.),t(C),rho(kgm^3) while running.
!   at a specified depth (m), and for a specified region.
!
!   If depth<0, sample the section at |depth| meters above bottom.
!
!   <label>:    Name of the new plot
!   <what>:     The field to be plotted, e.g. speed,u,v,wr,ua,va,s,t,rho
!   <depth>:    Meters below static 0 level, or if negative, above sea bed.
!        
!   <range>: Horisontal range. If 0,0,0,0 plot full range. the range
!            is x1,x2,y1,y2 in meters, where point1 and 2 are sw and
!            ne corner of a rectangle. 
!        
!   <speed>:    Logical. speed = .true overlays velocity vectors
!
        if(.not.z_set) call message("sigma layer distribution must be specified to do this",2)        
        call add_pgplot(name, splitline)


    case('pgplot_vsect')
! pgplot_vsect <label> <what> N x1 y1 ... xN yN
!        
!   Plot a color image of a field in the vertical section below the
!   lines connecting the above N points. The field can be one of
!        
!   speed (absolute velocity), un (normal velocity), ut (tangential velocity), wr (vertical velocity), 
!   s (salinity), t (temperature), rho (density), conc1..conc9 (concentration field #).
!
!   <label>:    Name of the new plot
!   <what>:     The field to be plotted, see above
!   <N>:        Number of the points defining the cross section
!   <x1>:       x coordinate for the first point
!   <y1>:       y coordinate for the first point
!   ...
!   <xN>:       x coordinate for the last point
!   <yN>:       y coordinate for the last point
!
!   Example:  pgplot_vsect "arnavaag" speed 4  4.5 39.4   5.5 17.5   6.5 15.5   9.5 1.5
!
!   If in 2D slice mode (JM=1), the y-coordinates all should be zero.
!
        call add_pgplot(name, splitline)

    case('pgplot_vprof')
! pgplot_vprof <label> <what> <x> [<y>]
!   Plot vertical profile of a field under the horisontal grid position x,y.
!   <what> can be one of u,v,wr,s,t,rho,speed (more on request...)
        call add_pgplot(name, splitline)

    case('pgplot_slice3d')
! pgplot_slice3d <label> <what> <slice> <ind> <range>
!   Color plot of a slice of an array with no processing or masking.
!     <label> a descriptive label for the plot, ends up in the title.       
!     <what> is a BOM 3D field, u,v,wr,s,t,rho,... most fields are supported. more on request.
!     <slice> is "i","j" or "k". which index plane to plot.
!     <ind>   slice index for 3d arrays.
!     <range> is the "Fortran" global array range, e.g. (<i1>:<i2>,<k1>:<k2>) for a "j" slice.
        call add_pgplot(name, splitline)
        
    case('pgplot_slice2d')
! pgplot_slice2d <label> <what> <range>
!   Color plot of a slice of an array "just the way it is".
!     <label> a descriptive label for the plot, ends up in the title.       
!     <what> is a BOM 2D field
!     <range> is the "Fortran" global array range (<i1>:<i2>,<j1>:<j2>)
        call add_pgplot(name, splitline)
        
    case('store_land_mask')
! store_land_mask <logical value>
!   If true, dump ascii map of the wet/dry points to the file "grid.data". Defaults to false.
!   Also show position of rivers and wells.
        call set_var(dodumpmask, splitline)
        dodumpmask_set=.true.

    case('store_sigma_layers')
! store_sigma_layers <logical value>
!   if true, dump sigma layers to the file "sigma-layers.data". defaults to false.
!
        call set_var(store_sigma_layers, splitline)
        store_sigma_layers_set=.true.


! diagnostic output of max and min as well as location for selected variables

    case('maxmin_nd')
! maxmin_nd <integer value>
!   maxmin values are dumped to screen every maxmin_nd'th step, by default 1.
        call set_var(maxmin_nd, splitline)

    case('maxmin_start')
! maxmin_start <integer value>
!   the timestep for which we start dumping maxmin values to screen, by default 1.
        call set_var(maxmin_start, splitline)

    case('maxmin_end')
! maxmin_end  <integer value>
!   the timestep where we stop dumping maxmin values to screen, by default numstep.
        call set_var(maxmin_end, splitline)

    case('maxmin')
! maxmin <label> <what> [<index range>]
!   Dump max, min and location to screen for 2d or 3d array <what>
!   Most of the 2d adn 3d arrays from state and mod_bound can be specified via <what>.
!   <label> is used in the printout for identification.        
!   Optionally compute max,min only in the given global index range. (NB not meters!)
!   The maxmin's are printed to screen at the end of each step by default.
!   If debugnormlevel is 2 maxmins are dumped for all 3d "dump_state" calls.
!   If debugnormlevel is 3 also for the 2d calls.
        call add_maxmin(splitline)

    case('monitor')
! monitor <label> <what> <index range>
!   Print values of <what> in the given <index range> to screen for debugging.
!   Most of the 2d adn 3d arrays from state and mod_bound can be specified via <what>.
!   The values are printed at the same times as for the maxmin keyword.       
        call add_maxmin(splitline)

    case('runtime_interval')
! runtime_interval <integer value>
!   Read the run time config file every runtime_interval'th time step. The file is erased after reading.
        call set_var(runtime_interval, splitline)
        
    case('checkpoint_interval')
! checkpoint_interval <integer value>
!   Time in seconds between checkpoints
!   If this variable is set, checkpointing is activated.
!   It is required to specify NDATE to use checkpointing.
        if(.not. ndate_set) call message("NDATE must be given before this varaible is set",2)
        call set_var(checkpoint_interval, splitline)
        
    case('checkpoint_time')
! checkpoint_time <ISO 8601 date> <label>
!   Add a single time at which to create a checkpoint. <label> is a string
!   used to name the checkpoint directory.
!   It is required to specify NDATE to use checkpointing.
        if(.not. ndate_set) call message("NDATE must be given before this varaible is set",2)
        call add_checkpoint_time(splitline)

    case('checkpoint_timestep')
! checkpoint_timestep <integer_value>
!   Specify on which timestep to do a checkpoint        
        call set_var(checkpoint_timestep, splitline)

    case('checkpoint_delay')
! checkpoint_delay <integer value>
!   in some systems, filesystem operations may be delayed, e.g. the
!   mkdir command creating the checkpoint directory.  to avoid failure
!   to write to non existing directories, the checkpoint routine can
!   be asked to wait a couple of seconds after the mkdir command
        call set_var(checkpoint_delay, splitline)
        
    case('cruise')
! cruise <label> <cruise file> [<coordinate type> <timecol> <lon col> <lat col>]
!   Store S,T along a cruise path specified as time and location in
!   the given text file. the cruise file is a text file where each        
!   line has time,x,y,z for each cruise sample point. there may be
!   more data on each line, if so, the columns where time etc can be
!   picked must be specified.        
!   If coordinate type=0 horisontal pos is lon,lat (geographical
!   coordinates), 1: coordinates are meters in the global grid, 2: grid        
!   coordinates. <timecol> <loncol> <latcol> says which column in the        
!   cruise file to fetch time, coordinate1, coordinate2
!   respectively. if not given 1 2 3 is assumed.       
        if(setlonlat .and. trim(splitline(4))=="0") call init_lonlat()
        call add_cruise(splitline)

    case('cruise_dump_interval')
! cruise_dump_interval <integer value>
!    the number of timesteps between each dump of cruise data        
        call set_var(cruise_dump_interval, splitline)
        
    case('test_forcing_mode')
! test_forcing_mode <logical value>
!   test run without numerics. reads init and forcing each time step, and does all output.
        call set_var(test_forcing_mode, splitline)        

    case('average_dir')
! average_dir <string>
!   the averaging keyword routines will dump files to this directory
        call set_var(average_dir, splitline)
        
    case('vsectaverage')
! vsectaverage <label> <what> <how often> <section: N x1 y1 ... xN yN>!
!   compute the average of a field on the given section and dump to
!   file at given intervals
!   <what> almost any BOM field, or "transport", "speed" ...
!   <how often> can be "hourly", "daily", "weekly", "monthly" or
!               "istep100" which will dump every 100th time step.
!   <section> a list of n x,y points defining a section of linear segments.
        call add_average(splitline)
        
    case('hsectaverage')
! hsectaverage <label> <what> <how often> <depth> <range: x1 x2 y1 y2>        
!   compute the average of a field at the given HSECT depth, and dump
!   to file at given intervals, like 'vesectaverage'        
!   <depth> is by default the z position below neutral 0. If <depth> is positive,
!   it is interpreted as the layer ABS(<depth>) meters above the sea floor.        
        call add_average(splitline)
        
    case('average')
! average <label> <what> <how often> <range: x1 x2 y1 y2>
!   compute the average of a field under the given horizontal range
!   and dump to file at given intervals.
!   <what> almost any BOM field, or "transport", "speed" ...
!   <how often> can be "hourly", "daily", "weekly", "monthly" or e.g.
!               "istep100" which will dump every 100th time step.
        call add_average(splitline)

    case('uvstaverage')
! uvstaverage <how often> <depth file> 
!   compute the average of U,V,S,T fields at the depths given in <depth file>, and dump
!   to file. <how often> can be "hourly", "daily", "weekly", "monthly" or e.g.
!   "istep100" which will dump every 100th time step.
        call add_average(splitline)

    case('vprofaverage')
! vprofaverage <how often> <x> <y> <depth file> 
!   compute the average of U,V,S,T fields at depths given in <depth file> under the given x,y position. Dump
!   to file at intervals given via <how often>, which can be "hourly", "daily", "weekly", "monthly" or e.g.
!   "istep100" which will dump every 100th time step. 
        call add_average(splitline)
        
    case('uadv')
! uadv <integer code> <data for initializing 3d array>
!   x component of divergence free velocity field used to advect fields. rarely needs initialization.
        call set_var(uadv, splitline)
        uadv_set=.true.
    case('vadv')
! vadv <integer code> <data for initializing 3d array>
!   y component of divergence free velocity field used to advect fields. rarely needs initialization.        
        call set_var(vadv, splitline)
        vadv_set=.true.
    case('wubot')
! wubot <integer code> <data for initializing 2d array>
!   x component of bottom stress. rarely needs to be initialized.
        call set_var(wubot, splitline)
        wubot_set=.true.
    case('wvbot')
! wvbot <integer code> <data for initializing 2d array>
!   y component of bottom stress. rarely needs to be initialized.
        call set_var(wvbot, splitline)
        wvbot_set=.true.
    case('mass')
! mass <real value>
!   (for testing purposes) explicitly specify the model average value of RHO-1000
        call set_var(given_mass, splitline)
    case('periodic_ew')
! periodic_ew <logical value>
!   if true, connect east and west edge of grid. all boundary conditions will be ignored.
!   it is required that MOD(IM,NUMBLI) == 0, i.e. the last cell is at the end of the last block.
!   this keyword must be specified before H.
        if(h_set) call message("Must be specified before H!",2)
        call set_var(periodic_ew, splitline)
        
! Now a section of variables used to restart the model, that do not need documentation.
        
    case('istep')
        if(.not.restart) call message("RESTART should be .true. to set ISTEP",1)
        call set_var(istep, splitline)
        first_step = istep + 1
    case('initial_time')
        if(.not.restart) call message("RESTART should be .true. to set INITIAL_TIME",2)
        call set_var(initial_time, splitline)
    case('reread_pgplot_timeseries')
        if(.not.restart) call message("RESTART should be .true. to use this keyword",2)
        call reread_pgplot_timeseries(istep)
    case('time')
        if(.not.restart) call message("RESTART should be .true. to set TIME",2)
        call set_var(time, splitline)
        if(restart) previous_checkpoint_time = time
    case('rhoaver')
        if(.not.restart) call message("RESTART should be .true. to set RHOAVER",2)
        call set_var(rhoaver, splitline)       
    case('wadv')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(wadv, splitline)
        wadv_set=.true.
    case('un')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(un, splitline)
        un_set=.true.
    case('delun')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(delun, splitline)
        delun_set=.true.
    case('vn')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(vn, splitline)
        vn_set=.true.
    case('delvn')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(delvn, splitline)
        delvn_set=.true.
    case('wn')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(wn, splitline)
        wn_set=.true.
    case('delwn')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(delwn, splitline)
        delwn_set=.true.
    case('rhon')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(rhon, splitline)
        rhon_set=.true.
    case('rhonm1')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(rhonm1, splitline)
        rhonm1_set=.true.
    case('sn')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(sn, splitline)
        sn_set=.true.
    case('snm1')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(snm1, splitline)
        snm1_set=.true.
    case('tn')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(tn, splitline)
        tn_set=.true.
    case('tnm1')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(tnm1, splitline)
        tnm1_set=.true.
    case('u_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(u_bound, splitline)
        u_bound_set=.true.
    case('v_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(v_bound, splitline)
        v_bound_set=.true.
    case('w_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(w_bound, splitline)
        w_bound_set=.true.
    case('s_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(s_bound, splitline)
        s_bound_set=.true.
    case('t_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(t_bound, splitline)
        t_bound_set=.true.
    case('s_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(s_initial, splitline)
        s_initial_set=.true.
    case('t_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(t_initial, splitline)
        t_initial_set=.true.
    case('rho_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(rho_initial, splitline)
        rho_initial_set=.true.
    case('q2_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(q2_bound, splitline)
        q2_bound_set=.true.
    case('q2l_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(q2l_bound, splitline)
        q2l_bound_set=.true.
    case('rho_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(rho_bound, splitline)
        rho_bound_set=.true.
    case('ua_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(ua_bound, splitline)
        ua_bound_set=.true.
    case('va_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(va_bound, splitline)
        va_bound_set=.true.
    case('eta_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(eta_bound, splitline)
        eta_bound_set=.true.
    case('etabfrs')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(etabfrs, splitline)
        etabfrs_set=.true.
    case('etatfrs')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        call set_var(etatfrs, splitline)
        etatfrs_set=.true.

    case('conc1_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(1)%allocated) call allocate_concentration(1)
        call set_var(co(1)%conc_bound, splitline)
    case('conc2_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(2)%allocated) call allocate_concentration(2)
        call set_var(co(2)%conc_bound, splitline)
    case('conc3_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(3)%allocated) call allocate_concentration(3)
        call set_var(co(3)%conc_bound, splitline)
    case('conc4_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(4)%allocated) call allocate_concentration(4)
        call set_var(co(4)%conc_bound, splitline)
    case('conc5_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(5)%allocated) call allocate_concentration(5)
        call set_var(co(5)%conc_bound, splitline)
    case('conc6_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(6)%allocated) call allocate_concentration(6)
        call set_var(co(6)%conc_bound, splitline)
    case('conc7_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(7)%allocated) call allocate_concentration(7)
        call set_var(co(7)%conc_bound, splitline)
    case('conc8_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(8)%allocated) call allocate_concentration(8)
        call set_var(co(8)%conc_bound, splitline)
    case('conc9_bound')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(9)%allocated) call allocate_concentration(9)
        call set_var(co(9)%conc_bound, splitline)

    case('conc1_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(1)%allocated) call allocate_concentration(1)
        call set_var(co(1)%conc_initial, splitline)
    case('conc2_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(2)%allocated) call allocate_concentration(2)
        call set_var(co(2)%conc_initial, splitline)
    case('conc3_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(3)%allocated) call allocate_concentration(3)
        call set_var(co(3)%conc_initial, splitline)
    case('conc4_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(4)%allocated) call allocate_concentration(4)
        call set_var(co(4)%conc_initial, splitline)
    case('conc5_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(5)%allocated) call allocate_concentration(5)
        call set_var(co(5)%conc_initial, splitline)
    case('conc6_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(6)%allocated) call allocate_concentration(6)
        call set_var(co(6)%conc_initial, splitline)
    case('conc7_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(7)%allocated) call allocate_concentration(7)
        call set_var(co(7)%conc_initial, splitline)
    case('conc8_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(8)%allocated) call allocate_concentration(8)
        call set_var(co(8)%conc_initial, splitline)
    case('conc9_initial')
        if(.not.restart) call message("RESTART should be .true. to set this variable",1)
        if(.not. co(9)%allocated) call allocate_concentration(9)
        call set_var(co(9)%conc_initial, splitline)

        
    case('timeseries_dump_start')
        if(.not.restart) call message("RESTART should be .true. to set this variable!",1)
        call set_var(timeseries_dump_start, splitline)
    case('timeseries_file_number')
        if(.not.restart) call message("RESTART should be .true. to set this variable",2)
        call set_var(timeseries_file_number, splitline)
        
    case default
        call message('Unknown keyword, ignored!',1)
        ierr=1
    end select


  contains

    integer function getzone()  ! (macro for getting zone number)
      read(splitline(2),*,err=101,end=101) getzone
      return
101   call message("couldn't get integer for boundary zone from column 2",2)
    end function getzone

  end subroutine parse_setup_line


  
  subroutine add_checkpoint_time(splitline)
    implicit none
    character(*) splitline(:)
    integer cptime(6)

    numcptime = numcptime + 1
    
    if(numcptime > max_checkpoint_times) &
         call message("too many individual checkpoint times, increase MAX_CHECKPOINT_TIMES and recompile",2)
    call get_iso8601(splitline(2), cptime)

    checkpoint_time(numcptime) = timediff(initial_time, cptime)

    if( checkpoint_time(numcptime)-time <= 0.5) then
        if(master) write(*,*) " Checkpoint time in seconds:",checkpoint_time(numcptime)
        if(master) write(*,*) " Initial model time in seconds:", time
        call message("The checkpoint will be discarded")
        checkpoint_time(numcptime) = -1
        numcptime = numcptime - 1
        return
    end if
    
    if(checkpoint_time(numcptime) < 0.0 .or. checkpoint_time(numcptime) > 63072000.0) then
        if(master) write(*,*) "checkpoint time in seconds (since initial NDATE) t=", checkpoint_time(numcptime)        
        call message("checkpoint time looks wrong, check that NDATE and the requested checkpoint time make sense",2)
    end if
    
    checkpoint_label = trim(splitline(3))
    
    if(debug > 0 .and. master) write(*,'(2x,a,f0.1,a,a)') "Adding checkpoint at t=",checkpoint_time(numcptime),&
         "s with label=",trim(checkpoint_label(numcptime))
  end subroutine add_checkpoint_time


  subroutine activate_std_array_setting()    
! Sometimes we want SET_VAR to read "normal" non distributed
! arrays. We then set ghost_width=0 temporarily to "re-use" set_var
! machinery. This routine is called just before such behaviour is
! wanted from SET_VAR, and should be turned off immideately after.
    implicit none
    backup_std_array_gwi = ghost_wi 
    backup_std_array_gwj = ghost_wj
    ghost_wi = 0
    ghost_wj = 0
    init_std_array = .true.
  end subroutine activate_std_array_setting


  subroutine deactivate_std_array_setting()
! Restore ghost width and reset switch.    
    implicit none
    ghost_wi = backup_std_array_gwi
    ghost_wj = backup_std_array_gwj
    init_std_array = .false.
  end subroutine deactivate_std_array_setting
  

  subroutine initialize_depth(splitline)
!
! h is special. it is used to initialize the mpi block decomposition
! which is needed to broadcast h. for very large cases, the init of h
! should also be the only time a global field is present on a single
! processor. below, the global h is allocated once, before the rest of
! the model is allocated.
!
    use autodecomp
    implicit none
    character(*) splitline(:)
    integer tmpint1, tmpint2
    real, allocatable :: tmph(:,:)
    integer start(10),end(10),i

    if(h_set) call message("H is only allowed to be set once",2)

    if(any((/im,jm,kb/) < 0)) call message("IM,JM,KB must all be set before H",2)

    if(master) then
        allocate(gh(im,jm))
        call activate_std_array_setting()
        call set_var(gh, splitline)
        call deactivate_std_array_setting()
    else
        allocate(gh(1,1))
    end if

    call close_boundary(gh)

    if(omit_dry_blocks)then
        if(numbli < 0 .or. numblj < 0) then
            if(master) call get_decomp(numproc, gh, numbli, numblj)
            call mbroadcast(numbli, numblj)
        end if
    else    
        if(numbli < 0 .and. numblj < 0) then
           ! find decomposition that match the number of processors
            call get_optimal_block_decomp(numbli,numblj,im,jm)
        else if(numblj < 0) then
            call message("keyword numblj not yet set, using default = 1",0)
            numblj = 1
        else if(numbli < 0) then
            call message("keyword numbli not yet set, using default = 1",0)
            numbli = 1
        end if
    end if

    if(periodic_ew .and. mod(im,numbli) .ne. 0) call message("when using 'PERIODIC_EW t' MOD(IM,NUMBLI) must be 0.",2)
    
    call init_mpilib(im, jm, kb, gh, 0.0)

! i1,i2 etc now set, so we can allocate depth

    allocate(h(i1:i2,j1:j2))
    h = 0.0
    
    if(numproc == 1) then
        h(1:im,1:jm) = gh(1:im,1:jm)
        if(periodic_ew) then
            h(-1:0,:) = h(im-1:im,:)
            h(im+1:im+2,:) = h(1:2,:)
        end if
    else
        call scatterfield(gh, h)
    end if

    h_set = .true.

    call dumpmask()
    
    call find_max_reals(1,h,depmax,loc=depmaxloc)
    allocate(tmph(i1:i2,j1:j2))
    tmph = h
    where(h < 0.00001) tmph = depmax + 1.0
    
    call find_min_reals(1,tmph,depmin,loc=depminloc)

    if(master)then
        write(*,'(/,2x,a,f0.1)') "The maximum depth in meters is ", depmax
        write(*,'(2x,a,f0.1,/)') "The minimum depth in meters is ", depmin
        if(depmax > 11033) then
            call message("The maximum depth is deeper than the Mariana Trench (11033m), so I think we stop here!",2)
        end if
        
        if(depmax <= 0.0) call message("All depths are <= 0. Check specification and format of input files",2)
        
        if(depmin < 1) then
            if(master) write(*,*)
            call message("The minimum depth is < 1m. In realistic cases it is not unheard of with &
                 &ETA < -1. BOOOM. DEPMIN=",1,rval=(/depmin/))
            if(master) write(*,'(a,/)') "(You have been warned....)"
        end if
    end if

    deallocate(tmph)


! while we still have gh allocated...
    if(num_zones_west == 0)then
        call initialize_boundary_zones(gh, 'west', num_zones_west, start, end)
        do i=1,num_zones_west
            bowest(i)%global_range = (/1,lb,start(i),end(i)/)
        end do
    else
        call message("will not auto detect boundary zones on western edge, as they are set already",1)
    end if
    if(num_zones_east == 0)then
        call initialize_boundary_zones(gh, 'east', num_zones_east, start, end)
        do i=1,num_zones_east
            boeast(i)%global_range = (/im-lb+1,im,start(i),end(i)/)
        end do
    else
        call message("will not auto detect boundary zones on eastern edge, as they are set already",1)
    end if
    if(num_zones_north == 0)then
        call initialize_boundary_zones(gh, 'north', num_zones_north, start, end)
        do i=1,num_zones_north
            bonorth(i)%global_range = (/start(i),end(i),jm-lb+1,jm/)
        end do
    else
        call message("will not auto detect boundary zones on northern edge, as they are set already",1)
    end if
    if(num_zones_south == 0)then
        call initialize_boundary_zones(gh, 'south', num_zones_south, start, end)
        do i=1,num_zones_south
            bosouth(i)%global_range = (/start(i),end(i),1,lb/)
        end do
    else
        call message("will not auto detect boundary zones on southern edge, as they are set already",1)
    end if

    if(adjust_bz) call adjust_boundary_zones(gh)


! figure out if some of our boundary zones are connected in the corners

    call check_b_zone_connections(sw_connected, nw_connected, ne_connected, se_connected)

! for MPI:
! convert zone coordinates from global to block local, save the global coordinates.

    call parallelize_boundary_zones()
    
    call allocate_bom()
    
! this is also a good place to set dx and dy
    if(.not.dx_set)then
        if(lx_set) then
            if(im-2 > 0)then
                dx = lx/(im-2)
            else
                call message("IM<2 is not supported",2)
            end if
        else
            call message("DX and LX not yet set, one of them should be!",2)
        end if
    else
        lx=(im-2)*dx
    end if
    if(.not.dy_set)then
        if(ly_set)then
            if(jm-2 > 0)then
                dy = ly/(jm-2)
            else
                call message("JM<2, ignoring requested ly",1)
            end if
        else
            if(jm > 1) then
                call message("DY and LY not yet set, one of them should be!",2)
            else
                ly = 0.0
                dy = real(1.0_8)
            end if
        end if
    else
        ly=(jm-2)*dy        
    end if
  end subroutine initialize_depth


  subroutine set_edge_closed(splitline)
! initialize the edge_closed array from a line like
!    close_boundary [east] [west] [south] [north]
! stuff in [] are optional as usual.
    character(*) splitline(:)
    character(50) tmpstring
    integer numel,i
    
    do i=1,4
        numel = i
        tmpstring = trim(adjustl(splitline(numel+2)))
        if(trim(splitline(numel+2)) == "" .or. tmpstring(1:1) == "!" ) exit
    end do
    do i=1,numel
        select case(splitline(i+1))
        case("east")
            if(debug>1.and.master) write(*,'(2x,a)') "closing east"
            edge_closed(2)=.true.
        case("west")
            if(debug>1.and.master) write(*,'(2x,a)') "closing west"
            edge_closed(1)=.true.
        case("north")
            if(debug>1.and.master) write(*,'(2x,a)') "closing north"            
            edge_closed(4)=.true.
        case("south")
            if(debug>1.and.master) write(*,'(2x,a)') "closing south"            
            edge_closed(3)=.true.
        case default
            write(*,*) "edge=", trim(splitline(i+1))
            call message("invalid edge",2)
        end select
    end do    
  end subroutine set_edge_closed

  
  subroutine close_boundary(gh)
    implicit none
    real gh(:,:)

! (kept for compability)    
    if(close_h) edge_closed(:) = .true.

    ! west east south north
    if(edge_closed(1)) gh(1,:)=0
    if(edge_closed(2)) gh(size(gh,1),:) = 0
    if(jm > 1) then
        if(edge_closed(3)) gh(:,1) = 0    
        if(edge_closed(4)) gh(:,size(gh,2)) = 0
    end if
  end subroutine close_boundary
  

  subroutine initialize_bin_regions(bin, frs, n)
! loop over all bin regions for an edge, if a bin range is set, initialize the bin.
    implicit none
    type(binreg_object) bin(:)
    type(boundary_object) frs(:)
    integer n,i,r(4),lr(4),ibin, fr(6), loc(3)
    real max_depth
    
    do i=1,n

! find the max depth of an area covering both the bin and the frs
        fr=(/&
             min(bin(i)%range(1), frs(i)%global_range(1)), &
             max(bin(i)%range(2), frs(i)%global_range(2)), &
             min(bin(i)%range(3), frs(i)%global_range(3)), &
             max(bin(i)%range(4), frs(i)%global_range(4)), &
             1, 1 /)

        call find_max_reals(1, h, max_depth, loc, range=fr)        
        
        r = bin(i)%range
        if(minval(r) > 0) then
            call globalrect2local(r(1),r(2),r(3),r(4),lr(1),lr(2),lr(3),lr(4),bin(i)%inblock)
            bin(i)%local_range = lr

            ! create bin%levels layers of equal thickness down to max_depth
            do ibin = 1,bin(i)%levels+1
                bin(i)%zlevels(ibin) = -(ibin-1)/real(bin(i)%levels) * max_depth
            end do
            
            ! the vertical coordinate of the center of each bin layer
            do ibin = 1,bin(i)%levels
                bin(i)%zinn(ibin) = 0.5*(bin(i)%zlevels(ibin) + bin(i)%zlevels(ibin+1))
            end do
            
            fr(1:4) = bin(i)%range
            fr(5:6) = (/1,1/)
            bin(i)%numwet = nint(psum(real(gridbox),im,jm,1, range=fr))
            
            bin(i)%initialized = .true.

            if(debug>0.and.master) write(*,'(2x,a,4(i0,1x),a,f0.1,a,i0)') "bin range=",bin(i)%range,&
                 " max depth=",max_depth," levels=",bin(i)%levels            
        end if
        
    end do
  end subroutine initialize_bin_regions


  subroutine add_open_boundary(edge, splitline)
! manually add open boundary zone, must be called before the initialize_depth routine.
! if this routine is called once for an edge, disable the automatic detection of zones for the edge.    
    implicit none
    character(*) edge, splitline(:)
    integer start,end

    read(splitline(3:4),*,err=101,end=101) start,end

    if(debug > 0 .and. master) then
        select case(edge)
        case("north","south")
            write(*,'(2x,a,a,a,i0,a,i0)') "edge=",edge,", starts on i=",start,", ends at i=",end
        case("east","west")
            write(*,'(2x,a,a,a,i0,a,i0)') "edge=",edge,", starts on j=",start,", ends at j=",end
        end select
    end if

    select case(edge)
    case('west')
        num_zones_west = num_zones_west + 1
        if(num_zones_west > max_zones) call message("too many open zones on western edge",2)
        bowest(num_zones_west)%global_range = (/1,lb,start,end/)
    case('east')
        num_zones_east = num_zones_east + 1
        if(num_zones_east > max_zones) call message("too many open zones on eastern edge",2)
        boeast(num_zones_east)%global_range = (/im-lb+1,im,start,end/)
    case('north')
        num_zones_north = num_zones_north + 1
        if(num_zones_north > max_zones) call message("too many open zones on northern edge",2)
        bonorth(num_zones_north)%global_range = (/start,end,jm-lb+1,jm/)
    case('south')
        num_zones_south = num_zones_south + 1
        if(num_zones_south > max_zones) call message("too many open zones on southern edge",2)
        bosouth(num_zones_south)%global_range = (/start,end,1,lb/)
    case default
        call message("unknown edge",2)
    end select
    return
101 call message("Could not read start,end indice for open boundary from columns 3 and 4",2)
  end subroutine add_open_boundary


  subroutine set_pre_setup2_defaults()
! when using the setup2 system, some scalar variables need default
! values that are different from the reasonable default given in
! the declaration, this is the place to do this.
    numbli=-1
    numblj=-1

    
! make all known pointers point to null() initially, otherwise ASSOCIATED() will not work.
    
! state    
    nullify(z,dz,dzr,dzz,zz)
    nullify(cor,fsm,dum,dvm,    &
      & eta,etap,h,hu,hv,d,du,dv,wusurf,wvsurf, &
      & ssurf,tsurf,rhosurf,wubot,wvbot,cbc,patm,am2d,  &
      & ua,va,husqr,hvsqr,longitude,latitude, &
      & sponge_cm2d,sponge_cm,sponge_km,sponge_kh,coast)
    nullify(u,v,w,wr,s,t,rho, &
      & am,ah,km,kh,kq,ddz,dudz,dvdz,    &
      & uadv,vadv,wadv,q2,q2l,deltau,deltav)
    nullify(un,unm1,delun,delustar,delunm1)
    nullify(vn,vnm1,delvn,delvstar,delvnm1)
    nullify(sn,snm1,delsn,delsstar,delsnm1)
    nullify(tn,tnm1,deltn,deltstar,deltnm1)
    nullify(wn,wnm1,delwn,delwstar,delwnm1)
    nullify(rhon,rhonm1,delrhon,delrhostar,delrhonm1)
    nullify(p)
    nullify(rhoback, soptu, soptv, masksoptu, masksoptv)
    
!mod_bound
    nullify(iriver,jriver,kriver1,kriver2,coderiver,riverformat)
    nullify(sriver,triver,friver,dzriver,depriver,criver)
    nullify(river_name,iwell,jwell,well_format,swell,twell,fwell)
    nullify(cwell,well_name,freq)
    nullify(etatamp,etatpha,uatamp)
    nullify(uatpha,vatamp,vatpha)
    nullify(wspeedu,wspeedv)
    nullify(alfa_u, alfa_v, alfa_eta, alfae_u, alfae_v, alfae_eta)
    nullify(u_bound, v_bound, w_bound, s_bound, t_bound, q2_bound, q2l_bound, rho_bound)
    nullify(ua_bound, va_bound, eta_bound, etabfrs,etatfrs)
    nullify(ua_tide, va_tide, eta_tide)
    nullify(s_initial, t_initial, rho_initial)
   
  end subroutine set_pre_setup2_defaults


  subroutine set_defaults()
!
! set defaults, and warn, for variables that have effect on the
! physics but are not initialized.
!
! uninitialized variables that deal with numerical choices or
! diagnostics like plotting etc. are not changed.
!
! set reasonable defaults for all uninitialized boundary arrays, for
! now this means to set them to the initial value of the corresponding
! field. for these arrays, only touch the boundary zones.
!   
    implicit none
    
    real :: default_kh=1e-4, default_km=1e-3, default_kq=1e-4, default_q2=1e-4, default_q2l=1e-4
    real :: default_cor=1e-4, default_eta=0.0, default_p=0.0, default_s = 33, default_t=10
    
    integer zi,time(8),jj

    if(master)then
        write(*,'(/,a,/)') "========================================================="
        write(*,'(a,/)') "Now we check if important keywords are set or not:"
    end if
        
    if(.not.ah_set)then
        if(.not.smag) call message('AH not set, setting SMAG to true')
        smag=.true.
    end if
    if(.not.am_set)then
        if(.not.smag) call message("AM not set, setting SMAG to true")
        smag=.true.
    end if
    if(.not.am2d_set)then
        if(.not.smag) call message("AM2D not set, setting SMAG to true")
        smag=.true.
    end if
    if(.not.atmpress_set)then
        call message("ATMPRESS not set, using default=", lval=(/atmpress/))
    end if
    if(.not.cbc_set)then
        call message("CBC not set, will compute it from MAX(CD,.16/LOG((ZZ(KB-1)-Z(KB))*H(I,J)/Z0)**2)")
    end if
    if(.not.cd_set)then
        call message("CD not set, using default =", rval=(/cd/))
    end if
    if(.not.ch_set)then
        call message("CH not set, using default =", rval=(/ch/))
    end if
    if(smag)then
        if(.not.cm_set)then
            call message("CM not set, using default =", rval=(/ch/))
        end if
        if(.not.cm2d_set)then
            call message("CM2D not set, using default =", rval=(/cm2d/))
        end if
    end if
    if(.not.cor_set)then
        call message("COR not set, using default constant value = 1e-4")
        cor = 1e-4
    end if
    if(.not.dx_set)then
        if(lx_set) then
            if(im-2 > 0)then
                dx = lx/(im-2)
            else
                call message("IM<2 is not supported",2)
            end if
        else
            call message("DX and LX not set, one of them should be!",2)
        end if
    end if
    if(.not.dy_set)then
        if(ly_set)then
            if(jm-2 > 0)then
                dy = ly/(jm-2)
            else
                call message("JM<2, ignoring requested ly",1)
            end if
        else
            if(jm > 1) then
                call message("DY and LY not set, one of them should be!",2)
            else
                ly = 0.0
                dy = real(1.0_8)
            end if
        end if
    end if

    if(.not.ndate_set) then
        call message("NDATE not set, using current time as start time: ", advance="no ")
        call date_and_time(values=time)
        ndate(1:3) = time(1:3)
        ndate(4:6) = time(5:7)
        initial_time = ndate
        if(master) write(*,'(i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2)') ndate
        call message("If any forcing uses absolute dates in ISO format, NDATE should be set to correspond!",1)
    end if

    if(.not.timezone_set) then
       call message("TIMEZONE not set, using GMT by default.",1)
       timezone = 0.
    end if

    if(.not.eta_set)then
        call message("ETA not set, using default = ", rval=(/default_eta/))
        eta = default_eta
    end if
    if(.not.kh_set)then
        call message("KH not set, using default = ", rval=(/default_kh/))
        kh = default_kh
    end if
    if(.not.khmin_set)then
        call message("KHMIN not set, using default = ", rval=(/khmin/))
    end if
    if(.not.km_set)then
        call message("KM not set, using default = ", rval=(/default_km/))
        km = default_km
    end if
    if(.not.kmmin_set)then
        call message("KMMIN not set, using default = ", rval=(/kmmin/))
    end if
    if(.not.kq_set)then
        call message("KQ not set, using default =", rval=(/default_kq/))
        kq = default_kq
    end if
    if(.not.n2d_set)then
        call message("N2D, the number of 2d steps per 3d step not set, setting it to 30")
        n2d = 30
    end if
    if(.not.ndivis_set)then
        if(depmax > 0.0 .and. dx > 0.0) then
            ndivis = nint(7200.*sqrt(2*grav*depmax)/dx/n2d)
            if(ndivis==0) then
                call message("With that crazy setupfile, NDIVIS was computed to 0, overriding it to be 1",1)
                ndivis=1
            end if
            if(mod(ndivis,2).ne.0) ndivis = ndivis+1
            dt = 3600./ndivis
        end if                
        call message("NDIVIS not set, using suggested NDIVIS =", ival=(/ndivis/))
    end if
    if(.not.nonhyd_set)then
        call message("NONHYD not set, using default = ", lval=(/nonhyd/))
    end if
    if(.not.internal_pressure_set .or. intp4_set) then
        call message("internal pressure algorithm not specified, using default: ", sval=(/trim(internal_pressure)/)) 
    end if
    if(.not.numstep_set)then
        call message("NUMSTEP not set, using default =", ival=(/numstep/))
    end if
    if(.not.p_set)then
        if(nonhyd)then
            call message("P not set, using default = ", rval=(/default_p/))
            p = default_p
        end if
    end if
    if(.not.patm_set)then
        patm = pmean
        call message("PATM not set, using PATM = PMEAN =", rval=(/pmean/))
    end if
    if(.not.pmean_set)then
        call message("PMEAN not set, using default = ", rval=(/pmean/))
    end if
    if(.not.q2_set.and.vertical_closure==1)then
        call message("Q2 not set, using default = ", rval=(/default_q2/))
        q2 = default_q2
    end if
    if(.not.q2l_set.and.vertical_closure==1)then
        call message("Q2L not set, using default = ", rval=(/default_q2l/))
        q2l = default_q2l
    end if
    if(.not.vertical_closure_set)then
        call message("VERTICAL_CLOSURE not set, the default is to use Mellor Yamada (vertical_closure=1)")
    end if
    if(.not.rho_set.and.advect_density)then
        call message("RHO not set, using default = ", rval=(/rho0/))
        rho = rho0 - 1000
    end if
    if(.not.rho0_set)then
        call message("RHO0 not set, using the default value =", rval=(/rho0/))
    end if
    if(.not.rhosurf_set .and. advect_density)then
        call message("RHOSURF not set, using default = 0")
        rhosurf = 0.0
    end if
    if(.not.rivlog_set)then
        call message("RIVLOG not set, using default = false")
    end if
    if(.not.roair_set)then
        call message("ROAIR not set, using default = ", rval=(/roair/))
    end if
    if(.not.s_set.and..not.advect_density)then
        call message("S not set, using default = 33 psu")
        s = 33.0
    end if
    if(.not.z_set)then
        call message("Z not set, calling initialize_sigma_layers:")
        call initialize_sigma_layers((/"sigma_layers","0           "/))
    end if
    if(.not.smag_set)then
        call message("SMAG not set, using default =", lval=(/smag/))
    end if
    if(.not.ssurf_set.and..not.advect_density)then
        call message("SSURF not set, using default = 0")
        ssurf = 0.0
    end if
    if(.not.t_set.and..not.advect_density)then
        call message("T not set, initializing to 10.0 degrees celsius")
        t = 10.0
    end if
    if(.not.theta2d_set)then
        call message("THETA2D not set, using default =", rval=(/theta2d/))
    end if
    if(.not.tide_set)then
        call message("TIDE not set, using default =", lval=(/tide/))
    end if
    if(.not.tsurf_set.and..not.advect_density)then
        call message("TSURF not set, using default = 0.0")
        tsurf = 0.0
    end if
    if(.not.u_set)then
        call message("U not set, using default = 0.0")
        u=0.0
    end if
    if(.not.v_set)then
        call message("V not set, using default = 0.0")
        v = 0.0
    end if
    if(.not.w_set)then
        call message("W not set, using default = 0.0")
        w = 0.0
    end if
    if(.not.wind_forcing_set)then
        if(wind_type.ne.0)then
            wind_forcing=.true.
        else
            call message("WIND_FORCING not set, using default =", lval=(/wind_forcing/))
        end if
    end if
    if(.not.wind_set)then
        call message("WIND type not set, turning wind off")
        wind_forcing = .false.
    end if
    if(.not.z0_set)then
        call message("Z0 not set, using default =", rval=(/z0/))
    end if

! Lateral stress coefficient (see eqn. 32 in the user guide)
! by default it is 0 (no lateral friction), if set to <0 use eqn below, otherwise use given value.    
    if(cdland < 0.0) then
        cdland = max(0.0025, 0.4**2/(log(0.5*dx/z0))**2)
        call message("Given CDLAND < 0, recomputed default value=", rval=(/cdland/))
    end if
    if(cdw < 0.0) then
        cdw = max(0.0025, 0.4**2/(log(0.5*dx/z0))**2)
        call message("Given CDW < 0, recomputed default value=", rval=(/cdland/))
    end if

    
    if(master) write(*,'(/,a,/)') "====================================================="
    
    bowest(:)%edge = west
    boeast(:)%edge = east
    bonorth(:)%edge = north
    bosouth(:)%edge = south
    
    bin_west(:)%edge = west
    bin_east(:)%edge = east
    bin_north(:)%edge = north
    bin_south(:)%edge = south   
    
! then check the boundary arrays, but only check and set in active zones.

    do zi=1,num_zones_west
        call set_default_bv(u,   u_bound, bowest(zi)%inblock, bowest(zi)%u_type,  bowest(zi)%u_set , bowest(zi)%local_range, kb)
        call set_default_bv(v,   v_bound, bowest(zi)%inblock, bowest(zi)%v_type,  bowest(zi)%v_set , bowest(zi)%local_range, kb)
        call set_default_bv(w,   w_bound, bowest(zi)%inblock, bowest(zi)%w_type,  bowest(zi)%w_set , bowest(zi)%local_range, kb)
        if(advect_density)then
            call set_default_bv(rho,   rho_bound, bowest(zi)%inblock, bowest(zi)%rho_type,  bowest(zi)%rho_set, bowest(zi)%local_range, kb)
        else
            call set_default_bv(s,   s_bound, bowest(zi)%inblock, bowest(zi)%s_type,  bowest(zi)%s_set, bowest(zi)%local_range, kb)
            call set_default_bv(t,   t_bound, bowest(zi)%inblock, bowest(zi)%t_type,  bowest(zi)%t_set, bowest(zi)%local_range, kb)
        end if
        call set_default_bv(ua,  ua_bound, bowest(zi)%inblock, bowest(zi)%ua_type, bowest(zi)%ua_set , bowest(zi)%local_range, 1)
        call set_default_bv(va,  va_bound, bowest(zi)%inblock, bowest(zi)%va_type, bowest(zi)%va_set , bowest(zi)%local_range, 1)
        call set_default_bv(eta, eta_bound, bowest(zi)%inblock, bowest(zi)%eta_type,bowest(zi)%eta_set, bowest(zi)%local_range, 1)
        if(vertical_closure==1) then
            call set_default_bv(q2,  q2_bound, bowest(zi)%inblock, bowest(zi)%q2_type, bowest(zi)%q2_set , bowest(zi)%local_range, kb)
            call set_default_bv(q2l, q2l_bound, bowest(zi)%inblock, bowest(zi)%q2l_type,bowest(zi)%q2l_set, bowest(zi)%local_range, kb)
        end if
        do jj=1,maxnumconc
            if(co(jj)%allocated) call set_default_bv(co(jj)%conc, co(jj)%conc_bound, bowest(zi)%inblock, bowest(zi)%conc_type(jj), bowest(zi)%conc_set(jj), bowest(zi)%local_range, kb)
        end do
    end do
    do zi=1,num_zones_east
        call set_default_bv(u,   u_bound, boeast(zi)%inblock, boeast(zi)%u_type,  boeast(zi)%u_set, boeast(zi)%local_range, kb)
        call set_default_bv(v,   v_bound, boeast(zi)%inblock, boeast(zi)%v_type,  boeast(zi)%v_set, boeast(zi)%local_range, kb)
        call set_default_bv(w,   w_bound, boeast(zi)%inblock, boeast(zi)%w_type,  boeast(zi)%w_set, boeast(zi)%local_range, kb)
        if(advect_density)then
            call set_default_bv(rho,   rho_bound, boeast(zi)%inblock, boeast(zi)%rho_type,  boeast(zi)%rho_set, boeast(zi)%local_range, kb)
        else
            call set_default_bv(s,   s_bound, boeast(zi)%inblock, boeast(zi)%s_type,  boeast(zi)%s_set, boeast(zi)%local_range, kb)
            call set_default_bv(t,   t_bound, boeast(zi)%inblock, boeast(zi)%t_type,  boeast(zi)%t_set, boeast(zi)%local_range, kb)
        end if
        call set_default_bv(ua,  ua_bound, boeast(zi)%inblock, boeast(zi)%ua_type, boeast(zi)%ua_set , boeast(zi)%local_range, 1)
        call set_default_bv(va,  va_bound, boeast(zi)%inblock, boeast(zi)%va_type, boeast(zi)%va_set , boeast(zi)%local_range, 1)
        call set_default_bv(eta, eta_bound, boeast(zi)%inblock, boeast(zi)%eta_type,boeast(zi)%eta_set, boeast(zi)%local_range, 1)
        if(vertical_closure==1) then
            call set_default_bv(q2,  q2_bound, boeast(zi)%inblock, boeast(zi)%q2_type, boeast(zi)%q2_set , boeast(zi)%local_range, kb)
            call set_default_bv(q2l, q2l_bound, boeast(zi)%inblock, boeast(zi)%q2l_type,boeast(zi)%q2l_set, boeast(zi)%local_range, kb)
        end if
        do jj=1,maxnumconc
            if(co(jj)%allocated) call set_default_bv(co(jj)%conc, co(jj)%conc_bound, boeast(zi)%inblock, boeast(zi)%conc_type(jj), boeast(zi)%conc_set(jj), boeast(zi)%local_range, kb)
        end do
    end do
    do zi=1,num_zones_south
        call set_default_bv(u,   u_bound, bosouth(zi)%inblock, bosouth(zi)%u_type,  bosouth(zi)%u_set, bosouth(zi)%local_range, kb)
        call set_default_bv(v,   v_bound, bosouth(zi)%inblock, bosouth(zi)%v_type,  bosouth(zi)%v_set, bosouth(zi)%local_range, kb)
        call set_default_bv(w,   w_bound, bosouth(zi)%inblock, bosouth(zi)%w_type,  bosouth(zi)%w_set, bosouth(zi)%local_range, kb)
        if(advect_density)then
            call set_default_bv(rho,   rho_bound, bosouth(zi)%inblock, bosouth(zi)%rho_type,  bosouth(zi)%rho_set, bosouth(zi)%local_range, kb)
        else
            call set_default_bv(s,   s_bound, bosouth(zi)%inblock, bosouth(zi)%s_type,  bosouth(zi)%s_set, bosouth(zi)%local_range, kb)
            call set_default_bv(t,   t_bound, bosouth(zi)%inblock, bosouth(zi)%t_type,  bosouth(zi)%t_set, bosouth(zi)%local_range, kb)
        end if
        call set_default_bv(ua,  ua_bound, bosouth(zi)%inblock, bosouth(zi)%ua_type, bosouth(zi)%ua_set , bosouth(zi)%local_range, 1)
        call set_default_bv(va,  va_bound, bosouth(zi)%inblock, bosouth(zi)%va_type, bosouth(zi)%va_set , bosouth(zi)%local_range, 1)
        call set_default_bv(eta, eta_bound, bosouth(zi)%inblock, bosouth(zi)%eta_type,bosouth(zi)%eta_set, bosouth(zi)%local_range, 1)
        if(vertical_closure==1) then
            call set_default_bv(q2,  q2_bound, bosouth(zi)%inblock, bosouth(zi)%q2_type, bosouth(zi)%q2_set , bosouth(zi)%local_range, kb)
            call set_default_bv(q2l, q2l_bound, bosouth(zi)%inblock, bosouth(zi)%q2l_type,bosouth(zi)%q2l_set, bosouth(zi)%local_range, kb)
        end if
        do jj=1,maxnumconc
            if(co(jj)%allocated) call set_default_bv(co(jj)%conc, co(jj)%conc_bound, bosouth(zi)%inblock, bosouth(zi)%conc_type(jj), bosouth(zi)%conc_set(jj), bosouth(zi)%local_range, kb)
        end do
    end do
    do zi=1,num_zones_north
        call set_default_bv(u,   u_bound, bonorth(zi)%inblock, bonorth(zi)%u_type,  bonorth(zi)%u_set, bonorth(zi)%local_range, kb)
        call set_default_bv(v,   v_bound, bonorth(zi)%inblock, bonorth(zi)%v_type,  bonorth(zi)%v_set, bonorth(zi)%local_range, kb)
        call set_default_bv(w,   w_bound, bonorth(zi)%inblock, bonorth(zi)%w_type,  bonorth(zi)%w_set, bonorth(zi)%local_range, kb)
        if(advect_density)then
            call set_default_bv(rho,   rho_bound, bonorth(zi)%inblock, bonorth(zi)%rho_type,  bonorth(zi)%rho_set, bonorth(zi)%local_range, kb)
        else
            call set_default_bv(s,   s_bound, bonorth(zi)%inblock, bonorth(zi)%s_type,  bonorth(zi)%s_set, bonorth(zi)%local_range, kb)
            call set_default_bv(t,   t_bound, bonorth(zi)%inblock, bonorth(zi)%t_type,  bonorth(zi)%t_set, bonorth(zi)%local_range, kb)
        end if
        call set_default_bv(ua,  ua_bound, bonorth(zi)%inblock, bonorth(zi)%ua_type, bonorth(zi)%ua_set ,bonorth(zi)%local_range, 1)
        call set_default_bv(va,  va_bound, bonorth(zi)%inblock, bonorth(zi)%va_type, bonorth(zi)%va_set ,bonorth(zi)%local_range, 1)
        call set_default_bv(eta, eta_bound, bonorth(zi)%inblock, bonorth(zi)%eta_type,bonorth(zi)%eta_set,bonorth(zi)%local_range, 1)
        if(vertical_closure==1) then
            call set_default_bv(q2,  q2_bound, bonorth(zi)%inblock, bonorth(zi)%q2_type, bonorth(zi)%q2_set ,bonorth(zi)%local_range, kb)
            call set_default_bv(q2l, q2l_bound, bonorth(zi)%inblock, bonorth(zi)%q2l_type,bonorth(zi)%q2l_set,bonorth(zi)%local_range, kb)
        end if
        do jj=1,maxnumconc
            if(co(jj)%allocated) call set_default_bv(co(jj)%conc, co(jj)%conc_bound, bonorth(zi)%inblock, bonorth(zi)%conc_type(jj), bonorth(zi)%conc_set(jj), bonorth(zi)%local_range, kb)
        end do
    end do
  end subroutine set_defaults


  subroutine set_default_bv(f, f_bound, inblock, type, bc_set, lr, kdim)
    implicit none
    integer lr(4), kdim
    real, dimension(i1:i2,j1:j2,kdim) :: f, f_bound
    logical inblock
    integer type
    logical bc_set
    integer i,j,k
    
    if(.not. bc_set) then
        !! when we restart, xxx_bound arrays are read directly
        if(.not.restart.and.inblock)then
            do k=1,kdim
                do j=lr(3),lr(4)
                    do i=lr(1),lr(2)
                        f_bound(i,j,k) = f(i,j,k)
                    end do
                end do
            end do
        end if
        bc_set = .true.
    end if
  end subroutine set_default_bv


  subroutine change_all_bc_type(bo, type)
! For all boundary objects in bo, (re)set all bc types to type
    implicit none
    type(boundary_object) bo(:)
    integer type, i,j
    
    do i=1,size(bo)
        bo(i)%u_type = type
        bo(i)%v_type = type
        bo(i)%w_type = type
        bo(i)%s_type = type
        bo(i)%t_type = type
        bo(i)%ua_type = type
        bo(i)%va_type = type
        bo(i)%eta_type = type
        bo(i)%q2_type = type
        bo(i)%q2l_type = type
        bo(i)%rho_type = type
        do j=1,maxnumconc
            bo(i)%conc_type(j) = type
        end do
    end do    
  end subroutine change_all_bc_type
  
  subroutine dump_bc_summary(edge, bo)
! writes an array of boundary objects to file for inspection
    implicit none
    character(*) edge
    type(boundary_object) bo(:)
    integer i

    write(1000010,'(a,i0,",",i0,a,a)') "Boundary summary for MPI block ",block%ibn,block%jbn," edge=", edge
    do i=1,size(bo)
        write(1000010,'(a,i0)') "Zone=", i
        write(1000010,'(a)')           "Variable      U   V   W   S   T  UA  VA ETA  Q2 Q2L RHO"
        write(1000010,'(a,12(i2,2x))') "Type         ", bo(i)%u_type, bo(i)%v_type, bo(i)%w_type, bo(i)%s_type, &
             bo(i)%t_type, bo(i)%ua_type, bo(i)%va_type, bo(i)%eta_type, bo(i)%q2_type, bo(i)%q2l_type, bo(i)%rho_type
        write(1000010,'(a,12(l2,2x))') "Set          ", bo(i)%u_set, bo(i)%v_set, bo(i)%w_set, bo(i)%s_set, &
             bo(i)%t_set, bo(i)%ua_set, bo(i)%va_set, bo(i)%eta_set, bo(i)%q2_set, bo(i)%q2l_set, bo(i)%rho_set
        write(1000010,'(a,l2)')        "In block?    ", bo(i)%inblock
        write(1000010,'(a,4(i0,1x))')  "Global range ", bo(i)%global_range
        write(1000010,'(a,4(i0,1x))')  "Local  range ", bo(i)%local_range
    end do
  end subroutine dump_bc_summary


  subroutine set_real4(f, splitline, r)
    implicit none
    real(4) f
    character(*) splitline(:)
    integer, optional :: r(:)

    read(splitline(2),*,err=101,end=101) f
    if(master .and. debug>0) write(*,*)' value=',f
    return
101 call message('set_real4 could not read float from column 2',2)
  end subroutine set_real4


  subroutine set_real8(f, splitline, r)
    implicit none
    real(8) f
    character(*) splitline(:)
    integer, optional :: r(:)

    read(splitline(2),*,err=101,end=101) f
    if(master.and.debug>0) write(*,*)' value=',f
    return
101 call message('set_real8 could not read float from column 2',2)
  end subroutine set_real8


  subroutine set_integer4(f, splitline, r)
    implicit none
    integer(4) f
    character(*) splitline(:)
    integer, optional :: r(:)

    read(splitline(2),*,err=101,end=101) f
    if(master.and.debug>0) write(*,*)' value=',f
    return
101 call message('set_integer4 could not read integer from column 2',2)
  end subroutine set_integer4


  subroutine set_integer8(f, splitline, r)
    implicit none
    integer(8) f
    character(*) splitline(:)
    integer, optional :: r(:)

    read(splitline(2),*,err=101,end=101) f
    if(master.and.debug>0) write(*,*)' value=',f
    return
101 call message('set_integer8 could not read integer from column 2',2)
  end subroutine set_integer8


  subroutine set_logical(f, splitline, r)
    implicit none
    logical f
    character(*) splitline(:)
    integer, optional :: r(:)

    read(splitline(2),*,err=101,end=101) f
    if(master.and.debug>0) write(*,*)' value=',f
    return
101 call message('set_logical could not read logical from column 2',2)
  end subroutine set_logical

  subroutine set_string(f, splitline, r)
    implicit none
    character(*) f, splitline(:)
    integer, optional :: r(:)

    f = trim(splitline(2))
    if(master.and.debug>0) write(*,'(3a)')'  value= "',trim(f),'"'
  end subroutine set_string


! 1d arrays are assumed to be vertical or "global" arrays for now.
  subroutine set_1dinteger4(f, splitline, r)
    implicit none
    integer(4) f(:)
    character(*) splitline(:)
    integer, optional :: r(:)
    integer code, val, si1,si2

    if(size(f)==0) call message("set_1dinteger4: cannot assign to array f of 0 size",2)

    if(present(r)) then
        if(debug>0) write(*,*)' range=',r
        if(size(r)<2) call message('provided range array too short',2)
    end if

    read(splitline(2),*,err=101) code

    select case(code)
    case(0)
        read(splitline(3),*,err=102) val
        f = val
        if(master.and.debug>0) write(*,*)' 1d integer4 array value=',val
    case(1)
        read(splitline(3:),*,end=103) f
        if(master.and.debug>0) write(*,*)' 1d integer4 array value=',f
    case(2)
        call message('integer array init not implemented',2)
!        read(splitline(3),*) col
!        if(master)
!          call get_fields(splitline(4), f, cols=(/col/))
!          call broadcast(f)

    case(7)
        read(splitline(3:),*,err=105) val,si1,si2
        f(si1:si2) = val
        if(master.and.debug>0) write(*,*)' 1d real8 array set to ',val
        if(master.and.debug>0) write(*,*)' in region =',si1,si2
    case(9)
! read global array from unformatted file (binary io)
!    <name> <code> <file>
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, form="unformatted")
            read(1000006, err=106) f
            close(1000006)
            if(debug >0) then
                if(size(f) > 10)then
                    write(*,'(3a,e12.4,a,e12.4)')'  1d integer4 from binary file ',trim(splitline(3)),&
                         ', max=',maxval(f), ' min=',minval(f)
                else
                    write(*,'(a,10(i0,1x))') '  1d integer4 from binary file: ', f
                end if
            end if
        end if
        call broadcast(f)
    case(12)
! read global array from stream access unformatted file (binary io)
!    <name> <code> <file>
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, access="stream", form="unformatted")
            read(1000006, err=106) f
            close(1000006)
            if(debug >0) then
                if(size(f) > 10)then
                    write(*,'(3a,e12.4,a,e12.4)')'  1d integer4 from stream access binary file ',trim(splitline(3)),&
                         ', max=',maxval(f), ' min=',minval(f)
                else
                    write(*,'(a,10(i0,1x))') '  1d integer4 from stream access binary file: ', f
                end if
            end if
        end if
        call broadcast(f)
    case default
        if(master) write(*,*)'line=', splitline
        if(master) write(*,*)'code=', code
        call message('set_1dinteger4: unknown code on line', 2)
    end select
    return
101 call message('set_1dinteger4: could not read integer from column 2',2)
102 call message('set_1dinteger4: could not read integer from column 3',2)
103 call message('set_1dinteger4: not enough data on line',2)
105 call message('set_1d_integer4: could not read integer for val from column 3 and onward.',2)
106 call message('set_1d_integer4: could not read unformatted file, check that kind and dimensions correspond',2)
112 call message('set_1d_integer4: could not open file for unformatted io',2)
    
  end subroutine set_1dinteger4


  subroutine set_1dreal4(f, splitline, r)
    implicit none
    real(4) f(:),val
    character(*) splitline(:)
    integer, optional :: r(:)
    integer type,err,col, si1, si2

    if(size(f)==0) call message("set_1dreal4: cannot assign to array f of 0 size",2)

    if(present(r)) then
        if(debug>0.and.master) write(*,*)' range=',r
        if(size(r)<2) call message('provided range array too short',2)
    end if

    read(splitline(2),*,err=101) type

    select case(type)
    case(0)
        read(splitline(3),*,err=102) val
        f = val
        if(debug>0.and.master) write(*,*)' 1d real4 array value=',val
    case(1)
        read(splitline(3:),*,end=103) f
        if(debug>0.and.master) write(*,*)' read 1d real4 array values=',f
    case(2)
        read(splitline(4),*, err=104) col
        if(master)then
            if(restart) splitline(3) = "../" // trim(splitline(3))
            call check_file(trim(data_path)//trim(splitline(3)))
            call get_fields(trim(data_path)//trim(splitline(3)), f, cols=(/col/),ierr=err)
            if(err>0) then
                write(*,*)" file=",splitline(3),"col=",col
                call message('get_1d_fields failed reading from file',2)
            end if
        end if
        call broadcast(f)
    case(7)
        read(splitline(3:),*,err=105) val,si1,si2
        f(si1:si2) = val
        if(debug>0.and.master)then
            write(*,*)' 1d real8 array set to ',val
            write(*,*)' in region =',si1,si2
        end if
    case(9)
! read global array from unformatted file (binary io)
!    <name> <code> <file>
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, form="unformatted")
            read(1000006, err=113) f
            close(1000006)
            if(debug >0) then
                if(size(f) > 10)then
                    write(*,'(3a,e12.4,a,e12.4)')'  1d real4 from binary file ',trim(splitline(3)),&
                         ', max=',maxval(f), ' min=',minval(f)
                else
                    write(*,'(a,10(f0.1,1x))')'  1d real4 from binary file ',f
                end if
            end if
        end if
        call broadcast(f)
    case(12)
! read global array from stream access unformatted file (binary io)
!    <name> <code> <file>
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, access="stream", form="unformatted")
            read(1000006, err=113) f
            close(1000006)
            if(debug >0) then
                if(size(f) > 10)then
                    write(*,'(3a,e12.4,a,e12.4)')'  1d real4 from stream binary file ',trim(splitline(3)),&
                         ', max=',maxval(f), ' min=',minval(f)
                else
                    write(*,'(a,10(f0.1,1x))')'  1d real4 from stream binary file ',f
                end if
            end if
        end if
        call broadcast(f)
    case default
        if(master) write(*,*)' type=',type
        call message('set_1d_real4: not yet taught how to handle this type',2)
    end select
    return
101 call message('set_1dreal4: could not read integer from column 2',2)
102 call message('set_1dreal4: could not read float from column 3',2)
103 call message('set_1dreal4: could not read array directly from line, end of line reached',2)
104 call message('set_1dreal4: could not read integer col from column 3',2)
105 call message('set_1dreal4: could not read float for val and integer range from column 3 and onward.',2)
112 call message('set_1dreal4: could not open file for unformatted io',2)
113 call message('set_1dreal4: could not read unformatted file, check that kind and dimensions correspond',2)
  end subroutine set_1dreal4



  subroutine set_1dreal8(f, splitline, r)
    implicit none
    real(8) f(:),val
    character(*) splitline(:)
    integer, optional :: r(:)
    integer type,col,err, si1, si2

    err=0
    if(size(f)==0) call message("set_1real8: cannot assign to array f of 0 size",2)

    if(present(r)) then
        if(debug>0) write(*,*)' range=',r
        if(size(r)<2) call message('provided range array too short',2)
    end if

    read(splitline(2),*,err=101) type

    select case(type)
    case(0)
        read(splitline(3),*,err=102) val
        f = val
        if(debug>0.and.master) write(*,*)' 1d real8 array value=',val
    case(1)
        read(splitline(3:),*,end=103) f
        if(debug>0.and.master) write(*,*)' read 1d real8 array value=',val
    case(2)
        read(splitline(4),*, err=104) col
        if(master)then
            call check_file(trim(data_path)//trim(splitline(3)))
            call get_fields(splitline(3), f, cols=(/col/),ierr=err)
            if(err>0) then
                write(*,*)" file=",splitline(3),"col=",col
                call message('get_1d_fields failed reading from file',2)
            end if
        end if
        call broadcast(f)
    case(7)
        read(splitline(3:),*,err=105) val,si1,si2
        f(si1:si2) = val
        if(debug>0.and.master)then
            write(*,*)' 1d real8 array set to ',val
            write(*,*)' in region =',si1,si2
        end if

    case(9)
! read global array from unformatted file (binary io)
!    <name> <code> <file>
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, form="unformatted")
            read(1000006, err=113) f
            close(1000006)
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')'  1d real8 from binary file ',trim(splitline(3)),&
                 ', max=',maxval(f), ' min=',minval(f)
            if(debug >0) then
                if(size(f) > 10)then
                    write(*,'(3a,e12.4,a,e12.4)')'  1d real8 from binary file ',trim(splitline(3)),&
                         ', max=',maxval(f), ' min=',minval(f)
                else
                    write(*,'(a,10(f0.1,1x))') '  1d real8 from binary file ',f
                end if
            end if
        end if
        call broadcast(f)
    case(12)
! read global array from stream access unformatted file (binary io)
!    <name> <code> <file>
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, access="stream", form="unformatted")
            read(1000006, err=113) f
            close(1000006)
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')'  1d real8 from stream access binary file ',trim(splitline(3)),&
                 ', max=',maxval(f), ' min=',minval(f)
            if(debug >0) then
                if(size(f) > 10)then
                    write(*,'(a3,e12.4,a,e12.4)')'  1d real8 from stream access binary file ',trim(splitline(3)),&
                         ', max=',maxval(f), ' min=',minval(f)
                else
                    write(*,'(a,10(f0.1,1x))') '  1d real8 from binary file ',f
                end if
            end if
        end if
        call broadcast(f)
    case default
        if(master) write(*,*)' type=',type
        call message('set_1d_real8: not yet taught how to handle this type',2)
    end select
    return
101 call message('set_1dreal8: could not read integer from column 2',2)
102 call message('set_1dreal8: could not read float from column 3',2)
103 call message('set_1dreal8: could not read array directly from line, end of line reached',2)
104 call message('set_1dreal8: could not read integer col from column 3',2)
105 call message('set_1dreal8: could not read float for val and integer range from column 3 and onward.',2)
112 call message('set_1dreal8: could not open file for unformatted io',2)
113 call message('set_1dreal8: could not read unformatted file, check that kind and dimensions correspond',2)
  end subroutine set_1dreal8

  
  subroutine set_2dreal4(f, splitline, r)
    implicit none
    real(4) f(1-ghost_wi:,1-ghost_wj:),val
    integer type, err
    character(*) splitline(:)
    integer, optional :: r(:)
    integer si1,si2,sj1,sj2, col, li1,li2,lj1,lj2,rc(4),k
    logical inblock
    real(4), allocatable :: tmpf(:,:)
    real reals(4)
    
    if(.not.init_std_array) call check_allocation()

    if(present(r)) then
        if(debug>0.and.master) write(*,*)' range=',r
        if(size(r)<4) call message('provided range array too short',2)
        call check_range(r)
        rc=r(1:4)
    end if

    read(splitline(2),*,err=101) type

    select case(type)
    case(0)
        read(splitline(3),*,err=102) val
        if(present(r)) then
            call globalrange2local(block%ibn, block%jbn, rc, inblock)
            if(inblock) f(rc(1):rc(2),rc(3):rc(4)) = val
        else
            f = val
        end if
        if(debug>0.and.master) write(*,*)' 2d real4 array value=',val
    case(2)

! this method should only be allowed for smallish grids, as it
! requires allocation of a global array on master and then a scatter
! operation
        read(splitline(4),*,err=104) col

        if(master)then
            allocate(tmpf(im,jm), stat=err)
            if(err > 0) call message('set_2dreal4: allocation of global 2d array on master failed',2)
            call check_file(trim(data_path)//trim(splitline(3)))

            if(present(r)) then
                call get_fields(trim(data_path)//trim(splitline(3)), tmpf(r(1):r(2),r(3):r(4)), cols=(/col/), ierr=err)
            else
                call get_fields(trim(data_path)//trim(splitline(3)), tmpf, cols=(/col/), ierr=err)
            end if
            if(err > 0) call message('set_2dreal4: get_2d_fields4 failed reading 2d field',2)
        else
            allocate(tmpf(1,1))
        end if

        if(.not.init_std_array) then
            call scatterfield(tmpf, f)
        else
            f = tmpf
        end if

        deallocate(tmpf)
        if(debug>0.and.master) write(*,*)' 2d real4 array from file=',trim(splitline(3)),' column=',col
        
    case(7)
        read(splitline(3:7),*,err=103) val,reals
        call apply0000convention(reals)
        call xyz2gind(reals(1),reals(3),1.0, si1,sj1,k)
        call xyz2gind(reals(2),reals(4),1.0, si2,sj2,k)
        call check_range((/si1,si2,sj1,sj2/))
        call globalrect2local(si1,si2,sj1,sj2,li1,li2,lj1,lj2,inblock)
        if(inblock) f(li1:li2,lj1:lj2) = val
        if(debug>0.and.master)then
            write(*,*)' 2d real4 array set to ',val
            write(*,*)' in region =',si1,si2,sj1,sj2
        end if
    case(9)
! read global array from unformatted file (binary io)
!    <name> <file>
        if(master)then
            allocate(tmpf(im,jm))
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, form="unformatted")
            read(1000006, err=113) tmpf
            close(1000006)
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')'  2d real from binary file ',trim(splitline(3)),&
                 ', max=',maxval(tmpf), ' min=',minval(tmpf)
        end if

        if(.not.init_std_array) then
            call scatterfield(tmpf, f)
        else
            f = tmpf
        end if

        if(master) deallocate(tmpf)
    case(12)
! read global array from stream access unformatted file (binary io)
!    <name> <file>
        if(master)then
            allocate(tmpf(im,jm))
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, access="stream", form="unformatted")
            read(1000006, err=113) tmpf
            close(1000006)
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')'  2d real from stream access binary file ',trim(splitline(3)),&
                 ', max=',maxval(tmpf), ' min=',minval(tmpf)
        end if

        if(.not.init_std_array) then
            call scatterfield(tmpf, f)
        else
            f = tmpf
        end if

        if(master) deallocate(tmpf)
    case default
        call message('set_2dreal4 unknown type',2)
    end select
    return
101 call message('set_2d_real4: could not read integer from column 2',2)
102 call message('set_2d_real4: could not read float from column 3',2)
103 call message('set_2d_real4: could not read val,i1,i2,j1,j2 from line',2)
104 call message('set_2d_real4: could not read integer from column 4',2)
112 call message('set_2dreal4: could not open file for unformatted io',2)
113 call message('set_2dreal4: could not read unformatted file, check that kind and dimensions correspond',2)
  end subroutine set_2dreal4


  subroutine set_2dreal8(f, splitline, r)
    implicit none
    real(8) f(1-ghost_wi:,1-ghost_wj:),val
    character(*) splitline(:)
    integer, optional :: r(:)
    integer type,col,si1,si2,sj1,sj2,err,li1,li2,lj1,lj2,rc(4),k
    real(8), allocatable :: tmpf(:,:)
    real reals(4)
    logical inblock

    err=0
    if(.not.init_std_array) call check_allocation()
    read(splitline(2),*,err=101) type

    if(present(r)) then
        if(debug>0) write(*,*)' range=',r
        if(size(r)<4) call message('provided range array too short',2)
        call check_range(r)
        rc=r(1:4)
    end if

    select case(type)
    case(0)
        read(splitline(3),*,err=102) val
        if(present(r)) then
            call globalrange2local(block%ibn, block%jbn, rc, inblock)
            if(inblock) f(rc(1):rc(2),rc(3):rc(4)) = val
        else
            f = val
        end if
        if(debug>0.and.master) write(*,*)' 2d real8 array value=',val
    case(2)
        read(splitline(4),*,err=104) col
        if(master)then
            allocate(tmpf(im,jm), stat=err)
            if(err > 0) call message('set_2dreal8: allocation of global 2d array on master failed',2)
            call check_file(trim(data_path)//trim(splitline(3)))

            if(present(r)) then
                call get_fields(trim(data_path)//trim(splitline(3)), tmpf(r(1):r(2),r(3):r(4)), cols=(/col/), ierr=err)
            else
                call get_fields(trim(data_path)//trim(splitline(3)), tmpf, cols=(/col/), ierr=err)
            end if
            if(err > 0) call message('error reading 2d field',2)
        else
            allocate(tmpf(1,1))
        end if

        if(.not.init_std_array) then
            call scatterfield(tmpf, f)
        else
            f = tmpf
        end if
        deallocate(tmpf)
        if(debug>0.and.master) write(*,*)' 2d real8 array from file=',trim(splitline(3)),' column=',col
    case(7)
        read(splitline(3:7),*,err=103) val,reals
        call apply0000convention(reals)
        call xyz2gind(reals(1),reals(3),1.0, si1,sj1,k)
        call xyz2gind(reals(2),reals(4),1.0, si2,sj2,k)
        
        call check_range((/si1,si2,sj1,sj2/))
        call globalrect2local(si1,si2,sj1,sj2,li1,li2,lj1,lj2,inblock)
        if(inblock) f(li1:li2,lj1:lj2) = val
        if(debug>0.and.master) then
            write(*,*)' 2d real8 array set to ',val
            write(*,*)' in region =',si1,si2,sj1,sj2
        end if

    case(9)
! read global array from unformatted file (binary io)
!    <name> <code> <file>
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, form="unformatted")
            allocate(tmpf(im,jm))
            read(1000006, err=113) tmpf
            close(1000006)
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')'  2d real from binary file ',trim(splitline(3)),&
                 ', max=',maxval(tmpf), ' min=',minval(tmpf)
        end if

        if(.not.init_std_array)then
            call scatterfield(tmpf, f)
        else
            f = tmpf
        end if

        if(master) deallocate(tmpf)

    case(12)
! read global array from stream access unformatted file (binary io)
!    <name> <code> <file>
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, access="stream", form="unformatted")
            allocate(tmpf(im,jm))
            read(1000006, err=113) tmpf
            close(1000006)
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')'  2d real from stream access binary file ',trim(splitline(3)),&
                 ', max=',maxval(tmpf), ' min=',minval(tmpf)
        end if

        if(.not.init_std_array)then
            call scatterfield(tmpf, f)
        else
            f = tmpf
        end if

        if(master) deallocate(tmpf)

    case default
        write(*,*)' type=',type
        call message('set_2d_real8: not yet taught how to handle this type',2)
    end select
    return
101 call message('set_2d_real8: could not read integer from column 2',2)
102 call message('set_2d_real8: could not read float from column 3',2)
103 call message('set_2d_real8: could not read val,i1,i2,j1,j2 from line',2)
104 call message('set_2d_real8: could not read integer from column 4',2)
112 call message('set_2dreal8: could not open file for unformatted io',2)
113 call message('set_2dreal8: could not read unformatted file, check that kind and dimensions correspond',2)
  end subroutine set_2dreal8


  subroutine set_3dreal4(f, splitline, r)
    implicit none
    real(4) f(1-ghost_wi:,1-ghost_wj:,:),val
    character(*) splitline(:)
    integer, optional :: r(:)
    integer type, is,ie,js,je,si1,si2,sj1,sj2,sk1,sk2, error, rc(6),li1,li2,lj1,lj2,k
    real(4) fmax,fmin,deltaz,zsymm,tmpreal
    logical inblock
    real(4), allocatable :: tmpf(:,:,:), tmpf2(:,:,:)
    real(4) z1,z2, invalid
    real reals(4)
    integer numlevs
    real(4), allocatable :: zlevs(:)
    
    call check_allocation()
    read(splitline(2),*,err=101) type

    if(present(r)) then
        if(debug>0.and.master) write(*,*)' range=',r
        if(size(r)<6) call message('provided range array to set_3dreal4 to short',2)
        call check_range(r)
        rc = r
        call globalrange2local(block%ibn, block%jbn, rc, inblock)
    end if

    select case(type)
    case(0)
! assign scalar <value> to the whole array.
!    <keyword> <code> <value>
        read(splitline(3),*,err=102) val

        if(present(r)) then
            if(inblock) f(rc(1):rc(2),rc(3):rc(4),rc(5):rc(6)) = val
        else
            f = val
        end if
        if(debug>0.and.master) write(*,*)' 3d real4 array value=',val

    case(1)
! read directly from rest of line, not likely, but if thats whats asked for...
!    <keyword> <code> <f(1,1,1)>,...,<f(im,jm,kb)>
        call message('set_3dreal4 read 3d array from setup line not implemented',2)
        read(splitline(3:),*,err=109,end=109) f

    case(2)
! read from spreadsheet "file", column "col";
!    <keyword> <code> <file> <col>
! if needed, extend this with range for data in file, and/or range in array to assign to.

        read(splitline(4),*,err=110,end=110) si1
        if(master)then
            allocate(tmpf(im,jm,kb), stat=error)
            if(error > 0) call message('set_3dreal4: allocation of global 3d array on master failed',2)
            call check_file(trim(data_path)//trim(splitline(3)))

            if(present(r)) then
                call get_fields(trim(data_path)//trim(splitline(3)), tmpf(r(1):r(2),r(3):r(4),r(5):r(6)), cols=(/si1/), ierr=error)
            else
                call get_fields(trim(data_path)//trim(splitline(3)), tmpf, cols=(/si1/), ierr=error)
            end if
            if(error.ne.0) call message('set_3d_real4 failed calling get_3d_fields',2)
        else
            allocate(tmpf(1,1,1))
        end if
        call scatterfield(tmpf, f)
        deallocate(tmpf)
        if(debug>0.and.master) write(*,*)' 3d real4 array from file=',trim(splitline(3)),' column=',si1

    case(3)
!    <keyword> <code> <file> <col1> <col2> <range>
        read(splitline(4:5),*,err=106,end=106) si1, si2
        if(si1<1 .or.si1>100) call message('col1 is not in [1,100]; spreadsheets with more than 100 columns are rare...',2)
        if(si2<1 .or.si2>100) call message('col2 is not in [1,100]; spreadsheets with more than 100 columns are rare...',2)
        if(.not.present(r)) then
            read(splitline(6:9),*,err=107,end=107) reals
            rc(1:4) = true2globrange(reals)
            call check_range(rc(1:4))
            call vprof_intpol(f, trim(data_path)//trim(splitline(3)), cols=(/si1,si2/), region=rc(1:4))            
        else
            call vprof_intpol(f, trim(data_path)//trim(splitline(3)), cols=(/si1,si2/), region=r(1:4))
        end if
    case(4)
!     <keyword> <code> <file> <weighting> <region>
! the WEIGHTING argument, by default -1.0 (inverse distance squared),
! is the exponent of the square of the distance between the current
! point and a interpolated profile value, used in the weighting. -0.5
! gives 1/distance as weight, -0.25 gives 1/sqrt(distance) as weight. read code for
! further info.
        read(splitline(4),*) tmpreal
        read(splitline(5:8),*,err=111) reals
        call apply0000convention(reals)        
        call xyz2gind(reals(1),reals(3),1.0, si1,sj1,k)
        call xyz2gind(reals(2),reals(4),1.0, si2,sj2,k)
        call check_range((/si1,si2,sj1,sj2/))
        call multivprof_intpol(f, splitline(3), region=(/si1,si2,sj1,sj2/), weighting=tmpreal)

    case(5)
!        <keyword> <code> ... horisontal interpolation from scattered points
        call message('this init type is only available for 2d arrays',2)
    case(6)
! horisontal extrapolation from tanh profile
!        <keyword> <code> <fmax> <fmin> <deltaz> <zsymm>
        read(splitline(3:),*,end=103,err=103) fmax,fmin,deltaz,zsymm
        if(present(r)) then
            call tanhprofile4(f,fmax,fmin,deltaz,zsymm,r)
        else
            call tanhprofile4(f,fmax,fmin,deltaz,zsymm)
        end if
    case(7)
! assign value to array region and specific sigma layers
!        <keyword> <code> <value> <x1> <x2> <y1> <y2> <k1> <k2>
        read(splitline(3:9),*,err=105) val,reals,sk1,sk2
        call apply0000convention(reals)        
        call xyz2gind(reals(1),reals(3),1.0, si1,sj1,k)
        call xyz2gind(reals(2),reals(4),1.0, si2,sj2,k)       
        call check_range((/si1,si2,sj1,sj2,sk1,sk2/))
        call globalrect2local(si1,si2,sj1,sj2,li1,li2,lj1,lj2,inblock)
        if(inblock) f(li1:li2,lj1:lj2,sk1:sk2) = val

        if(debug>0.and.master)then
            write(*,*)' 3d real4 array set to ',val
            write(*,*)' in region =',si1,si2,sj1,sj2,sk1,sk2
        end if
    case(8)
! horisontal extrapolation from exp profile
!        <keyword> <code> <fmax> <fmin> <deltaz>
        read(splitline(3:),*,end=104,err=104) fmax,fmin,deltaz
        if(present(r)) then
            call check_range(r)
            call expprofile4(f,fmax,fmin,deltaz,r)
        else
            call expprofile4(f,fmax,fmin,deltaz)
        end if
    case(9)
! read global array directly from unformatted file (binary io)
!    <keyword> <code> <file>
        allocate(tmpf(im,jm,kb))
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, form="unformatted")
            read(1000006, err=113) tmpf
            close(1000006)
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')' 3d real4 from binary file ',trim(splitline(3)),&
                 ' max=',maxval(tmpf),' min=',minval(tmpf)
        end if
        call scatterfield(tmpf, f)
        deallocate(tmpf)
    case(10)
! assign value to a horisontal array region, and from depth z1 to z2.
!    <keyword> <code> <value> <x1> <x2> <y1> <y2> <Z1> <Z2>
! z1,z2 must be specified as negative depths, i.e. a correct spec is from z1=-10 to z2=-5, so that z2>z1.
        read(splitline(3:7),*,err=105) val,reals
        call apply0000convention(reals)
        read(splitline(8:9),*,err=114) z1, z2
        if(z1 > z2 .or. (z1>0 .and. z2>0)) call message("z2 must be larger than z1, both must be negative",2)
        call xyz2gind(reals(1),reals(3),1.0, si1,sj1,k)
        call xyz2gind(reals(2),reals(4),1.0, si2,sj2,k)
        call check_range((/si1,si2,sj1,sj2/))
        call globalrect2local(si1,si2,sj1,sj2,li1,li2,lj1,lj2,inblock)

        if(inblock) call assign2zlayer4(val, z1, z2, (/li1,li2,lj1,lj2/), f)

        if(debug>0.and.master)then
            write(*,*)' 3d real4 array set to ',val
            write(*,*)' in region =',si1,si2,sj1,sj2, 'from depth z1=',z1,' to depth z2=',z2
        end if
        
    case(11)
! read global Z coordinate array from single precision float, unformatted
! stream files (binary io), interpolate to sigma levels
!    <keyword> <code> <file> <numlevels> <level file>

        read(splitline(4),*) numlevs
        allocate(zlevs(numlevs))

        if(master)then
            open(unit=1000006, file=(trim(data_path)//trim(splitline(5))), status="old", err=115, form="unformatted", access="stream")
            read(1000006) zlevs
            close(1000006)
            if(debug > 1) write(*,'(1x,a,30(f0.01,1x))') "3d z levels:",zlevs
        end if
        call broadcast(zlevs)
        
        allocate(tmpf(im,jm,numlevs))
        if(master)then
            open(unit=1000006, file=(trim(data_path)//trim(splitline(3))), status="old", form="unformatted", access="stream")
            read(1000006, err=113) tmpf
            close(1000006)
            
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')' 3d z coordinate real4 field from binary file ',trim(splitline(3)),&
                 ' max=',maxval(tmpf),' min=',minval(tmpf)            
        end if
        
        allocate(tmpf2(i1:i2,j1:j2,numlevs))
        
        call scatterfield(tmpf, tmpf2)

        invalid = -32767.00
        call z2sigma(1, f, tmpf2, zlevs, numlevs, invalid)
        
        deallocate(tmpf, tmpf2, zlevs)
    case(12)
! read global array directly from stream access unformatted file (binary io)
!    <keyword> <code> <file>
        allocate(tmpf(im,jm,kb))
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, access="stream", form="unformatted")
            read(1000006, err=113) tmpf
            close(1000006)
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')' 3d real4 from stream access binary file ',trim(splitline(3)),&
                 ' max=',maxval(tmpf),' min=',minval(tmpf)
        end if
        call scatterfield(tmpf, f)
        deallocate(tmpf)
        
    case default
        write(*,*)' type=',type
        call message('set_3d_real4: not yet taught how to handle this type',2)
    end select
    return
101 call message('set_3dreal4: could not read integer from column 2',2)
102 call message('set_3dreal4: could not read float from column 3',2)
103 call message('set_3dreal4: tanhprofile failed reading fmax,fmin,deltaz,zsymm from column 3 to 6',2)
104 call message('set_3dreal4: expprofile failed reading fmax,fmin,deltaz from column 3 to 5',2)
105 call message('set_3dreal4: could not read float value for f and 6 numbers for range from column 3:',2)
106 call message('set_3dreal4: could not get 2 integers for col1,col2 from column 4,5',2)
107 call message('set_3dreal4: could not get 4 numbers for range from column 6-9',2)
108 call message('set_3dreal4: failed to read all values from input file line',2)
109 call message('set_3dreal4: failed reading all values directly from line',2)
110 call message('set_3dreal4: failed reading integer for spreadsheet column from column 3',2)
111 call message('set_3dreal4: could not get 4 numbers for range from column 5-8',2)
112 call message('set_3dreal4: could not open file for unformatted io',2)
113 call message('set_3dreal4: could not read unformatted file, check that kind and dimensions correspond',2)
114 call message('set_3dreal4: could not read 2 floats for z1,z2 from column 8,9',2)
115 call message('set_3dreal4: could not open 3d z file',2)
  end subroutine set_3dreal4



  subroutine set_3dreal8(f, splitline, r)
    implicit none
    real(8) f(1-ghost_wi:,1-ghost_wj:,:), val
    character(*) splitline(:)
    integer, optional :: r(:)
    integer type,si1,si2,sj1,sj2,sk1,sk2, error, rc(6),li1,li2,lj1,lj2,k
    integer is,ie,js,je
    real(8) fmax,fmin,deltaz,zsymm
    real(4) tmpreal
    logical inblock
    real(8), allocatable :: tmpf(:,:,:), tmpf2(:,:,:)
    real(4), allocatable :: tmpf_real4(:,:,:)
    real(8) z1,z2,invalid
    real reals(4)
    integer numlevs, lr(4)
    real(8), allocatable :: zlevs(:)
    
    call check_allocation()
    read(splitline(2),*,err=101) type

    if(present(r)) then
        if(debug>0.and.master) write(*,*)' range=',r
        if(size(r)<6) call message('provided range array to set_3dreal8 to short',2)
        call check_range(r)
        rc = r
        call globalrange2local(block%ibn, block%jbn, rc, inblock)
    end if

    select case(type)

    case(0)
! assign scalar <value> to the whole array.
!    <keyword> <code> <value>
        read(splitline(3),*,err=102) val

        if(present(r)) then
            if(inblock) f(rc(1):rc(2),rc(3):rc(4),rc(5):rc(6)) = val
        else
            f = val
        end if
        if(debug>0.and.master) write(*,*)' 3d real8 array value=',val

    case(1)
! read directly from rest of line, not likely, but if thats whats asked for...
!    <keyword> <code> <f(1,1,1)>,...,<f(im,jm,kb)>
        call message('set_3dreal8 read 3d array from setup line not implemented',2)
!        allocate(tmpf(im,jm,kb)
!        read(splitline(3:),*,err=109,end=109) tmpf
!        call scatter(tmpf, f, im,jm,kb)

    case(2)
! read from spreadsheet "file", column "col";
!    <keyword> <code> <file> <col>
! if needed, extend this with range for data in file, and/or range in array to assign to.

        read(splitline(4),*,err=110,end=110) si1
        if(master)then
            allocate(tmpf(im,jm,kb), stat=error)
            if(error > 0) call message('set_3dreal8: allocation of global 3d array on master failed',2)

            call check_file(trim(data_path)//trim(splitline(3)))

            if(present(r)) then
                call get_fields(trim(data_path)//trim(splitline(3)), tmpf(r(1):r(2),r(3):r(4),r(5):r(6)), cols=(/si1/), ierr=error)
            else
                call get_fields(trim(data_path)//trim(splitline(3)), tmpf, cols=(/si1/), ierr=error)
            end if
            if(error.ne.0) call message('set_3d_real8 failed calling get_3d_fields8',2)
        else
            allocate(tmpf(1,1,1))
        end if
        call scatterfield(tmpf, f)
        deallocate(tmpf)
        if(debug>0.and.master) write(*,*)' 3d real8 array from file=',trim(splitline(3)),' column=',si1
    case(3)
!    <keyword> <code> <file> <col1> <col2> <range>
        read(splitline(4:5),*,err=106,end=106) si1, si2
        if(si1<1 .or.si1>100) call message('col1 is not in [1,100]; spreadsheets with more than 100 columns are rare...',2)
        if(si2<1 .or.si2>100) call message('col2 is not in [1,100]; spreadsheets with more than 100 columns are rare...',2)
        if(.not.present(r))then
            read(splitline(6:9),*,err=107,end=107) reals
            rc(1:4) = true2globrange(reals)
            call check_range(rc(1:4))
            lr = rc(1:4)
            call globalrange2local(block%ibn, block%jbn, lr, inblock)            
            allocate(tmpf_real4(i1:i2,j1:j2,kb))
            call vprof_intpol(tmpf_real4, trim(data_path)//trim(splitline(3)), cols=(/si1,si2/), region=rc(1:4))
            if(inblock) f(lr(1):lr(2),lr(3):lr(4),:) = tmpf_real4(lr(1):lr(2),lr(3):lr(4),:)
            deallocate(tmpf_real4)
        else
            allocate(tmpf_real4(i1:i2,j1:j2,kb))
            call vprof_intpol(tmpf_real4, trim(data_path)//trim(splitline(3)), cols=(/si1,si2/), region=r(1:4))
            if(inblock) f(rc(1):rc(2),rc(3):rc(4),:) = tmpf_real4(rc(1):rc(2),rc(3):rc(4),:)
            deallocate(tmpf_real4)
        end if
    case(4)
!     <keyword> <code> <file> <weighting> <region>
! the WEIGHTING argument, by default -1.0 (inverse distance squared),
! is the exponent of the square of the distance between the current
! point and a interpolated profile value, used in the weighting. -0.5
! gives 1/distance as weight, -0.25 gives 1/sqrt(distance) as weight. read code for
! further info.
        read(splitline(4),*) tmpreal
        read(splitline(5:8),*,err=111) reals
        call apply0000convention(reals)
        call xyz2gind(reals(1),reals(3),1.0, si1,sj1,k)
        call xyz2gind(reals(2),reals(4),1.0, si2,sj2,k)
        
        call check_range((/si1,si2,sj1,sj2/))
        call globalrect2local(si1,si2,sj1,sj2,li1,li2,lj1,lj2,inblock)

! dp is necessary mostly in calculations that require large precision,
! it is rare to store dp numbers in ascii, so we read as real(kind=4)
! and copy into f:
        allocate(tmpf_real4(i1:i2,j1:j2,kb))
        call multivprof_intpol(tmpf_real4, splitline(3), region=(/si1,si2,sj1,sj2/), weighting=tmpreal)
        if(inblock) f(li1:li2,lj1:lj2,:) = tmpf_real4(li1:li2,lj1:lj2,:)
        deallocate(tmpf_real4)
    case(5)
!        <keyword> <code> ... horisontal interpolation from scattered points
        call message('this init type is only available for 2d arrays',2)
    case(6)
! horisontal extrapolation from tanh profile
!        <keyword> <code> <fmax> <fmin> <deltaz> <zsymm>
        read(splitline(3:),*,end=103,err=103) fmax,fmin,deltaz,zsymm
        if(present(r)) then
            call tanhprofile8(f,fmax,fmin,deltaz,zsymm,r)
        else
            call tanhprofile8(f,fmax,fmin,deltaz,zsymm)
        end if
    case(7)
! assign value to array region and specific sigma layers
!        <keyword> <code> <value> <x1> <x2> <y1> <y2> <k1> <k2>
        read(splitline(3:9),*,err=105) val,reals,sk1,sk2
        call apply0000convention(reals)        
        call xyz2gind(reals(1),reals(3),1.0, si1,sj1,k)
        call xyz2gind(reals(2),reals(4),1.0, si2,sj2,k)       
        call check_range((/si1,si2,sj1,sj2,sk1,sk2/))
        call globalrect2local(si1,si2,sj1,sj2,li1,li2,lj1,lj2,inblock)

        if(inblock) f(li1:li2,lj1:lj2,sk1:sk2) = val

        if(debug>0.and.master)then
            write(*,*)' 3d real8 array set to ',val
            write(*,*)' in region =',si1,si2,sj1,sj2,sk1,sk2
        end if
    case(8)
! horisontal extrapolation from exp profile
!        <keyword> <code> <fmax> <fmin> <deltaz>
        read(splitline(3:),*,end=104,err=104) fmax,fmin,deltaz
        if(present(r)) then
            call check_range(r)
            call expprofile8(f,fmax,fmin,deltaz,r)
        else
            call expprofile8(f,fmax,fmin,deltaz)
        end if
    case(9)
! read global array directly from unformatted file (binary io)
!    <keyword> <code> <file>
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, form="unformatted")
            allocate(tmpf(im,jm,kb))
            read(1000006, err=113) tmpf
            close(1000006)
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')' 3d real8 from binary file ',trim(splitline(3)),&
                 ' max=',maxval(tmpf),' min=',minval(tmpf)
        end if
        call scatterfield(tmpf, f)
        if(master) deallocate(tmpf)
    case(10)
! assign value to a horisontal array region, and from depth z1 to z2.
!    <keyword> <code> <value> <x1> <x2> <y1> <y2> <Z1> <Z2>
! z1,z2 must be specified as negative depths, i.e. a correct spec is from z1=-10 to z2=-5, so that z2>z1.
        read(splitline(3:7),*,err=105) val,reals
        call apply0000convention(reals)
        read(splitline(8:9),*,err=114) z1, z2
        if(z1 > z2 .or. (z1>0 .and. z2>0)) call message("z2 must be larger than z1, both must be negative",2)

        call xyz2gind(reals(1),reals(3),1.0, si1,sj1,k)
        call xyz2gind(reals(2),reals(4),1.0, si2,sj2,k)
        call check_range((/si1,si2,sj1,sj2/))
        call globalrect2local(si1,si2,sj1,sj2,li1,li2,lj1,lj2,inblock)

        if(inblock) call assign2zlayer8(val, z1, z2, (/li1,li2,lj1,lj2/), f)

        if(debug>0.and.master)then
            write(*,*)' 3d real8 array set to ',val
            write(*,*)' in region =',si1,si2,sj1,sj2, 'from depth z1=',z1,' to depth z2=',z2
        end if

    case(11)
! read global Z coordinate array from double precision float, unformatted
! stream files (binary io), interpolate to sigma levels
!    <keyword> <code> <file> <numlevels> <level file>

        read(splitline(4),*) numlevs
        allocate(zlevs(numlevs))

        if(master)then
            open(unit=1000006, file=(trim(data_path)//trim(splitline(5))), status="old", err=115, form="unformatted", access="stream")
            read(1000006) zlevs
            close(1000006)
        end if
        call broadcast(zlevs)
        
        allocate(tmpf(im,jm,numlevs))
        if(master)then
            allocate(tmpf_real4(im,jm,numlevs))
            
            open(unit=1000006, file=(trim(data_path)//trim(splitline(3))), status="old", form="unformatted", access="stream")
            read(1000006, err=113) tmpf_real4
            close(1000006)
            
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')' 3d z coordinate real4 field from binary file ',trim(splitline(3)),&
                 ' max=',maxval(tmpf_real4),' min=',minval(tmpf_real4)
            
            tmpf = tmpf_real4
            
            deallocate(tmpf_real4)
        end if
        
        allocate(tmpf2(i1:i2,j1:j2,numlevs))
        
        call scatterfield(tmpf, tmpf2)

        invalid = -32767._8
        call z2sigma(1, f, tmpf2, zlevs, numlevs, invalid)
        
        deallocate(tmpf, tmpf2, zlevs)
        
    case(12)
! read global array directly from stream access unformatted file (binary io)
!    <keyword> <code> <file>
        if(master)then
            open(unit=1000006, file=(trim(data_path)//splitline(3)), status="old", err=112, access="stream", form="unformatted")
            allocate(tmpf(im,jm,kb))
            read(1000006, err=113) tmpf
            close(1000006)
            if(debug >0) write(*,'(3a,e12.4,a,e12.4)')' 3d real8 from stream access binary file ',trim(splitline(3)),&
                 ' max=',maxval(tmpf),' min=',minval(tmpf)
        end if
        call scatterfield(tmpf, f)
        if(master) deallocate(tmpf)

    case default
        write(*,*)' type=',type
        call message('set_3d_real8: not yet taught how to handle this type',2)
    end select
    return
101 call message('set_3dreal8: could not read integer from column 2',2)
102 call message('set_3dreal8: could not read float from column 3',2)
103 call message('set_3dreal8: failed reading fmax,fmin,deltaz,zsymm from column 3 to 6',2)
104 call message('set_3dreal8: expprofile failed reading fmax,fmin,deltaz from column 3 to 5',2)
105 call message('set_3dreal8: could not read float value for f and 6 numbers for range from column 3:',2)
106 call message('set_3dreal8: could not get 2 integers for col1,col2 from column 4,5',2)
107 call message('set_3dreal8: could not get 4 numbers for range from column 6-9',2)
110 call message('set_3dreal8: failed reading integer for spreadsheet column from column 3',2)
111 call message('set_3dreal8: could not get 4 numbers for range from column 5-8',2)
112 call message('set_3dreal8: could not open file for unformatted io',2)
113 call message('set_3dreal8: could not read unformatted file, check that kind and dimensions correspond',2)
114 call message('set_3dreal8: could not read 2 floats for z1,z2 from column 8,9',2)
115 call message('set_3dreal8: could not open 3d z file',2)
  end subroutine set_3dreal8


  subroutine assign2zlayer4(val, z1, z2, r, f)
!
! Assign value VAL to a horisontal array region R, and from depth z1
! to z2. z1,z2 must be specified as negative depths, i.e. a correct
! spec is from z1=-10 to z2=-5, so that z2>z1.
!
    implicit none
    real(4) val, f(1-ghost_wi:,1-ghost_wj:,:), z1, z2, dep
    integer r(4),i,j,k

    do k = 1,kb-1
        do j = r(3),r(4)
            do i = r(1),r(2)
                if (gridbox(i,j)==1) then
                    dep = zz(k)*h(i,j)
                    if (dep > z1 .and. dep < z2) then
                        f(i,j,k) = val
                    endif
                endif
            end do
        end do
    end do
  end subroutine assign2zlayer4

  
  subroutine assign2zlayer8(val, z1, z2, r, f)
!
! Assign value VAL to a horisontal array region R, and from depth z1
! to z2. z1,z2 must be specified as negative depths, i.e. a correct
! spec is from z1=-10 to z2=-5, so that z2>z1.
!
    implicit none
    real(8) val, f(1-ghost_wi:,1-ghost_wj:,:), z1, z2, dep
    integer r(4),i,j,k

    do k = 1,kb-1
        do j = r(3),r(4)
            do i = r(1),r(2)
                if (gridbox(i,j)==1) then
                    dep = zz(k)*h(i,j)
                    if (dep > z1 .and. dep < z2) then
                        f(i,j,k) = val
                    endif
                endif
            end do
        end do
    end do
  end subroutine assign2zlayer8


  subroutine ggrid(x, r)
!
! use a geometrical series to generate a grid x, where r is a
! parameter for the stretching; 
!    r == 1   uniform grid
!    r < 1    dense near x(1)
!    r > 1    dense near x(size(x))
!      
    implicit none
    integer j,n
    real  x(:), r

    n = size(x)
  
    if (r == 1.) then
        do j = 1, n
            x(j) = real(j-1)/real(n-1)
        end do
    else
        do j = 1, n
            x(j) = (r**(j-1) - 1.) / (r**(n-1) - 1.)
        end do
    end if
  end subroutine ggrid
  

  subroutine initialize_sigma_layers(splitline)
! compute Z from the given splitline
    implicit none
    character(*) splitline(:)
    integer type,k,err,col
    real dsbot, deltaeps, pi, aa, eps, stretching
    real(4) tmpz(1000)

    call check_allocation()

    read(splitline(2),*) type

    select case(type)
    case(0)
        if(debug>0.and.master) write(*,*)' Uniform distribution of sigma layers'
        do k = 1,kb
            z(k) = -(k-1)/real(kb-1)
        end do

    case(1)
        read(splitline(3),*) dsbot
        if(dsbot>1.0 .or. dsbot<0.0) call message('dsbot is >1 or <0',2)

        if(debug>0.and.master) then
            write(*,*)' Initialize sigma layers from Lynch distribution'
            write(*,*)'  dsbot=', dsbot
        end if
!
!   Here we are using the algorithm due to Lynch et.al,1995.
!   Convergence Studies of Tidally-Rectified Circulation on Georges Bank.
!   In: Quantitative Skill Assessment for Coastal Ocean Models,
!   Lynch,D.R. and Davies,A.M. (editors), American Geophysical Union.
!
        DELTAEPS = 1.0/(KB-1.)
        PI = 4.*ATAN(1.0)
        AA = (DELTAEPS - DSBOT)/SIN(2.*PI*DELTAEPS)
        DO K = 1,KB
            EPS = (K-1.)/(KB-1.)
            Z(KB + 1 - K) = -1. + EPS - AA*SIN(2*PI*EPS)
        END DO

    case(2)
        col=1
        read(splitline(4),*,err=101,end=101) col

        if(master)then
            if(debug>0) write(*,*)' Reading sigma layers from file=',trim(splitline(3)),' column=',col
            call get_fields(trim(data_path)//trim(splitline(3)), tmpz(1:kb), cols=(/col/), ierr=err)
            if(err.ne.0) call message('failed reading sigma layers',2)
            z(1:kb) = tmpz(1:kb)
        end if

        call broadcast(z)

    case(3)
        read(splitline(3),*) stretching
        if(master.and.debug>0) write(*,*) "stretching r=", stretching
        if(stretching < 0.0 .or. stretching > 4.0) call message("stretching parameter r for ggrid looks wrong",2)
        
        call ggrid(z, stretching)
        z = -z
        
    case default
        if(master)write(*,*)' type code=', type
        call message('the type code for initializing sigma layers is unknown',2)
    end select

!! We can now define layer thicknesses

    call depth(kb,z,dz,dzr,dzz,zz)

    return
    101 call message('could not get column number for sigma z from column 4',2)
  end subroutine initialize_sigma_layers



  subroutine expprofile4(f, fmax, fmin, deltaz, r)
    implicit none
    integer, parameter :: localkind=4
    include "expprofile.f90"
  end subroutine

  subroutine expprofile8(f, fmax, fmin, deltaz, r)
    implicit none
    integer, parameter :: localkind=8
    include "expprofile.f90"
  end subroutine


  subroutine tanhprofile4(f, fmax, fmin, deltaz, zsymm, r)
!
!  Initialize a 3d array using a vertical tanh profile
!
    implicit none
    integer, parameter :: localkind=4
    include "tanhprofile.f90"
  end subroutine

  subroutine tanhprofile8(f, fmax, fmin, deltaz, zsymm, r)
!
!  Initialize a 3d array using a vertical tanh profile
!
    implicit none
    integer, parameter :: localkind=8
    include "tanhprofile.f90"
  end subroutine


  subroutine initialize_boundary_zones(depth, label, numzones, start, end)
!
! loops along each edge to define boundary zone ranges
!   call extract_boundary_zones('west', numzones_west, west_start, west_end)
!   ...

    use state
    implicit none
    integer numzones, numzones2
    character(len=*) label
    real depth(:,:)
    integer start(:), end(:), length, ind, i,j
    logical prev_cell_land, cur_cell_water

    select case(label)
    case('west','east')
        if(periodic_ew) return
        length=jm
    case('north','south')
        length=im
        if(jm == 1)then
            numzones = 0
            return
        end if
    case default
        write(*,*)'extract bnd zone, should not happen'
        stop
    end select

    if(master)then

!! first look for start index
    numzones = 0
    prev_cell_land = .true.

    do ind=1,length
        select case(label)
        case('west')
            i=1
            j=ind
        case('east')
            i=im
            j=ind
        case('north')
            i=ind
            j=jm
        case('south')
            i=ind
            j=1
        end select

        cur_cell_water = .false.
        if(depth(i,j) > 0.0) cur_cell_water = .true.

        if(prev_cell_land .and. cur_cell_water) then
            numzones = numzones + 1
            if(numzones > max_zones) call message("numzones > max_zones, something looks fishy w.r. bathymetry!",2)
            start(numzones) = ind
        end if
        prev_cell_land = .not. cur_cell_water
    end do

!! then look for end index
    numzones2 = 0
    prev_cell_land = .true.

    do ind=1,length
        select case(label)
        case('west')
            i=1
            j=ind
        case('east')
            i=im
            j=ind
        case('north')
            i=ind
            j=jm
        case('south')
            i=ind
            j=1
        end select

        cur_cell_water=.false.
        if(depth(i,j) > 0.0) cur_cell_water=.true.

!! if the previous cell was water, and the current is land
!! the previous cell was the end of a boundary zone
        if((.not.prev_cell_land) .and. (.not.cur_cell_water)) then
            numzones2 = numzones2 + 1
            end(numzones2) = ind-1
        end if
! special case if the whole edge is open
        if(ind==length .and. cur_cell_water) then
            numzones2 = numzones2 + 1
            end(numzones2) = ind
        end if

        prev_cell_land = .not. cur_cell_water
    end do

    if(numzones .ne. numzones2) then
        write(*,*)'extract bnd zones error: numzones mismatch n1=',numzones,'n2',numzones2,'side=',label
        write(*,*)'start=', start
        write(*,*)'end=', end
        stop
    end if

    if(debug >0) then
        do ind=1,numzones
            write(*,*)' open boundary ',label,' start=',start(ind),' end=',end(ind)
        end do
    end if

    end if  !! master

    call broadcast(numzones)
    call broadcast(start)
    call broadcast(end)
  end subroutine initialize_boundary_zones



  subroutine adjust_boundary_zones(gh)
!
! in the case of expanding inlets: the current zone sizes are
! calculated from the size of the inlets at the edge, not the size in
! the interiour. this routine tries to expand frs zones to cover
! inlets also in the interior.
!
    real gh(:,:)
    integer zone,is,ie,js,je,i,j
    logical :: north_tweaked=.false., west_tweaked=.false.

    if(master) then
! west
        do zone=1,num_zones_west

            !! zone corners
            is=1
            ie=lb
            js=bowest(zone)%global_range(3)
            je=bowest(zone)%global_range(4)

            !!move start southward from i,j if possible
            i=ie ; j=js
            do
                if(zone > 1)then
                    if(j-1 <= bowest(zone-1)%global_range(4)) exit
                end if
                if(j-1 < 1) exit
                if(gh(i,j-1) > 0.00001) then
                    !! wet
                    j=j-1
                    bowest(zone)%global_range(3)=j
                    west_tweaked=.true.
                else
                    exit
                end if
            end do

            !!move zone end "northward" from i,j if possible
            i=ie ; j=je
            do
                if(zone < num_zones_west)then
                    if(j+1 >= bowest(zone+1)%global_range(3)) exit
                end if
                if(j+1 > jm)exit
                if(gh(i,j+1) > 0.00001) then
                    !! wet
                    j=j+1
                    bowest(zone)%global_range(4)=j
                    west_tweaked=.true.
                else
                    exit
                end if
            end do

        end do
! north
        do zone=1,num_zones_north
            !! global zone corners
            is=bonorth(zone)%global_range(1)
            ie=bonorth(zone)%global_range(2)
            js=jm-lb+1
            je=jm

            !! move beginning westward
            i=is ; j=js
            do
                if(zone > 1)then
                    if(i-1 <= bonorth(zone-1)%global_range(2)) exit
                end if
                if(i-1 < 1) exit
                if(gh(i-1,j) > 0.00001) then
                    i=i-1
                    bonorth(zone)%global_range(1)=i
                    north_tweaked=.true.
                else
                    exit
                end if
            end do

            !! move end eastward
            i=ie ; j=js
            do
                if(zone < num_zones_north)then
                    if(i+1 >= bonorth(zone+1)%global_range(1)) exit
                end if
                if(i+1 > im) exit
                if(gh(i+1,j) > 0.00001) then
                    i=i+1
                    bonorth(zone)%global_range(2)=i
                    north_tweaked=.true.
                else
                    exit
                end if
            end do
! etc. for east and south
        end do

        do zone=1,num_zones_east
            if(master) write(*,*)' Warning, if you have expanding inlets to the east&
                 & , the frs zones have not been adjusted to cover these fully'
        end do
        do zone=1,num_zones_south
            if(master) write(*,*)' Warning, if you have expanding inlets to the south&
                 &, the frs zones have not been adjusted to cover these fully'
        end do

        if(debug>0.and.north_tweaked.or.west_tweaked)then
            write(*,*)' adjust_boundary_zones has done some tweaking:'
            call print_zones("west", num_zones_west, bowest)
            call print_zones("east", num_zones_east, boeast)
            call print_zones("south", num_zones_south, bosouth)
            call print_zones("north", num_zones_north, bonorth)
            write(*,*)' if this looks crazy, turn off the automatic adjustment by setting'
            write(*,*)' ADJUST_BZ to false, before H is set.'
        end if
    end if

    do zone=1,num_zones_west
        call broadcast(bowest(zone)%global_range)
    end do
    do zone=1,num_zones_north
        call broadcast(bonorth(zone)%global_range)
    end do
  contains

    subroutine print_zones(name, n, bo)
      implicit none
      character(*) name
      type(boundary_object) bo(:)
      integer n, ind
      if(n>0) then
        do ind=1,n
            write(*,'(4x,3a,4(i0,1x))')'open boundary ',name,' range=',bo(ind)%global_range
        end do
      end if
    end subroutine print_zones

  end subroutine adjust_boundary_zones



  subroutine check_b_zone_connections(sw_connected, nw_connected, ne_connected, se_connected)
!
! Figure out if boundary zones are connected or overlapping in the
! corners.  if connected, north and south zones that overlap with
! east/west, are shrunk horisontally to fit, in order to cover the
! same ranges as the original relax.f90 for "all sides".
!
    logical :: sw_connected, nw_connected, ne_connected, se_connected
    logical :: bz_tweaked

    bz_tweaked=.false.

! first the simple cases
    if(num_zones_west==0)then
        sw_connected=.false.
        nw_connected=.false.
    end if
    if(num_zones_east==0)then
        ne_connected=.false.
        se_connected=.false.
    end if
    if(num_zones_north==0)then
        ne_connected=.false.
        nw_connected=.false.
    end if
    if(num_zones_south==0)then
        se_connected=.false.
        sw_connected=.false.
    end if

    if(num_zones_west.ne.0 .and. num_zones_north.ne.0) then
        if(bowest(num_zones_west)%global_range(4) == jm .and. bonorth(1)%global_range(1)==1) then
            nw_connected = .true.
            if(bonorth(1)%global_range(2) < lb+1) then
                if(master) write(*,*)'WARNING north zone1 ends inside western frs, not good.'
            else
                if(master) write(*,*)'  INFO: Start of north zone1 moved to ',lb+1
                bonorth(1)%global_range(1)=lb+1
                bz_tweaked=.true.
            end if
        else if(bowest(num_zones_west)%global_range(4) >= jm-lb .and. bonorth(1)%global_range(1)<=lb) then
            if(master)then
            write(*,*)'error: crash between boundary zones in north west corner, not connected, but overlap.'
            write(*,*)'end of last zone west=',bowest(num_zones_west)%global_range(4),&
                 ' beginning of first zone north=',bonorth(1)%global_range(1)
            write(*,*)'please redesign open boundaries'
            end if
            call abort()
        end if
    end if
    if(num_zones_east.ne.0 .and. num_zones_north.ne.0) then
        if(boeast(num_zones_east)%global_range(4) == jm .and. bonorth(num_zones_east)%global_range(2)==im) then
            ne_connected = .true.
            if(bonorth(num_zones_north)%global_range(1) >= im-lb+1)then
                if(master)write(*,*)'WARNING last north zone begins inside eastern frs, not good.'
            else
                if(master)write(*,*)'  INFO: end of last north zone moved to ',im-lb
                bonorth(num_zones_north)%global_range(2) = im-lb
                bz_tweaked=.true.
            end if
        else if(boeast(num_zones_east)%global_range(4) >= jm-lb+1 .and. bonorth(num_zones_north)%global_range(2) >= im-lb+1) then
            if(master)then
            write(*,*)'error: crash between boundary zones in north east corner, not connected, but overlap.'
            write(*,*)'end of last zone east=',boeast(num_zones_east)%global_range(4),&
                 ' end of last zone north=',bonorth(num_zones_north)%global_range(2)
            write(*,*)'please redesign open boundaries'
            end if
            call abort()
        end if
    end if
    if(num_zones_east.ne.0 .and. num_zones_south.ne.0) then
        if(boeast(1)%global_range(3) == 1 .and. bosouth(num_zones_south)%global_range(2)==im) then
            se_connected = .true.
            if(bosouth(num_zones_south)%global_range(1) >= im-lb+1)then
                if(master)write(*,*)'WARNING last south zone begins inside eastern frs, not good.'
            else
                if(master)write(*,*)'   INFO: end of last south zone moved to ',im-lb
                bosouth(num_zones_south)%global_range(2) = im-lb
                bz_tweaked=.true.
            end if

        else if(boeast(1)%global_range(3) <= lb .and. bosouth(num_zones_south)%global_range(2) >= im-lb+1) then
            if(master)then
            write(*,*)'error: crash between boundary zones in south east corner, not connected, but overlap.'
            write(*,*)'beginning of first zone east=',boeast(1)%global_range(3),&
                 ' end of last zone south=',bosouth(num_zones_south)%global_range(2)
            write(*,*)'please redesign open boundaries'
            end if
            call abort()
        end if
    end if
    if(num_zones_west.ne.0 .and. num_zones_south.ne.0) then
        if(bowest(1)%global_range(3) == 1 .and. bosouth(1)%global_range(1)==1) then
            se_connected = .true.
            if(bosouth(1)%global_range(2) < lb+1) then
                if(master)write(*,*)'WARNING south zone1 ends inside western frs, not good.'
            else
                if(master)write(*,*)'   INFO: Start of south zone1 moved to ',lb+1
                bosouth(1)%global_range(1)=lb+1
                bz_tweaked=.true.
            end if
        else if(bowest(1)%global_range(3) <= lb .and. bosouth(1)%global_range(1) <= lb) then
            if(master)then
            write(*,*)'error: crash between boundary zones in south west corner, not connected, but overlap.'
            write(*,*)'beginning of first zone west=',bowest(1)%global_range(3),&
                 ' beginning of first zone south=',bosouth(1)%global_range(1)
            write(*,*)'please redesign open boundaries'
            end if
            call abort()
        end if
    end if

    if(debug>0.and.bz_tweaked.and.master)then
        write(*,*)' NCALFA modified boundary zones:'
        call print_zones("west", num_zones_west, bowest)
        call print_zones("east", num_zones_east, boeast)
        call print_zones("south", num_zones_south, bosouth)
        call print_zones("north", num_zones_north, bonorth)
        if(nw_connected) write(*,*) "  NW corner zones are connected"
        if(ne_connected) write(*,*) "  NE corner zones are connected"
        if(sw_connected) write(*,*) "  SW corner zones are connected"
        if(se_connected) write(*,*) "  SE corner zones are connected"
    end if

  contains

    subroutine print_zones(name, n, bo)
      implicit none
      character(*) name
      type(boundary_object) bo(:)
      integer n, ind
      if(n>0) then
        do ind=1,n
            write(*,'(a,a,a,4(i0,1x))')' open boundary ',name,' range=',bo(ind)%global_range
        end do
      end if
    end subroutine print_zones

  end subroutine check_b_zone_connections


  subroutine parallelize_boundary_zones()
!
! on each processor, convert the global coordinates for boundary
! zones into "local coordinates"
!
    implicit none
    integer i, is,ie,js,je,r(4)

    do i=1,num_zones_west
        r = bowest(i)%global_range
        call globalrect2local(r(1),r(2),r(3),r(4),is,ie,js,je,bowest(i)%inblock)
        bowest(i)%local_range = (/is,ie,js,je/)
    end do

    do i=1,num_zones_east
        r = boeast(i)%global_range
        call globalrect2local(r(1),r(2),r(3),r(4),is,ie,js,je,boeast(i)%inblock)
        boeast(i)%local_range = (/is,ie,js,je/)
    end do

    do i=1,num_zones_north
        r = bonorth(i)%global_range
        call globalrect2local(r(1),r(2),r(3),r(4),is,ie,js,je,bonorth(i)%inblock)
        bonorth(i)%local_range = (/is,ie,js,je/)
    end do

    do i=1,num_zones_south
        r = bosouth(i)%global_range
        call globalrect2local(r(1),r(2),r(3),r(4),is,ie,js,je,bosouth(i)%inblock)
        bosouth(i)%local_range = (/is,ie,js,je/)
    end do
  end subroutine parallelize_boundary_zones


  subroutine set_concentration(id, splitline)
    implicit none
    integer id    
    character(*) splitline(:)
          
    call allocate_concentration(id)
    write(co(id)%name,'(a,i0)') "conc",id
    call set_var(co(id)%conc, splitline)
  end subroutine set_concentration
  
  
  subroutine set_conc_ah(splitline)
! conc_ah <concentration id number> <data for initialization of 3d array>
    implicit none
    character(*) splitline(:)
    integer id

    read(splitline(2),*) id
    
    if(id <1 .or. id > MAXNUMCONC) then
        call message("ERROR: Concentration id is out of range, ignoring keyword!",1)
        return
    end if
    
    if(.not. co(id)%ALLOCATED) then
        call allocate_concentration(id)
        call message("allocating concentration, conc id=",1, ival=(/id/))
    end if

    allocate(co(id)%ah(i1:i2,j1:j2,kb))
    call message("concentration AH allocated, conc id=", ival=(/id/))

    splitline(2) = "conc" // trim(splitline(2)) // "%AH"
    call set_var(co(id)%ah, splitline(2:))
  end subroutine set_conc_ah
  
  subroutine set_conc_kh(splitline)
! conc_kh <concentration id number> <data for initialization of 3d array>
    implicit none
    character(*) splitline(:)
    integer id
    read(splitline(2),*) id
!    if(id <1 .or. id > MAXNUMCONC) then
    if(id <1 .or. id > 1) then        
        call message("ERROR: Concentration id is out of supported range, ignoring keyword!",1)
        return
    end if
    if(.not. co(id)%ALLOCATED) then
        call allocate_concentration(id)
        call message("allocating concentration, conc id=",1, ival=(/id/), margin=2)
    end if
    if(.not. associated(co(id)%kh)) then
        allocate(co(id)%kh(i1:i2,j1:j2,kb))
        call message("concentration KH allocated, conc id=", ival=(/id/), margin=2)
    end if
    ! the 2d background field
    if(.not.associated(sponge_kh_c1)) then
        allocate(sponge_kh_c1(i1:i2,j1:j2))
        sponge_kh_c1 = khmin
        sponge_set = .true.  !   (to trigger the shapiro filter)
    end if
    splitline(2) = "conc" // trim(splitline(2)) // "%KH"
    call set_var(sponge_kh_c1, splitline(2:))
  end subroutine set_conc_kh
  

  subroutine add_concentration_time(splitline)
    implicit none
    character(*) splitline(:)
    integer id    
    integer conctime(6)

    read(splitline(2),*) id
    if(id <1 .or. id > MAXNUMCONC) then
        call message("ERROR: Concentration id is out of range",2)
    else if(.not. co(id)%ALLOCATED) then
        call message("setting activation time for concentration that is not (yet?) defined",1,margin=2)
    end if

    if(index(splitline(3),"T") > 0) then
        call get_iso8601(splitline(3), conctime)
        co(id)%activation_time = timediff(initial_time, conctime)
    else
        read(splitline(3),*) co(id)%activation_time
    end if

    if( co(id)%activation_time-time <= 0.5) then
        call message("Concentration activation time in seconds:", rval=(/co(id)%activation_time/), margin=2)
        call message("Initial model time in seconds:", rval8=(/time/), margin=2)
        call message("The concentration will be activated at initial model time", margin=2)
    else if(co(id)%activation_time < 0.0 .or. co(id)%activation_time > 10000000) then
        call message("concentration activation time in seconds (since initial NDATE) t=", rval=(/co(id)%activation_time/), margin=2)
        call message("concentration activation time looks wrong, check that NDATE and the requested concentration activation time make sense",2)
    end if
    
    if(debug > 0 .and. master) write(*,'(2x,a,i0,a,f0.1,a,a)') "Concentration ",id," will be activated at t=", co(id)%activation_time
  end subroutine add_concentration_time

  
  subroutine add_binreg(edge, binreg, splitline)
!
! Get binreg_<edge> <zone> <levels> <x1> <x2> <y1> <y2>
!
    implicit none
    character(*) edge
    type(binreg_object) binreg(:)
    character(*) splitline(:)
    integer zone
    real reals(4)

    read(splitline(2),*,err=101,end=101) zone
    if(zone < 0 .or. zone > max_zones) call message("zone number is out of allowed range",2)
    
    read(splitline(3),*,err=103) binreg(zone)%levels
    if(binreg(zone)%levels > maxnumbin) call message("too many bin levels, increase maxnumbin in mod_bound if you need this",2)
    
    read(splitline(4:7),*,err=102,end=102) reals
    
! maybe later?    
!     binreg(zone)%range = nint(reals/(/dx,dx,dy,dy/)+1.5)
    binreg(zone)%range = nint(reals)

    call allocate_binreg()
    
    if(debug>0.and.master) write(*,*) "binreg, edge=",edge," zone=",zone," range=",binreg(zone)%range

    return
101 call message("could not read binreg zone from column 2",2)
102 call message("could not read 4 numbers for binreg region from col 4-7",2)
103 call message("could not read integer for the number of binreg levels from col 3",2)
  end subroutine add_binreg


  subroutine lag_forcing(splitline)
! lag_forcing <forcing id>    
    implicit none
    character(*) splitline(:)
    integer forcing_id

    read(splitline(2),*,end=1,err=1) forcing_id    
    if(forcing_id < 1 .or. forcing_id > numforcings) call message("invalid forcing id",2)
    forcing(forcing_id)%lag = .true.
    if(debug>0.and.master) write(*,*) "  id=",forcing_id
    return
1   call message("could not read integer for forcing id from column 2",2)
  end subroutine lag_forcing


  subroutine add_river(splitline)
! parse a line with the river keyword:
!   river <name> <x> <y> <code> <depth> <format> <data>
    implicit none
    character(*) splitline(:)
    integer k,i,li,lj,gi,gj
    logical inblock
    real dep,x,y,glon,glat
    integer linenumbers(20000), forcing_time(6), numwords, prevnumwords
    character(200) forcing_lines(20000), forcing_splitline(6)
    character(30) tmpstring
    logical wordwarning, convgeo, reloc
    
    if(.not. h_set .or. .not. z_set) then
        call message('Too early to specify rivers, specify z (or sigma_layers) and h first!',2)
    end if

    if(rivlog_set .and. .not. rivlog) return
    
    numc = numc + 1
    if(numc > maxnumc) then
        if(master)write(*,*)' numc=', numc
        call message('Too many rivers! you should increase MAXNUMC in mod_bound.f90 and recompile.',2)
    end if

!  INTEGER, DIMENSION(:),ALLOCATABLE :: IRIVER,JRIVER,KRIVER1,KRIVER2,CODERIVER
!  REAL, DIMENSION(:),ALLOCATABLE :: SRIVER,TRIVER,FRIVER,DZRIVER
!
!      name           x       y     code    depth  format   friver   triver  sriver
!river "Banten river" 20      3       2       5.0    0        20.98    29.0    25.0

    river_name(numc) = splitline(2)

    convgeo=.false.
    if(index(lower(splitline(3)),"lon") > 0) then
        tmpstring = splitline(3)
        read(tmpstring(4:),*,err=101,end=101) glon
        convgeo=.true.
    else
        read(splitline(3),*,err=101,end=101) x
! first guess for cell position
        x = x/dx + 1.
        iriver(numc) = nint(x + 0.5)
    end if
    
    if(index(lower(splitline(4)),"lat") > 0) then
        tmpstring = splitline(4)
        read(tmpstring(4:),*,err=102,end=102) glat
        convgeo=.true.
    else
        read(splitline(4),*,err=102,end=102) y
        y = y/dy + 1.
        jriver(numc) = nint(y + 0.5)
    end if

    if(convgeo) then
        call sph2gr(glon, glat, (xpole+dx)/dx, (ypole+dy)/dy, dx*0.001, dy*0.001, ylong, x, y, ierr)
        call message("  got longitude,latitude:",rval=(/glon,glat/))
        call message("  converted to grid position:",rval=(/x,y/))

        iriver(numc) = nint(x + 0.5)
        jriver(numc) = nint(y + 0.5)
    end if

    read(splitline(5),*,err=103,end=103) coderiver(numc)
    read(splitline(6),*,err=104,end=104) depriver(numc)
    read(splitline(7),*,err=105,end=105) riverformat(numc)

    if(coderiver(numc) < 1 .or. coderiver(numc) > 4) call message("invalid river code, please check column 4",2)

    if(master)then
        reloc=.false.
        gi=iriver(numc) ; gj=jriver(numc)
        call snapriver2coast(code=coderiver(numc), x=x, y=y, gi=iriver(numc), gj=jriver(numc))
        if(gi.ne.iriver(numc) .or. gj.ne.jriver(numc)) reloc=.true.
    end if        
    call broadcast(reloc)
    if(reloc) call mbroadcast(iriver(numc), jriver(numc))

    
! Define outlet in the upper approx depriver(numc) meters
    
    kriver1(numc) = 1
    kriver2(numc) = kb-1

    li = iriver(numc)  ;   lj = jriver(numc)
    call globalind2local(li,lj,inblock,ghost=.true.)
    
    if(inblock)then
        do k = 1,kb-1
            dep = -z(k+1)*h(li,lj)
            if (depriver(numc) .lt. dep) then
                kriver2(numc) = k
                exit
            endif
        end do
        
! Compute the fraction of the vertical covered by the river outlet
        dzriver(numc) = -z(kriver2(numc)+1)+z(kriver1(numc))
    end if


    select case(riverformat(numc))
    case(0)
        read(splitline(8),*,err=106,end=106) friver(numc)
        if(advect_density)then
            ! river density read as true value. subtract 1000+rhoaver later.
            read(splitline(9),*,err=112,end=112) sriver(numc)
        else
            read(splitline(9),*,err=107,end=107) triver(numc)
            read(splitline(10),*,err=108,end=108) sriver(numc)
        end if
        if(debug > 0 .and. master) then
            write(*,'(2x,a,": i=",i0," j=",i0," code=",i0," depth=",f0.1," Flux=",f0.1," T=",f0.1, " S=",f0.1)') &
                 trim(river_name(numc)), iriver(numc), jriver(numc), coderiver(numc), depriver(numc), &
                 friver(numc), triver(numc), sriver(numc)
        end if
    case(1)
        if(advect_density) call message("time dependent density rivers TODO",2)
        numforcings = numforcings + 1
        if(numforcings > max_numforcings) &
             call message("add_river: too many bc's with input from file, increase max_numforcings",2)

        if(master) write(*,'(3x,a,i0)') "Forcing id=", numforcings
        
        forcing(numforcings)%file_name = (trim(data_path)//splitline(8))

        call replace_yyyy_with_year(forcing(numforcings)%file_name, ndate)
        
        forcing(numforcings)%id = numc           
        forcing(numforcings)%type = 4
        forcing(numforcings)%name = trim(river_name(numc))//"(river)"

        if(debug > 0 .and. master)then
            write(*,'(2x,a,": i=",i0," j=",i0," code=",i0," depth=",f0.1," Forcing file=",a)') trim(river_name(numc)), &
                 iriver(numc), jriver(numc), coderiver(numc), depriver(numc), forcing(numforcings)%file_name
        end if
        
        if(master) call get_lines(forcing(numforcings)%file_name, forcing(numforcings)%numfields, forcing_lines, linenumbers)
        call broadcast(forcing(numforcings)%numfields)
        call broadcast(forcing_lines)
        call broadcast(linenumbers)

        allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))

        prevnumwords = 0 ; numwords = 0 ; wordwarning=.false.
        
        do i=1,forcing(numforcings)%numfields
            prevnumwords = numwords
            call split(forcing_lines(i), forcing_splitline, numwords)
            
            if(i > 1 .and. numwords .ne. prevnumwords) wordwarning = .true.
            
            ! if there is a t in the forcing time string, it must be in iso format
            if(index(forcing_splitline(1),"T") > 0) then
                call get_iso8601(forcing_splitline(1), forcing_time)
                forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
            else
                read(forcing_splitline(1),*,err=110) forcing(numforcings)%field(i)%time
            end if

            if(numwords < 4) call message("error in forcing file, found < 3 reals on a line",2)
            read(forcing_splitline(2:4),*,err=109) &
                 forcing(numforcings)%field(i)%data, &
                 forcing(numforcings)%field(i)%data2,&
                 forcing(numforcings)%field(i)%data3
            
! then a chicken and egg problem. the add_river_conc below must be
! called after add_river because it is optional and uses the river
! name.  but right here is the spot where the time dependent river
! conc should be read, but since it is defined later, how to decide?
! first attempt is to always read data if they are in the file. the
! cost is that we have to count the number of reals on each line. in
! apply forcing there should be a test to see if a certain river conc
! is time dependent.            
            if(numwords > 4+maxnumconc) numwords = 4+maxnumconc
            if(numwords > 4) then
                read(forcing_splitline(5:numwords),*,err=111) forcing(numforcings)%field(i)%conc(1:numwords-4)
            end if
        end do

        if(debug > 0 .and. master)then
            write(*,*) " got",forcing(numforcings)%numfields,"forcing times with ",numwords," columns"
            i=forcing(numforcings)%numfields
            write(*,*) " first time=",forcing(numforcings)%field(1)%time," last time=",forcing(numforcings)%field(i)%time
        end if
        
        if(wordwarning) call message("the number of reals for concentration was not the same on all lines",1)
        
    case default
        call message('unknown river format',2)
    end select

    return

101 call message('could not read integer iriver from column 3',2)
102 call message('could not read integer jriver from column 4',2)
103 call message('could not read integer coderiver from column 5',2)
104 call message('could not read float value for depriver from column 6',2)
105 call message('could not read integer "river format" from column 7',2)
106 call message('could not read float value for friver from column 8',2)
107 call message('could not read float value for triver from column 9',2)
108 call message('could not read float value for sriver from column 10',2)
109 if(master) write(*,*) "river forcing file=", splitline(8)," line=", linenumbers(i)
    if(master) write(*,*) "i=",i," line=", trim(forcing_lines(i))
    call message('failed to read river forcing data: time flux s t, from above line',2)
110 if(master) write(*,*) "forcing file=",forcing(numforcings)%file_name,"line number=",linenumbers(i)
    if(master) write(*,*) "forcing line=",trim(forcing_lines(i))
    call message("could not read real value for time from column 1 at the given line in the forcing file",2)
111 if(master) write(*,*) "numwords=",numwords,"forcing line=",trim(forcing_lines(i))
    call message("failed reading numwords reals for river concentration from column 5 and out on this line",2)
112 call message("could not read float value for river density from column 9",2)
113 call message('could not get longitude from column 3',2)
114 call message('could not get latitude from column 4',2)
    
  end subroutine add_river


  subroutine override_river_position(splitline)
! river_cell <name> <iriver> <jriver>    
    implicit none
    character(*) splitline(:)
    integer i

    do i=1,numc
        if(splitline(2) == river_name(i))then
            read(splitline(3),*,err=101,end=101) iriver(i)
            read(splitline(4),*,err=102,end=102) jriver(i)
            if(master .and. debug > 0) &
                 write(*,'(a,a,a,i0,",",i0)') "  river ",trim(river_name(i))," manually relocated to cell ",iriver(i),jriver(i)
            return
        end if
    end do
101 call message("failed to read iriver from column 3",2)       
102 call message("failed to read jriver from column 4",2)       
  end subroutine override_river_position


  subroutine override_river_geometry(splitline)
! river_geometry <name> <iriver> <jriver> <kriver1> <kriver2> <coderiver>
    implicit none
    character(*) splitline(:)
    integer i

    do i=1,numc
        if(splitline(2) == river_name(i))then
            read(splitline(3),*,err=101,end=101) iriver(i)
            read(splitline(4),*,err=102,end=102) jriver(i)
            read(splitline(5),*,err=103,end=103) kriver1(i)
            read(splitline(6),*,err=104,end=104) kriver2(i)
            read(splitline(7),*,err=105,end=105) coderiver(i)
            
            if(master .and. debug > 0) then
                write(*,'(a,a,a,i0)') "  river ",trim(river_name(i)), " river number=",i
                write(*,'(a,i0,",",i0)') " IRIVER,JRIVER=",iriver(i),jriver(i)
                write(*,'(a,i0,",",i0)') " KRIVER1,KRIVER2=",kriver1(i),kriver2(i)
                write(*,'(a,i0)') " CODERIVER=",coderiver(i)
                write(*,'(a,f0.5)') " DZRIVER=", dzriver(i)
            end if

            if(kriver2(i)>kb-1 .or. kriver1(i) > kb-1) &
                 call message("hard coded kriver out of range, KB,KRIV1,2=",2,ival=(/kb,kriver1(i),kriver2(i)/))
            
            dzriver(i) = -z(kriver2(i)+1) + z(kriver1(i))            
            return
        end if
    end do
101 call message("failed to read iriver from column 3",2)       
102 call message("failed to read jriver from column 4",2)       
103 call message("failed to read kriver1 from column 5",2)       
104 call message("failed to read kriver2 from column 6",2)       
105 call message("failed to read coderiver from column 7",2)       
  end subroutine override_river_geometry
  

  subroutine submerge_river(splitline)
!   river_submerge <name> <depth>
    implicit none
    character(*) splitline(:)
    integer i,k,li,lj
    real depth,dep
    logical inblock
    
    do i=1,numc
        if(splitline(2) == river_name(i))then
            ! i is the river
            read(splitline(3),*,err=1,end=1) depth
            if(depth > 0) call message("the z level of the top of the dived river inlet must be < 0",2)
            if(depth < -depriver(i)) call message("top of dived river inlet cannot be below bottom of inlet",2)
            
            kriver1(i) = 1
            li = iriver(numc)  ;   lj = jriver(numc)
            call globalind2local(li,lj,inblock,ghost=.false.)
    
            if(inblock)then
                do k = 1,kb-1
                    dep = z(k+1)*h(li,lj)
                    if (dep < depth) then
                        kriver1(i) = k
                        exit
                    endif
                end do
        
               ! recompute the fraction of the vertical covered by the river outlet
                dzriver(i) = -z(kriver2(i)+1)+z(kriver1(i))
                if(debug > 0) then
                    write(*,'(2x,a,f0.1)') "target top depth of dived river=",depth
                    write(*,'(2x,a,i0,1x,i0,a,f0.2)') "recomputed k1,k2=",kriver1(i),kriver2(i)," dzriver=",dzriver(i)
                end if
            end if
            return
        end if
    end do
2   call message("could not find a river with the given name",2)    
1   call message("could not read real for dived depth from column 3",2)
  end subroutine submerge_river
  

  subroutine set_river_conc(splitline)
! river_conc <string with river name> <integer value for conc field> [<real value for concentration>]
    implicit none
    character(*) splitline(:)
    integer i,j,c

    do i=1,numc
        if(splitline(2) == river_name(i))then
            ! i is the river
            read(splitline(3),*,err=101,end=101) c
            if(c < 1 .or. c > maxnumconc) call message("invalid integer for concentration id",2)
           
            select case(riverformat(i))
            case(0)
                read(splitline(4),*, err=102, end=102) criver(c,i)
                if(criver(c,i) < 0 .or. criver(c,i) > 1.0) call message("invalid concentration",2)
                if(master .and. debug > 0) &
                     write(*,'(2x,a,i0,a,f0.2,a,a)') "set river conc",c," to ",criver(c,i)," for river ",trim(river_name(i))
            case(1)
! which forcing has river number i
                do j=1,numforcings
                    if(forcing(j)%type == 4 .and. forcing(j)%id == i) then
                        forcing(j)%timedepconc(c) = .true.
                        if(master .and. debug > 0) &
                             write(*,'(2x,a,i0,a,a)') "set river conc",c," as time dependent for river ",trim(river_name(i))
                    end if
                end do
            end select           
            return
        end if
    end do
    call message("invalid river name, make sure river is defined before adding concentration",2)
101 call message("failed to read integer for concentration from column 3",2)
102 call message("failed to read real value for constant river concentration from column 4",2)
  end subroutine set_river_conc

  
  subroutine set_well_conc(splitline)
! well_conc <string with well name> <integer value for conc field> [<real value for concentration>]
    implicit none
    character(*) splitline(:)
    integer i,j,c

    do i=1,numwell
        if(splitline(2) == well_name(i))then
            read(splitline(3),*,err=101,end=101) c
            if(c < 1 .or. c > maxnumconc) call message("invalid integer for concentration id",2)
           
            select case(well_format(i))
            case(0)
                read(splitline(4),*, err=102, end=102) cwell(c,i)
                if(master .and. debug > 0) &
                     write(*,'(2x,a,i0,a,f0.2,a,a)') "set well conc",c," to ",cwell(c,i)," for well ",trim(well_name(i))
            case(1)
! which forcing has well number i
                do j=1,numforcings
                    if(forcing(j)%type == 5 .and. forcing(j)%id == i) then
                        forcing(j)%timedepconc(c) = .true.
                        if(master .and. debug > 0) &
                             write(*,'(2x,a,i0,a,a)') "set well conc",c," as time dependent for well ",trim(well_name(i))
                    end if
                end do                
            end select
            return
        end if
    end do
    call message("invalid well name, make sure well is defined before specifying concentration",2)
101 call message("failed to read integer for concentration from column 3",2)
102 call message("failed to read real value for constant well concentration from column 4",2)
  end subroutine set_well_conc


  subroutine snapriver2coast(code, x, y, gi, gj)
!    
! river location x,y may not always be in a wet box with a land box on
! the side of the river outlet, as given by code. code=1 left, code=2
! south, etc.
!
! we try to have river outlet positions given in geogr coordinates
! with 3 decimals. 1deg=60minutes~100km, so 0.001 degrees is roughly
! 100m, with dx minimum 10m we should search at least 10 boxes wide.
!
! this should be done on the global bathymetry
!    
    implicit none
    real x,y
    integer code, gi, gj, initialgi,initialgj,i,j
    integer :: del
    real :: uncertainty=14, dist, mindist
    
    initialgi=gi ; initialgj=gj

    mindist = (2*uncertainty)**2
    dist = mindist
    del = nint(uncertainty)

    if(jm > 1) then
    
        do j=max(initialgj-del,2),min(initialgj+del,jm-1)
            do i=max(initialgi-del,2),min(initialgi+del,im-1)
                select case(code)
                case(1)
! if code is 1, we only compare with boxes that have coast on the left edge
                    if(gh(i,j)>0..and.gh(i-1,j)<=0.)  dist = rdist(i-1.,j-0.5,x,y)
                case(2)
                    if(gh(i,j)>0..and.gh(i,j-1)<=0.)  dist = rdist(i-0.5,j-1.,x,y)
                case(3)
                    if(gh(i,j)>0..and.gh(i+1,j)<=0.)  dist = rdist(i*1.,j-0.5,x,y)
                case(4)
                    if(gh(i,j)>0..and.gh(i,j+1)<=0.)  dist = rdist(i-0.5,j*1.,x,y)
                end select
!            write(*,'(6(i0," "),5(f0.1,1x))') i,j,initialgi,initialgj,dist,mindist,gh(i,j),gh(i-1,j),gh(i,j-1),gh(i+1,j),gh(i,j+1)
                
                if(dist < mindist)then
                    gi=i
                    gj=j
                    mindist = dist
                end if
            end do
        end do
    else
        j = 1
        do i=max(initialgi-del,2),min(initialgi+del,im-1)
            select case(code)
            case(1)
                if(gh(i,j)>0..and.gh(i-1,j)<=0.)  dist = rdist(i-1.,0.,x,y)
            case(3)
                if(gh(i,j)>0..and.gh(i+1,j)<=0.)  dist = rdist(i*1.,0.,x,y)
            end select
!            write(*,'(6(i0," "),5(f0.1,1x))') i,j,initialgi,initialgj,dist,mindist,gh(i,j),gh(i-1,j),gh(i,j-1),gh(i+1,j),gh(i,j+1)

            if(dist < mindist)then
                gi=i
                gj=j
                mindist = dist
            end if
        end do        
    end if

    if(mindist == (2*uncertainty)**2) then
        if(debug>0) write(*,*) "initial cell for river=",initialgi,initialgj
        call message("did not find coastal cell. use keyword RIVER_CELL to manually position river in grid",1)
    else
        if(debug>0 .and. (gi.ne.initialgi .or. gj.ne.initialgj)) then
            write(*,'(2x,a,i0,",",i0,a,i0,",",i0)') "River with provided position in cell ",initialgi,initialgj,&
                 " was relocated to coastal cell ",gi,gj
        end if
    end if
    
  contains
    
    real function rdist(x1,y1,x2,y2)
      implicit none
      real x1,y1,x2,y2
      rdist = (x2-x1)**2 + (y2-y1)**2
    end function rdist    
  end subroutine snapriver2coast


  subroutine add_well(splitline)
    implicit none
    character(*) splitline(:)
    integer i
    real x,y,glon,glat
    integer linenumbers(50), forcing_time(6), numwords, prevnumwords
    character(200) forcing_lines(50), forcing_splitline(10),tmpstring
    logical wordwarning, convgeo

    if(.not. h_set) then
        call message('Too early to specify well, set h first!',2)
    end if

    numwell = numwell + 1
    if(numwell > maxnumwell) then
        if(master)write(*,*)'numwell=', numwell
        call message('Too many wells! you should increase MAXNUMWELL in mod_bound.f90 and recompile.',2)
    end if

!         name     iwell,jwell   type     fwell    twell   swell
!  well "Banten"   20,3           0       20.98    29.0    25.0

    well_name(numwell) = splitline(2)

    convgeo=.false.
    if(index(lower(splitline(3)),"lon") > 0) then
        tmpstring = splitline(3)
        read(tmpstring(4:),*,err=101,end=101) glon
        convgeo=.true.
    else
        read(splitline(3),*,err=101,end=101) x
      ! first guess for cell position
        iwell(numwell) = nint(x/dx + 1.5)
    end if
    
    if(index(lower(splitline(4)),"lat") > 0) then
        tmpstring = splitline(4)
        read(tmpstring(4:),*,err=102,end=102) glat
        convgeo=.true.
    else
        read(splitline(4),*,err=102,end=102) y
        jwell(numwell) = nint(y/dy + 1.5)
    end if

    if(convgeo) then
        call sph2gr(glon, glat, (xpole+dx)/dx, (ypole+dy)/dy, dx*0.001, dy*0.001, ylong, x, y, ierr)
        call message("  got longitude,latitude:",rval=(/glon,glat/))
        call message("  converted to grid position:",rval=(/x,y/))

        iwell(numwell) = nint(x + 0.5)
        jwell(numwell) = nint(y + 0.5)
    end if
    
    read(splitline(5),*,err=106) well_format(numwell)

    select case(well_format(numwell))
    case(0)
        read(splitline(6),*,err=103) fwell(numwell)
        read(splitline(7),*,err=104) twell(numwell)
        read(splitline(8),*,err=105) swell(numwell)

! cwell
        
    case(1)
        numforcings = numforcings + 1
        if(numforcings > max_numforcings) &
             call message("add_well: too many bc's with input from file, increase max_numforcings",2)

        if(master) write(*,'(3x,a,i0)') "Forcing id=", numforcings
        
        forcing(numforcings)%file_name = (trim(data_path)//splitline(6))

        forcing(numforcings)%type = 5
        forcing(numforcings)%id = numwell
        forcing(numforcings)%name = splitline(1)
        
        if(master) call get_lines(forcing(numforcings)%file_name, forcing(numforcings)%numfields, forcing_lines, linenumbers)
        call broadcast(forcing(numforcings)%numfields)
        call broadcast(forcing_lines)
        call broadcast(linenumbers)

        allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))

        prevnumwords = 0 ; numwords = 0 ; wordwarning = .false.
        
        do i=1,forcing(numforcings)%numfields
            prevnumwords = numwords
            call split(forcing_lines(i), forcing_splitline, numwords)
            if(i > 1 .and. numwords .ne. prevnumwords) wordwarning = .true.

            ! if there is a t in the forcing time string, it must be in iso format
            if(index(forcing_splitline(1),"T") > 0) then
                call get_iso8601(forcing_splitline(1), forcing_time)
                forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
            else
                read(forcing_splitline(1),*,err=110) forcing(numforcings)%field(i)%time
            end if

            if(numwords < 4) call message("error in forcing file, found < 3 reals on line",2)
            
            read(forcing_splitline(2:4),*,err=107) &
                 forcing(numforcings)%field(i)%data, &
                 forcing(numforcings)%field(i)%data2,&
                 forcing(numforcings)%field(i)%data3

            if(numwords > 4+maxnumconc) numwords = 4+maxnumconc
            if(numwords > 4) then
                read(forcing_splitline(5:numwords),*,err=111) forcing(numforcings)%field(i)%conc(1:numwords-4)
            end if
        end do
            
        if(wordwarning) call message("the number of reals for concentration was not the same on all lines",1)
    case default
        call message('unknown well format',2)
    end select

    if(debug > 0 .and. master) then
        select case(well_format(numwell))
        case(0)
            write(*,'(2x,a,": i=",i0," j=",i0," Flux=",f0.1," T=",f0.1, " S=",f0.1)') &
                 trim(well_name(numwell)), iwell(numwell), jwell(numwell), fwell(numwell), twell(numwell), swell(numwell)
        case(1)
            write(*,'(2x,a,": i=",i0," j=",i0," Forcing file=",a)') &
                 trim(well_name(numwell)), iwell(numwell), jwell(numwell), trim(forcing(numforcings)%file_name)       
            write(*,*) " got",forcing(numforcings)%numfields,"forcing times with ",numwords," columns"
            i=forcing(numforcings)%numfields
            write(*,*) " first time=",forcing(numforcings)%field(1)%time," last time=",forcing(numforcings)%field(i)%time
        end select
    end if

    return

101 call message('could not read integer iwell from column 3',2)
102 call message('could not read integer jwell from column 4',2)
103 call message('could not read float value for fwell from column 6',2)
104 call message('could not read float value for twell from column 7',2)
105 call message('could not read float value for swell from column 8',2)
106 call message('could not read integer for the well format from column 5',2)
107 if(master) write(*,*) "well forcing file=", splitline(6)," line=", linenumbers(i)
    if(master) write(*,*) "i=",i," line=", trim(forcing_lines(i))
    call message('failed to read well forcing data: time flux s t, from above line',2)
110 if(master) write(*,*) "forcing file=",forcing(numforcings)%file_name,"line number=",linenumbers(i)
    if(master) write(*,*) "forcing line=",trim(forcing_lines(i))
    call message("could not read real value for time from column 1 at the given line in the forcing file",2)
111 if(master) write(*,*) "numwords=",numwords,"forcing line=",trim(forcing_lines(i))
    call message("failed reading numwords reals for well concentration from column 5 and out on this line",2)
  end subroutine add_well


  subroutine initialize_wind(splitline)
    implicit none
    character(*) splitline(:)
    integer linenumbers(10000),i,j, forcing_time(6)
    character(250) forcing_lines(10000)
    character(100) forcing_splitline(5)
    real(4), allocatable :: tmp2d(:,:), tmp1dr4(:)

    if(wind_forcing_set .and. .not. wind_forcing) return
    
    if(wind_set) call message("this keyword is only allowed to be set once",2)
    
    read(splitline(2),*,err=106) wind_type
    select case(wind_type)
    case(0)
        wind_forcing=.false.
        if(debug>0) write(*,*)' wind was turned off'
    case(1)
        read(splitline(3),*,err=101) wind1_amplitude
        read(splitline(4),*,err=102) wind1_period
        read(splitline(5),*,err=103) wind1_direction
        read(splitline(6),*,err=104) wind1_hourly_rotation
        read(splitline(7),*,err=105) wind1_rampup_time
        if(debug>0.and.master) write(*,*)' wind1 selected: amplitude,period,direction,rotspeed,rampup=',&
             wind1_amplitude, wind1_period, wind1_direction, wind1_hourly_rotation, wind1_rampup_time
    case(2)
        read(splitline(3),*,err=107) wind2_strength
        read(splitline(4),*,err=108) wind2_radius
        read(splitline(5),*,err=109) wind2_startx
        read(splitline(6),*,err=110) wind2_starty
        read(splitline(7),*,err=111) wind2_center_speed
        read(splitline(8),*,err=112) wind2_center_direction
        read(splitline(9),*,err=113) wind2_rampup_time
        if(debug>0.and.master) write(*,*)' wind2 selected: strength,radius,x0,y0,centerspeed,centerdir,rampup=', &
             wind2_strength, wind2_radius, wind2_startx, wind2_starty, wind2_center_speed, &
             wind2_center_direction, wind2_rampup_time
    case(3)
        if(.not.bom_allocated) then
            call message('arrays not allocated, initialize this type of wind after im,jm,kb are set.',2)
        end if
        wind3_data_file = trim(data_path)//splitline(3)

        call replace_yyyy_with_year(wind3_data_file, ndate)
        
        if(debug>0.and.master) write(*,*)' wind3 selected, forcing file=', wind3_data_file

        numforcings = numforcings + 1
        if(numforcings > max_numforcings) call message("too many bc's with input from file, increase max_numforcings",2)

        if(master) write(*,'(2x,a,i0)') "Forcing id=", numforcings
        
        if(master)call get_lines(wind3_data_file, forcing(numforcings)%numfields, forcing_lines, linenumbers)
        call broadcast(forcing(numforcings)%numfields)
        call broadcast(forcing_lines)
        call broadcast(linenumbers)

        allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))
        
        do i=1,forcing(numforcings)%numfields
            forcing(numforcings)%name = splitline(1)
            call split(forcing_lines(i), forcing_splitline)
            
           ! if there is a t in the forcing time string, it must be in iso format
            if(index(forcing_splitline(1),"T") > 0) then
                call get_iso8601(forcing_splitline(1), forcing_time)
                forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
            else
                read(forcing_splitline(1),*,err=110) forcing(numforcings)%field(i)%time
            end if

            read(forcing_splitline(2:3),*) &
                 forcing(numforcings)%field(i)%data, &
                 forcing(numforcings)%field(i)%data2
            forcing(numforcings)%type = 6
        end do
        
    case(4)
        
! atmospheric forcing from ncep netcdf files
!   WIND  TYPE  FILE
        if(.not.bom_allocated) then
            call message('arrays not allocated, initialize this type of wind after im,jm,kb are set.',2)
        end if
        wind3_data_file=trim(data_path)//splitline(3)
        call replace_yyyy_with_year(wind3_data_file, ndate)

        if(debug>0.and.master) write(*,*)' Atmospheric forcing from NCEP, forcing file=', trim(wind3_data_file)

        numforcings = numforcings + 1
        if(numforcings > max_numforcings) call message("too many bc's with input from file, increase max_numforcings",2)

        if(master) write(*,'(3x,a,i0)') "Forcing id=", numforcings
        
        if(master) call get_lines(wind3_data_file, forcing(numforcings)%numfields, forcing_lines, linenumbers)

        if(debug > 0 .and. master) write(*,*) " got",forcing(numforcings)%numfields,"fields"
        
        call broadcast(forcing(numforcings)%numfields)
        call broadcast(forcing_lines)
        call broadcast(linenumbers)

        allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))
        
! the file format for the forcing file is lines like this
!   xxxx-xx-xxTxx:xx:xx  fileu  filev  filep  timeindex
! the time index is the index in each netcdf files time dimension that correspond to the time in the first column.
! we do this to avoid linking in udunits.
        
        do i=1,forcing(numforcings)%numfields
            forcing(numforcings)%name = splitline(1)
            call split(forcing_lines(i), forcing_splitline)

           !! if there is a t in the forcing time string, it must be in iso format
            if(index(forcing_splitline(1),"T") > 0) then
                call get_iso8601(forcing_splitline(1), forcing_time)
                forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
            else
                read(forcing_splitline(1),*,err=114,end=114) forcing(numforcings)%field(i)%time
            end if
            
           !! the netcdf files containing the u, v and p fields for the given time
            forcing(numforcings)%field(i)%file = forcing_splitline(2)
            forcing(numforcings)%field(i)%file2 = forcing_splitline(3)
            forcing(numforcings)%field(i)%file3 = forcing_splitline(4)

           !! use col to store the time index of the field. should be the same for all files 
            read(forcing_splitline(5),*,err=115,end=115) forcing(numforcings)%field(i)%col           
            
            forcing(numforcings)%type = 7
        end do
        
! brute force read dimensions from the first file, assume all 2d fields to be of the same dimension.
        call getdimensions(forcing(numforcings)%field(1)%file, forcing(numforcings)%dim1, forcing(numforcings)%dim2)
        
! find and store lon lat for all bom grid points, to be used for interpolation.
        if(setlonlat) call init_lonlat()

    case(5,6)

! atmospheric forcing from binary stream files
!   WIND  N  FILE LONDIM LATDIM LONFILE LATFILE
        
        if(.not.bom_allocated) then
            call message('arrays not allocated, initialize this type of wind after im,jm,kb are set.',2)
        end if
        wind3_data_file = trim(data_path)//splitline(3)
        call replace_yyyy_with_year(wind3_data_file, ndate)

        if(debug>0.and.master) write(*,*)' Atmospheric forcing from binary stream forcing file=', trim(wind3_data_file)

        numforcings = numforcings + 1
        if(numforcings > max_numforcings) call message("too many bc's with input from file, increase max_numforcings",2)

        if(master) write(*,'(2x,a,i0)') "Forcing id=", numforcings

        forcing(numforcings)%name = splitline(1)
        read(splitline(4:5),*) forcing(numforcings)%dim1, forcing(numforcings)%dim2
        
        if(master) call get_lines(wind3_data_file, forcing(numforcings)%numfields, forcing_lines, linenumbers)

        if(debug > 0 .and. master) write(*,*) " got",forcing(numforcings)%numfields,"fields"
        
        call broadcast(forcing(numforcings)%numfields)
        call broadcast(forcing_lines)
        call broadcast(linenumbers)

        allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))
        
! the file format for the forcing file is lines like this
!   xxxx-xx-xxTxx:xx:xx  file1  file2  file3
! the time index is the index in each netcdf files time dimension that correspond to the time in the first column.
! we do this to avoid linking in udunits.
        
        do i=1,forcing(numforcings)%numfields
            call split(forcing_lines(i), forcing_splitline)

           !! if there is a t in the forcing time string, it must be in iso format
            if(index(forcing_splitline(1),"T") > 0) then
                call get_iso8601(forcing_splitline(1), forcing_time)
                forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
            else
                read(forcing_splitline(1),*,err=114,end=114) forcing(numforcings)%field(i)%time
            end if
            
           !! the files containing e.g. the u, v and p fields for the given time
            forcing(numforcings)%field(i)%file = forcing_splitline(2)
            forcing(numforcings)%field(i)%file2 = forcing_splitline(3)
            forcing(numforcings)%field(i)%file3 = forcing_splitline(4)
        end do
        
! find and store lon lat for all bom grid points, to be used for interpolation.
        if(setlonlat) call init_lonlat()

        if(debug>1.and.master)then
            write(*,*) " shape of lon,lat arrays= ",forcing(numforcings)%dim1, forcing(numforcings)%dim2
            write(*,*) " longitude file= ",trim(splitline(6))
            write(*,*) " latitude file= ",trim(splitline(7))
        end if
        
        if(wind_type == 5) then
            forcing(numforcings)%type = 8  ! these moving target integer codes will turn into a problem... clean later.
            
            allocate(ncep_lons(forcing(numforcings)%dim1), ncep_lats(forcing(numforcings)%dim2))
            if(master)then
                open(1000012, file=(trim(data_path)//splitline(6)), access="stream", form="unformatted", status="old")
                if(kind(ncep_lons).ne.4)then
                    allocate(tmp1dr4(forcing(numforcings)%dim1))
                    read(1000012) tmp1dr4
                    ncep_lons = tmp1dr4
                    deallocate(tmp1dr4)
                else
                    read(1000012) ncep_lons
                end if
                close(1000012)
            end if
            call broadcast(ncep_lons)
            if(master)then
                open(1000012, file=(trim(data_path)//splitline(7)), access="stream", form="unformatted", status="old")
                if(kind(ncep_lats).ne.4)then
                    allocate(tmp1dr4(forcing(numforcings)%dim2))
                    read(1000012) tmp1dr4
                    ncep_lats = tmp1dr4
                    deallocate(tmp1dr4)
                else
                    read(1000012) ncep_lats
                end if
                close(1000012)
            end if
            call broadcast(ncep_lats)

!           compute interpolation weights for gip            
!            allocate(forcing%map(i1:i2,j1:j2))
!            call getbilinmap(ncep_lons, ncep_lats, forcing%map)
        end if

        if(wind_type == 6) then
         !! allocate and fetch lon,lat for the met forcing source grid
            
            forcing(numforcings)%type = 11

            allocate(met_lons(forcing(numforcings)%dim1, forcing(numforcings)%dim2))
            allocate(met_lats(forcing(numforcings)%dim1, forcing(numforcings)%dim2))
            
            if(master)then
                open(1000012, file=(trim(data_path)//splitline(6)), access="stream", form="unformatted", status="old")
                read(1000012) met_lons
                close(1000012)
            end if
            call broadcast(met_lons)
            if(master)then
                open(1000012, file=(trim(data_path)//splitline(7)), access="stream", form="unformatted", status="old")
                read(1000012) met_lats
                close(1000012)
            end if            
            call broadcast(met_lats)
            
            !! allocate data structures for triangular interpolation.
            allocate(forcing(numforcings)%map(i1:i2,j1:j2))
            call gettrimap3(real(met_lons), real(met_lats), forcing(numforcings)%map)
            deallocate(met_lats)
        end if

    case(7)

        if(.not.bom_allocated) then
            call message('arrays not allocated, initialize this type of wind after im,jm,kb are set.',2)
        end if
        wind3_data_file = trim(data_path)//splitline(3)
        call replace_yyyy_with_year(wind3_data_file, ndate)

        if(debug>0.and.master) write(*,*)' Wind fields from gridded files, forcing file=', trim(wind3_data_file)

        numforcings = numforcings + 1
        if(numforcings > max_numforcings) call message("too many bc's with input from file, increase max_numforcings",2)

        if(master) write(*,'(3x,a,i0)') "Forcing id=", numforcings
        
        if(master) call get_lines(wind3_data_file, forcing(numforcings)%numfields, forcing_lines, linenumbers)

        if(debug > 0 .and. master) write(*,*) " got",forcing(numforcings)%numfields,"fields"
        
        call broadcast(forcing(numforcings)%numfields)
        call broadcast(forcing_lines)
        call broadcast(linenumbers)

        allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))
        
! the file format for the forcing file is lines like this
!   xxxx-xx-xxTxx:xx:xx  fileu  filev
        
        do i=1,forcing(numforcings)%numfields
            forcing(numforcings)%name = splitline(1)
            call split(forcing_lines(i), forcing_splitline)

           !! if there is a t in the forcing time string, it must be in iso format
            if(index(forcing_splitline(1),"T") > 0) then
                call get_iso8601(forcing_splitline(1), forcing_time)
                forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
            else
                read(forcing_splitline(1),*,err=114,end=114) forcing(numforcings)%field(i)%time
            end if
            
           !! the netcdf files containing the u, v and p fields for the given time
            forcing(numforcings)%field(i)%file = forcing_splitline(2)
            forcing(numforcings)%field(i)%file2 = forcing_splitline(3)

            forcing(numforcings)%type = 15
        end do
        
    case default
        call message('unknown wind forcing type',2)
    end select
    
    return
    
101 call message('could not get float for wind1 strength(amplitude) from column 3',2)
102 call message('could not get float for wind1 period from column 4',2)
103 call message('could not get float for wind1 direction from column 5',2)
104 call message('could not get float for hourly rotation of wind1 from column 6',2)
105 call message('could not get float for rampup_time for wind1 strength from column 7',2)
106 call message('could not get integer for wind type from column 2',2)

107 call message('could not get float for wind2 strength from column 3',2)
108 call message('could not get float for wind2 radius from column 4',2)
109 call message('could not get float for wind2 start x position from column 5',2)
110 call message('could not get float for wind2 start y position from column 6',2)
111 call message('could not get float for wind2 low pressure center speed from column 7',2)
112 call message('could not get float for wind2 low pressure center direction from column 8',2)
113 call message('could not get float for wind2 low pressure ramp-up time from column 9',2)

114 call message('wind4 error reading time from column 1',2)
115 call message('wind4 error reading file index from column 5',2)
    
  end subroutine initialize_wind

  
  subroutine init_lonlat
! initialize the BOM grid lon, lat arrays
    real gxpole, gypole,x,y
    integer i,j,ifail

    if(.not. setlonlat) return
    
    if(.not.all((/xpole_set,ypole_set,ylong_set/))) call message("xpole,ypole and ylong must all be set before this line",2)
    if(.not.bom_allocated) call message("BOM arrays not allocated, check that H is declared before this keyword",2)

    allocate(longitude(i1:i2,j1:j2), latitude(i1:i2,j1:j2))
    gxpole = (xpole+dx)/dx
    gypole = (ypole+dy)/dy
    do j=j1,j2
        do i=i1,i2
           ! global grid coordinate
            x = block%fi + i - 1 - 0.5
            y = block%fj + j - 1 - 0.5
            call gr2sph(x, y, gxpole, gypole, dx*0.001, dy*0.001, ylong, longitude(i,j), latitude(i,j), ifail)
        end do
    end do

    minlon = minval(longitude)
    maxlon = maxval(longitude)
    minlat = minval(latitude)
    maxlat = maxval(latitude)

    call reduce('min', minlon, minlat)
    call reduce('max', maxlon, maxlat)

    if(debug > 1 .and. master) write(*,1) "init_lonlat: Grid longitude range=",minlon,maxlon," latitude range=",minlat,maxlat
1   format(2x,a,f0.2,",",f0.2,a,f0.2,",",f0.2)
    setlonlat = .false.
  end subroutine init_lonlat


  subroutine set_bc2d(zone, bctype, bcset, edge, f, kdum, edge_maxzone, splitline)
! Identical to set_bc3d but only takes care of 2d arrays    
! initialize the boundary condition data structures for a given
! variable, zone etc. from the configuration line given in splitline, which is
!     bc_<variable>_<edge> <zone> <type> <setting_method> [<data>]
!
    implicit none
    integer, intent(in) :: zone
    integer, intent(out) :: bctype
    logical, intent(out) :: bcset
    character(*), intent(in) :: edge
    integer, intent(in) :: edge_maxzone, kdum
    real f(i1:,j1:)
    character(*) splitline(:)
    integer setting_method, range(6), f_id, dim1,dim2,r(6),kdim
    logical inblock

    kdim = 1

    if(periodic_ew .and. (edge=="west" .or. edge=="east")) then
        call message("ignoring BC as PERIODIC_EW is requested",1)
        return
    end if
    
    if(edge_maxzone <=0 ) call message('according to bathymetry there are no open boundaries here',2)

    if(zone>0 .and. zone<=edge_maxzone) then
        read(splitline(3),*,err=102) bctype

        if(debug>0.and.master) write(*,'(1x,a,i0,a,a,a,a)')' setting zone=',zone,' edge=',edge,&
             ' to type=',trim(bcname(bctype))
        if(bctype<-1 .or. bctype>4) call message('invalid bc type',2)
    else
        if(master)write(*,'(1x,a,i0)') ' zone=',zone
        call message('invalid zone number',2)
    end if

    
    if(bctype == 4) binreg_set=.true.
    if(restart .and. bctype==4) call allocate_binreg()
    
    read(splitline(4),*,err=103, end=103) setting_method
   
    range(5:6) = (/1,kdim/)
    select case(edge)
    case('west')
        range(1:4)=bowest(zone)%global_range
        if(any(setting_method == (/13,14/))) range(3:4) = (/1,jm/)
    case('east')
        range(1:4)=boeast(zone)%global_range
        if(any(setting_method == (/13,14/))) range(3:4) = (/1,jm/)
    case('south')
        range(1:4)=bosouth(zone)%global_range
        if(any(setting_method == (/13,14/))) range(1:2) = (/1,im/)
    case('north')
        range(1:4)=bonorth(zone)%global_range
        if(any(setting_method == (/13,14/))) range(1:2) = (/1,im/)
    case default
        if(master)write(*,*)'edge=',edge
        call message('unknown edge specifier. try one of east,west,south or north',2)
    end select

    bcset = .true.



    if(bctype == 2) then

! periodic cond. get the corresponding zone on the opposite side. must be same length.

        if(.not. h_set) call message('you must set h before specifying periodic conditions',2)

        select case(edge)
        case('west')
            bowest(zone)%periodic = setting_method
            dim1 = bowest(zone)%global_range(4) - bowest(zone)%global_range(3)
            dim2 = boeast(setting_method)%global_range(4) - boeast(setting_method)%global_range(3)
            if(dim1 .ne. dim2) then
                if(master)write(*,*)'zone west=',zone
                if(master)write(*,*)'corresponding zone east=',setting_method
                call message('boundary zone lengths are not the same',2)
            end if
        case('east')
            boeast(zone)%periodic = setting_method
            dim1 = boeast(zone)%global_range(4) - boeast(zone)%global_range(3)
            dim2 = bowest(setting_method)%global_range(4) - bowest(setting_method)%global_range(3)
            if(dim1 .ne. dim2) then
                if(master)write(*,*)'zone east=',zone
                if(master)write(*,*)'corresponding zone west=',setting_method
                call message('boundary zone lengths are not the same',2)
            end if
        case('south')
            bosouth(zone)%periodic = setting_method
            dim1 = bosouth(zone)%global_range(4) - bosouth(zone)%global_range(3)
            dim2 = bonorth(setting_method)%global_range(4) - bonorth(setting_method)%global_range(3)
            if(dim1 .ne. dim2) then
                if(master)write(*,*)'zone south=',zone
                if(master)write(*,*)'corresponding zone north=',setting_method
                call message('boundary zone lengths are not the same',2)
            end if
        case('north')
            bonorth(zone)%periodic = setting_method
            dim1 = bonorth(zone)%global_range(4) - bonorth(zone)%global_range(3)
            dim2 = bosouth(setting_method)%global_range(4) - bosouth(setting_method)%global_range(3)
            if(dim1 .ne. dim2) then
                if(master)write(*,*)'zone north=',zone
                if(master)write(*,*)'corresponding zone south=',setting_method
                call message('boundary zone lengths are not the same',2)
            end if
        end select
        return
    end if


    
    select case(setting_method)
    case(-1)
        if(debug > 0 .and. master) write(*,*)' boundary array will be set to initial condition for field'
        bcset = .false.
        
    case(0,3,4,6,7,8,9,10,11)
!
! constant boundary condition:
!   initialize xx_bound by reusing set_var from above: we have to use
!   only part of splitline and we also pass only the section of the
!   array that match the frs zone. we have read splitline(3), so it can
!   safely be overwritten to get the label right in set_var.
!
        call check_allocation()
        splitline(3)=splitline(1)
        call set_var(f, splitline(3:), r=range)
        
    case(1)
!
! simple, time dependent bc: set field to a scalar value. the file
! format is the time in seconds then scalar value.
!
       call simple_forcing(f, splitline, range, f_id)
!
! only flag bc as set if the time of the first forcing field is before
! the time for the first time step. if the first forcing field comes
! after, we need to use the initial value of the forcing field in the
! interpolation, and therefore it needs to get its value set from the
! "mother" variable.
       if(forcing(f_id)%field(1)%time >= dt) then
           bcset = .false.
       end if

    case(12)
! Rampup forcing:
!  initially, during ramp up time, the bc for the variable is interpolated between state 1 and state 2.
!   BC_<VARIABLE>_<EDGE> <ZONE> <TYPE> 12 <RAMP UP TIME (S)> <TYPE STATE 1> <DATA STATE 1> <TYPE STATE 2> <DATA STATE 2>
!  example        
        call add_rampup(f, kdim, range, splitline)
        
    case(13)
! Read one row of values from file, quite similar to simple forcing
! but give header file names instead of scalar values.
        call edge_from_file2d(f, splitline, range, f_id)
        if(forcing(f_id)%field(1)%time >= dt) then
            bcset = .false.
        end if
        
    case(14)
! Special type of forcing only used by UA and VA. Sets UA_BOUND and
! VA_BOUND to be the depth average of U_BOUND, V_BOUND respecitively.
        
! It is important that this forcing spec is added AFTER the
! corresponding spec for U, V, since we iterate linearly through the
! list of forcings to apply
        
        if(index(splitline(1), "bc_ua") > 0) then
            call copy3dto2dforcing(trim(splitline(1)), u_bound, f, range)
            if(master.and.debug >1) write(*,*) " UA_BOUND will be depth integrated from U_BOUND"
            select case(edge)
            case("east")
                if(.not. boeast(zone)%u_set) call message("BC for U east has not been set for this zone",2)
            case("west")
                if(.not. bowest(zone)%u_set) call message("BC for U west has not been set for this zone",2)
            case("north")
                if(.not. bonorth(zone)%u_set) call message("BC for U north has not been set for this zone",2)            
            case("south")
                if(.not. bosouth(zone)%u_set) call message("BC for U south has not been set for this zone",2)                        
            end select            
        end if
        if(index(splitline(1), "bc_va") > 0) then
            call copy3dto2dforcing(trim(splitline(1)), v_bound, f, range)           
            if(master.and.debug >1) write(*,*) " VA_BOUND will be depth integrated from V_BOUND"
            select case(edge)
            case("east")
                if(.not. boeast(zone)%v_set) call message("BC for V east has not been set for this zone",2)
            case("west")
                if(.not. bowest(zone)%v_set) call message("BC for V west has not been set for this zone",2)
            case("north")
                if(.not. bonorth(zone)%v_set) call message("BC for V north has not been set for this zone",2)            
            case("south")
                if(.not. bosouth(zone)%v_set) call message("BC for V south has not been set for this zone",2)                        
            end select            
        end if
        
    case(15)
        call message("this method only available for 3d arrays for now",2)
        
    case default
        if(master)write(*,*)' setting method=',setting_method,' zone=',zone
        call message('unknown method',2)
    end select

    return
101 call message('could not read integer value for zone from column 2',2)
102 call message('could not read integer value for bc type from column 3',2)
103 call message('could not read integer value for initialization type from column 4',2)
104 call message('',2)
  end subroutine set_bc2d


  subroutine copy3dto2dforcing(name, f3d_bound, f2d_bound, range)    
! connect the 2d _bound array to the source 3d _bound array
! used e.g. to depth integrate U_BOUND and put the result into UA_BOUND
    implicit none
    character(*) name
    real, target :: f3d_bound(i1:,j1:,:), f2d_bound(i1:,j1:)
    integer range(:)
    character pointtype

    numforcings = numforcings + 1
    if(numforcings > max_numforcings) call message("too many bc's with input from file, increase max_numforcings",2)
    if(master) write(*,'(3x,a,i0)') "Forcing id=", numforcings
    forcing(numforcings)%name = name
    forcing(numforcings)%range = range
    call globalrange2local(block%ibn, block%jbn, forcing(numforcings)%range, forcing(numforcings)%inblock)
    forcing(numforcings)%target2d => f2d_bound
    forcing(numforcings)%source3d => f3d_bound
    forcing(numforcings)%type = 9
  end subroutine copy3dto2dforcing

  
  subroutine set_bc3d(zone, bctype, bcset, edge, f, kdum, edge_maxzone, splitline)
!    
! initialize the boundary condition data structures for a given 3d
! variable, zone etc. from the configuration line given in splitline, which is
!     bc_<variable>_<edge> <zone> <type> <setting_method> [<data>]
!
    implicit none
    integer, intent(in) :: zone
    integer, intent(out) :: bctype
    logical, intent(out) :: bcset
    character(*), intent(in) :: edge
    integer, intent(in) :: edge_maxzone,kdum
    real f(i1:,j1:,:)
    character(*) splitline(:)
    integer setting_method, range(6), f_id, dim1,dim2,r(6),kdim
    logical inblock

    if(periodic_ew .and. (edge=="west" .or. edge=="east")) then
        call message("ignoring BC as PERIODIC_EW is requested",1)
        return
    end if
    
    kdim = size(f,3)
    
    if(edge_maxzone <=0 ) call message('according to bathymetry there are no open boundaries here',2)

    if(zone>0 .and. zone<=edge_maxzone) then
        read(splitline(3),*,err=102) bctype

        if(debug>0.and.master) write(*,'(1x,a,i0,a,a,a,a)')' setting zone=',zone,' edge=',edge,&
             ' to type=',trim(bcname(bctype))
        if(bctype<-1 .or. bctype>4) call message('invalid bc type',2)
    else
        if(master)write(*,'(1x,a,i0)') ' zone=',zone
        call message('invalid zone number',2)
    end if

    
    if(bctype == 4) binreg_set=.true.
    if(restart .and. bctype==4) call allocate_binreg()

    
    read(splitline(4),*,err=103, end=103) setting_method
    
   
    range(5:6) = (/1,kdim/)
    select case(edge)
    case('west')
        range(1:4)=bowest(zone)%global_range
        if(any(setting_method == (/13,14/))) range(3:4) = (/1,jm/)
    case('east')
        range(1:4)=boeast(zone)%global_range
        if(any(setting_method == (/13,14/))) range(3:4) = (/1,jm/)
    case('south')
        range(1:4)=bosouth(zone)%global_range
        if(any(setting_method == (/13,14/))) range(1:2) = (/1,im/)
    case('north')
        range(1:4)=bonorth(zone)%global_range
        if(any(setting_method == (/13,14/))) range(1:2) = (/1,im/)
    case default
        if(master)write(*,*)'edge=',edge
        call message('unknown edge specifier. try one of east,west,south or north',2)
    end select

    bcset = .true.


    if(bctype == 2) then

! periodic cond. get the corresponding zone on the opposite side. must be same length.

        if(.not. h_set) call message('you must set h before specifying periodic conditions',2)

        select case(edge)
        case('west')
            bowest(zone)%periodic = setting_method
            dim1 = bowest(zone)%global_range(4) - bowest(zone)%global_range(3)
            dim2 = boeast(setting_method)%global_range(4) - boeast(setting_method)%global_range(3)
            if(dim1 .ne. dim2) then
                if(master)write(*,*)'zone west=',zone
                if(master)write(*,*)'corresponding zone east=',setting_method
                call message('boundary zone lengths are not the same',2)
            end if
        case('east')
            boeast(zone)%periodic = setting_method
            dim1 = boeast(zone)%global_range(4) - boeast(zone)%global_range(3)
            dim2 = bowest(setting_method)%global_range(4) - bowest(setting_method)%global_range(3)
            if(dim1 .ne. dim2) then
                if(master)write(*,*)'zone east=',zone
                if(master)write(*,*)'corresponding zone west=',setting_method
                call message('boundary zone lengths are not the same',2)
            end if
        case('south')
            bosouth(zone)%periodic = setting_method
            dim1 = bosouth(zone)%global_range(4) - bosouth(zone)%global_range(3)
            dim2 = bonorth(setting_method)%global_range(4) - bonorth(setting_method)%global_range(3)
            if(dim1 .ne. dim2) then
                if(master)write(*,*)'zone south=',zone
                if(master)write(*,*)'corresponding zone north=',setting_method
                call message('boundary zone lengths are not the same',2)
            end if
        case('north')
            bonorth(zone)%periodic = setting_method
            dim1 = bonorth(zone)%global_range(4) - bonorth(zone)%global_range(3)
            dim2 = bosouth(setting_method)%global_range(4) - bosouth(setting_method)%global_range(3)
            if(dim1 .ne. dim2) then
                if(master)write(*,*)'zone north=',zone
                if(master)write(*,*)'corresponding zone south=',setting_method
                call message('boundary zone lengths are not the same',2)
            end if
        end select
        return
    end if
    
    
    select case(setting_method)
    case(-1)
        if(debug > 0 .and. master) write(*,*)' boundary array will be set to initial condition for field'
        bcset = .false.
        
    case(0,3,4,6,7,8,9,10,11)
!
! constant boundary condition:
!   initialize xx_bound by reusing set_var from above: we have to use
!   only part of splitline and we also pass only the section of the
!   array that match the frs zone. we have read splitline(3), so it can
!   safely be overwritten to get the label right in set_var.
!
        call check_allocation()
        splitline(3)=splitline(1)
        call set_var(f, splitline(3:), r=range)

! RHO is a special case, since we initialize it with "physical"
! magnitude, need to subtract 1000, and later also subtract MASS from
! RHO_BOUND
        r = range
        call globalrange2local(block%ibn,block%jbn,r,inblock)
        if(splitline(1)(4:6) == "rho".and.inblock) f(r(1):r(2),r(3):r(4),r(5):r(6)) = f(r(1):r(2),r(3):r(4),r(5):r(6)) - 1000.0 
        
    case(1)
!
! simple, time dependent bc: set field to a scalar value. the file
! format is the time in seconds then scalar value.
!
       call simple_forcing(f, splitline, range, f_id)
!
! only flag bc as set if the time of the first forcing field is before
! the time for the first time step. if the first forcing field comes
! after, we need to use the initial value of the forcing field in the
! interpolation, and therefore it needs to get its value set from the
! "mother" variable.
       if(forcing(f_id)%field(1)%time >= dt) then
           bcset = .false.
       end if

    case(12)
! Rampup forcing:
!  initially, during ramp up time, the bc for the variable is interpolated between state 1 and state 2.
!   BC_<VARIABLE>_<EDGE> <ZONE> <TYPE> 12 <RAMP UP TIME (S)> <TYPE STATE 1> <DATA STATE 1> <TYPE STATE 2> <DATA STATE 2>
!  example        
        call add_rampup(f, kdim, range, splitline)
        
    case(13)
! Read one row of values from file, quite similar to simple forcing
! but give header file names instead of scalar values.
        call edge_from_file(f, splitline, range, f_id)
        if(forcing(f_id)%field(1)%time >= dt) then
            bcset = .false.
        end if
   
    case(14)
        call message("forcing type only available for UA, VA",2)

    case(15)
! read and interpolate (vertically) 3d z coordinate forcing data from file into boundary arrays.
        call zone_from_file(f, splitline, range, f_id)
        if(forcing(f_id)%field(1)%time >= dt) then
            bcset = .false.
        end if
        
    case default
        if(master)write(*,*)' setting method=',setting_method,' zone=',zone
        call message('unknown method',2)
    end select

    return
101 call message('could not read integer value for zone from column 2',2)
102 call message('could not read integer value for bc type from column 3',2)
103 call message('could not read integer value for initialization type from column 4',2)
104 call message('',2)
  end subroutine set_bc3d


  subroutine add_3d_simple_forcing(f, splitline, range, forcing_id)
! this is the forcing reader for type=1 forcing from file, when the variable to be forced is a 3d array.
!     bc_<variable>_<edge> <zone> <type> <setting_method> [<data>]
!
    implicit none
    real, target :: f(i1:,j1:,:)
    character(*) splitline(:)
    integer range(6), rc(6)
    integer linenumbers(400),i, forcing_id,forcing_time(6)
    character(200) forcing_lines(400), forcing_splitline(2)
    logical inblock

    numforcings = numforcings + 1
    if(numforcings > max_numforcings) call message("too many bc's with input from file, increase max_numforcings",2)

    if(master) write(*,'(3x,a,i0)') "Forcing id=", numforcings
    
    forcing(numforcings)%file_name = trim(data_path)//trim(splitline(5))
    if(master)call get_lines(forcing(numforcings)%file_name, forcing(numforcings)%numfields, forcing_lines, linenumbers)
    if(master.and.debug>1) write(*,*) "  Simple forcing from file:",trim(splitline(5))," #fields=",forcing(numforcings)%numfields
    call broadcast(forcing(numforcings)%numfields)
    call broadcast(forcing_lines)
    call broadcast(linenumbers)
    allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))

    do i=1,forcing(numforcings)%numfields
        forcing(numforcings)%name = splitline(1)

        call split(forcing_lines(i), forcing_splitline)

        ! if there is a t in the forcing time string, it must be in iso format
        if(index(forcing_splitline(1),"T") > 0) then
            call get_iso8601(forcing_splitline(1), forcing_time)
            forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
        else
            read(forcing_splitline(1),*,err=1) forcing(numforcings)%field(i)%time
        end if
        read(forcing_splitline(2),*) forcing(numforcings)%field(i)%data

        forcing(numforcings)%type = 0
    end do

    if(master.and.debug>1) then
        write(*,*) "  Simple forcing first time, data:",&
             forcing(numforcings)%field(1)%time, forcing(numforcings)%field(1)%data
        write(*,*) "  Simple forcing last time, data:",&
             forcing(numforcings)%field(forcing(numforcings)%numfields)%time, &
             forcing(numforcings)%field(forcing(numforcings)%numfields)%data    
    end if
    
    forcing(numforcings)%range = range

    call globalrange2local(block%ibn, block%jbn, forcing(numforcings)%range, forcing(numforcings)%inblock)

    forcing(numforcings)%target3d => f
    forcing_id = numforcings
    return
1   if(master) write(*,*) "forcing file=",forcing(numforcings)%file_name,"line number=",linenumbers(i)
    if(master) write(*,*) "forcing line=",trim(forcing_lines(i))
    call message("could not read real value for time from column 1 at the given line in the forcing file",2)
  end subroutine add_3d_simple_forcing

  
  subroutine add_2d_simple_forcing(f, splitline, range, forcing_id)
!
! this is the forcing reader for type=1 forcing from file, when the
! variable to be forced is a 2d array, otherwise the body is identical
! to the routine above. consider using include as in other routines here.
!
!     bc_<variable>_<edge> <zone> <type> <setting_method> [<data>]
!
    implicit none
    real, target :: f(i1:,j1:)
    character(*) splitline(:)
    integer range(6), rc(6), forcing_id
    integer linenumbers(400),i,forcing_time(6)
    character(200) forcing_lines(400), forcing_splitline(2)
    logical inblock

    numforcings = numforcings + 1
    if(numforcings > max_numforcings) call message("too many bc's with input from file, increase max_numforcings",2)

    if(master) write(*,'(3x,a,i0)') "Forcing id=", numforcings
    
    forcing(numforcings)%file_name = trim(data_path)//trim(splitline(5))

    if(master)call get_lines(forcing(numforcings)%file_name, forcing(numforcings)%numfields, forcing_lines, linenumbers)
    if(master.and.debug>1) write(*,*) "  Simple forcing from file:",trim(splitline(5))," #fields=",forcing(numforcings)%numfields
    call broadcast(forcing(numforcings)%numfields)
    call broadcast(forcing_lines)
    call broadcast(linenumbers)
    allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))

    do i=1,forcing(numforcings)%numfields
        forcing(numforcings)%name = splitline(1)

        call split(forcing_lines(i), forcing_splitline)

        ! if there is a t in the forcing time string, it must be in iso format
        if(index(forcing_splitline(1),"T") > 0) then
            call get_iso8601(forcing_splitline(1), forcing_time)
            forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
        else
            read(forcing_splitline(1),*,err=1) forcing(numforcings)%field(i)%time
        end if
        read(forcing_splitline(2),*) forcing(numforcings)%field(i)%data

        forcing(numforcings)%type = 0
    end do

    if(master.and.debug>1) then
        write(*,*) "  Simple forcing first time, data:",&
             forcing(numforcings)%field(1)%time, forcing(numforcings)%field(1)%data
        write(*,*) "  Simple forcing last time, data:",&
             forcing(numforcings)%field(forcing(numforcings)%numfields)%time, &
             forcing(numforcings)%field(forcing(numforcings)%numfields)%data    
    end if
    
    rc = range
    call globalrange2local(block%ibn, block%jbn, rc, inblock)

    forcing(numforcings)%range = rc
    forcing(numforcings)%inblock = inblock
    forcing(numforcings)%target2d => f
    forcing_id = numforcings
    return

1   if(master) write(*,*) "forcing file=",forcing(numforcings)%file_name,"line number=",linenumbers(i)
    if(master) write(*,*) "forcing line=",trim(forcing_lines(i))
    call message("could not read real value for time from column 1 at the given line in the forcing file",2)
  end subroutine add_2d_simple_forcing


  subroutine relaxsurftemp(splitline, code)
! relax_tsurf <time constant C1> <forcing file>
!    
! add forcing that relaxes surface temperature towards climatology as in Oey&Chen 1992
!   TSURF is compute as  C1*(TCLIMATE - TMODEL)
!            c1 = 1.736E-5
!            wtsurf = c1*(  - t(:,:,1))
    character(*) splitline(:)
    integer, intent(in) :: code
    
    integer linenumbers(800),i, forcing_id,forcing_time(6)
    character(200) forcing_lines(800), forcing_splitline(2)

    
    numforcings = numforcings + 1
    if(numforcings > max_numforcings) call message("too many bc's with input from file, increase max_numforcings",2)

    if(master) write(*,'(3x,a,i0)') "Forcing id=", numforcings

    select case(code)
    case(0)        
        forcing(numforcings)%type = 10
    case(1)
        forcing(numforcings)%type = 12
    case default
        call message("invalid code for relaxsurftemp:", 2, ival=(/code/))
    end select
        
    
    ! get the constant c1
    read(splitline(2),*) forcing(numforcings)%rdata

    forcing(numforcings)%file_name = trim(data_path)//splitline(3)

    call replace_yyyy_with_year(forcing(numforcings)%file_name, ndate)
    
    if(master)call get_lines(forcing(numforcings)%file_name, forcing(numforcings)%numfields, forcing_lines, linenumbers)
    if(master.and.debug>1) then
        write(*,*) "  relaxsurftemp: header file= ",trim(splitline(3)),"  # fields=",forcing(numforcings)%numfields
        write(*,*) "  C1 relaxation time constant=", forcing(numforcings)%rdata
    end if
        
    call broadcast(forcing(numforcings)%numfields)
    call broadcast(forcing_lines)
    call broadcast(linenumbers)
    
    allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))

    forcing(numforcings)%name = splitline(1)
    
    do i=1,forcing(numforcings)%numfields
        call split(forcing_lines(i), forcing_splitline)
        
        ! if there is a t in the forcing time string, it must be in iso format
        if(index(forcing_splitline(1),"T") > 0) then
            call get_iso8601(forcing_splitline(1), forcing_time)
            forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
        else
            read(forcing_splitline(1),*,err=1) forcing(numforcings)%field(i)%time
        end if

        select case(code)
        case(0)
            forcing(numforcings)%field(i)%file = trim(forcing_splitline(2))
        case(1)
            read(forcing_splitline(2),*,err=2) forcing(numforcings)%field(i)%data
        end select
    end do
    return
1   call message("relaxsurftemp: could not read C1 from column 2",2)
2   call message("relaxsurftemp: could not read temperature from column 2",2)
  end subroutine relaxsurftemp


  subroutine zone_from_file(f, splitline, range, forcing_id)
!    
! this is the forcing reader for setting method=15 forcing from
! file. the variable to be forced is a 3d array.
!                             2      3    4       5           6          7            8           9     10     11
!     bc_<variable>_<edge> <zone> <type> 15 <header file> <i offset> <j offset> <z level file> <dim1> <dim2> <dim3>
!   
! <header file> is a file with 2 columns, where col1 is time, col2 is the 
!               file name with unformatted data for the edge. lines beginning
!               with ! are ignored, blank lines are allowed. each file 
!               contains a 3d array with dimensions <dim1> <dim2> <dim3>                   
! <offsetx> is the x distance in meters from origo to the first grid
!           point of the forcing. similar for y. the horisontal resolution is
!           assumed to be the same as the model grid
! <z level file> is a file with a 1d array of z levels of dimension <dim3>
!    
! as usual we follow the convention that all distances and coordinates
! in the setup file is in meters, using the coordinate system where
! origo is in the "northeast" corner of model grid box 1,1. the model
! mostly use "grid coordinates" internally, where origo is the south
! west corner of the grid cell 1,1. im,jm is the coordinate of the
! north east corner of grid box im,jm. in this system it is easier to
! map between index space and grid coordinates.
!    
    implicit none
    real, target :: f(i1:,j1:,:)
    character(*) splitline(:)
    integer range(6), rc(6)
    integer linenumbers(400),i, forcing_id,forcing_time(6)
    character(200) forcing_lines(400), forcing_splitline(2)
    logical inblock

    numforcings = numforcings + 1
    if(numforcings > max_numforcings) call message("too many bc's with input from file, increase max_numforcings",2)

    if(master) write(*,'(3x,a,i0)') "Forcing id=", numforcings
    
    forcing(numforcings)%target3d => f
    
    forcing(numforcings)%file_name = trim(data_path)//trim(splitline(5))

    if(index(splitline(1),"east") > 0) forcing(numforcings)%edge=east
    if(index(splitline(1),"west") > 0) forcing(numforcings)%edge=west
    if(index(splitline(1),"north") > 0) forcing(numforcings)%edge=north
    if(index(splitline(1),"south") > 0) forcing(numforcings)%edge=south
    
! this should match u,v,ua,va, we default to p point:
    forcing(numforcings)%pointtype="p"
    if(index(forcing_splitline(1),"bc_u") > 0) forcing(numforcings)%pointtype="u"
    if(index(forcing_splitline(1),"bc_v") > 0) forcing(numforcings)%pointtype="v"

! read the header file from column 5    
    if(master)call get_lines(forcing(numforcings)%file_name, forcing(numforcings)%numfields, forcing_lines, linenumbers)
    if(master.and.debug>1) write(*,*) "  zone_from_file: header file= ",trim(splitline(5)),"  # fields=",forcing(numforcings)%numfields
    call broadcast(forcing(numforcings)%numfields)
    call broadcast(forcing_lines)
    call broadcast(linenumbers)
    allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))

    forcing(numforcings)%name = splitline(1)
    
    do i=1,forcing(numforcings)%numfields
        call split(forcing_lines(i), forcing_splitline)
        
        ! if there is a t in the forcing time string, it must be in iso format
        if(index(forcing_splitline(1),"T") > 0) then
            call get_iso8601(forcing_splitline(1), forcing_time)
            forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
        else
            read(forcing_splitline(1),*,err=1) forcing(numforcings)%field(i)%time
        end if
        
        forcing(numforcings)%field(i)%file = trim(forcing_splitline(2))
    end do

    read(splitline(6:7),*,err=3,end=3) forcing(numforcings)%ioffset, forcing(numforcings)%joffset

    if(master.and.debug>1) then
        write(*,'(3x,a,1x,f0.1,",",f0.1)') "zone_from_file: i,j offset=", forcing(numforcings)%ioffset,forcing(numforcings)%joffset
    end if
    
    forcing(numforcings)%zdatafile = trim(splitline(8))        
    forcing(numforcings)%type = 3  ! forcing from 3d fields

    read(splitline(9:11),*,end=2,err=2) forcing(numforcings)%dim1, forcing(numforcings)%dim2, forcing(numforcings)%dim3
    
    if(master.and.debug>1) then
        write(*,'(3x,a,1x,i0,"x",i0)') "zone_from_file: forcing dimensions=", &
             forcing(numforcings)%dim1, forcing(numforcings)%dim2, forcing(numforcings)%dim3
        write(*,'(3x,a,f0.2,1x,a)') "zone_from_file: <first time> <datafile> = ",&
             forcing(numforcings)%field(1)%time, trim(forcing(numforcings)%field(1)%file)
        write(*,'(a,f0.2,1x,a)') "   zone_from_file: <last time>  <datafile> = ",&
             forcing(numforcings)%field(forcing(numforcings)%numfields)%time, &
             trim(forcing(numforcings)%field(forcing(numforcings)%numfields)%file)
    end if
    
    forcing(numforcings)%range = range
    call globalrange2local(block%ibn, block%jbn, forcing(numforcings)%range, forcing(numforcings)%inblock)

    if(master.and.debug>1) write(*,'(3x,a,6(i0,1x))') "edge_from_file: range=", range
    
    forcing_id = numforcings
    return
    
! error handling    
1   if(master) write(*,*) "forcing file=",forcing(numforcings)%file_name,"line number=",linenumbers(i)
    if(master) write(*,*) "forcing line=",trim(forcing_lines(i))
    call message("could not read real value for time from column 1 at the given line in the forcing file",2)
2   call message("could not read forcing field dimension1 and 2 from col9 and 10",2)
3   call message("could not read forcing field offsets from col6-7",2)
4   call message("could not open data file with z levels",2)
5   call message("could not read data file with z levels",2)
  end subroutine zone_from_file
  

  subroutine edge_from_file(f, splitline, range, forcing_id)
!    
! this is the forcing reader for setting method=13 forcing from
! file. the variable to be forced is a 3d array.    
!     bc_<variable>_<edge> <zone> <type> 14 <header file> <offset> <resolution> <z level file> <dim1> <dim2>    
! <data> is a file with 2 columns, where col1 is time, col2 is the
! file name with unformatted data for the edge.    
! <offset> is the distance in meters from origo to the first grid
! point of the forcing. <resolution> is the distance between the    
! forcing grid points in meters.
!    
! as usual we follow the convention that all distances and coordinates
! in the setup file is in meters, using the coordinate system where
! origo is in the "northeast" corner of model grid box 1,1. the model
! mostly use "grid coordinates" internally, where origo is the south
! west corner of the grid cell 1,1. im,jm is the coordinate of the
! north east corner of grid box im,jm. in this system it is easier to
! map between index space and grid coordinates.
!    
    implicit none
    real, target :: f(i1:,j1:,:)
    character(*) splitline(:)
    integer range(6), rc(6)
    integer linenumbers(400),i, forcing_id,forcing_time(6)
    character(200) forcing_lines(400), forcing_splitline(2)
    logical inblock

    numforcings = numforcings + 1
    if(numforcings > max_numforcings) call message("too many bc's with input from file, increase max_numforcings",2)

    if(master) write(*,'(3x,a,i0)') "Forcing id=", numforcings
    
    forcing(numforcings)%target3d => f
    
    forcing(numforcings)%file_name = trim(data_path)//trim(splitline(5))

    if(index(splitline(1),"east") > 0) forcing(numforcings)%edge=east
    if(index(splitline(1),"west") > 0) forcing(numforcings)%edge=west
    if(index(splitline(1),"north") > 0) forcing(numforcings)%edge=north
    if(index(splitline(1),"south") > 0) forcing(numforcings)%edge=south
    
    if(master)call get_lines(forcing(numforcings)%file_name, forcing(numforcings)%numfields, forcing_lines, linenumbers)
    if(master.and.debug>1) write(*,*) "  edge_from_file: header file= ",trim(splitline(5)),"  # fields=",forcing(numforcings)%numfields
    call broadcast(forcing(numforcings)%numfields)
    call broadcast(forcing_lines)
    call broadcast(linenumbers)
    allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))

    forcing(numforcings)%name = splitline(1)
    
    do i=1,forcing(numforcings)%numfields
        call split(forcing_lines(i), forcing_splitline)
        
        ! if there is a t in the forcing time string, it must be in iso format
        if(index(forcing_splitline(1),"T") > 0) then
            call get_iso8601(forcing_splitline(1), forcing_time)
            forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
        else
            read(forcing_splitline(1),*,err=1) forcing(numforcings)%field(i)%time
        end if
        
        forcing(numforcings)%field(i)%file = trim(forcing_splitline(2))
    end do

    select case(forcing(numforcings)%edge)
    case(east,west)
        read(splitline(6:7),*,err=3,end=3) forcing(numforcings)%yoffset, forcing(numforcings)%dy
        if(master.and.debug>1) then
            write(*,'(3x,a,1x,f0.1,",",f0.1)') "edge_from_file: yoffset, dy=", forcing(numforcings)%yoffset, forcing(numforcings)%dy
        end if
    case(north,south)
        read(splitline(6:7),*,err=3,end=3) forcing(numforcings)%xoffset, forcing(numforcings)%dx
        if(master.and.debug>1) then
            write(*,'(3x,a,1x,f0.1,",",f0.1)') "edge_from_file: xoffset, dx=", forcing(numforcings)%xoffset, forcing(numforcings)%dx
        end if
    end select
    
    forcing(numforcings)%zdatafile = trim(splitline(8))        
    forcing(numforcings)%type = 2  ! forcing from 2d fields

    read(splitline(9:10),*,end=2,err=2) forcing(numforcings)%dim1, forcing(numforcings)%dim2
    
    if(master.and.debug>1) then
        write(*,'(3x,a,1x,i0,"x",i0)') "edge_from_file: forcing dimensions=", forcing(numforcings)%dim1, forcing(numforcings)%dim2
        write(*,'(3x,a,f0.2,1x,a)') "edge_from_file: <first time> <datafile> = ",&
             forcing(numforcings)%field(1)%time, trim(forcing(numforcings)%field(1)%file)
        write(*,'(a,f0.2,1x,a)') "   edge_from_file: <last time>  <datafile> = ",&
             forcing(numforcings)%field(forcing(numforcings)%numfields)%time, &
             trim(forcing(numforcings)%field(forcing(numforcings)%numfields)%file)
    end if
    
    forcing(numforcings)%range = range
    call globalrange2local(block%ibn, block%jbn, forcing(numforcings)%range, forcing(numforcings)%inblock)

    if(master.and.debug>1) write(*,'(3x,a,6(i0,1x))') "edge_from_file: range=", range
    
    forcing_id = numforcings
    return
    
! error handling    
1   if(master) write(*,*) "forcing file=",forcing(numforcings)%file_name,"line number=",linenumbers(i)
    if(master) write(*,*) "forcing line=",trim(forcing_lines(i))
    call message("could not read real value for time from column 1 at the given line in the forcing file",2)
2   call message("could not read forcing field dimension1 and 2 from col9 and 10",2)
3   call message("could not read forcing field offset and resolution from col6 and 7",2)
4   call message("could not open data file with z levels",2)
5   call message("could not read data file with z levels",2)
  end subroutine edge_from_file


  subroutine edge_from_file2d(f, splitline, range, forcing_id)
!    
! this is the forcing reader for setting method=13 forcing from
! file. the variable to be forced is a 2d array.    
!     bc_<variable>_<edge> <zone> <type> 13 <header file> <offset> <resolution> <dim1> 
! <data> is a file with 2 columns, where col1 is time, col2 is the
! file name with unformatted data for the edge.    
! <offset> is the distance in meters from model origo to the first grid
! point of the forcing. <resolution> is the distance between the    
! forcing grid points in meters.
!    
! as usual we follow the convention that all distances and coordinates
! in the setup file is in meters, using the coordinate system where
! origo is in the "northeast" corner of model grid box 1,1. the model
! mostly use "grid coordinates" internally, where origo is the south
! west corner of the grid cell 1,1. im,jm is the coordinate of the
! north east corner of grid box im,jm. in this system it is easier to
! map between index space and grid coordinates.
!    
    implicit none
    real, target :: f(i1:,j1:)
    character(*) splitline(:)
    integer range(6), rc(6)
    integer linenumbers(400),i, forcing_id,forcing_time(6)
    character(200) forcing_lines(400), forcing_splitline(2)
    logical inblock

    numforcings = numforcings + 1
    if(numforcings > max_numforcings) call message("too many bc's with input from file, increase max_numforcings",2)

    if(master) write(*,'(3x,a,i0)') "Forcing id=", numforcings
    
    forcing(numforcings)%target2d => f
    
    forcing(numforcings)%file_name = trim(data_path)//trim(splitline(5))

    if(index(splitline(1),"east") > 0) forcing(numforcings)%edge=east
    if(index(splitline(1),"west") > 0) forcing(numforcings)%edge=west
    if(index(splitline(1),"north") > 0) forcing(numforcings)%edge=north
    if(index(splitline(1),"south") > 0) forcing(numforcings)%edge=south
    
    if(master)call get_lines(forcing(numforcings)%file_name, forcing(numforcings)%numfields, forcing_lines, linenumbers)
    if(master.and.debug>1) write(*,*) "  edge_from_file2d: header file= ",trim(splitline(5)),"  # fields=",forcing(numforcings)%numfields
    call broadcast(forcing(numforcings)%numfields)
    call broadcast(forcing_lines)
    call broadcast(linenumbers)
    allocate(forcing(numforcings)%field(forcing(numforcings)%numfields))

    forcing(numforcings)%name = splitline(1)
    
    do i=1,forcing(numforcings)%numfields
        call split(forcing_lines(i), forcing_splitline)
        
        ! if there is a t in the forcing time string, it must be in iso format
        if(index(forcing_splitline(1),"T") > 0) then
            call get_iso8601(forcing_splitline(1), forcing_time)
            forcing(numforcings)%field(i)%time = timediff(initial_time, forcing_time)
        else
            read(forcing_splitline(1),*,err=1) forcing(numforcings)%field(i)%time
        end if
        
        forcing(numforcings)%field(i)%file = trim(forcing_splitline(2))
    end do

    select case(forcing(numforcings)%edge)
    case(east,west)
        read(splitline(6:7),*,err=3,end=3) forcing(numforcings)%yoffset, forcing(numforcings)%dy
        if(master.and.debug>1) then
            write(*,'(3x,a,1x,f0.1,",",f0.1)') "edge_from_file2d: yoffset, dy=", forcing(numforcings)%yoffset, forcing(numforcings)%dy
        end if
    case(north,south)
        read(splitline(6:7),*,err=3,end=3) forcing(numforcings)%xoffset, forcing(numforcings)%dx
        if(master.and.debug>1) then
            write(*,'(3x,a,1x,f0.1,",",f0.1)') "edge_from_file2d: xoffset, dx=", forcing(numforcings)%xoffset, forcing(numforcings)%dx
        end if
    end select
    
    forcing(numforcings)%type = 14  ! forcing from 2d fields

    read(splitline(8),*,end=2,err=2) forcing(numforcings)%dim1
    
    if(master.and.debug>1) then
        write(*,'(3x,a,1x,i0,"x",i0)') "edge_from_file2d: forcing dimension=", forcing(numforcings)%dim1
        write(*,'(3x,a,f0.2,1x,a)') "edge_from_file2d: <first time> <datafile> = ",&
             forcing(numforcings)%field(1)%time, trim(forcing(numforcings)%field(1)%file)
        write(*,'(a,f0.2,1x,a)') "   edge_from_file2d: <last time>  <datafile> = ",&
             forcing(numforcings)%field(forcing(numforcings)%numfields)%time, &
             trim(forcing(numforcings)%field(forcing(numforcings)%numfields)%file)
    end if
    
    forcing(numforcings)%range = range
    call globalrange2local(block%ibn, block%jbn, forcing(numforcings)%range, forcing(numforcings)%inblock)

    if(master.and.debug>1) write(*,'(3x,a,6(i0,1x))') "edge_from_file2d: range=", range
    
    forcing_id = numforcings
    return
    
! error handling    
1   if(master) write(*,*) "forcing file=",forcing(numforcings)%file_name,"line number=",linenumbers(i)
    if(master) write(*,*) "forcing line=",trim(forcing_lines(i))
    call message("could not read real value for time from column 1 at the given line in the forcing file",2)
2   call message("could not read forcing field dimension from col8",2)
3   call message("could not read forcing field offset and resolution from col6 and 7",2)
  end subroutine edge_from_file2d


  
  subroutine add_rampup(f, kdim, r, splitline)
!    
! parse rampup line, columns 5 and up.    
!   BC_<VARIABLE>_<EDGE> <ZONE> <TYPE> 12 <RAMP UP TIME (S)> <TYPE STATE 1> <DATA STATE 1> <TYPE STATE 2> <DATA STATE 2>
!
    use mod_rampup
    implicit none
    real, target :: f(i1:j2,j1:j2,kdim)
    real(4), allocatable :: tmpf_real4(:,:,:), hcoo(:), vcoo(:), tmp2dr4(:,:)
    integer idim,jdim,kdim,r(6),type1,type2
    character(*) splitline(:)
    logical inblock
    integer type1col,type2col, colz, coldat
    real val
    
    if(num_rampup == max_rampup) call message("max number of ramp-ups reached, increase max_rampup in mod_bound, recompile",2)
    num_rampup = num_rampup + 1

    rampup(num_rampup)%globrange = r
    call globalrange2local(block%ibn, block%jbn, r, inblock)    
    rampup(num_rampup)%range = r
    rampup(num_rampup)%inblock = inblock
    
    read(splitline(5),*,err=1,end=1) rampup(num_rampup)%rampup_time

!    select case(kdim)
!    case(1)
!        if(inblock)then
!            rampup(num_rampup)%f2d => f
!            allocate(rampup(num_rampup)%s1f2d(r(1):r(2),r(3):r(4)), rampup(num_rampup)%s1f2d(r(1):r(2),r(3):r(4)))
!            allocate(rampup(num_rampup)%s2f2d(r(1):r(2),r(3):r(4)), rampup(num_rampup)%s2f2d(r(1):r(2),r(3):r(4)))
!        end if
!    case(kb)
        if(inblock)then
            rampup(num_rampup)%f3d => f
            allocate(rampup(num_rampup)%s1f3d(r(1):r(2),r(3):r(4),r(5):r(6)), &
                 rampup(num_rampup)%s2f3d(r(1):r(2),r(3):r(4),r(5):r(6)))
        end if
!    case default
!        if(master) write(*,*) " kdim=",kdim
!        call message("invalid kdim given to add_rampup",2)
!    end select

! first get state 1. store column of type1 in type1col. type2col will
! depend on the number of arguments for type1. (xml was invented for a reason...)
    type1col = 6
    read(splitline(type1col),*,err=2,end=2) type1
    
    select case(type1)
    case(0)
      ! scalar to be applied to whole region
        type2col = 8                                   ! start reading state 2 in this column
        read(splitline(type1col+1),*,err=3,end=3) val
!        select case(kdim)
!        case(1)
!            if(inblock) rampup(num_rampup)%s1f2d = val
!        case(kb)
            if(inblock) rampup(num_rampup)%s1f3d = val
!        end select
    case(1)
        call message("rampup from 1d z data not finished",1)
        type2col = type1col+4                          ! start reading state 2 in this column
        
!vprof_intpol(f, filename, cols, region, report)
! ... <type> <file> <colz> <coldat>
        read(splitline(type1col+2),*,err=4,end=4) colz
        read(splitline(type1col+3),*,err=5,end=5) coldat       
        allocate(tmpf_real4(i1:i2,j1:j2,kdim))
        call vprof_intpol(tmpf_real4, trim(data_path)//trim(splitline(type1col+1)), (/colz, coldat/), rampup(num_rampup)%globrange)       
        f = tmpf_real4
        deallocate(tmpf_real4)
    case(2)
        call message("rampup from 2d z data not yet done",1)
        
!        <type1> <file> <x file> <z file>
        call readfield(hcoo, splitline(type1col+2))
        call readfield(vcoo, splitline(type1col+3))
        idim = size(hcoo)
        kdim = size(vcoo)
        call readfield(tmp2dr4, splitline(type1col+1))

        if(idim.ne.size(tmp2dr4,1)) call message("dimension 1 of 2d field does not correspond to array for horisontal position",2)
        if(kdim.ne.size(tmp2dr4,2)) call message("dimension 2 of 2d field does not correspond to array for vertical position",2)

        ! interpolate 2d z slice onto 3d sigma
        call zslice2dtosigma3d(tmp2dr4,hcoo,vcoo, rampup(num_rampup)%s1f3d, r)
    case(3)
        call message("rampup from 3d z data not yet implemented",2)
    case default
        call message("unknown data type for ramp up",2)
    end select

! then state 2
    
    read(splitline(type2col),*,err=2,end=2) type2
    
    select case(type2)
    case(0)
       ! scalar to be applied to whole region
        read(splitline(type2col+1),*,err=3,end=3) val
!        select case(kdim)
!        case(1)
!            if(inblock) rampup(num_rampup)%s2f2d = val
!        case(kb)
            if(inblock) rampup(num_rampup)%s2f3d = val
!        end select
    case(1)
        call message("rampup towards 1d z data not finished",1)
!vprof_intpol(f, filename, cols, region, report)
! ... <type> <file> <colz> <coldat>
        read(splitline(type2col+2),*,err=4,end=4) colz
        read(splitline(type2col+3),*,err=5,end=5) coldat
        allocate(tmpf_real4(i1:i2,j1:j2,kdim))
        call vprof_intpol(tmpf_real4, trim(data_path)//trim(splitline(type2col+1)), (/colz, coldat/), rampup(num_rampup)%globrange)
        f = tmpf_real4
        deallocate(tmpf_real4)
    case(2)
        call message("rampup towards 2d z data not finished",1)
        
!        <type2> <file> <x file> <z file>
        call readfield(hcoo, splitline(type2col+2))
        call readfield(vcoo, splitline(type2col+3))
        idim = size(hcoo)
        kdim = size(vcoo)
        call readfield(tmp2dr4, splitline(type2col+1))

        if(idim.ne.size(tmp2dr4,1)) call message("dimension 1 of 2d field does not correspond to array for horisontal position",2)
        if(kdim.ne.size(tmp2dr4,2)) call message("dimension 2 of 2d field does not correspond to array for vertical position",2)
        
        ! interpolate 2d z slice onto 3d sigma
        call zslice2dtosigma3d(tmp2dr4,hcoo,vcoo, rampup(num_rampup)%s2f3d, r)
        
    case(3)
        call message("rampup towards 3d z data not yet implemented",2)
    case default
        call message("unknown data type for ramp up",2)
    end select
    
    return
1   call message("could not read ramp up time from column 5",2)
2   call message("could not read integer ramp up type 1 from column 6",2)
3   call message("could not read scalar value for initial condition column 6",2)
4   call message("could not read integer for column with z-positions from column 8",2)
5   call message("could not read integer for column with z-data from column 9",2)    
  end subroutine add_rampup
  


  logical function frequency_ok(f, current_fs, type)
!! returns true if no numbers closer than 0.0001 to f is found in current_fs, and f not very close to 0.
    real pi
    real f, current_fs(:), tmp_f
    integer i, type

    pi=4.0*atan(1.0)
    select case(type)
    case(0)
        tmp_f=degperhr2radians(f)
    case(1,2)
        tmp_f=periodinhrs2radians(f)
    end select

    frequency_ok = .true.
    do i=1,size(current_fs)
        if(abs(tmp_f).gt.0.00000001 .and. abs(current_fs(i)-tmp_f) < 0.0000001) frequency_ok=.false.
    end do
  end function frequency_ok


  real function degperhr2radians(f)
! given a tidal frequency in degrees per hour f, return radians per sec.
    real f, pi
    pi = 4.0*atan(1.0)
    degperhr2radians = f*pi/(180.*3600.)
  end function degperhr2radians



  real function periodinhrs2radians(f)
    ! given a tidal period in hours f, return "frequency" in radians per
    ! second
    real f, pi

    pi = 4.0*atan(1.0)
    periodinhrs2radians = 2*pi/(f*3600.)
  end function periodinhrs2radians


  subroutine set_tide(edge, edge_maxzone, splitline)
    use adcirc_tide_fac

    implicit none
    character(*) edge, splitline(:)
    integer edge_maxzone
    integer zone, type
    real frequency, etaamp, etaphase, uaamp, uaphase, vaamp, vaphase, pi
    character(len=8) :: const(1)
    real eqarg(1), nodfac(1), tmpfreq
    integer r(4)

    if(.not.numstep_set) call message("numstep must be set before this keyword",2)
    if(.not.ndate_set) call message("ndate must be set before this keyword",2)

    read(splitline(2),*,err=101) zone
    if(edge_maxzone==0) then
        if(master)write(*,*)' edge=',edge,' zone=',zone,' edge num zones=', edge_maxzone
        call message('set_tide: could not find an open boundary on this edge',2)
    end if
    if(zone > edge_maxzone) then
        if(master)write(*,*)' zone=',zone
        call message('set_tide: invalid zone for this edge',2)
    end if

    read(splitline(3),*,err=102) type

    pi = 4.0*atan(1.0)

    select case(type)
    case(0,1)
        read(splitline(4:),*,err=103) frequency, etaamp, etaphase, uaamp, uaphase, vaamp, vaphase

        ! adjust phase and amplitudes with astronomical arguments
        if(tide_fac) then
            if(.not.ndate_set .or. .not.timezone_set) then
               call message('Can not use tide_fac without first setting the initial time (ndate) and timezone!',1)
            else if(type==0) then
                call get_astronomical_arguments(initial_time, numstep, timezone, dt,  &
                     (/frequency/), const, eqarg, nodfac)
                etaphase = etaphase - frequency*timezone
                uaphase = uaphase - frequency*timezone
                vaphase = vaphase - frequency*timezone
            else if(type==1) then
                tmpfreq = 360./frequency
                call get_astronomical_arguments(initial_time, numstep, timezone, dt,  &
                     (/tmpfreq/), const, eqarg, nodfac)
                etaphase = etaphase - tmpfreq*timezone
                uaphase = uaphase - tmpfreq*timezone
                vaphase = vaphase - tmpfreq*timezone
            end if

            if(master.and.debug>0) write(*,*) ' const=',const(1),' NodFac=', nodfac(1),' EqArg=', eqarg(1)
        
            etaamp = etaamp*nodfac(1)
            etaphase = etaphase - eqarg(1)
            uaphase = uaphase - eqarg(1)
            vaphase = vaphase - eqarg(1)
        end if

        select case(edge)
        case('west')
            if(kcon_west(zone) == maxkcon) call message("too many constituents, increase parameter MAXKCON in mod_bound.f90",2)
            tidetype_west(zone) = 0
            if(frequency_ok(frequency, freq_west(zone,1:kcon_west(zone)), type)) then
                kcon_west(zone) = kcon_west(zone) + 1
                if(master.and.debug>0) write(*,*)' frequency=',frequency,' zone=',zone,' edge=',edge

                ! frequency given as degrees per hour
                if(type==0) freq_west(zone,kcon_west(zone)) = degperhr2radians(frequency)

                ! frequency is the period in hours
                if(type==1) freq_west(zone,kcon_west(zone)) = periodinhrs2radians(frequency)

            else
                if(master)write(*,*)' edge=',edge,' zone=',zone,' freq=', frequency
                call message('a constituent of this frequency already set for this zone',2)
            end if

            if(bowest(zone)%inblock)then
                r = bowest(zone)%local_range
                etatamp(r(1):r(2),r(3):r(4),kcon_west(zone)) = etaamp
                etatpha(r(1):r(2),r(3):r(4),kcon_west(zone)) = etaphase*pi/180.
                uatamp (r(1):r(2),r(3):r(4),kcon_west(zone)) = uaamp
                uatpha (r(1):r(2),r(3):r(4),kcon_west(zone)) = uaphase*pi/180.
                vatamp (r(1):r(2),r(3):r(4),kcon_west(zone)) = vaamp
                vatpha (r(1):r(2),r(3):r(4),kcon_west(zone)) = vaphase*pi/180.
            end if
        case('east')
            if(kcon_east(zone) == maxkcon) call message("too many constituents, increase parameter MAXKCON in mod_bound.f90",2)
            tidetype_east(zone) = 0

            if( frequency_ok(frequency, freq_east(zone,1:kcon_east(zone)), type)) then
                kcon_east(zone) = kcon_east(zone) + 1
                if(debug>0.and.master) write(*,*)' frequency=',frequency,' zone=',zone,' edge=',edge
                if(type==0) freq_east(zone,kcon_east(zone)) = degperhr2radians(frequency)
                if(type==1) freq_east(zone,kcon_east(zone)) = periodinhrs2radians(frequency)
            else
                if(master)write(*,*)' edge=',edge,' zone=',zone,' freq=', frequency
                call message('a constituent of this frequency already set for this zone',2)
            end if

            if(boeast(zone)%inblock)then
                r = boeast(zone)%local_range
                etatamp(r(1):r(2),r(3):r(4),kcon_east(zone)) = etaamp
                etatpha(r(1):r(2),r(3):r(4),kcon_east(zone)) = etaphase*pi/180.
                uatamp (r(1):r(2),r(3):r(4),kcon_east(zone)) = uaamp
                uatpha (r(1):r(2),r(3):r(4),kcon_east(zone)) = uaphase*pi/180.
                vatamp (r(1):r(2),r(3):r(4),kcon_east(zone)) = vaamp
                vatpha (r(1):r(2),r(3):r(4),kcon_east(zone)) = vaphase*pi/180.
            end if
        case('south')
            if(kcon_south(zone) == maxkcon) call message("too many constituents, increase parameter MAXKCON in mod_bound.f90",2)
            tidetype_south(zone) = 0

            if( frequency_ok(frequency, freq_south(zone,1:kcon_south(zone)), type)) then
                kcon_south(zone) = kcon_south(zone) + 1
                if(debug>0.and.master) write(*,*)' frequency=',frequency,' zone=',zone,' edge=',edge
                if(type==0) freq_south(zone,kcon_south(zone)) = degperhr2radians(frequency)
                if(type==1) freq_south(zone,kcon_south(zone)) = periodinhrs2radians(frequency)
            else
                if(master)write(*,*)' edge=',edge,' zone=',zone,' freq=', frequency
                call message('a constituent of this frequency already set for this zone',2)
            end if

            if(bosouth(zone)%inblock)then
                r = bosouth(zone)%local_range
                etatamp(r(1):r(2),r(3):r(4),kcon_south(zone)) = etaamp
                etatpha(r(1):r(2),r(3):r(4),kcon_south(zone)) = etaphase*pi/180.
                uatamp (r(1):r(2),r(3):r(4),kcon_south(zone)) = uaamp
                uatpha (r(1):r(2),r(3):r(4),kcon_south(zone)) = uaphase*pi/180.
                vatamp (r(1):r(2),r(3):r(4),kcon_south(zone)) = vaamp
                vatpha (r(1):r(2),r(3):r(4),kcon_south(zone)) = vaphase*pi/180.
            end if
        case('north')
            if(kcon_north(zone) == maxkcon) call message("too many constituents, increase parameter MAXKCON in mod_bound.f90",2)
            tidetype_north(zone) = 0

            if( frequency_ok(frequency, freq_north(zone,1:kcon_north(zone)), type)) then
                kcon_north(zone) = kcon_north(zone) + 1
                if(debug>0.and.master) write(*,*)' frequency=',frequency,' zone=',zone,' edge=',edge
                if(type==0) freq_north(zone,kcon_north(zone)) = degperhr2radians(frequency)
                if(type==1) freq_north(zone,kcon_north(zone)) = periodinhrs2radians(frequency)
            else
                if(master)write(*,*)' edge=',edge,' zone=',zone,' freq=', frequency
                call message('a constituent of this frequency already set for this zone',2)
            end if

            if(bonorth(zone)%inblock)then
                r = bonorth(zone)%local_range
                etatamp(r(1):r(2),r(3):r(4),kcon_north(zone)) = etaamp
                etatpha(r(1):r(2),r(3):r(4),kcon_north(zone)) = etaphase*pi/180.
                uatamp (r(1):r(2),r(3):r(4),kcon_north(zone)) = uaamp
                uatpha (r(1):r(2),r(3):r(4),kcon_north(zone)) = uaphase*pi/180.
                vatamp (r(1):r(2),r(3):r(4),kcon_north(zone)) = vaamp
                vatpha (r(1):r(2),r(3):r(4),kcon_north(zone)) = vaphase*pi/180.
            end if
        case default
            if(master)write(*,*)' edge=',edge
            call message('unknown edge specifier. try one of east,west,south or north',2)
        end select
    case(2)
        call message("not yet implemented zone wise, use keyword tide_from_chart to read binary tidal fields",2)
    case default
        if(master)write(*,*)' type=',type
        call message('invalid init type for tide',2)
    end select

    return
101 call message('set_tide: could not read integer for zone from column 2',2)
102 call message('set_tide: could not read integer for tide init type from column 3',2)
103 call message('set_tide: could not read 7 floats for type=0 init from columns 3 and onward',2)
  end subroutine set_tide


  subroutine set_bintide(splitline)
!    
! set the tidal arrays for all open boundary zones from global 2d chart for a given constituent
!   tide_from_chart <frequencey> <file with data format "stream" of unformatted real4>
!    
    use adcirc_tide_fac
    implicit none
    character(*) splitline(:)
    integer edge_maxzone
    integer zone, type
    real frequency, etaamp, etaphase, uaamp, uaphase, vaamp, vaphase, pi, freq_rad
    character(len=8) :: const(1)
    real eqarg(1), nodfac(1), tmpfreq
    integer r(4)

    if(.not.numstep_set) call message("numstep must be set before this keyword",2)
    if(.not.ndate_set) call message("ndate must be set before this keyword",2)

    pi = 4.0*atan(1.0)
    read(splitline(2),*,err=101) frequency
    if(master.and.debug>0) write(*,*)' set_bintide: frequency=',frequency
    
! in this routine we do the same for all open boundaries, first set the type to 2
    if(kcon == maxkcon) call message("too many constituents, increase parameter MAXKCON in mod_bound.f90",2)
    type = 2    
    tidetype_west(:) = type
    tidetype_east(:) = type
    tidetype_north(:) = type
    tidetype_south(:) = type    
    
! then increment global constituent counter and set frequency for all edges
    if(frequency_ok(frequency, freq(1:kcon), type)) then
        kcon = kcon + 1
        kcon_west(:) = kcon
        kcon_east(:) = kcon
        kcon_north(:) = kcon
        kcon_south(:) = kcon
        
        freq(kcon) = periodinhrs2radians(frequency)
        freq_west(:, kcon) = freq(kcon)
        freq_east(:, kcon) = freq(kcon)
        freq_north(:, kcon) = freq(kcon)
        freq_south(:, kcon) = freq(kcon)
    else
        call message('a constituent of this frequency already set',2)
    end if

! get the binary fields of tidal charts
    call get_binary_tide_fields(trim(splitline(3)), etatamp(:,:,kcon), etatpha(:,:,kcon), &
         uatamp(:,:,kcon), uatpha(:,:,kcon), vatamp(:,:,kcon), vatpha(:,:,kcon))
    
! adjust phase and amplitudes with astronomical arguments
    if(tide_fac) then
        if((.not.ndate_set).or.(.not.timezone_set)) call message('Cannot use tide_fac without first setting ndate and timezone!',2)
            
        tmpfreq = 360./frequency
        
        call get_astronomical_arguments(initial_time, numstep, timezone, dt, (/tmpfreq/), const, eqarg, nodfac)
        
        if(master.and.debug>0) write(*,*) ' const=',const(1),' NodFac=', nodfac(1),' EqArg=', eqarg(1)        
        
        etatpha(:,:,kcon) = etatpha(:,:,kcon) - tmpfreq*timezone
        uatpha(:,:,kcon) = uatpha(:,:,kcon) - tmpfreq*timezone
        vatpha(:,:,kcon) = vatpha(:,:,kcon) - tmpfreq*timezone
        
        etatamp(:,:,kcon) = etatamp(:,:,kcon)*nodfac(1)
        etatpha(:,:,kcon) = etatpha(:,:,kcon) - eqarg(1)
        uatpha(:,:,kcon) = uatpha(:,:,kcon) - eqarg(1)
        vatpha(:,:,kcon) = vatpha(:,:,kcon) - eqarg(1)
    end if
    
! finally, convert to radians
    etatpha(:,:,kcon) = etatpha(:,:,kcon)*pi/180.
    uatpha(:,:,kcon) = uatpha(:,:,kcon)*pi/180.
    vatpha(:,:,kcon) = vatpha(:,:,kcon)*pi/180.
    return
101 call message("could not read frequency from column 2",2)
  end subroutine set_bintide
  

  subroutine get_binary_tide_fields(filename, etatamp, etatphase, uatamp, uatphase, vatamp, vatphase)
!    
! using preinterpolated fields may seem like a waste, but makes it
! possible to carefully inspect the fields for errors.
!    
    implicit none
    character(*) filename
    real, dimension(i1:i2,j1:j2) :: etatamp, etatphase, uatamp, uatphase, vatamp, vatphase
    real(4), allocatable, dimension(:,:) :: f1, f2, f3, f4, f5, f6
    integer i,j
    
    if(master) then
        allocate(f1(im,jm), f2(im,jm), f3(im,jm), f4(im,jm), f5(im,jm), f6(im,jm))
        open(1000012, file=(trim(data_path)//filename), access="stream", form="unformatted", err=101)
        read(1000012, err=102, end=103) ((f1(i,j), f2(i,j), f3(i,j), f4(i,j), f5(i,j), f6(i,j),i=1,im),j=1,jm)
        close(1000012)
    else
        allocate(f1(1,1), f2(1,1), f3(1,1), f4(1,1), f5(1,1), f6(1,1))
    end if
    ! use real to convert f1 which is always real4 to default real, which can be 4 or 8.
    call scatterfield(real(f1), etatamp)
    call scatterfield(real(f2), etatphase)
    call scatterfield(real(f3), uatamp)
    call scatterfield(real(f4), uatphase)
    call scatterfield(real(f5), vatamp)
    call scatterfield(real(f6), vatphase)
    deallocate(f1, f2, f3, f4, f5, f6)
    return
101 call message("could not open stream real4 file with tidal constituent",2)
102 call message("could not read binary real4 stream from file",2)
103 call message("end of file while reading binary real4 file",2)
  end subroutine get_binary_tide_fields
  
  
  subroutine add_sponge(splitline)
! add_sponge <variable> <value> <range>
!
! this routine initializes the background values for the sponge
! viscosities/diffusivities, along with the logical switches needed to
! pick them up. the routine depends on cm,cm2d,ammin,am2dmin,kmmin and
! khmin to have their final values.
!    
    implicit none
    character(*) splitline(:)
    character(4) name
    real reals(4), value
    integer range(4), r(4)
    logical inblock
    
    name = lower(trim(splitline(2)))   
    read(splitline(3),*,err=2,end=2) value
    read(splitline(4:7),*,err=3,end=3) reals
    
    if(.not. any(name == (/"am  ","am2d","km  ","kh  "/))) call message("invalid variable in column 2",2)    
    if(value < 0.0 .or. value > 1e7) call message("value out of range?",2)
    
!    call apply0000convention(reals)
    range = true2globrange(reals)
    if(range(1) == 2) range(1)=1
    if(range(2) == im-1) range(2)=im
    if(range(3) == 2) range(3)=1
    if(range(4) == jm-1) range(4)=jm

    call check_range(range)
    
    r = range
    call globalrange2local(block%ibn, block%jbn, r, inblock)

    if(master.and.debug >0) then
        write(*,'(2x,a,a,a,f0.7,a,4(i0,1x))') "adding sponge region for ", name, " value=", value, " global index range=", range
    end if
    
    select case(name)
    case("am")
        if(.not.associated(sponge_cm)) then
            allocate(sponge_cm(i1:i2,j1:j2))
            if(smagorinsky_sponge_type=="cm")then
                sponge_cm = cm
            else
                sponge_cm = ammin
            end if
        end if
        if(smagorinsky_sponge_type=="cm")then
            cmsponge = .true.
        else
            amsponge = .true.
        end if
        if(inblock) sponge_cm(r(1):r(2),r(3):r(4)) = value
        if(master.and.debug>0) write(*,'(2x,a,l1,a,l1)') "cmsponge=",cmsponge," amsponge=",amsponge
    case("am2d")
        if(.not.associated(sponge_cm2d)) then
            allocate(sponge_cm2d(i1:i2,j1:j2))
            if(smagorinsky_sponge_type=="cm")then
                sponge_cm2d = cm2d
            else
                sponge_cm2d = am2dmin
            end if
        end if
        if(smagorinsky_sponge_type=="cm")then
            cm2dsponge = .true.
        else
            am2dsponge = .true.
        end if
        if(inblock) sponge_cm2d(r(1):r(2),r(3):r(4)) = value
        if(master.and.debug>0) write(*,'(2x,a,l1,a,l1)') "cm2dsponge=",cm2dsponge," am2dsponge=",am2dsponge
    case("km")
        if(.not.associated(sponge_km)) then
            allocate(sponge_km(i1:i2,j1:j2))
            sponge_km = kmmin
        end if
        if(inblock) sponge_km(r(1):r(2),r(3):r(4)) = value
        kmsponge = .true.        
        if(master.and.debug>0) write(*,'(2x,a,l1,a,l1)') "kmsponge=",kmsponge
    case("kh")
        if(.not.associated(sponge_kh)) then
            allocate(sponge_kh(i1:i2,j1:j2))
            sponge_kh = khmin
        end if
        if(inblock) sponge_kh(r(1):r(2),r(3):r(4)) = value
        khsponge = .true.
        if(master.and.debug>0) write(*,'(2x,a,l1,a,l1)') "khsponge=",khsponge
    end select

    if(master.and.debug>0) write(*,'(a)') 
    
    return
2   call message("could not read real for value from column 3",2)
3   call message("could not read 4 reals for the horisontal sponge range",2)   
  end subroutine add_sponge

  
  subroutine smooth_sponges()
    implicit none
    integer k
    if(master.and.debug >0) write(*,'(/,1x,a,i0,/)') "Applying Shapiro filter to sponge arrays, NFILTER=",sponge_shapiro
    if(cm2dsponge .or. am2dsponge) call shapiro(sponge_shapiro, sponge_cm2d, i1,i2,j1,j2,1)
    if(cmsponge .or. amsponge)     call shapiro(sponge_shapiro, sponge_cm, i1,i2,j1,j2,1)
    if(kmsponge)                   call shapiro(sponge_shapiro, sponge_km, i1,i2,j1,j2,1)
    if(khsponge)                   call shapiro(sponge_shapiro, sponge_kh, i1,i2,j1,j2,1)
    if(associated(sponge_kh_c1)) then
        if(master.and.debug >0) write(*,*) "Shapiro filtering CONC1 KH"
        call shapiro(sponge_shapiro, sponge_kh_c1, i1,i2,j1,j2,1)
        do k=1,kb-1
            co(1)%kh(:,:,k) = sponge_kh_c1(:,:)
        end do
    end if
  end subroutine smooth_sponges

  
  subroutine initialize_parallelization()
    implicit none

    logical file_is_there 
    character(200) :: lines(50)=" "
    character(100) :: splitline(10)=" "  
    integer        :: i,j, kw, keywords, srcline(50)   ! the number of keywords in the above lists.
    real tot

    call start_qmpi()

! From this point QMPI_COMM_WORLD is MPI_COMM_WORLD, i.e. the
! global communicator over all grids/processors

    if(master) inquire(file="multibom.dat", exist=file_is_there)

    call broadcast(file_is_there)

    if(file_is_there)then
        call message("multibom.dat found, will try to run BOM on multiple grids")
        if(master) call get_lines("multibom.dat", keywords, lines, srcline)
    else
        grid(1)%init_file = "setupfile2.dat"
        return
    end if

    call broadcast(keywords)
    call broadcast(lines)
    call broadcast(srcline)

! read multibom.dat, and assign processors to grids with the given
! distribution

    numgrids = 0
    do kw=1,keywords
        call split(lines(kw), splitline)

        select case(lower(trim(splitline(1))))          
        case('grid')
!   grid <label> <percentage of processors> [<name of setupfile>]
            numgrids = numgrids +1
            grid(numgrids)%label = trim(splitline(2))
            read(splitline(3),*) grid(numgrids)%workload
            if(grid(numgrids)%workload <= 0) call message("processor workload must be a positive real",2)

            if(trim(splitline(4)) .ne. "") then
                grid(numgrids)%init_file = trim(splitline(4))
            else
                write(grid(numgrids)%init_file,'("setupfile",i0,".dat")') numgrids
            end if

! prepare for nesting of grids, which will happen with keywords ala this
!      case('connect')
!   connect <gridlabel1> <gridlabel2> <connection type>
!      specify how grids should be coupled. one-way, two-way,   
        case default
            call message("unknown keyword",1)
        end select
    end do


    if(numgrids == 0) then
        call message("multibom.dat read, no grids found, now trying the usual setupfile2.dat",1)
        grid(1)%init_file = "setupfile2.dat"
        return
    end if


! normalize workload

    tot = 0
    do i=1,numgrids
        tot = tot + grid(i)%workload
    end do
    do i=1,numgrids
        grid(i)%workload = grid(i)%workload/tot
    end do

! assign ranks to each grid according to workload

    do i=1,numgrids
        if(i==1)then
            grid(i)%rank1 = 0
        else
            grid(i)%rank1 = grid(i-1)%rank2 + 1
        end if
        grid(i)%rank2 = grid(i)%rank1 + max(0,nint(grid(i)%workload*numproc) - 1)
        if(i==numgrids) grid(i)%rank2 = numproc-1
    end do
    do i=1,numgrids
        if(master) write(*,'("grid ",i0,": rank1,rank2=",i4,",",i4," label=",a," fraction=",f0.1," file=",a)') &
             i,grid(i)%rank1,grid(i)%rank2,trim(grid(i)%label),grid(i)%workload*100,trim(grid(i)%init_file)
    end do

! construct a communicator for each grid rank range

    do i=1,numgrids
        call initialize_communicator((/(j,j=grid(i)%rank1,grid(i)%rank2)/), gridcomm(i))
    end do

! on each processor we now reassign myid, numproc and qmpi_comm_world
! to be that of the local grid. the global myid and numproc are
! (still) qmpi_proc_num and qmpi_num_proc.

    do i=1,numgrids
        if(myid >= grid(i)%rank1 .and. myid <= grid(i)%rank2) then
            mygid = i
            myid = gridcomm(i)%myid
            numproc = gridcomm(i)%size
            qmpi_comm_world = gridcomm(i)%comm
            exit
        end if
    end do

    master=.false.
    if(myid==0) master=.true.

! from now on QMPI_COMM_WORLD is a communicator for the grid of the
! current processor, MPI_COMM_WORLD covers all processors.
    if(debug > 2) write(*,*)' myid=',myid,'np=',numproc,'mygid=',mygid  
  end subroutine initialize_parallelization

  
  subroutine replace_yyyy_with_year(string, date)
! search for YYYY from the end of string, replace with year from date
     implicit none
     character(*) string
     integer date(:), pos
     character(4) year

     pos = index(string, 'YYYY', back=.true.)
     
     if(pos > 0 .and. date(1) > 1000 .and. date(1) < 3000) then
         write(year,'(i4)') date(1)
         string = string(1:pos-1) // year // string(pos+4:)
     end if
  end subroutine replace_yyyy_with_year

  
end module setup2
